# Copyright (C) 2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

package CDG::DepPlot;


use strict;
use warnings;
use Carp;
use GD;

require Exporter;
@CDG::DepPlot::ISA = qw(Exporter);

# Export this by default
@CDG::DepPlot::EXPORT = ();

# Export this upon request
@CDG::DepPlot::EXPORT_OK = qw(parse graph $INPUT_TYPE_LOG $INPUT_TYPE_CDA $INPUT_TYPE_XML);

##### CONSTS ###########################################################

# set this to 1 to enable debugging
our $DEBUG = 0;

our $INPUT_TYPE_LOG = 'log';
our $INPUT_TYPE_CDA = 'cda';
our $INPUT_TYPE_XML = 'xml';

##### COMMON ##########################################################

sub error {
	if ($DEBUG) {
		die(shift);
	} else {
		croak(shift);
	} 
}

##### RENDERER ########################################################

###############################################################################
# return the level for a text node depending on the number of its children
sub getLevel {
  my $currentNode = $_[0]; # hash key for node to process
  my $textNodes = $_[1]; # reference to a hash of all text nodes
  if (!@{$textNodes->{$currentNode}{children}}) {
    return 0;
  } else {
    my $level = 0;
    foreach my $child (@{$textNodes->{$currentNode}{children}}) {
      my $newLevel = &getLevel($child, $textNodes);
      $level = $newLevel + 1 if ($newLevel >= $level);
    }
    return $level;
  }
}

###############################################################################
# sort function for text nodes
sub sortNodes {
#  $a =~ /\(([0-9]+)\-[0-9]+\)/;
#  my $aPos = $1;
#  $b =~ /\(([0-9]+)\-[0-9]+\)/;
#  my $bPos = $1;
#
#  return -1 if ($aPos <  $bPos);
#  return  0 if ($aPos == $bPos);
#  return  1 if ($aPos >  $bPos);

  return $a <=> $b;
}

###############################################################################
# returns the key of the highest node of a given %tNodes hash
sub getHighest {
  my (%tNodes) = @_;
  my ($highest) = keys %tNodes; # begin somewhere
  foreach my $node (keys %tNodes) {
    $highest = $node if ($tNodes{$node}{level} > $tNodes{$highest}{level});
  }
  return $highest;
}

###############################################################################
# follow all possible paths from a given node and if a cycle is found, break it

sub fixCycle {
  my $currentNode = shift;
  my $allNodes = shift; # hash of all nodes
  my $history = shift; # list of nodes, already visited
  my $edgesref = shift;
  my %edges = %{$edgesref};
  
  foreach my $child (@{$allNodes->{$currentNode}{children}}) {
    # check if this node is already in the history
    my $inHist = 0;
    foreach my $hist (@{$history}) {
      if ($hist == $child) { # this node was already visited -> this is a cycle
	#&writeDebug("Cycle detected \(node $child\)");
	# remove this child from the children list -> break the cycle
	my @backup = @{$allNodes->{$currentNode}{children}};
	@{$allNodes->{$currentNode}{children}} = ();
	foreach $_ (@backup) {
	  push @{$allNodes->{$currentNode}{children}}, $_ if ($_ ne $child);
	}
	# mark the edge(s) that we removed at this way as 'disabled'
	foreach $_ (keys %edges) {
	  if ($edges{$_}{toPos} eq $currentNode && $edges{$_}{fromPos} eq $child) {
	    $edges{$_}{disabled} = 1;
	    #&writeDebug("disabled edge $_");
	  }
	}
	$inHist = 1;
      } 
    }
    if (! $inHist) { 
      push @{$history}, $child;
      &fixCycle($child, $allNodes, $history,\%edges);
    }
  }
}


###############################################################################
sub graph {

  my $parseResult = shift;	# data to render
  my $imFile = shift;		# Filename to write to
  
  # generate a hash for all words
  # $word{wordpos} -> { word
  #		        level
  #                     @children }

  #my $edges = ${$parseResult}{'Edges'};
  #my $violations = ${$parseResult}{'Violations'};
  
  my %edges = %{$parseResult->{'Edges'}};
  my %violations = %{$parseResult->{'Violations'}};
  
  
  my %tNodes;
  foreach my $edge (keys %edges) {
    my $pos = $edges{$edge}{fromPos};
    $tNodes{$pos} = { word => $edges{$edge}->{fromWord},
	             level => 0,
		     children => [ ]};
  }
  
  # do some modifications to the %edge and %tNodes hash
  foreach my $ce (keys %edges) {
    # collect children
    my $pos = $edges{$ce}{toPos};
    push @{$tNodes{$pos}->{children}}, $edges{$ce}{fromPos} if $pos;
    # store parent
    if ($pos) {
      $tNodes{$edges{$ce}{fromPos}}{parent} = $pos;
    } else {
      $tNodes{$edges{$ce}{fromPos}}{parent} = 'NIL';
    }
    # make NIL the same position as its child
    if ($edges{$ce}{toWord} eq 'NIL') {
      $edges{$ce}{toPos} = $edges{$ce}{fromPos};
    }
  }

  # get level for each text node but first detect and break any cycles which may
  # occur
  foreach my $tNode (keys %tNodes) {
    my @history = (); # needed for tracking cycles
    &fixCycle($tNode, \%tNodes, \@history,\%edges);
    $tNodes{$tNode}{level} = &getLevel($tNode, \%tNodes);
  }

  # raise levels: we have to raise the level of a node if:
  # - he has a sister with a higher level which is on his right (left) and
  #   their parent is on the right (left) side of her.
  # - it is connected to NIL and there is an edge crossing it.
  

  foreach my $cn (keys %tNodes) {
    if ($tNodes{$cn}{parent} ne 'NIL') { # skip nodes connected to nil first
      # get sister with the highest level which is relevant for raising
      my @sisters = @{$tNodes{$tNodes{$cn}{parent}}->{children}};
      # &writeDebug("Found sisters for Node $cn:" . join(', ', @sisters));
      my $hSis = $cn; # we begin with our selves
      foreach my $sis (@sisters) {
	if ($tNodes{$sis}{level} > $tNodes{$hSis}{level}) {
	  if (($cn < $tNodes{$cn}{parent} && $sis < $tNodes{$cn}{parent}) || 
	    ($cn > $tNodes{$cn}{parent} && $sis > $tNodes{$cn}{parent})) { 
	  $hSis = $sis;
	  }
	}
      }
      if ($hSis != $cn) {
	$tNodes{$cn}{level} = $tNodes{$hSis}{level};
	#&writeDebug("raised level of node $cn (" . $tNodes{$cn}{word} . ")"
	#  . " to level " . $tNodes{$hSis}{level});
      }
    } else { # this is a node going to nil
      # first get the level of the highest node in the net
      my $newLevel = $tNodes{getHighest(%tNodes)}{level};
      # then raise the node, if it is necessary
      my $crossing = 0;
      foreach my $ce (keys %edges) {
	my $from = $edges{$ce}{fromPos};
	my $to = $edges{$ce}{toPos};
	if ( ($from < $cn && $cn < $to) ||
	     ($from > $cn && $cn > $to) ) {
	  $crossing = 1;
	  last;
	}
      }
      $tNodes{$cn}{level} = $newLevel if $crossing;
    }
  }

            ######### generate image #########

  my $tFont = gdGiantFont; # used font for text nodes
  my $lFont = gdTinyFont; # used font for labels
  my $tSpacing = 20; # space between text nodes
  my $imMargW = 20; # width of image margin
  my $levelH = 40; # height of each level
  my $gNSize = 5; # size of a graphical node
  my $minViolation = 5.000e-01; # a violation will only be marked if its value
			        # is lower than this
  
#  my $imW = our $imH = 300; # width and heigth of the image
  
  my $imW = $imMargW;
  my $imH = $imMargW;

  # count positions for all text nodes and compute
  # image height and width
  my $highestLevel = 0;
  foreach my $cn (sort sortNodes keys %tNodes) {
    $tNodes{$cn}{lIndex} = $imW;
    $imW += length ($tNodes{$cn}{word}) * $tFont->width;
    $tNodes{$cn}{rIndex} = $imW;
    $imW += $tSpacing;
    $highestLevel = $tNodes{$cn}{level} if ($tNodes{$cn}{level} > $highestLevel);
    $tNodes{$cn}{middle} = $tNodes{$cn}{lIndex} + int(($tNodes{$cn}{rIndex} - 
      $tNodes{$cn}{lIndex}) / 2);
  }
  $imW += $imMargW;
  $imH += ($highestLevel + 1) * $levelH + $imMargW; # we need to add 1 level to have
						    # enough room for root nodes
  
  
  # create a new image
  my $im = new GD::Image($imW, $imH);

  # allocate some colors
  my $white = $im->colorAllocate(255,255,255);
  my $black = $im->colorAllocate(0,0,0);       
  my $red = $im->colorAllocate(255,0,0);      
  my $blue = $im->colorAllocate(0,0,255);
  my $grey = $im->colorAllocate(220,220,220);

  # set colors for specific elements
  my $tColor = $black; # color for text nodes
  my $eColor = $black; # color for edges
  my $labelColor = $blue; # color for edge labels
  my $vColor = $red; # color for violations

  # draw text line
  foreach my $cn (keys %tNodes) {
    $im->string($tFont, $tNodes{$cn}{lIndex}, $imH - $imMargW, $tNodes{$cn}{word}, $tColor);
  }
  
  # draw vertical lines above text nodes if they have children
  foreach my $cn (keys %tNodes) {
    if ($tNodes{$cn}{level} > 0) {
      my $x1 = $tNodes{$cn}{lIndex} + int(($tNodes{$cn}{rIndex} - $tNodes{$cn}{lIndex}) / 2);
      my $y1 = $imH - $tFont->height - $imMargW;
      my $x2 = $x1;
      my $y2 = $y1 - $levelH * $tNodes{$cn}{level};
      $im->line($x1, $y1, $x2, $y2, $grey);
    }
  }

  # draw edges, inner nodes and labels
  foreach my $ce (keys %edges) {
    # we change the color of an edge if there is a constraint
    # violation lower then $minViolation
    #print "Violations for edge $ce: ". Dumper($violations{$ce});
    my $c;
    if ($violations{$ce}) {
      foreach my $cv (@{$violations{$ce}}) {
	$c = $vColor if ($cv->{value} < $minViolation);
      }
    }
    $c = $eColor if !$c;
    my $fromPos = $edges{$ce}{fromPos};
    my $toPos = $edges{$ce}{toPos};
    my $x1 = $tNodes{$fromPos}{middle};
    my $y1 = $imH - $imMargW - $tFont->height - $levelH * $tNodes{$fromPos}{level};
    my $x2 = $tNodes{$toPos}{middle};
    my $y2;
    # raise level of toPos if this edge is going to NIL
    if ($toPos eq $fromPos) {
      $y2 = $imH - $imMargW - $tFont->height - $levelH * ($tNodes{$toPos}{level} + 0.5);
    } else {
      $y2 = $imH - $imMargW - $tFont->height - $levelH * $tNodes{$toPos}{level};
    }	
    # draw node
    $im->arc($x1, $y1, $gNSize, $gNSize, 0, 360, $black);
    $im->fill($x1, $y1, $black);
    $im->arc($x2, $y2, $gNSize, $gNSize, 0, 360, $black);
    $im->fill($x2, $y2, $black);
    # draw line. if this edge was disabled, we draw a grey, dashed line
    if ($edges{$ce}{disabled}) {
      $im->dashedLine($x1, $y1, $x2, $y2, $grey);
    } else {
      $im->line($x1, $y1, $x2, $y2, $c);
    }
    # draw label
    my $xMiddle = $x1 + ($x2 - $x1)/ 2;
    my $yMiddle = $y1 + ($y2 - $y1)/ 2;
    my $label = $edges{$ce}{label};
    my $xLabel; # the upper left corner of the label
    my $yLabel; # ...
    # the position of the label depends on the direction of the edge
    if ($fromPos < $toPos) { # from lower left to upper right
      $xLabel = $xMiddle - ($lFont->width * length($label));
      $yLabel = $yMiddle - $lFont->height;
    } elsif ($fromPos == $toPos) { # vertical edge
      $xLabel = $xMiddle - ($lFont->width * length($label));
      $yLabel = $yMiddle - ($lFont->height / 2);
    } elsif ($fromPos > $toPos) { # from upper left to lower right
      $xLabel = $xMiddle;
      $yLabel = $yMiddle - $lFont->height;
    }  
    # we use an other color if this edge was disabled
    if ($edges{$ce}{disabled}) {
      $im->string($lFont, $xLabel, $yLabel, $label, $grey);
    } else {
      $im->string($lFont, $xLabel, $yLabel, $label, $labelColor);
    }
  }

  ########### write to file and make it available to TWiki ######
  #my $imFile = &TWiki::Func::getPubDir() . "/$web/$topic/parseImage.$parseNum.png";
  #&writeDebug("generated imageFileName: $imFile");
  #my $imageUrl = &TWiki::Func::getPubUrlPath() . "/$web/$topic/parseImage.$parseNum.png";

  # Convert the image to PNG 
  $imFile =~ s/$TWiki::securityFilter//go;
  $imFile =~ /(.*)/;
  $imFile = $1;  # untaint variable
  open(IMAGE , ">$imFile");
  print IMAGE $im->png;
  close IMAGE;
  
  #return $imageUrl;
}

##### PARSER ##########################################################

sub parse {
	my $input = shift;
	my $type = shift;
	
	if (!$type || !($type =~ /log|cda|xml/)) {
		&error("Unknown input type $type\nstopped");
	}
	
	my %parseEdgesViolations;
        my %parseProfileScoreInf = &getProfileScoreInf($input);
	if ($type eq $INPUT_TYPE_LOG) {
            %parseEdgesViolations = &parseCdgLog($input); 
	    return(
	           'Edges' => \%{$parseEdgesViolations{'Edges'}},
	           'Violations' => \%{$parseEdgesViolations{'Violations'}},
	           'rawEdges' => \@{$parseEdgesViolations{'rawEdges'}},
	           'rawViolations' => \@{$parseEdgesViolations{'rawViolations'}},
		   'profile' => $parseProfileScoreInf{'profile'},
		   'score' => $parseProfileScoreInf{'score'}
	           );
	} elsif ($type eq $INPUT_TYPE_XML){
	        %parseEdgesViolations = &parseCdgXml($input); 
	    return(
	           'Edges' => \%{$parseEdgesViolations{'Edges'}},
	           'Violations' => \%{$parseEdgesViolations{'Violations'}},
	           'rawEdges' => \@{$parseEdgesViolations{'rawEdges'}},
	           'rawViolations' => \@{$parseEdgesViolations{'rawViolations'}},
		   'profile' => $parseProfileScoreInf{'profile'},
		   'score' => $parseProfileScoreInf{'score'}
	           );	
	} else {  
		%parseEdgesViolations = &parseCdgCda($input); 
	    return(
	           'Edges' => \%{$parseEdgesViolations{'Edges'}},
	           'Violations' => \%{$parseEdgesViolations{'Violations'}},
	           'rawEdges' => \@{$parseEdgesViolations{'rawEdges'}},
	           'rawViolations' => \@{$parseEdgesViolations{'rawViolations'}},
		   'profile' => $parseProfileScoreInf{'profile'},
		   'score' => $parseProfileScoreInf{'score'}
	           );
	}
}

sub parseCdgLog {
  my $cdgpOut = $_[0];

  # get resulting syntax edges out of cdgp output
  my @rawEdges = grep (/[0-9]{3}\s+SYN:\s.*$/, split ("\n", $cdgpOut));

  #get resulting violation out of cdgp output
  $cdgpOut =~ /Violations: *\n(.*)?$/s;
  # &writeDebug("violations: $1");
  my @rawViolations = ();
  @rawViolations = split("\n", $1) if $1; # there may be no violations at all

  # construct a hash of all edges
  my $levelRegex = qr/([0-9]{3})/;
  my $labelRegex = qr/--([A-Z0-9]*)?-->/;
  my $posRegex = qr/(\(\d+-\d+\))?/;
  my $wordRegex = qr/(.+?)/;

  my %edges;
  my $edgeRegex = '\s*'.$levelRegex . '\s*SYN:\s*' . $wordRegex . $posRegex . $labelRegex . $wordRegex . $posRegex . '\s*$';
  foreach my $rawEdge (@rawEdges) {
    $rawEdge =~ /$edgeRegex/;
    $edges{$1} = { fromWord => $2,
		   fromPos  => $3,
		   label    => $4,
		   toWord   => $5,
		   toPos    => $6};
#    if ($6) {
#      &writeDebug("matched: $2, $3, $4, $5, $6.");
#    } else {
#      &writeDebug("matched: $2, $3, $4, $5, ' ',.");
#    }
  }

  # adjust some fields in the edge hash
  foreach my $ce (keys %edges) {
    # separate word and category
    if (! ($edges{$ce}{fromWord} eq '_')) {
      if ($edges{$ce}{fromWord} =~ /^(.+?)_(.+)$/) {
	$edges{$ce}{fromWord} = $1;
	$edges{$ce}{fromCat} = $2;
      }
    }
    if (! ($edges{$ce}{toWord} eq '_')) {
      if ($edges{$ce}{toWord} =~ /^(.+?)_(.+)$/) {
	$edges{$ce}{toWord} = $1;
	$edges{$ce}{toCat} = $2;
      }
    }

    # only use the second number of the position thing
    $edges{$ce}{fromPos} =~ /\([0-9]+\-([0-9]+)\)/;
    $edges{$ce}{fromPos} = $1;
    if ($edges{$ce}{toPos}) {
      $edges{$ce}{toPos} =~ /\([0-9]+\-([0-9]+)\)/;
      $edges{$ce}{toPos} = $1;
    }
  }

  # construct a hash of violations
  # every key, which is also a key in the %edges hash, points to
  # an array containing two element hashes, describing the violations
  my $violationsRegex = qr/([0-9]{3}) : ([0-9]{3})?(?: : )?([0-9.e+-]+) : (.*)/;
#  our %violations;
  my %violations;
  foreach my $rawViolation (@rawViolations) {
    $rawViolation =~ /$violationsRegex/;
    if ($1 && $2) {
      $violations{$1} = [] if !$violations{$1};
      push @{$violations{$1}}, { value => $3,
			         name  => $4};
      $violations{$2} = [] if !$violations{$2};
      push @{$violations{$2}}, { value => $3,
			         name  => $4};
    }
    if ($1 && !$2) {
     $violations{$1} = [] if !$violations{$1};
     push @{$violations{$1}}, { value => $3,
			        name  => $4};
    }
  }

  
  return (
	'Edges'		=> \%edges,
	'Violations'	=> \%violations,
	'rawEdges'	=> \@rawEdges,
	'rawViolations'	=> \@rawViolations
	);

}

sub parseCdgXml {
  my $cdgpOut = $_[0];

  $cdgpOut =~ /<annotation.*?>.*?(<arc.*?>.*?<\/arc>)\s*<\/annotation>/s;
  my $rawXmlEdgeString=$1;
  
  # produce resulting syntax edges from xml
  my @rawXmlArcs=($rawXmlEdgeString=~m/<arc .*?>.*?<\/arc>/gs);
  
  $cdgpOut =~ /<violations.*?>\s*?((<constraint .*?\/>\s*?)*)\s*<\/violations>/s;
  my $rawXmlViolationsString=$1;
  
  # produce resulting violations from xml
  my @rawXmlViolations = ($rawXmlViolationsString=~m/<constraint .*?\/>/gs);
  
  
  # fill the %edges hash
  my %edges=();
  my $index=0;
  my $level;
  my $depSYN;
  foreach my $rawXmlArcs(@rawXmlArcs){
      #stick one or two zeros at the beginning of the number
       if(2*$index>99){
           $level=2*$index;
       }   
       elsif(2*$index>9){
           $level="0".2*$index;
       }
       else {
          $level="00".2*$index;
       }   
       #for xml elements position does matter
       if(!($rawXmlArcs =~ /<arc .*?>\s*(<tag .*?\/>\s*)+(<dep .*?<{0}.*?\/>\s*)+\s*<\/arc>/)){
           &error("not valid Xml\nstopped");
       }
       
       
       
       #for xml attributes position doesn't matter
       $rawXmlArcs =~ /<arc.+?to="(.+?)".*?>/;#only second number really needed
       $edges{$level}{fromPos}=$1;
       $rawXmlArcs =~ /<arc.+?word="(.+?)".*?>/;
       $edges{$level}{fromWord}=$1;
       $rawXmlArcs =~ /<tag.+?cat="(.+?)".*?\/>/;
       $edges{$level}{fromCat}=$1;
       $rawXmlArcs =~ /(<dep.+?level="SYN".*?\/>)/;
       $depSYN = $1;
       $depSYN =~ /<dep.+?label="(.*?)".*?\/>/;
       $edges{$level}{label}=$1;
       $depSYN =~ /<dep.+?modifiee="(.*?)".*?\/>/;
       $edges{$level}{toPos}= $1 eq "0" ? "" : $1;
       
       $index++;
  
  }
  
  
  # fill the remainig fields (toWord, toCat) of every element of the %edges hash
  my $toWordLevel;
  foreach my $key (keys %edges){
  
      if($edges{$key}{toPos}){
          #stick one or two zeros at the beginning of the number
          if(2*($edges{$key}{toPos}-1)>99){
             $toWordLevel = 2*($edges{$key}{toPos}-1);
          }   
          elsif(2*($edges{$key}{toPos}-1)>9){
             $toWordLevel = "0".2*($edges{$key}{toPos}-1);
          }
          else {
            $toWordLevel = "00".2*($edges{$key}{toPos}-1);
          }   
	  
	  $edges{$key}{toWord} = $edges{$toWordLevel}{fromWord};
          $edges{$key}{toCat} = $edges{$toWordLevel}{fromCat};
          
      }   
      else{
      
          $edges{$key}{toWord} = "NIL";
          $edges{$key}{toCat} = "";
          
      } 
  
  }  
  
  
  #fill the @rawEdges array 
  my @rawEdges = @rawXmlArcs;
  
  
  #fill the @rawViolations array
  my @rawViolations = ();
  $index=0;
  my $violationLevel;
  foreach my $rawXmlViolations (@rawXmlViolations){
      $rawXmlViolations=~/node="(.*?)"/;
      #stick one or two zeros at the beginning of the number
       if($1>99){
           $violationLevel=$1;
       }   
       elsif($1>9){
           $violationLevel="0".$1;
       }
       else {
          $violationLevel="00".$1;
       } 
      
      $rawViolations[$index]=" ".$violationLevel." :";
      
      
      
      if($rawXmlViolations=~/node2="(.*?)"/){
          if($1>99){
             $violationLevel=$1;
          }   
          elsif($1>9){
             $violationLevel="0".$1;
          }
          else {
             $violationLevel="00".$1;
          }
          $rawViolations[$index].=" ".$violationLevel." :";
       }
       $rawXmlViolations=~/penalty="(.*?)"/;
       $rawViolations[$index].=" ".$1." :";
      
       $rawXmlViolations=~/id="(.*?)"/;
       $rawViolations[$index].=" ".$1;
       
       $index++;
  }
  
  
  
  
  # construct a hash of violations
  # every key, which is also a key in the %edges hash, points to
  # an array containing two element hashes, describing the violations
  my $violationsRegex = qr/([0-9]{3}) : ([0-9]{3})?(?: : )?([0-9.e+-]+) : (.*)/;
#  our %violations;
  my %violations;
  foreach my $rawViolation (@rawViolations) {
    $rawViolation =~ /$violationsRegex/;
    if ($1 && $2) {
      $violations{$1} = [] if !$violations{$1};
      push @{$violations{$1}}, { value => $3,
			         name  => $4};
      $violations{$2} = [] if !$violations{$2};
      push @{$violations{$2}}, { value => $3,
			         name  => $4};
    }
    if ($1 && !$2) {
     $violations{$1} = [] if !$violations{$1};
     push @{$violations{$1}}, { value => $3,
			        name  => $4};
    }
  }
  
  @rawViolations = @rawXmlViolations;
  
  return (
	'Edges'		=> \%edges,
	'Violations'	=> \%violations,
	'rawEdges'	=> \@rawEdges,
	'rawViolations'	=> \@rawViolations
	);

}


sub parseCdgCda {
  my $cdgpOut = $_[0];

  $cdgpOut =~ /(wordgraph0 : wordgraph0 : <->.*REF->->0)/s;
  my $rawCdaEdgeString=$1;
  
  # produce resulting syntax edges from cda
  my @rawCdaArcs=($rawCdaEdgeString=~m/[0-9]{3} [0-9]{3}.*?:.*?REF->->0/gs);
  
  
  #$cdgpOut =~ /<violations.*?>\s*?((<constraint .*?\/>\s*?)*)\s*<\/violations>/s;
  #my $rawXmlViolationsString=$1;
  
  # produce resulting violations from xml
  #my @rawXmlViolations=($rawXmlViolationsString=~m/<constraint .*?\/>/gs);
  
  
  # fill the %edges hash
  my %edges=();
  my $index;
  my $level;
  my $depSYN;
  my $fromWord;
  foreach my $rawCdaArcs(@rawCdaArcs){
      #get the number out of which the level will be computed
      $rawCdaArcs =~ /([0-9]{3}) [0-9]{3} (.*?):\s*/;
      $index = $1;
      $fromWord = $2;
      $index =~/^(0)*([0-9]+?)$/;
      $index = $2;
      #stick one or two zeros at the beginning of the number
       if(2*$index>99){
           $level=2*$index;
       }   
       elsif(2*$index>9){
           $level="0".2*$index;
       }
       else {
          $level="00".2*$index;
       }   
       
       $edges{$level}{fromPos}=$index+1;
       $edges{$level}{fromWord}=$fromWord;
       $rawCdaArcs =~ /cat\/(.*?) *\n/;
       $edges{$level}{fromCat}=$1;
       $rawCdaArcs =~ /SYN->(.*?)->([0-9]+)\s*/;
       $edges{$level}{label}=$1;
       $edges{$level}{toPos}= $2 eq "0" ? "" : $2;
  
  }
  
  
  # fill the remainig fields (toWord, toCat) of every element of the %edges hash
  my $toWordLevel;
  foreach my $key (keys %edges){
  
      if($edges{$key}{toPos}){
          #stick one or two zeros at the beginning of the number
          if(2*($edges{$key}{toPos}-1)>99){
             $toWordLevel = 2*($edges{$key}{toPos}-1);
          }   
          elsif(2*($edges{$key}{toPos}-1)>9){
             $toWordLevel = "0".2*($edges{$key}{toPos}-1);
          }
          else {
            $toWordLevel = "00".2*($edges{$key}{toPos}-1);
          }   
	  
	  $edges{$key}{toWord} = $edges{$toWordLevel}{fromWord};
          $edges{$key}{toCat} = $edges{$toWordLevel}{fromCat};
          
      }   
      else{
      
          $edges{$key}{toWord} = "NIL";
          $edges{$key}{toCat} = "";
          
      } 
  
  }  
  
  
  #fill the @rawEdges array 
  my @rawEdges = @rawCdaArcs;  
  
  #contruct an empty @rawViolations array
  my @rawViolations=();
  
  # construct an empty hash of violations
  my %violations=();
  
  return (
	'Edges'		=> \%edges,
	'Violations'	=> \%violations,
	'rawEdges'	=> \@rawEdges,
	'rawViolations'	=> \@rawViolations
	);

}

#helper method to be used from other methods to get additional information out of the cdp return 
sub getProfileScoreInf{

  # now we get some additional information out of the cdp return
  my $cdgpOut = shift;
  
  # get profile information
  my $profile = -1;
  $cdgpOut =~ /PROFILE: newnet took ([0-9]+)ms/o; # profile for newnet
  my $newnetPro = $1;
  if (!$newnetPro) {
	&error("no network information found\nstopping");
  }

#  if ($solver =~ /frobbing/) { # profile for solver 'frobbing'
#  } elsif ($solver eq 'gls') { # profile for solver 'gls'
#  }
  
  # solver autodetection
  my $solverPro;
  if ($cdgpOut =~ /PROFILE: ([0-9.]+)s spent frobbing/o) {
    $solverPro = $1 * 1000; # to ms
  } elsif ($cdgpOut =~ /total\s+:\s*([0-9.]+) ms.*cycles.*repairs.*penalties/o) {
    $solverPro = $1;
  }
  if (!$solverPro) {
	&error("no solver information found\nstopping");
  }
  $profile = $newnetPro + $solverPro if ($newnetPro >= 0 && $solverPro >= 0);
  
  # get total score of this parse tree
  my $totalScore = -1;
  if($cdgpOut =~ /score\s+:\s*([0-9.e+-]+)/o){
      $totalScore = $1 if $1;
  }
  else{
      $cdgpOut =~ /solution\s+:\s*.*?ms.*?cycles.*?%\s*([0-9.e+-]+) score/o;
      $totalScore = $1 if $1;
  }
  return('profile'	=> $profile,
	'score'		=> $totalScore);

}

1;
