#!/usr/bin/env perl

# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# cdg/utils/extract-grammar.pl


#
# This program tries to deduce constraints from a corpus of annotated
# sentences by computing the number of times that certain features
# (label, direction, category etc.) coincide. It is extremely
# incomplete yet.
use warnings;
use strict;
BEGIN {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0); 
  require 'cdgsetup.pl';
}
use CDG;
use CDG::Utils qw( :functions);

##############################################################################
# loads the annotations
sub readAnnos {
  commandEval "set verbosity 0 ";
  commandEval "load @ARGV";
}

##############################################################################
# import annotations and fill global variables stats, labels and cats
#
# returns
# a) stats:
# - LabelCatCat
# - CatLabel
# - LabelCat
# - CatCat
# b) hash of seen dependency labels
# c) hash of seen pos tags;
#
sub createStats {
  my %stats;
  my %labels;
  my %cats;
  
  my $annos = InputStruct_annotations_get($CDG::inputCurrentGrammar);
  my $n = hashSize($annos);
  for (my $l = hashListOfKeys($annos); $l; $l = ListStruct_next_get($l)) {
    my $annoEntry = hashGet($annos, ListStruct_item_get($l));
    my $anno = AnnoEntry2hash $annoEntry;
    my $id = AnnoEntryStruct_id_get $annoEntry;

    for my $edge (@{$anno->{edges}}) {
      my $modifier = $edge->{'@'};
      my $modifiee = $edge->{'^'};
      my $label = $edge->{label};

      # collect all categories and all labels
      $cats{$modifier->{cat}}++;
      $labels{$label}++;

      # collect stats on unary edges

      # cat x label x cat
      if ($modifiee->{cat}) {
	$stats{LabelCatCat}{$label}{$modifier->{cat}}{$modifiee->{cat}}++;
      } else {
	$stats{LabelCatCat}{$label}{$modifier->{cat}}{''}++;
      }

      # cat x label
      $stats{CatLabel}{$modifier->{cat}}{$label}++;

      # label x cat
      $modifiee->{cat} = '' if ! $modifiee->{cat};
      $stats{LabelCat}{$modifiee->{cat}}{$label}++;

      # cat x cat
      $stats{CatCat}{$modifier->{cat}}{$modifiee->{cat}}++;

    }
  }

  return (\%stats, \%labels, \%cats);
}

##############################################################################
# print level declaration
sub writeLevelDecl {
  my $labels = shift;

  print "SYN # ";
  my $isFirst = 1;
  for my $label (sort keys %$labels) {
    print ", " unless $isFirst; $isFirst = 0;
    print format_string($label);
  }
  print ";\n\n";

}


##############################################################################
# print label initialization constraints
# uses LabelCatCat statistics
sub writeLabelDefinitionConstraints {
  my ($stats, $labels, $cats) = @_;

  
  my $isFirstLabel = 1;
  for my $label (sort keys %{$stats->{LabelCatCat}}) {

    if ($isFirstLabel) {
      $isFirstLabel = 0;
    } else {
      print ");\n\n";
    }
    print "{X:SYN} : '" . (($label eq '')?"NoLabel":$label) . 
      " definition' : auto : 0.0 :\n";
    print "X.label = " . format_string($label) . " -> \n";
    
    my $isFirstModifier = 1;
    for my $modifierCat (sort keys %{$stats->{LabelCatCat}{$label}}) {

      if ($isFirstModifier) {
	$isFirstModifier  = 0;
      } else {
	print ") | \n";
      }

      print "  (X\@cat = " . format_string($modifierCat) . " & (";

      my $isFirstModifiee = 1;
      for my $modifieeCat (sort keys %{$stats->{LabelCatCat}{$label}{$modifierCat}}) {
	if ($isFirstModifiee) {
	  $isFirstModifiee = 0;
	} else {
	  print " | "
	}
	if ($modifieeCat eq '') {
	  print "root(X^id)";
	} else {
	  print "X^cat = " . format_string($modifieeCat);
	}
      }
      print ")";
    }
  }
  print ");\n\n";


}

##############################################################################
# creates 
sub createVerbStats {
  my %stats;

  my $annos = InputStruct_annotations_get($CDG::inputCurrentGrammar);
  my $n = hashSize($annos);
  for (my $l = hashListOfKeys($annos); $l; $l = ListStruct_next_get($l)) {
    my $annoEntry = hashGet($annos, ListStruct_item_get($l));
    my $anno = AnnoEntry2hash $annoEntry;
    my $id = AnnoEntryStruct_id_get $annoEntry;

    for my $edge (@{$anno->{edges}}) {
      my $modifier = $edge->{'@'};
      next if $modifier->{cat} !~ /VB/;

      # verb counter
      my $verbId = "$modifier->{word}/$modifier->{cat}";
      $stats{Verbs}{$verbId}++;

      for my $edge2 (@{$anno->{edges}}) {
	my $modifier2 = $edge2->{'@'}; 
	my $modifiee2 = $edge2->{'^'}; 
	my $label = $edge2->{label};

	next if $modifiee2->{no} != $modifier->{no};
	next if $modifier2->{cat} =~ /RB/;

      }
    }
  }

  return %stats;
}

##############################################################################
# uses VerbFrame stats
sub writeVerbStats {
  my (%stats) = @_;
  for my $verb (sort keys %{$stats{VerbFrame}}) {
    print "$verb($stats{Verbs}{$verb}) :\n";
    for my $label (sort keys %{$stats{VerbFrame}{$verb}}) {
      print "  -$label-> ";
      for my $cat (sort keys %{$stats{VerbFrame}{$verb}{$label}}) {
	print "$cat($stats{VerbFrame}{$verb}{$label}{$cat}) ";
      }
      print "\n";
    }
  }
}


##############################################################################
sub main {
  readAnnos();
  my ($stats, $labels, $cats) = createStats();

  #writeLevelDecl($labels);
  writeLabelDefinitionConstraints($stats, $labels, $cats);
  #createVerbStats();
  #writeVerbStats();
}

main();
