# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

# cdg/utils/extract-grammar.pl
#
# author: Michael Daum (see also THANKS and AUTHORS)
# initial revision: 2003-09-24
#
# Plugin for the dependency synthesiser to convert the wsj of the penn treebank.
#
#------------------------------------------------------------------------------

=head1 NAME

PennPlugin - WSJ plugin for DEPSY

=head1 DESCRIPTION

This is a plugin for DEPSY to convert the Wallstreet Journal
Corpus of the Penn Treebank to a dependency treebank.

=head1 TODOs

Problems collection:

- The "more than" Alternations
  
  * s4: (ADVP (NP (QP more/RBR than 30) years) ago)
    within a quantifier phrase
    
  * s88: (NP More/JJR (PP than (NP a few CEOs)))
    putative quantifier phrase (anno error)

  * s135: (ADVP (ADVP (NP (QP about $ 55 million)) more/JJR) (SBAR than (S previously (VP ordered ...))))
    conjunction
    
  * s163: (S Sales ... (VP (ADVP more than) doubled (PP to (NP 30,841 units)) ... ))
    pure adverbial
- The "consider" Alternations
   
- APPOs
   * s1: "Pierre Vinken, (ADJ 61 years old)," should be
         "Pierre Vinken, (APP 61 years old),"
   * same in s3, s68

- Coordinations
  * s29: "(cotton fibers) and (acetate fibers)" vs.
         "(cotton) and (acetat fibers)"
  * s66: "(this ((US sales) and marketing) arm)" vs. 
         "(this (US (sales and marketing) arm))" vs.
         "(this US (sales and marketing) arm)" vs.
  * s67
  * 132
- sentences only separated by ; (semicolon) should be fragments:
  e.g.: s50, s103, s107
- mark second object differently whenever we get two

- Annotation errors/inconsistencies:
  * flat coordination: 
    - s66: "this U.S. sales and marketing arm" 
    - s67: "Mazda's U.S. sales, services, parts and marketing operations"
    - s69: "a sales and marketing executive"
    - s100
  * flat NP: s76: "a guest pianist-comedian Victor Borge" -- "Victor Borge" should be APP
  * flat NP: s66: "(Japanese (auto maker) (Mazda Motor Corp))" 
- -ADV relation
  * s35: "(NP-ADV a fraction)" is an object, isn't it
  * s88: (S...(VP tempts (S (NP-SBJ them) (VP to ...)))) should be
       (S...(VP tempts (NP them) (S (VP to ..)))
  * s92: (NP (ADJP (QP a mere 0.7) %) increase) should be
       (NP (NP   (QP a mere 0.7) %) increase)

- predicative forms: were (ADJP-PRD)
   * s11: Neither ... were -PRD-> aware of any research ...

- stacked PPs (not coordinated PPs): ---> actually this is translated quite ok
  * "from ... to": s15, s96
  * "from ... to ...to": s323
  * "more ... than": s73
  * "at ... in": s463
  * "if not for money, the for pride": s758
  * "from ... through": s825
  * "more from ... than": s1089

=head1 IMPLEMENTATION

=head2 Tiny lexicon

This is needed to distinguish active from passive voice

=cut

 my %tagLexicon=('VB'=>{'gcat'=>'verb',
			'vform'=>'base'},
		 'MD'=>{'gcat'=>'verb',
			'aux'=>'modal'},
		 'VBP'=>{'gcat'=>'verb',
			 'vform'=>'present'},
		 'VBZ'=>{'gcat'=>'verb',
			 'vform'=>'present_3sg'},
		 'VBD'=>{'gcat'=>'verb',
			 'vform'=>'past'},
		 'VBN'=>{'gcat'=>'verb',
			 'vform'=>'past_part'},
		 'VBG'=>{'gcat'=>'verb',
			 'vform'=>'pres_part'},
                 'JJ'=>{'gcat'=>'adj',
			'aform'=>'base'},
                 'JJR'=>{'gcat'=>'adj',
			'aform'=>'comparative'},
                 'JJS'=>{'gcat'=>'adj',
			'aform'=>'superlative'},
                 'RB'=>{'gcat'=>'adv',
			'aform'=>'base'},
                 'RBR'=>{'gcat'=>'adv',
			'aform'=>'comparative'},
                 'RBS'=>{'gcat'=>'adv',
			'aform'=>'superlative'}
);

 my %lexicon=('VBD'=>{'was'=>{'lemma'=>'be'},
		      'were'=>{'lemma'=>'be'},
		      'had'=>{'lemma'=>'have'}},
	      'VBP'=>{'are'=>{'lemma'=>'be'},
		      "'re"=>{'lemma'=>'be'},
		      'am'=>{'lemma'=>'be'},
		      "'m"=>{'lemma'=>'be'}},
	      'VBZ'=>{'is'=>{'lemma'=>'be'},
		      'has'=>{'lemma'=>'have'},
		      "'s"=>{'lemma'=>'be'}});  
	  
=head2 Configuration Tables

=over 10

=item C<%callbacks>

This table defines implementations of used general callbacks in this plugin.

=cut

my $debug = 0; # set this to 1 to be verbose, or to 0 to shut up

%callbacks = ("initialize" => \&pennInitialize,
	      "finalize" => \&pennFinalize,
	      "writeLexiconEntry" => \&pennWriteLexiconEntry,
	      "readSentence" => \&pennReadSentence,
	      "postProcess" => \&pennPostProcess,
	      "getFeatures" => \&pennGetFeatures
);

=item C<%headTable>

This is the head table definition.

format:

   <nodeTag> => ["r|l", "<nodeTag>-<edgeTag>", "<nodeTag>-<edgeTag>", ...]

=cut

if (1) { # set to 0 to use collins head table
%headTable = (
  "ADJP"	=> ["r", '^NNS-', 
			 '^NN-', # nouns before quantifier (fixes at least s92)
			 '^QP-', #
			 '^ADVP-', '^JJ-', '^VBN-', 
			 '^VBG-', '^ADJP-', '^JJR-', '^NP-', '^JJS-', '^DT-', '^FW-', 
			 '^RBR-', '^RBS-', '^SBAR-', '^RB-',],
			 
  "ADVP"	=> ["r", '^RB-', '^RBR-', '^RBS-', '^FW-', '^ADVP-', '^TO-', '^CD-', 
                         '^JJR-', '^JJ-', '^IN-', '^NP-', '^JJS-', '^NN-', ],

  "CDP"		=> ["r", ".*",], # artificial cardinal numbers phrase
  
  "CONJP"	=> ["r", '^CC-', '^RB-', '^IN-', ],

  "CRDP"	=> ["l", '^CC-', '^VP-' ], # artificial coordination node
  "CNP"		=> ["l", '^N-', ], # artificial coordination node
  
  "FRAG"	=> ["l",  # collins searches from right, w/o any priorities
			 "^NP", # lets take at least the first NP
			 ],
  
  "INTJ"	=> ["l"],

  "LST"		=> ["l", '^LS-', '^:-',], # collins searches from right
  
  "NAC"		=> ["l", '^NN-', '^NS-', '^NNP-', '^NNPS-', '^NP-', '^NAAC-', '^EX-', 
		         '^CD-', '^QP-', '^PRP-', '^VBG-', '^JJ-', '^JJS-', 
			 '^JJR-', '^ADJP-', '^FW-', ],
			 
  "NP"		=> ["r", # collins likes POS and JJR to be the head
			 '^(NN|NNP|NNPS|NNS|NX)-', 
			 '^NP-HD', 
			 '^NP-', 
			 '^(\$|ADJP|PRN)-', '^CD-', '^(JJ|JJS|RB|QP)-',], 
			 # see npHeadDetection also

  "NX"		=> ["r", # collins likes POS and JJR to be the head :(
			 '^(NN|NNP|NNPS|NNS|NX)-', 
			 '^NP-HD', '^NP-', 
			 '^(\$|ADJP|PRN)-', '^CD-', '^(JJ|JJS|RB|QP)-',], 
  
  "PP"		=> ["l", '^IN-', '^TO-', '^VBG-', '^VBN-', '^RP-', '^FW-',], 
			 # collins searches from right
  
  "PRN"		=> ["l", ".*"],
  
  "PRT"		=> ["l", '^RP-'], # collins searches from right
  
  "QP"		=> ["r", # collins uses different priorities here
			 '^\$-', # $ 25
			 '^NNS-', '^NN-', 
			 '^CD-', 
			 '^DT-', 
			 '^QP-', 
			 #'^JJR-', this is adverbial only
			 #'^JJS-', this is adverbial only
			 #'^IN-', this is adverbial only
			 #'^JJ-', this is adverbial only
			 #'^RB-', this is adverbial only
			 ],

  "RRC"		=> ["l", '^VP-', 
			 #'^NP-', what's this
			 "ADVP", "ADJP", 
			 #"PP",
			 ], #collins searches from right

  # diff to collins: TO or IN is never a head of an S, it's the finit verb
  "S"		=> ["l", '^VP-', '^S-', '^SBAR-', 
			 '-SBJ', # diff to collins: the subject becomes the head if the verb is missing
			 '^ADJP-', '^UCP-',
			 '^NP-'],

  "SBAR"	=> ["l", #'^WHNP-', '^WHPP-', '^WHADVP-', '^WHADJP-', 
			 '^IN-', # s7: although preliminar findings were reported ...
			 '^DT-', 
			 '^S-', '^SQ-', '^SINV-', '^SBAR-', '^FRAG-', ],
  
  "SBARQ"	=> ["l", '^SQ-', '^S-', '^SINV-', '^SBARQ-', '^FRAG-', ],

  "SINV"	=> ["l", '^VBZ-', '^VBD-', '^VBP-', '^VB-', '^MD-', '^VP-', '^S-', 
			 '^SINV-', '^ADJP-', '^NP-', ],

  "SQ"		=> ["l", '^VBZ-', '^VBD-', '^VBP-', '^VB-', '^MD-', '^VP-', '^SQ-', ],
  
  "UCP"		=> ["l", ".*"], # collins searches from right

  # diff to collins: TO is never a head of a VPm it's the verb
  "VP"		=> ["l", '^VBD', '^VBN', '^MD', '^VBZ', '^VB', '^VBG', 
			 '^VBP', '^TO', '^VP', '^ADJP', '^NN', '^NNS', '^NP'], 

  "WHADJP"	=> ["l", '^CC-', '^WRB-', '^JJ-', '^ADJP-',],

  "WHADVP"	=> ["r", '^CC-', '^WRB-',],

  "WHNP"	=> ["l", '^WDT-', '^WP-', '^WP\$-', '^WADJP-', '^WHPP-', '^WHNP-', ],

  "WHPP"	=> ["l", '^IN-', '^TO-', '^FW-'],
  "CMP1"        => ["l", '^JJR', '^RBR'],
  "X"		=> ["l"],
);
# collins' head table 
} else {
%headTable = (
  "ADJP"	=> ["l", '^NNS-', '^QP-', '^NN-', '^\$-', '^ADVP-', '^JJ-', '^VBN-', 
			 '^VBG-', '^ADJP-', '^JJR-', '^NP-', '^JJS-', '^DT-', '^FW-', 
			 '^RBR-', '^RBS-', '^SBAR-', '^RB-',],
			 
  "ADVP"	=> ["r", '^RB-', '^RBR-', '^RBS-', '^FW-', '^ADVP-', '^TO-', '^CD-', 
                         '^JJR-', '^JJ-', '^IN-', '^NP-', '^JJS-', '^NN-', ],
			 
  "CONJP"	=> ["r", '^CC-', '^RB-', '^IN-', ],
  
  "FRAG"	=> ["r", ".*" ], # why from right
  
  "INTJ"	=> ["l"],

  "LST"		=> ["r", '^LS-', '^:-',], # why from right
  
  "NAC"		=> ["l", '^NN-', '^NS-', '^NNP-', '^NNPS-', '^NP-', '^NAAC-', '^EX-', 
		         '^\$-', '^CD-', '^QP-', '^PRP-', '^VBG-', '^JJ-', '^JJS-', 
			 '^JJR-', '^ADJP-', '^FW-', ],
			 
		    
  "NP"		=> ["r", # if the last word is POS return this TODO write a conversion
			 '^(NN|NNP|NNPS|NNS|NX|POS|JJR)-', 
			 '^NP-', # search this NP _left to right_
			 '^(\$|ADJP|PRN)-', '^CD-', '^(JJ|JJS|RB|QP)-',], 
			 # see npHeadDetection also

  "NX"		=> ["r", # if the last word is POS return this TODO write a conversion
			 '^(NN|NNP|NNPS|NNS|NX|POS|JJR)-', 
			 '^NP-', # search this NP _left to right_
			 '^(\$|ADJP|PRN)-', '^CD-', '^(JJ|JJS|RB|QP)-',], 
  
  "PP"		=> ["r", '^IN-', '^TO-', '^VBG-', '^VBN-', '^RP-', '^FW-',], # why from right
  
  "PRN"		=> ["l", ".*"],
  
  "PRT"		=> ["r", '^RP-'], # why from right
  
  "QP"		=> ["l", '^\$-', '^IN-', '^NNS-', '^NN-', '^JJ-', '^RB-', '^DT-', 
			 '^CD-', '^NCD-', '^QP-', '^JJR-', '^JJS-', ],

  "RRC"		=> ["r", '^VP-', '^NP-', "ADVP", "ADJP", "PP",], # why from right

  "S"		=> ["l", '^TO-', '^IN-', '^VP-', '^S-', '^SBAR-', '^ADJP-', '^UCP-', 
			 '^NP-', ],

  "SBAR"	=> ["l", '^WHNP-', '^WHPP-', '^WHADVP-', '^WHADJP-', '^IN-', '^DT-', 
			 '^S-', '^SQ-', '^SINV-', '^SBAR-', '^FRAG-', ],
  
  "SBARQ"	=> ["l", '^SQ-', '^S-', '^SINV-', '^SBARQ-', '^FRAG-', ],

  "SINV"	=> ["l", '^VBZ-', '^VBD-', '^VBP-', '^VB-', '^MD-', '^VP-', '^S-', 
			 '^SINV-', '^ADJP-', '^NP-', ],

  "SQ"		=> ["l", '^VBZ-', '^VBD-', '^VBP-', '^VB-', '^MD-', '^VP-', '^SQ-', ],
  
  "UCP"		=> ["r", ".*"], # why from right

  "VP"		=> ["l", '^TO-', '^VBD', '^VBN', '^MD', '^VBZ', '^VB', '^VBG', 
			 '^VBP', '^VP', '^ADJP', '^NN', '^NNS', '^NP'], 

  "WHADJP"	=> ["l", '^CC-', '^WRB-', '^JJ-', '^ADJP-',],

  "WHADVP"	=> ["r", '^CC-', '^WRB-',],

  "WHNP"	=> ["l", '^WDT-', '^WP-', '^WP\$-', '^WADJP-', '^WHPP-', '^WHNP-', ],

  "WHPP"	=> ["l", '^IN-', '^TO-', '^FW-'],
  "X"		=> ["l"],
);
}

=item C<@labelTable>

This maps penn treebank to dependency edge labels. 

format:

   1: <dependency label>
   2: <modifier-posTag>-<modifier-edgeTag>
   3: <modifiee-posTag>-<modifiee-edgeTag>
   4: <domiNode-nodeTag>-<domiNode-edgeTag>

 Example:
(NP Legislation (SBAR (WHNP-1) (S (NP-1) (VP to (VP lift (NP ... ))))))

When considering the adjunction of "lift" as head of the SBAR phrase onto
"Legislation" as head of the NP:

the modifier is "lift" with "VB-VP" as pos- and edgeTag
the modifiee is "Legislation" with "NN-NP" as pos- and edgeTag
the domiNode is the SBAR phrase with "SBAR-" as pos and edgeTag
(setEdgeTag sets the edge tag to be the name of the phrase, so
the edge tags of "lift" and "Legislation" are the labels of the containing phrases).

=cut

@labelTable = (
 [ '', '^(,|\.|\'\'|"|``|:)-', 'ROOT-ROOT', '.*'], # punctuation has no dependency
 [ 'RELCLAUSE', ".*", ".*", "^RRC-"], # properly marked relative clauses
 [ 'S', "^(VBN|VBG)-VP", "-NP", ".*"], # relative clauses without object extraction
 [ 'RELCLAUSE', "^(VBD|VBZ|VBP)-VP", "-NP", ".*"], # relative clauses with an extracted object
 [ 'SUBJ', "-WHNP", "^(VBD|VBZ|VBP)", ".*"], # WH sentences
 [ 'S', "^VB-VP", "-NP", '^SBAR-'], # s63: Legislation to lift the debt ceiling is ...
 [ 'POBJ', '^IN-SBAR', 'IN-PP', '.'], # whether users are at risk
 [ 'CONJ', "IN-SBAR", ".*", ".*"], # must preceed S rules
 [ 'CJ', "^VB", "IN-SBAR", ".*"], # must preceed S rules
 [ 'S', ".*", "ROOT-ROOT", ".*"], 
 [ 'S', "^VB", ".*", "S-"], 
 [ 'S', ".*", "-SBAR", "S-"], 
 [ 'S', "^(VB|MD)", "-WH", ".*"], # s115: ... under which it would ...
 [ 'S', "-VP", "-VP", "^(S|SBAR)-"], 
 [ 'PAR', '-VP', '-VP', 'PRN-'],
 [ 'S', '^NN', '^VB', 'S-ADV'],
 [ 'S', '^TO', '.', '^S'],
 [ 'S', '-VP', '-VP', 'ADJP-PRD'], #s121: that got hard to take
 [ 'REL', "-VP", ".*", '^SBAR-'], 

 [ 'OSUBJ', '.*', '^(?:VBG|TO)', '-SBJ'],
 [ 'SUBJ', '.*', '.*', '-SBJ'],
 [ 'POS', '^POS-', '.*', '.*'],

 [ 'PMOD', '^IN', '^NN', '^CRDP'], # fix for s75

 
 [ 'CONJ', '^CC-', '.*', '.*'], # label conjunction
 [ 'CJ', '.*', '.*', '^CRDP-'], # label conjuncts
 [ 'CJ', '.*', '^CC-', '.*'], # label conjuncts

 [ 'PRTCL', ".*", ".*", "^PRT"],
 [ 'INF', "-VP", "^TO", "."],
 [ 'PRT', "^IN---", "^IN-(PP|--)", ".*"],

 [ 'POBJ', "^NN", "^IN---", "^NP-"],
 [ 'POBJ', ".*", "^(IN|TO|VBG)-PP", "NP-"],
 [ 'PMOD', '^(IN|TO|VBG)-PP', '.*', '^PP-'], # must preceed ADJ
 [ 'PMOD', '^IN-PP', '^NN', '^PRN'], # fix for s119

 [ 'MODAL', '^VB-VP', '^MD-VP', 'VP-'],
 [ 'PERFECT', '^VBN-VP', '^VB.?-VP', 'VP-'],
 [ 'PROGRESSIVE', '^VBG-VP', '^VB.?-VP', 'VP-'],

 [ 'ADV', '-NP', '.*', '-TMP'], # Nov 29.

 [ 'DIROBJ', '-NP', '^VB', '.*'],
 [ 'DIROBJ', '^(NN|NNS|NNP|NNPS|\$)', '^VB', '^NP-'], # must preceed ADJ

 [ 'CMP', '^IN', '^(?:JJ|RB)R-', '.'],
 [ 'CMP', '^(?:JJ|RB)R-', '.', '.'],
 [ 'ADV', '.*', '.*', 'ADJP-PRD'], # Nov 29.
 [ 'APMOD', '^NN', '-ADVP', '.*'], # 33 years -APMOD-> ago
 [ 'ADV', '^VBN-', '^VB-VP', '.*'],
 [ 'ADV', "^RB-", "DT-ADJP", ".*"], 
 [ 'ADV', '^RB-', '^(JJ|JJR|JJS)-ADJP', '^RB'], # usually resilient
 [ 'ADV', '.*', '-VP', 'ADJP-'],
 [ 'ADV', "^(RB|RBR|RBS|IN)-", "^(V|IN|RB|MD)", ".*"],
 [ 'ADV', ".*", ".*", "^ADV"], 

 [ 'COMP', '-COMP', '.*', '.*'], # flagged by the compoundHandler
 [ 'COMP', '.*', '-COMP', '.*'], # flagged by the compoundHandler
 [ 'COMP', '^(?:JJ|NNP)', '^JJ-ADJP', '.'],
 [ 'AMOD', '-NP', '^\$', 'NP-ADV'], # $ 27 a share
 [ 'ADV', '.*', '.*', '-ADV'],

 [ 'APMOD', '^NNS', '^(?:JJ|RB)', '.'], # 10 feet deep
 [ 'AMOD', '.*', '.*', 'NP-ADV'],
 [ 'AMOD', '^JJ', '.*', '^JJ'],
 [ 'AMOD', '^VBG-', '-NP', '.*'], # publishing group
 [ 'AMOD', '^VBG-', '^NN.*', '^VBG'],
 [ 'AMOD', '^VBN-', '.*', '^(VBN|ADJP)'], # the expected number
 [ 'AMOD', '^JJ-ADJP', '^NN', '.*'], 
 #[ 'AMOD', '-NP', '-ADJP', '^NP'],
 [ 'AMOD', '^NN', '.*', 'ADJP-'], # an 8 % -ADJ-> increase
 [ 'ADV', '^RB', '^(NN|CD|DT)', '^RB'], # even brief exposures
 [ 'AMOD', '^IN', '^(CD|DT|NN|NNS|\$)' , '^IN'], # wired quantifier thingy
 
 [ 'AMOD', '^NNP-', '^NN-NP', 'NNP---' ], # s8: a Lorillard spokeswoman
 [ 'AMOD', '^NNP-', '^NN-NP', 'NAC-LOC'], # s20: West Groton .... factory

 [ 'DET', '^(DT|PRP\$)-', '.*', '.*'],
 [ 'NUMBER', '^CD', '^CD', '.'],
 [ 'DATE', '^CD-RCD', '^NN', '.'],
 [ 'UNIT', '^CD', '^NN', '.'],
 [ 'QNT', '.*', '.*', '^QP'],
 [ 'QNT', '^CD-', '.*', '.*'],

 [ 'NAME', '^NNP-', '^NNP-NP', '^NNP-'],
 [ 'COMP', '^NNP-', '^NNS-NP', '^NNP-'],
 [ 'COMP', '^NN-', '^(NN|NNS)-NP', '^NN-'],
 [ 'COMP', '^NNP---', '^NNP---', '^NNP---'],
 [ 'COMP', '^NNS---', '^NNS-NP', '^NNS---'], # s58 controls systems
 [ 'COMP', '^NNS---', '^(NN|NNS)-NP', '^CNP---'], # s66 sales and marketing arm
 [ 'COMP', '^NNP---', '^(NN|NNP)-NP', '^(CNP|NAC)---'], # s21 Hollingsworth ... Co.
 [ 'APPOSITION', '^NN-NP', '^(NN|NNP)-NP', '^NP-'],
 [ 'APPOSITION', '^JJS-NP', '-NP', '^NP-'], # s37
 [ 'NAME', '^NNP', '^NNP', '.'],
 [ 'APPOSITION', '-NP', '-NP', 'PRN---'], # s40
 [ 'APPOSITION', '^NNP-NP', '^NNP-NP', '^NP---'], # s58
# [ 'APPOSITION', '-NP', '-NP', 'NP-'],
# [ 'APPOSITION', '^(NN|NNS|NNP|NNPS|\$)-', '^(NN|NNS|NNP|NNPS|\$)-', '.*'],
);

=item C<%phraseConvTable>

phrase tree conversion rules

=cut

%phraseConvTable = (

  ","		=> [\&puncHandler],
  "."		=> [\&puncHandler],
  ":"		=> [\&puncHandler],
  "''"		=> [\&puncHandler],
  "``"		=> [\&puncHandler],
  "\""		=> [\&puncHandler],

  "ADJP"	=> [\&setEdgeTag, \&coordinationHandler,
		    \&adjCompoundHandler],
  "ADVP"	=> [\&setEdgeTag, \&coordinationHandler],
  "CRDP"	=> [\&coordinationHandler],
  "CNP"		=> [\&coordinationHandler],
  "CONJP"	=> [\&coordinationHandler],
  "FRAG"	=> [\&coordinationHandler],
  "INTJ"	=> [\&coordinationHandler],
  "NAC"		=> [\&coordinationHandler],
  "NP"		=> [\&compoundHandler, \&adjCompoundHandler,
		    \&adjpHandler, \&npHeadDetection, \&setEdgeTag,
		    \&npCoordinationHandler],
  "NX"		=> [\&coordinationHandler],
  "PP"		=> [\&setEdgeTag, \&coordinationHandler],
  "PRN"		=> [\&coordinationHandler],
  "PRT"		=> [\&coordinationHandler],
  "QP"		=> [\&qpHandler, \&numberHandler, \&coordinationHandler],
  "RRC"		=> [\&coordinationHandler],
  "S"		=> [\&coordinationHandler,\&sbarFixHandler],
  "SBAR"	=> [\&setEdgeTag, \&coordinationHandler],
  "SBARQ"	=> [\&coordinationHandler],
  "SINV"	=> [\&coordinationHandler],
  "SQ"		=> [\&coordinationHandler],
  "UCP"		=> [\&coordinationHandler],
  "VP"		=> [\&setEdgeTag, \&coordinationHandler],
  "WHADJP"	=> [\&coordinationHandler],
  "WHADVP"	=> [\&setEdgeTag, \&coordinationHandler],
  "WHNP"	=> [\&setEdgeTag, \&coordinationHandler],
  "WHPP"	=> [\&setEdgeTag, \&coordinationHandler],
  "X"		=> [\&coordinationHandler],
);

=item C<%depConvTable>

dependency tree conversion rules

=cut

%depConvTable = (
  "AMOD" => [\&appoHandler],
  "PAR" => [\&parenthesisHandler],
  "POS" => [\&posHandler],
  "COMP" => [\&appoFixHandler],
  "PERFECT" => [\&passiveHandler],
  "DIROBJ" => [\&predHandler],
  "ADV" => [\&predHandler]
);

=item C<%knownCompounds>

this is a collection of known compounds easying the head detection

=back

=cut

my @knownCompounds = (
  ["auto", "maker"],
  ["Mazda", "Motor", "Corp"],
  ["guest", "pianist"],
);



# -----------------------------------------------------------------------------

=head2 Functions

=over 10

=cut

# -----------------------------------------------------------------------------

=item C<appoFixHandler()>

Handle cases that are not expressable by the labelTable only, i.e. things
that are an apposition and not a compound

=cut

sub appoFixHandler {
  my ($edge, $deps, $extradeps) = @_;
  my ($modifier, $label, $modifiee) = @$edge;
  my $debug = 0;

  if ($modifier->{min} > $modifiee->{min}) {
    print "DEBUG: fixing $modifier->{word} --> $modifiee->{word}\n" if $debug;
    $$edge[1] = "APPOSITION";
    return 1;
  }

  return 0;
}

# ------------------------------------------------------------------------------

=item C<passiveHandler()>

changes PERFECT attachments to PASSIVE where there is no "have"

=cut

sub passiveHandler
{
  my ($edge, $deps, $extradeps) = @_;
  my ($modifier, $label, $modifiee) = @$edge;

  if ($modifier->{dep}[1] eq "PERFECT" &&
      $modifier->{dep}[2]->{features}->{lemma} eq "be")
  {
      $modifier->{dep}[1]="PASSIVE";
      1;
  }
  0;
}

# ------------------------------------------------------------------------------

=item C<predHandler()>

changes ADV and DIROBJ attachments to PRD when the main verb is a form of "be"

=cut

sub predHandler
{
  my ($edge, $deps, $extradeps) = @_;
  my ($modifier, $label, $modifiee) = @$edge;

  return 0 unless $modifier->{dep}[2]->{features}->{lemma};

  if ($modifier->{dep}[1] =~ /^(?:ADV|DIROBJ)$/ &&
      $modifier->{dep}[2]->{features}->{lemma} eq "be" &&
      $modifier->{nodeTag} !~ /^RB/)
  {
      #$Data::Dumper::Maxdepth=3;
      #$Data::Dumper::Varname="edge";
      #print Dumper($edge);
      $modifier->{dep}[1]="PRD";
      1;
  }
  0;
}

# -----------------------------------------------------------------------------

=item C<posHandler()>

A posessive nominal attachment is an adjective, not an apposition. 
e.g.: "The scientists' -ADJ-> dictionary."
This is a decision that is fundend by examining two nodes, so the head table
does not work.

=cut

sub posHandler {
  my ($edge, $deps, $extradeps) = @_;
  my ($modifier, $label, $modifiee) = @$edge;

  print "DEBUG: called posHandler\n" if $debug;

  if ($modifiee->{dep}[2]->{nodeTag} =~ /^(NN|NNS|NNP|NNPS|CD|NX)/) {
    $modifiee->{dep}[1] = "GMOD";
    return 1;
  } else {
    return 0;
  }
}


# -----------------------------------------------------------------------------

=item C<appoHandler()>

marks appositional adjective and noun phrases as such.
=cut

sub appoHandler {
    my ($edge, $deps, $extradeps) = @_;
    my ($modifier, $label, $modifiee) = @$edge;
    if ($modifier->{id} > $modifiee->{id})
    {
	$modifier->{dep}[1]="APPOSITION";
	return 1;
    }
    0;
}


# -----------------------------------------------------------------------------

=item C<parenthesisHandler()>

By default, a parenthesis is attached to the head of the matrix constituency.
This conversion shifts it to the first element just before the parenthesis.

=cut

sub parenthesisHandler {
  my ($edge, $deps, $extradeps) = @_;
  my ($modifier, $label, $modifiee) = @$edge;

  print "DEBUG: called parenthesisHandler\n" if $debug;

  # find new modifiee
  my $pos = $modifier->{min} - 2;
  return 0 if $pos < 0;

  my $newModifiee = $rootNode->{leaves}[$pos];

  &moveEdge($edge, $newModifiee);

  return 1;
}

# -----------------------------------------------------------------------------

=item C<npHeadDetection()>

searches from left to right for an NP

=cut
sub npHeadDetection {
  my $node = shift;

  print "DEBUG: called npHeadDetection for $node\n" if $debug;

  # detect two competing NP children
  my $headNP;
  foreach my $child (@{$node->{children}}) {
    if ($child->{nodeTag} eq "NP") {
      if (!$headNP || $headNP->{id} > $child->{id}) {
	$headNP = $child;
      }
    }
  }
  if ($headNP) {
    $headNP->{edgeTag} = "HD";
  }
}

# -----------------------------------------------------------------------------

=item C<setEdgeTag()>

propagates the nodeTag to the edgeTag of the head of a phrase; most of the
time, the edgeTag is empty anyway

=cut

sub setEdgeTag {
  my $node = shift;

  print "DEBUG: called setEdgeTag for $node->{nodeTag}-$node->{edgeTag}\n" if $debug;

  foreach my $child (@{$node->{children}}) {
    if ($child->{isLexHead}) {
      if ($child->{edgeTag} ne '--' && $child->{edgeTag} ne $node->{nodeTag}) {
	print "WARNING: overwriting edgeTag " . 
	  $child->{edgeTag} . " with " . $node->{nodeTag} . "\n";
      }
      $child->{edgeTag} = $node->{nodeTag};
      print "DEBUG: setting $child->{nodeTag}-$child->{edgeTag}\n" if $debug;
      return 1
    }
  }

  return 0;
}

# -----------------------------------------------------------------------------

=item C<compoundHandler()>

bundle well known compounds as such and separate them from the rest of the
constituency

=cut

sub compoundHandler {
  my $node = shift;
  my $debug = 0;

  print "DEBUG: called compoundHandler for $node->{nodeTag}-$node->{edgeTag}\n" if $debug;

  # loop thru all child nodes
  for (my $childIndex = 0; $childIndex < @{$node->{children}}; $childIndex++) {
    my $child = $node->{children}[$childIndex];
    next if ! $child->{word} || $child->{word} eq "NIL" || $child->{edgeTag} eq "COMP";

    print "DEBUG: reading word $child->{word}\n" if $debug;

    # loop thru all known compounds
    my $oldChildIndex = $childIndex;
    COMP: foreach my $compound (@knownCompounds) {
      my $found = 1;
      my @compoundNodes;

      # loop thru all compound items
      ITEM: for (my $compIndex = 0; $compIndex < @$compound; $compIndex++) {
	my $child = $node->{children}[$childIndex];
	my $item = $compound->[$compIndex];
	if (!$child->{word} || $child->{word} eq "NIL" || $child->{edgeTag} eq "COMP" ||
	    $child->{word} !~ /$item/) {
	  $found = 0;
	  last ITEM;
	} else {
	  push @compoundNodes, $child;
	  $childIndex++;
	}
      }

      if ($found) {
	my $newNode = insertInnerNode($rootNode, $node, { nodeTag => "NP", edgeTag => '--' });
	print "DEBUG: found compound " if $debug;
	foreach my $item (@compoundNodes) {
	  print "$item->{word} " if $debug;
	  $item->{edgeTag} = "COMP";
	  moveNode($item, $newNode);
	}
	print "\n" if $debug;
	&compoundHandler($node); # recurse
	return 1;
      } else {
	$childIndex = $oldChildIndex;
      }
    }
  }

  print "DEBUG: done\n" if $debug;
  return 0;
}

# -----------------------------------------------------------------------------

=item C<adjpHandler()>

Handle cases of unannotated adjective phrases like in "(NP a fairly systematic effort)"

=cut

sub adjpHandler {
  my $node = shift;
  my $debug = 0;

  print "DEBUG: called adjpHandler for $node->{nodeTag}-$node->{edgeTag}\n" if $debug;

  # loop thru all child nodes until we find something or fall off the end
  return 0 unless (@{$node->{children}}>2);
  for (my $childIndex = 0; $childIndex < @{$node->{children}}-1; $childIndex++) {
      my $n1=$node->{children}[$childIndex];
      my $n2=$node->{children}[$childIndex+1];
      if ($n1->{nodeTag} =~ /^RB/ &&
	  $n2->{nodeTag} =~ /^JJ/)
      {
	  my $newNode=insertInnerNode($rootNode, $node, { nodeTag => "ADJP", edgeTag => 'NP' });
	  moveNode($n1,$newNode);
	  moveNode($n2,$newNode);
	  $n2->{isLexHead}=1;
	  print "DEBUG: moved $n1->{word} $n2->{word} to a seperate ADJP\n" if $debug;
	  setEdgeTag($node);
	  return 1;
      }
  }

  print "DEBUG: done\n" if $debug;
  return 0;
}

# -----------------------------------------------------------------------------

=item C<adjCompoundHandler()>

for New York-based, North American-made, South African-controlled,
    Old Testament-style compounds that comprise an adjective.

=cut

sub adjCompoundHandler {
  my $node = shift;
  my $debug = 0;

  print "DEBUG: called adjpHandler for $node->{nodeTag}-$node->{edgeTag}\n" if $debug;

  return 0 unless (@{$node->{children}}>2);
  for (my $childIndex = 0; $childIndex < @{$node->{children}}-1; $childIndex++) {
      my $n1=$node->{children}[$childIndex];
      my $n2=$node->{children}[$childIndex+1];
      if ($n1->{nodeTag} =~ /^[NJ]/ &&
	  $n1->{word} =~ /^[A-Z]/ &&
	  $n2->{nodeTag} =~ /^JJ/ &&
	  $n2->{word} =~ /^[A-Z].*-.+/)
      {
	  my $newNode=insertInnerNode($rootNode, $node, { nodeTag => "ADJP", edgeTag => '--' });
	  moveNode($n1,$newNode);
	  moveNode($n2,$newNode);
	  $n2->{isLexHead}=1;
          print "DEBUG: moved $n1->{word} $n2->{word} to a seperate ADJP\n";
	  $n1->{edgeTag}="COMP";
	  $n2->{edgeTag}="COMP";	  
	  return 1;
      }
  }

  print "DEBUG: done\n" if $debug;
  return 0;
}

# -----------------------------------------------------------------------------

=item C<qpHandler()>

Handle quantifier phrases like
(QP more than 3) months

=cut

sub qpHandler {
  my $node = shift;
  my $debug = 0;

  print "DEBUG: called adjpHandler for $node->{nodeTag}-$node->{edgeTag}\n" if $debug;

  # loop thru all child nodes until we find something or fall off the end
  return 0 unless (@{$node->{children}}>2);
  for (my $childIndex = 0; $childIndex < @{$node->{children}}-1; $childIndex++) {
      my $n1=$node->{children}[$childIndex];
      my $n2=$node->{children}[$childIndex+1];
      # handle mistagging of more
      if (($n1->{nodeTag} =~ /^RBR/ || $n1->{word} eq "more") &&
	  $n2->{word} eq "than")
      {
	  my $newNode=insertInnerNode($rootNode, $node, { nodeTag => "CMP1", edgeTag => 'QP' });
	  moveNode($n1,$newNode);
	  moveNode($n2,$newNode);
	  $n1->{isLexHead}=1;
	  print "DEBUG: moved $n1->{word} $n2->{word} to a CMP1 phrase\n" if $debug;
	  setEdgeTag($node);
	  return 1;
      }
  }

  print "DEBUG: done\n" if $debug;
  return 0;
}

# -----------------------------------------------------------------------------

=item C<npCoordinationHandler()>

fix falt coordinated NPs

This is a special version of C<coordinationHandler()> because NPs in the
Penn Treebank are annotated very lazily.

=cut

sub npCoordinationHandler {
  my $node = shift;
  my $debug = 0;

  # see if there's work
  my $found = 0;
  foreach my $child (@{$node->{children}}) {
    if ($child->{nodeTag} =~ /CC|CONJP/) {
      $found = 1;
      last;
    }
  }
  return &coordinationHandler($node) if ! $found;

  # no possibility to disambiguate a coordnation btw
  # (cotton and acetate) fibers and
  # cotton and (acetate fibers) based on the given annotation
  if ($sentenceIndex eq "29") {
    print "DEBUG: no npCoordinationHandler for s29\n" if $debug;
    return &coordinationHandler($node);
  }

  print "DEBUG: called npCoordinationHandler for $node->{nodeTag}-$node->{edgeTag}\n" if $debug;

  # 1. skip pre-modifiers
  # 2. get noun kernel including coordination
  # 3. skip post-modifiers
  my @nounKernel;
  my $state = 0;
  my $nrSkipped = 0;
  my $cat = "CNP";
  my $prevChild;
  foreach my $child (@{$node->{children}}) {

    if ($state == 0) {
      if ($child->{nodeTag} =~ /^(DT|PRP|NP|JJ)/) {
	# skip pre-modifiers
	$nrSkipped++;
	print "DEBUG: skipping pre-modifier $child->{word}/$child->{nodeTag}\n" if $debug;
	$prevChild = $child;
	next;
      }
      if ($child->{nodeTag} =~ /^,/) {
	# found noun kernel boundary but not part of it
	$state = 1;
	$nrSkipped++;
	$cat = "CRDP";
	$prevChild = $child;
	next;
      }
      if ($child->{word} eq "U.S.") {
	# always an apposition, never part of the kernel
	# fixes s66, s67
	$nrSkipped++;
	$prevChild = $child;
	next;
      }
      if ($child->{nodeTag} =~ /^(NN|CC)/) {
	# found noun kernel boundary
	$state = 1;
      }
    }

    
    if ($state == 1) {
      if ($child->{isLexHead} && !(!$prevChild || $prevChild->{nodeTag} =~ /CC|,/)) {
	# found noun kernel boundary
	$state = 2;
      } else {
	# collect noun kernel 
	push @nounKernel, $child;
	print "DEBUG: nounkernel - $child->{word}/$child->{nodeTag}\n" if $debug;
      }
    }

    if ($state == 2) {
      # skip rest
      $nrSkipped++;
    }
    $prevChild = $child;
  }

  return &coordinationHandler($node) if @nounKernel < 2 || $nrSkipped == 0;

  my $newNode = insertInnerNode($rootNode, $node, { nodeTag => $cat, edgeTag => '--' });

  foreach my $child (@nounKernel) {
    moveNode($child, $newNode);
  }


  return 1;
}

# -----------------------------------------------------------------------------

=item C<coordinationHandler()>

build our model of a coordination.

This phraseconversion constructs a binary branching phrase structure for all
members of the coordination, where the first element c-commands a CRDP phrase
covering the rest of the conjuncts, including the conjunction.  The head of a
CRDP is always the leftmost element. Thereby we get the dependency chain that
we'd like to model in D-Space.

=cut

sub coordinationHandler {
  my $node = shift;

  # see, if there's work
  my $found = 0;
  foreach my $child (@{$node->{children}}) {
    if ($child->{nodeTag} =~ /CC|CONJP/) {
      $found = 1;
      last;
    }
  }
  return 0 if ! $found;

  print "DEBUG: called coordinationHandler for $node->{nodeTag}-$node->{edgeTag}\n" if $debug;

  my @otherChildren;
  my $state = 0;
  my $nrSkipped = 0;
  foreach my $child (@{$node->{children}}) {

    # find start CRDP
    if ($state == 0) {

      # found but not part of it
      if ($child->{nodeTag} eq ",") {
	$state = 1;
	next;
      }

      # found but not part of it
      if ($child->{nodeTag} =~ /CC|CONJP/ && $nrSkipped == 0) {
	$state = 1;
	$nrSkipped++;
	next;
      }

      # found and part of it
      if ($child->{nodeTag} =~ /CC|CONJP/) {
	$state = 1;
      } else {
	$nrSkipped++;
	next;
      }
    }

    # collect CRDP
    if ($state == 1) {
      push @otherChildren, $child;
#      if ($child->{min} == $child->{max}) {
#	print "DEBUG: collected $child->{word}\n";
#      } else {
#	print "DEBUG: collected $rootNode->{leaves}[$child->{min}]->{word}..$rootNode->{leaves}[$child->{max}]->{word}\n";
#      }
    }
  }

  return 0 if @otherChildren < 2;# || $nrSkipped == 0;

  
  my $cat = "CRDP";

  # check for an embedded NP (rather dump test ;)
  my $foundNP = 1;
  foreach my $child (@otherChildren) {
    if ($child->{nodeTag} !~ /^(NN|NNS|NNP|NNPS|DT|ADJ|ADV|CD|NX)/) {
      $foundNP = 0;
      last;
    }
  }
  $cat = "NP" if $foundNP;

  my $newNode = insertInnerNode($rootNode, $node, { nodeTag => $cat, edgeTag => '--' });

  foreach my $child (@otherChildren) {
    moveNode($child, $newNode);
  }

  return 1;
}


#------------------------------------------------------------------------------

=item C<puncHandler>

=cut

sub puncHandler {
  my $node = shift;

  print "DEBUG: called puncHandler\n" if $debug;

  &moveNode($node, $rootNode);

  return 1;  
}

# -----------------------------------------------------------------------------

=item C<numberHandler>

This phrase conversion bundles adjacent cardinal numbers as one CDP, a 
cardinal number phrase.

=cut

sub numberHandler {
  my $node = shift;


  print "DEBUG: called numberHandler\n" if $debug;

  my $state = 0;
  my @numbers;
  for (my $i = $node->{min}; $i <= $node->{max}; $i++) {
    my $word = $rootNode->{leaves}[$i];

    # find first CD, skip non-CDs
    if ($state == 0) {
      if ($word->{nodeTag} eq "CD") {
	$state = 1;
      } else {
	next;
      }
    }

    # gather numbers
    if ($state == 1) {
      if ($word->{nodeTag} eq "CD") {
	push @numbers, $word;
      } else {
	$state = 2;
      }
      if ($i == $node->{max}) {
	$state = 2;
      }
    }


    # generate an inner node for all collected numbers
    if ($state == 2) {
      if (@numbers > 1) {
	my $newNode = &insertInnerNode($rootNode, $node, 
	  {"nodeTag" => "CDP", "edgeTag" => "--"});

	foreach my $number (@numbers) {
	  &moveNode($number, $newNode);
	}

	#print "DEBUG: generating new node $newNode->{nodeTag} for:\n";
	#foreach my $number (@numbers) {
	#  print "DEBUG: $number->{word}\n";
	#}
      }
      @numbers = ();
      $state = 0;
    }
  }

  return 1;
}

# -----------------------------------------------------------------------------

=item C<sbarFixHandler>

A (S (S...) (so/IN) (S...)) should be (S (S...) (SBAR so/IN (S...))).
e.g.: s79, 1458, s2136, s5042, s5401 and more

Fixing tagging: 'so' is a subordinating conjunction (IN) in s2136, s3976, s5401

=cut

sub sbarFixHandler {
  my $node = shift;

  print "DEBUG: called sbarFixHandler\n" if $debug;

  # get a (so/...) leaf node under an S node
  my $soNode;
  my $sNode;
  my $state = 0;
  foreach my $child (@{$node->{children}}) {

    # search 'so'
    if ($state == 0) {
      if ($child->{word} && $child->{word} eq "so") {
	print "DEBUG: found 'so' node\n" if $debug;
	$soNode = $child;
	$state = 1;
	next;
      }
    }

    # search adjacent S node
    if ($state == 1) {
      if ($child->{nodeTag} eq "S") {
	print "DEBUG: found S node\n" if $debug;
	$sNode = $child;
      }
      last;
    }
  }

  if (!$soNode || !$sNode) {
    print "DEBUG: bailing out\n" if $debug;
    return 0;
  }

  my $sbarNode = &insertInnerNode($rootNode, $node,
    {"nodeTag" => "SBAR", "edgeTag" => "--"});

  $soNode->{nodeTag} = "IN";
  $soNode->{edgeTag} = "SBAR";

  &moveNode($soNode, $sbarNode);
  &moveNode($sNode, $sbarNode);

  return 1;
}

# -----------------------------------------------------------------------------

=item C<pennReadSentence()>

get rid of some anoying pos tagging errors

- s2: "Dutch" should be JJ (compare to "British")

=cut

sub pennReadSentence {
    my $nodes = shift;
    
    foreach my $node (values %$nodes) {
	if ($node->{word}) {
	    if ($node->{word} eq "Dutch") {
		$node->{nodeTag} = "JJ";
	    }
	}
    }
    
    return 1;
}

# -----------------------------------------------------------------------------

=item C<pennGetFeatures>

handle for the getFeatures callback

=cut

sub pennGetFeatures {
    my ($word, $nodeTag, $morphTag) = @_;
    my %features=();
    my $f;
    $f=$tagLexicon{$nodeTag};
    $features{$_}=$f->{$_} foreach (keys(%$f));
    if ($nodeTag eq "VB")
    {
	$features{lemma}=$word;
    }
    $f=$lexicon{$nodeTag};
    if (ref($f))
    {
	$f=$f->{$word};
	if (ref($f))
	{
	    $features{$_}=$f->{$_} foreach (keys(%$f));
	}
    }
    return \%features;
}

# -----------------------------------------------------------------------------

=item C<pennPostProcess>

=cut

sub pennPostProcess {

  if ($nolabelWarning) {
    my $edges = shift;

    my $found = 0;
    for my $edge (@{$edges}) {
      my ($modifier, $label, $modifiee) = @$edge;
  #    if ((!$label || $label eq '') && $modifiee->{id} >= 0) {
      if ($label eq '--') {
	#print "no label btw $modifier->{word}:$modifier->{id} and $modifiee->{word}:$modifiee->{id}\n";
	$found = 1;
	last;
      }
    }

    if ($found) {
      print STDERR "WARNING: $nrSentences still unset labels present\n";
    } else {
      print STDERR "         $nrSentences - no undefined labels left, ignore previous warning\n";
    }
  }

  return 1;
}

# -----------------------------------------------------------------------------

=item C<pennWriteLexiconEntry>

=cut

sub pennWriteLexiconEntry {

  my $node = shift;

  return 0 if $node->{nodeTag} =~ /^(DT|WDT|PRP|PRP\$|WP|WP\$|CD)$/;
  print "DEBUG: writing lexical entry for $node->{word}/$node->{nodeTag}\n" if $debug;
  return 1;
}

# -----------------------------------------------------------------------------

=item C<initializer()>

plugin initializer

=cut

sub pennInitialize {
}

# -----------------------------------------------------------------------------

=item finalize

plugin finalizer

=cut

sub pennFinalize {
}

=back

=cut

# -----------------------------------------------------------------------------

=head1 AUTHORS

Michael Daum E<lt>micha@nats.informatik.uni-hamburg.deE<gt>

=head1 COPYRIGHT

Copyright (C) 1997-2004 The CDG Team E<lt>cdg@nats.informatik.uni-hamburg.deE<gt>

This file is free software; as a special exception the author gives
unlimited permission to copy and/or distribute it, with or without 
modifications, as long as this notice is preserved.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

=cut

