/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Ingo Schroeder (see also AUTHORS and THANKS for more)
 * Birth: 24/2/97
 *
 * $Id: cdg.c,v 1.93 2006-03-10 14:51:50 foth Exp $
 */

/* -------------------------------------------------------------------------
 * @addtogroup Cdg Cdg - The Root Module
 *
 * Initialization of the \c CDG system on startup.
 *
 * This is the so called root module of the library with the following purpose:
 * - initialization of the \c CDG system at startup 
 * - finalization at shutdown 
 * - container for globally used functions and definitions
 * - owner of the symbol table 
 *
 * All other modules depend on this module and include cdg.h
 * @{
 */

/* ---------------------------------------------------------------------- */
#include <config.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <pwd.h>
#include <readline/readline.h>
#include "cdg.h"
#include "parsing.h"
#include "input.h"
#include "command.h"
#include "statistics.h"
#include "constraintnet.h"
#include "hook.h"
#include "timer.h"
#include "incrementalcompletion.h"
#include "eval.h"
#include "parse.h"
#include "frobbing.h"
#include "write.h"
#include "search.h"
#ifndef DARWIN
#include "compile.h"
#endif
#include "netsearch.h"
#include "set.h"
#include "gls.h"
#include "tagger.h"
#include "chunker.h"
#include "cdgdb.h"
#include "shift.h"

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* -- VARIABLES --------------------------------------------------------- */


/* -------------------------------------------------------------------------
 * indicator flag set by the signal handler of ctrl-c.
 * If \ref cdgCtrlCAllowed is TRUE and the user presses ctrl-c
 * the flag \c cdgCtrlCTrapped is set to TRUE. It's the responsibility
 * of the current function to re-act to the interruption.
 */
Boolean cdgCtrlCTrapped = FALSE;

/* -------------------------------------------------------------------------
 * allow a ctrl-c event.
 * This flag is used to allow the signal handler to indicate a ctrl-c event.
 * If TRUE the signal handler will set \ref cdgCtrlCTrapped, if FALSE it 
 * will not.
 */
Boolean cdgCtrlCAllowed = FALSE;

/* -------------------------------------------------------------------------
 * xcdg indicator flag.
 * If TRUE we're running under XCDG.
 */
Boolean cdgXCDG = FALSE;

/* -------------------------------------------------------------------------
 * time limit in milliseconds used by netsearch, frobbing and gls.
 * The value of \c cdgTimeLimit is normaly used to install a timer alarm 
 * as in \c timerSetAlarm(cdgTimeLimit);
 */
unsigned long cdgTimeLimit = 0;

/* -------------------------------------------------------------------------
   If set, pseudo-umlaut sequences typed into the CDG shell will
   be turned into real 8bit umlauts: "a ==> 
 */
Boolean cdgEncodeUmlauts = FALSE;

/* -------------------------------------------------------------------------
   Program to call on the arguments to inputwordgraph.
 */
String cdgTokenizer = NULL;

/* -------------------------------------------------------------------------
 * user name currently running the \c CDG system.
 * On startup the user name is stored into this place and
 * propagated to newly computed parses.
 */
String cdgUser = NULL;

/* -------------------------------------------------------------------------
   When set to cdgMultiply, penalties are multiplied to yield scores.
   When set to cdgMinimum, scores are computed with min().
*/
AggregationMode cdgAggregation = cdgMultiply;

Hashtable cdgNets;                      /**< all existing ConstraintNets */
Hashtable cdgParses;                    /**< all existing Parses */
Hashtable cdgProblems;                  /**< all existing Problems */
Vector cdgPredictors;                   /**< all defined predictors */

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * wrapper for strDelete
 * This function is our wrapper arround the strDelete() of the blah lib.
 * By default this is a NOP, that is: we don't call strDelete(). You might
 * switch on the call to strDelete() at compile time, if you are brave enuf
 * and wanna help to debug string programming.
 */
void cdgFreeString(String str)
{
#if 0
  strDelete(str);
#endif
}


/* ----------------------------------------------------------------------
 * helper function for agInsert.
 * This function calls agInsert and emits truncation warnings using
 * the cdgPrintf mechanism.
 */
inline void cdgAgInsert(Agenda a, double score, Pointer state)
{
  if (!agInsert(a, score, state)) {
    if (agVerbosity(a)) {
      cdgPrintf(CDG_WARNING,
		"WARNING: truncating agenda, result may be incorrect\n");
    }
  }
}

/* ----------------------------------------------------------------------
 * system initialization.
 * This is the initialization of the root module that initializes all
 * other modules containing an initialization callback. It is absolutely
 * necessary to call this function before starting to use the \c CDG
 * library. In general this is done by the cdgp commandline frontend
 * or by the tcl and perl language extension while loading the bindings.
 * When finishing the \c CDG library the cousin of this function - cdgFinalize() -
 * is to be called.
 */
void cdgInitialize(void)
{
  String userId;
  struct passwd *pwdLine;

  /* initialize the blah library */
  blahInitialize();

  /* force an early readline initialization:
   * this is needed cus readline overwrites our locale settings
   * the first tine readline() is executed in commandLoop()
   */
  rl_initialize();

  /* identify the cdg user */
#ifdef __USE_XOPEN
  userId = cuserid(NULL);
#else
  userId = getenv("USER");
  if (userId == NULL)
    userId = getenv("LOGNAME");
#endif
  
  /* test if we could be identified in the system */
  if (userId != NULL) {
    pwdLine = getpwnam(userId);
    if (pwdLine != NULL) {
      cdgUser = strRegister(pwdLine->pw_gecos);
    }
  } 

  /* default user */
  if (cdgUser == NULL) {
    cdgUser = strRegister("Mac Hine");
  }
  endpwent();

  /* initialize modules */
  hkInitialize();
  inputInitialize();
  evalInitialize();
  icInitialize();
  writeInitialize();
  timerInitialize();
#ifndef DARWIN
  comInitialize();
#endif
  nsInitialize();
  cnInitialize();
  statInitialize();
  lgInitialize();
  scInitialize();
  glsInitialize();
  parseInitialize();
  taggerInitialize();
  frobInitialize();
  chunkerInitialize();
  shiftInitialize();
  dbInitialize();
  
  setRegister("encode-umlauts", SET_BOOL, &cdgEncodeUmlauts, NULL, NULL, NULL, NULL);
  setRegister("timelimit", SET_UNSLONG, &cdgTimeLimit, NULL, NULL, NULL, NULL);
  setRegister("tokenizer", SET_STRING, &cdgTokenizer, NULL, NULL, NULL, NULL);

  setRegister("aggregation", SET_ENUM, &cdgAggregation,
	      NULL, NULL, NULL,
	      "min", cdgMinimum,
	      "mult", cdgMultiply,
              NULL);

  
  cdgNets =
    hashNew( 500, 0.8, hashStringHashFunction, hashStringEqualFunction );
  cdgParses =
    hashNew( 500, 0.8, hashStringHashFunction, hashStringEqualFunction );
  cdgProblems =
    hashNew( 500, 0.8, hashStringHashFunction, hashStringEqualFunction );
  cdgPredictors = vectorNew(16);
  constStr_POS=strRegister("POS");
}

/* ----------------------------------------------------------------------
   Delete all existing constraint nets, parses and problems.

   This should always be used after changing the constraint grammar in any
   way, because there is no guarantee that the previously computed
   structures still make sense with the new grammar defnition. A constraint
   net might contain an LV that used to be allowed, but is now forbidden by
   a unary hard constraint so that it would not have been built at all under
   the new regime. Conversely, if the constraint that forbade building a
   particular LV were turned off, an existing constraint net would become
   incomplete, and solving it could miss a solution that would be possible
   under the new conditions.
   
*/
void cdgDeleteComputed(void) {

  List l, m;

  /* Notify Tcl.

     Tcl keeps annotated versions of core CDG structures, in particular,
     graphical representations of Parses. Since this function deletes
     the core structures, we have to notify Tcl to remove all dangling
     references.
  */
  cdgExecHook(HOOK_RESET);

  if(hashSize(cdgNets) > 0 ||
     hashSize(cdgParses) > 0 ||
     hashSize(cdgProblems) > 0) {
    cdgPrintf(CDG_INFO, "INFO: deleting all constraint nets\n");
  }

  l = hashListOfKeys(cdgNets);
  for(m = l; m != NULL; m = m->next) {
    String id = m->item;
    ConstraintNet net = hashRemove(cdgNets, id);
    cnDelete(net);
  }
  listDelete(m);

  l = hashListOfKeys(cdgParses);
  for(m = l; m != NULL; m = m->next) {
    String id = m->item;
    Parse p = hashRemove(cdgParses, id);
    parseDelete(p);
  }
  listDelete(m);

  l = hashListOfKeys(cdgProblems);
  for(m = l; m != NULL; m = m->next) {
    String id = m->item;
    Problem P = hashRemove(cdgProblems, id);
    deleteProblem(P);
  }
  listDelete(m);

  cnMostRecentlyCreatedNet = NULL;
  parseMostRecentlyCreatedParse = NULL;
  inputMostRecentlyCreatedLattice = NULL;

}


/* ----------------------------------------------------------------------
   Execute all #pragma commands seen during last load operation.

   The #pragma commands are stored in the input as they are seen during
   parsing, but they are not executed until after the load. There are two
   reasons for this:

   1. a #pragma can refer to structures that are loaded from the same
   file as the #pragma -- so executing it immediately would be too early.

   2. if a load operation fails, it should not affect the state of the
   program at all. 
*/
void cdgExecPragmas(List pragmas) {
  List l;
  String command;

  for(l = pragmas; l != NULL; l = l->next) {
    command = l->item;
    commandEval(command);
  }
}


/* ----------------------------------------------------------------------
 * finalize the \c CDG library. 
 * When shutting down the system each module gets the chance to clean
 * up things that might be worth it calling their finalization callback.
 * Be aware that order of finalization might matter. So the hook module
 * gets finalized at last.
 */
void cdgFinalize(void)
{
  /* use this define for memory debugging only: 
   * but be warned that a module which definitely needs finalization
   * should not be covered by this ifdef
   */
//#define SERIOUS_FINALIZATION

  /* finalize modules */
  writeFinalize();
  
#ifdef SERIOUS_FINALIZATION

  /* delete any remaining parsing structures */
  cdgDeleteComputed();

#ifndef DARWIN
  comFinalize();
#endif
  predictFinalize();
  inputFinalize();
  evalFinalize();
  parseFinalize();
  frobbingFinalize();
  icFinalize();
  parseFinalize();
  taggerFinalize();
  setFinalize();
  cdgFreeString(cdgUser);
  shiftFinalize();
#endif

  dbFinalize();
  /* must be last because cdgPrintf() uses hkHooks! */
  hkFinalize();
  blahFinalize();
}

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/** @} */
