/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Ingo Schroeder (see also AUTHORS and THANKS for more)
 * Purpose: header file for some global variables
 *          other globals are declared elsewhere, e.g. constraintnet.h
 * Birth: 13/3/97
 *
 * $Id: cdg.h,v 1.71 2006-03-10 14:51:50 foth Exp $
 */

/* ------------------------------------------------------------------- */
/** \mainpage The CDG Reference Manual

\author Michael Daum, Ingo Schrder, Kilian A. Foth

\section Purpose Purpose of this document  

This document is a thorough description of all data structures and
functions used in the CDG system and serves as a technical documentation of
the parser.

Most of the core functionality is exported by the modules in the directory \c
libcdg, which can be viewed as yet another C library, although a very
specialized one. Application-specific functionality must be implemented by the
application that uses the \c libcdg library. Thus, the command-line parser \c
cdgp includes code to parse its command-line options and to do I/O on a
terminal, and the graphical parser \c xcdg defines extensive routines to
display input structures graphically. This document will eventually cover all
data structures and functions of the constraint parsing library \c libcdg and the
text-based parser \c cdgp.

\c CDG is written entirely in C (and automatically
generated C in some cases). It relies on several non-ANSI extensions
to the C programming language (such as local functions and dynamic
arrays). The GNU C compiler can compile all of these features; others
may or may not.

\section Overview Overview

The \c CDG library is organized into several modules each of which offers a
specialized service to the environment. The set of all modules can be
further divided into those offering the basic infrastructure for \c CDG
parsing whereas others are build on top and offer the different parsing flavours
that are available.

\subsection BasicModules Basic Modules
  - \ref Cdg: This the root layer of the \c CDG library. 
  - \ref Chunker: This module offers an interface for an external chunker.
  - \ref Cdgdb: This module provides access to a berkeley database for retrieving lexical entries.
  - \ref Command: This module implements the \c CDG scripting interface.
  - \ref Compiler: This module implements a compiler in order to translate constraints into C code.
  - \ref Constraintnet: Herein the basic and most central datastructure of \c CDG is implemented, the ConstraintNet.
  - \ref Eval: The Eval module actually implements the constraint interpreter
  - ref Functions: Implementation of all function calls 
  - \ref Hook: The Hook module connects the core of the library to its outside world by a callback mechanism.  
  - ref Input:This module defines the data structures and access functions for constraint grammar.
  - ref Interpreter: This module implements the constraint grammar language it consists of the following submodules
  - ref Levelvalue
  - \ref Lexemgraph
  - ref Parse
  - ref Predicates: Implementation of all predicates
  - \ref Scache
  - \ref Scorematrix: This module implements the score matrix for a constraint edge.
  - ref Set: This module provides a uniform interface for all user-visible variables.
  - ref Statistics
  - ref Tagger
  - ref Testing
  - \ref Timer
  - ref Variables: This module provides a generic interface for variable encapsulation, and is exclusively used by the set module.
  - ref Write


\subsection ParsingFlavours Parsing Flavours
   - ref Arcconsistency
   - ref Frobbing
   - ref MGLS
   - ref Increment
   - ref Incrementalcompletion
   - ref Netsearch
   - ref Nodeconsistency
   - ref Search

\section Guidelines General guidelines

\subsection FileStructure File structure  

Every module consists of a declaration
part \c \<moduleName\>.h and an implementation part \c \<moduleName\>.c . 
Another module can use the exported services of this
modules by including its declarations in \c \<moduleName\>.h .
Note that many of the functions covered in this manual are not exported and
therefore cannot be used from other modules at all. The files
\c skel.c and skel.h provide a skeleton for a new module.

\subsection Terms Terms used in this document

While some of the functions in the \c libcdg library explicitly
use variable-length argument lists, other have a prototype like the following:
\code int no, char **args \endcode 

In this case it is always assumed that
\a args is an array of valid zero-terminated strings, and that
\a no specifies the number of these strings. This is an alternate
way of passing a varying number of additional arguments. The strings
contained in the array are called \em command words in this
document to distinguish them from the actual function arguments.

\subsection CodingStyle Coding style

Most identifiers of data structures and algorithms are complete
English phrases such as `ConstraintNet' or `printLexiconItem'. When a
module deals primarily with one data structure, it is common practice
to abbreviate the name of this structure and use it as a prefix to all
exported identifiers, as in `lvNew', `lvPrint', and `lvDelete' (rather
than `newLevelValue', `printLevelValue', and `deleteLevelValue').

To avoid the explicit use of pointer variables, most modules export
\c typedef 'd aliases for all pointer types. These aliases have
meaningful names, while the underlying \c struct types have names
ending in \c Struct:

\code
typedef struct {
   String id;
   Boolean active;
   int counter;
} SectionStruct;
typedef SectionStruct *Section;
\endcode

Although ANSI C allows the programmer to collapse these two
\c typedef statements into one, this construct cannot be parsed by
the interface generator SWIG, used for building \c xcdg. Therefore it is
important to use exactly this way of defining pointers to structures.

One consequence of this definition style is that although very few
pointer symbols are used in the code, most \c libcdg data types
obey reference semantics: A function called on a variable \c s can
usually change the underlying structure, even though C function calls
actually use value semantics.

A more practical demonstration of the coding style is given in \ref Skel.

*/

#ifndef CDG_H
#define CDG_H

/* -------------------------------------------------------------------------
 * @addtogroup Cdg 
 * @{
 */

/* -- TYPE DEFINITIONS -------------------------------------------------- */
/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>
#include <unistd.h>
#include <stdlib.h>
#include "blah.h"

#ifdef DMALLOC
  #include <dmalloc.h>
#endif


/* -- MACROS ------------------------------------------------------------ */

#ifndef SWIG
/* ----------------------------------------------------------------------
 * maximum of \c a or \c b.
 * Be aware that the expressions \c a and \c b might be evaluated twice
 * depending on wether \c a or \c b are greater.
 */
#define max(a, b) (a) > (b) ? (a) : (b)

/* ----------------------------------------------------------------------
 * minimum of \c a or \c b.
 * Be aware that the expressions \c a and \c b might be evaluated twice
 * depending on wether \c a or \c b are smaller.
 */
#define min(a, b) (a) < (b) ? (a) : (b)
#endif


/* ----------------------------------------------------------------------
 * null, zero, nadda, nothing.
 * \c NULL has the numeric value 0.
 */
#ifndef NULL
#define NULL 0
#endif

/* ----------------------------------------------------------------------
 * this is true.
 * \c TRUE has the numeric value 1.
 */
#ifndef TRUE
#define TRUE 1
#endif

/* ----------------------------------------------------------------------
 * this is false.
 * \c FALSE has the numeric value 0.
 */
#ifndef FALSE
#define FALSE 0
#endif

/* -- FORWARD TYPE DEFINITIONS ------------------------------------------ */

/** @name Forward type definitions
 * This set of declarations help to deal with cirular C-structures which
 * introduce circular header file dependencies as a consequence. All
 * type definitions of that kind are moved to this place here to get
 * a one-stop forward declaration.
 * @{ */

struct ChunkerStruct;
typedef struct ChunkerStruct ChunkerStruct;
  /**< type of a chunker structure */
typedef ChunkerStruct *Chunker; 
  /**< type of a chunker pointer*/

struct ChunkStruct;
typedef struct ChunkStruct ChunkStruct;
  /**< type of a chunk structure */
typedef ChunkStruct *Chunk; 
  /**< type of a chunk pointer */

struct LexemNodeStruct;
typedef struct LexemNodeStruct LexemNodeStruct;
  /**< type of a lexem node structure */
typedef LexemNodeStruct *LexemNode;
  /**< type of a lexem node pointer */

struct GraphemNodeStruct;
typedef struct GraphemNodeStruct GraphemNodeStruct;
  /**< type of a graphem node structure */
typedef GraphemNodeStruct *GraphemNode;
  /**< type of a graphem node pointer */

/** @} */

typedef enum {
  cdgMultiply,
  cdgMinimum
} AggregationMode;



/* -- VARIABLES --------------------------------------------------------- */
extern AggregationMode cdgAggregation;
extern Boolean cdgCtrlCTrapped;
extern Boolean cdgCtrlCAllowed;
extern Boolean cdgXCDG;
extern unsigned long cdgTimeLimit;
extern Boolean cdgEncodeUmlauts;
extern String cdgTokenizer;
extern String cdgUser;
extern Hashtable cdgNets;
extern Hashtable cdgParses;
extern Hashtable cdgProblems;
extern Vector cdgPredictors;

extern String constStr_POS;

/* -- FUNCTIONS --------------------------------------------------------- */
#ifndef SWIG
extern inline void cdgAgInsert(Agenda, double, Pointer);
extern void cdgInitialize( void );
extern void cdgDeleteComputed(void);
extern void cdgFreeString(String str);
#endif
extern void cdgFinalize(void);
extern void cdgExecPragmas(List pragmas);

/* ---------------------------------------------------------------------- */
/** @} */
#endif /* don't insert anything after this #endif */
