/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Birth: 3/3/97 
*/

/* -------------------------------------------------------------------------
  @addtogroup Command Command - The CDG Scripting Language
  @author Ingo Schroeder (see also AUTHORS and THANKS for more)

  $Id: command.c,v 1.331 2006-01-17 15:59:49 foth Exp $
 
  This module implements the \c CDG scripting interface, that is, all the
  commands that are available on the \c cdgp prompt or executable via
  commandEval().

  Note: do not confuse the scripting interface language with the
  constraint grammar language or the language bindings. They are all
  separate things.
 
  Each valid \c command corresponds to a C function \c cmdCommand that
  implements this command all commands are listed in \ref
  interface_commands. Each command is defined by a \ref Command structure,
  where the name, the interface completion commands, a little help
  message for that command etc are defined.

  The following commands exist:
  - activate, cmdActivate(): activates a section
  - annotation, cmdAnnotation(): prints out one or all annotation
  - anno2parse, cmdAnno2Parse(): build a parse from an annotation
    and evaluate it
  - chunk, cmdChunk(): call a chunk parser on a lattice
  - closedb, cmdCloseDB(): closes the database
  - compareparses, cmdCompareParses(): compares the structures of two parses
  - compile, cmdCompile(): compiles the current grammar
  - constraint, cmdConstraint(): prints out one or all constraints
  - deactivate, cmdDeactivate(): deactivates a section
  - distance, cmdDistance(): prints the distance matrix of a net
  - edges, cmdEdges(): prints out one or all edges of a constraint net
  - frobbing, cmdFrobbing(): transforms an arbitrary assignment into a solution
  - gls, cmdGls(): solve a constraintnet using guided local search
  - help, cmdHelp(): gives a short help
  - hierarchy, cmdHierarchy(): prints out one or all hierarchies
  - hook, cmdHook(): toggle hook-system on or off"
  - incrementalcompletion, cmdIncrementalCompletion():
    incremental completion of parses
  - inputwordgraph, cmdInputwordgraph(): creates new wordgraph
  - isearch, cmdISearch(): incremental parsing of a wordgraph
  - level, cmdLevel(): prints out one or all level declarations
  - levelsort, cmdLevelsort(): set/check the priority of the defined levels
  - lexicon, cmdLexicon(): prints out one or all lexical entries
  - license, cmdLicense(): prints out the license
  - load, cmdLoad(): loads a file
  - ls, cmdLs(): lists one or more files
  - net, cmdNet(): prints out one or all constraint net
  - netdelete, cmdNetdelete(): deletes a constraint net
  - netsearch, cmdNetsearch(): searches a net for globally best solutions
  - newnet, cmdNewnet(): builds a new constraint net from a wordgraph
  - nonspeccompatible, cmdNonSpecCompatible():
    check compatibility of constraints for incremental parsing
  - parsedelete, cmdParsedelete(): deletes a parse
  - printparse, cmdPrintParse(): prints the structure of a parse
  - printparses, cmdPrintParses(): prints out all parses of a constraint net
  - quit, cmdQuit(): quits
  - renewnet, cmdRenewnet(): restores a net to pristine state
  - reset, cmdReset(): discards all loaded and computed information
  - section, cmdSection(): prints out one or all section
  - set, cmdSet(): set various system variables
  - showlevel, cmdShowlevel(): toggles whether a level is shown or not
  - status, cmdStatus(): describes the status of the system
  - tagger, cmdTagger(): start/stop POS tagger
  - testing, cmdTesting(): undocumented command for testing
  - useconstraint, cmdUseconstraint():
    toggles whether a constraint is used or not
  - uselevel, cmdUselevel(): toggles whether a level is used or not
  - uselexicon, cmdUseLexicon():
    specify which lexicon and which index database to use
  - version, cmdVersion(): prints version
  - verify, cmdVerify(): compares parse to annotation
  - weight, cmdWeight(): changes a constraint's weight
  - wordgraph, cmdWordgraph(): prints out one or all wordgraph
  - writenet, cmdWriteNet(): writes a TeX representation of the net
  - writeparses, cmdWriteParses(): writes XFig and LaTeX files with the parses
  - writewordgraph, cmdWriteWordgraph(): writes a wordgraph in an XFig file
   @{
*/

/* ---------------------------------------------------------------------- */
#include <config.h>

#include <blah.h>
#include <errno.h>
#include <libgen.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <glob.h>
#include <sys/resource.h>
#include <sys/stat.h>

#ifndef DARWIN
#include "compile.h"
#endif

#include "cdg.h"
#include "chart.h"
#include "chunker.h"
#include "command.h"
#include "constraintnet.h"
#include "eval.h"
#include "frobbing.h"
#include "gls.h"
  
#ifdef USE_NEWGLS
#include "newgls.h"
#endif

#include "hook.h"
#include "increment.h"
#include "incrementalcompletion.h"
#include "input.h"
#include "lexemgraph.h"
#include "netsearch.h"
#include "parse.h"
#include "parsing.h"
#include "predict.h"
#include "scache.h"
#include "search.h"
#include "set.h"
#include "shift.h"
#include "statistics.h"
#include "tagger.h"
#include "testing.h"
#include "timer.h"
#include "write.h"
#include "cdgdb.h"

/* -- TYPE DEFINITIONS -------------------------------------------------- */

/* -------------------------------------------------------------------------
   type of a function that implements a scripting command.
  
   The first parameter \c no is the number of arguments entered by the user
   after the command name. The array \c args holds these arguments as a
   list of plain strings. For example, the command "\c newnet \c s1"
   contains one command and argument. To handle it, the function
   cmdNewnet() will be called with two arguments: the number 1 (as we have
   one argument) and an array containing the argument strings, which is one
   here: "s1".
  
   The job of a CommandFunction() is to check the arguments for consistency
   and then either execute the action specified by the command or give reasons
   why they are incorrect.
  
   Each command returns
   a Boolean which is \ref TRUE on a successful command execution and otherwise
   \ref FALSE, that is when an error occurred.
*/
typedef Boolean CommandFunction(int no, char **args);

/* -------------------------------------------------------------------------
   command definition for the \ref interface_commands.
*/
typedef struct {

  String name;                          /**< the name of the command */

  rl_compentry_func_t *args[MAXARGNO];  /**< array of completion functions for each argument */

  CommandFunction *function;            /**< function that implements the command */

  Boolean same;  /**< completion flag. This is set to 1 for those commands which use the same 
                  * completion-function for all of their arguments */

  String doc;                           /**< little help text*/
} Command;

/* -- VARIABLES --------------------------------------------------------- */

/** \todo Write an explanation of the interface and the completion gadget */
static int interface_index = 0;

/** \todo Write an explanation of the interface and the completion gadget */
static int interface_length = 0;

/** \todo Write an explanation of the interface and the completion gadget */
static List interface_list = NULL;

/** \todo Write an explanation of the interface and the completion gadget */
static List interface_cell = NULL;

/** \todo Write an explanation of the interface and the completion gadget */
static HashIterator interface_hashiterator = NULL;

/* ------------------------------------------------------------------------
   flag indicating the end of the command main loop.
   Entering the commandLoop() this is set to \ref TRUE. commandLoop() exits
   if this global variable is set to FALSE.
*/
static Boolean commandLoopFlag = TRUE;

/* -- FORWARD DECLARATIONS ---------------------------------------------- */
char **interface_completion(String text, int start, int end);
char *annotation_completion_function(const char *, int);
char *command_completion_function(const char *, int);
char *constraint_completion_function(const char *, int);
char *frob_method_completion_function(const char *, int);
char *hierarchy_completion_function(const char *, int);
char *hook_completion_function(const char *, int);
char *level_completion_function(const char *, int);
char *levelsort_completion_function(const char *, int);
char *lexicon_completion_function(const char *, int);
char *make_rl_string(char *);
char *net_completion_function(const char *, int);
char *parse_completion_function(const char *, int);
char *search_method_completion_function(const char *, int);
char *section_completion_function(const char *, int);
char *set_completion_function(const char *, int);
char *word_completion_function(const char *, int);
char *wordgraph_completion_function(const char *, int);

/* ----------------------------------------------------------------------
   array of all registered wcdg scripting commands.
*/
Command interface_commands[] = {
  {"activate",
   {section_completion_function},
   cmdActivate, 1, "activates a section"},
  {"annotation",
   {annotation_completion_function},
   cmdAnnotation, 0, "prints out one or all annotation"},
  {"anno2parse",
   {annotation_completion_function},
   cmdAnno2Parse, 1, "build a parse from an annotation and evaluate it"},
  {"chart",
   {wordgraph_completion_function},
   cmdChart, 1, "search for solutions with a chart"},
  {"chunk",
   {wordgraph_completion_function},
   cmdChunk, 1, "call a chunk parser on a lattice"},
  {"closedb",
   {command_completion_function},
   cmdCloseDB, 1, "close database and write contents to disk"},
  {"compareparses",
   {parse_completion_function, parse_completion_function},
   cmdCompareParses, 0, "compares the structures of two parses"},
  {"compile",
   {NULL},
   cmdCompile, 0, "compiles the current grammar"},
  {"constraint",
   {constraint_completion_function},
   cmdConstraint, 1, "prints out one or all constraints"},
  {"deactivate",
   {section_completion_function},
   cmdDeactivate, 1, "deactivates a section"},
  {"distance",
   {net_completion_function},
   cmdDistance, 0, "prints the distance matrix of a net"},
  {"edges",
   {net_completion_function},
   cmdEdges, 0, "prints out one or all edges of a constraint net"},
  {"frobbing",
   {net_completion_function, frob_method_completion_function},
   cmdFrobbing, 0, "transforms an arbitrary assignment into a solution"},
  {"gls",
   {net_completion_function},
   cmdGls, 0, "solve a constraintnet using guided local search"},
#ifdef USE_NEWGLS
  {"ngls",
   {net_completion_function},
   cmdNewGls, 0, "solve a constraintnet using the new guided local search"},
#endif
  {"help",
   {command_completion_function},
   cmdHelp, 1, "gives a short help"},
  {"hierarchy",
   {hierarchy_completion_function},
   cmdHierarchy, 1, "prints out one or all hierarchies"},
  {"hook",
   {hook_completion_function},
   cmdHook, 0, "toggle hook-system on or off"},
  {"incrementalcompletion",
   {wordgraph_completion_function},
   cmdIncrementalCompletion, 0, "incremental completion of parses"},
  {"inputwordgraph",
   {word_completion_function},
   cmdInputwordgraph, 1, "creates new wordgraph"},
  {"isearch",
   {wordgraph_completion_function},
   cmdISearch, 0, "incremental parsing of a wordgraph"},
  {"level",
   {level_completion_function},
   cmdLevel, 1, "prints out one or all level declarations"},
  {"levelsort",
   {levelsort_completion_function},
   cmdLevelsort, 1, "set/check the priority of the defined levels"},
  {"lexicon",
   {lexicon_completion_function},
   cmdLexicon, 1, "prints out one or all lexical entries"},
  {"license",
   {NULL},
   cmdLicense, 0, "prints out the license"},
  {"load",
   {rl_filename_completion_function},
   cmdLoad, 1, "loads a file"},
  {"ls",
   {rl_filename_completion_function},
   cmdLs, 1, "lists one or more files"},
  {"net",
   {net_completion_function},
   cmdNet, 1, "prints out one or all constraint net"},
  {"netdelete",
   {net_completion_function},
   cmdNetdelete, 1, "deletes a constraint net"},
  {"netsearch",
   {net_completion_function, search_method_completion_function},
   cmdNetsearch, 0, "searches a net for globally best solutions"},
  {"newnet",
   {wordgraph_completion_function},
   cmdNewnet, 0, "builds a new constraint net from a wordgraph"},
  {"nonspeccompatible",
   {constraint_completion_function},
   cmdNonSpecCompatible, 0,
   "check compatibility of constraints for incremental parsing"},
  {"parsedelete",
   {parse_completion_function},
   cmdParsedelete, 1, "deletes a parse"},
  {"predict",
   {NULL},
   cmdPredict, 0, "registers an external predictor"},
  {"printparse",
   {parse_completion_function},
   cmdPrintParse, 0, "prints the structure of a parse"},
  {"printparses",
   {net_completion_function},
   cmdPrintParses, 0, "prints out all parses of a constraint net"},
  {"quit",
   {NULL},
   cmdQuit, 0, "quits"},
  {"renewnet",
   {net_completion_function},
   cmdRenewnet, 0, "restores a net to pristine state"},
  {"reset",
   {NULL},
   cmdReset, 0, "discards all loaded and computed information"},
  {"section",
   {section_completion_function},
   cmdSection, 0, "prints out one or all section"},
  {"set",
   {set_completion_function},
   cmdSet, 0, "set various system variables"},
  {"shift-reduce",
   {net_completion_function},
   cmdShift, 0, "builds parse tree as by shift and reduce actions"},
  {"showlevel",
   {level_completion_function},
   cmdShowlevel, 1, "toggles whether a level is shown or not"},
  {"status",
   {NULL},
   cmdStatus, 0, "describes the status of the system"},
  {"tagger",
   {NULL},
   cmdTagger, 0, "start/stop POS tagger"},
  {"testing",
   {NULL},
   cmdTesting, 0, "undocumented command for testing"},
  {"useconstraint",
   {constraint_completion_function},
   cmdUseconstraint, 1, "toggles whether a constraint is used or not"},
  {"uselevel",
   {level_completion_function},
   cmdUselevel, 1, "toggles whether a level is used or not"},
  {"uselexicon",
   {rl_filename_completion_function},
   cmdUseLexicon, 0, "specifies lexicon and its index database"},
  {"version",
   {NULL},
   cmdVersion, 0, "prints version"},
  {"verify",
   {parse_completion_function},
   cmdVerify, 0, "compares parse to annotation"},
  {"weight",
   {constraint_completion_function},
   cmdWeight, 1, "changes a constraint's weight"},
  {"wordgraph",
   {wordgraph_completion_function},
   cmdWordgraph, 1, "prints out one or all wordgraph"},
  {"writeannotation",
   {net_completion_function},
   cmdWriteAnno, 0, "writes best parse of a net to disk as an annotation"},
  {"writenet",
   {net_completion_function},
   cmdWriteNet, 0, "writes a TeX representation of the net"},
  {"writeparses",
   {net_completion_function},
   cmdWriteParses, 0, "writes XFig and LaTeX files with the parses"},
  {"writewordgraph",
   {wordgraph_completion_function},
   cmdWriteWordgraph, 1, "writes a wordgraph in an XFig file"},
  {"parses2prolog",
   {NULL},
   cmdParses2prolog, 0, "saves all parses to a file in Prolog format"},
  {"annos2prolog",
   {NULL},
   cmdAnnos2prolog, 0, "saves all annotations to a file in Prolog format"},

  /* end of array marker */
  {(char *)NULL,
   {NULL},
   NULL, 0, (char *)NULL}
};

/* ---------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
   Produce a string suitable for internal processing by readline.

   This function serves two purposes:
   
   -# it accepts a string and returns a clone of it. This is necessary
   because the readline library insists on doing its own deallocating.

   -# if the string passed contains whitespace, it wraps quotes around it
   so that the CDG shell will parse it as one string later. That way,
   completion works even for identifiers with whitespace in them (highly
   discouraged but possible).
   
   TODO: This is not yet quite correct. It can handle identifiers like
   "SYN", "um zu", and "it's". But it can fail if a string contains both
   whitespace and single quotes:

   "Hand me the 'scope, skipper!"

   My excuse is that I can think of no reason whatsoever why anybody would
   use that sort of identifier in a constraint grammar.
   
*/
char *make_rl_string(char *s)
{
  /* Usually, we just clone the argument. */
  if (strpbrk(s, rl_basic_word_break_characters) == NULL) {
    return (strCopy(s));
  }

  /* But if the string retrieved has whitespace in it, we wrap quotes around
   * it. */
  else {
    return strCopy(strPrintf("'%s'", s));
  }
}

/* ---------------------------------------------------------------------------
   Compute the completions for commands.

   This is one of many functions provided to pass information to the
   readline library. These functions all have basically the same structure
   which is required by readline: they must return a newly allocated string
   every time they are called. They vary only in which strings are returned
   as possible completions.

   The parameter "text" is the partial string typed by the user so far. The
   completion function must now offer the readline library suitable strings
   to complete it. This particular function knows what commands the CDG
   shell understands, so it iterates over the array interface_commands[]
   internally. The other completion functions will iterate over other
   structures, obviously.
*/
char *command_completion_function(const char *text, int state)
{

  String name;

  /* The parameter "state" indicates what state the readline computations are 
     currently in. state == 0 indicates that this is the first call to this
     function during an attempted completion. Therefore we now initialise the 
     iterator. */
  if (state == 0) {
    interface_index = 0;
    interface_length = strlen(text);
  }

  /* Note that this loop will exit the entire function as soon as the first
     match is found. Therefore, it may take several calls to this function
     until the loop terminates. This coroutine-like behaviour is slightly
     weird, but readline absolutely demands it. This is also why we use a
     global variable as an iterator. */
  while ((name = interface_commands[interface_index].name) != NULL) {
    interface_index++;

    /* A match is found if the beginning of the string retrieved from the
       internals of CDG matches the partial string typed so far. */
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }

  /* after the loop has terminated, we return NULL to tell the readline
     library no further matches will follow. */
  return (NULL);
}

/* ----------------------------------------------------------------------
   Compute completions for constraints.
*/
char *constraint_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_cell = interface_list =
      hashListOfKeys(inputCurrentGrammar->constraints);
    interface_length = strlen(text);
  }
  while (interface_cell) {
    name = interface_cell->item;
    interface_cell = interface_cell->next;
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }

  listDelete(interface_list);
  return (NULL);
}

/* ----------------------------------------------------------------------
   Compute completions for the command level.
*/
char *level_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_list = inputCurrentGrammar->levels;
    interface_length = strlen(text);
  }
  while (interface_list != NULL &&
         (name = ((Level) interface_list->item)->id) != NULL) {
    interface_list = interface_list->next;
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }
  return (NULL);
}

/* ----------------------------------------------------------------------
   Compute completions for the command levelsort.
*/
char *levelsort_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_list = inputCurrentGrammar->levels;
    interface_length = strlen(text);
  }
  while (interface_list != NULL &&
         (name = ((Level) interface_list->item)->id) != NULL) {
    interface_list = interface_list->next;
    /* Each argument is only to be used once. Therefore, the following */
    /* line checks if an argument is already in the rl_line_buffer. */
    if (strstr(rl_line_buffer, name))
      continue;
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }
  return (NULL);
}

/* ----------------------------------------------------------------------
   Compute completions for lexical entries.
*/
char *lexicon_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_hashiterator =
      hashIteratorNew(inputCurrentGrammar->lexicon);
    interface_length = strlen(text);
  }

  while (NULL != (name = (String) hashIteratorNextKey(interface_hashiterator))) {
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }

  hashIteratorDelete(interface_hashiterator);
  interface_hashiterator = NULL;
  return (NULL);
}

/* ----------------------------------------------------------------------
   Compute completions for orthographic words.
*/
char *word_completion_function(const char *text, int state)
{
  String name;
  String form;

  if (state == 0) {
    interface_hashiterator =
      hashIteratorNew(inputCurrentGrammar->lexicon);
    interface_length = strlen(text);
  }

  while (NULL != (name = (String) hashIteratorNextKey(interface_hashiterator))) {
    form = name;
    if (interface_length == 0 || strncmp(form, text, interface_length) == 0) {
      return (strCopy(form));
    }
  }

  hashIteratorDelete(interface_hashiterator);
  interface_hashiterator = NULL;
  return (NULL);
}


/* ----------------------------------------------------------------------
   Compute completions for wordgraphs.
*/
char *wordgraph_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_list = inputCurrentGrammar->lattices;
    interface_length = strlen(text);
  }
  while (interface_list != NULL &&
         (name = ((Level) interface_list->item)->id) != NULL) {
    interface_list = interface_list->next;
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }

  listDelete(interface_list);
  return (NULL);
}

/* ----------------------------------------------------------------------
   Compute completions for annotations.
*/
char *annotation_completion_function(const char *text, int state)
{

  if (state == 0) {
    // initialization
    // TODO: make hashListOfKeys interruptable ?
    interface_cell = interface_list =
      hashListOfKeys(inputCurrentGrammar->annotations);
    interface_index = -1;	       // means this List is a List of
    // Strings, so no index is neccessary
    interface_length = strlen(text);

  }

  if (interface_index == -1) {
    // is it a List of Strings ? -> we're in the loaded annotations List,
    // then pass the items from the List that have text as prefix
    while (interface_cell) {
      String name = interface_cell->item;

      interface_cell = interface_cell->next;
      if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
	return (make_rl_string(name));
      }
    }

    // End of the List -> Free it
    listDelete(interface_list);

    // Build the List of not yet loaded annotations 
    // - can be interrupted, so an incomplete List is passed
    interface_cell = interface_list = getListOfAllAnnotations((String)text);
    interface_index = 0;
  }
 
  // it is not a List of Strings ? -> this must be the GlobList
  while (interface_cell) {

    glob_t *glob = (glob_t *) interface_cell->item;

    if (interface_index < glob->gl_pathc) {
      String fname = glob->gl_pathv[interface_index];
      char *ext_begin = strrchr(fname, '.');
      char *name_begin = strrchr(fname, '/');
      int length = (int)ext_begin - (int)name_begin;

      if (length <= 0) {
	return strCopy(""); // must be a free()able pointer
      }

      char *annoId = memMalloc(sizeof(char) * length);

      memcpy(annoId, &name_begin[1], length);
      annoId[length - 1] = '\0';

      interface_index++;

      /* Uh, don't use make_rl_string() here as this will duplicate
       * our string once again and might register the annoId globally.
       * The string allocated here is already readline_ready.
       */
      if (strpbrk(annoId, rl_basic_word_break_characters) == NULL) {
	return annoId;
      } else {
	/* add quotes arround the result */
	annoId = memRealloc(annoId, sizeof(char) * (length + 2));
	memmove(annoId, &annoId[1], length);
	annoId[0] = '\''; annoId[length - 1] = '\''; annoId[length] = '\0';
	return annoId;
      }
    }

    interface_cell = interface_cell->next;
    interface_index = 0;
  }

  // End of GlobList -> Free it
  listForEachDelete(interface_list, &globfree);

  return (NULL);

}

/* ----------------------------------------------------------------------
   Compute completions for hierarchies.
*/
char *hierarchy_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_cell = interface_list =
      hashListOfKeys(inputCurrentGrammar->hierarchies);
    interface_length = strlen(text);
  }

  while (interface_cell) {
    name = interface_cell->item;
    interface_cell = interface_cell->next;
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }

  listDelete(interface_list);
  return (NULL);
}

/* ----------------------------------------------------------------------
   Compute completions for constraint nets.
*/
char *net_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_cell = interface_list =
      hashListOfKeys(cdgNets);
    interface_length = strlen(text);
  }

  while (interface_cell) {
    name = interface_cell->item;
    interface_cell = interface_cell->next;
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }

  listDelete(interface_list);
  return NULL;
}

/* ----------------------------------------------------------------------
   Compute completions for search methods.
*/
char *search_method_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_index = 0;
    interface_length = strlen(text);
  }
  while ((name = search_methods[interface_index].id) != NULL) {
    interface_index++;
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }
  return (NULL);
}

/* ----------------------------------------------------------------------
   Compute completions for sections.
*/
char *section_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_list = inputSections;
    interface_length = strlen(text);
  }
  while (interface_list != NULL &&
         (name = ((Section) interface_list->item)->id) != NULL) {
    interface_list = interface_list->next;
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }
  return (NULL);
}

/* ----------------------------------------------------------------------
   Compute completions for the set command
*/
char *set_completion_function(const char *text, int state)
{
  String name, result = NULL;
  List names = setReturnListOfNames();
  Vector options = listToVector(names);

  listDelete(names);

  if (state == 0) {
    interface_length = strlen(text);
    interface_index = 0;
  }

  while (interface_index < vectorSize(options)) {
    name = vectorElement(options, interface_index++);
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      result = make_rl_string(name);
      break;
    }
  }

  vectorDelete(options);
  return result;
}

/* ----------------------------------------------------------------------
   Compute completions for frobbing methods.
*/
char *frob_method_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_index = 0;
    interface_length = strlen(text);
  }
  while ((name = frobMethods[interface_index].id) != NULL) {
    interface_index++;
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }

  return (NULL);
}

/* ----------------------------------------------------------------------
   Compute completions for known hooks.
*/
char *hook_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_index = 0;
    interface_length = strlen(text);
  }
  while (interface_index < vectorSize(hkHooks)) {
    name = ((Hook) vectorElement(hkHooks, interface_index))->name;
    interface_index++;
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }

  return (NULL);
}

/* ----------------------------------------------------------------------
   Compute completions for parses.
*/
char *parse_completion_function(const char *text, int state)
{
  String name;

  if (state == 0) {
    interface_cell = interface_list =
      hashListOfKeys(cdgParses);
    interface_length = strlen(text);
  }

  while (interface_cell) {
    name = interface_cell->item;
    interface_cell = interface_cell->next;
    if (interface_length == 0 || strncmp(name, text, interface_length) == 0) {
      return (make_rl_string(name));
    }
  }

  return (NULL);
}

/* ----------------------------------------------------------------------
   Return a copy of the next command word in S.

   Command words are searched in the part of S starting at INDEX and not
   exceeding STOP. Command words are separated by instances of the
   characters defined in rl_basic_word_break_characters by the readline
   library.
*/
String getNextArgument(String s, int *index, int stop)
{
  int start;
  char buffer[1024];
  int bIndex = 0;
  int escaped = 0;

  while (s[*index] != '\0' &&
         s[*index] != '\'' &&
         strchr(rl_basic_word_break_characters, s[*index]) != NULL &&
         *index < stop) {
    (*index)++;
  }

  if (*index >= stop || s[*index] == '\0') {
    return (NULL);
  }

  if (s[*index] == '\'') {
    (*index)++;
    start = *index;
    if (s[*index] == '\\') {
      escaped = 1;
      (*index)++;
    }
    while (*index < stop && s[*index] != '\0' && !(s[*index] == '\'' && ! escaped)) {
      if (s[*index] == '\\' && ! escaped) {
	escaped = 1;
	(*index)++;
      } else {
	escaped = 0;
	buffer[bIndex++] = s[(*index)++];
      }
    }
    if (*index >= stop || s[*index] == '\0' || start == *index)
      return (NULL);
    (*index)++;
    if (*index - start > 1000) {
      cdgPrintf(CDG_WARNING, "WARNING: argument too long\n");
      return (NULL);
    }
    /**
    (void)strncpy(buffer, &s[start], *index - start - 1);
    buffer[*index - start - 1] = '\0';
    **/
    buffer[bIndex] = '\0';
  } else {
    start = *index;
    while (*index < stop &&
           s[*index] != '\0' &&
           strchr(rl_basic_word_break_characters, s[*index]) == NULL)
      (*index)++;
    if (*index >= stop || start == *index)
      return (NULL);

    if (*index - start + 1 > 1000) {
      cdgPrintf(CDG_WARNING, "WARNING: argument too long\n");
      return (NULL);
    }
    (void)strncpy(buffer, &s[start], *index - start);
    buffer[*index - start] = '\0';
  }

  return inputEncodeUmlauts(cdgXCDG ? strDecode(buffer) : buffer);
}

/* ----------------------------------------------------------------------
   Provide context sensitive command completion.

   This function is passed to the readline library as the
   rl_attempted_completion_function. See the readline documentation for
   details of its behaviour.
*/
char **interface_completion(String text, int start, int end)
{
  int i, j;
  String s, t;
  char **matches;

  i = 0;
  /* get command */
  s = getNextArgument(rl_line_buffer, &i, start);

  if (s == NULL) {
    /* no argument yet */
    matches = rl_completion_matches(text, command_completion_function);
  } else {
    /* count args */
    j = 0;
    while (NULL != (t = getNextArgument(rl_line_buffer, &i, start))) {
      /* hacky */
      cdgFreeString(t);
      j++;
    }

    if (j >= MAXARGNO) {
      matches = (char **)NULL;
    } else {
      /* find command entry */
      i = 0;
      while (NULL != (t = interface_commands[i].name) && strcmp(s, t) != 0) {
        i++;
      }

      /* CDG command not found? */
      if (t == NULL) {
        matches = (char **)NULL;
      } else {
        /* is the command one of those which have a fixed number of arguments 
         * ? */
        /* is there no completion-function-entry for array args, index j ? */
        if (interface_commands[i].args[j] == NULL &&
            interface_commands[i].same != 1) {
          matches = (char **)NULL;
        } else {
          if (interface_commands[i].same == 1) {
            /* this command is of the type which uses the same
               completion-function for all arguments */
            matches =
              rl_completion_matches(text, interface_commands[i].args[0]);
          } else {
            /* this command has a fixed number of arguments */
            matches =
              rl_completion_matches(text, interface_commands[i].args[j]);
          }
        }
      }
    }

    cdgFreeString(s);
  }

  return (matches);
}

/* ----------------------------------------------------------------------
   CDG main event loop.

   This function reads and executes commands and terminates
   when the \ref commandLoopFlag has been set to \ref FALSE.
   \param prompt to be displayed at the command line
   \returns the boolean result of the last command, that is \ref TRUE on
   success and else \ref FALSE.
*/
Boolean commandLoop(String prompt)
{
  String line;
  char buffer[MAXBUFFER];
  Boolean useReadline;
  Boolean result = TRUE;

  useReadline = isatty(fileno(stdin));

  /* completion function */
  rl_attempted_completion_function = (CPPFunction *) interface_completion;

  line = NULL;
  while (commandLoopFlag) {
    /* read from commandline */
    if (useReadline) {
      if (line) {
        memFree(line);
      }
      line = readline(prompt);
    } else {
      /* read from pipe */
      /* 
       * from manpage of fgets:
       * 
       * RETURN VALUES If end-of-file is encountered and no characters have
       * been read, no characters are transferred to s and a null pointer is
       * returned.  If a read error occurs, such as trying to use these
       * functions on a file that has not been opened for read- ing, a null
       * pointer is returned and the error indicator for the stream is set.
       * If end-of-file is encountered, the EOF indicator for the stream is
       * set.  Otherwise s is returned. */
      if (!fgets(buffer, MAXBUFFER, stdin)) {
        line = NULL;
      } else if (strlen(buffer) >= MAXBUFFER - 1) {
        cdgPrintf(CDG_WARNING, "WARNING: line too long, skipping\n");
        continue;                      /* with while(TRUE) */
      } else {
        line = buffer;
        /* Print line when reading from pipe (like initialization file). */
        cdgPrintf(CDG_DEFAULT, "%s", buffer);
      }
    }

    if (line == NULL) {
      /* arguments are not valid or stdin ended */
      cdgPrintf(CDG_DEFAULT, "\n");
      cmdQuit(0, NULL);
    }

    /* String ``line'' not registered anymore. Free'ing was buggy anyway. */
    result = commandEval(line);
  }
  if (useReadline && line) {
    memFree(line);
  }

  return result;
}

/* ----------------------------------------------------------------------
   CDG command dispatch routine.

   This function interprets one command of the scripting language.
   \param line the command line to be evaluated
   \returns \ref TRUE on a successful command execution, otherwise \ref FALSE.
*/
Boolean commandEval(String line)
{
  int len;
  int i = 0, j;
  String s, t;
  String args[MAXARGNO];
  Boolean result = TRUE;

  if (!line)
    return TRUE;

  len = strlen(line);
  /* get command */
  s = getNextArgument(line, &i, len + 1);

  if (s == NULL) {
    if (len != 0 && strcmp(line, "\n") != 0)
      cdgPrintf(CDG_ERROR, "ERROR: Oops, no valid command found\n");
    result = FALSE;
  } else if (s[0] == '#' || (s[0] == '/' && s[1] == '/')) {
    /* skip comments */
    result = TRUE;
  } else {

    add_history(line);

    for (j = 0; j < MAXARGNO; j++)
      args[j] = NULL;

    /* get arguments */
    for (j = 0;
         j < MAXARGNO &&
         (args[j] = getNextArgument(line, &i, len + 1)) != NULL; j++)
      /* do nothing */
      ;

    if (j == MAXARGNO && i < len) {
      cdgPrintf(CDG_ERROR, "ERROR: only %d arguments are allowed per line\n",
                MAXARGNO);
      return FALSE;
    }
    for (i = 0;
         (t = interface_commands[i].name) != NULL && strcmp(s, t) != 0; i++)
      /* do nothing */
      ;

    if (t == NULL || interface_commands[i].function == NULL) {
      cdgPrintf(CDG_ERROR, "ERROR: Oops, command `%s' unknown\n", s);
      result = FALSE;
    } else {
      Boolean tagFlag = FALSE;

      /* surround every command by xml tags */
      if (hkVerbosity & CDG_XML) {
        writeXmlCommand(line);
        tagFlag = TRUE;
      }

      /* call to command function */
      result = interface_commands[i].function(j, args);

      /* close the command tag */
      if (tagFlag && (hkVerbosity & CDG_XML)) {
        writeXmlEndTag();
      }
    }

    cdgFreeString(s);
    for (j = 0; j < MAXARGNO && args[j] != NULL; j++)
      cdgFreeString(args[j]);
  }
  cdgFlush();

  return result;
}

/* ----------------------------------------------------------------------
   Leave the command loop.

   @see commandLoop
*/
Boolean cmdQuit(int no, char **args)
{
  commandLoopFlag = FALSE;
  return TRUE;
}


/* ----------------------------------------------------------------------
   List one or more files.

   This simply spawns the system command \c ls. 
*/
Boolean cmdLs(int no, char **args)
{
  String lscmd = "ls -laF";
  int i;
  char buffer[1024];
  Boolean result = TRUE;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdLs: ERROR: argument is NULL\n");
    return FALSE;
  }

  if (no == 0) {
    result = (system(lscmd) != -1);
  } else {
    for (i = 0; i < no && args[i] != NULL; i++) {
      /* protection against buffer overflow */
      if (strlen(args[i]) > 512) {
        cdgPrintf(CDG_WARNING,
                  "cmdLs: argument number %d too large, ignoring\n", i);
      } else {
        sprintf(buffer, "%s %s", lscmd, args[i]);
        result = (system(buffer) != -1);
      }
    }
  }

  return result;
}

/* ----------------------------------------------------------------------
   Compile current grammar.

   Currently broken.
*/
Boolean cmdCompile(int no, char **args)
{
  Boolean result = TRUE;

#ifdef DARWIN
  cdgPrintf(CDG_ERROR,
            "Sorry, the MACOS X version currently doesn't support compiling grammars!\n");
#else
  String sofilename = comCompile(no, args);

  if (sofilename) {
    Input input = comLoad(sofilename);

    if (input) {
      cdgPrintf(CDG_INFO, "INFO: hooking compiled constraints\n");
      mergeInput(inputCurrentGrammar, input);
      inputCacheInput(inputCurrentGrammar);
    }
    memFree(input);
    evalEvaluationMethod = EMTCompiled;
  } else {
    result = FALSE;
  }
#endif

  return result;
}

/* ----------------------------------------------------------------------
   Print the specified constraints.
*/
Boolean cmdConstraint(int no, char **args)
{
  Constraint c;
  int i;
  Boolean result = FALSE;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdConstraint: ERROR: argument is NULL\n");
    return FALSE;
  }

  for (i = 0; i < no; i++) {
    if (NULL != (c = hashGet(inputCurrentGrammar->constraints, args[i]))) {
      printConstraint(CDG_DEFAULT, c);
      result = TRUE;
    }
  }

  return result;
}

/* ----------------------------------------------------------------------
   Print the specified level declarations.
*/
Boolean cmdLevel(int no, char **args)
{
  List l;
  Level c;
  int i, j, index, noLevels;
  Boolean result = FALSE;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdLevel: ERROR: argument is NULL\n");
    return FALSE;
  }

  /* show all requested levels */
  for (l = inputCurrentGrammar->levels; l; l = l->next) {
    c = (Level) l->item;
    if (no > 0) {
      for (i = 0; i < no && args[i] != NULL; i++) {
        if (strcmp(c->id, args[i]) == 0) {
          printLevel(CDG_DEFAULT, c);
          result = TRUE;
          break;
        }
      }
    } else {
      printLevel(CDG_DEFAULT, c);
      result = TRUE;
    }
  }

  if (no > 0)
    return result;

  cdgPrintf(CDG_DEFAULT,
            "\nnumber of binary constraints between values of given levels:\n\n");
  noLevels = listSize(inputCurrentGrammar->levels);
  cdgPrintf(CDG_DEFAULT, "       |");
  for (i = 0; i < noLevels; i++)
    cdgPrintf(CDG_DEFAULT, "%3d ", i);
  cdgPrintf(CDG_DEFAULT, "\n");
  cdgPrintf(CDG_DEFAULT, "   ----|");
  for (i = 0; i < noLevels; i++)
    cdgPrintf(CDG_DEFAULT, "----");
  cdgPrintf(CDG_DEFAULT, "\n");
  for (i = 0; i < noLevels; i++) {
    cdgPrintf(CDG_DEFAULT, "   %3d |", i);
    for (j = 0; j < noLevels; j++) {
      index = i * noLevels + j;
      cdgPrintf(CDG_DEFAULT, "%3d ",
                (int)vectorElement(inputCurrentGrammar->levelMatrixCounter,
                                     index));
    }
    cdgPrintf(CDG_DEFAULT, "\n");
  }

  c = inputGetMainlevel(inputCurrentGrammar);
  if (c) {
    cdgPrintf(CDG_DEFAULT, "\n\nMain level is %s\n", c->id);
  }

  return result;
}

/* ----------------------------------------------------------------------
   Simulate shift-reduce parsing.
*/
Boolean cmdShift(int no, char **args)
{
  Boolean result = TRUE;

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  /* set the timelimit for this command */
  if (cdgTimeLimit != 0) {
    timerSetAlarm(cdgTimeLimit);
  }

  result = shift_reduce(no, args);

  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;

  /* stop the alarm clock on suspicion */
  timerStopAlarm();

  return result;
}

/* ----------------------------------------------------------------------
   Toggle whether the specified level is shown or not.
*/
Boolean cmdShowlevel(int no, char **args)
{
  List l;
  Level c;
  int i;
  Boolean result = FALSE;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdShowlevel: ERROR: argument is NULL\n");
    return FALSE;
  }

  for (l = inputCurrentGrammar->levels; l; l = l->next) {
    c = (Level) l->item;
    if (no > 0) {
      for (i = 0; i < no && args[i] != NULL; i++) {
        if (strcmp(c->id, args[i]) == 0) {
          c->showflag = !c->showflag;
          cdgPrintf(CDG_INFO, "INFO: level `%s' %s shown now\n",
                    c->id, c->showflag ? "is" : "isn't");
          break;
          result = TRUE;
        }
      }
    } else {
      c->showflag = !c->showflag;
      cdgPrintf(CDG_INFO, "INFO: level `%s' %s shown now\n",
                c->id, c->showflag ? "is" : "isn't");
      result = TRUE;
    }
  }
  return result;
}

/* ----------------------------------------------------------------------
   Toggle whether the specified constraint is used or not.
*/
Boolean cmdUseconstraint(int no, char **args)
{
  Constraint c;
  int i;
  Boolean change = FALSE;
  
  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdUseconstraint: ERROR: argument is NULL\n");
    return FALSE;
  }

  for (i = 0; i < no; i++) {
    if (NULL != (c = hashGet(inputCurrentGrammar->constraints, args[i]))) {
      c->active = !c->active;
      change = TRUE;
      cdgPrintf(CDG_INFO, "INFO: constraint `%s' %s used now\n",
                c->id, c->active ? "is" : "isn't");
    }
  }

  /* delete all constraint nets */
  if (change) {
    cdgDeleteComputed();
  }

  return change;
}

/* ----------------------------------------------------------------------
   Toggle whether the specified level is used or not.
*/
Boolean cmdUselevel(int no, char **args)
{
  Level c;
  int i;
  Boolean change = FALSE;
  List l;
  
  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdUselevel: ERROR: argument is NULL\n");
    return FALSE;
  }

  for (l = inputCurrentGrammar->levels; l; l = l->next) {
    c = (Level) l->item;
    if (no > 0) {
      for (i = 0; i < no && args[i] != NULL; i++) {
        if (strcmp(c->id, args[i]) == 0) {
          c->useflag = !c->useflag;
          change = TRUE;
          cdgPrintf(CDG_INFO, "INFO: level `%s' %s used now\n",
                    c->id, c->useflag ? "is" : "isn't");
          break;
        }
      }
    } else {
      c->useflag = !c->useflag;
      change = TRUE;
      cdgPrintf(CDG_INFO, "INFO: level `%s' %s used now\n",
                c->id, c->useflag ? "is" : "isn't");
    }
  }

  /* delete all constraint nets */
  if (change) {
    cdgDeleteComputed();
  }

  return change;
}

/* ----------------------------------------------------------------------
   Print the specified lexicon entries.
*/
Boolean cmdLexicon(int no, char **args)
{
  List l;
  Boolean result = FALSE;

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  if (no == 0) {
    HashIterator hi = hashIteratorNew(inputCurrentGrammar->lexicon);
    String form;

    while (NULL != (form = hashIteratorNextKey(hi))) {
      List items = inputLexiconGet(form);

      for (l = items; l != NULL; l = l->next) {
        printLexiconItem(CDG_DEFAULT, l->item);
        cdgPrintf(CDG_INFO, "\n");
        result = TRUE;
      }
      listForEachDelete(items, freeLexiconItem);
      if (cdgCtrlCTrapped) {
        break;
      }
    }
    hashIteratorDelete(hi);

  } else {
    int i;
    List l, lexicon;

    for (i = 0; i < no; i++) {
      lexicon =
        inputLexiconGet(args[i]);
      for (l = lexicon; l != NULL; l = l->next) {
        printLexiconItem(CDG_DEFAULT, l->item);
        result = TRUE;
        cdgPrintf(CDG_INFO, "\n");
        if (cdgCtrlCTrapped) {
          break;
        }
      }
      listForEachDelete(lexicon, freeLexiconItem);
    }
  }

  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;

  return result;
}

/* ----------------------------------------------------------------------
   Print wordgraphs.
*/
Boolean cmdWordgraph(int no, char **args)
{
  Lattice lat;
  int i;
  Boolean result = FALSE;
  // cdgPrintf(CDG_DEFAULT, args[0]);
  // cdgPrintf(CDG_DEFAULT, args[1]);

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdWordgraph: ERROR: argument is NULL\n");
    return FALSE;
  }
  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  if (no == 0) {
    // Show all wordgraphs
    List l;
    for(l = inputCurrentGrammar->lattices; l != NULL; l = l->next) {
      Lattice lat = l->item;
      printLattice(CDG_DEFAULT, lat);
      writeXmlLattice(lat);
      result = TRUE;
    }
  }
  else if (strcmp(args[0],"-c") == 0) 
    {
      int found=0;
      // Show all wordgraphs matching the key word
      List l;
      for(l = inputCurrentGrammar->lattices; l != NULL; l = l->next) {
	Lattice lat = l->item;
	List l2;
	for (l2 = lat->arcs; l2 != NULL; l2 = l2->next) {
	  Arc arc=l2->item;
	  if (strcmp(args[1],arc->word) == 0)
	    {
	      // found match --> print wordgraph
	      printLattice(CDG_DEFAULT, lat);
	      writeXmlLattice(lat);
	      found=1;
	      break;
	    }
	}
      }
       if (found==0) { cdgPrintf(CDG_DEFAULT,"Sorry, no matches.\n"); }
       result = TRUE;
    }
  else {
    // Show specific wordgraphs
    for (i = 0; i < no && ! cdgCtrlCTrapped; i++) {
      if (NULL != (lat = findLattice(args[i]))) {
        printLattice(CDG_DEFAULT, lat);
        writeXmlLattice(lat);
        result = TRUE;
      }
    }
  }

  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;

  return result;
}

/* ----------------------------------------------------------------------
   Write the specified wordgraphs to XFig file.
*/
Boolean cmdWriteWordgraph(int no, char **args)
{
  List l;
  Lattice lat;
  Boolean withWeights = FALSE;
  Boolean result = FALSE;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdWriteWordgraph: ERROR: argument is NULL\n");
    return FALSE;
  }

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;
  switch (no) {
    case 0:
    case 1:
      cdgPrintf(CDG_ERROR,
                "cmdWriteWordgraph: ERROR: writewordgraph wordgraph XFig_filename [-w]\n");
      cdgPrintf(CDG_ERROR,
                "                          -w = writes arc weights;\n");
      cdgPrintf(CDG_ERROR,
                "                          by default, the weights are not written\n");
      result = FALSE;
    default:
      if (no >= 3 && strcmp(args[2], "-w") == 0)
        withWeights = TRUE;

      for(l = inputCurrentGrammar->lattices;
          l != NULL && !cdgCtrlCTrapped;
          l = l->next) {
        lat = l->item;
        if (strcmp(lat->id, args[0]) == 0 || graphContainsWord(lat, args[0])) {
          writeLattice(lat, withWeights, args[1]);
          result = TRUE;
        }
      }
  }
  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;

  return result;
}

/* ----------------------------------------------------------------------
   Print the specified annotation entries.
*/
Boolean cmdAnnotation(int no, char **args)
{
  AnnoEntry ae;
  int i;
  Boolean result = FALSE;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdAnnotation: ERROR: argument is NULL\n");
    abort();
  }
  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  if (no == 0) {
    HashIterator hi = hashIteratorNew(inputCurrentGrammar->annotations);
    while (NULL != (ae = hashIteratorNextValue(hi)) && ! cdgCtrlCTrapped) {
      printAnnoEntry(CDG_DEFAULT, ae);
      writeXmlAnnoEntry(ae);
      result = TRUE;
    }
    hashIteratorDelete(hi);
  } else {
    for (i = 0; i < no && ! cdgCtrlCTrapped; i++) {
      if (NULL != (ae = findAnnoByName(args[i], TRUE))) {
        printAnnoEntry(CDG_DEFAULT, ae);
        writeXmlAnnoEntry(ae);
        result = TRUE;
      }
    }
  }

  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;

  return result;
}

/* ----------------------------------------------------------------------
   Build a parse from an annotation and evaluate it against the current 
   grammar.
*/
Boolean cmdAnno2Parse(int no, char **args)
{
  AnnoEntry ae = NULL;
  Parse parse = NULL;
  Lattice lattice = NULL;
  List annos = NULL;
  LexemGraph lg = NULL;
  List l;
  int i, noParses;
  Boolean result = TRUE;

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  /* Make list of annotations to process */
  if (no == 0) {
    HashIterator hi = hashIteratorNew(inputCurrentGrammar->annotations);
    List l;
    Hashtable ht;
    
    /* process all */
    while (NULL != (ae = hashIteratorNextValue(hi))) {
      annos = listPrependElement(annos, ae);
    }
    hashIteratorDelete(hi);

    if(dbAvailable()) {

      /* If all available annotations are to be converted, and lexicon items
         need to be autoloaded, then there will probably be a great lot of
         them. Therefore it pays to optimize for this case and issue one huge
         autoload instead of 12374 individual ones. */
      ht = hashNew(500, 0.8, hashStringHashFunction, hashStringEqualFunction);

      for(l = annos; l != NULL; l = l->next) {
        AnnoEntry ae = l->item;
        List m;
        for(m = ae->annos; m != NULL; m = m->next) {
          Annotation a = m->item;
          hashSet(ht, a->word, (Pointer)1);
        }
      }

      l = hashListOfKeys(ht);
      dbLoadAll(l);
      listDelete(l);
      hashDelete(ht);
    }
  }
    
  else {
    /* process some */
    for (i = 0; i < no && args[i]; i++) {
      if ((ae = findAnnoByName(args[i], TRUE))) {
        if (!parseFind(ae->id)) {
          annos = listPrependElement(annos, ae);
        } else {
          cdgPrintf(CDG_WARNING,
                    "WARNING: parse for anno `%s' already exists...skipping\n",
                    ae->id);
        }
      }
    }
  }

  /* now process all annos in the list */
  noParses = 0;
  for (l = annos; l != NULL; l = l->next) {
    ae = l->item;
    parse = parseFromAnno(ae);
    if (!parse)
      continue;

    /* search lattice for this anno */
    lattice = findLattice(ae->lattice);
    if (!lattice) {
      lattice = latticeFromAnno(ae);
    }
    if (NULL != (lg = lgNew(lattice))) {
      if (!parseDecorate(parse, lg, ae)) {
        cdgPrintf(CDG_WARNING,
                  "WARNING: could not decorate parse for annotation `%s'.\n",
                  ae->id);
      } else {
        parseEval(parse);
        lgDelete(lg);

        /* check for word mismatch between annotation and associate lattice */
        if(!latticeBranches(lattice) &&
           vectorSize(parse->words) != listSize(lattice->arcs)) {
          cdgPrintf(CDG_WARNING,
                    "WARNING: Number of words is not the same in lattice and annotation!\n");
        }
        
        
      }
    }
    parseRegister(parse);
    noParses++;

    if(0 == noParses % 100) {
      cdgPrintf(CDG_PROGRESS,
                "PROGRESS: generated %d parse%s\n",
                noParses,
                noParses == 1 ? "" : "s");
    }
    
    if(cdgCtrlCTrapped) {
      break;
    }

  }

  if (noParses == 0) {
    cdgPrintf(CDG_ERROR, "ERROR: no parses generated \n");
    result = FALSE;
  } else {
    cdgPrintf(CDG_INFO,
              "INFO: generated %d parse%s\n",
              noParses,
              noParses == 1 ? "" : "s");

    result = TRUE;
  }

  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;
  
  if (annos)
    listDelete(annos);

  return result;
}

/* ----------------------------------------------------------------------
   Print the specified hierarchies.
*/
Boolean cmdHierarchy(int no, char **args)
{
  Hierarchy h;
  int i;
  Boolean result = FALSE;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdHierarchy: ERROR: argument is NULL\n");
    return FALSE;
  }

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  for (i = 0; i < no && ! cdgCtrlCTrapped; i++) {
    if ((h = hashGet(inputCurrentGrammar->hierarchies, args[i]))) {
      printHierarchy(CDG_DEFAULT, h);
      /* TODO: writeXmlHierarchy */
      result = TRUE;
    }
  }

  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;

  return result;
}

/* ----------------------------------------------------------------------
   Print the specified constraint sections.
*/
Boolean cmdSection(int no, char **args)
{
  List l;
  Section s;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdSection: ERROR: argument is NULL\n");
    return FALSE;
  }
  for (l = inputSections; l; l = l->next) {
    s = (Section) l->item;
    cdgPrintf(CDG_DEFAULT, "%s, %5d constraint(s), %s\n",
              s->active ? "  active" : "inactive", s->counter, s->id);
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Set the specified CDG variable.
*/
Boolean cmdSet(int no, char **args)
{
  String p, v;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdSet: ERROR: argument is NULL\n");
    return FALSE;
  }

  if (no == 1) {
    return setPrint(CDG_DEFAULT, args[0]);
  }

  if (no == 2) {
    p = args[0];
    v = args[1];
    return setValue(p, v);
  }

  cdgPrintf(CDG_ERROR,
            "ERROR: `set' expects one or two arguments, but got %d\n", no);

  return FALSE;
}

/* ----------------------------------------------------------------------
   Deactivate constraint classes.

   All constraint classes that are named as arguments are deactivated, that
   is, their field ->active is set to FALSE. The effect is that these
   constraints are not used for evaluation, as if they had never been
   defined. Since this changes the basis of evaluation, this is treated
   like a change in the grammar, i.e. all previously computed structures
   become invalid and are deleted.
*/
Boolean cmdDeactivate(int no, char **args)
{
  Section s;
  int i;
  Boolean change = FALSE;
  
  if (0 == no) {
    List l;
    change = TRUE;
    for(l = inputSections; l != NULL; l = l->next) {
      s = l->item;
      s->active = FALSE;
    }
  }

  for (i = 0; i < no && args[i] != NULL; i++) {
    if (NULL != (s = findSection(args[i]))) {
      change = TRUE;
      s->active = FALSE;
    }
  }

  /* delete all constraint nets */
  if (change) {
    cdgDeleteComputed();
  }

  return change;
}

/* ----------------------------------------------------------------------
   Activate constraint classes.

   All constraint classes that are named as arguments are activated, that
   is, their field ->active is set to TRUE. The effect is that these
   constraints are used (again) for evaluation. Since this changes the
   basis of evaluation, this is treated like a change in the grammar, i.e. 
   all previously computed structures become invalid and are deleted.
*/
Boolean cmdActivate(int no, char **args)
{
  Section s;
  int i;
  Boolean change = FALSE;
  
  if (0 == no) {
    List l;
    change = TRUE;
    for(l = inputSections; l != NULL; l = l->next) {
      s = l->item;
      s->active = TRUE;
    }
  }

  for (i = 0; i < no && args[i] != NULL; i++) {
    if (NULL != (s = findSection(args[i]))) {
      change = TRUE;
      s->active = TRUE;
    }
  }
  /* delete all constraint nets */
  if (change) {
    cdgDeleteComputed();
  }

  return change;
}

/* ----------------------------------------------------------------------
   Compare the specified parses.
*/
Boolean cmdCompareParses(int no, char **args)
{
  Parse p1 = NULL;
  Parse p2 = NULL;
  ParseCmpStatStruct stat;
  Boolean result = TRUE;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdCompareParses: ERROR: argument is NULL\n");
    return FALSE;
  }

  if (0 == no || no > 2) {
    cdgPrintf(CDG_ERROR,
              "ERROR: `compareparses' needs two parses: a reference and a test\n");
    return FALSE;
  }

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  p1 = hashGet(cdgParses, args[0]);
  p2 = hashGet(cdgParses, args[1]);
  if (no == 1) {
    p2 = parseMostRecentlyCreatedParse;
  }

  if (p1 == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: can't find a parse `%s'\n", args[0]);
    return FALSE;
  }
  if (p2 == NULL) {
    if (no == 2) {
      cdgPrintf(CDG_ERROR, "ERROR: can't find a parse `%s'\n", args[1]);
    } else {
      cdgPrintf(CDG_ERROR, "ERROR: can't find a recent parse\n");
    }
    return FALSE;
  }

  result = (parseCompare(p1, p2, &stat, PC_ALL, PC_STRICT) >= 0);
  writeXmlParseComparison(p1, p2, &stat);

  return result;
}

/* ----------------------------------------------------------------------
   Introduces an external knowledge source.
*/
Boolean cmdPredict(int no, char **args)
{
  Predictor p = NULL;
  List features = NULL;
  String name;
  int i;
  int pos;
  if (2 > no) {
    cdgPrintf(CDG_ERROR, "ERROR: `predict' takes at least two arguments.\n");
    return FALSE;
  }
  
  name=strRegister(args[0]);
  pos=predictFindPredictor(name);

  /* `off' switches off a predictor */
  if(!strcmp("off", args[1])) {
    if (pos > 0) {
      p=(Predictor)vectorElement(cdgPredictors,pos-1);
      predictFree(p);
      vectorRemoveElementAt(cdgPredictors,pos-1);
      cdgPrintf(CDG_ERROR, "INFO: predictor `%s' removed.\n", args[0]);
    } else {
      cdgPrintf(CDG_ERROR, "INFO: no predictor %s found to remove.\n",
		args[0]);
    }
    return TRUE;
  }
  
  for(i = 2; i < no; i++) {
    features = listAppendElement(features, strRegister(args[i]));
  }

  p = predictNew(args[0], args[1], features);

  if (pos == 0) {
      cdgPrintf(CDG_ERROR, "Warning: cannot replace the POS tagger (`%s').\n", args[0]);
  } else if (pos > 0) {
    Predictor p2=(Predictor)vectorSetElement(cdgPredictors,p,pos-1);
    predictFree(p2);
  } else {
    vectorAddElement(cdgPredictors,p);
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Prints the specified Parse on stdout.
*/
Boolean cmdPrintParse(int no, char **args)
{
  Parse p = NULL;
  AnnoEntry anno = NULL;
  int format = 0; /* 0: log, 1: cda, 2: xml */
  unsigned long int oldHkVerbosity;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdPrintParse: ERROR: argument is NULL\n");
    return FALSE;
  }

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;
  if (no == 0) {
    p = parseMostRecentlyCreatedParse;
    if (p == NULL) {
      cdgPrintf(CDG_ERROR, "ERROR: can't find a recent parse\n");
      return FALSE;
    }
  } else if (no == 1) {
    p = hashGet(cdgParses, args[0]);
    if (p == NULL) {
      cdgPrintf(CDG_ERROR, "ERROR: can't find a parse `%s'\n", args[0]);
      return FALSE;
    }
  } else if (no >= 2) {
    if (strcmp(args[0], "-f") == 0) {
      if (strcmp(args[1], "log") == 0) {
	format = 0;
      } else if (strcmp(args[1], "cda") == 0) {
	format = 1;
      } else if (strcmp(args[1], "xml") == 0) {
	format = 2;
      } else {
	cdgPrintf(CDG_ERROR, 
	  "ERROR: invalid format spec `%s', should be log, cda or xml\n", args[1]);
	return FALSE;
      }
      if (no == 3) {
	p = hashGet(cdgParses, args[2]);
      } else {
	p = parseMostRecentlyCreatedParse;
      }
      if (p == NULL) {
	cdgPrintf(CDG_ERROR, "ERROR: can't find a parse `%s'\n", args[2]);
	return FALSE;
      }
    }
  } else {
    cdgPrintf(CDG_ERROR, "ERROR: `printparse' needs one parse\n");
    return FALSE;
  }

  oldHkVerbosity = hkVerbosity;
  if (format == 0) {
    parsePrint(p);
  }
  if (format == 2) {
    hkVerbosity |= CDG_XML;
  }
  writeXmlParse(p);
  

  /* write extra annotation info */
  anno = parse2annotation(p);
  if (anno) {                          /* for sure */
    if (format == 1) {
      printAnnoEntry(CDG_INFO, anno);
    }
    writeXmlAnnoEntry(anno);
    freeAnnoEntry(anno);
    hkVerbosity = oldHkVerbosity;
    return TRUE;
  }

  hkVerbosity = oldHkVerbosity;
  return FALSE;
}

/* ----------------------------------------------------------------------
   Print the specified constraint net.
*/
Boolean cmdNet(int no, char **args)
{
  int i;
  ConstraintNet net = NULL;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdNet: ERROR: argument is NULL\n");
    return FALSE;
  }

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;
  if (no > 0) {
    for (i = 0; i < no && args[i] != NULL && !cdgCtrlCTrapped; i++) {
      net = hashGet(cdgNets, args[i]);
      if (!net) {
        cdgPrintf(CDG_ERROR, "ERROR: can't find a net `%s'\n", args[i]);
        return FALSE;
      }
    }
  } else {
    if (cnMostRecentlyCreatedNet != NULL) {
      net = cnMostRecentlyCreatedNet;
    }
  }

  if (net) {
    cnPrint(CDG_DEFAULT, net);
    return TRUE;
  } else {
    cdgPrintf(CDG_ERROR,
              "ERROR: `net' needs a constraint net id or a default net\n");
    return FALSE;
  }
}

/* ----------------------------------------------------------------------
   Adjust the weight of a constraint.
*/
Boolean cmdWeight(int no, char **args)
{
  Number x;
  HashIterator hi;
  Constraint c;
  Section s;
  static char *usage = "usage: weight [<constraint> [[absolute] <weight>]]\n";

  switch (no) {
    case 0:
      hi = hashIteratorNew(inputCurrentGrammar->constraints);
      while ((c = hashIteratorNextValue(hi))) {
        cdgPrintf(CDG_INFO, "%s: %7.5f\n", c->id, inputGetWeight(c));
      }
      hashIteratorDelete(hi);
      break;

    case 1:
      if ((c = hashGet(inputCurrentGrammar->constraints, args[0]))) {
        cdgPrintf(CDG_INFO, "%s: %4.3e\n", c->id, inputGetWeight(c));
      }
      break;

    case 2:
      if (!sscanf(args[1], "%lf", &x)) {
        cdgPrintf(CDG_ERROR, "ERROR: can't read weight `%s'\n", args[1]);
        return FALSE;
      }
      if ((c = hashGet(inputCurrentGrammar->constraints, args[0]))) {
        inputSetWeight(c, x, FALSE);
      } else if (NULL != (s = findSection(args[0]))) {
        inputSetGroupWeight(s, x);
      } else {
        cdgPrintf(CDG_ERROR, "ERROR: can't use identifier `%s'\n", args[0]);
        return FALSE;
      }
      break;
    case 3:
      if ((c = hashGet(inputCurrentGrammar->constraints, args[0]))) {
        if (strcmp(args[1], "absolute")) {
          cdgPrintf(CDG_ERROR, usage);
          return FALSE;
        }
        if (!sscanf(args[2], "%lf", &x)) {
          cdgPrintf(CDG_ERROR, "ERROR: can't read weight `%s'\n", args[2]);
          return FALSE;
        }
        inputSetWeight(c, x, TRUE);
      }
      break;
    default:
      cdgPrintf(CDG_ERROR, usage);
      return FALSE;
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Perform a complete search for globally best solution.
*/
Boolean cmdNetsearch(int no, char **args)
{
  ConstraintNet net;
  Boolean result;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdNetsearch: ERROR: argument is NULL\n");
    return FALSE;
  }

  if (no < 1 || args[0] == NULL) {
    if (cnMostRecentlyCreatedNet == NULL) {
      cdgPrintf(CDG_ERROR,
                "ERROR: `netsearch' needs a constraint net id or a default net\n");
      return FALSE;
    }
    net = cnMostRecentlyCreatedNet;
    no = 0;
    args = NULL;
    cdgPrintf(CDG_INFO,
              "INFO: using most recently created net `%s'\n", net->id);

  } else {
    net = hashGet(cdgNets, args[0]);
    if (!net) {
      cdgPrintf(CDG_ERROR,
                "ERROR: `netsearch': constraint net `%s' unknown\n", args[0]);
      return FALSE;
    }
    no--;
    args = &args[1];
  }

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;
  result = (netsearch(net, no, args) >= 0);
  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;

  /* print solution(s) */
  if (hkVerbosity & CDG_SEARCHRESULT) {
    cnPrintParses(net);
  }
  cnPrintInfo(net);

  writeXmlParses(net->parses);
  writeXmlNetInfo(net);

  return result;
}

/* ----------------------------------------------------------------------
   Start or stop the tagger.
*/
Boolean cmdTagger(int no, char **args)
{

  if(1 != no) {
    cdgPrintf(CDG_ERROR, "ERROR: please use `tagger on' of `tagger off'.\n");
    return FALSE;
  }

  if(!strcmp("on", args[0]) && !taggerUp()) {
    taggerStartStop();
  }
  if(!strcmp("off", args[0]) && taggerUp()) {
    taggerStartStop();
  }
  return TRUE;

}

/* ----------------------------------------------------------------------
   Undocumented command for testing
   mode=1: determine base line for accuracy
   mode=2: determine base line for accuracy (TODO: is that correct??)
   mode=3: compare two lexica bit for bit
   mode=4: approve compiled grammar
*/
Boolean cmdTesting(int no, char **args)
{
  int mode;
  Boolean result = TRUE;

  if (no <= 0 || !args[0]) {
    cdgPrintf(CDG_ERROR, "ERROR: wrong number of arguments\n");
    return FALSE;
  }
  if (!sscanf(args[0], "%d", &mode)) {
    cdgPrintf(CDG_ERROR, "ERROR: cannot determine testing mode\n");
    return FALSE;
  }

  switch (mode) {
    case 1:
      tst1(no - 1, &args[1]);
      break;
    case 2:
      tst2(no - 1, &args[1]);
      break;
    case 3:
      tst3(no - 1, &args[1]);
      break;
#ifndef DARWIN
    case 4:
      comApprove(no - 1, &args[1]);
      break;
#endif
    case 5:
      tst5(no - 1, &args[1]);
      break;
    case 7:
      tst7(no - 1, &args[1]);
      break;
    default:
      cdgPrintf(CDG_ERROR, "ERROR: Which test do you want? Mode %d unknown\n",
                mode);
      result = FALSE;
      break;
  }

  return result;
}

/* ----------------------------------------------------------------------
   Try to find a complete solution by transformation.
*/
Boolean cmdFrobbing(int no, char **args)
{
  Boolean result = TRUE;

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  /* set the timelimit for this command */
  if (cdgTimeLimit != 0) {
    timerSetAlarm(cdgTimeLimit);
  }

  result = frobbing(no, args);

  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;

  /* stop the alarm clock on suspicion */
  timerStopAlarm();

  return result;
}

/* ----------------------------------------------------------------------
   Print parses of the specified net.
*/
Boolean cmdPrintParses(int no, char **args)
{
  ConstraintNet net = NULL;
  int i;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdPrintParses: ERROR: argument is NULL\n");
    return FALSE;
  }
  if (no > 0) {
    for (i = 0; i < no && args[i] != NULL; i++) {
      ConstraintNet net = cnFindNet(args[i]);
      if (!net) {
        return FALSE;
      }
    }
  } else if (cnMostRecentlyCreatedNet != NULL) {
    net = cnMostRecentlyCreatedNet;
  } else {
    cdgPrintf(CDG_ERROR,
              "ERROR: `printparses' needs a constraint net id or a default net\n");
    return FALSE;
  }

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  cnPrintParses(net);
  writeXmlParses(net->parses);

  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;
  cnPrintInfo(net);

  return TRUE;
}

/* ----------------------------------------------------------------------
   Delete the specified parse.
*/
Boolean cmdParsedelete(int no, char **args)
{
  Boolean force = FALSE;
  String wordgraph = NULL;
  Parse p = NULL;
  List l, m;
  int i = 0;
  int n = 0;
  String usage = "Usage: parsedelete parse1 parse2 ...\n"
    "       parsedelete -w wordgraph [-f]\n";

  /* no arguments on the command line */
  if (no == 0) {
    l = hashListOfKeys(cdgParses);
    for(m = l; m != NULL; m = m->next) {
      String id = m->item;
      Parse p = hashRemove(cdgParses, id);
      parseDelete(p);
    }
    listDelete(m);
    return TRUE;
  }

  if (strcmp(args[0], "-w") == 0) {

    /* use the wordgraph variant */
    if (no != 2) {
      cdgPrintf(CDG_ERROR, usage);
      return FALSE;
    }
    if (no == 3 && strcmp(args[2], "-f") == 0) {
      force = TRUE;
    }
    wordgraph = strRegister(args[1]);

    for(l = m = hashListOfKeys(cdgParses);
        l != NULL; l = l->next) {
      p = hashGet(cdgParses, l->item);
      if (!p->latticeId || p->latticeId != wordgraph) {
        continue;
      }

      /* keep parses made from annos */
      if (p->searchStrategy != NULL &&
          strcmp(p->searchStrategy, "anno2parse") == 0) {
        if (force) {
          cdgPrintf(CDG_INFO, "INFO: forcing deletion of `%s'\n", p->id);
        } else {
          continue;
        }
      }
      hashRemove(cdgParses, l->item);
      parseDelete(p);
      n++;
      if (parseMostRecentlyCreatedParse == p) {
        parseMostRecentlyCreatedParse = NULL;
      }
    }
  } else {

    /* use the parse list */
    for (i = 0; i < no && args[i]; i++) {
      Parse p = hashGet(cdgParses, args[i]);
      if(p) {
        hashRemove(cdgParses, args[i]);
        parseDelete(p);
        n++;
        if (parseMostRecentlyCreatedParse == p) {
          parseMostRecentlyCreatedParse = NULL;
        }
      }
    }
  }
  cdgPrintf(CDG_INFO, "INFO: deleted %d parse(s)\n", n);
  
  return (n > 0);
}

/* ----------------------------------------------------------------------
   Delete constraint nets.

   If no argument is given, delete all known constraint nets.
   Otherwise, delete all named nets.
*/
Boolean cmdNetdelete(int no, char **args)
{
  int i;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdNetdelete: ERROR: argument is NULL\n");
    return FALSE;
  }

  if(0 == no) {

    /* delete all nets */
    List l, m;
    l = hashListOfKeys(cdgNets);
    for(m = l; m != NULL; m = m->next) {
      String id = m->item;
      ConstraintNet net = hashRemove(cdgNets, id);
      cnDelete(net);
    }
    listDelete(m);
    cnMostRecentlyCreatedNet = NULL;

    l = hashListOfKeys(cdgProblems);
    for(m = l; m != NULL; m = m->next) {
      String id = m->item;
      Problem P = hashRemove(cdgProblems, id);
      deleteProblem(P);
    }
    listDelete(m);
    
  } else {
    for (i = 0; i < no && args[i] != NULL; i++) {
      ConstraintNet net = hashGet(cdgNets, args[i]);
      if (!net) {
        cdgPrintf(CDG_ERROR, "ERROR: can't find a net `%s'\n", args[i]);
        return FALSE;
      }
      if(net == cnMostRecentlyCreatedNet) {
        cnMostRecentlyCreatedNet = NULL;
      }
      cnDelete(net);
      hashRemove(cdgNets, args[i]);
    }
  } 

  return FALSE;
}

/* ----------------------------------------------------------------------
   Print edges of the specified constraint net.
*/
Boolean cmdEdges(int no, char **args)
{
  int i;
  ConstraintNet net = NULL;
  ConstraintEdge e;
  ConstraintNode start = NULL;
  ConstraintNode stop = NULL;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdEdges: ERROR: argument is NULL\n");
    return FALSE;
  }
  if (no < 1 || no > 3) {
    cdgPrintf(CDG_ERROR,
              "ERROR: `edges' needs a net id and optionally two node indices\n");
    return FALSE;
  }

  net = hashGet(cdgNets, args[0]);
  if (!net) {
    cdgPrintf(CDG_ERROR,
              "ERROR: `edges': no such constraint net `%s'\n", args[0]);
    return FALSE;
  }

  if (!net->edges) {
    cdgPrintf(CDG_ERROR, "ERROR: no edges were built for net `%s'\n", net->id);
    return FALSE;
  }

  if (no > 1 && args[1] != NULL) {
    sscanf(args[1], "%d", &i);
    if (i < 0 || i > vectorSize(net->nodes) - 1) {
      cdgPrintf(CDG_ERROR, "ERROR: invalid start index `%d'\n", i);
      return FALSE;
    }
    start = (ConstraintNode) vectorElement(net->nodes, i);
  }
  if (no > 2 && args[2] != NULL) {
    sscanf(args[2], "%d", &i);
    if (i < 0 || i > vectorSize(net->nodes) - 1) {
      cdgPrintf(CDG_ERROR, "ERROR: invalid stop index `%d'\n", i);
      return FALSE;
    }
    stop = (ConstraintNode) vectorElement(net->nodes, i);
  }
  for (i = 0; i < vectorSize(net->edges); i++) {
    e = (ConstraintEdge) vectorElement(net->edges, i);
    if ((start == NULL || e->start == start)
        && (stop == NULL || e->stop == stop))
      cnPrintEdge(CDG_DEFAULT, e);
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Write the specified constraint net to an XFig file.
*/
Boolean cmdWriteNet(int no, char **args)
{
  ConstraintNet net;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdWriteNet: ERROR: argument is NULL\n");
    return FALSE;
  }

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  switch (no) {
    case 0:
      cdgPrintf(CDG_ERROR,
                "cmdWritenet: ERROR: writenet [net] LaTeX_filename\n");
      return FALSE;
    case 1:
      if (cnMostRecentlyCreatedNet != NULL)
        writeConstraintNet(cnMostRecentlyCreatedNet, args[0]);
      else
        cdgPrintf(CDG_ERROR,
                  "ERROR: `writenet' needs a constraint net id or a default net\n");
      return FALSE;
      break;
    default:
      net = hashGet(cdgNets, args[0]);
      if(net) {
        writeConstraintNet(net, args[1]);
      } else {
        cdgPrintf(CDG_ERROR, "ERROR: cannot find a net '%s'\n", args[0]);
        return FALSE;
      }
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Write best Parse of a net to disk as an annotation.
*/
Boolean cmdWriteAnno(int no, char **args) {
  ConstraintNet net;
  Parse best = NULL;
  List l;
  AnnoEntry ae;
  
  if(0 == no) {
    if(cnMostRecentlyCreatedNet) {
      net = cnMostRecentlyCreatedNet;
    } else {
      cdgPrintf(CDG_ERROR,
                "ERROR: `writeannotation' needs a constraint net id or a default net\n");
      return FALSE;
    }
  } else {
    net = hashGet(cdgNets, args[0]);
  }

  if(!net) {
    cdgPrintf(CDG_ERROR, "ERROR: cannot find a net `%s'\n", args[0]);
    return FALSE;
  }

  /* find best Parse */
  for(l = net->parses; l != NULL; l = l->next) {
    Parse p = l->item;
    if(!best || bCompare(p->badness, best->badness)) {
      best = p;
    }
  }

  if(!best) {
    cdgPrintf(CDG_ERROR, "ERROR: no parse in net `%s'\n", net->id);
    return FALSE;
  }

  ae = parse2annotation(best);
  writeAnnotation(ae, best->latticeId);
  freeAnnoEntry(ae);
  
  return TRUE;
}

/* ----------------------------------------------------------------------
   Write the specified parses to XFig and LaTeX files.
*/
Boolean cmdWriteParses(int no, char **args)
{
  String main_level = NULL;
  String graph_file = NULL;
  String info_file = NULL;
  ConstraintNet net = NULL;
  Level level = inputGetMainlevel(inputCurrentGrammar);

  if (level) {
    main_level = level->id;
  }

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdWriteParses: ERROR: argument is NULL\n");
    return FALSE;
  }
  switch (no) {
    case 0:
    case 1:
      cdgPrintf(CDG_ERROR,
                "cmdWriteParses: ERROR: writeparses [net] [main_level] Graphic-filename Info-filename\n");
      cdgPrintf(CDG_ERROR,
                "                       the filenames will have appended '-i.fig' and '-i.tex',\n");
      cdgPrintf(CDG_ERROR,
                "                       where i stands for the current solution number.\n");
      return FALSE;
    case 2:
      graph_file = args[0];
      info_file = args[1];
      if (cnMostRecentlyCreatedNet != NULL)
        net = cnMostRecentlyCreatedNet;
      else {
        net = NULL;
        cdgPrintf(CDG_ERROR,
                  "ERROR: `writeparses' needs a constraint net id or a default net\n");
        return FALSE;
      }
      break;
    case 3:
      graph_file = args[1];
      info_file = args[2];
      net = cnFindNet(args[0]);
      if (!net) {
        cdgPrintf(CDG_ERROR, "ERROR: can't find a net `%s'\n", args[0]);
        return FALSE;
      }
      break;
    case 4:
    default:
      main_level = args[1];
      graph_file = args[2];
      info_file = args[3];
      net = cnFindNet(args[0]);
      if (!net) {
        cdgPrintf(CDG_ERROR, "ERROR: can't find a net `%s'\n", args[0]);
        return FALSE;
      }
      break;
  }

  if (net) {
    cdgCtrlCAllowed = TRUE;
    cdgCtrlCTrapped = FALSE;
    writeParses(net->parses, main_level, graph_file, info_file);
    cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Print status information.
*/
Boolean cmdStatus(int no, char **args)
{
  List l;
  Level level = inputGetMainlevel(inputCurrentGrammar);
  HashIterator hi;
  int counter = 0;

  if (no > 0) {
    cdgPrintf(CDG_INFO, "INFO: `status' has no arguments\n");
  }

  cdgPrintf(CDG_DEFAULT, "         constraints: %d\n",
            hashSize(inputCurrentGrammar->constraints));
  cdgPrintf(CDG_DEFAULT, "  level declarations: %d\n",
            listSize(inputCurrentGrammar->levels));
  cdgPrintf(CDG_DEFAULT, "          word forms: %d\n",
            hashSize(inputCurrentGrammar->lexicon));

  counter = 0;
  hi = hashIteratorNew(inputCurrentGrammar->lexicon);
  while ((l = hashIteratorNextValue(hi))) {
    counter += listSize(l);
  }

  hashIteratorDelete(hi);
  cdgPrintf(CDG_DEFAULT, "     lexical entries: %d\n", counter);
  cdgPrintf(CDG_DEFAULT, "          wordgraphs: %d\n",
            listSize(inputCurrentGrammar->lattices));
  cdgPrintf(CDG_DEFAULT, "         annotations: %d\n",
            hashSize(inputCurrentGrammar->annotations));
  cdgPrintf(CDG_DEFAULT, "         hierarchies: %d\n",
            hashSize(inputCurrentGrammar->hierarchies));
  cdgPrintf(CDG_DEFAULT, "          parameters: %d\n",
            listSize(inputCurrentGrammar->parameters));
  cdgPrintf(CDG_DEFAULT, "\n");

  cdgPrintf(CDG_DEFAULT, "              parses: %d\n",
            hashSize(cdgParses));
  cdgPrintf(CDG_DEFAULT, "\n");

  cdgPrintf(CDG_DEFAULT, "     constraint nets: %d\n",
            hashSize(cdgNets));
  cdgPrintf(CDG_DEFAULT, "\n");
  cdgPrintf(CDG_DEFAULT, "           verbosity: %ld\n", hkVerbosity);
  cdgPrintf(CDG_DEFAULT, " show deleted values: %s\n",
            cnShowDeletedFlag ? "yes" : "no");
  cdgPrintf(CDG_DEFAULT, "      use statistics: %s\n",
            statUseStatisticsFlag ? "yes" : "no");
  cdgPrintf(CDG_DEFAULT, " search modifies net: %s\n",
            nsSearchModifiesNetFlag ? "yes" : "no");
  cdgPrintf(CDG_DEFAULT, "automatic comparison: %s\n",
            nsAutoCompare ? "yes" : "no");
  cdgPrintf(CDG_DEFAULT, "       normalization: %s\n",
            nsNormalization == NSNOff ? "off" : nsNormalization ==
            NSNLinear ? "linear" : nsNormalization ==
            NSNSquare ? "square" : nsNormalization ==
            NSNDepth ? "depth" : nsNormalization ==
            NSNBreadth ? "breadth" : "unknown");
  cdgPrintf(CDG_DEFAULT, "   subsumes warnings: %s\n",
            evalSloppySubsumesWarnings ? "sloppy" : "full");
  cdgPrintf(CDG_DEFAULT, "         build edges: %s\n",
            cnEdgesFlag == cnEdgesOff ? "no" : "yes");
  cdgPrintf(CDG_DEFAULT, "unary pruning factor: %f\n", cnUnaryPruningFraction);
  cdgPrintf(CDG_DEFAULT, "             caching: %s\n",
            scUseCache ? "yes" : "no");

  cdgPrintf(CDG_DEFAULT, "compact level values: %s\n",
            lgCompactLVs ? "yes" : "no");
  cdgPrintf(CDG_DEFAULT, "   evaluation method: %s\n",
            evalEvaluationMethod ==
            EMTCompiled ? "compiled" : evalEvaluationMethod ==
            EMTInterpreted ? "interpreted" : "invalid");
  cdgPrintf(CDG_DEFAULT, "   peek value method: %s\n",
            evalPeekValueMethod ==
            EMTCompiled ? "compiled" : evalPeekValueMethod ==
            EMTInterpreted ? "interpreted" : "invalid");
  cdgPrintf(CDG_DEFAULT, "    acoustics method: %s\n",
            vmAcoustics == vmAcOff ? "off" : vmAcoustics ==
            vmAcLinear ? "linear" : vmAcoustics ==
            vmAcScaled ? "scaled" : "invalid");

  cdgPrintf(CDG_DEFAULT, "          sort nodes: %s\n",
            cnSortNodesMethod == 0 ? "no" :
            cnSortNodesMethod == 1 ? "By level priority" :
            cnSortNodesMethod == 2 ? "By increasing domain size" : "unknown");

  cdgPrintf(CDG_DEFAULT, "          main level: %s\n",
            level ? level->id : "<undefined>");

  cdgPrintf(CDG_DEFAULT, "     #shared strings: %d\n", strStoreSize());

#if 0
#ifndef __CYGWIN__
  {
    struct rlimit cpu;
    struct rlimit mem;
    struct rusage ru;

    if (getrusage(RUSAGE_SELF, &ru) ||
        getrlimit(RLIMIT_CPU, &cpu) || getrlimit(RLIMIT_AS, &mem)) {
      cdgPrintf(CDG_DEFAULT,
                "WARNING: can't get rlimit or rusage: %s\n", strerror(errno));
    } else {
      /* 
       * I'm not sure how to compute current memory usage. ru.ru_maxrss gives the ``maximum
       * resident set size'' in pages. Currently it doesn't work. */
      cdgPrintf(CDG_DEFAULT,
                "           cpu usage: %d.%06d sec ",
                ru.ru_utime.tv_sec + ru.ru_stime.tv_sec,
                ru.ru_utime.tv_usec + ru.ru_stime.tv_usec);
      if (cpu.rlim_cur == RLIM_INFINITY) {
        cdgPrintf(CDG_DEFAULT, "(unlimited)\n");
      } else {
        cdgPrintf(CDG_DEFAULT, "of %d sec\n", cpu.rlim_cur);
      }
      cdgPrintf(CDG_DEFAULT,
                "        memory usage: %d bytes ",
                ru.ru_maxrss * getpagesize());
      if (mem.rlim_cur == RLIM_INFINITY) {
        cdgPrintf(CDG_DEFAULT, "(unlimited)\n");
      } else {
        cdgPrintf(CDG_DEFAULT, "of %d bytes\n", mem.rlim_cur);
      }
      cdgPrintf(CDG_DEFAULT, "\n");
    }
  }
#endif
#endif

  cdgPrintf(CDG_DEFAULT, "   %2d loaded file(s):\n",
            listSize(inputCurrentGrammar->files));
  for (l = inputCurrentGrammar->files; l != NULL; l = l->next) {
    cdgPrintf(CDG_DEFAULT, "                       %s\n", (String) l->item);
  }

  cdgPrintf(CDG_DEFAULT, "\n");

#if 0
  cdgPrintf(CDG_DEFAULT, "CDG_HINT         %s\n",
            hkVerbosity & CDG_HINT ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_INFO         %s\n",
            hkVerbosity & CDG_INFO ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_WARNING      %s\n",
            hkVerbosity & CDG_WARNING ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_PROLOG       %s\n",
            hkVerbosity & CDG_PROLOG ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_EVAL         %s\n",
            hkVerbosity & CDG_EVAL ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_SEARCHRESULT %s\n",
            hkVerbosity & CDG_SEARCHRESULT ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_PROFILE      %s\n",
            hkVerbosity & CDG_PROFILE ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_HOOK         %s\n",
            hkVerbosity & CDG_HOOK ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_ERROR        %s\n",
            hkVerbosity & CDG_ERROR ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_DEBUG        %s\n",
            hkVerbosity & CDG_DEBUG ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_DEFAULT      %s\n",
            hkVerbosity & CDG_DEFAULT ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_PROGRESS     %s\n",
            hkVerbosity & CDG_PROGRESS ? "on" : "off");
  cdgPrintf(CDG_DEFAULT, "CDG_XML          %s\n",
            hkVerbosity & CDG_XML ? "on" : "off");
#endif

  /* print out the symbol table */
  {
    String name;
    List names = setReturnListOfNames();
    List l;

    cdgPrintf(CDG_DEFAULT, "\nsymbol table:\n");

    for (l = names; l; l = l->next) {
      name = l->item;
      cdgPrintf(CDG_DEFAULT, "   ");
      setPrint(CDG_DEFAULT, name);
    }
    listDelete(names);
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Discard all loaded and computed information.
*/
Boolean cmdReset(int no, char **args)
{

  /* just to be sure */
  if (no > 0) {
    cdgPrintf(CDG_ERROR, "ERROR: `reset' takes no arguments.\n");
    return FALSE;
  }

  cdgDeleteComputed();
  
  /* free loaded items */
  inputFreeInput(inputCurrentGrammar);
  inputCurrentGrammar = inputNewInput();

  /* A lot changed since we were last in London... */
  inputCacheInput(inputCurrentGrammar);

  return TRUE;
}

/* ----------------------------------------------------------------------
   Load one or more files.
*/
Boolean cmdLoad(int no, char **args)
{
  int i;
  Boolean success = FALSE;
  Boolean result = TRUE;
  Input meanResult = inputNewInput();
  
  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdLoad: ERROR: argument is NULL\n");
    return FALSE;
  }

  /* load all files */
  for (i = 0; i < no && args[i] != NULL; i++) {

    result = inputLoad(args[i]);
    if (!result)
      break;

    if (parseNoOfErrors == 0) {
      if (strlen(parseCurrentFile) >= 3 &&
          strcmp(&(parseCurrentFile[strlen(parseCurrentFile) - 3]),
                 ".so") == 0) {
        cdgPrintf(CDG_INFO, "INFO: hooking compiled constraints\n");
        evalEvaluationMethod = EMTCompiled;
      } else {
        cdgPrintf(CDG_INFO, "INFO: file `%s' loaded: ", parseCurrentFile);
        cdgPrintf(CDG_INFO, "%d/%d/%d/%d/%d/%d/%d/%d\n",
                  hashSize(parseResult->constraints),
                  listSize(parseResult->levels),
                  hashSize(parseResult->lexicon),
                  listSize(parseResult->lattices),
                  hashSize(parseResult->annotations),
                  hashSize(parseResult->hierarchies),
                  listSize(parseResult->parameters),
                  hashSize(parseResult->maps));
      }
      if (parseNoOfWarnings != 0) {
        cdgPrintf(CDG_INFO, "INFO: %d warning(s)\n", parseNoOfWarnings);
      }

      /* combine already loaded items and just loaded items */
      mergeInput(meanResult, parseResult);

      /* All the stuff except parseResult is free'd in mergeInput. */
      memFree(parseResult);
      success = TRUE;
    } else {
      /* There were errors. Free and ignore the new items. */
      cdgPrintf(CDG_INFO, "ERROR: %d error(s), input ignored\n",
                parseNoOfErrors);
      inputFreeInput(parseResult);
      parseResult = NULL;
      result = FALSE;
      break;
    }
    parseResult = NULL;
  }

  /* we loaded at least one file */
  if (success) {
    
    cdgDeleteComputed();
    
    /* promote new data to current data */
    mergeInput(inputCurrentGrammar, meanResult);
    cdgExecPragmas(meanResult->pragmas);
    inputCacheInput(inputCurrentGrammar);
    memFree(meanResult);
  }
  
  return result;
}

/* ----------------------------------------------------------------------
   Print the General Public License.
*/
Boolean cmdLicense(int no, char **args)
{
  cdgPrintf(CDG_DEFAULT,
            "\n             GNU GENERAL PUBLIC LICENSE\n"
            "                  Version 2, June 1991\n"
            "\n"
            " Copyright (C) 1989, 1991 Free Software Foundation, Inc.\n"
            "                          675 Mass Ave, Cambridge, MA 02139, USA\n"
            " Everyone is permitted to copy and distribute verbatim copies\n"
            " of this license document, but changing it is not allowed.\n"
            "\n"
            "                       Preamble\n"
            "\n"
            "  The licenses for most software are designed to take away your\n"
            "freedom to share and change it.  By contrast, the GNU General Public\n"
            "License is intended to guarantee your freedom to share and change free\n"
            "software--to make sure the software is free for all its users.  This\n"
            "General Public License applies to most of the Free Software\n"
            "Foundation's software and to any other program whose authors commit to\n"
            "using it.  (Some other Free Software Foundation software is covered by\n"
            "the GNU Library General Public License instead.)  You can apply it to\n"
            "your programs, too.\n"
            "\n"
            "  When we speak of free software, we are referring to freedom, not\n"
            "price.  Our General Public Licenses are designed to make sure that you\n"
            "have the freedom to distribute copies of free software (and charge for\n"
            "this service if you wish), that you receive source code or can get it\n"
            "if you want it, that you can change the software or use pieces of it\n"
            "in new free programs; and that you know you can do these things.\n"
            "\n"
            "  To protect your rights, we need to make restrictions that forbid\n"
            "anyone to deny you these rights or to ask you to surrender the rights.\n"
            "These restrictions translate to certain responsibilities for you if you\n"
            "distribute copies of the software, or if you modify it.\n"
            "\n"
            "  For example, if you distribute copies of such a program, whether\n"
            "gratis or for a fee, you must give the recipients all the rights that\n"
            "you have.  You must make sure that they, too, receive or can get the\n"
            "source code.  And you must show them these terms so they know their\n"
            "rights.\n"
            "\n"
            "  We protect your rights with two steps: (1) copyright the software, and\n"
            "(2) offer you this license which gives you legal permission to copy,\n"
            "distribute and/or modify the software.\n"
            "\n"
            "  Also, for each author's protection and ours, we want to make certain\n"
            "that everyone understands that there is no warranty for this free\n"
            "software.  If the software is modified by someone else and passed on, we\n"
            "want its recipients to know that what they have is not the original, so\n"
            "that any problems introduced by others will not reflect on the original\n"
            "authors' reputations.\n"
            "\n"
            "  Finally, any free program is threatened constantly by software\n"
            "patents.  We wish to avoid the danger that redistributors of a free\n"
            "program will individually obtain patent licenses, in effect making the\n"
            "program proprietary.  To prevent this, we have made it clear that any\n"
            "patent must be licensed for everyone's free use or not licensed at all.\n"
            "\n"
            "  The precise terms and conditions for copying, distribution and\n"
            "modification follow.\n"
            "\n"
            "               GNU GENERAL PUBLIC LICENSE\n"
            "   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION\n"
            "\n"
            "  0. This License applies to any program or other work which contains\n"
            "a notice placed by the copyright holder saying it may be distributed\n"
            "under the terms of this General Public License.  The \"Program\", below,\n"
            "refers to any such program or work, and a \"work based on the Program\"\n"
            "means either the Program or any derivative work under copyright law:\n"
            "that is to say, a work containing the Program or a portion of it,\n"
            "either verbatim or with modifications and/or translated into another\n"
            "language.  (Hereinafter, translation is included without limitation in\n"
            "the term \"modification\".)  Each licensee is addressed as \"you\".\n"
            "\n"
            "Activities other than copying, distribution and modification are not\n"
            "covered by this License; they are outside its scope.  The act of\n"
            "running the Program is not restricted, and the output from the Program\n"
            "is covered only if its contents constitute a work based on the\n"
            "Program (independent of having been made by running the Program).\n"
            "Whether that is true depends on what the Program does.\n"
            "\n"
            "  1. You may copy and distribute verbatim copies of the Program's\n"
            "source code as you receive it, in any medium, provided that you\n"
            "conspicuously and appropriately publish on each copy an appropriate\n"
            "copyright notice and disclaimer of warranty; keep intact all the\n"
            "notices that refer to this License and to the absence of any warranty;\n"
            "and give any other recipients of the Program a copy of this License\n"
            "along with the Program.\n"
            "\n"
            "You may charge a fee for the physical act of transferring a copy, and\n"
            "you may at your option offer warranty protection in exchange for a fee.\n"
            "\n"
            "  2. You may modify your copy or copies of the Program or any portion\n"
            "of it, thus forming a work based on the Program, and copy and\n"
            "distribute such modifications or work under the terms of Section 1\n"
            "above, provided that you also meet all of these conditions:\n"
            "\n"
            "    a) You must cause the modified files to carry prominent notices\n"
            "    stating that you changed the files and the date of any change.\n"
            "\n"
            "    b) You must cause any work that you distribute or publish, that in\n"
            "    whole or in part contains or is derived from the Program or any\n"
            "    part thereof, to be licensed as a whole at no charge to all third\n"
            "    parties under the terms of this License.\n"
            "\n"
            "    c) If the modified program normally reads commands interactively\n"
            "    when run, you must cause it, when started running for such\n"
            "    interactive use in the most ordinary way, to print or display an\n"
            "    announcement including an appropriate copyright notice and a\n"
            "    notice that there is no warranty (or else, saying that you provide\n"
            "    a warranty) and that users may redistribute the program under\n"
            "    these conditions, and telling the user how to view a copy of this\n"
            "    License.  (Exception: if the Program itself is interactive but\n"
            "    does not normally print such an announcement, your work based on\n"
            "    the Program is not required to print an announcement.)\n"
            "\n"
            "These requirements apply to the modified work as a whole.  If\n"
            "identifiable sections of that work are not derived from the Program,\n"
            "and can be reasonably considered independent and separate works in\n"
            "themselves, then this License, and its terms, do not apply to those\n"
            "sections when you distribute them as separate works.  But when you\n"
            "distribute the same sections as part of a whole which is a work based\n"
            "on the Program, the distribution of the whole must be on the terms of\n"
            "this License, whose permissions for other licensees extend to the\n"
            "entire whole, and thus to each and every part regardless of who wrote it.\n"
            "\n"
            "Thus, it is not the intent of this section to claim rights or contest\n"
            "your rights to work written entirely by you; rather, the intent is to\n"
            "exercise the right to control the distribution of derivative or\n"
            "collective works based on the Program.\n"
            "\n"
            "In addition, mere aggregation of another work not based on the Program\n"
            "with the Program (or with a work based on the Program) on a volume of\n"
            "a storage or distribution medium does not bring the other work under\n"
            "the scope of this License.\n"
            "\n"
            "  3. You may copy and distribute the Program (or a work based on it,\n"
            "under Section 2) in object code or executable form under the terms of\n"
            "Sections 1 and 2 above provided that you also do one of the following:\n"
            "\n"
            "    a) Accompany it with the complete corresponding machine-readable\n"
            "    source code, which must be distributed under the terms of Sections\n"
            "    1 and 2 above on a medium customarily used for software interchange; or,\n"
            "\n"
            "    b) Accompany it with a written offer, valid for at least three\n"
            "    years, to give any third party, for a charge no more than your\n"
            "    cost of physically performing source distribution, a complete\n"
            "    machine-readable copy of the corresponding source code, to be\n"
            "    distributed under the terms of Sections 1 and 2 above on a medium\n"
            "    customarily used for software interchange; or,\n"
            "\n"
            "    c) Accompany it with the information you received as to the offer\n"
            "    to distribute corresponding source code.  (This alternative is\n"
            "    allowed only for noncommercial distribution and only if you\n"
            "    received the program in object code or executable form with such\n"
            "    an offer, in accord with Subsection b above.)\n"
            "\n"
            "The source code for a work means the preferred form of the work for\n"
            "making modifications to it.  For an executable work, complete source\n"
            "code means all the source code for all modules it contains, plus any\n"
            "associated interface definition files, plus the scripts used to\n"
            "control compilation and installation of the executable.  However, as a\n"
            "special exception, the source code distributed need not include\n"
            "anything that is normally distributed (in either source or binary\n"
            "form) with the major components (compiler, kernel, and so on) of the\n"
            "operating system on which the executable runs, unless that component\n"
            "itself accompanies the executable.\n"
            "\n"
            "If distribution of executable or object code is made by offering\n"
            "access to copy from a designated place, then offering equivalent\n"
            "access to copy the source code from the same place counts as\n"
            "distribution of the source code, even though third parties are not\n"
            "compelled to copy the source along with the object code.\n"
            "\n"
            "4. You may not copy, modify, sublicense, or distribute the Program\n"
            "except as expressly provided under this License.  Any attempt\n"
            "otherwise to copy, modify, sublicense or distribute the Program is\n"
            "void, and will automatically terminate your rights under this License.\n"
            "However, parties who have received copies, or rights, from you under\n"
            "this License will not have their licenses terminated so long as such\n"
            "parties remain in full compliance.\n"
            "\n"
            "  5. You are not required to accept this License, since you have not\n"
            "signed it.  However, nothing else grants you permission to modify or\n"
            "distribute the Program or its derivative works.  These actions are\n"
            "prohibited by law if you do not accept this License.  Therefore, by\n"
            "modifying or distributing the Program (or any work based on the\n"
            "Program), you indicate your acceptance of this License to do so, and\n"
            "all its terms and conditions for copying, distributing or modifying\n"
            "the Program or works based on it.\n"
            "\n"
            "  6. Each time you redistribute the Program (or any work based on the\n"
            "Program), the recipient automatically receives a license from the\n"
            "original licensor to copy, distribute or modify the Program subject to\n"
            "these terms and conditions.  You may not impose any further\n"
            "restrictions on the recipients' exercise of the rights granted herein.\n"
            "You are not responsible for enforcing compliance by third parties to\n"
            "this License.\n"
            "\n"
            "  7. If, as a consequence of a court judgment or allegation of patent\n"
            "infringement or for any other reason (not limited to patent issues),\n"
            "conditions are imposed on you (whether by court order, agreement or\n"
            "otherwise) that contradict the conditions of this License, they do not\n"
            "excuse you from the conditions of this License.  If you cannot\n"
            "distribute so as to satisfy simultaneously your obligations under this\n"
            "License and any other pertinent obligations, then as a consequence you\n"
            "may not distribute the Program at all.  For example, if a patent\n"
            "license would not permit royalty-free redistribution of the Program by\n"
            "all those who receive copies directly or indirectly through you, then\n"
            "the only way you could satisfy both it and this License would be to\n"
            "refrain entirely from distribution of the Program.\n"
            "\n"
            "If any portion of this section is held invalid or unenforceable under\n"
            "any particular circumstance, the balance of the section is intended to\n"
            "apply and the section as a whole is intended to apply in other\n"
            "circumstances.\n"
            "\n"
            "It is not the purpose of this section to induce you to infringe any\n"
            "patents or other property right claims or to contest validity of any\n"
            "such claims; this section has the sole purpose of protecting the\n"
            "integrity of the free software distribution system, which is\n"
            "implemented by public license practices.  Many people have made\n"
            "generous contributions to the wide range of software distributed\n"
            "through that system in reliance on consistent application of that\n"
            "system; it is up to the author/donor to decide if he or she is willing\n"
            "to distribute software through any other system and a licensee cannot\n"
            "impose that choice.\n"
            "\n"
            "This section is intended to make thoroughly clear what is believed to\n"
            "be a consequence of the rest of this License.\n"
            "\n"
            "  8. If the distribution and/or use of the Program is restricted in\n"
            "certain countries either by patents or by copyrighted interfaces, the\n"
            "original copyright holder who places the Program under this License\n"
            "may add an explicit geographical distribution limitation excluding\n"
            "those countries, so that distribution is permitted only in or among\n"
            "countries not thus excluded.  In such case, this License incorporates\n"
            "the limitation as if written in the body of this License.\n"
            "\n"
            "  9. The Free Software Foundation may publish revised and/or new versions\n"
            "of the General Public License from time to time.  Such new versions will\n"
            "be similar in spirit to the present version, but may differ in detail to\n"
            "address new problems or concerns.\n"
            "\n"
            "Each version is given a distinguishing version number.  If the Program\n"
            "specifies a version number of this License which applies to it and \"any\n"
            "later version\", you have the option of following the terms and conditions\n"
            "either of that version or of any later version published by the Free\n"
            "Software Foundation.  If the Program does not specify a version number of\n"
            "this License, you may choose any version ever published by the Free Software\n"
            "Foundation.\n"
            "\n"
            "  10. If you wish to incorporate parts of the Program into other free\n"
            "programs whose distribution conditions are different, write to the author\n"
            "to ask for permission.  For software which is copyrighted by the Free\n"
            "Software Foundation, write to the Free Software Foundation; we sometimes\n"
            "make exceptions for this.  Our decision will be guided by the two goals\n"
            "of preserving the free status of all derivatives of our free software and\n"
            "of promoting the sharing and reuse of software generally.\n"
            "\n"
            "                       NO WARRANTY\n"
            "\n"
            "  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY\n"
            "FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN\n"
            "OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES\n"
            "PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED\n"
            "OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n"
            "MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS\n"
            "TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE\n"
            "PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,\n"
            "REPAIR OR CORRECTION.\n"
            "\n"
            "  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING\n"
            "WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR\n"
            "REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,\n"
            "INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING\n"
            "OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED\n"
            "TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY\n"
            "YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER\n"
            "PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE\n"
            "POSSIBILITY OF SUCH DAMAGES.\n"
            "\n" "                   END OF TERMS AND CONDITIONS\n" "\n");
  return TRUE;
}

/* ----------------------------------------------------------------------
   Print the CDG version.
*/
Boolean cmdVersion(int no, char **args)
{
  cdgPrintf(CDG_DEFAULT, STARTUPMSG);

  return TRUE;
}

/* ----------------------------------------------------------------------
   Build a new constraint net.
*/
Boolean cmdNewnet(int no, char **args)
{
  ConstraintNet net;
  String id;
  Lattice lat;
  Timer profileTime;
  Boolean result = TRUE;
  unsigned long duration;
  Boolean buildLVs = TRUE;
  
  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdNewnet: ERROR: argument is NULL\n");
    return FALSE;
  }
  if (no < 1 || (id = args[0]) == NULL) {
    if (inputMostRecentlyCreatedLattice == NULL) {
      cdgPrintf(CDG_ERROR,
                "ERROR: `newnet' needs a word graph id or a default word graph.\n");
      return FALSE;
    }
    cdgPrintf(CDG_INFO,
              "INFO: using most recently created word graph `%s'\n",
              inputMostRecentlyCreatedLattice->id);
    lat = inputMostRecentlyCreatedLattice;
  } else {

    if (!(lat = findLattice(id))) {
      cdgPrintf(CDG_ERROR, "ERROR: `newnet': wordgraph `%s' unknown\n", id);
      return FALSE;
    }
  }

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  profileTime = timerNew();
  net = cnBuild(lat, buildLVs);
  duration = timerElapsed(profileTime);
  timerFree(profileTime);

  if (net == NULL) {
    if (!cdgCtrlCTrapped) {
      cdgPrintf(CDG_WARNING, "ERROR: invalid net\n");
    }
    result = FALSE;
  } else {
    cdgPrintf(CDG_PROFILE, "PROFILE: newnet took %lums\n", duration);
    cnPrintInfo(net);
    writeXmlStatistics(1,"newnet","general statistics on newnet",
    "totalTime", (double) duration,
    NULL);
    writeXmlNetInfo(net);
  }

  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;

  return result;
}

/* ----------------------------------------------------------------------
   Restore the specified net to original state.
*/
Boolean cmdRenewnet(int no, char **args)
{
  ConstraintNet net;
  int i;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdRenewNet: ERROR: argument is NULL\n");
    return FALSE;
  }

  /* get net */
  net = NULL;
  if (no > 0) {
    for (i = 0; i < no && args[i] != NULL; i++) {
      net = cnFindNet(args[i]);
      if (!net) {
        cdgPrintf(CDG_INFO, "INFO: can't find a net `%s'\n", args[i]);
        return FALSE;
      }
    }
  } else {
    net = cnMostRecentlyCreatedNet;
  }

  if (!net) {
    cdgPrintf(CDG_ERROR,
              "ERROR: `renewnet' needs a constraint net id or a default net\n");
    return FALSE;
  } else {
    cnRenew(net);
    cnPrintInfo(net);
    writeXmlNetInfo(net);
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Build a new wordgraph.
*/
Boolean cmdInputwordgraph(int no, char **args)
{
  Lattice lat;
  Arc arc;
  int i;
  static int counter = 0;
  List typed = NULL;
  List tokenized = NULL;
  List l;
  
  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdInputwordgraph: ERROR: argument is NULL\n");
    return FALSE;
  }

  for (i = 0; i < no && args[i] != NULL; i++) {
    typed = listAppendElement(typed, strRegister(args[i]));
  }

  /* If the user has requested it, pass all arguments through a tokenizer
     to turn e.g. "Stop!" into " Stop ! " (four arcs). */
  if(cdgTokenizer) {
    Boolean ok = TRUE;
    char inputname[] = "/tmp/cdg-XXXXXX";
    char outputname[] = "/tmp/cdg-XXXXXX";
    int indescriptor;
    int outdescriptor;
    FILE *input;
    FILE *output;
    List l;

    /* create temp files */
    indescriptor = mkstemp(inputname);
    if(NULL == (input = fdopen(indescriptor, "w"))) {
      cdgPrintf(CDG_ERROR, "ERROR: Can't write tokenizer input file `%s': %s!\n",
                inputname, strerror(errno));
      unlink(inputname);
      close(indescriptor);
      ok = FALSE;
    }
    outdescriptor = mkstemp(outputname);
    if(NULL == (output = fdopen(outdescriptor, "r"))) {
      cdgPrintf(CDG_ERROR, "ERROR: Can't read tokenizer output file: %s!\n",
                strerror(errno));
      unlink(outputname);
      close(outdescriptor);
      ok = FALSE;
    }

    /* call external tokenizer */
    if(ok) {
      String call;
      for(l = typed; l != NULL; l = l->next) {
        fprintf(input, "%s ", (String)l->item);
      }
      fflush(input);
      call = strPrintf("%s < %s > %s", cdgTokenizer, inputname, outputname); 
      if(system(call)) {
        cdgPrintf(CDG_ERROR, "ERROR: Can't call tokenizer - using arguments literally,\n");
        ok = FALSE;
      }
    }

    /* read lines of output */
    if(ok) {
      char buffer[1024];
      tokenized = NULL;
      while(fgets(buffer, 1024, output)) {
        int newline = strcspn(buffer, "\n");

        /* empty line signals end of tokenizer output */
        if(0 == newline) {
          break;
        }
        
        /* No newline after 1024 characters? That spells trouble. */
        if(newline == strlen(buffer)) {
          cdgPrintf(CDG_ERROR, "ERROR: tokenizer output too wide!\n");
          ok = FALSE;
          break;
        }
        buffer[newline] = '\0';
        tokenized = listAppendElement(tokenized, strRegister(buffer));
      }
    }

    /* clean up */
    unlink(inputname);
    unlink(outputname);
    close(indescriptor);
    close(outdescriptor);

    if(ok) {
      listDelete(typed);
    } else {
      listDelete(tokenized);
      tokenized = typed;
    }
  }

  else {
    /* use args literally */
    tokenized = typed;
  }

  /* need at least one token */
  if (!tokenized) {
    cdgPrintf(CDG_ERROR,
              "`inputwordgraph' needs at least one word as an argument.\n");
    return FALSE;
  }

  lat = (Lattice) memMalloc(sizeof (LatticeStruct));
  lat->id = strPrintf("wordgraph%d", counter++);
  lat->filename = strRegister("*USER*");
  lat->lineNo = 0;

  lat->arcs = NULL;
  i = 0;
  for (l = tokenized; l; l = l->next) {
    arc = newArc(i, i + 1, i,
                 strRegister(l->item), 0.0, NULL, 0, 0);
    lat->arcs = listAppendElement(lat->arcs, arc);
    i++;
  }
  listDelete(tokenized);

  inputScanLattice(lat);

  inputCurrentGrammar->lattices =
    listAppendElement(inputCurrentGrammar->lattices, lat);
  cdgPrintf(CDG_INFO, "INFO: wordgraph id: %s, #arcs: %d\n", lat->id, i);
  inputMostRecentlyCreatedLattice = lat;
  inputCurrentGrammar->xcdgwordgraphsToDate = FALSE;

  return TRUE;
}

/* ----------------------------------------------------------------------
   Print the distance matrix of the specified net.
*/
Boolean cmdDistance(int no, char **args)
{
  LexemGraph lg;
  LexemNode a, b;
  ConstraintNet net;
  int i, j, k;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "cmdDistance: ERROR: argument is NULL\n");
    return FALSE;
  }

  if (no < 1 || args[0] == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: `distance' needs a constraint net id\n");
    return FALSE;
  }

  net = cnFindNet(args[0]);
  if (!net) {
    cdgPrintf(CDG_ERROR,
              "ERROR: `distance': constraint net `%s' unknown\n", args[0]);
    return FALSE;
  }

  lg = net->lexemgraph;

  cdgPrintf(CDG_DEFAULT, "       ");
  for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
    a = ((GraphemNode) vectorElement(lg->graphemnodes, i))->lexemes->item;
    cdgPrintf(CDG_DEFAULT, "%6.6s ", a->lexem->word);
  }
  cdgPrintf(CDG_DEFAULT, "\n\n");

  for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
    a = ((GraphemNode) vectorElement(lg->graphemnodes, i))->lexemes->item;
    cdgPrintf(CDG_DEFAULT, "%6.6s  ", a->lexem->word);
    for (j = 0; j < vectorSize(lg->graphemnodes); j++) {
      b = ((GraphemNode) vectorElement(lg->graphemnodes, j))->lexemes->item;
      k = lgDistanceOfNodes(lg, a, b);
      if (k == 0)
        cdgPrintf(CDG_DEFAULT, "       ");
      else
        cdgPrintf(CDG_DEFAULT, "%+4.4d  ", k);
    }
    cdgPrintf(CDG_DEFAULT, "\n");
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Report whether the specified constraints are suitable
   for processing underspecified subordinations.
*/
Boolean cmdNonSpecCompatible(int no, char **args)
{
  List constraints, l;
  Constraint c, foundConstraint;
  char usage[] = "usage: nonspeccompatible [constraint]\n";

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: argument is NULL\n");
    return FALSE;
  }

  if (no > 1) {
    cdgPrintf(CDG_ERROR, "ERROR: wrong no of parameters\n" "       %s", usage);
    return FALSE;
  }

  constraints = incGetNonSpecIncompatibles();

  if (no == 0) {
    if (listSize(constraints) == 0) {
      cdgPrintf(CDG_DEFAULT, "all %d constraints are compatible.\n",
                hashSize(inputCurrentGrammar->constraints));
    } else {
      for (l = constraints; l; l = l->next) {
        c = l->item;
        cdgPrintf(CDG_DEFAULT, "%s:%d: error in %s\n",
                  c->filename, c->lineNo, c->id);
      }
      cdgPrintf(CDG_INFO, "INFO: %d incompatible constraints out of %d:\n",
                listSize(constraints),
                hashSize(inputCurrentGrammar->constraints));
    }
  } else {
    /* find constraint */
    foundConstraint = hashGet(inputCurrentGrammar->constraints, args[0]);
    if (!foundConstraint) {
      cdgPrintf(CDG_ERROR, "ERROR: unknown constraint `%s'\n", args[0]);
      return FALSE;
    } else {
      cdgPrintf(CDG_DEFAULT, "INFO: constraint `%s' is %s.\n",
                args[0],
                (listIndex(constraints, foundConstraint)) ? "incompatible" :
                "ok");
    }
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Parsing problems are created incrementally over prefixes 
   of the word graph and solved individually, with the solution of one
   problem guiding the search in the next one.
*/
Boolean cmdIncrementalCompletion(int no, char **args)
{
  String id;
  Lattice lat = NULL;
  Timer profileTime;
  CnEdgesType oldcnEdgesFlag;
  Boolean oldscUseCache;
  Boolean oldnsSearchModifiesNetFlag;
  Boolean result;

  /* interactive method is called with `interactive' as wordgraph */
  if (no == 0) {
    cdgPrintf(CDG_ERROR, "ERROR: incrementalcompletion word_graph\n");
    cdgPrintf(CDG_ERROR,
              "       use interactive as word_graph to run the interactive method\n");
    return FALSE;
  }

  id = args[0];
  no--;
  args = &args[1];
  if (strcmp(id, "interactive") == 0) {
    lat = NULL;
    cdgPrintf(CDG_INFO, "INFO: Running in interactive mode\n");
  } else {
    if (!(lat = findLattice(id))) {
      cdgPrintf(CDG_ERROR,
                "ERROR: `incrementalcompletion': wordgraph `%s' unknown\n", id);
      return FALSE;
    }
  }

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;
  oldcnEdgesFlag = cnEdgesFlag;
  cnEdgesFlag = cnEdgesOff;
  oldscUseCache = scUseCache;
  scUseCache = FALSE;
  oldnsSearchModifiesNetFlag = nsSearchModifiesNetFlag;
  nsSearchModifiesNetFlag = FALSE;
  profileTime = timerNew();
  result = (icIncrementalCompletion(lat, icParams, no, args) >= 0);
  cnEdgesFlag = oldcnEdgesFlag;
  scUseCache = oldscUseCache;
  nsSearchModifiesNetFlag = oldnsSearchModifiesNetFlag;

  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;

  cdgPrintf(CDG_PROFILE,
            "PROFILE: incrementalcompletion took %lums\n",
            timerElapsed(profileTime));
  timerFree(profileTime);

  return result;
}

/* ----------------------------------------------------------------------
   Perform a complete search on increasing prefixes of the
   word graph.
*/
Boolean cmdISearch(int no, char **args)
{
  String id = NULL;
  ConstraintNet net;
  int i;
  Lattice lattice = NULL;
  Timer profileTime;

  int agSize = 1000;
  Number agThreshold = -1.0;
  Number nonSpecThreshold = 0.0;

  char usage[] =
    "usage: isearch <wordgraph>\n"
    "       <agenda-size>       (default 1000)\n"
    "       <agenda-threshold>  (default 0.0)\n"
    "       <nonspec-threshold> (default 0.0)\n";

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: argument is NULL\n");
    return FALSE;
  }

  if (no < 1 || (id = args[0]) == NULL) {
    if (inputMostRecentlyCreatedLattice == NULL) {
      cdgPrintf(CDG_ERROR,
                "ERROR: `isearch' needs a word graph id or a default word graph.\n%s",
                usage);
      return FALSE;
    }
    cdgPrintf(CDG_INFO,
              "INFO: using most recently created word graph `%s'\n",
              inputMostRecentlyCreatedLattice->id);
    lattice = inputMostRecentlyCreatedLattice;
  }

  if (no < 1 || no > 4) {
    cdgPrintf(CDG_ERROR, "ERROR: wrong no of parameters\n%s", usage);
    return FALSE;
  }

  for (i = 0; i < no; i++) {
    switch (i) {
      case 0:
        id = args[i];
        if (!(lattice = findLattice(id))) {
          cdgPrintf(CDG_ERROR, "ERROR: wordgraph '%s' unknown\n%s", id, usage);
          return FALSE;
        }
        break;
      case 1:                          /* size */
        agSize = atoi(args[i]);
        break;
      case 2:                          /* threshold */
        sscanf(args[i], "%lf", &agThreshold);
        break;
      case 3:                          /* nonspec-threshold */
        sscanf(args[i], "%lf", &nonSpecThreshold);
        break;
      default:
        cdgPrintf(CDG_ERROR, "ERROR: unknown option '%s'\n%s", args[i], usage);
        return FALSE;
    }
  }

  cdgPrintf(CDG_INFO,
            "INFO: agenda-size=%d\n"
            "      agenda-threshold=%4.3e\n"
            "      nonSpec-threshold=%4.3e\n",
            agSize, agThreshold, nonSpecThreshold);

  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  profileTime = timerNew();
  net = incSearch(lattice, agSize, agThreshold, nonSpecThreshold);
  cdgPrintf(CDG_PROFILE,
            "PROFILE: isearch took %lums\n", timerElapsed(profileTime));
  timerFree(profileTime);

  cdgCtrlCTrapped = FALSE;
  cdgCtrlCAllowed = FALSE;
  cnMostRecentlyCreatedNet = net;

  if (!net) {
    cdgPrintf(CDG_ERROR, "ERROR: invalid net\n");
    return FALSE;
  }
  hashSet(cdgNets, net->id, net);

  return TRUE;
}

/* ----------------------------------------------------------------------
   Sort all levels according to the specified order and
   display the new order.
*/
Boolean cmdLevelsort(int no, char **args)
{
  if (no != listSize(inputCurrentGrammar->levels)) {
    cdgPrintf(CDG_ERROR,
              "ERROR: please specify all %d levels\n",
              listSize(inputCurrentGrammar->levels));
    return FALSE;
  }

  if (setLevelSort(args)) {
    cdgPrintf(CDG_INFO, "INFO: order of levels: %s\n", getLevelSort());

    /* re-sorting levels invalidates constraint hashing,
       so it must be repeated. */
    inputCacheConstraints(inputCurrentGrammar);
    
    return TRUE;
  }

  return FALSE;
}

/* ----------------------------------------------------------------------
   Set or query hooks.
*/
Boolean cmdHook(int no, char **args)
{
  int i;
  Hook hook;
  int hookNo;
  const char *usage = "hook [ on|off|reset | <HookName> [on|off|reset] ]";

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: argument is NULL\n");
    return FALSE;
  }

  if (no > 2) {
    cdgPrintf(CDG_ERROR, "ERROR: wrong # args\n" "       %s\n", usage);
    return FALSE;
  }

  if (no < 1 || args[0] == NULL) {
    cdgPrintf(CDG_DEFAULT, "\n  No Name                 State    Count\n");
    cdgPrintf(CDG_DEFAULT, "  -----------------------------------------\n");
    for (i = 0; i < vectorSize(hkHooks); i++) {
      hook = (Hook) vectorElement(hkHooks, i);
      if (!hook)
        continue;

      cdgPrintf(CDG_DEFAULT, "  %02d %-20s %-8s %d\n",
                i, hook->name,
                (hook->active ? "enabled" : "disabled"), hook->count);
    }
    cdgPrintf(CDG_DEFAULT, "\n  hooking %s.\n\n",
              (hkVerbosity & CDG_HOOK ? "enabled" : "disabled"));
  }

  if (no == 1) {
    if (strcmp(args[0], "on") == 0) {
      cdgPrintf(CDG_INFO, "INFO: hooking enabled\n");
      hkVerbosity |= CDG_HOOK;
    } else if (strcmp(args[0], "off") == 0) {
      cdgPrintf(CDG_INFO, "INFO: hooking disabled\n");
      hkVerbosity &= ~CDG_HOOK;
    } else if (strcmp(args[0], "reset") == 0) {
      cdgPrintf(CDG_INFO, "INFO: resetting counter of all hooks\n");
      for (i = 0; i < vectorSize(hkHooks); i++) {
        hook = (Hook) vectorElement(hkHooks, i);
        hook->count = 0;
      }
    } else {
      hookNo = hkFindNoOfHook(args[0]);
      if (hookNo < 0) {
        cdgPrintf(CDG_ERROR, "ERROR: unknown hook `%s'\n", args[0]);
        return FALSE;
      }
      hook = (Hook) vectorElement(hkHooks, hookNo);
      hook->active ^= TRUE;
      cdgPrintf(CDG_INFO, "INFO: hook `%s' is %s\n",
                hook->name, (hook->active ? "enabled" : "disabled"));
    }
  }

  if (no == 2) {
    hookNo = hkFindNoOfHook(args[0]);
    if (hookNo < 0) {
      cdgPrintf(CDG_ERROR, "ERROR: unknown hook `%s'\n", args[0]);
      return FALSE;
    }
    hook = (Hook) vectorElement(hkHooks, hookNo);

    if (strcmp(args[1], "on") == 0) {
      hook->active = TRUE;
      cdgPrintf(CDG_INFO, "INFO: hook `%s' is enabled\n", hook->name);
    } else if (strcmp(args[1], "off") == 0) {
      hook->active = FALSE;
      cdgPrintf(CDG_INFO, "INFO: hook `%s' is disabled\n", hook->name);
    } else if (strcmp(args[1], "reset") == 0) {
      hook->count = 0;
      cdgPrintf(CDG_INFO, "INFO: hook `%s' is reseted\n", hook->name);
    } else {
      cdgPrintf(CDG_ERROR, "ERROR: unknown option `%s'\n"
                "       %s\n", args[1], usage);
      return FALSE;
    }
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Subject the specified net to a guided local search 
   for global solutions.
*/
Boolean cmdGls(int no, char **args)
{
  return (glsCommand(no, args) >= 0);
}

#ifdef USE_NEWGLS
/* ----------------------------------------------------------------------
 * forward command to gls v2 module
 */
Boolean cmdNewGls(int no, char **args)
{
  return nglsCommand(no, args);
}
#endif

/* ----------------------------------------------------------------------
 * Compare parse to annotation
*/
Boolean cmdVerify(int no, char **args)
{
  Parse p = NULL;
  AnnoEntry ae = NULL;
  String id;
  ParseVerification pv;
  Lattice lat;

  if (no > 0) {
    if (!(p = parseFind(args[0]))) {
      cdgPrintf(CDG_ERROR, "ERROR: no such Parse, `%s'\n", args[0]);
      return FALSE;
    }
  } else {
    if (!(p = parseMostRecentlyCreatedParse)) {
      cdgPrintf(CDG_ERROR, "ERROR: no Parse available.\n");
      return FALSE;
    }
  }

  if (no > 1) {
    id = strRegister(args[1]);
  } else {
    id = p->latticeId;
  }

  if (!(lat = findLattice(id))) {
    cdgPrintf(CDG_ERROR, "ERROR: no such lattice.\n");
    return FALSE;
  }
    
  if (!(ae = findAnnoForLattice(lat->id, TRUE))) {
    cdgPrintf(CDG_ERROR, "ERROR: no such annotation.\n");
    return FALSE;
  }

  cdgPrintf(CDG_INFO, "Accuracy of %s:\n", p->id);
  pv = parseVerify(p, ae);
  writeXmlParseVerification(pv);
  parseVerificationDelete(pv);

  return TRUE;
}

/* ----------------------------------------------------------------------
   Chunk the specified wordgraph.
*/
Boolean cmdChunk(int no, char **args)
{
  Lattice lattice;
  int i;
  Chunker chunker;
  List chunks = NULL;
  LexemGraph lg;
  Timer profileTime;
  int noLattices;
  List l;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: argument is NULL\n");
    return FALSE;
  }

  profileTime = timerNew();
  noLattices = 0;
  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  for(l = inputCurrentGrammar->lattices; l != NULL; l = l->next) {
    lattice = l->item;
    
    if (no > 0) {
      for (i = 0; i < no && args[i] != NULL; i++) {
        if (lattice->id == strRegister(args[i])) {
          lg = lgNew(lattice);
          if (!lg) {
            timerFree(profileTime);
            return FALSE;
          }
          chunker = chunkerNew(DefaultChunker, lg);
          if (!chunker) {
            cdgPrintf(CDG_ERROR, "ERROR: chunking of '%s' failed\n",
                      lattice->id);
            timerFree(profileTime);
            lgDelete(lg);
            return FALSE;
          }
          chunks = chunkerChunk(chunker);
          chunkerPrintChunks(CDG_INFO, chunks);
          lgDelete(lg);
          chunkerDelete(chunker);
          noLattices++;
          break;
        }
      }
    } else {
      lg = lgNew(lattice);
      if (!lg) {
        timerFree(profileTime);
        return FALSE;
      }
      chunker = chunkerNew(DefaultChunker, lg);
      if (!chunker) {
        cdgPrintf(CDG_ERROR, "ERROR: chunking of '%s' failed\n", lattice->id);
        timerFree(profileTime);
        lgDelete(lg);
        return FALSE;
      }
      chunks = chunkerChunk(chunker);
      chunkerPrintChunks(CDG_INFO, chunks);
      lgDelete(lg);
      noLattices++;
    }
  }
  cdgPrintf(CDG_PROFILE, "\nPROFILE: chunking %d lattice(s) took %lums\n",
            noLattices, timerElapsed(profileTime));
  timerFree(profileTime);
  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;

  return (chunks != NULL);
}

/* ----------------------------------------------------------------------
   Print the help string for the specified CDG command.
*/
Boolean cmdHelp(int no, char **args)
{
  int i, j;

  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: argument is NULL\n");
    return FALSE;
  }

  for (i = 0; interface_commands[i].name != NULL; i++) {
    if (no > 0) {
      for (j = 0; j < no && args[j] != NULL; j++) {
        if (strcmp(interface_commands[i].name, args[j]) == 0) {
          cdgPrintf(CDG_DEFAULT,
                    "%15s %s\n", args[j], interface_commands[i].doc);
          break;
        }
      }
    } else {
      cdgPrintf(CDG_DEFAULT,
                "%15s %s\n",
                interface_commands[i].name, interface_commands[i].doc);
    }
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Specifies a file to be used for autoloading lexicon items.

   The parameter must be string that is used as the file name. The file of
   cdg input must be called S.cdg, and the index must be called S.db.

   If these files are not in the current working directory they are also
   searched in other directories that input was loaded from.
*/
Boolean cmdUseLexicon(int no, char **args) {
  String name;
  
  if (no != 1) {
    cdgPrintf(CDG_ERROR, "ERROR: `uselexicon' needs one string argument.\n");
    return FALSE;
  }
 
  name = strCopy(args[0]);
  dbOpen(name);
 
  return TRUE;
}

/* ----------------------------------------------------------------------
   Close the lexicon database.
*/

Boolean cmdCloseDB()
{
  return dbClose();
}

/* ----------------------------------------------------------------------
   Save all parses to a file in Prolog format.
*/

Boolean cmdParses2prolog(int no, char **args)
{
  char filename[PATH_MAX];
  FILE *s;
  List l, m;
  int i = 0;

  if (no != 1) {
    cdgPrintf(CDG_ERROR, "ERROR: please specify one file name\n");
    return FALSE;
  }

  realpath(args[0], filename);
  s = fopen(filename, "w");

  if (s == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: cannot open %s\n", filename);
    return FALSE;
  }
  /* cdgPrintf(CDG_DEFAULT, "File '%s' successfully opened.\n", cdgFilepath); */

  for (l = m = hashListOfKeys(cdgParses); l != NULL; l = l->next) {
    String id = l->item;
    Parse p = hashGet(cdgParses, id);

    writePrologParse(s, p);
    i++;
  }
  listDelete(m);
  (void)fclose(s);

  cdgPrintf(CDG_INFO, "%i parse(s) written to `%s'\n", i, filename);

  return TRUE;
}

/* ----------------------------------------------------------------------
   Save all annotations to a file in Prolog format.
*/
Boolean cmdAnnos2prolog(int no, char **args)
{
  char filename[PATH_MAX];
  FILE* s;
  List l, m;
  int i = 0;

  if (no != 1)
  {
    cdgPrintf(CDG_ERROR, "ERROR: please specify a file name\n");
    return FALSE;
  }

  realpath(args[0], filename);
  s = fopen(filename, "w");

  if (s == NULL)
  {
    cdgPrintf(CDG_ERROR, "ERROR: cannot open %s\n", filename);
    return FALSE;
  }
  /*cdgPrintf(CDG_DEFAULT, "File '%s' successfully opened.\n", cdgFilepath);*/

  for(l = m = hashListOfKeys(inputCurrentGrammar->annotations); l != NULL; l = l->next)
  {
    String id = l->item;
    AnnoEntry ae = hashGet(inputCurrentGrammar->annotations, id);
    writePrologAE(s,ae);
    i++;
  }
  listDelete(m);
  (void)fclose(s);

  cdgPrintf(CDG_INFO,"%i annotation(s) written to `%s'\n",i,filename);

  return TRUE;
}

/* ----------------------------------------------------------------------
   Perform chart-based search.
*/
Boolean cmdChart(int no, char **args) {
  return chart_search(no, args);
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */

/** @} */
