/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: constraintnet.c,v 1.189 2006-03-10 14:51:50 foth Exp $
 */

/* -------------------------------------------------------------------------
   @addtogroup Constraintnet Constraintnet - maintainance of constraint nets
   @author Ingo Schroeder
   @date 6/3/97
   @{
 */

/* ---------------------------------------------------------------------- */
#include <config.h>

#include <stdio.h>
#include <blah.h>
#include <math.h>
#include <string.h>
#include "cdg.h"
#include "input.h"
#include "eval.h"
#include "statistics.h"
#include "levelvalue.h"
#include "constraintnet.h"
#include "lexemgraph.h"
#include "hook.h"
#include "timer.h"
#include "increment.h"
#include "parse.h"
#include "set.h"
#include "tagger.h"
#include "chunker.h"

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */

/** Counts the constraint nets created so far and appears
   in the name of every constraint net. */
int cnCounter = 0;

/** cnMostRecentlyCreatedNet points to last constraint net created
(or to NULL). This variable is used as an implicit argument to
commands that expect the name of a constraint net. */
ConstraintNet cnMostRecentlyCreatedNet = NULL;

/** This variable implements the CDG variable edges. */
CnEdgesType cnEdgesFlag = cnEdgesOff;

/** If this flag is set, deleted LVs will be bracketed with [ ] in
    all output. If it is not set, they will not be shown at all. */
Boolean cnShowDeletedFlag = FALSE;

/** if this Flag is set, newly created constraint networks will contain
 * non-specific dependency edges aswell */
Boolean cnUseNonSpec = FALSE;

/** This variable gives the approximate ratio of LVs retained while
    building a constraint net. It defaults to~1, so that no LVs are
    deleted upon building. If it is smaller, then from any set <i>W</i>
    of LVs the worst floor(|<i>W</i>|   x)) LVs will   be
    removed. */
Number cnUnaryPruningFraction = 1.0;

/** This variable implements the CDG variable sortnodes. */
int cnSortNodesMethod = 0;

/* -- FUNCTIONS --------------------------------------------------------- */
/* ----------------------------------------------------------------------
   This function returns a new ConstraintNet with a unique name,
   but with all fields set to meaningless default values. Building a
   complete constraint net is a multi-stage process that involves several
   of the other functions of this module.
 */
ConstraintNet cnBuildInit()
{
  ConstraintNet net;

  net = (ConstraintNet) memMalloc(sizeof (ConstraintNetStruct));
  /* There is no garantee that no longer ids are needed but ... */
  net->id = strPrintf("net%d", cnCounter++);
  net->isBuilt = NULL;
  net->totalNumberOfValues = 0;
  net->values = net->nodes = net->edges = NULL;
  net->parses = (List) NULL;
  net->searchagenda = (Agenda) NULL;
  net->evalUnary = 0;
  net->evalBinary = 0;
  net->statUnary = 0;
  net->cache = NULL;
  net->lvTotals = NULL;

  return net;
}

/* ----------------------------------------------------------------------
   Build lexeme graph and add it to NET.
*/
Boolean cnTag(ConstraintNet net, Lattice lat) {

  net->lexemgraph = lgNew(lat);

  /* invalid graph? */
  if (net->lexemgraph == NULL) {
    cnDelete(net);
    return FALSE;
  }

  cdgPrintf(CDG_INFO, "INFO: grapheme graph: #nodes %d, min %d, max %d\n",
            vectorSize(net->lexemgraph->graphemnodes),
            net->lexemgraph->min, net->lexemgraph->max);
  cdgPrintf(CDG_INFO,
            "INFO: lexem graph: #nodes %d, min %d, max %d, #paths %lld\n",
            vectorSize(net->lexemgraph->nodes),
            net->lexemgraph->min,
            net->lexemgraph->max,
            net->lexemgraph->noOfPaths);

  return TRUE;
}

/* ----------------------------------------------------------------------
   Final improvements: optimize, sort, unary pruning, score cache and edges

   This function performs various clean-up operations on a net for which
   all constraint nodes have been built:
   -  It removes unnecessary structures from the net using
     \b cnOptimizeNet().
   -  If \b cnSortNodes is set, it sorts the constraint nodes
     using \b cnSortNodes().
   -  It applies \b cnUnaryPruning() to each constraint node.
   -  If \b scUseCache is set, it initializes \b net->cache
     to a new cache returned by \b scNew().
   -  If \b cnEdgesFlag is not \b cnEdgesOff, it applies
     \b cnBuildEdges() to build the edges of the constraint net.

   The function returns \b FALSE iff any of the subsidiary functions
   returned \b FALSE.
 */
ConstraintNet cnBuildFinal(ConstraintNet net, Boolean buildLVs)
{
  int i, j, k;
  int max = net->lexemgraph + cnUseNonSpec ? 2 : 1;

  /* We set all cells in isBuilt here, instead of doing it in
     cnBuildLevelValues, because that function does not do what isBuilt
     counts. isBuilt cares about whether all LVs that subordinate word 3
     under word 5 have been built ot not. cnBuildLevelValues() only builds
     LVs for that bind one particular grapheme, of which word 3 may have
     several, but for that grapheme it uses all possible regents.

     To actually build only the LVs connecting word 3 to word 5, use
     cnBuildTriple().
  */
  for(i = net->lexemgraph->min; i < net->lexemgraph->max; i++) {
    for(j = 0; j < max+1; j++) {
      for(k = 0; k < inputCurrentGrammar->noOfLevels; k++) {
        arraySetElement(net->isBuilt, (Pointer)(int)buildLVs, i, j, k);
      }
    }
  }


  /*
     checks net for deletable values and lexemes
   */
  if (buildLVs && cnOptimizeNet(net) < 0) {
    net = NULL;
    return net;
  }

  /*
     sorts nodes of constraintnet
   */
  if (cnSortNodesMethod != 0) {
    cnSortNodes(net);
  }

  /*
     unary pruning
   */
  for (i = 0; i < vectorSize(net->nodes); i++) {
    ConstraintNode node = (ConstraintNode) vectorElement(net->nodes, i);

    cnUnaryPruning(node);
  }

  /* sort LVs in each node according to their score */
  cnSortLVs(net);

  /* builds the score cache */
  if (scUseCache) {
    net->cache = scNew(vectorSize(net->values));
  } else {
    net->cache = NULL;
  }

  /* build edges TODO: in theory edges should be built incrementally too */
  if (cnEdgesFlag != cnEdgesOff) {
    cnBuildEdges(net);
    if (net->edges == NULL || vectorSize(net->edges) == 0) {
      cdgPrintf(CDG_WARNING, "WARNING: invalid net, net contains no edges\n");
      cnDelete(net);
      cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;
      net = NULL;
      return net;
    }
  }

  return net;
}

/* ----------------------------------------------------------------------
   Build a constraint net from LAT.

   This function returns a new constraint net for a word graph. This
   function uses \b cnBuildInit(), \b cnBuildNodes(), and
   \b cnBuildFinal(). The resulting net contains no edges. The function
   performs these initializations:
   -  \b id is set to \b net<cnCounter>
   -  the \b lexemgraph is built
   -  \b state is set to \b NSCreated

   If BUILDLVS is set, the nodes in the constraint net will be filled with
   LVs, otherwise the net will contain only constraint nodes, no LVs.
*/
ConstraintNet cnBuild(Lattice lat, Boolean buildLVs)
{
  ConstraintNet net;

  net = cnBuildInit();

  if(!cnTag(net,lat)) {
    return NULL;
  }

  /* build nodes */
  if (!cnBuildNodes(net, buildLVs)) {
    cdgPrintf(CDG_WARNING, "WARNING: invalid net, net contains no nodes\n");
    cnDelete(net);
    return NULL;
  }

  if (cdgCtrlCTrapped) {
    cdgPrintf(CDG_WARNING, "WARNING: interrupt while building a constraintnet\n");
    cnDelete(net);
    return NULL;
  }

  net = cnBuildFinal(net, buildLVs);
  if (net)
    cnMostRecentlyCreatedNet = net;
  else
    net = NULL;

  hashSet(cdgNets, net->id, net);

  return net;
}

/* ----------------------------------------------------------------------
   Print a constraint net.

   This function displays a constraint net in textual form. Output is
   suppressed if \b hkVerbosity does not have the bit \b mode
   set. The function uses \b cnPrintNode() and \b lvPrint().
   Constraint nodes from levels that have their \b showflag reset are
   skipped. Deleted LVs are only shown if \b cnShowDeletedFlag is set.
 */
void cnPrint(long unsigned int mode, ConstraintNet net)
{
   int i, j;
   ConstraintNode n;
   int noOfNotDeletedNodes = 0;
   int vmin = 99999999;
   int vmax = 0;
   int vtotal = 0;
   int vnum;

   cdgPrintf(mode, "------------------------------------------------------------\n");
   cdgPrintf(mode, "    id: %s\n", net->id);
   cdgPrintf(mode, " nodes:\n");
   for (i = 0; i < vectorSize(net->nodes) && !cdgCtrlCTrapped; i++) {
      n = (ConstraintNode) vectorElement(net->nodes, i);

      if (!n->level->showflag)
         continue;

      if (!lgAreDeletedNodes(net->lexemgraph, n->gn->lexemes))
         noOfNotDeletedNodes++;

      cdgPrintf(mode, "%2d ", i);
      cnPrintNode(mode, n);

      cdgPrintf(mode, " %d:\n", vectorSize(n->values));
      vnum = n->noValidValues;
      for (j = 0; j < vectorSize(n->values) && !cdgCtrlCTrapped; j++) {
         LevelValue lv = (LevelValue) vectorElement(n->values, j);

         cdgPrintf(mode, "   ");

         if (!lv->isDeleted || cnShowDeletedFlag) {
           cdgPrintf(mode, lv->isDeleted ? "[" : " ");
           lvPrint(mode, lv, net->lexemgraph->isDeletedNode, 1);
           cdgPrintf(mode, lv->isDeleted ? "] " : " ");
           cdgPrintf(mode, "\n");
         }
      }
      vmin = min(vmin, vnum == 0 ? vmin : vnum);
      vmax = max(vmax, vnum);
      vtotal += vnum;
      cdgPrintf(mode, "\n");
   }
   cdgPrintf(mode, "#solutions(s): %d\n", listSize(net->parses));
   cdgPrintf(mode, "#nodes: %d/%d\n", noOfNotDeletedNodes, vectorSize(net->nodes));
   cdgPrintf(mode, "#paths: %lld\n", net->lexemgraph->noOfPaths);
   cdgPrintf(mode, "values: #min %d, #max %d, #total %d, average %5.2f\n",
      vmin, vmax, vtotal, vtotal / (float)noOfNotDeletedNodes);

   cdgPrintf(mode, "#edges: %d\n", net->edges == NULL ? 0 : vectorSize(net->edges));
/*
   for (i = 0; i < vectorSize(net->nodes); i++)
   {
      n = (ConstraintNode) vectorElement(net->nodes, i);
      cdgPrintf(mode,"%2d %s(%d,%d)-%s: ", i,
         n->lexemnode->lexem->description,
         n->lexemnode->arc->from,
         n->lexemnode->arc->to,
         n->level->id);
      for (j = 0; j < vectorSize(net->edges); j++)
      {
         ConstraintEdge e = (ConstraintEdge) vectorElement(net->edges, j);

         if (n == e->start)
         {
            cdgPrintf(mode,"%s(%d,%d)-%s ",
               e->stop->lexemnode->lexem->description,
               e->stop->lexemnode->arc->from,
               e->stop->lexemnode->arc->to,
               e->stop->level->id);
         }
      }
      cdgPrintf(mode,"\n");
   }
*/

   cdgPrintf(mode, "------------------------------------------------------------\n");


   if (net->lexemgraph->chunks) {
     cdgPrintf(mode, "chunks:\n");
     chunkerPrintChunks(mode, net->lexemgraph->chunks);
   }
}

/* ----------------------------------------------------------------------
   Print a constraint edge.

   This function displays a constraint edge and the associated matrix
   \b scores. It reacts to the display flag \b mode in the same
   way as \b cnPrint() and all other display functions.
 */
void cnPrintEdge(long unsigned int mode, ConstraintEdge e)
{
   int i, j;
   LevelValue lv;
   String pseudoLexeme;                /* either a lexeme id or NIL or NONSPEC */

   /* the two nodes */
   cdgPrintf(mode, "%s(%d,%d)-%s  --->  %s(%d,%d)-%s\n",
      e->start->gn->arc->word,
      e->start->gn->arc->from,
      e->start->gn->arc->to,
      e->start->level->id,
      e->stop->gn->arc->word,
      e->stop->gn->arc->from,
      e->stop->gn->arc->to,
      e->stop->level->id);

   /* matrix of combinational scores */

   /* top row */
   cdgPrintf(mode, "start v    stop >| ");
   for (j = 0; j < vectorSize(e->stop->values); j++) {
      lv = (LevelValue) vectorElement(e->stop->values, j);

      if (lv->regents == NULL) {
         pseudoLexeme = "NIL";
      } else if (lv->regents == NONSPEC) {
         pseudoLexeme = "NONSPEC";
      } else {
         pseudoLexeme = ((LexemNode)(lv->regents->item))->lexem->description;
      }

      cdgPrintf(mode, "%5.5s/%-10.10s ", lv->label, pseudoLexeme);


   }
   cdgPrintf(mode, "\n--------------------------------------------------");
   cdgPrintf(mode, "--------------------------------------------------\n");

   /* other rows */
   for (i = 0; i < vectorSize(e->start->values); i++) {
      lv = (LevelValue) vectorElement(e->start->values, i);

      if (lv->regents == NULL) {
         pseudoLexeme = "NIL";
      } else if (lv->regents == NONSPEC) {
         pseudoLexeme = "NONSPEC";
      } else {
         pseudoLexeme = ((LexemNode)(lv->regents->item))->lexem->description;
      }

      /* left column */
      cdgPrintf(mode, "%5.5s/%-10.10s | ", lv->label, pseudoLexeme);

      /* cells */
      for (j = 0; j < vectorSize(e->stop->values); j++) {
         cdgPrintf(mode, "  %.6e   ", smGetScore(e->scores, i, j));
      }
      cdgPrintf(mode, "\n");
   }

   cdgPrintf(mode, "\n");

}

/* ----------------------------------------------------------------------
   Print a constraint node.

   A constraint node is essentially a collection of LVs that bind the
   same grapheme node on the same level. If a node binds the word
   `ich' on the level REF, it might look like this:

     ich(0-1)/REF

*/
void cnPrintNode(long unsigned int mode, ConstraintNode cn)
{
  cdgPrintf(mode, "%s", cn->gn->arc->word);
  cdgPrintf(mode, "(%d-%d)",
            cn->gn->arc->from,
            cn->gn->arc->to);
  
  cdgPrintf(mode, "/%s", cn->level->id);
}

/* -------------------------------------------------------------------
   comparison function for cnUnaryPruning()

   This function compares two LVs by their scores. It returns \b TRUE
   iff \b a->score < b->score.
 */
Boolean cnUnaryPruningCompare(LevelValue a, LevelValue b)
{
   return (a->score < b->score);
}

/* ----------------------------------------------------------------------
   Remove a given percentage of values from a node.

   This function finds the values with the worst scores from a
   constraint node and deletes them. Let <i>W</i> be the set of LVs in \b node.
   All elements of <i>W</i> are deleted whose score is smaller than the score
   of the <i>n</i>th element, where

   n = \b (int) =|W| * (1 - \b cnUnaryPruningFraction)

   This function temporarily sorts the set $W$ using the function
   \b cnUnaryPruningCompare()=.
 */
void cnUnaryPruning(ConstraintNode node)
{
  Vector v;
  int i, j;
  Number limit = 0.0;
  LevelValue lv;

  if (node->noValidValues == 0)
    return;

  v = vectorClone(node->values);
  vectorSort(v, cnUnaryPruningCompare);

  i = (int)((node->noValidValues - 1) * (1.0 - cnUnaryPruningFraction));

  /* get the i-th valid value of the domain */
  for (j = 0; j < vectorSize(v); j++) {
    lv = (LevelValue) vectorElement(v, j);
    if (!lv->isDeleted)
      i--;

    if (i == 0) {
      limit = lv->score;
      break;
    }
  }

  vectorDelete(v);

  /* mark levelvalues as deleted if they fall below the limit */
  for (i = 0; i < vectorSize(node->values); i++) {
    lv = (LevelValue) vectorElement(node->values, i);
    if (lv->isDeleted)
      continue;

    if (lv->score < limit) {
      cdgPrintf(CDG_INFO, "INFO: deleting ");
      lvPrint(CDG_INFO, lv, NULL, 2);
      cdgPrintf(CDG_INFO, "\n");

      lv->isDeleted = TRUE;
      node->noValidValues--;
    }
  }
}

/* -------------------------------------------------------------------
   Compare constraint nodes by the ->no of their levels.
 */
Boolean cnNodeComparePrio(ConstraintNode a, ConstraintNode b)
{
   return a->level->no > b->level->no;
}

/* -------------------------------------------------------------------
   Comparison function for cnSortNodes() based on domain size.

   This function is used by \b cnSortNodes() to compare two
   constraint nodes according to the value of their
   \b totalNumberOfValues= fields. It returns \b TRUE iff \b a
   has a lower value or the same values as \b b.
 */
Boolean cnNodeCompareSmallest(ConstraintNode a, ConstraintNode b)
{
  return ( a->noValidValues <= b->noValidValues);
}

/* -------------------------------------------------------------------
   Sort constraint nodes.

   This function sorts the constraint nodes of \b net using the
   function \b cnNodeCompare().
 */
void cnSortNodes(ConstraintNet net)
{
  ConstraintNode node;
  int i;

  if (hkVerbosity & CDG_DEBUG) {
    cdgPrintf(CDG_DEBUG, "DEBUG: old sorting:\n");
    for (i = 0; i < vectorSize(net->nodes); i++) {
      node = (ConstraintNode) vectorElement(net->nodes, i);
      cdgPrintf(CDG_DEBUG, "   %04d: [%3d] ", i, node->noValidValues);
      cnPrintNode(CDG_DEBUG, node);
      cdgPrintf(CDG_DEBUG, "\n");
    }
  }

  if (cnSortNodesMethod==1) {
    vectorSort(net->nodes, cnNodeComparePrio);
  } else if (cnSortNodesMethod==2) {
    vectorSort(net->nodes, cnNodeCompareSmallest);
  }

  if (hkVerbosity & CDG_DEBUG) {
    cdgPrintf(CDG_DEBUG, "DEBUG: new sorting:\n");
    for (i = 0; i < vectorSize(net->nodes); i++) {
      node = (ConstraintNode) vectorElement(net->nodes, i);
      cdgPrintf(CDG_DEBUG, "   %04d: [%3d] ", i, node->noValidValues);
      cnPrintNode(CDG_DEBUG, node);
      cdgPrintf(CDG_DEBUG, "\n");
    }
  }

}

/* ----------------------------------------------------------------------
   Build a new levelvalue in NODE.

   This function creates exactly one LV with the specified fields using
   \b lvNew() and stores it in \b node, incrementing all relevant
   counters properly.
 */
void cnBuildLv(ConstraintNode node, List dependents, Level level, String label, List regents)
{
  LevelValue newLv = lvNew(dependents, level, label, regents);
  ConstraintNet net = node->net;

  newLv->score = evalUnary(newLv, net, NULL, FALSE, NULL, NULL);
  if (statUseStatisticsFlag) {
    newLv->score *= statUnary(newLv, net);
  }

#ifdef DEBUGCNBUILDLV
  cdgPrintf(CDG_DEBUG, "DEBUG: Built LV ");
  lvPrint(CDG_DEBUG, newLv, NULL, 2);
#endif

  if (newLv->score == 0.0) {
#ifdef DEBUGCNBUILDLV
    cdgPrintf(CDG_DEBUG, "... invalid.\n");
#endif
    lvDelete(newLv);
  } else {
#ifdef DEBUGCNBUILDLV
    cdgPrintf(CDG_DEBUG, "... OK.\n");
#endif
    newLv->limit = newLv->score;
    vectorAddElement(node->values, newLv);
    newLv->indexWRTNet = vectorAddElement(net->values, newLv);
    net->totalNumberOfValues++;
    node->totalNumberOfValues++;
  }
}

/* ----------------------------------------------------------------------
   Builds all level values for LEVEL, MODIFIER and MODIFIEE.

   This function builds all LVs that represent tuples composed from the
   parameters it receives. This function is used as the innermost loop by
   \b cnBuildIter().
 */
void cnBuildLevelValues(ConstraintNode node, Level level,
                        GraphemNode modifier, GraphemNode modifiee)
{
  ConstraintNet net = node->net;
  Direction dir;
  List l;

  /* guard against self-modification */
  if(!lgMayModify(net->lexemgraph, modifier, modifiee)) {
    return;
  }

  /* compute edge direction */
  if (modifiee == NONSPEC) {
    dir = Right;
  } else if(!modifiee) {
    dir = Nil;
  } else if(modifier->arc->from < modifiee->arc->from) {
    dir = Right;
  } else {
    dir = Left;
  }

  /* iterate over labels */
  for(l = level->labels; l != NULL; l = l->next) {

    /* compute edge type */
    String label = l->item;
    EdgeType et = etEncode(level,label,dir);
    BitString downRelevant = vectorElement(inputCurrentGrammar->downFeatures, et);
    BitString upRelevant = vectorElement(inputCurrentGrammar->upFeatures, et);
    List downClasses = lgPartitions(modifier, downRelevant);
    List upClasses = lgPartitions(modifiee, upRelevant);

    List m, n;
    for (m = downClasses; m != NULL; m = m->next) {
      List dependents = m->item;
      for (n = upClasses; n != NULL; n = n->next) {
        List regents = n->item;
        cnBuildLv(node, dependents, level, label, regents);
      }
    }
    if (spec(modifiee)) {
      listForEachDelete(upClasses, listDelete);
    } else {
      listDelete(upClasses);
    }
    listForEachDelete(downClasses, listDelete);
  }
}

/* ----------------------------------------------------------------------
   Build all LVs on LEVEL for the pair of time points (A,B) in NET.

   All LVs that connect a word that starts at time point A as a modifier to
   a word that starts at time point B are built and inserted into the
   constraint net at the appropriate places. B == -1 means NIL, B == -2
   means NONSPEC.

   Returns FALSE if such LVs have already been built.

   This function has nothing to do with incremental processing. It is
   intended for use by parsing algorithms that do not want the entire set
   of LVs built before they can start working, although the entire input
   is known.
*/
Boolean cnBuildTriple(ConstraintNet net, int a, int b, int levelno) {
  LexemGraph lg = net->lexemgraph;
  int i,j;
  int bb = b;
  List l;
  /* warp sentinel values */
  if(b < 0) {
    bb = lg->max - b;
  }

  /* refuse unnecessary work */
  if(arrayElement(net->isBuilt, a, bb, levelno)) {
    return FALSE;
  }

  /* do necessary work */
  for (l = inputCurrentGrammar->levels; l != NULL; l = l->next) {
    Level level = (Level) l->item;
    if (!level->useflag) continue;
    if(level->no != levelno) continue;

    for(i = 0; i < vectorSize(lg->graphemnodes); i++) {
      GraphemNode gn = vectorElement(lg->graphemnodes, i);
      ConstraintNode node = NULL;
      if(a != gn->arc->from) continue;

      /* find pertinent constraint node */
      for(j = 0; j < vectorSize(net->nodes); j++) {
        node = vectorElement(net->nodes, j);
        if(node->level != level) continue;
        if(node->gn != gn) continue;
        break;
      }

      if(-1 == b) {
        cnBuildLevelValues(node, level, gn, NULL);
      }
      else if(-2 == b) {
        cnBuildLevelValues(node, level, gn, NONSPEC);
      }
      else for(j = 0; j < vectorSize(lg->graphemnodes); j++) {
        GraphemNode modifiee = vectorElement(lg->graphemnodes, j);
        if(modifiee->arc->from != b) continue;
        cnBuildLevelValues(node, level, gn, modifiee);
      }
      node->noValidValues = vectorSize(node->values);
      node->totalNumberOfValues = node->noValidValues;
    }
  }

  /* be proud of it */
  arraySetElement(net->isBuilt, (Pointer)TRUE, a, bb, levelno);
  return TRUE;

}


/* ----------------------------------------------------------------------
  Builds the constraint node corresponding to GN in NET.

  This function really performs most of the work that was documented
  under cnBuildNodes() for simplicity.
 */
Boolean cnBuildIter(ConstraintNet net, GraphemNode gn, Boolean buildLVs)
{
  Level level;
  GraphemNode modifiee;
  ConstraintNode node;
  List l;
  int k, z;
  int ggs = vectorSize(net->lexemgraph->graphemnodes);
  int lgs = vectorSize(net->lexemgraph->nodes);
  int lvs = listSize(inputCurrentGrammar->levels);

  if (gn->lexemes == NULL)
    return TRUE;

  cdgFlush();

  /* loop through all levels */
  for (l = inputCurrentGrammar->levels; l != NULL; l = l->next) {
    level = (Level) l->item;
    if (!level->useflag)
      continue;

    /* check for ^C here */
    if (cdgCtrlCTrapped) {

      /* delete all constraint-nodes */
      for (z = 0; z < vectorSize(net->nodes); z++) {
        node = (ConstraintNode) vectorElement(net->nodes, z);
        cnDeleteNode(node);
      }
      vectorDelete(net->nodes);
      net->nodes = NULL;
      return FALSE;
    }
    cdgFlush();

    /* if an entire grapheme is deleted from the graph already,
       do not even build the constraint node. */
    if (lgAreDeletedNodes(net->lexemgraph, gn->lexemes)) {
      continue;
    }

    /* allocate new constraint node */
    node = (ConstraintNode) memMalloc(sizeof (ConstraintNodeStruct));
    node->net = net;
    node->level = level;
    node->gn = gn;
    node->values = vectorNew(lgs * lvs);
    node->totalNumberOfValues = 0;

    /* fill it */
    if(buildLVs) {
      for (k = 0; k <= ggs; k++) {
        if(k == ggs) {
          modifiee = NULL;
        } else {
          modifiee = vectorElement(net->lexemgraph->graphemnodes, k);
        }
        cnBuildLevelValues(node, level, gn, modifiee);
      }
      if (cnUseNonSpec) {
        cnBuildLevelValues(node, level, gn, NONSPEC);
      }
    }

    /* add it to the net */
    node->noValidValues = vectorSize(node->values);
    node->totalNumberOfValues = node->noValidValues;
    vectorAddElement(net->nodes, node);
  }
  return TRUE;
}


/* ----------------------------------------------------------------------
   Builds the constraint nodes of NET. If BUILDLVS is set, the nodes are
   immediately filled with all possible LVs, otherwise they remain empty.

   Returns FALSE if the constraintnet is invalid.

   Basically it performs the following steps:

   -  allocates the Vectors \b nodes and \b values
   -  partitions the set of lexeme nodes created from each word
      hypothesis according to each level by using \b lgPartitions()
   -  allocates a \b ConstraintNode for each of the partitions <i>k</i>
      and inserts it into the Vector \b nodes
   -  checks whether the subordination <i>(k, l, m)</i> is possible for each
      triples of modifier set, label and modifiee set, and constructs the
      corresponding LV
   -  checks whether the subordination <i>(k, l, \b NIL)</i> is possible for each
      pair of modifier set and label, and constructs the corresponding LV
   -  inserts all new LVs into the respective constraint nodes and the
      Vector \b values
   -  sorts the LVs in each constraint node by their limit.

   However, some complications apply:

   -  Under \b CDG_DEBUG, each word in the word graph is printed
      as it is used to build LVs.
   -  If a level has its \b useflag reset, it is ignored completely.
   -  Some of the LVs created may be destroyed again by
      \b cnUnaryPruning() if the variable
      \b cnUnaryPruningFraction is smaller than~1. Such LVs are not
      stored in the constraint node nor in the constraint net.
   -  Each iteration executes the hook \b HOOK_CNBUILDNODES.
   -  If \b cnSortNodes is set, \b cnSortNodes() is called
      after building all nodes.
   -  If \b scUseCache is set, a new cache is allocated for the
      constraint net.
   -  The time elapsed is printed as a \b CDG_PROFILE message.
   -  This function is interruptible by \b C-c in much the same
      way as \b cnBuildEdges().
 */
Boolean cnBuildNodes(ConstraintNet net, Boolean buildLVs)
{
  Timer profileTime;
  GraphemNode gn;
  int i;
  int ggs = vectorSize(net->lexemgraph->graphemnodes);
  int lgs = vectorSize(net->lexemgraph->nodes);
  int lvs = listSize(inputCurrentGrammar->levels);
  LexemGraph lg = net->lexemgraph;

  profileTime = timerNew();
  net->nodes = vectorNew(lgs * lvs);
  net->values = vectorNew(lgs * lvs);
  net->isBuilt = arrayNew(lg->max, (2+lg->max), lvs, 0);

  for (i = 0; i < ggs; i++) {
    gn = (GraphemNode) vectorElement(net->lexemgraph->graphemnodes, i);
    if (gn->lexemes == NULL)
      continue;

    cdgPrintf(CDG_DEBUG, "%s ",
              ((LexemNode)( gn->lexemes->item))->lexem->word);

    if (!cnBuildIter(net, gn, buildLVs))
      break;
  }

  cdgPrintf(CDG_DEBUG, "\n");
  cdgPrintf(CDG_PROFILE,
            "PROFILE: building nodes took %ldms\n", timerElapsed(profileTime));
  if (hkVerbosity & CDG_HOOK)
    cdgExecHook(HOOK_CNBUILDNODES, net);

  timerFree(profileTime);
  return TRUE;
}

/* ----------------------------------------------------------------------
   Optimize constraint node by deleting lexemes and values.
   @returns > 0 : we have changed the constraintnode
   @returns 0   : no changes have been made
   @returns < 0 : the node is invalid

   This function serves to discard structures in a constraint node that
   cannot appear in any solution. It returns the number of deletion
   operations that it has performed, or -1 if an inconsistency was found.
   This function is repeatedly called by \b cnOptimizeNet().
 */
int cnOptimizeNode(ConstraintNet net, ConstraintNode node)
{
  int k;
  LevelValue lv;
  int noChanges = 0;
  LexemGraph lg = net->lexemgraph;
  GraphemNode gn = node->gn;
  List l;
  Boolean pruneModifiees = TRUE;
  GraphemNode theModifiee = NULL;
  BitString down = bitNew(vectorSize(lg->nodes));
  BitString up = bitNew(vectorSize(lg->nodes));

  bitClearAll(down);
  bitClearAll(up);

#ifdef DEBUGCNOPTIMIZENODE
  {
    int i;

    cdgPrintf(CDG_DEBUG,
	      "DEBUG: optimizing constraint node %s/%s(%d,%d)\n",
	      gn->arc->word,
	      node->level->id, gn->arc->from, gn->arc->to);
    for (i = 0; i < vectorSize(node->values); i++) {
      LevelValue lv = vectorElement(node->values, i);

      if (lv->isDeleted) {
	cdgPrintf(CDG_DEBUG, "[");
      }
      lvPrint(CDG_DEBUG, lv, NULL, 0);
      if (lv->isDeleted) {
	cdgPrintf(CDG_DEBUG, "]");
      }
      cdgPrintf(CDG_DEBUG, "\n");
    }
  }
#endif
  /* condition (a) */
  if (node->noValidValues > 0) {

    /* test domain */
    for (k = 0; k < vectorSize(node->values); k++) {
      lv = (LevelValue) vectorElement(node->values, k);
      if (lv->isDeleted) {
	continue;
      }

      /* mark invalid levelvalues as deleted */
      if (lgAreDeletedNodes(lg, lv->dependents)) {
#ifdef DEBUGCNOPTIMIZENODE
	cdgPrintf(CDG_DEBUG, "DEBUG: Deleting LV ");
	lvPrint(CDG_DEBUG, lv, NULL, 0);
	cdgPrintf(CDG_DEBUG, " because of invalid dependents.\n");
#endif
	lv->isDeleted = TRUE;
	node->noValidValues--;
	noChanges++;
      } else if ((spec(lv->regents) && lgAreDeletedNodes(lg, lv->regents)) ||
 	  (spec(lv->regents) && 
	   !lgCompatibleNodes(lg, lv->dependents->item, lv->regents->item))) {
#ifdef DEBUGCNOPTIMIZENODE
	cdgPrintf(CDG_DEBUG, "DEBUG: Deleting LV ");
	lvPrint(CDG_DEBUG, lv, NULL, 0);
	cdgPrintf(CDG_DEBUG, " because of invalid regents.\n");
#endif
	lv->isDeleted = TRUE;
	node->noValidValues--;
	noChanges++;
      }

      if (lv->isDeleted) {
	continue;
      }

      /* detect that a lexeme node can still be bound */
      for (l = lv->dependents; l != NULL; l = l->next) {
	LexemNode ln = l->item;
	bitSet(down, ln->no);
      }

      /** if all of the LVs in this node modify the same grapheme node, then
       * we can delete all variants of their regents.
       * Otherwise we can't be sure, so we don't do anything. */

      if (spec(lv->regents)) {
	if (pruneModifiees) {
	  if (!theModifiee) {
	    theModifiee = ((LexemNode) (lv->regents->item))->grapheme;
#ifdef DEBUGCNOPTIMIZENODE
	    cdgPrintf(CDG_DEBUG,
		      "DEBUG: no modifiee seen yet, taking `%s'(%d-%d)\n",
		      theModifiee->arc->word,
		      theModifiee->arc->from, theModifiee->arc->to);
#endif
	  } else if (theModifiee != ((LexemNode) (lv->regents->item))->grapheme) {
#ifdef DEBUGCNOPTIMIZENODE
	    cdgPrintf(CDG_DEBUG,
		      "DEBUG: That's a different modifiee -- no modifiee pruning.\n");
#endif
	    pruneModifiees = FALSE;
	  }
	}
	if (pruneModifiees) {
	  for (l = lv->regents; l != NULL; l = l->next) {
	    LexemNode ln = l->item;
	    bitSet(up, ln->no);
	  }
	}
      } else {
#ifdef DEBUGCNOPTIMIZENODE
	if (pruneModifiees) {
	  cdgPrintf(CDG_DEBUG,
		    "DEBUG: NIL binding seen -- no modifiee pruning.\n");
	}
#endif
	pruneModifiees = FALSE;
      }
    }
  }

  if (node->noValidValues == 0) {
    if (!lgIsDeletableNode(gn)) {
      cdgPrintf(CDG_WARNING,
		"WARNING: can't find a value for %s/%s(%d,%d), invalid net\n",
		gn->arc->word,
		node->level->id, gn->arc->from, gn->arc->to);

      return -1;		       /* this indicates an invalid net */
    }
  }

  /* condition (b) */
  for (l = gn->lexemes; l != NULL; l = l->next) {
    LexemNode ln = l->item;

    if (!lgIsDeletedNode(lg, ln) && !bitGet(down, ln->no)) {
#ifdef DEBUGCNOPTIMIZENODE
      cdgPrintf(CDG_DEBUG, "DEBUG: deleting ");
      lgPrintNode(CDG_DEBUG, ln);
      cdgPrintf(CDG_DEBUG, " because of the modifier criterion.\n");
#endif
      lgDeleteNode(lg, ln);
      noChanges++;
    }
  }
  if (pruneModifiees && theModifiee) {
    for (l = theModifiee->lexemes; l != NULL; l = l->next) {
      LexemNode ln = l->item;

      if (!lgIsDeletedNode(lg, ln) && !bitGet(up, ln->no)) {
#ifdef DEBUGCNOPTIMIZENODE
	cdgPrintf(CDG_DEBUG, "DEBUG: deleting ");
	lgPrintNode(CDG_DEBUG, ln);
	cdgPrintf(CDG_DEBUG, " because of the modifiee criterion.\n");
#endif
	lgDeleteNode(lg, ln);
	noChanges++;
      }
    }
  }
  bitDelete(up);
  bitDelete(down);

  /* condition (c) */
  if (lg->noOfPathsFromStart[gn->no] * lg->noOfPathsToEnd[gn->no] == 0 &&
      !lgAreDeletedNodes(lg, gn->lexemes)) {
#ifdef DEBUGCNOPTIMIZENODE
    cdgPrintf(CDG_DEBUG,
	      "DEBUG: Deleting entire grapheme node because of the path criterion.\n");
#endif
    lgDeleteNodes(lg, gn->lexemes);
    noChanges++;
  }

  return noChanges;
}

/* ----------------------------------------------------------------------
   This function returns the number of undeleted LVs in \b net.
   It is a helper fo cnOptimizeNet 
 */
int countValidValues(ConstraintNet net)
{
  int i;
  LevelValue lv;
  int result = 0;

  for (i = 0; i < vectorSize(net->values); i++) {
    lv = (LevelValue)vectorElement(net->values, i);
    if (!lv->isDeleted)
      result++;
  }

  return result;
}


/* ----------------------------------------------------------------------
   optimize constraint net by deleting lexemes and values

   This function deletes lexeme nodes and LVs from a constraint net that
   cannot possibly appear in any solution. It uses several rules for
   deleting structures:

   -  An LV can be deleted if it binds a set of lexeme nodes that are
     all deleted.
   -  An LV can be deleted if is modifier and modifiee are
     incompatible.
   -  A lexeme node can be deleted if it it cannot be bound by any LV
     on one level.
   -  A lexeme can be deleted if there is no path through the lexeme
     graph in which it appears, and which is totally undeleted.

   Since these conditions can trigger each other, the function
   \b cnOptimizeNode() is called repeatedly on each node in the
   constraint net until no progress has been made. The function returns
   the total number of changes returned by the calls to
   \b cnOptimizeNode(), or -1 if none of them did so.

   @returns > 0 : we have changed the constraintnet
   @returns 0   : no changes have been made
   @returns < 0 : the net is invalid
 */
int cnOptimizeNet(ConstraintNet net)
{
  ConstraintNode node;
  int i, n, result = 0;
  Boolean progress = TRUE;
#ifdef DEBUG_CNOPTIMIZENET
  int noDeletedValues;
  int oldNoValues;
  oldNoValues = countValidValues(net);
#endif

  /* loop while progress has been made */
  while (progress) {
    progress = FALSE;

    /* loop over all nodes and call cnOptimizeNode */
    for (i = 0; i < vectorSize(net->nodes); i++) {
      node = (ConstraintNode)vectorElement(net->nodes, i);

      n = cnOptimizeNode(net, node);
      if (n < 0) {
        return -1;
      }
      else if (n > 0) {
        progress = TRUE;
        result += n;
      }
    }
  }

#ifdef DEBUG_CNOPTIMIZENET
  noDeletedValues = oldNoValues - countValidValues(net);
  if (noDeletedValues) {
    cdgPrintf(CDG_DEBUG, "DEBUG: deleted %d levelvalues\n",
              noDeletedValues);
  }
#endif

  return result;
}

/* ----------------------------------------------------------------------
   Are two constraint nodes connected by an arc?
   This function checks whether two constraint nodes should be connected by an
   edge or not. This check is always performed before allocating a
   constraint edge. Two nodes fail this test if they bind the same lexeme
   and belong to the same level.  
 */
Boolean cnConnectedByArc(ConstraintNode a, ConstraintNode b)
{
   /**
      Think it over which nodes are connected by an arc.
      Do we need arcs in both directions?
    */
   return (a->gn != b->gn || a->level != b->level);

/**
   return ((a->lexemnode == b->lexemnode && a->level != b->level)
           ||
           (a->level == b->level
            &&
            lgDistanceOfNodes (a->lexemnode->lexemgraph,
                               a->lexemnode,
                               b->lexemnode) != 0
           )
   );
*/
}

/* ----------------------------------------------------------------------
   Return a constraint net to untouched state

   This function undoes all changes that have been made to a constraint
   net since it was built. In particular, it restores all deleted
   lexeme nodes and LVs to a constraint net. All parses of
   \b net are deallocated. This returns the net to the state that it
   was in immediately after being optimized. \b TRUE is returned only
   if at least one value remains for each pair of word and level.
 */
Boolean cnRenew(ConstraintNet net)
{
  int i, j;
  LevelValue lv;
  LexemNode ln;
  ConstraintNode node;

  /* refresh constraint nodes and LVs */
  for (i = 0; i < vectorSize(net->nodes); i++) {
    node = (ConstraintNode) vectorElement(net->nodes, i);

    node->noValidValues = 0;
    for (j = 0; j < vectorSize(node->values); j++) {
      lv = vectorElement(node->values, j);

      if (!cnUseNonSpec && lv->regents == NONSPEC) {
        lv->isDeleted = TRUE;
      } else {
        lv->isDeleted = FALSE;
        node->noValidValues++;
      }
      lv->limit = lv->score;
      lv->no = -1;
    }
  }

  /* refresh individual lexemes */
  for (i = 0; i < vectorSize(net->lexemgraph->nodes); i++) {
    ln = vectorElement(net->lexemgraph->nodes, i);
    ln->limit = 1.0;
  }

  /* undo lexem deletions */
  bvSetAllElements(net->lexemgraph->isDeletedNode, FALSE);
  lgComputeNoOfPaths(net->lexemgraph);
  lgComputeDistances(net->lexemgraph);

  /* throw away score-cache and build a new one */
  scDelete(net->cache);
  if (scUseCache) {
    net->cache = scNew(vectorSize(net->values));
  } else {
    net->cache = NULL;
  }

  /* throw away all solutions */
  /* Parses are owned by inputCurrentGrammar and are not freed here. */
  listDelete(net->parses);
  net->parses = NULL;

  /* re-sort the LVs by unary score */
  cnSortLVs(net);

  /* optimize the net */
  if(cnOptimizeNet(net) < 0) {
    return FALSE;
  };

  return TRUE;
}

/* ----------------------------------------------------------------------
   builds edges in a constraint net

   This function computes all edges in a constraint net:

   -  The Vector \b edges is allocated.
   -  A constraint edge is built for every pair of constraint nodes
     and inserted into \b edges.
   -  The fields \b isMarked and \b scores are allocated for
     each edge.
   -  All pairs of LVs from the two constraint nodes are evaluated
     jointly and the result is store in the matrix \b scores. The result
     will be~0 if the LVs are no \b lvCompatible(), otherwise it is
     the product of the combined binary score and the two unary scores.
   -  A mapping from all edges to their reverses is computed and
     stored in the field \b reverse.

   \b C-c interrupts this function, deletes all partial results, and
   displays the total time elapsed as a \b CDG_PROFILE message.
 */
void cnBuildEdges(ConstraintNet net)
{
  LexemGraph lg = net->lexemgraph;
  int lgs = vectorSize(lg->nodes);
  int lvs = listSize(inputCurrentGrammar->levels);
  int i, j, k, l;
  int total, done = 0;
  ConstraintNode in, jn;
  ConstraintEdge edge, reverse;
  Timer profileTime;
  Vector reverseTable;

  profileTime = timerNew();

  reverseTable = vectorNew(vectorSize(net->nodes) * vectorSize(net->nodes));
  vectorSetAllElements(reverseTable, NULL);

  /* estimate total # of steps necessary */
  total = vectorSize(net->nodes) * vectorSize(net->nodes);
  if (!cdgXCDG) {
    cdgPrintf(CDG_DEBUG, "DEBUG: building edges...   0%%");
  }

  /*
     check whether there are already edges, free them
   */

  /*
     How many edges will we probably end up with?
     (lgs*lvs)^2 is the maximum number
   */
  net->edges = vectorNew(lgs * lvs);
  for (i = 0; i < vectorSize(net->nodes); i++) {
    in = (ConstraintNode) vectorElement(net->nodes, i);
    for (j = 0; j < vectorSize(net->nodes); j++) {
      jn = (ConstraintNode) vectorElement(net->nodes, j);

      /* progress message */
      if (!cdgXCDG) {
        cdgPrintf(CDG_DEBUG, "%3.0f%%", 100 * (++done / (float)total));
      }

      /* check for ^C here */
      if (cdgCtrlCTrapped) {
        /* compare to cnDelete */
        for (i = 0; i < vectorSize(net->edges); i++) {
          edge = (ConstraintEdge) vectorElement(net->edges, i);
          memFree(edge->isMarked);
          smDelete(edge->scores);
          memFree(edge);
        }
        vectorDelete(net->edges);
        /* end comparison */
        net->edges = NULL;
        return;
      }
      /* TEST TEST TEST TEST TEST TEST TEST TEST */
      /* not really necessary */
      if (in->level->no <= jn->level->no)
        k = in->level->no * lvs + jn->level->no;
      else
        k = jn->level->no * lvs + in->level->no;

      if (cnEdgesFlag == cnEdgesFew && inputCurrentGrammar->levelMatrixCounter != NULL
          && (int)vectorElement(inputCurrentGrammar->levelMatrixCounter,
                                k) == 0) continue;
      /* TEST TEST TEST TEST TEST TEST TEST TEST */

      /* connected by arc should be configurable */
      if (cnConnectedByArc(in, jn)) {
        edge = (ConstraintEdge) memMalloc(sizeof (ConstraintEdgeStruct));
        edge->start = in;
        edge->stop = jn;
        edge->reverse = (ConstraintEdge) NULL;

        /* 7/31/97 IS - added isMarked to general edge setup */
        edge->isMarked =
          (Boolean *) memMalloc(sizeof (Boolean) * vectorSize(in->values));
        edge->scores = smNew(vectorSize(in->values), vectorSize(jn->values));
        for (k = 0; k < vectorSize(in->values); k++) {
          LevelValue ik = (LevelValue) vectorElement(in->values, k);

          edge->isMarked[k] = FALSE;
          for (l = 0; l < vectorSize(jn->values); l++) {
            LevelValue jl = (LevelValue) vectorElement(jn->values, l);
            double d;

            /* check whether the level values can exist on a path */
            if (lvCompatible(lg, ik, jl)) {
              if(cdgMultiply == cdgAggregation) {
                d = ik->score * jl->score * evalBinary(ik, jl, net, NULL, FALSE, NULL, NULL);
              } else {
                d = evalBinary(ik, jl, net, NULL, FALSE, NULL, NULL);
                d = min(d, ik->score);
                d = min(d, jl->score);
              }
            } else {
              d = 0.0;
            }
            smSetScore(edge->scores, d, k, l);
            /* 30.7.97 changed initialization FALSE -> TRUE */
            smSetFlag(edge->scores, TRUE, k, l);
          }
        }
        vectorAddElement(net->edges, edge);

        k = i * vectorSize(net->nodes) + j;
        vectorSetElement(reverseTable, edge, k);
        k = i + vectorSize(net->nodes) * j;
        reverse = vectorElement(reverseTable, k);
        if (reverse != NULL) {
          edge->reverse = reverse;
          reverse->reverse = edge;
        }
      }
    }
    cdgFlush();
  }

  vectorDelete(reverseTable);

  if (!cdgXCDG) {
    cdgPrintf(CDG_DEBUG, "done.\n");
  }

  cdgPrintf(CDG_PROFILE, "PROFILE: building edges took %ldms\n",
            timerElapsed(profileTime));

  timerFree(profileTime);
}

/* ----------------------------------------------------------------------
   Does this node bind the earliest lexeme in the lattice?
   The functions
   \b lgIsStartNode() and \b lgIsEndNode() are used for this
   purpose.
 */
Boolean cnIsStartNode(ConstraintNode n)
{
   return (n->level == inputCurrentGrammar->levels->item
      &&
      lgIsStartNode(n->gn));
}

/* ----------------------------------------------------------------------
   Does this node bind the latest lexeme in the lattice?
   The functions
   \b lgIsStartNode() and \b lgIsEndNode() are used for this
   purpose.  */
Boolean cnIsEndNode(ConstraintNode n)
{
   return (n->level == listLastElement(inputCurrentGrammar->levels)
      &&
      lgIsEndNode(n->gn));
}

/* ----------------------------------------------------------------------
   cnDeleteBinding

   This function deallocates \b NodeBinding structure. Both
   components of the pair are shallow copies and cannot be deallocated.
 */
void cnDeleteBinding(NodeBinding nb)
{
   memFree(nb);
}

/* ----------------------------------------------------------------------------
   Show what the conflict means. This function prints out the conflict 
   itself, the LVs that actually cause it, the constraint that is
   violated, and the participating lexemes.
*/
void cvAnalyse(ConstraintViolation cv, Vector context)
{

  Boolean inverse;
  int i;
  List l;
  LevelValue lv;
  List lexemes = NULL;
  LexemNode ln;

  VarInfo var;

  /* show conflict */
  cdgPrintf(CDG_INFO, "\n\nThe conflict is:\n================\n\n");
  cvPrint(CDG_INFO, cv);

  /* show constraint */
  cdgPrintf(CDG_INFO, "\n\n\nThe constraint is:\n==================\n\n");
  printConstraint(CDG_INFO, cv->constraint);

  /* is the order of LVs in the signature different from the order in the
   * conflict? */
  inverse =
    listSize(cv->constraint->vars) == 2
    && evalBinaryConstraint(cv->constraint, NULL, context, cv->lv1, cv->lv2);

  /* it is safe to assume that these two vector elements are not NULL because 
   * ConstraintViolations never point to NULL slots. */

  /* show LVs */
  cdgPrintf(CDG_INFO, "\n\nThe LVs concerned are:\n======================\n\n");
  for (i = 1, l = cv->constraint->vars; l; l = l->next, i++) {
    var = (VarInfo) l->item;
    if (listSize(cv->constraint->vars) == 1) {
      lv = cv->lv1;
    } else if ((i == 1 && !inverse) || (i == 2 && inverse)) {
      lv = cv->lv1;
    } else {
      lv = cv->lv2;
    }
    cdgPrintf(CDG_INFO, "%s == ", var->varname);
    lvPrint(CDG_INFO, lv, NULL, 0);
    cdgPrintf(CDG_INFO, "\n");

    /* note down the participating lexemes */
    if (lv->regents != NULL && lv->regents != NONSPEC) {
      lexemes = listAddUniqueElement(lexemes, lv->regents->item);
    }
    lexemes = listAddUniqueElement(lexemes, lv->dependents->item);
  }

  /* show participating lexemes */
  cdgPrintf(CDG_INFO,
	    "\n\nThe lexemes concerned are:\n==========================\n\n");
  for (l = lexemes; l; l = l->next) {
    ln = l->item;
    cdgPrintf(CDG_INFO, "%s: ", ln->lexem->word);
    printValue(CDG_INFO, ln->lexem->value, 0);
    cdgPrintf(CDG_INFO, "\n\n");
  }

  listDelete(lexemes);

}

/* ----------------------------------------------------------------------
   cvDelete

   This function deallocates a structure of type
   \b ConstraintViolation. Since all components of this type are
   shallow copies, only the structure itself is deallocated.
 */
void cvDelete(ConstraintViolation cv)
{
  /* Usually, all LVs are owned by the net in which they occur. If an LV
     with indexWRTNet -1 appears in a constraint violation, it must be an
     LV that exists only temporarily (not in any net). These LVs are owned
     by the cv, so they must be deallocated here. */
  if(cv->lv1->indexWRTNet == -1) {
    lvDelete(cv->lv1);
  }

  if(cv->lv2 && cv->lv2->indexWRTNet == -1) {
    lvDelete(cv->lv2);
  }

  memFree(cv);
}

/* ----------------------------------------------------------------------
   deletes a constraint-node

   This function deallocates a single constraint node and all LVs
   contained in it. Note that pointers to these LVs may remain in the
   field \b values of the enclosing constraint net, so
   \b cnDeleteNode() should only be called immediately before
   deleting the net itself.
 */
void cnDeleteNode(ConstraintNode node)
{
  int j;
  LevelValue lv;

  for (j = 0; j < vectorSize(node->values); j++) {
    lv = (LevelValue) vectorElement(node->values, j);
    lvDelete(lv);
  }

  vectorDelete(node->values);

  memFree(node);
}

/* ----------------------------------------------------------------------
   deletes a constraint-edge

   This function deallocates a constraint edge. Apart from the
   \b ConstraintEdgeStruct itself, only the field \b scores
   is deallocated by calling \b smDelete().
 */
void cnDeleteEdge(ConstraintEdge edge)
{
   memFree(edge->isMarked);
   smDelete(edge->scores);

   memFree(edge);
}

/* ----------------------------------------------------------------------
   deletes constraint net

   This function deallocates an entire constraint net. Note that while
   \b cnBuild() only allocates a few fields, this function allocates
   everything. The following functions are used for the purpose:

   -  \b lgDelete() for the \b lexemgraph
   -  \b cnDeleteNode() for the elements of \b nodes
   -  \b cnDeleteEdge() for the elements of \b edges
   -  \b agDelete() for the \b searchagenda
   -  \b deleteParse() for the elements of \b parses
   -  \b scDelete() for the \b cache

   If \b net is identical to \b cnMostRecentlyCreatedNet, that
   variable is reset.
 */
void cnDelete(ConstraintNet net)
{
   int i;
   ConstraintNode node;
   ConstraintEdge edge;

   if (net == NULL) {
      cdgPrintf(CDG_WARNING, "WARNING: cnDelete: net is NULL\n");
      return;
   }
   if (cnMostRecentlyCreatedNet == net)
      cnMostRecentlyCreatedNet = NULL;

   cdgFreeString(net->id);
   if (net->lexemgraph != NULL)
      lgDelete(net->lexemgraph);
   if (net->values != NULL)
      vectorDelete(net->values);
   if (net->nodes != NULL) {
      for (i = 0; i < vectorSize(net->nodes); i++) {
         node = (ConstraintNode) vectorElement(net->nodes, i);
         cnDeleteNode(node);
      }
      vectorDelete(net->nodes);
   }
   if (net->edges != NULL) {
      for (i = 0; i < vectorSize(net->edges); i++) {
         edge = (ConstraintEdge) vectorElement(net->edges, i);
         cnDeleteEdge(edge);
      }
      vectorDelete(net->edges);
   }

   if (net->searchagenda != NULL)
      agDelete(net->searchagenda);

   if (net->lvTotals != NULL)
     vectorDelete(net->lvTotals);

   /* Parses are owned by inputCurrentGrammar and are not freed here. */
   listDelete(net->parses);
   net->parses = NULL;

   scDelete(net->cache);
   memFree(net);
}

/* ----------------------------------------------------------------------
   print out some information about this net

   This function displays some general information about \b net.
   Apart from the fields of the structure themselves, the minimal,
   average, and maximal number of LVs per node is calculated and
   displayed. This function is called in various places to summarize
   briefly the state of a constraint net.
 */
void cnPrintInfo(ConstraintNet net)
{
  int i, vmin, vmax, vtotal;
  ConstraintNode node;
  long double candidates;

  if (!(hkVerbosity & CDG_INFO))
    return;

  vmin = 99999999;
  vtotal = vmax = 0;
  candidates = 0;
  for (i = 0; i < vectorSize(net->nodes); i++) {
    node = (ConstraintNode) vectorElement(net->nodes, i);
    vmin = min(vmin, node->noValidValues);
    vmax = max(vmax, node->noValidValues);
    vtotal += node->noValidValues;
    if (node->noValidValues) {
      if (!candidates) {
        candidates = node->noValidValues;
      } else {
        candidates *= node->noValidValues;
        if (candidates == HUGE_VAL) {
          cdgPrintf(CDG_WARNING, "WARNING: very large problem\n");
        }
      }
    }
  }

  cdgPrintf(CDG_INFO, "INFO: net: id %s, wordgraph %s\n",
            net->id, net->lexemgraph->lattice->id);
  cdgPrintf(CDG_INFO, "      #nodes %d, #edges %d\n",
            vectorSize(net->nodes),
            net->edges == NULL ? 0 : vectorSize(net->edges));
  cdgPrintf(CDG_INFO,
            "      #evaluations: %d unary, %d statistics, %d binary\n",
            net->evalUnary, net->statUnary, net->evalBinary);
  cdgPrintf(CDG_INFO,
            "      #values: min %d, max %d, total %d, average %5.2f\n", vmin,
            vmax, vtotal, vtotal / (float)vectorSize(net->nodes));
  cdgPrintf(CDG_INFO, "      #candidates: %.4Lg\n", candidates);
  if (net->cache && net->cache->size > 0) {
    cdgPrintf(CDG_INFO,
              "      cache: size %d #values %d, #hits %d, %1.1f on average, usage %1.2f%%\n",
              net->cache->capacity,
              net->cache->size,
              net->cache->hits,
              (float)net->cache->hits / (float)net->cache->size,
              (float)net->cache->size / (float)net->cache->capacity * 100);
  }
}

/* ----------------------------------------------------------------------
   Allocate a new constraint violation.
   This function allocates a new \b ConstraintViolation. The
   parameters are installed in the corresponding fields of the new
   structure. All fields of a \b ConstraintViolation are shallow
   copies.

 */
ConstraintViolation cvNew(Constraint c, LevelValue lva, LevelValue lvb)
{

  ConstraintViolation result = memMalloc(sizeof(ConstraintViolationStruct));
  result->constraint = c;
  result->penalty = c->penalty;
  result->lv1 = lva;
  result->lv2 = lvb;
  /* If an LV with indexWRTNet -1 appears in a constraint violation,
     it must be an LV that exists only temporarily (not in any net). These
     LVs are deallocated when they are no longer needed, so we clone them
     here in case the violation is preserved longer than the structure that
     caused it. */
  if(result->lv1->indexWRTNet == -1) {
    result->lv1 = lvClone(result->lv1);
  }
  if(result->lv2 && result->lv2->indexWRTNet == -1) {
    result->lv2 = lvClone(result->lv2);
  }


  result->nodeBindingIndex1 = lvIndex(lva);
  result->nodeBindingIndex2 = lvb ? lvIndex(lvb) : -1;
  /* Note that the default initialization to -1 is meaningful. At several
     points in the program, cv structures are compared by comparing their
     fields nodeBindingIndex1 and nodeBindingIndex2. If nodeBindingIndex2
     were left undefined for unary constraints, two identical unary
     violations might get classified as different.  */

  return result;

}

/* ----------------------------------------------------------------------
   clone a constraint violation
   @returns a clone of \b cv.
 */
ConstraintViolation cvClone(ConstraintViolation cv) {

  ConstraintViolation result = cvNew(cv->constraint, cv->lv1, cv->lv2);
  result->penalty = cv->penalty;
  return result;

}


/* ----------------------------------------------------------------------
   re-sorts the LVs in each constraint node by limit rather than by score

   This function sorts the Vectors  \b values=of all constraint nodes
   of \b net using \b lvCompare().
 */
void cnSortLVs(ConstraintNet net)
{
   int i;
   ConstraintNode n;

   for (i = 0; i < vectorSize(net->nodes); i++) {
      n = (ConstraintNode) vectorElement(net->nodes, i);
      vectorSort(n->values, lvCompare);
   }

}

/* ----------------------------------------------------------------------
   compares two violations by penalty, then by domain index
   This function compares two structures of type
   \b ConstraintViolation, comparing first the penalty of the
   violation, after that the fields
   \b nodeBindingIndex1 and \b nodeBindingIndex2 and finally the
   names of the constraints.

   TODO: this function is never used, see cvCompare for a similar function
 */
Boolean cnCompareViolation(ConstraintViolation a, ConstraintViolation b) {

  if(a->penalty < b->penalty) {
    return TRUE;
  } else if(a->penalty > b->penalty) {
    return FALSE;
  }

  if(a->nodeBindingIndex1 < b->nodeBindingIndex1) {
    return TRUE;
  } else if(a->nodeBindingIndex1 > b->nodeBindingIndex1) {
    return FALSE;
  }

  if(a->nodeBindingIndex2 < b->nodeBindingIndex2) {
    return TRUE;
  } else if(a->nodeBindingIndex2 > b->nodeBindingIndex2) {
    return FALSE;
  }

  if(strcmp(a->constraint->id, b->constraint->id) < 0) {
    return TRUE;
  }

  return FALSE;

}

/* ----------------------------------------------------------------------
   prints all Parses of a net

   This function applies \b parsePrint() to all entries of
   \b net->parses.
    */
void cnPrintParses(ConstraintNet net)
{
  List l;

  for (l = net->parses; l != NULL; l = l->next) {
    parsePrint(l->item);
    cdgPrintf(CDG_INFO, "\n\n");
  }
}

/* ----------------------------------------------------------------------
    Find ConstraintNode of an LV. This finds LVs by comparing their IWRT
    numbers, so cannot give meaningful results if lv was not originally
    built for net.
   @returns the constraint node in \b net that holds
   \b lv (or \b NULL).
 */
ConstraintNode cnFindNode(ConstraintNet net, LevelValue lv) {

  ConstraintNode cn;
  LevelValue lv2;
  int i, j;

  for (i = 0; i < vectorSize(net->nodes); i++) {
    cn = (ConstraintNode) vectorElement(net->nodes, i);
    for (j = 0; j < vectorSize(cn->values); j++) {
      lv2 = (LevelValue) vectorElement(cn->values, j);
      if(lv->indexWRTNet == lv2->indexWRTNet) {
        return cn;
      }
    }
  }

  cdgPrintf(CDG_WARNING,
            "WARNING: can't find LV #%d in net!\n",
            lv->indexWRTNet);

  return NULL;

}

/* ----------------------------------------------------------------------
   Print a conflict.

   This function displays a conflict  using \b cdgPrintf(). If
   \b width is~0, the constraint identifier is printed in full,
   otherwise it is truncated to \b width characters.
 */
void cvPrint(unsigned long mode, ConstraintViolation cv) {

  if (cv->nodeBindingIndex2 >= 0) {
    cdgPrintf(mode, "%03d : %03d : %4.3e : %s",
              cv->nodeBindingIndex1, cv->nodeBindingIndex2,
              cv->penalty, cv->constraint->id);
  } else {
    cdgPrintf(mode, "      %03d : %4.3e : %s",
              cv->nodeBindingIndex1, cv->penalty, cv->constraint->id);
  }

}

/* ----------------------------------------------------------------------
  Does a conflict occur in a list?

  This function checks whether a conflict equivalent to \b cv=
  appears in the List. \b TRUE= is returned if both the constraint
  and the fields \b nodeBindingIndex1= and \b nodeBindingIndex2=
  match (the penalty need not match).
*/
Boolean cvContains(List conflicts, ConstraintViolation cv) {

  List l;
  ConstraintViolation cv2;

  for (l = conflicts; l; l = l->next) {
    cv2 = l->item;
    if (cv->constraint->no == cv2->constraint->no &&
        ((cv->nodeBindingIndex1 == cv2->nodeBindingIndex1 &&
          cv->nodeBindingIndex2 == cv2->nodeBindingIndex2)
         ||
         (cv->nodeBindingIndex2 == cv2->nodeBindingIndex1 &&
          cv->nodeBindingIndex1 == cv2->nodeBindingIndex2))) {
      return TRUE;
    }
  }

  return FALSE;
}

/* ----------------------------------------------------------------------
   compares two violations by natural order

   This function compares two \b ConstraintViolation= structures by
   the position of the affected LVs. If both affect the same LVs, it
   compares them by the constraint name.
 */
Boolean cvCompareNatural(ConstraintViolation a, ConstraintViolation b)
{
  if (a->nodeBindingIndex1 < b->nodeBindingIndex1) {
    return TRUE;
  } else if (a->nodeBindingIndex1 > b->nodeBindingIndex1) {
    return FALSE;
  }

  if (a->nodeBindingIndex2 < b->nodeBindingIndex2) {
    return TRUE;
  } else if (a->nodeBindingIndex2 > b->nodeBindingIndex2) {
    return FALSE;
  }

  if (strcmp(a->constraint->id, b->constraint->id) < 0) {
    return TRUE;
  }

  return FALSE;
}


/* --------------------------------------------------------------------
   Constraint violations should be sorted first by penalty, then by arity,
   finally by natural order.

   This function compares two conflicts. It returns \b TRUE
    - if \b a has the lower penalty
    - otherwise, if \b a is unary and \b b is binary
    - otherwise, if the first LV in \b a precedes the first lv in
     \b b according to natural order.
 */
Boolean cvCompare(ConstraintViolation a, ConstraintViolation b)
{
  int indexa, indexb;

  /* compare by penalty */
  if (a->penalty < b->penalty)
    return TRUE;
  if (a->penalty > b->penalty)
    return FALSE;

  /* compare by arity */
  if (!a->lv2 && b->lv2)
    return TRUE;
  if (a->lv2 && !b->lv2)
    return FALSE;

  /* compare names */
  indexa = strcmp(a->constraint->id, b->constraint->id);
  if (indexa < 0)
    return FALSE;
  if (indexa > 0)
    return TRUE;

  /* compare by natural order */
  indexa = lvIndex(a->lv1);
  indexb = lvIndex(b->lv1);
  if (indexa < indexb)
    return TRUE;

  return FALSE;
}

/* ----------------------------------------------------------------------
   Looks for a constraint net with id ID and return it or NULL if
   it can't find the net in inputCurrentGrammar.

   @returns the net with the specified \b id in the
   \b inputCurrentGrammar structure, or \b NULL.
 */
ConstraintNet cnFindNet(String id)
{
  return hashGet(cdgNets, id);
}

/* ----------------------------------------------------------------------
   Looks for a constraint net about LATTICE.
 */
ConstraintNet cnFindNetForLattice(String id) {
  ConstraintNet result;
  HashIterator hi = hashIteratorNew(cdgNets);
  String s = strRegister(id);
  while((result = hashIteratorNextValue(hi))) {
    if(s == result->lexemgraph->lattice->id) {
      hashIteratorDelete(hi);
      return result;
    }
  }
  hashIteratorDelete(hi);
  return NULL;
}

/* ----------------------------------------------------------------------
   prints the active levelvalues

   This function applies \b lvPrint() to all LVs in the specified
   constraint net that are not deleted.
 */
void cnPrintActiveLVs(ConstraintNet net)
{
  int i, k;
  int j = 0;

  for (i=0; i < vectorSize(net->nodes); i++) {
    ConstraintNode cn = (ConstraintNode)vectorElement(net->nodes, i);
    for (k=0; k < vectorSize(cn->values); k++) {
      LevelValue lv = (LevelValue)vectorElement(cn->values, k);
      if (!lv->isDeleted) {
        lvPrint(CDG_DEBUG, lv, net->lexemgraph->isDeletedNode, 0);
        cdgPrintf(CDG_DEBUG, "\n");
        j++;
      }
    }
  }
  cdgPrintf(CDG_INFO, "INFO: number of levelvalues: %d found; %d active\n", vectorSize(net->values), j);
}

/* ----------------------------------------------------------------------
   undelete all levelvalues in a constraint net

   This function resets the field \b isDeleted for all LVs in the
   specified constraint net.
 */
void cnUndeleteAllLVs(ConstraintNet net)
{
  int i;
  LevelValue lv;

  for (i=0; i<vectorSize(net->values); i++) {
    lv = (LevelValue) vectorElement(net->values, i);
    lv->isDeleted = FALSE;
  }
}

/* ----------------------------------------------------------------------
   delete all levelvalues in a constraint net

   This function sets the field \b isDeleted for all LVs in the
   specified constraint net.
 */
void cnDeleteAllLVs(ConstraintNet net)
{
  int i;
  LevelValue lv;

  for (i=0; i<vectorSize(net->values); i++) {
    lv = (LevelValue) vectorElement(net->values, i);
    lv->isDeleted = TRUE;
  }
}

/* ----------------------------------------------------------------------
   Update NET with the arcs in LISTARCS.

   This function is used by \b incrementalcompletion to extend an
   existing constraint net by the structures corresponding to the
   specified \b Arc structures. */
Boolean cnBuildUpdateArcs(ConstraintNet net, List listArcs)
{
  int i;
  List m;

  for (i = 0; i < vectorSize(net->nodes); i++) {
    ConstraintNode cn = (ConstraintNode) vectorElement(net->nodes, i);
    Level level = cn->level;
    GraphemNode modifier = cn->gn;

#ifdef DEBUGCNBUILDNODES
    cdgPrintf(CDG_DEBUG, "\nDEBUG: generating LevelValues for node ");
    cnPrintNode(CDG_DEBUG, cn);
    cdgPrintf(CDG_DEBUG, "\n");
#endif

    for (m = listArcs; m; m = m->next) {
      Arc arc = (Arc) m->item;
      GraphemNode modifiee = cnGetGraphemNodeFromArc(net, arc);
      if (modifiee == NULL) {
        continue;
      }
      cnBuildLevelValues(cn, level, modifier, modifiee);
    }
    cn->noValidValues = vectorSize(cn->values);
    cn->totalNumberOfValues = cn->noValidValues;
  }

  for (m = listArcs; m; m = m->next) {
    Arc arc = (Arc) m->item;
    GraphemNode gn = cnGetGraphemNodeFromArc(net, arc);
    if (gn == NULL) {
      continue;
    }
    if (gn->lexemes == NULL) {
      continue;
    }
    if (!cnBuildIter(net, gn, TRUE)) {
      return FALSE;
    }
    /*** add levelvalues for NONSPEC ***/
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Return the graphem node in NET that points to ARC.
 */
GraphemNode cnGetGraphemNodeFromArc(ConstraintNet net, Arc arc)
{
  int i;
  GraphemNode gn;

  for (i=0; i<vectorSize(net->lexemgraph->graphemnodes); i++) {
    gn = (GraphemNode) vectorElement(net->lexemgraph->graphemnodes, i);
    if (gn->arc == arc)
      return gn;
  }

  return (GraphemNode) NULL;
}


/* ----------------------------------------------------------------------
   Callback function for cnUnaryPruningFraction.

   This function is used as the callback function for the CDG variable
   \b unaryFraction. It prints a notification of the change to the
   CDG shell.
 */
void cnCallback(String name, float *var)
{
  if (strcmp(name, "unaryFraction") == 0) {
    cdgPrintf(CDG_INFO, "INFO: unary pruning fraction set to %f\n",
              *var);
  }
}

/* ----------------------------------------------------------------------
   Initialize the module constraintnet.

   This function initializes the module \ref Constraintnet. It
   merely registers the module's CDG variables.
 */
void cnInitialize()
{
  setRegister("showdeleted",   SET_BOOL, &cnShowDeletedFlag, NULL, NULL, NULL, NULL);
  setRegister("edges",         SET_ENUM, &cnEdgesFlag, NULL, NULL, NULL,
                               "on", cnEdgesOn, "off", cnEdgesOff, "few", cnEdgesFew, "all", cnEdgesAll, NULL);
  setRegister("sortnodes",     SET_ENUM, &cnSortNodesMethod, NULL, NULL, NULL,
                               "off", 0, "on", 1, "prio", 1, "smallest", 2, NULL);
  setRegister("unaryFraction", SET_FLOAT, &cnUnaryPruningFraction, NULL, NULL, cnCallback, NULL);
  setRegister("usenonspec", SET_BOOL, &cnUseNonSpec, NULL, NULL, NULL, NULL);
}

/* ----------------------------------------------------------------------
   Needed by XCDG because it can't talk to LexemGraphs directly.

   This function simply returns \b cn->lexemgraph->lattice. It exists
   because XCDG needs that field, and XCDG cannot talk directly to lexeme
   graphs (they contain arrays of long long ints, which SWIG doesn't
   handle correctly).
 */
Lattice cnGetLattice(ConstraintNet cn) {
  return cn->lexemgraph->lattice;
}

/* ----------------------------------------------------------------------
   Delete all LVs of LEVEL at WORD whose label does not match with that of P.
*/
void cnLockLabel(ConstraintNet net, String label, int index, String levelname) {
  int i;
  Level level = findLevel(levelname);

  for (i = 0; i < vectorSize(net->nodes); i++) {
    ConstraintNode node = vectorElement(net->nodes, i);
    int j;
    if(level != node->level) continue;
    if(index != node->gn->arc->from) continue;
    for(j = 0; j < vectorSize(node->values); j++) {
      LevelValue lv = vectorElement(node->values, j);
      if(strcmp(lv->label,label) && !lv->isDeleted) {
        lv->isDeleted =TRUE;
        node->noValidValues--;
        if(0 == node->noValidValues) {
          cdgPrintf(CDG_WARNING, "WARNING: cnLockLabel() deleted last value\n");
        }
        
#ifdef DEBUGCNLOCK
        cdgPrintf(CDG_DEBUG, "DEBUG: deleting ");
        lvPrint(CDG_DEBUG, lv, NULL, 0);
        cdgPrintf(CDG_DEBUG, "\n");
      } else {
        if(!lv->isDeleted) {
          cdgPrintf(CDG_DEBUG, "DEBUG: retaining ");
          lvPrint(CDG_DEBUG, lv, NULL, 0);
          cdgPrintf(CDG_DEBUG, "\n");
        }
#endif
      }
    }
  }  
}

/* ----------------------------------------------------------------------
   Delete all LVs that do not bind lexeme ID at timepoint INDEX.
*/
void cnLockWord(ConstraintNet net, String description, int index) {

  int i;
  List l;
  for (i = 0; i < vectorSize(net->nodes); i++) {
    ConstraintNode node = vectorElement(net->nodes, i);
    int j;
    if(index != node->gn->arc->from) continue;

    for(j = 0; j < vectorSize(node->values); j++) {
      LevelValue lv = vectorElement(node->values, j);
      Boolean ok = FALSE;

      for(l = lv->dependents; l != NULL; l = l->next) {
        LexemNode ln = l->item;
        if(!strcmp(ln->lexem->description, description)) {
          ok = TRUE;
          break;
        }
      }
      
      if(!ok && !lv->isDeleted) {
        lv->isDeleted =TRUE;
        node->noValidValues--;
        if(0 == node->noValidValues) {
          cdgPrintf(CDG_WARNING, "WARNING: cnLockWord() deleted last value\n");
        }

#ifdef DEBUGCNLOCK
        cdgPrintf(CDG_DEBUG, "DEBUG: deleting ");
        lvPrint(CDG_DEBUG, lv, NULL, 0);
        cdgPrintf(CDG_DEBUG, "\n");
      } else {
        if(!lv->isDeleted) {
          cdgPrintf(CDG_DEBUG, "DEBUG: retaining ");
          lvPrint(CDG_DEBUG, lv, NULL, 0);
          cdgPrintf(CDG_DEBUG, "\n");
        }
#endif
      }
    }
  }
}


/* ----------------------------------------------------------------------
   Delete all LVs that do not subordinate word INDEX to REGENT on LEVELNAME.
*/
void cnLockEdge(ConstraintNet net, int regent, int index, String levelname) {

  int i;
  Level level = findLevel(levelname);

  for (i = 0; i < vectorSize(net->nodes); i++) {
    ConstraintNode node = vectorElement(net->nodes, i);
    int j;
    if(level != node->level) continue;
    if(index != node->gn->arc->from) continue;

    for(j = 0; j < vectorSize(node->values); j++) {
      LevelValue lv = vectorElement(node->values, j);
      int r = -1;
      if(lv->regents) {
        LexemNode ln = lv->regents->item;
        r = ln->arc->from;
      }
      
      if(regent != r && !lv->isDeleted) {
        lv->isDeleted =TRUE;
        node->noValidValues--;
        if(0 == node->noValidValues) {
          cdgPrintf(CDG_WARNING, "WARNING: cnLockEdge() deleted last value\n");
        }

#ifdef DEBUGCNLOCK
        cdgPrintf(CDG_DEBUG, "DEBUG: deleting ");
        lvPrint(CDG_DEBUG, lv, NULL, 0);
        cdgPrintf(CDG_DEBUG, "\n");
      } else {
        if(!lv->isDeleted) {
          cdgPrintf(CDG_DEBUG, "DEBUG: retaining ");
          lvPrint(CDG_DEBUG, lv, NULL, 0);
          cdgPrintf(CDG_DEBUG, "\n");
        }
#endif
      }
    }
  }
}


/* ---------------------------------------------------------------------- */
/** @} */
/* -- ENDOFFILE --------------------------------------------------------- */
