/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: functions.c,v 1.70 2006-05-02 14:21:56 micha Exp $
 */

/*
 * File: functions.c
 * Author: Ingo Schroeder
 * Purpose: CDG functions
 * Birth: 17/3/97
 *
 */

/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>

#include <stdio.h>
#include <string.h>
#include <blah.h>
#include <math.h>
#include "cdg.h"
#include "lexemgraph.h"
#include "functions.h"
#include "increment.h"
#include "hook.h"
#include "predict.h"

/* -- MACROS ------------------------------------------------------------ */

/* -- FORWARD DECLARATIONS ---------------------------------------------- */

/* -- VARIABLES --------------------------------------------------------- */
CDGFunctionStruct eval_functions[] = {
  {"abs", funcAbs, FALSE, "1",
   "funcAbs", "returns the absolute value"},
  {"acoustics", funcAcoustics, FALSE, "1",
   "funcAcoustics", "returns normalized acoustic score"},
  {"exp", funcExp, FALSE, "1",
    "funcExp", "returns the value of e raised to the power of x"},
  {"distance", funcDistance, FALSE, "2",
   "funcDistance", "returns the distance between the lexem nodes"},
  {"match", funcMatch, FALSE, "3", "funcMatch", "read the manual"},
  {"max", funcMax, FALSE, ">0", "funcMax", "returns the maximum"},
  {"min", funcMin, FALSE, ">0", "funcMin", "returns the minimum"},
  {"predict", funcPredict, FALSE, "3", "returns external predictions"},
  {"lookup", funcLookup, FALSE, ">0", "funcLookup" "holds user-defined data"},
  {"quotes", funcQuotes, FALSE, "1", "funcQuotes", 
   "return level of quotation of a word"},
  {"parens", funcParens, FALSE, "1", "funcParens", 
   "return level of parentheses of a word"},
  {"phrasequotes", funcPhrasequotes, FALSE, "1", "funcPhrasequotes", 
   "return level of phrase quotation of a word"},
  {"parent", funcParent, TRUE, "1-2", "funcParent", 
   "return parent of a word"},
  {"height", funcHeight, TRUE, "1", "funcHeight", 
   "return height of subtree under a word"},



  {(String) NULL, (VTFunction *) NULL, FALSE, (String) NULL}
};

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * return a pointer to the function implementing a functor
 */
CDGFunction funcGetFunction(String functor)
{
  int i;

  for (i = 0; eval_functions[i].functor != NULL; i++) {
    if (strcmp(functor, eval_functions[i].functor) == 0) {
      return &eval_functions[i];
      break;
    }
  }

  return (CDGFunction)NULL;
}

/* ----------------------------------------------------------------------
 * return the name of the function implementing a functor
 */
String funcGetFunctionName(String functor)
{
  int i;

  for (i = 0; eval_functions[i].functor != NULL; i++) {
    if (strcmp(functor, eval_functions[i].functor) == 0) {
      return eval_functions[i].functionName;
      break;
    }
  }

  return NULL;
}


/* ----------------------------------------------------------------------
 * returns distance between lexem nodes
 */
Value funcDistance(List args, Value val, LexemGraph lg, Vector context)
{
   Value a, b;
   ValueStruct vsa, vsb;


   a = evalTerm((Term) args->item, &vsa, lg, context);
   b = evalTerm((Term) args->next->item, &vsb, lg, context);

   if (a->type != VTLexemNode || b->type != VTLexemNode) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: arguments of function `distance' not all ids in constraint `%s'\n",
	 evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
      return ErrorValue;
   }
   if (a->data.lexemnode == NULL || b->data.lexemnode == NULL) {
      val->data.number = 0.0;
   } else {
     val->data.number = (Number)lgDistanceOfNodes(lg, 
                                                  (LexemNode)a->data.lexemnode, 
                                                  (LexemNode)b->data.lexemnode);
   }
   val->type = VTNumber;
   return val;
}

/* ----------------------------------------------------------------------
 * returns absolute value of its argument
 */
Value funcAbs(List args, Value val, LexemGraph lg, Vector context)
{
  Value v;

   v = evalTerm((Term) args->item, val, lg, context);
   switch (v->type) {
      case VTNumber:
        if (v->data.number < 0.0) {
	    v->data.number = -v->data.number;
        }
        break;
      default:
	 cdgPrintf(CDG_WARNING,
	    "WARNING: argument of function `abs' not a number in constraint `%s'\n",
	    evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
	 return ErrorValue;
   }

   return v;
}
/* ----------------------------------------------------------------------
 * returns the value of e raised to the power of x
 */
Value funcExp(List args, Value val, LexemGraph lg, Vector context)
{
  Value v;

  v = evalTerm((Term) args->item, val, lg, context);
  switch (v->type) {
    case VTNumber:
      v->data.number = exp(v->data.number);
      break;
    default:
      cdgPrintf(CDG_WARNING,
		"WARNING: argument of function `exp' not a number in constraint `%s'\n",
		evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return ErrorValue;
  }

  return v;
}

/* ----------------------------------------------------------------------
 * returns the minimum
 */
Value funcMin(List args, Value val, LexemGraph lg, Vector context)
{
  ValueStruct vsa;
   Value v;
   Number minValue=0.0;
   List l;


   l = args;
   if (l != NULL && l->item != NULL) {
     v = evalTerm((Term) l->item, &vsa, lg, context);
     if (v->type == VTNumber) {
       minValue = v->data.number;
     } else {
       cdgPrintf(CDG_WARNING,
                 "WARNING: argument of function `min' not a number in constraint `%s'\n",
		 evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
       return ErrorValue;
     }    
     l = l->next;
   }

   for (;
      l != NULL && l->item != NULL;
      l = l->next) {

      v = evalTerm((Term) l->item, &vsa, lg, context);
      if (v->type == VTNumber) {
	       minValue = min(minValue, v->data.number);
      } else {
        cdgPrintf(CDG_WARNING,
                  "WARNING: argument of function `min' not a number in constraint `%s'\n",
	 	  evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
        return ErrorValue;
      }
   }
   val->data.number = minValue;
   val->type = VTNumber;

   return val;
}

/* ----------------------------------------------------------------------
 * returns the maximum
 */
Value funcMax(List args, Value val, LexemGraph lg, Vector context)
{
  ValueStruct vsa;
   Value v;
   Number maxValue=0.0;
   List l;

   l = args;
   if (l != NULL && l->item != NULL) {
     v = evalTerm((Term) l->item, &vsa, lg, context);
     if (v->type == VTNumber) {
       maxValue = v->data.number;
     } else {
       cdgPrintf(CDG_WARNING,
                 "WARNING: argument of function `max' not a number in constraint `%s'\n",
		 evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
       return ErrorValue;
     }    
     l = l->next;
   }

   for (;
      l != NULL && l->item != NULL;
      l = l->next) {

      v = evalTerm((Term) l->item, &vsa, lg, context);
      if (v->type == VTNumber) {
	       maxValue = max(maxValue, v->data.number);
      } else {
        cdgPrintf(CDG_WARNING,
                  "WARNING: argument of function `max' not a number in constraint `%s'\n",
		  evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
        return ErrorValue;
      }
   }
   val->data.number = maxValue;
   val->type = VTNumber;

   return val;
}

/* ----------------------------------------------------------------------
 * Returns the number corresponding to the first sort that subsumes 
 * the additional argument.
 *
 * Function assumes well-formedness of VTList argument 
 * (it's enforced in parser.y):
 *
 *   VTString VTNumber VYString VTNumber ...
 */
Value funcMatch( List args, Value val, LexemGraph lg, Vector context)
{
  ValueStruct vsa, vsb, vsh;
  Value va, vb, vh;
  Hierarchy h;
  int a, b;
  List l;
  

  vh = evalTerm((Term) listNthElement(args, 1), &vsh, lg, context);
  if ( vh->type != VTString ) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (1st arg) for `match' in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return ErrorValue;
  }
  
  va = evalTerm((Term) listNthElement(args, 2), &vsa, lg, context);
  if ( va->type != VTString && va->type != VTList ) {
    if (!evalSloppySubsumesWarnings)
      cdgPrintf(CDG_WARNING,
		"WARNING: type mismatch (2nd arg) for `match' in constraint `%s'\n",
		evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return ErrorValue;
  }

  vb = evalTerm((Term) listNthElement(args, 3), &vsb, lg, context);
  if (vb->type != VTString) {
    if (!evalSloppySubsumesWarnings)
      cdgPrintf(CDG_WARNING,
		"WARNING: type mismatch (3rd arg) for `match' in constraint `%s'\n",
		evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return ErrorValue;
  }
  
  /* get hierarchy */
  h = findHierarchy(vh->data.string);
  if (!h) {
    cdgPrintf(CDG_WARNING,
              "WARNING: hierarchy `%s' not found in constraint `%s'\n", vh->data.string,
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return ErrorValue;
  }
  
   /* find second sort */
   if ( 0 > ( b = (int)hashGet( h->types, vb->data.string ) - 1 ) ) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type `%s' not found in hierarchy `%s' in constraint `%s'\n",
	 vb->data.string, vh->data.string,
	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return ErrorValue;
   }

   if ( va->type == VTString ) {
     /* This is a MATCH( STRING, STRING, STRING) call. */
     if ( 0 > ( a = (int)hashGet( h->types, va->data.string ) - 1) ) {
       cdgPrintf(CDG_WARNING,
		 "WARNING: type `%s' not found in hierarchy `%s' in constraint `%s'\n",
		 va->data.string, vh->data.string,
		 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
       return ErrorValue;
     }
     if ( (int)arrayElement( h->distance, a, b ) >= 0 ) {
       val->data.number = 1.0;
       val->type = VTNumber;
       return val;
     }     
   } else {
     /* This is a MATCH( STRING, LIST, STRING) call. */

     /* Loop through list */
     for ( l = va->data.list; l != NULL; l = l->next->next ) {
       /* find first sort */
       if ( 0 > ( a = (int)hashGet( h->types, ((Value)l->item)->data.string ) - 1 ) ) {
	 cdgPrintf(CDG_WARNING,
		   "WARNING: type `%s' not found in hierarchy `%s' in constraint `%s'\n",
		   ((Value)l->item)->data.string, vh->data.string,
		   evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
	 return ErrorValue;
       }
       if ( (int)arrayElement( h->distance, a, b ) >= 0 ) {
	 return (Value)l->next->item;
       }     
     }
   }
   val->data.number = 0.0;
   val->type = VTNumber;
   return val;     
}

/* ----------------------------------------------------------------------
 * Returns the normalized acoustic score.
 */
Value funcAcoustics(List args, Value val, LexemGraph lg, Vector context)
{
   Value a;
   ValueStruct vsa;
   LexemNode ln;


   a=evalTerm((Term)args->item, &vsa, lg, context);

   if (a->type!=VTLexemNode) {
     cdgPrintf(CDG_WARNING,
	       "WARNING: argument of function `acoustics' not an id in constraint `%s'\n",
	       evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
     return ErrorValue;
   }

   ln=(LexemNode)a->data.lexemnode;
   
   if (!ln || !ln->arc) {
     cdgPrintf(CDG_WARNING,
	       "WARNING: argument of function `acoustics' not complete in constraint `%s'\n",
	       evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
     return ErrorValue;
   }

   val->data.number=ln->arc->penalty;
   val->type=VTNumber;
   return val;
}

/* ----------------------------------------------------------------------
   Returns user-defined data.

   The first argument must be the name of a defined map. The second
   argument is a key into this map, and the corresponding Value is
   returned. 
*/
Value funcLookup(List args, Value val, LexemGraph lg, Vector context) {
  ValueStruct vsm;
  Value vm;
  String key = NULL;
  Value result;
  Hashtable h;
  List l;
  int counter;
  String constraint =
    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*";
  String sep = strRegister(":");
  

  vm = evalTerm(args->item, &vsm, lg, context);
  if ( vm->type != VTString ) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (1st arg) for `lookup' in constraint `%s'\n",
              constraint);
    return ErrorValue;
  }
#ifdef DEBUGLOOKUP
  cdgPrintf(CDG_DEBUG, "DEBUG: lookup(%s,", vm->data.string);
#endif
  /* find map */
  h = hashGet(inputCurrentGrammar->maps, vm->data.string);

  if (h) {
    /* construct key */
    counter = 1;
    for(l = args->next; l != NULL; l = l->next) {
      Term t = l->item;
      ValueStruct vs;
      Value v = evalTerm(t, &vs, lg, context);
      String s = v->data.string;
      counter++;
      if (v->type != VTString &&
          v->type != VTNumber) {
	cdgPrintf(CDG_WARNING,
		  "WARNING: type mismatch (arg %d) for `lookup' in constraint `%s'\n",
		  counter, constraint);
	return ErrorValue;
      }    
      if(VTNumber == v->type) {
        s = strPrintf("%.0f", v->data.number);
      }
      key = key ? strAppend(key, sep, s, NULL) : s;
    }
#ifdef DEBUGLOOKUP
  cdgPrintf(CDG_DEBUG, "%s) ", key);
#endif

    if((result = hashGet(h, key))) {
#ifdef DEBUGLOOKUP
      cdgPrintf(CDG_DEBUG, "== %s\n", result->data.string);
#endif
      return result;
    } else {
#ifdef DEBUGLOOKUP
      cdgPrintf(CDG_DEBUG, "failed.\n");
#endif
    }
  } else {
    cdgPrintf(CDG_WARNING,
              "WARNING: map `%s' not found in constraint `%s'\n",
              vm->data.string, constraint);
  }
    
  /* Try to be smart about whether to return
     0 or "" upon lookup failure */
  {
    Boolean numeric = FALSE;
    HashIterator hi;
    if(h) {
      hi = hashIteratorNew(h);
      result = hashIteratorNextValue(hi);
      hashIteratorDelete(hi);
      if(result && result->type == VTNumber) {
        numeric = TRUE;
      }
    }
    if(numeric) {
      val->type = VTNumber;
      val->data.number = 0;
    } else {
      val->type = VTString;
      val->data.string = strRegister("");
    }
  }
  return val;
  
}


/* ----------------------------------------------------------------------
   How many levels of sentence quotation does the word occupy?
 */
Value funcQuotes(List l, Value val, LexemGraph lg, Vector context)
{
  ValueStruct vas;
  Value a;
  LexemNode ln;

  a = evalTerm((Term) l->item, &vas, lg, context);
  if (a->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of function `quotes' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return ErrorValue;
  }
  ln = (LexemNode) a->data.lexemnode;
  if (!spec(ln)) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of function `quotes' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return ErrorValue;
  }
  val->data.number = ln->arc->quotes;
  val->type = VTNumber;
  return val;
}

/* ----------------------------------------------------------------------
   How many levels of parentheses does the word occupy?
 */
Value funcParens(List l, Value val, LexemGraph lg, Vector context) {
  ValueStruct vas;
  Value a;
  LexemNode ln;

  a = evalTerm((Term) l->item, &vas, lg, context);
  if (a->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of function `parens' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*"); 
    return ErrorValue;
  }
  ln = (LexemNode)a->data.lexemnode;
  if (!spec(ln)) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of function `parens' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*"); 
    return ErrorValue;
  }
  val->data.number = ln->arc->parens;
  val->type = VTNumber;
  return val;  
}

/* ----------------------------------------------------------------------
   How many levels of phrase quotation does the word occupy?
 */
Value funcPhrasequotes(List l, Value val, LexemGraph lg, Vector context)
{
  ValueStruct vas;
  Value a;
  LexemNode ln;

  a = evalTerm((Term) l->item, &vas, lg, context);
  if (a->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of function `phrasequotes' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*"); 
    return ErrorValue;
  }
  ln = (LexemNode) a->data.lexemnode;
  if (!spec(ln)) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of function `phrasequotes' not a lexeme node in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*"); 
    return ErrorValue;
  }
  val->data.number = ln->arc->phraseQuotes;
  val->type = VTNumber;
  return val;
}

/* ----------------------------------------------------------------------
   Perform the work of funcParent().
*/
int funcParentImpl(LexemGraph lg, Vector context, int i, int levelno, String condition) {
  int nrlevels = inputCurrentGrammar->noOfLevels;
  LexemNode ln;
  LevelValue lv;
  int counter = 0;

  Constraint c = NULL;

  /* if the string is the name of a unary constraint,
     use this constraint instead of label match  */
  if(condition) {
    c = hashGet(inputCurrentGrammar->constraints, condition);
    if(c && 1 != listSize(c->vars)) {
      c = NULL;
    }
  }
  
  /* unary call: look one edge above */
  if(!condition) {
    lv = vectorElement(context, nrlevels * i + levelno);
    if(!lv) {
      return 0;
    }
    ln = spec(lv->regents) ? lv->regents->item : lv->regents;
    if(nonspec(ln)) {
      return lg->max + 1;
    } else if (!ln) {
      return 0;
    } else {
      return ln->arc->to;
    }
  }

  /* binary call: ascend edges with the correct label,
     then return the dependent */
  while(1) {
    if(++counter > lg->max) {
#ifdef DEBUGPARENT
      cdgPrintf(CDG_DEBUG, " cycle!\n");
#endif
      /* cycle */
      return -1;
    }
    lv = vectorElement(context, nrlevels * i + levelno);
    if(!lv) {
      return 0;
#ifdef DEBUGPARENT
      cdgPrintf(CDG_DEBUG, "[no LV!] ");
#endif
    }
    if(spec(lv->regents)) {
      i = ((LexemNode)lv->regents->item)->arc->from;
#ifdef DEBUGPARENT
      cdgPrintf(CDG_DEBUG, "[%d] ", i);
#endif
    }

    if(!lv || !spec(lv->regents)) break;

    /* call constraint */
    if(c) {
      Constraint thisConstraint = evalCurrentConstraint;
      if(evalUnaryConstraint(c, NULL, context, lv)) {
        evalCurrentConstraint = thisConstraint;
      } else {
        evalCurrentConstraint = thisConstraint;
        break;
      }
    }

    /* compare label */
    else {
      if(lv->label != condition) break;
    }
  }
#ifdef DEBUGPARENT
  cdgPrintf(CDG_DEBUG, "[returning %d] ", ((LexemNode)lv->dependents->item)->arc->to);
#endif
  return ((LexemNode)lv->dependents->item)->arc->to;

}

/* ----------------------------------------------------------------------
   Return the timepoint of the regent of a given word.

   If there is a second argument, the search is repeated as long as the
   edges in question bear that label.

   If the second argument is the name of a unary constraint, this
   constraint is called on each edge and the search continues until it
   fails.

*/
Value funcParent(List args, Value val, LexemGraph lg, Vector context) {
  Value v;
  int a;
  LexemNode ln;
  VarInfo vi = evalCurrentConstraint->vars->item;
  String condition = NULL;
  
  v = evalTerm((Term) args->item, val, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (1st arg) for `parent' in constraint `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return ErrorValue;
  }
  ln = (LexemNode) val->data.lexemnode;
  if (!spec(ln)) {
    return ErrorValue;
  }
  a = ln->arc->from;
  
#ifdef DEBUGPARENT
  cdgPrintf(CDG_DEBUG, "parent(%s", ln->arc->word);
#endif
  
  if(args->next) {
    args = args->next;
    v = evalTerm((Term) args->item, val, lg, context);
    if (val->type != VTString) {
      cdgPrintf(CDG_WARNING,
                "WARNING: type mismatch (2st arg) for `parent' in constraint `%s'\n",
                evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      printTypeError(CDG_WARNING, val, VTString);
      return ErrorValue;
    }
  condition = val->data.string;

#ifdef DEBUGPARENT
  cdgPrintf(CDG_DEBUG, ", %s", condition);
#endif
  }

#ifdef DEBUGPARENT
  cdgPrintf(CDG_DEBUG, ") = ");
#endif

  v->type = VTNumber;
  v->data.number = funcParentImpl(lg, context, a, vi->level->no, condition);

#ifdef DEBUGPARENT
  cdgPrintf(CDG_DEBUG, "%.0f.\n", v->data.number);
#endif

  return v;
}



/* ----------------------------------------------------------------------
   Return the maximal length of dependency chains under a given word.
*/
Value funcHeight(List args, Value val, LexemGraph lg, Vector context)
{
  Value v;
  LexemNode ln;
  Term t = args->item;
  Level level;

  v = evalTerm(t, val, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (1st arg) for `height' in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return ErrorValue;
  }
  ln = (LexemNode) val->data.lexemnode;
  if (!spec(ln)) {
    return ErrorValue;
  }

  /* deduce level to search */
  if (NULL == (level = findLevel(t->data.peek.varinfo->levelname))) {
    cdgPrintf(CDG_ERROR, "ERROR: level `%s' not found in constraint `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return ErrorValue;
  }

  v->data.number = funcHeightImpl(ln, level->no, lg, context);
  v->type = VTNumber;
  return v;
}

/* ----------------------------------------------------------------------
   Perform the work of funcHeight().
*/
int funcHeightImpl(LexemNode ln, int levelno, LexemGraph lg, Vector context) {
  int i;
  LevelValue lv;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  int result = 0;
  
  for(i = lg->min; i <= lg->max; i++) {
    int index = nrlevels * i + levelno;
    int height = 0;
    if(index >= vectorSize(context)) break;
    
    while(1) {
      height++;
      if(height > lg->max) break;
      lv = vectorElement(context, index);
      if(!lv || !spec(lv->regents)) break;
      if(((LexemNode)(lv->regents->item)) == ln) {
        if(height > result) {
          result = height;
        }
        break;
      }
      index = nrlevels * ((LexemNode)(lv->regents->item))->arc->from + levelno;
    }
  }
  return result;
}


/* ----------------------------------------------------------------------
 * Returns predictions attached by gnWritePrediction().
 */
Value funcPredict(List args, Value val, LexemGraph lg, Vector context)
{
   Value a;
   ValueStruct vsa;
   LexemNode ln;
   int slot;
   Value key;
   
   a=evalTerm((Term)args->item, &vsa, lg, context);
   if (a->type!=VTLexemNode) {
     cdgPrintf(CDG_WARNING,
	       "WARNING: argument 1 of function `predict' not an id in constraint `%s'\n",
	       evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
     return ErrorValue;
   }
   ln=(LexemNode)a->data.lexemnode;

   args = args->next;

   if (((Term)args->item)->type!=TTString) {
     cdgPrintf(CDG_WARNING,
	       "WARNING: argument 2 of function `predict' not a string in constraint `%s'\n",
	       evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
     return ErrorValue;
   }

   slot=predictFindPredictor(((Term)args->item)->data.string);

   args=args->next;

   key=evalTerm((Term)args->item, &vsa, lg, context);
   if (key->type == VTString) {
     a = gnReadPrediction(ln->grapheme, slot, key->data.string);
   } else if (key->type == VTNumber) {
     String k = strPrintf("%.0f", key->data.number);
     a = gnReadPrediction(ln->grapheme, slot, k);
   } else {
     cdgPrintf(CDG_WARNING,
	       "WARNING: argument 3 of function `predict' not a string in constraint `%s'\n",
	       evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
     return ErrorValue;
   }
   if(!a) {
     val->data.number = 1.0;
     val->type = VTNumber; 
   } else {
     val->type = a->type;
     if(VTString == a->type) {
       val->data.string = a->data.string;
     } else {
       val->data.number = a->data.number;
     }
   }
   return val;
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
