/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/* -------------------------------------------------------------------------
 * @addtogroup LevelValue LevelValue - maintainance of level values
 * @author Ingo Schroeder (see also AUTHORS and THANKS for more)
 * @date 24/4/97
 *
 * $Id: levelvalue.c,v 1.73 2006-01-13 20:09:13 micha Exp $
 *
 * @{
 */

/* ---------------------------------------------------------------------- */
#include <config.h>

#include <stdio.h>
#include <string.h>
#include <blah.h>
#include "cdg.h"
#include "input.h"
#include "eval.h"
#include "statistics.h"
#include "levelvalue.h"
#include "lexemgraph.h"
#include "increment.h"
#include "hook.h"
#include "constraintnet.h"

/* ---------------------------------------------------------------------- */
/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * checks whether two level value are mutually compatible
 */
Boolean lvCompatible(LexemGraph lg, LevelValue a, LevelValue b)
{
   List atop = a->regents;
   List abot = a->dependents;

   List btop = b->regents;
   List bbot = b->dependents;

#define LVCOMPATIBLE 0

   if (!lgCompatibleSets(lg, abot, bbot)) {
#if LVCOMPATIBLE
     cdgPrintf(CDG_INFO, "*** lgcs abot bbot\n");
#endif
     return (FALSE);
   }

   if (atop != NULL && atop != NONSPEC) {
     if (!lgCompatibleSets(lg, atop, bbot)) {
#if LVCOMPATIBLE
       cdgPrintf(CDG_INFO, "*** lgcs atop bbot\n");
#endif
       return (FALSE);
     }

      if (btop != NULL && btop != NONSPEC) {
	 if (!lgCompatibleSets(lg, abot, btop)) {
#if LVCOMPATIBLE
	   cdgPrintf(CDG_INFO, "*** lgcs abot btop\n");
#endif
	   return (FALSE);
	 }
	 if (!lgCompatibleSets(lg, atop, btop)) {
#if LVCOMPATIBLE
	   cdgPrintf(CDG_INFO, "*** lgcs atop btop\n");
#endif
	   return (FALSE);
	 }
      }
   } else {
      if (btop != NULL && btop != NONSPEC &&
	  !lgCompatibleSets(lg, btop, abot)) {
#if LVCOMPATIBLE
	cdgPrintf(CDG_INFO, "*** lgcs btop abot\n");
#endif
	return (FALSE);
      }
   }
   return (TRUE);
}

/* ----------------------------------------------------------------------
 * builds level value for a constraint node, level and lexem nodes
 */
LevelValue lvNew(List dependents, Level level, String label, List regents)
{
  LevelValue lv;

  lv = (LevelValue) memMalloc(sizeof (LevelValueStruct));
  lv->constraints = NULL;
  lv->no = -1;
  lv->score = 0.0;
  lv->limit = 1.0;
  lv->isDeleted = FALSE;
  lv->indexWRTNet = 0;
  lv->level = level;
  lv->label = strRegister(label);

  /* check for special regents */
  if (spec(regents)) {
    lv->regents = listClone(regents);
  } else {
    lv->regents = regents;
  }

  /* dependents are never special */
  lv->dependents = listClone(dependents);

  return lv;
}

/* ----------------------------------------------------------------------
 * reset the values of an already existing levelvalue
 * - all list-containers are owned by the levelvalue 
 * - old dependents- and regents-lists are reused
 */
LevelValue lvSet(LevelValue lv, List dependents, Level level, String label,
		 List regents)
{
  lv->no = -1;
  lv->score = 0.0;
  lv->limit = 1.0;
  lv->isDeleted = FALSE;
  lv->indexWRTNet = 0;
  lv->level = level;
  lv->label = label;

  /* check for special regents */
  if (regents != NULL && regents != NONSPEC) {
    lv->regents = listCopy(lv->regents, regents);
  } else {
    lv->regents = NULL;
  }

  /* dependents are never special */
  lv->dependents = listCopy(lv->dependents, dependents);

  /* old violations are deleted */
  listDelete(lv->constraints);
  lv->constraints = NULL;

  return lv;
}

/* ----------------------------------------------------------------------
 * delete a levelvalue and all owned containers
 */
void lvDelete(LevelValue lv) 
{
  if (!lv) {
    return;
  }

  listDelete(lv->dependents); 
  if (spec(lv->regents)) {
    listDelete(lv->regents);
  }
  listDelete(lv->constraints);
  memFree(lv);
}

/* ----------------------------------------------------------------------
   Prints a representation of the level value LV.

   If ISDELETED is set, it is used to suppress the display of lexeme nodes
   that are currently deleted. If an LV has multiple dependents, and some of
   them are deleted, those are not shown even if PRINTALTERNATIVES is set.
   However, if *all* dependents or regents are deleted, one of them is
   shown and enclosed in square brackets to indicate invalidity.

   If PRINTALTERNATIVES is 0, only one modifier will ever be printed.
   It it is 1, an indication of how many dependents are actually present
   will be added to the output, like {3}. If it is 2, all dependents and
   regents will be printed. Since this can take huge amounts of space, it
   is only ever useful for debugging algorithms that have to maintain the
   path criterion.
   
 */
void lvPrint(long unsigned int mode, LevelValue lv,
             ByteVector isDeleted, int printAlternatives) {
   List l;
   Boolean first;
   int i, j, k;
   List showme = NULL;
   
   cdgPrintf(mode, "%0.2f (%02d-%02d) ", 
      lv->score,
      ((LexemNode)(lv->dependents->item))->arc->from,
      ((LexemNode)(lv->dependents->item))->arc->to);
   if (spec(lv->regents)) {
      cdgPrintf(mode, "(%02d-%02d) ",
	 ((LexemNode)(lv->regents->item))->arc->from,
	 ((LexemNode)(lv->regents->item))->arc->to);
   } else {
      cdgPrintf(mode, "        ");
   }
   cdgPrintf(mode, "%10s: ", lv->level->id);


   /* print modifier */
   for (l = lv->dependents; l != NULL; l = l->next) {
      LexemNode ln = l->item;
      if (isDeleted && bvElement(isDeleted, ln->no))
	 continue;
      showme = listAppendElement(showme, ln);
   }

   first = TRUE;
   for(l = showme; l != NULL; l = l->next) {
     LexemNode ln = l->item;
     if(!first) {
       cdgPrintf(mode, "/");
     }
     cdgPrintf(mode, "%s", ln->lexem->description);
     if(0 == printAlternatives) {
       break;
     }
     if(1 == printAlternatives) {
       int len = listSize(showme);
       if(1 != len) {
         cdgPrintf(mode, "{%d}", len);
       }
       break;
     }
     first = FALSE;
   }
   if(!showme) {
     cdgPrintf(mode, "[%s]", ((LexemNode)(lv->dependents->item))->lexem->description);
   }
   listDelete(showme);
   
   /* take care to make all --LABEL--> arrows the same length */
   cdgPrintf(mode, "--");

   i = inputCurrentGrammar->maxLabelLength - strlen(lv->label);
   j = i / 2;
   for(k = 0; k < j; k++) cdgPrintf(mode, "-");

   cdgPrintf(mode, "%s", lv->label);

   if(i % 2 == 1) j++;
   for(k = 0; k < j; k++) cdgPrintf(mode, "-");

   cdgPrintf(mode, "-->");

   /* print modifiee */
   if (lv->regents == NONSPEC) {
     cdgPrintf(mode, "NONSPEC");
     return;
   }
   if (lv->regents == NULL) {
     cdgPrintf(mode, "NIL");
     return;
   }

   showme = NULL;
   for (l = lv->regents; l != NULL; l = l->next) {
     LexemNode ln = l->item;
     if (isDeleted && bvElement(isDeleted, ln->no))
       continue;
     showme = listAppendElement(showme, ln);
   }
   first = TRUE;
   for(l = showme; l != NULL; l = l->next) {
     LexemNode ln = l->item;
     if(!first) {
       cdgPrintf(mode, "/");
     }
     cdgPrintf(mode, "%s", ln->lexem->description);
     if(0 == printAlternatives) {
       break;
     }
     if(1 == printAlternatives) {
       int len = listSize(showme);
       if(1 != len) {
         cdgPrintf(mode, "{%d}", len);
       }
       break;
     }
     first = FALSE;
   }
   if(spec(lv->regents) && !showme) {
     cdgPrintf(mode, "[%s]", ((LexemNode)(lv->regents->item))->lexem->description);
   }
   listDelete(showme);
}

/* ----------------------------------------------------------------------
 * return the Direction of an LV
 */
Direction getDir(LevelValue x)
{
  if (x->regents == NULL)
    return Nil;
  else if (x->regents == NONSPEC)
    return Right;
  else if (((LexemNode) (x->regents->item))->arc->from <
	   ((LexemNode) (x->dependents->item))->arc->from)
    return Left;
  else
    return Right;
}


/* ----------------------------------------------------------------------
 * return the Connexion of two LVs
 */
Connexion getCon(LevelValue a, LevelValue b)
{
  GraphemNode atop = NULL, btop = NULL;

  GraphemNode
    abot = ((LexemNode) (a->dependents->item))->grapheme,
    bbot = ((LexemNode) (b->dependents->item))->grapheme;

  if (a->regents == NONSPEC)
    atop = NONSPEC;

  if (b->regents == NONSPEC)
    btop = NONSPEC;

  if (a->regents != NULL && a->regents != NONSPEC)
    atop = ((LexemNode) (a->regents->item))->grapheme;
  if (b->regents != NULL && b->regents != NONSPEC)
    btop = ((LexemNode) (b->regents->item))->grapheme;

  if (atop == btop && abot == bbot && atop != NONSPEC)
    return Identical;

  if (atop == bbot && abot == btop)
    return Inverse;

  if (atop == bbot)
    return Under;

  if (abot == btop)
    return Over;

  if (atop == btop && atop != NONSPEC)
    return Top;

  if (abot == bbot)
    return Bottom;

  return None;

}


/* ----------------------------------------------------------------------
 * Does this Signature apply to this LV?
 */
Boolean matchUnarySignature(LevelValue lv, Signature x)
{

   return (matchLevel(x->l1,lv->level) &&
      subsumesDirection(x->d1, getDir(lv)));

}

/* ----------------------------------------------------------------------
 * Does this Signature apply to these LVs?
 */
Boolean matchBinarySignature(LevelValue lva, LevelValue lvb, Signature x)
{

   return (matchLevel(x->l1, lva->level) &&
           matchLevel(x->l2, lvb->level) &&
      subsumesConnexion(x->conn, getCon(lva, lvb)) &&
      subsumesDirection(x->d1, getDir(lva)) &&
      subsumesDirection(x->d2, getDir(lvb)));

}


/* ----------------------------------------------------------------------
 * Define total ordering on LVs:
    compare by scores, and secondarily by their indexWRTNet
 */
Boolean lvCompare(LevelValue a, LevelValue b)
{
  if(a->score > b->score) return TRUE;
  if(a->score < b->score) return FALSE;
  return a->indexWRTNet < b->indexWRTNet;
}

/* ----------------------------------------------------------------------
 * like lvCompare(), but sorts only by limits.
 */
Boolean lvCompareByLimit(LevelValue a, LevelValue b)
{
   return (a->limit > b->limit);
}

/* ----------------------------------------------------------------------
 * like lvCompare(), but sorts by level & by time.
 */
Boolean lvCompareByNaturalOrder(LevelValue a, LevelValue b)
{
  if (((LexemNode)(a->dependents->item))->arc->from <
      ((LexemNode)(b->dependents->item))->arc->from) {
    return TRUE;
  }
  if (((LexemNode)(a->dependents->item))->arc->from >
      ((LexemNode)(b->dependents->item))->arc->from) {
    return FALSE;
  }
  if (a->level->no < b->level->no) {
    return TRUE;
  }

  return FALSE;

}

/* ----------------------------------------------------------------------
 * clone an LV
 */
LevelValue lvClone(LevelValue lv) {
  LevelValue result;

  if (!lv)
    return NULL;

  result = lvNew(lv->dependents,
		 lv->level, lv->label, lv->regents);

  result->no = lv->no;
  result->indexWRTNet = lv->indexWRTNet;
  result->score = lv->score;
  result->limit = lv->limit;

  return result;
}

/* ----------------------------------------------------------------------
   Is the LV physically compatible with the lexemes allowed in the
   Vector?
  
   This uses a Vector of Boolean as produced by lgRequireLexeme().
 */
Boolean lvVectorCompatible(LevelValue lv, ByteVector taboo)
{
  List l;
  Boolean found = FALSE;
  LexemNode ln;

  /* one modifiee must be allowed */
  if (spec(lv->regents)) {
    for (l = lv->regents; l; l = l->next) {
      ln = l->item;
      if (!bvElement(taboo, ln->no)) {
	found = TRUE;
	break;
      }
    }
    if (!found) {
      return FALSE;
    }
  }

  /* one modifier must be allowed */
  for (l = lv->dependents; l; l = l->next) {
    ln = l->item;
    if (!bvElement(taboo, ln->no)) {
      return TRUE;
    }
  }

  return FALSE;

}


/* ----------------------------------------------------------------------
   This function computes the `logical' number of an LV in the set of LVs
   that constitute a dependency analysis: it is the position it will have if
   the LVs are sorted by time index first, and by level second. Some
   indices may not be assigned in two cases:

   (a) if there are lexeme nodes that span more than one time interval,
   lvIndex() will leave out those that correspond to positions within the
   long word.

   (b) if individual levels have their useFlag reset, the indices
   corresponding to LVs of these levels will also be skipped.

   This allows the function to determine exactly what position an LV holds,
   without any other context, but can be wasteful on space. The structure
   Analysis from the module frobbing uses this indexing method; the
   structure Parse uses a compact method that always gives consecutive index
   numbers. */
int lvIndex(LevelValue lv) {
  return
    inputCurrentGrammar->noOfLevels *
    ((LexemNode)(lv->dependents->item))->arc->from + lv->level->no;
}

/* ----------------------------------------------------------------------
 * Returns FALSE if all the dependents or all the regents are deleted
 */
Boolean lvTestAllDeleted(LevelValue lv, ByteVector isDeleted)
{
  List l;
  Boolean modifiers_ok = FALSE;

  for (l = lv->dependents; l; l = l->next) {
    LexemNode ln = l->item;
    if (bvElement(isDeleted, ln->no) == FALSE)
      modifiers_ok = TRUE;
  }

  if (modifiers_ok == FALSE)
    return FALSE;
  
  if (lv->regents == NULL)
    return TRUE;
  
  if (lv->regents == NONSPEC)
    return TRUE;

  for (l = lv->regents; l; l = l->next) {
    LexemNode ln = l->item;
    if (bvElement(isDeleted, ln->no) == FALSE)
      return TRUE;
  }
  return FALSE;
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
/** @} */
