/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/* ---------------------------------------------------------------------------
 * @addtogroup Lexemgraph Lexemgraph - maintainance of lexem graphs
 * @author Ingo Schroeder (see also AUTHORS and THANKS for more)
 * @date 1997-03-04
 *
 * $Id: lexemgraph.c,v 1.157 2006-02-22 16:47:16 foth Exp $
 * @{
 */

/* ------------------------------------------------------------------------- */
#include <config.h>

#include <ctype.h>
#include <stdio.h>
#include <blah.h>
#include <string.h>
#include <limits.h>
#include "hook.h"
#include "input.h"
#include "lexemgraph.h"
#include "predict.h"
#include "eval.h"
#include "increment.h"
#include "set.h"
#include "tagger.h"
#include "parse.h"
#include "chunker.h"

/* -- VARIABLES ------------------------------------------------------------ */

/**
  This variable controls whether the levelvalues should be deflated if
  they are equivalent. Usually, we want this switched on, only for
  testing a value of FALSE might be appropriate.
*/
Boolean lgCompactLVs=TRUE;


/* -- STATIC FUNCTIONS ----------------------------------------------------- */
static long long computeNoOfPathsFromStart(LexemGraph lg, GraphemNode gn,
					   long long sofar);
static long long computeNoOfPathsToEnd(LexemGraph lg, GraphemNode gn,
				       long long sofar);

/* ----------------------------------------------------------------------------
  @brief (re-)computes the distance matrix LexemGraph::distance

  This function computes the distance between any two lexeme nodes in @a lg and
  stores the result in @a lg->distance.
*/
void lgComputeDistances(LexemGraph lg)
{
  GraphemNode gn, gnNext;
  int noOfNodes = vectorSize(lg->graphemnodes);
  int i, j, k, ji, ik, jk, ij;

  /* arrayNew initializes the cells with NULL */
  if (!lg->distance) {
    lg->distance = arrayNew(noOfNodes, noOfNodes, 0);
  } else {
    arraySetAllElements(lg->distance, (Pointer) 0);
  }

  for (i = 0; i < noOfNodes; i++) {
    gn = (GraphemNode) vectorElement(lg->graphemnodes, i);

    /* If this gn has been deleted from the problem, leave its distances all
     * at 0. This is necessary because the distance info does double duty as
     * compatibility info. */
    if (lgAreDeletedNodes(lg, gn->lexemes)) {
      continue;
    }
    for (j = 0; j < noOfNodes; j++) {
      gnNext = (GraphemNode) vectorElement(lg->graphemnodes, j);
      if (lgAreDeletedNodes(lg, gnNext->lexemes)) {
	continue;
      }
      if (gn->arc->to == gnNext->arc->from) {
	arraySetElement(lg->distance, (Pointer) 1, i, j);
      }
    }
  }
  /* Warshall's algorithm for transitive closure */
  for (i = 0; i < noOfNodes; i++) {
    for (j = 0; j < noOfNodes; j++) {
      ji = (int)arrayElement(lg->distance, j, i);

      if (ji > 0) {
	for (k = 0; k < noOfNodes; k++) {
	  ik = (int)arrayElement(lg->distance, i, k);
	  jk = (int)arrayElement(lg->distance, j, k);

	  if (ik > 0 && (jk == 0 || ji + ik < jk)) {
	    arraySetElement(lg->distance, (Pointer) (ji + ik), j, k);
	  }
	}
      }
    }
  }

  for (i = 0; i < noOfNodes; i++) {
    for (j = 0; j < noOfNodes; j++) {
      ij = (int)arrayElement(lg->distance, i, j);
      ji = (int)arrayElement(lg->distance, j, i);

      if (ij == 0 && ji > 0) {
	arraySetElement(lg->distance, (Pointer) (-ji), i, j);
      }
    }
  }
}


/* ----------------------------------------------------------------------------
  @brief computes LexemGraph::noOfPathsFromStart

  This function computes the number of paths leading to @a gn from the start of
  @a lg. If @a gn corresponds to a start node, this is simply the number of
  lexeme nodes sprung from @a gn. Otherwise it is that number multiplied by the
  sum of the numbers of paths leading from the start to immediately preceding
  grapheme nodes. If @a gn is deleted, the number is always zero.
*/
long long computeNoOfPathsFromStart(LexemGraph lg, GraphemNode gn,
				    long long sofar)
{
  long long i, sum;
  GraphemNode prevNode;
  long long noOfPathsFromStart;
  int maximal = vectorSize(lg->graphemnodes);

  noOfPathsFromStart = lg->noOfPathsFromStart[gn->no];

  if (noOfPathsFromStart >= 0) {
    return noOfPathsFromStart;
  }

  if (!gn->live) {
    return lg->noOfPathsFromStart[gn->no] = 0;
  }

  if (sofar > maximal) {
    cdgPrintf(CDG_ERROR, "ERROR: lexem lattice contains cycles\n");
    abort();
  }

  if (gn->arc->from == lg->min) {
    return lg->noOfPathsFromStart[gn->no] = 1;
  }
  sum = 0;
  for (i = 0; i < maximal; i++) {
    prevNode = (GraphemNode) vectorElement(lg->graphemnodes, i);

    if (!prevNode->live || prevNode->arc->to != gn->arc->from)
      continue;

    sum += computeNoOfPathsFromStart(lg, prevNode, sofar + 1);

    if (sum < 0) {
      cdgPrintf(CDG_ERROR,
		"PANIC: integer overflow in computeNoOfPathsFromStart()!\n");
      abort();
    }
  }
  return lg->noOfPathsFromStart[gn->no] = sum;
}

/* ----------------------------------------------------------------------------
  @brief computes LexemGraph::noOfPathsToEnd

  This function computes the number of paths leading from @a gn to the end of
  @a lg. If @a gn corresponds to an end node, this is simply the number of
  lexeme nodes sprung from @a g. Otherwise it is that number multiplied by the
  sum of the numbers of paths leading to the end from immediately following
  grapheme nodes. If @a gn is deleted, the number is always zero.
*/
long long computeNoOfPathsToEnd(LexemGraph lg, GraphemNode gn,
				long long sofar)
{
  long long i, sum;
  GraphemNode nextNode;
  long long noOfPathsToEnd;
  int maximal = vectorSize(lg->graphemnodes);

  noOfPathsToEnd = lg->noOfPathsToEnd[gn->no];

  if (noOfPathsToEnd >= 0)
    return noOfPathsToEnd;

  if (!gn->live) {
    return lg->noOfPathsToEnd[gn->no] = 0;
  }

  if (sofar > maximal) {
    cdgPrintf(CDG_ERROR, "ERROR: lexem lattice contains cycles\n");
    abort();
  }

  if (gn->arc->to == lg->max) {
    return lg->noOfPathsToEnd[gn->no] = 1;
  }

  sum = 0;
  for (i = 0; i < maximal; i++) {
    nextNode = (GraphemNode) vectorElement(lg->graphemnodes, i);

    if (!nextNode->live || nextNode->arc->from != gn->arc->to)
      continue;

    sum += computeNoOfPathsToEnd(lg, nextNode, sofar + 1);
    if (sum < 0) {
      cdgPrintf(CDG_ERROR,
		"PANIC: integer overflow in computeNoOfPathsFromStart()!\n");
      abort();
    }
  }
  return lg->noOfPathsToEnd[gn->no] = sum;
}

/* ----------------------------------------------------------------------------
  @brief computes # of paths possible in the graph, given the current state of
  deletions.

  This function computes the number of paths possible in @a lg, according to
  the state of its Vector LexemGraph::isDeletedNode. It calls
  computeNoOfPathsToEnd() and computeNoOfPathsFromStart() for each lexeme node.
  The total number of all paths is the sum of all numbers of paths leading to
  grapheme nodes that are end nodes.
*/
void lgComputeNoOfPaths(LexemGraph lg)
{
  int i;
  LexemNode ln;
  GraphemNode gn;
  List l;
  int noOfNodes = vectorSize(lg->graphemnodes);

  /* append cells needed by iSearch */
  for (i = bvSize(lg->isDeletedNode); i < vectorSize(lg->nodes); i++) {
    bvSetElement(lg->isDeletedNode, FALSE, i);
  }

  /* maybe allocate the arrays */
  if (!lg->noOfPathsFromStart) {
    lg->noOfPathsFromStart = memMalloc(sizeof (long long) * noOfNodes);
  }
  if (!lg->noOfPathsToEnd) {
    lg->noOfPathsToEnd = memMalloc(sizeof (long long) * noOfNodes);
  }

  /* recompute ambiguities of each word */
  for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
    gn = vectorElement(lg->graphemnodes, i);
    gn->live = FALSE;
    for (l = gn->lexemes; l != NULL; l = l->next) {
      ln = l->item;
      if (!bvElement(lg->isDeletedNode, ln->no)) {
	gn->live = TRUE;
	break;
      }
    }

    /* no paths from start/to end yet */
    lg->noOfPathsFromStart[i] = -1;
    lg->noOfPathsToEnd[i] = -1;
  }

  /* compute # of paths for all nodes */
  lg->noOfPaths = 0;
  for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
    gn = (GraphemNode) vectorElement(lg->graphemnodes, i);

    if (lg->noOfPathsFromStart[gn->no] < 0) {
      computeNoOfPathsFromStart(lg, gn, 0);
    }
    if (lg->noOfPathsToEnd[gn->no] < 0) {
      computeNoOfPathsToEnd(lg, gn, 0);
    }

    /* count # of paths in lg */
    if (gn->live && gn->arc->from == lg->min)
      lg->noOfPaths += lg->noOfPathsToEnd[gn->no];
  }
}

/* ----------------------------------------------------------------------------
  @brief partitions a set of lexeme nodes into equivalence classes

  This function partitions the set of lexeme nodes of @a gn into equivalence
  classes. The equivalence relation used is the function inputCompareLeByAtts()
  with the argument @a features. The function returns a new List of new lists
  of lexemes. (The latter are re-used in ConstraintNode structures, the former
  are deallocated by cnBuildNodes().)
*/
List lgPartitions(GraphemNode gn, BitString features)
{

/** @code

  result = [];

  FOR each lexeme l:
   IF l fits into one of the known classes,
    insert l there;
   ELSE
    create new class [lexem];
    insert the new class into result;
   FI
  ROF

  return result.

  @endcode
*/

   List result, p;
   List q;
   List class;
   LexemNode ln;
   LexiconItem lexeme1 = NULL, lexeme2 = NULL;

   result = NULL;

   /* partitioning the empty set
      yields the set with one class: { {} } */
   if(!gn) {
     return listPrependElement(NULL, NULL);
   } else if (gn == NONSPEC) {
     return listPrependElement(NULL, NONSPEC);
   }
#ifdef DEBUGPARTITION
   cdgPrintf(CDG_DEBUG,
	     "DEBUG: about to classify %d lexemes\n",
	     listSize(gn->lexemes));

   for (p = gn->lexemes; p != NULL; p = p->next) {
      cdgPrintf(CDG_DEBUG, "%s ", ((LexemNode)( p->item))->lexem->description);
   }
   cdgPrintf(CDG_DEBUG, "\n");
#endif

   /* partitioning a normal set yields sets of equivalent lexemes:
      { { das_ART_nom,das_ART_acc }, { das_PREL_nom,das_PREL_acc} } */
   for (p = gn->lexemes; p != NULL; p = p->next) {
      ln = p->item;
      lexeme1 = ln->lexem;

#ifdef DEBUGPARTITION
      cdgPrintf(CDG_DEBUG, "DEBUG: classifying lexeme %s\n", lexeme1->description);
#endif

      /* compare to the first item in each class */

      for (q = result; q != NULL; q = q->next) {
	 class = (List) q->item;
	 lexeme2 = ((LexemNode)( class->item))->lexem;

#ifdef DEBUGPARTITION
	 cdgPrintf(CDG_DEBUG, "DEBUG: comparing to lexeme %s... ", lexeme2->description);
#endif

	 /* insert into correct class */
	 if (lgCompactLVs &&
             lexeme1->word == lexeme2->word &&
             !inputCompareLeByAtts(lexeme1, lexeme2, features)) {
#ifdef DEBUGPARTITION
	    cdgPrintf(CDG_DEBUG, "equivalent.\n");
	    cdgPrintf(CDG_DEBUG, "DEBUG: adding lexeme to this class.\n");
#endif
	    class = listAppendElement(class, ln);
	    break;
	 } else {
#ifdef DEBUGPARTITION
	    cdgPrintf(CDG_DEBUG, "differs.\n");
#endif
	 }
      }

      /* open new equivalence class */

      if (q == NULL) {

#ifdef DEBUGPARTITION
	 cdgPrintf(CDG_DEBUG, "DEBUG: creating new equivalence class.\n");
#endif

	 class = listPrependElement(NULL, ln);
	 result = listAppendElement(result, class);

      }
#ifdef DEBUGPARTITION
      cdgPrintf(CDG_DEBUG, "DEBUG: partition now looks like this:  ");
      for (q = result; q != NULL; q = q->next)
	 cdgPrintf(CDG_DEBUG, "%d ", listSize(q->item));
      cdgPrintf(CDG_DEBUG, "\n");
#endif

   }

#ifdef DEBUGPARTITION
   cdgPrintf(CDG_DEBUG, "DEBUG: %d lexemes partitioned into %d classes.\n\n",
      listSize(gn->lexemes), listSize(result));
#endif


   return (result);

}

 
/* ----------------------------------------------------------------------------
  @brief initializes the lexemgraph

  This function returns a new LexemGraph structure with all fields initialized
  to meaningless values. In particular, it contains no nodes whatsoever.
*/
LexemGraph lgNewInit()
{
  LexemGraph lg;

  lg = (LexemGraph) memMalloc(sizeof(LexemGraphStruct));
  lg->lattice = NULL;
  lg->graphemnodes = vectorNew(20);
  lg->max = 0;
  lg->min = INT_MAX;
  lg->distance = NULL;
  lg->isDeletedNode = bvNew(20);
  lg->noOfPathsFromStart = NULL;
  lg->noOfPathsToEnd = NULL;
  lg->nodes = vectorNew(20);
  lg->chunks = NULL;
  return lg;
}


/* ----------------------------------------------------------------------
   Allocate a new GraphemNode.
*/
GraphemNode gnNew(void) {
  GraphemNode result = memMalloc(sizeof (GraphemNodeStruct));
  result->lexemes = NULL;
  result->predictions = NULL;
  result->chunk = NULL;
  return result;
}


/* ----------------------------------------------------------------------
   Return a printable representation of predictions made for GN.
*/
String gnFormatPredictions(GraphemNode gn) {

  String result = NULL;
  if(gn->predictions) {
    int i;
    List arcs = gn->lexemgraph->lattice->arcs;

    for (i=0; i < vectorSize(gn->predictions); i++) {
      Hashtable ht;
      String name;
      if(0 == i) {
        name = "POS";
      } else {
        Predictor p = vectorElement(cdgPredictors, i-1);
        name = p->name;
      }

      /* retrieve all predictions made */
      if ((ht=(Hashtable)vectorElement(gn->predictions,i))!=NULL) {
	List keys = listSort(hashListOfKeys(ht), strVersionCompare);
        List l;
	for(l = keys; l != NULL; l = l->next) {
          String key = l->item;
	  Value value = hashGet(ht, key);
          String keyrep = key;

          /* if the key is an integer, it is probably intended to denote a
             regent. Therefore, print the word as well as its position for
             debugging. */
          if(strspn(key, "0123456789") == strlen(key)) {
            int x;
            sscanf(key,"%d",&x);
            if(x+1 > 0 && listSize(arcs) >= x+1) {
              Arc arc = listNthElement(arcs, x+1);
              keyrep = strPrintf("%s(%s)", key, arc->word);
            }
          }

          result = strCat(result,
                          strPrintf("%s: %s = %s\n", name, keyrep,
                                    formatValue(value, 0)));
        }
	listDelete(keys);
      }
    }
    return result;
  }
  else {
    return strRegister("");
  }
}

/* ----------------------------------------------------------------------
   Free a GraphemNode.
*/
void gnDelete(GraphemNode gn) {
  listDelete(gn->lexemes);
  if(gn->predictions) {
    int i;
    Hashtable pred;
    for (i=0; i < vectorSize(gn->predictions); i++) {
      if ((pred=(Hashtable)vectorElement(gn->predictions,i))!=NULL) {
	HashIterator hi = hashIteratorNew(pred);
	String key;
	while (NULL != (key = hashIteratorNextKey(hi))) {
	  String value = hashGet(pred, key);
	  cdgFreeString(key);
	  cdgFreeString(value);
	}
	hashIteratorDelete(hi);
	hashDelete(pred);
      }
    }
    vectorDelete(gn->predictions);
  }
  memFree(gn);
}

/* ----------------------------------------------------------------------
   Attach a prediction to a GraphemNode.
*/
void gnWriteNumber(GraphemNode gn, int slot, String key, Number x) {
  Value v;
  Hashtable pred;
  if (slot < 0) return;
  if(!gn->predictions) {
    gn->predictions = vectorNew(vectorSize(cdgPredictors)+1);
  }
  if (vectorSize(gn->predictions)<=slot ||
      (pred=(Hashtable)vectorElement(gn->predictions,slot))==NULL) {
    pred=hashNew(10, 0.8, hashStringHashFunction, hashStringEqualFunction);
    vectorSetElement(gn->predictions, pred, slot);
  }
  v = memMalloc(sizeof(ValueStruct));
  v->type = VTNumber;
  v->data.number = x;
  hashSet(pred, key, v);
}

/* ----------------------------------------------------------------------
   Attach a prediction to a GraphemNode.
*/
void gnWriteString(GraphemNode gn, int slot, String key, String s) {
  Value v;
  Hashtable pred;
  if (slot < 0) return;
  if(!gn->predictions) {
    gn->predictions = vectorNew(vectorSize(cdgPredictors)+1);
  }
  if (vectorSize(gn->predictions) <= slot ||
      (pred=(Hashtable)vectorElement(gn->predictions,slot))==NULL) {
    pred=hashNew(10, 0.8, hashStringHashFunction, hashStringEqualFunction);
    vectorSetElement(gn->predictions, pred, slot);
  }
  v = memMalloc(sizeof(ValueStruct));
  v->type = VTString;
  v->data.string = s;
  hashSet(pred, key, v);
}

/* ----------------------------------------------------------------------
   Read a prediction from a GraphemNode.
*/
Value gnReadPrediction(GraphemNode gn, int slot, String key) {
  Hashtable pred;
  if (slot < 0 || !gn->predictions ||
      vectorSize(gn->predictions) <= slot) {
    return NULL;
  }
  pred=(Hashtable)vectorElement(gn->predictions,slot);
  if(!pred) {
    return NULL;
  }
  return hashGet(pred, key);
}


/* ----------------------------------------------------------------------
   Which key carries the highest score as a prediction?
*/
String gnHighestPrediction(GraphemNode gn) {

  HashIterator hi;
  Hashtable pred;
  String key, result = NULL;
  Number record = -1;
  
  if(!gn->predictions || vectorSize(gn->predictions)==0) {
    return NULL;
  }
  pred=vectorElement(gn->predictions,0);
  hi = hashIteratorNew(pred);
  while (NULL != (key = hashIteratorNextKey(hi))) {
    Value v = hashGet(pred, key);
    if(v->type == VTNumber && v->data.number > record) {
      record = v->data.number;
      result = key;
    }
  }
  hashIteratorDelete(hi);

  return result;
}




/* ----------------------------------------------------------------------
   Add information from partial parsers to a lexeme graph.
*/
void lgTag(LexemGraph lg) {

  Chunker chunker;
  List chunks;
  
  /* tagging */
  taggerTag(lg);

  /* chunking */
  if(RealChunker == chunkerMode) {
    chunker = chunkerNew(DefaultChunker, lg);
    chunks = chunkerChunk(chunker);
    if (hkVerbosity & CDG_DEBUG) {
      chunkerPrintChunks(CDG_DEBUG, chunks);
    }
    chunkerDelete(chunker);
  }

  /* other predictors */
  predictAll(lg);

}


/* ----------------------------------------------------------------------------
  @brief Insert lexeme nodes into the LexemGraph that correspond the Arc.

  This function builds all possible lexeme nodes for the specific
  @a arc and adds them to @a lg. It fails if there is no
  matching entry in the lexicon.
*/
Boolean lgNewIter(LexemGraph lg, Arc arc)
{
  GraphemNode gn;
  LexemNode ln;
  LexiconItem le;
  List j, entries = NULL;

  /** Maybe undo capitalisation introduced by orthographic convention.

     If the written word is uppercase, but that uppercase-ness is suspect
     because it is at the start of a phrase and might be mere orthographic
     convention, we have to decide which version we use for lexicon lookup.

     If our lexicon contains items for the lower-case version but none for
     the upper-case versions, we use only those; if it contains only
     items for the upper-case version, we use those; and if it contains
     neither, we allow both and hope that there is a lexical template which
     will catch this word.

     We do not use the obvious solution - look up both versions whenever a
     word is spurious - because it has the following defect: If a sentence
     starts with `Der', some naive lexical template could introduce a noun
     reading, and if POS tagging allows, it might actually survive even
     though it is exceedingly unlikely. Since we do not want this, we
     effectively force the reading to be `der'.

     Moral: If you really need to have open-class items in your lexicon
     that are near-homonymous with closed-class items, you can bloody well
     write proper lexicon items for them and not templates. */

  if(lgSpuriousUppercase(lg, arc)) {
    String low = strCopy(arc->word);
    Boolean have_upper, have_lower;

    /* skip initial ( or " */
    int index = 0;
    while(((unsigned char *)arc->word)[index] == '"' ||
          ((unsigned char *)arc->word)[index] == '(') {
      index++;
    }
    
    low[index] = tolower(low[index]);
    have_upper = inputLexiconQuery(arc->word);
    have_lower = inputLexiconQuery(low);

    if(have_upper && !have_lower) {
      entries = listAppendList(entries, inputLexiconGet(arc->word));
    }
    else if(have_lower && !have_upper) {
      entries = listAppendList(entries, inputLexiconGet(low));
      arc->word = strCopy(low);
    } else {
      entries = listAppendList(entries, inputLexiconGet(low));
      entries = listAppendList(entries, inputLexiconGet(arc->word));
    }
    memFree(low);
  }

  /** Much the same goes for words in ALL UPPER CAPS, except that those can
     occur anywhere in a sentence, not only at the start, and we have to
     check three different spellings instead of two. */
  else if(inputALLUPPER(arc->word)) {
        String low = strCopy(arc->word);
        String startUpper = strCopy(arc->word);
        Boolean have_upper, have_lower, have_startUpper;
        int i;
    for (i=0;  i < strlen((char *)arc->word); i++){
        low[i] = tolower(low[i]);
    }
    for (i=1;  i < strlen((char *)arc->word); i++){
        startUpper[i] = tolower(low[i]);
    }

    have_lower = inputLexiconQuery(low);
    have_startUpper = inputLexiconQuery(startUpper);
    have_upper = inputLexiconQuery(arc->word);

    /** In one-letter words, the intermediate version is
       indistinguishable from the third one, so we suppress it.  */
    if(1 == strlen(arc->word)) {
      have_startUpper = FALSE;
    }
    
    if( have_lower ){
         entries = listAppendList(entries, inputLexiconGet(low));
    }
    if( have_startUpper ){
         entries = listAppendList(entries, inputLexiconGet(startUpper));
    }
    if( have_upper ){
         entries = listAppendList(entries, inputLexiconGet(arc->word));
    }
    if(!have_lower && !have_startUpper && !have_upper){
         entries = listAppendList(entries, inputLexiconGet(low));
         entries = listAppendList(entries, inputLexiconGet(startUpper));
         entries = listAppendList(entries, inputLexiconGet(arc->word));
    }

    if(have_startUpper && !have_upper && !have_lower) {
      arc->word = strCopy(startUpper);
    }
    else if(have_lower && !have_upper && !have_startUpper) {
      arc->word = strCopy(low);
    }
    else if(!have_upper) {
      arc->word = strCopy(startUpper);
    }
 
    memFree(startUpper);
    memFree(low);
  }

  else {
    entries = inputLexiconGet(arc->word);
  }
  
  if (!entries) {
    cdgPrintf(CDG_WARNING, "WARNING: no lexical entry for `%s'.\n", arc->word);
    return FALSE;
  }

  lg->min = min(lg->min, arc->from);
  lg->max = max(lg->max, arc->to);

  gn = gnNew();
  gn->no = vectorAddElement(lg->graphemnodes, gn);
  gn->lexemgraph = lg;
  gn->arc = arc;
  
  /* force consistent order among homonyms */
  j = listSort(entries, leCompareByName);
  listDelete(entries);
  entries = j;

  for (j = entries; j; j = j->next) {
    le = (LexiconItem) j->item;
    ln = (LexemNode) memMalloc(sizeof (LexemNodeStruct));
    ln->no = vectorAddElement(lg->nodes, ln);
    ln->lexemgraph = lg;
    ln->arc = gn->arc;
    ln->lexem = le;
    if(-1 != taggerCategoryIndex) {
      Value v = ln->lexem->values[taggerCategoryIndex];
      if (v->type == VTString) {
        String cat = v->data.string;
        gnWriteNumber(gn, 0, cat, 1.0);
      }
    }
    ln->limit = 1.0;
    bvSetElement(lg->isDeletedNode, FALSE, ln->no);

    /* link the nodes to each other */
    ln->grapheme = gn;
    gn->lexemes = listAppendElement(gn->lexemes, ln);
  }
  listDelete(entries);

  return TRUE;
}

/* ----------------------------------------------------------------------------
  @brief does the final computations for the lexemgraph

  This function sets those fields of @a lg that can only be computed after all
  lexeme nodes are present:
  - It initializes LexemGraph::isDeletedNode to all FALSE.
  - It uses lgComputeDistances() to compute the distance
    between any to nodes in the graph.
  - It applies lgComputeNoOfPaths() to compute the number of paths to and from
    all lexeme nodes in the graph. If no complete path exists, @a lg is
    deallocated with a warning.

  The function can fail returning FALSE if there is no valid path through the
  lexeme graph.
*/
Boolean lgNewFinal(LexemGraph lg)
{
  int noOfNodes;

  /* compute noOfPathsFromStart and noOfPathsToEnd */
  noOfNodes = vectorSize(lg->graphemnodes);
  lgComputeNoOfPaths(lg);

  /* no path? */
  if (lg->noOfPaths == 0) {
    cdgPrintf(CDG_WARNING, "WARNING: no path possible, invalid graph\n");
    return FALSE;
  }
  bvSetAllElements(lg->isDeletedNode, FALSE);
  lgComputeDistances(lg);

  /* build tags cheat sheet */
  {
    AnnoEntry ae = findAnnoForLattice(lg->lattice->id, TRUE);
    List l, m;

    if (ae) {
      for (l = ae->annos; l != NULL; l = l->next) {
	Annotation a = l->item;

	for (m = a->specs; m != NULL; m = m->next) {
	  Specification s = m->item;
	  if (s->type == STTag &&
              taggerCategoryPath == s->kind) {
            int i;
            for(i = 0; i < vectorSize(lg->graphemnodes); i++) {
              GraphemNode gn = vectorElement(lg->graphemnodes, i);
              if(gn->arc->from == a->from) {
                /* CAUTION: this does not work reliably on wordgraphs */
                gnWriteString(gn, 0, taggerCategoryPath, s->name);
                break;
              }
            }
	  }
	}
      }
    }
  }

  lgTag(lg);
  
#ifdef DEBUGLGNEW
  lgPrint(CDG_INFO, lg);
#endif

  return TRUE;
}

/* ----------------------------------------------------------------------------
  This function creates a lexeme graph from a Lattice @a lat and a cdg lexicon.
  For each arcs of the lattice a grapheme node is allocated and annotated with
  all possible lexical entries. (If there is no lexical entry for an arc, a
  warning is given, but processing continues.)

  For each grapheme node, as many lexeme nodes are created as there are
  lexical alternatives in the lexicon.

  Furthermore:
  - LexemGraph::isDeletedNode is initialized to FALSE
  - LexemGraph::noOfPathsFromStart and LexemGraph::noOfPathsFromStart=are
    computed using computeNoOfPaths().
  - LexemGraph::distance is computed using lgComputeDistances().
*/
LexemGraph lgNew(Lattice lat)
{
  LexemGraph lg = lgNewInit();
  List l;

  int width = listSize(inputCurrentGrammar->allAttributes);

  lg->lattice = lat;
  for (l = lat->arcs; l; l = l->next) {
    lgNewIter(lg, (Arc) l->item);
  }

  if (!lgNewFinal(lg)) {
    lgDelete(lg);
    lg = NULL;
  }

  /* Caution! lgNewIter() might have changed the number of features in the
     grammar by autoloading a LexiconItem with a feature not previously
     present. If this happened, some of the ->values caches in the lexeme
     graph are outdated, and inputCacheInput() did not notice that,
     because it does not yet know about _this_ lexeme graph, because it is
     not yet registered in cdgNets. Therefore, if new features turned up
     during this invocation, we have to cache lg ourselves.*/
  if(width < listSize(inputCurrentGrammar->allAttributes)) {
    inputCurrentGrammar->attsCurrent = FALSE;
    inputCacheLexemGraph(inputCurrentGrammar, lg->nodes);
  }

  return lg;
}

/* ----------------------------------------------------------------------------
  @brief print lexem graph

  This function displays a textual representation of the lexeme graph @a lg.
*/
void lgPrint(long unsigned int mode, LexemGraph lg)
{
  Lattice lat = lg->lattice;
  int i;
  LexemNode ln;
  Arc a;
  Boolean deleted;

  cdgPrintf(mode, "\n%s : \n", lat->id);
  for (i = 0; i < vectorSize(lg->nodes); i++) {
    ln = (LexemNode) vectorElement(lg->nodes, i);
    deleted = bvElement(lg->isDeletedNode, ln->no);
    a = ln->arc;

    cdgPrintf(mode, "%%%d:%s%s(%d-%d)%s   %4.3e\n",
	      ln->no,
	      deleted ? "[" : " ",
	      ln->lexem->description,
	      a->from, a->to, deleted ? "]" : " ", lgTagscore(ln));
  }

  if (lg->chunks) {
    cdgPrintf(mode, "chunks:\n");
    chunkerPrintChunks(mode, lg->chunks);
  }
}

/* ----------------------------------------------------------------------------
  @brief returns TRUE if node is a start node

  This function checks whether @c n->arc->from is equal to the minimal time
  point in the lexeme graph.
*/
Boolean lgIsStartNode(GraphemNode n)
{
   if(n == NULL || n == NONSPEC)
      return (FALSE);
   else
      return (n->arc->from == n->lexemgraph->min);
}

/* ---------------------------------------------------------------------------
  @brief returns TRUE if node is an end node

  This function checks whether @c n->arc->to is equal to the maximal time point
  in the lexeme graph.

  @todo should nodes on segment boundaries of incremental parsed input be
  considered to be endnodes
*/
Boolean lgIsEndNode(GraphemNode n)
{
   if(n == NULL || n == NONSPEC)
      return (FALSE);
   else
      return (n->arc->to == n->lexemgraph->max);
}

/* ----------------------------------------------------------------------------
  @brief returns a distance measure for two lexem nodes

  This function computes the logical distance between @a a and @a b, measured
  in words. Usually this is just the corresponding element of
  LexemGraph::distance. If either of the nodes is underspecified it is treated
  as if it followed the latest specified lexeme node directly. Hence, the
  return value may be greater than value in LexemGraph::distance. Two
  underspecified lexeme nodes are considered to have distance zero.
*/
int lgDistanceOfNodes(LexemGraph lg, LexemNode a, LexemNode b)
{
  int maxDistance, i, maxDim, distance;

  if (a == NULL || b == NULL || lg == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: lgDistanceOfNodes: argument is NULL\n");
    abort();
  }
  if (a == NONSPEC) {
    if (b == NONSPEC)
      return 0;

    maxDim = arrayDimension(lg->distance, 1);
    maxDistance = 0;
    for (i = 0; i < maxDim; i++) {
      distance = (int)arrayElement(lg->distance, b->grapheme->no, i);
      if (maxDistance < distance) {
	maxDistance = distance;
      }
    }
    return -maxDistance - 1;
  }

  if (b == NONSPEC) {
    maxDim = arrayDimension(lg->distance, 1);
    maxDistance = 0;
    for (i = 0; i < maxDim; i++) {
      distance = (int)arrayElement(lg->distance, a->grapheme->no, i);
      if (maxDistance < distance) {
	maxDistance = distance;
      }
    }
    return maxDistance + 1;
  }

  return ((int)arrayElement(lg->distance, a->grapheme->no, b->grapheme->no));
}

/* ----------------------------------------------------------------------------
  @brief may these words modify each other?

  \pre @a lexemes must be a List of lexeme nodes with identical time spans. If
  this is not the case, the behaviour is undefined.

  This function checks whether an LevelValue can exist with the dependents @a
  exemes and a modifiee from @a gn. This is the case iff both can coexist on
  one path and do not overlap.
*/
Boolean lgMayModify(LexemGraph lg, GraphemNode down, GraphemNode up)
{
   LexemNode lna, lnb;
   if(!spec(up)) {
     return TRUE;
   }
   lna = down->lexemes->item;
   lnb = up->lexemes->item;
   
   return (lgDistanceOfNodes(lg, lna, lnb) != 0);
}

/* ---------------------------------------------------------------------------
   @brief returns TRUE if lexem nodes a and b exist on one path.

   This function checks whether, in principle, a complete path can exist
   through @a lg that includes both @a a and @a b. This is independent of the
   current state of deletions. In fact, the function merely checks whether the
   distance between the nodes is not 0 by using lgDistanceOfNodes(). Note that
   two nodes are not automatically compatible merely because they do not
   overlap in time.  Also, a lexeme node is not compatible with itself by this
   definition.
 */
Boolean lgCompatibleNodes(LexemGraph lg, LexemNode a, LexemNode b)
{
  if (lg == NULL || a == NULL || b == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: lgCompatibleNodes: argument is NULL\n");
    abort();
  }
  return (a == b || lgDistanceOfNodes(lg, a, b) != 0);
}

/* ----------------------------------------------------------------------------
  @brief does existence of these lexemes exclude that lexeme.
 
  \pre @a lexemes must be a List of lexeme nodes with identical time spans. If
  this is not the case, the behaviour is undefined.

  This function checks whether the List @a lexemes and the lexeme node @a ln
  can be selected in a solution. This is the case if either of the following
  holds:
  - @a lexemes is empty
  - @a ln is compatible with the first element of @a lexemes
  - @a lexemes contains @a ln

  In these cases FALSE is returned (@a ln is not forbidden).  Otherwise TRUE is
  returned.
 */
Boolean lgForbiddenBy(LexemGraph lg, LexemNode ln, List lexemes)
{
  if (!spec(lexemes))
    return FALSE;

  if (!lgCompatibleNodes(lg, ln, lexemes->item)
      && !listContains(lexemes, ln)) {
    return (TRUE);
  } else {
    return (FALSE);
  }
}


/* ----------------------------------------------------------------------------
   @brief checks if these sets of lexemes are compatible, i.e. either unrelated
   or intersecting?
 
   \pre Both @a a and @a b must be Lists of lexeme nodes spanning the same
   respective time interval. If this is not the case, the behaviour is
   undefined.
  
   This function checks whether both sets of lexeme nodes may be selected in a
   solution. This is defined as follows:
   - If either set is empty, the result is TRUE
   - If the first elements of @a a and @a b are compatible, the result is TRUE
   - If the sets intersect, the result is TRUE
   - Otherwise the result is FALS
 */


Boolean lgCompatibleSets(LexemGraph lg, List a, List b)
{

   LexemNode lna, lnb;
   List p;

   /* empty sets are compatible */

   if (a == NULL || b == NULL)
      return (TRUE);

   lna = a->item;
   lnb = b->item;

   /* unrelated sets are compatible */

   if (lgCompatibleNodes(lg, lna, lnb))
      return (TRUE);

   /* if we get here, the only chance for compatibility
      is for the lists to intersect. This is only possible
      if their elements are simultaneous. */

   if(lna->arc->from != lnb->arc->from ||
      lna->arc->to != lnb->arc->to)
     return FALSE;
   
   /* OK, the sets are simultaneous. Do they actually intersect? */

   for (p = a; p != NULL; p = p->next) {
      if (listContains(b, p->item))
	 return (TRUE);
   }

   /* mutually exclusive sets are not compatible */

   return (FALSE);

}

/* ----------------------------------------------------------------------------
 * This checks if a lexem node has been deleted TRUE is returned, if not
 * FALSE.
 */
Boolean lgIsDeletedNode(LexemGraph lg, LexemNode ln)
{
   return (bvElement(lg->isDeletedNode, ln->no));
}


/* ----------------------------------------------------------------------------
 * This checks if the @a lexemes have been deleted TRUE is returned, if not
 * FALSE
 */
Boolean lgAreDeletedNodes(LexemGraph lg, List lexemes)
{
  List l;

  for (l = lexemes; l != NULL; l = l->next) {
    if (!lgIsDeletedNode(lg, l->item))
      return FALSE;
  }
  return TRUE;
}

/* ----------------------------------------------------------------------------
 * checks whether not all paths passing this GraphemNode will be broken by
 * its deletion
 */
Boolean lgIsDeletableNode(GraphemNode gn)
{
  LexemGraph lg = gn->lexemgraph;
  int affectedPaths =
    lg->noOfPathsFromStart[gn->no] * lg->noOfPathsToEnd[gn->no];

  return affectedPaths < lg->noOfPaths;
}

/* ----------------------------------------------------------------------------
  @brief deletes a node from the lexeme graph itself.

  This function marks a lexeme node as deleted. It does this by setting the
  cell @a ln->no in the Vector @a lg->isDeletedNode. If this destroys the last
  possible path through @a lg=, a warning is displayed. This function always
  re-computes the number of remaining paths in @a lg.
*/
void lgDeleteNode(LexemGraph lg, LexemNode ln)
{

  int oldNo = lg->noOfPaths;

  /* already deleted, ok */
  if (bvElement(lg->isDeletedNode, ln->no))
    return;

  bvSetElement(lg->isDeletedNode, TRUE, ln->no);
  lgComputeNoOfPaths(lg);

  if (lg->noOfPaths == 0 && oldNo > 0) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: lgDeleteNode: deletion of node `%s' destroys last path\n",
	      ln->lexem->description);
  }

  /* have to do this because two nodes may become more distant by the
   * deletion of the shortest connecting path */
  lgComputeDistances(lg);
}

/* ----------------------------------------------------------------------------
  @brief delete a list of lexeme nodes

  This function behaves as lgDeleteNode() were called on each element of the @a
  nodes, but it is more efficient since it only re-computes the number of
  remaining paths once.
*/
void lgDeleteNodes(LexemGraph lg, List nodes)
{
  List p;
  int oldNo = lg->noOfPaths;
  LexemNode ln = NULL;

  /* So we inline the relevant code into a loop here... */
  for (p = nodes; p != NULL; p = p->next) {
    ln = p->item;
    bvSetElement(lg->isDeletedNode, TRUE, ln->no);
  }

  /* ...and then do one call to computeNoOfPaths(). */
  lgComputeNoOfPaths(lg);
  if (oldNo > 0 && lg->noOfPaths == 0) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: lgDeleteNodes: deletion of node `%s' destroys last path\n",
	      ln->lexem->description);
  }

  /* have to do this because two nodes may become more distant by the
   * deletion of the shortest connecting path */
  lgComputeDistances(lg);
}


/* ----------------------------------------------------------------------------
  @brief do the lexemes span the same time interval?

  This function checks whether @a a and @a b cover the same time span. An
  argument of NONSPEC always causes TRUE to be returned. However, the NULL node
  is not simultaneous to any lexeme node, not even to another root node.
*/
Boolean lgSimultaneous(LexemNode a, LexemNode b)
{

   if (a == NULL && b != NULL) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: lgSimultaneous: first argument must not be NULL\n");
      cdgPrintf(CDG_WARNING,
	 "WARNING: the other argument was `%s'\n", b->lexem->word);
   }
   if (b == NULL && a != NULL) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: lgSimultaneous: second argument must not be NULL\n");
      cdgPrintf(CDG_WARNING,
	 "WARNING: the other argument was `%s'\n", a->lexem->word);
   }
   if (a == NULL && b == NULL)
      cdgPrintf(CDG_WARNING,
	 "WARNING: lgSimultaneous: both arguments must not be NULL\n");

   if(a == NONSPEC || b == NONSPEC)
      return(TRUE);
   else if(a == NULL || b == NULL)
      return (FALSE);
   else
      return (a->arc->from == b->arc->from
	      && a->arc->to == b->arc->to);
}

/* ----------------------------------------------------------------------------
  @brief is this lexeme a member of the this set?

  \pre @a lexemes must be a List of lexeme nodes with identical time spans. If
  this is not the case, the behaviour is undefined.

  This function checks whether @a ln is an element of @a lexemes.

  In the following cases @a ln is not a member (return FALSE);
  - a NIL binding (@a ln is NULL) is not an element of anything
  - an empty set (@a lexemes is NULL) has no member,
  - @a ln belongs to another timespan
  - @a ln is not contained litteraly in the set 
  
  Otherwise TRUE is returned.
*/
Boolean lgMember(LexemNode ln, List lexemes)
{
   if (ln == NULL)
      return (FALSE);

   if (lexemes == NULL)
      return (FALSE);

   if (!lgSimultaneous(lexemes->item, ln))
      return (FALSE);

   return (listContains(lexemes, ln));
}

/* ----------------------------------------------------------------------------
  This function checks whether @a a is a subset of @a b.

  \pre Both @a a and @a b must be Lists of lexeme nodes spanning the same
  respective time interval. If this is not the case, the behaviour is
  undefined.
*/
Boolean lgSubset(List a, List b)
{

  List l;
  LexemNode ln;

/* empty set is always a subset */
  if (a == NULL)
    return (TRUE);

/* empty set has no proper subset */
  if (b == NULL)
    return (FALSE);

/* if a belongs to a different timespan, it's not a subset */
  if (!lgSimultaneous(a->item, b->item))
    return (FALSE);

/* find it out the hard way */
  for (l = a; l != NULL; l = l->next) {
    ln = l->item;
    if (!lgMember(ln, b))
      return (FALSE);
  }

  return (TRUE);

}

/* ----------------------------------------------------------------------------
  @brief Do two lexeme lists intersect.

  \pre Both @a a and @a b must be Lists of lexeme nodes spanning the same
  respective time interval. If this is not the case, the behaviour is
  undefined.

  This function checks whether @a a and @a b intersect.
 */
Boolean lgIntersectingSets(List a, List b)
{

   List l;
   LexemNode ln;

   if (a == NULL && b != NULL) { return (FALSE); }

   if (a != NULL && b == NULL) { return (FALSE); }
   
   if (a == NULL && b == NULL) { return (TRUE); }

/* if a belongs to a different timespan, no intersection */
   if (!lgSimultaneous(a->item, b->item)) {
      return (FALSE);
   }
   
/* find it out the hard way */
   for (l = a; l != NULL; l = l->next) {
      ln = l->item;
      if (listContains(b, ln)) {
	 return (TRUE);
      }
   }

   return (FALSE);

}

/* ----------------------------------------------------------------------------
  Takes a set of lexeme nodes, and extends it to a complete path through the
  graph, composed of undeleted LexemNodes.  Returns NULL if this is impossible,
  It returns a List of lexeme nodes that
  -# is a superset of @a nodes
  -# corresponds to a complete path through the graph and
  -# contains only undeleted lexeme nodes.

  If this is not possible, NULL is returned.
*/
List lgMakePath(LexemGraph lg, List nodes) {

  LexemNode lna, lnb;
  List result = NULL;
  List l, m;
  int i, nrBoundPoints = 0;
  Boolean okay;
  
  /** We do this by simply appending arbitrary non-contradictory nodes
     until we have bound all time points. Note that for this approach
     to be correct, there must not be any undeleted dangling nodes in
     the graph. This condition must have ensured by cnOptimizeNet(). */
     
  /* is the list of nodes self-contradictory? */
  for(l = nodes; l != NULL; l = l->next) {
    lna = l->item;
    nrBoundPoints += (lna->arc->to - lna->arc->from);
    for(m = l->next; m != NULL; m = m->next) {
      lnb = m->item;
      if(!lgCompatibleNodes(lg, lna, lnb)) {
	return NULL;
      }
    }
  }
  
  /* choose arbitrary nodes until all time points are bound */  
  result = listClone(nodes);
  for(i = 0; i < vectorSize(lg->nodes); i++) {
    okay = TRUE;
    lna = vectorElement(lg->nodes, i);
    if(lgIsDeletedNode(lg, lna)) {
      continue;
    }
    for(l = result; l != NULL; l = l->next) {
      lnb = l->item;
      if(lna == lnb ||
	 !lgCompatibleNodes(lg, lna, lnb)) {
	okay = FALSE;
	break;
      }
    }
    if(okay) {
      result = listAppendElement(result, lna);
      nrBoundPoints += (lna->arc->to - lna->arc->from);
    }
  }

  if(nrBoundPoints != lg->max - lg->min) {
    listDelete(result);
    result = NULL;
  }
 
  return result;
}

/* ----------------------------------------------------------------------------
  @brief deletes LexemGraph

  This function deallocates a lexeme graph. This deallocates all parts of the
  structure, even the lexeme nodes and lexical entries themselves. The lexicon
  remains unchanged as the LexicalEntry structures are merely clones of the
  structures in @ref inputCurrentGrammar.
*/
void lgDelete(LexemGraph lg)
{
  int i;
  GraphemNode gn;
  LexemNode ln;

  if (!lg) {
    return;
  }

  for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
    gn = vectorElement(lg->graphemnodes, i);
    gnDelete(gn);
  }
  vectorDelete(lg->graphemnodes);

  for (i = 0; i < vectorSize(lg->nodes); i++) {
    ln = (LexemNode) vectorElement(lg->nodes, i);
    freeLexiconItem(ln->lexem);
    memFree(ln);
  }
  vectorDelete(lg->nodes);

  if (lg->distance != NULL) {
    arrayDelete(lg->distance);
  }

  memFree(lg->noOfPathsFromStart);
  memFree(lg->noOfPathsToEnd);
  bvDelete(lg->isDeletedNode);

  listForEachDelete(lg->chunks, chunkerChunkDelete);

  memFree(lg);
}


/* ----------------------------------------------------------------------------
  @returns maximal ambiguity per time point

  This function computes the maximal number of overlapping lexeme nodes for any
  time point in @a lg. Thus it gives an upper bound (not an estimate) of the
  average acoustical and lexical ambiguity of the graph.
*/
int lgWidth(LexemGraph lg)
{

  int result = 0;
  int i, j;
  int *ambiguity = memMalloc(sizeof (int) * lg->max);
  LexemNode ln;

  for (i = 0; i < lg->max; i++) {
    ambiguity[i] = 0;
  }

  for (i = 0; i < vectorSize(lg->nodes); i++) {
    ln = vectorElement(lg->nodes, i);
    for (j = ln->arc->from; j < ln->arc->to; j++) {
      ambiguity[j]++;
      if (ambiguity[j] > result) {
	result = ambiguity[j];
      }
    }
  }

  memFree(ambiguity);
  return result;

}

/* ----------------------------------------------------------------------------
  @brief prints out a lexeme node

  This function displays the identifier and the time span of @a ln in the
  format @c der_nom(0,1).
*/
void lgPrintNode(unsigned long mode, LexemNode ln)
{
  Boolean dead = bvElement(ln->lexemgraph->isDeletedNode, ln->no);

  cdgPrintf(mode, "%s`%s'(%d-%d)%s",
	    dead ? "[" : "",
	    ln->lexem->description,
	    ln->arc->from, ln->arc->to, dead ? "]" : "");
}


/* ----------------------------------------------------------------------------
  @brief Do these lexeme nodes overlap?

  Returns TRUE if the two lexeme nodes have at least one time point in common.

  This is subtly different from the more common question, "Can the two nodes
  coexist on one path?": two nodes can be compatible although they overlap if
  they are identical. Conversely, a and b may be incompatible even if they do
  not overlap if there is no path between them.
*/
Boolean lgOverlap(LexemNode a, LexemNode b) 
{
  return((a->arc->to > b->arc->from && a->arc->from <= b->arc->from) ||
	 (b->arc->to > a->arc->from && b->arc->from <= a->arc->from));
}

/* ----------------------------------------------------------------------------
  Takes a Vector of Boolean, and sets all cells that correspond to the numbers
  of nodes incompatible with ln. This function can be used in combination with
  lvVectorCompatible() to decide whether an LV is compatible with a set of
  other LVs.
*/
void lgRequireLexeme(LexemGraph lg, ByteVector v, LexemNode ln)
{
  int i;
  LexemNode ln2;

  if (!spec(ln))
    return;

  for (i = 0; i < vectorSize(lg->nodes); i++) {
    ln2 = vectorElement(lg->nodes, i);
    if (!lgCompatibleNodes(lg, ln, ln2)) {
      bvSetElement(v, TRUE, ln2->no);
    }
  }
}

/* ----------------------------------------------------------------------------
  This function is similar to lgRequireLexeme(), but takes a List of lexeme
  nodes. It marks all those lexeme nodes that are incompatible with all
  lexemnodes of @a which.
*/
void lgRequireLexemes(LexemGraph lg, ByteVector v, List which)
{
  int i;
  LexemNode ln1, ln2;

  if (!spec(which))
    return;

  ln1 = which->item;

  if (!spec(ln1))
    return;

  for (i = 0; i < vectorSize(lg->nodes); i++) {
    ln2 = vectorElement(lg->nodes, i);
    if (!lgCompatibleNodes(lg, ln1, ln2) && !listContains(which, ln2)) {
      bvSetElement(v, TRUE, ln2->no);
    }
  }
}

/* ----------------------------------------------------------------------------
  @brief updates the partial lexemgraph with the incoming arcs.

  This function extends a lexeme graph by the Arc structures contained in @a
  listArcs.
*/
Boolean lgUpdateArcs(LexemGraph lg, Lattice lat, List listArcs)
{
  List m;

  for (m = listArcs; m; m = m->next) {
    lgNewIter(lg, (Arc) m->item);
  }
  if (lg->distance != NULL) {
    arrayDelete(lg->distance);
    lg->distance = NULL;
  }
  
  lg->lattice = lat;
  return lgNewFinal(lg);
}

/* ----------------------------------------------------------------------------
   This function copies information that a lexeme graph acquired by being
   decorated through external programs to its clone. This is only useful to
   save repeated invocation of taggerTag() or similar commands for two
   graphs produced from the same lattice.
*/
void lgCopyTagScores(LexemGraph destination, LexemGraph source)
{
  List l;

  void backlink(Chunk chunk) {
    List l;

    for (l = chunk->nodes; l; l = l->next) {
      GraphemNode gn = l->item;

      gn->chunk = chunk;
    }
    for (l = chunk->subChunks; l; l = l->next) {
      backlink(l->item);
    }
  }

  /* copy chunk information */
  for (l = source->chunks; l; l = l->next) {
    Chunk chunk = l->item;

    chunk = chunkerCloneChunk(chunk);
    destination->chunks = listAppendElement(destination->chunks, chunk);
    chunkerReplaceGraphemes(chunk, destination);
    backlink(chunk);
  }
}

/* ----------------------------------------------------------------------
   Read the score that a POS tagger assigned to this particular lexeme.
*/
Number lgTagscore(LexemNode ln) {
  GraphemNode gn = ln->grapheme;
  Value v;
  if(-1 == taggerCategoryIndex) {
    return 0.0;
  }
  v = ln->lexem->values[taggerCategoryIndex];
  if(v->type != VTString) {
    return 0.0;
  }
  v = gnReadPrediction(gn, 0, v->data.string);
  if(v->type != VTNumber) {
    return 0.0;
  }
  return v->data.number;
}


/* ----------------------------------------------------------------------------
  This function checks whether at least one of the LexemNode
  structures in @a list points to a lexicon element @a le
*/
Boolean lgLexemeInLexemNodeList(LexiconItem le, List list)
{
  List m;
  for (m = list; m != NULL; m = m->next) {
    LexemNode ln = m->item;
    if (ln->lexem == le)
      return TRUE;
  }
  return FALSE;
}

/* ----------------------------------------------------------------------------
  @brief Select the path in DST whose parts most closely match SRC.

  This function inspects the undeleted words in @a source and undeletes those
  words in @a destination that most closely correspond to them. (This is
  necessary because two lexeme graphs built from the same lattice may have
  their nodes in different order, so you cannot simply re-use an
  LexemGraph::isDeletedNode vector across lexeme graphs.)
*/
Boolean lgCopySelection(LexemGraph destination, LexemGraph source)
{

  int i, j, k, record;
  LexemNode ln1, ln2, best;

  if (destination == source) {
    return TRUE;
  }

  bvSetAllElements(destination->isDeletedNode, TRUE);
  for (i = 0; i < bvSize(source->isDeletedNode); i++) {
    if (bvElement(source->isDeletedNode, i)) {
      continue;
    }

    /* OK, find the node in DST that corresponds to this one. */
    ln1 = vectorElement(source->nodes, i);
    best = NULL;
    record = INT_MAX;
    for (j = 0; j < vectorSize(destination->nodes); j++) {
      ln2 = vectorElement(destination->nodes, j);
      if (ln2->arc->from != ln1->arc->from ||
          ln2->arc->to != ln1->arc->to ||
          ln2->lexem->word != ln1->lexem->word) {
	continue;
      }
      k = leCompare(inputCurrentGrammar, ln1->lexem, ln2->lexem, FALSE);
      if (k < record) {
	record = k;
	best = ln2;
      }
    }
    if (!best) {
      return FALSE;
    }
    bvSetElement(destination->isDeletedNode, FALSE, best->no);
  }

  return TRUE;
}

/* ----------------------------------------------------------------------------
  @brief Initialize the input module

  This function initializes the module Lexemgraph and registers the variable 
  @a compactlevelvalues.
*/
void lgInitialize()
{
  setRegister("compactlevelvalues", SET_BOOL, &lgCompactLVs, NULL, NULL, NULL,
	      NULL);
}

/* ---------------------------------------------------------------------------
  @brief What categories can this node represent? (Needed while tagging.)

  This function queries the lexicon about what syntactical categories @a gn can
  represent. (The syntactical category is that feature whose index is 
  @ref taggerCategoryIndex.) This function is used to check whether an
  assignment by the tagger can be honored by the lexicon.
*/
List lgQueryCat(LexemGraph lg, GraphemNode gn)
{

  List l;
  List result = NULL;

  for (l = gn->lexemes; l != NULL; l = l->next) {
    LexemNode ln = l->item;
    LexiconItem li = ln->lexem;
    Value v = li->values[taggerCategoryIndex];

    if (v->type != VTString) {
      /* This should definitely not happen, but it will already generate
       * errors elsewhere, so we just ignore it here. */
      continue;
    }
    result = listAddUniqueElement(result, strRegister(v->data.string));
  }
  return result;
}

/* ----------------------------------------------------------------------------
  @brief Does a lexemgraph contain at least one instance of a given form? 
 
  This function checks whether @a lg contains at least one instance of the
  form @a form. Capitalized versions of @a form are permissible if they are
  spurious (cf. lgSpuriousUppercase()).
*/
Boolean lgContains(LexemGraph lg, String form) {
  int i;
  for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
    GraphemNode gn = vectorElement(lg->graphemnodes, i);
    if(form == gn->arc->word) {
      return TRUE;
    }
    if(lgSpuriousUppercase(lg, gn->arc) &&
       !strcasecmp(form, gn->arc->word)) {
      return TRUE;
    }

  }
  return FALSE;
}
   
/* ----------------------------------------------------------------------------
  Returns the most probable path, as defined by tagging scores.
*/
List lgMostProbablePath(LexemGraph lg)
{
  List result = NULL;
  int i;
  List l;

  if (latticeBranches(lg->lattice)) {
    cdgPrintf(CDG_WARNING,
	      "ERROR: lgSelectMostProbablePath() called no branching lattice!\n");
    return NULL;
  }

  for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
    GraphemNode gn = vectorElement(lg->graphemnodes, i);
    Number record = 0.0;
    LexemNode best = NULL;

    for (l = gn->lexemes; l != NULL; l = l->next) {
      LexemNode ln = l->item;
      Number score = lgTagscore(ln);
      if (score > record) {
	record = score;
	best = ln;
      }
    }
    result = listPrependElement(result, best);
  }
  return result;
}

/* ----------------------------------------------------------------------------
  @brief Clone a grapheme node.

  The field GraphemNode::lexemes is not set; the caller has to do that. (The
  two-way links between grapheme nodes and lexeme nodes can be set easier when
  all nodes are known.)
*/
GraphemNode gnClone(GraphemNode gn, Lattice lat)
{
  GraphemNode result = gnNew();

  result->no = gn->no;
  result->lexemgraph = gn->lexemgraph;
  result->arc = inputFindArc(lat, gn->arc->no);
  result->lexemes = NULL;

  if(gn->predictions) {
    int i;
    Hashtable pred;
    result->predictions=vectorNew(vectorCapacity(gn->predictions));
    for (i=0; i < vectorSize(gn->predictions); i++) {
      if ((pred=(Hashtable)vectorElement(gn->predictions,i))!=NULL) {
	Hashtable pred2=hashNew(10,0.8,
				hashStringHashFunction,
				hashStringEqualFunction);
	HashIterator hi = hashIteratorNew(pred);
	String key;
	while (NULL != (key = hashIteratorNextKey(hi))) {
	  String value = hashGet(pred, key);
	  hashSet(pred2,key,value);
	}
	hashIteratorDelete(hi);
	vectorSetElement(result->predictions,pred2,i);
      }
    }
  }
  result->chunk = NULL;

  return result;
}


/* ----------------------------------------------------------------------------
  @brief Clone a lexeme graph.

  This performs a totally deep copy; even the underlying lattice, lexicon
  items etc. are cloned.
*/
LexemGraph lgClone(LexemGraph lg) {

  int i;
  LexemGraph result = memMalloc(sizeof(LexemGraphStruct));

  result->lattice = inputCloneLattice(lg->lattice);
  result->graphemnodes = vectorNew(20);
  for(i = 0; i < vectorSize(lg->graphemnodes); i++) {
    GraphemNode gn = vectorElement(lg->graphemnodes, i);
    GraphemNode new = gnClone(gn, result->lattice);
    vectorAddElement(result->graphemnodes, new);
  }
  
  result->nodes = vectorNew(20);
  for(i = 0; i < vectorSize(lg->nodes); i++) {
    LexemNode ln = vectorElement(lg->nodes, i);
    LexemNode new = memMalloc(sizeof (LexemNodeStruct));
    new->no = ln->no;
    new->lexemgraph = result;
    new->arc = inputFindArc(result->lattice, ln->arc->no);
    new->lexem = inputCloneLI(ln->lexem);
    new->limit = ln->limit;
    new->grapheme = vectorElement(result->graphemnodes, ln->grapheme->no);
    new->grapheme->lexemes = listAppendElement(new->grapheme->lexemes, new);
    vectorAddElement(result->nodes, new);
  }

  result->max = lg->max;
  result->min = lg->min;

  result->isDeletedNode = bvClone(lg->isDeletedNode);

  result->distance = NULL;
  lgComputeDistances(result);

  result->noOfPathsToEnd = NULL;
  result->noOfPathsFromStart = NULL;
  lgComputeNoOfPaths(result);
  result->chunks = NULL;
  lgCopyTagScores(result, lg);

  return result;
    
}


/* ----------------------------------------------------------------------------
  Might this be a lowercase word that is spelled in upper case
  because of orthographic convention?
*/
Boolean lgSpuriousUppercase(LexemGraph lg, Arc arc)
{
  int index = 0;
  
  /* skip over leading punctuation in a word */
  while(((unsigned char *)arc->word)[index] == '"' ||
        ((unsigned char *)arc->word)[index] == '(') {
    index++;
  }
  
  /** Spurious uppercase must be an upper case letter... */
  if (!isupper(((unsigned char *)arc->word)[index])) {
    return FALSE;
  }

  /** ... followed by a lower case letter. */
  if (strlen(arc->word) > 1 + index &&
      !islower(((unsigned char *)arc->word)[1 + index])) {
    return FALSE;
  }
  
  /** This is another instance of the "wordgraphs start at 0" assumption.
 
     Ordinarily, this would be wrong, since the lexeme graph might start at
     some other time point. However, at this time lg->min may not be
     initialized, so we can't check it. Since spurious upper case only
     occurs in written text, and weird time points occur mainly in
     recognizer output for spoken text, I'm letting it pass here.  */
  if (arc->from == 0) {
    return TRUE;
  }

  return inputSentenceBoundary(inputLatticePrevArc(lg->lattice, arc));
}


/* -- ENDOFFILE --------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/** @} */
