/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/* -- INCLUDES ------------------------------------------------------------- */
#include <stdio.h>
#include <blah.h>
#include <string.h>
#include "cdg.h"
#include "hook.h"
#include "lexemgraph.h"
#include "tagger.h"
#include "predict.h"

/* -- VARIABLES ------------------------------------------------------------ */

/* the string "POS".
   initialized in cdg.c */
String constStr_POS=NULL;

/* -- FUNCTIONS ------------------------------------------------------------ */

#define BUFSIZE 32000

/* -------------------------------------------------------------------------
   Call an external program and record its predictions in the lexeme graph.
 */
int predictCall(Predictor p, int slot, LexemGraph lg) {

  int gi; 
  int i;
  int result = 0;
  char inputname[] = "/tmp/cdg-XXXXXX";
  char outputname[] = "/tmp/cdg-XXXXXX";
  int indescriptor;
  int outdescriptor;
  String call;
  FILE *input;
  FILE *output;
  char buffer [BUFSIZE];

  /* fail on branching lattices */
  if (latticeBranches(lg->lattice)) {
    cdgPrintf(CDG_ERROR, "ERROR: can't process branching lattices\n");
    return -1;
  }
  
  /* open input and output and construct the invocation */
  indescriptor = mkstemp(inputname);
  if(NULL == (input = fdopen(indescriptor, "w"))) {
    cdgPrintf(CDG_ERROR, "ERROR: Can't write input file `%s'!\n", inputname);
    return -1;
  }

  outdescriptor = mkstemp(outputname);
  if(NULL == (output = fdopen(outdescriptor, "r"))) {
    cdgPrintf(CDG_ERROR, "ERROR: Can't read output file!\n");
    unlink(inputname);
    close(indescriptor);
    return -1;
  }

  /* write the lattice name as additional information */
  fprintf(input, "#sentence: %s\n", lg->lattice->id);

  /* iterate through words of the lattice 
     and write them to temporary file */
  for(i = 0; i < vectorSize(lg->graphemnodes); i++) {
    GraphemNode gn = vectorElement(lg->graphemnodes, i);
    Arc a = (Arc) gn->arc;
    char b[BUFSIZE]; 
    char *c;
    List l;
    String cat;
    LexiconItem favored = NULL;
    
    /* Words with embedded spaces confuse most programs.
       Temporarily replace ' ' with an unprintable control character. */
    strncpy(b,a->word,BUFSIZE);
    while(NULL != (c = strchr(b,' '))) {
      *c = 255;
    }
    
    /* print current word */
    fprintf(input, "%s", b);

    /* Try to find the preferred lexeme for the grapheme, even if `cat'
       is not explicitly requested.
       
       The reason for this is to find the most relevant value to print.
       For instance, statistical analyzers probably want to normalize
       words to their base forms. A predictor might ask for the value of
       the feature `base'. But different homonyms can have different
       values for that; for instance, the word `gefllt' may either be a
       finite verb (VVFIN) with the base form `gefallen', or a participle
       (VVPP) with the base form `fllen'. If POS tagging has made the
       right decision, this will also disambiguate the base form. */
    if(-1 != taggerCategoryIndex &&
       NULL != (cat = gnHighestPrediction(gn))) {
      List l;
      for(l = gn->lexemes; l != NULL; l = l->next) {
        LexemNode ln = l->item;
        Value v = ln->lexem->values[taggerCategoryIndex];
        if(VTString == v->type &&
           !strcmp(cat,v->data.string)) {
          favored = ln->lexem;
          break;
        }
      }
    }

    /* Print additional information.

       For instance, a PP attacher might want to know not just the surface
       form but also the base form of each word. The List p->input
       specifies the set of things that the predictor wants to know. The
       token `cat' is special and refers to `the category that was scored
       most highly by the POS tagger'. All other tokens are interpreted as
       attribute names in the grammar, and the value of the first lexeme
       of this grapheme which has that attribute will be printed.
    */
    for(l = p->input; l != NULL; l = l->next) {
      String value = NULL;
      String s = l->item;
      List path = listPrependElement(NULL, s);
      int index = inputLookupPath(path, inputCurrentGrammar);
      listDelete(path);

      if(!strcmp("cat", s)) {
        
        /* print predicted POS tag */
        value = gnHighestPrediction(gn);
        /* This assumes that only the tagger was yet called! */

      } else if (-1 != index) {
        List m;

        if(favored) {
          Value v = favored->values[index];
          if(v->type == VTString) {
            value = v->data.string;
          }
        } else {
          for(m = gn->lexemes; m != NULL; m = m->next) {
            Value v;
            LexemNode ln = m->item;
            v = ln->lexem->values[index];
            
            if(v->type == VTString) {
              value = v->data.string;
              break;
            }
          }
        }
      }

      if(value) {
        fprintf(input, "\t%s", value);
      } else {
        fprintf(input, "\t-");
      }
    }
    fprintf(input, "\n");
  }
  fflush(input);

  /* call the program */
  cdgPrintf(CDG_INFO, "INFO: calling predictor `%s'\n", p->name);
  call = strPrintf("%s < %s > %s", p->command, inputname, outputname); 
  if(system(call)) {
    cdgPrintf(CDG_ERROR, "ERROR: Can't call predictor `%s'!\n", p->name);
    unlink(inputname);
    unlink(outputname);
    close(indescriptor);
    close(outdescriptor);
    return -1;
  }
  
  /* go through the response and record predictions */
  gi = 0;
  while(fgets(buffer, BUFSIZE, output)) {

    GraphemNode gn;
    String w;
    char b[BUFSIZE];
    char *p;

    /* No newline? That means the program wrote a line that is longer than
       our buffer. Exit in distress. */
    if(strcspn(buffer, "\n") == strlen(buffer)) {
      cdgPrintf(CDG_ERROR,
                "ERROR: Program output too wide! Processing stopped for this sentence.\n");
      cdgPrintf(CDG_ERROR,
                "ERROR: The bad line of output was: `%s'\n", buffer);
      return -1;
    }

    /* Misbehaving programs might print more lines of output than input,
       guard against this */
    if(gi >= vectorSize(lg->graphemnodes)) {
      cdgPrintf(CDG_ERROR,
                "ERROR: Too many lines of output! Processing stopped for this sentence.\n");
      return -1;
    }
    gn = (GraphemNode) vectorElement(lg->graphemnodes, gi);

    w = strtok(buffer, " \t\n");
    if (!w) {
      cdgPrintf(CDG_ERROR, "ERROR: can't find word\n");
      result = -1;
      break;
    }

    /* Undo space escaping that we might have done */
    strncpy(b,w,BUFSIZE);
    while(NULL != (p = strchr(b,255))) {
      *p = ' ';
    }

    if (strcmp(b, gn->arc->word)) {
      cdgPrintf(CDG_ERROR,
                "ERROR: word mismatch! `%s' should be `%s'\n",
                b, gn->arc->word);
      result = -1;
      break;
    }

    /* loop through all tokens in a response line */
    String key = NULL, value = NULL;
    do {

      key = strtok(NULL, " \t\n");
      if(key) {
        value = strtok(NULL, " \t\n");
      
        /* read one prediction */
        if(key && value) {
          Number f;
#ifdef DEBUGPREDICTCALL
          cdgPrintf(CDG_DEBUG, "DEBUG: %s  %s  %s\n", b, key, value);
#endif
          if(strspn(value, "+-e.0123456789") == strlen(value) &&
             1 == sscanf(value, "%lf",&f)) {
            gnWriteNumber(gn, slot, strRegister(key), f);
          } else {
            gnWriteString(gn, slot, strRegister(key), strRegister(value));
          }
        }
      }
    } while(key && value);
    gi++;
  }
  unlink(inputname);
  unlink(outputname);
  close(indescriptor);
  close(outdescriptor);
  
  return result;
}

/* ----------------------------------------------------------------------
   Find a Predictor by its name
   the returned value is either 0 (for the POS tagger)
   or one more than an index into the cdgPredictors vector
   or -1 if nothing can be found
*/
int predictFindPredictor(String name)
{
  int i;
  if (name==constStr_POS)
    return 0;
  for (i=0; i<vectorSize(cdgPredictors); i++) {
    if (((Predictor)vectorElement(cdgPredictors,i))->name == name)
      return i+1;
  }
  return -1;
}

/* ----------------------------------------------------------------------
   Create a new Predictor.
*/
Predictor predictNew(String name, String command, List features) {
  Predictor result = memMalloc(sizeof(PredictorStruct));
  result->name = strRegister(name);
  result->input = features;
  result->command = command;
  return result;
}

/* ----------------------------------------------------------------------
   Deallocate a Predictor.
*/
void predictFree(Predictor p) {
  cdgFreeString(p->name);
  listDelete(p->input);
  cdgFreeString(p->command);
  memFree(p);
}

/* ----------------------------------------------------------------------
   Deallocate all Predictors.
*/
void predictFinalize() {
  int i;
  for(i = 0; i < vectorSize(cdgPredictors); i++) {
    Predictor p = vectorElement(cdgPredictors,i);
    predictFree(p);
  }
  vectorDelete(cdgPredictors);
  cdgPredictors=NULL;
}

/* ----------------------------------------------------------------------
   Call all predictors.
*/
void predictAll(LexemGraph lg) {
  int i;
  for(i = 0; i < vectorSize(cdgPredictors); i++) {
    predictCall((Predictor)vectorElement(cdgPredictors,i),
 		i+1, lg);
  }
}

/* ------------------------------------------------------------------- */
/* ------------------------------------------------------------------- */

