/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: scorematrix.c,v 1.21 2005-12-01 16:19:04 micha Exp $ 
 */

/* -------------------------------------------------------------------------
 * @addtogroup Scorematrix Scorematrix - The matrix of scores appearing in each ConstraintEdge
 *
 * @author Ingo Schroeder
 * @date 6/3/97
 *
 *
 * \b Note:
 * The functions for accessing these matrices check whether the matrix is
 * non- \b NULL, but they do not check whether the index is
 * meaningful. Hence the behaviour of these functions is undefined if an
 * invalid array index is passed to them.
 * 
 * @{
 */

/* ---------------------------------------------------------------------- */
#include <config.h>

#include <stdio.h>
#include "scorematrix.h"
#include "cdg.h"
#include "hook.h"

/* ---------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- 
 * Entry in the matrix.
 * All entries in the matrix are pairs of a score and a flag indicating
 * whether this pair is valid or not.
 */
struct SMEntryStruct {
  double score;	 /**< score of this entry*/
  Boolean flag;  /**< indicating if this is a valid pari */
};

typedef struct SMEntryStruct SMEntryStruct; 
  /**< type of matrix entry structure */
typedef SMEntryStruct *SMEntry; 
  /**< type of matrix entry pointer */

/* ---------------------------------------------------------------------- 
 * The score matrix
 *
 * Since not all matrices in the edges of a constraint net are of equal
 * size, each matrix administrates the number of its rows and columns
 * explicitly in addition to the array itself. The element <i>(r,c)</i> in 
 * a score matrix is always referenced by
 *
 * <b> sm->entries[c * sm->rows + r] </b>
 *
 * The last element of a matrix always has the index
 *
 * <b> sm->rows * sm->cols -1 </b>
 * 
 * The type \b ScoreMatrix is only used in a \b ConstraintEdge. A
 * similar function is fulfilled by the structure \b ScoreCache
 * defined by the module \ref Scache, which has the advantage of
 * being usable even without constraint edges.
 */
struct ScoreMatrixStruct {
  int rows;   /**< number of rows in the matrix */
  int cols;   /**< number of columns in the matrix */
  SMEntry *entries; /**< entries in the matrix */
};

typedef struct ScoreMatrixStruct ScoreMatrixStruct;
  /**< type of score matrix structure */

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * creates and returns a new score matrix
 * 
 * This function allocates a new \b ScoreMatrix and returns a pointer
 * to it. It allocates an array of size <b> r * c </b>. All elements
 * are initialized to pairs of the form <b> (0.0, FALSE) </b>.
 */
ScoreMatrix smNew(int r, int c)
{
  ScoreMatrix sm;
  int i;
  int total = r * c;

  sm = (ScoreMatrix)memMalloc( sizeof(ScoreMatrixStruct) );
  sm->rows = r;
  sm->cols = c;

  sm->entries = (SMEntry *)memMalloc( sizeof(SMEntry) * total );

  for ( i = 0; i < total; i++ ) {
    sm->entries[i] = (SMEntry)memMalloc( sizeof(SMEntryStruct) );
    sm->entries[i]->score = 0.0;
    sm->entries[i]->flag = FALSE;
  }

  return( sm );
}

/* ----------------------------------------------------------------------
 * deletes score matrix
 * 
 * This function deallocates a \b ScoreMatrix. It first deallocates
 * all elements of <b> sm->entries[] </b> and then the array itself as well
 * as the \b ScoreMatrixStruct.
 */
void smDelete(ScoreMatrix sm)
{
  int i;

  if ( sm == NULL ) {
    cdgPrintf(CDG_ERROR, "smDelete: argument must not be NULL\n");
    abort();
  }

  for ( i = 0; i < sm->rows * sm->cols; i++ )
    memFree( sm->entries[i] );

  memFree( sm->entries );
  memFree( sm );
}

/* ----------------------------------------------------------------------
 * sets a score matrix element to a new value, returns old value
 */
double smSetScore(ScoreMatrix sm, double score, int r, int c)
{
  double old;

  if ( sm == NULL ) {
    cdgPrintf(CDG_ERROR, "smSet: argument must not be NULL\n");
    abort();
  }

  old = sm->entries[ c * sm->rows + r ]->score;
  sm->entries[ c * sm->rows + r ]->score = score;

  return( old );
}

/* ----------------------------------------------------------------------
 * sets a flag matrix element to a new value, returns old value
 */
Boolean smSetFlag(ScoreMatrix sm, Boolean flag, int r, int c)
{
  Boolean old;

  if ( sm == NULL ) {
    cdgPrintf(CDG_ERROR, "smSet: argument must not be NULL\n");
    abort();
  }

  old = sm->entries[ c * sm->rows + r ]->flag;
  sm->entries[ c * sm->rows + r ]->flag = flag;

  return( old );
}

/* ----------------------------------------------------------------------
 * sets flags of all matrix elements to a new value
 * 
 * This function sets the \b flag of all elements of
 * \b sm->entries[] to \b flag.
 */
void smSetAllFlags(ScoreMatrix sm, Boolean flag)
{
  int i;

  if ( sm == NULL ) {
    cdgPrintf(CDG_ERROR, "smSet: argument must not be NULL\n");
    abort();
  }

  for ( i = 0; i < sm->rows * sm->cols; i++ ) 
    sm->entries[ i ]->flag = flag;

}

/* ----------------------------------------------------------------------
 * retrieves a score matrix element score
 */
double smGetScore(ScoreMatrix sm, int r, int c)
{
  if ( sm == NULL ) {
    cdgPrintf(CDG_ERROR, "smGet: argument must not be NULL\n");
    abort();
  }

  return( sm->entries[ c * sm->rows + r ]->score );
}

/* ----------------------------------------------------------------------
 * retrieves a score matrix element flag
 * @returns the \b flag of the specified element.
 */
Boolean smGetFlag(ScoreMatrix sm, int r, int c)
{
  if ( sm == NULL ) {
    cdgPrintf(CDG_ERROR, "smGet: argument must not be NULL\n");
    abort();
  }

  return( sm->entries[ c * sm->rows + r ]->flag );
}

/* ---------------------------------------------------------------------- */
/** @} */
