/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: shift.c,v 1.20 2006-03-10 14:51:51 foth Exp $
 */


/* -------------------------------------------------------------------------
  @addtogroup Shift
  @{

  This module provides a method of solving the parsing problem via the
  shift-reduce algorithm for dependency structures. This algorithm
  employs two basic structures: a pointer into the input string and a
  stack of words already processed but still available for further
  processing.

  At any stage in parsing, up to four actions are possible:

   - SHIFT: the word under the input pointer is pushed onto the stack
     and the pointer is advanced by one.

   - REDUCE: the stack is popped.

   - LEFT: the top stack word is subordinated under the current input word.

   - RIGHT: the current input word is subordinated under the top stack word.

  By choosing the correct sequence of actions, any projective dependency
  tree can be constructed; see Nivre(2003) (IWPT 2003, Nancy) for details.
  The critical decision is not only whether or not to establish a
  dependency link, but also when to reduce the stack (ideally, immediately
  after the word has accepted its last modifier) and when to shift
  (ideally, only if the input word will modify a later word).

  The function shift_reduce() provides an implementation of this
  algorithm for constraint nets. Only the declared main level is
  constructed via shift/reduce; other levels are computed on the fly
  via complete search, if at all. A number of different policies is
  available to steer the parser, including statistical information
  learnt from other corpora.

*/

/* -- INCLUDES ------------------------------------------------------------- */
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "cdg.h"
#include "transform.h"
#include "hook.h"
#include "tagger.h"
#include "set.h"

/* -- MACROS --------------------------------------------------------------- */

/* -- TYPE DEFINITIONS ----------------------------------------------------- */

/* The set of possible transitions and their names. */
#define nr_actions 4
typedef enum { left, right, reduce, shift } Action;
String names[nr_actions] = {"LEFT","RIGHT","REDUCE","SHIFT"};

/* Representation of action counts for an abstract parse state.

   Each CorpusCount records how often each possible decision was taken for a
   particular class of parse states. It does not record *which* parse state
   this was; the association is done by key/value pairing in the global
   parse table.

   counts[] is indexed with Action variables; each cell holds the
   number of times that this action was performed in this parse state.

   The field total is the sum of all four fields, and the field best
   holds the Action with the highest count.
*/
typedef struct {
  int count[nr_actions];
  int total;
  Action best;
} CorpusCountStruct;
typedef CorpusCountStruct *CorpusCount;

/* Representation of a concrete state during parsing.

   A and P are the normal transformation structures. The parse state owns
   A, but not P.

   STRUCTURE is an array of subordination information that is maintained
   redundantly along with a->slots.

   DOT is the value arc->from of the next input value.

   STACK is a stack of the arc->from values of words on the stack.

   SCORE is the probability of the derivation history of the state.
   This has nothing to do with grammar penalties; to assess parse states
   according to their linguistic merit, read a->b instead.

   NO is a unique (within a parsing run) number.

*/
typedef struct {
  Analysis a;           /* partially built syntactic analysis */
  Problem P;            /* enclosing Problem */
  int *structure;       /* array of dependency structure */
  int dot;              /* arc->from of the next word in the input */
  List stack;           /* stack of arc->from values symbolizing words  */
  Number score;         /* number for ordering states in the agenda */
  int no;               /* unique identification number */

} ParseStateStruct;
typedef ParseStateStruct *ParseState;

/* Representation of a possible transition between parse states.

   A transition differs from an Action mainly because there can be several
   possible LEFT or RIGHT actions possible at the same time, and the field
   p details which is meant. It can also carry information about the
   probability of this transition at this point, which cannot be computed
   from either the current or the following state alone.
*/
typedef struct {
  Action a;             /* type of transition */
  ParseState from;      /* state to apply this Transition to */
  Patch p;              /* corresponding Patch (for LEFT/RIGHT only) */
  Number score;         /* score of this transition (not of the state itself) */
} TransitionStruct;
typedef TransitionStruct *Transition;

/* -- VARIABLES ------------------------------------------------------------ */

/* reduction strategy for shift-reduce parsing */
String reductionPolicy;

/* String of characters indicating predetermined transitions.

   This is merely a pointer into the middle of reductionPolicy.
*/
String program = NULL;

/* number of choices to try at each stage */
int beam_width = 1;

/* weight (out of 1.0) given to corpus information in hybrid mode */
float corpusstrength = 0.5;

/* name of file with corpus counts */
String table_file = NULL;

/* in-memory representation of table_file */
Hashtable table = NULL;

/* number of main level of current grammar */
int main_level_no = 0;

/* module-local instance of worstBadness() */
Badness min;

ConstraintNet net = NULL;
Problem P = NULL;

static int statecount;

/* -- FUNCTIONS ------------------------------------------------------------ */

/* ----------------------------------------------------------------------
   Translate String to Action.

   The only allowed names for actions are LEFT, RIGHT, SHIFT and REDUCE.
   Everything else is regarded as a funny way of specifying SHIFT.
*/
Action string2action(String s) {
  int i;
  for(i = 0; i < nr_actions; i++) {
    if(!strcmp(names[i], s)) {
      return (Action)i;
    }
  }
  return shift;
}

/* ----------------------------------------------------------------------
   Return initialized CorpusCount.

   All counts are 0; ->best is initially SHIFT.
*/
CorpusCount newCorpusCount(void) {
  CorpusCount result = memMalloc(sizeof(CorpusCountStruct));
  int i;
  for(i = 0; i < nr_actions; i++) {
    result->count[i] = 0;
  }
  result->total = 0;
  result->best = shift;
  return result;
}

/* ----------------------------------------------------------------------
   Delete a CorpusCount.
*/
void freeCorpusCount(CorpusCount cc) {
  memFree(cc);
}

/* ----------------------------------------------------------------------
   Return initialized ParseState.
*/
ParseState newParseState(Analysis a) {
  ParseState ps = memMalloc(sizeof(ParseStateStruct));
  ps->a = aClone(a);
  ps->P = a->P;
  ps->structure = memMalloc(sizeof(int) * a->P->maxTimepoint);
  ps->dot = 0;
  ps->stack = NULL;
  ps->score = 1.0;
  ps->no = statecount++;
  return ps;
}

/* ----------------------------------------------------------------------
   Return somewhat deep copy of a ParseState.
*/
ParseState psClone(ParseState ps) {
  int i;
  ParseState new = memMalloc(sizeof(ParseStateStruct));

  new->a = aClone(ps->a);
  new->P = ps->P;
  new->structure = memMalloc(sizeof(int) * ps->P->maxTimepoint);
  for(i = 0; i < ps->P->maxTimepoint; i++) {
    new->structure[i] = ps->structure[i];
  }
  new->dot = ps->dot;
  new->stack = listClone(ps->stack);
  new->score = ps->score;
  new->no = statecount++;

  return new;
}

/* ----------------------------------------------------------------------
   Delete a ParseState.
*/
void freeParseState(ParseState ps) {
  listDelete(ps->stack);
  aDelete(ps->a);
  memFree(ps->structure);
  memFree(ps);
}

/* ----------------------------------------------------------------------
   Return new Transition.
*/
Transition newTransition(ParseState ps, Action a) {
  Transition result = memMalloc(sizeof(TransitionStruct));

  result->a = a;
  result->from = ps;
  result->p = NULL;
  result->score = 1.0;

  return result;
}

/* ----------------------------------------------------------------------
   Free a Transition.
*/
void freeTransition(Transition t) {
  pDelete(t->p);
  memFree(t);
}

/* ----------------------------------------------------------------------
   Print representation of a Transition.
*/
void tPrint(unsigned long mode, Transition t) {
  cdgPrintf(mode, "%s: ", names[t->a]);
  if(t->p) {
    lvPrint(CDG_INFO, vectorElement(t->p->selection, 0), NULL, 0);
    cdgPrintf(CDG_INFO, "\n");
  } else {
    cdgPrintf(CDG_INFO, "\n");
  }
}

/* ----------------------------------------------------------------------
   Compare two Transitions.
*/
Boolean tCompare(Transition a, Transition b) {

  /* The Badness that results from a Transition is either the field
     `predicted' of its Patch, or just the current Badness. */
  Badness ba = a->p ? a->p->predicted : a->from->a->b;
  Badness bb = b->p ? b->p->predicted : b->from->a->b;
  
  if(!strcmp(reductionPolicy,"best")) {

    /* compare by Badness first  */
    if(bCompare(ba,bb)) {
      return TRUE;
    }
    if(bCompare(bb,ba)) {
      return FALSE;
    }
  }
  
  /* compare by transition scores first */
  if(a->score > b->score) {
    return TRUE;
  }
  if(a->score < b->score) {
    return FALSE;
  }

  /* compare by Badness next */
  if(ba && bb) {
    if(bCompare(ba,bb)) {
      return TRUE;
    }
    if(bCompare(bb,ba)) {
      return FALSE;
    }
  }
  
  /* then by transition type */
  if((int)a->a > (int)b->a) {
    return TRUE;
  }
  if((int)a->a < (int)b->a) {
    return FALSE;
  }
  
  /* force stable sort */
  if(a->p && b->p) {
    LevelValue lva = vectorElement(a->p->selection,0);
    LevelValue lvb = vectorElement(b->p->selection,0);
    if(lva->indexWRTNet > lvb->indexWRTNet) {
      return TRUE;
    }
    if(lva->indexWRTNet < lvb->indexWRTNet) {
      return FALSE;
    }
  }

  cdgPrintf(CDG_INFO, "WARNING: using unstable sorting in tCompare()\n");

  return FALSE;

}

/* ----------------------------------------------------------------------
   Free a transition table.
*/
void freeTable(Hashtable h) {
  List l,m;

  if(h) {
    for(l = m = hashListOfKeys(h); l != NULL; l = l->next) {
      CorpusCount cc = hashGet(table, l->item);
      freeCorpusCount(cc);
    }
    hashDelete(h);
    listDelete(m);
  }
}

/* ----------------------------------------------------------------------
   Read transition information from file.

   This function reads the file FILENAME and fills table with corresponding
   CorpusCount elements.

   If the same filename was given to the previous invocation, it succeeds
   without reading. (If the file has changed in the mean time this is not
   noticed. Don't do that.)
*/
Boolean read_table(String filename) {
  static String last_read_table = NULL;
  FILE *file;
  char buffer[1024];

  /* catch undefined table */
  if(!filename) {
    cdgPrintf(CDG_ERROR, "%s%s",
              "ERROR: No transition table defined. ",
              "Use `set shift-reduce-table <filename>'.\n");
    return FALSE;
  }

  /* skip already loaded table */
  if(last_read_table && !strcmp(filename, last_read_table)) return TRUE;

  /* reject nonexistent table */
  if (NULL == (file = fopen(filename, "r"))) {
    cdgPrintf(CDG_ERROR, "ERROR: could not open table `%s'\n", filename);
    return FALSE;
  }

  /* free previous table */
  if(table) {
    freeTable(table);
  }

  /* actually read */
  table = hashNew(100, 0.7, hashStringHashFunction, hashStringEqualFunction);
  while(!feof(file)) {
    String state;
    String token;
    int count, total, record;
    CorpusCount cc;

    fgets(buffer, 1024, file);

    /* read `transition' */
    token = strtok(buffer, " ");
    if(token && strcmp(token, "transition")) {
      continue;
    }

    total = record = 0;

    /* read width */
    token = strtok(NULL, " ");
    if(!token) {
      continue;
    }

    /* read state */
    state = strtok(NULL, " ");
    if(!state) {
      continue;
    }

    cc = newCorpusCount();

    /* read counts */
    while((token = strtok(NULL, " "))) {
      Action a = string2action(token);
      token = strtok(NULL, " ");
      count = 0;
      sscanf(token, "%d", &count);
      if(record < count) {
        record = count;
        cc->best = a;
      }
      total += count;
      cc->count[a] = count;
    }
    cc->total = total;
    hashSet(table, strRegister(state), cc);
  }

  fclose(file);

  cdgPrintf(CDG_INFO, "INFO: read %d stats from `%s'\n",
            hashSize(table), filename);

  last_read_table = strRegister(filename);

  return TRUE;
}


/* ----------------------------------------------------------------------
   Parse command words.

   The following named parameters are defined:

     beam=N: sets beam_width

     net=X: chooses the constraint to parse

     policy=X: chooses disambiguation policy (see decide())

     corpus=N: sets corpusstrength

   Returns FALSE only if a parameter is obviously unusable (e.g. the
   net does not exist).
*/
Boolean parse_options(int no, char **args) {
  int i;

  net = cnMostRecentlyCreatedNet;
  reductionPolicy = "table";

  for(i = 0; i < no; i++) {
    String s = strRegister(args[i]);
    char *key, *value;

    /* find named assignments */
    if(NULL != (key = strtok(s, "=")) &&
       NULL != (value = strtok(NULL, "="))) {

      if(!strcmp(key, "beam")) {
        sscanf(value, "%d", &beam_width);
      }
      else if(!strcmp(key, "net")) {
        if(!(net = cnFindNet(value))) {
          cdgPrintf(CDG_ERROR, "ERROR: net `%s' not known.\n", value);
          return FALSE;
        }
        P = findProblem(net);
      }
      else if(!strcmp(key, "policy")) {
        reductionPolicy = strRegister(value);
      }
      else if(!strcmp(key, "corpus")) {
        sscanf(value, "%f", &corpusstrength);
      }
      else {
        cdgPrintf(CDG_ERROR, "ERROR: can't use named parameter `%s'\n", key);
        return FALSE;
      }
      continue;
    }

    else {
      cdgPrintf(CDG_ERROR, "ERROR: can't use parameter `%s'\n", s);
      return FALSE;
    }
  }

  if (!net) {
    cdgPrintf(CDG_ERROR,
              "ERROR: `shift-reduce' needs a constraint net id or a default net\n");
    return FALSE;
  }

  /* verify that a known policy is in effect */
  if(!strncmp(reductionPolicy, "exact:", 6)) {
    program =reductionPolicy+6;
  }

  else if(!strcmp(reductionPolicy, "table")) {
    if(!read_table(table_file)) {
      return FALSE;
    }
  }
  else if(!strncmp(reductionPolicy, "hybrid", 6)) {
    if(!read_table(table_file)) {
      return FALSE;
    }
  }

  else if(!strcmp(reductionPolicy, "best") ||
          !strcmp(reductionPolicy, "nivre")) {
    /* do nothing */
  }

  else {
    cdgPrintf(CDG_ERROR, "ERROR: Unknown policy `%s'.\n", reductionPolicy);
    return FALSE;
  }

  cdgPrintf(CDG_ERROR, "INFO: using reduction policy `%s'\n", reductionPolicy);


  P = findProblem(net);

  return TRUE;
}


/* ----------------------------------------------------------------------
   Turn the LV in slot WHERE to NIL, choosing the best way of doing so
*/
void turn_to_nil(ParseState ps, int where) {
  Patch p;
  if(!slotActive(ps->a, where)) return;
  aRedirect(ps->a, where, -1);
  p = optimizeDomains(ps->a, (Pointer)where,
                      OD_INT | OD_PREDICT |
                      OD_LABEL | OD_LEXICAL | OD_DELETED);
  if(p) {
    apply(ps->a, p);
  } else {
    cdgPrintf(CDG_WARNING, "WARNING: can't force domain @%d to NIL!\n", where);
  }
  if(where % ps->P->nrLevels == main_level_no) {
    ps->structure[where / ps->P->nrLevels] = -1;
  }
}

/* ----------------------------------------------------------------------
   Return top item of the stack of PS, or -1.
*/
int top(ParseState ps) {
  if(ps->stack) {
    return (int)ps->stack->item;
  }
  return -1;
}

/* ----------------------------------------------------------------------
   Perform the pop operation on PS.
*/
void pop(ParseState ps) {
  ps->stack = listDeleteElement(ps->stack, ps->stack->item);
}

/* ----------------------------------------------------------------------
   Perform the push operation on PS.
*/
void push(ParseState ps, int i) {
  ps->stack = listPrependElement(ps->stack, (Pointer)i);
}

/* ----------------------------------------------------------------------
   Print representation of entire parse state on stdout.

   This prints the readings of the stack and input words, not just their
   categories.
*/
void dump_state(ParseState ps) {
  List r = listReverse(ps->stack);
  List l;

  cdgPrintf(CDG_DEBUG, "[%d] ", ps->no);
  cdgPrintf(CDG_DEBUG, "[ ");
  for(l = r; l != NULL; l = l->next) {
    int i = (int)l->item;
    LexemNode ln = selectedAt(ps->a, i);
    if(ln) {
      cdgPrintf(CDG_DEBUG, "%s ", ln->lexem->word);
    } else {
      cdgPrintf(CDG_DEBUG, "%d ", i);
    }
  }
  cdgPrintf(CDG_DEBUG, "] . ");
  listDelete(r);
  if(ps->dot < ps->P->nrTimepoints) {
    cdgPrintf(CDG_DEBUG, "%s ",  selectedAt(ps->a, ps->dot)->lexem->word);
  }

  bPrint(CDG_DEBUG, ps->a->b);
  cdgPrintf(CDG_DEBUG, " ((%4.3e))\n", ps->score);

}

/* ----------------------------------------------------------------------
   Get syntactic category of word at timepoint WHERE.

   The category is defined as the feature taggerCategoryIndex of the
   currently selected word that starts at that time point. If the lookup
   somehow fails, NULL is returned.

   Note that during transformational analysis, the category of a word can
   change over time as different lexical variants are swapped in.
*/
String get_cat(ParseState ps, int where) {
  LexemNode ln = selectedAt(ps->a, where);
  if(-1 == taggerCategoryIndex) return NULL;
  if(ln) {
    return ((Value)ln->lexem->values[taggerCategoryIndex])->data.string;
  } else {
    return NULL;
  }
}

/* ----------------------------------------------------------------------
   Return string representation of current parse state.

   This is not used for screen output, but as keys into the hash table of
   corpus statistic, therefore the format is more terse than that of
   dump_state(). It might look like this: ART,000/NN|ADV. See
   shift-reduce.pl for explanation.
   
   The representation is currently hard-coded to that generated by
   shift-reduce.pl with the options -i -a -l2 (i.e., encode top word
   environment, encode distance between visible words, encode two
   words of lookahead.) Tables generated with other options will
   silently fail to work.

   If NRFEATURES is smaller than 4, the feature set used is reduced in
   the same way as in shift-reduce.pl, i.e. first the distance feature
   goes, then the lookahead is reduced, finally the top word environment
   goes.

   The result is written into RESULT, using not more than LENGTH
   characters. This may result in a truncated result if the caller gives
   too low a limit.

*/
void encode_state(ParseState ps, int nrfeatures, String result, int length) {

  String topcat, dotcat, lookaheadcat;
  char attach[4];
  int dist = 0;
  int top_no = top(ps);
  int i;

  topcat = dotcat = lookaheadcat = NULL;

  /* top word category */
  if(ps->stack) {
    topcat = get_cat(ps, top_no);
  } else {
    topcat = "NIL";
  }

  /* top word environment */
  if(nrfeatures > 0) {
    strcpy(attach, "000");
    if(ps->stack) {
      if(ps->structure[top_no] != -1) {
        attach[0] = '1';
      }
      for(i = 0; i < top_no; i++) {
        if(ps->structure[i] == top_no) {
          attach[1] = '1';
          break;
        }
      }
      for(i = top_no; i < ps->dot; i++) {
        if(ps->structure[i] == top_no) {
          attach[2] = '1';
          break;
        }
      }
    }
  }

  /* lookahead categories */
  if(nrfeatures > 1) {
    if(ps->dot < ps->P->maxTimepoint) {
      dotcat = get_cat(ps, ps->dot);
    } else {
      dotcat = "$";
    }
  }
  if(nrfeatures > 2) {
    if(ps->dot+1 < ps->P->maxTimepoint) {
      lookaheadcat = get_cat(ps, ps->dot+1);
    } else {
      lookaheadcat = "$";
    }
  }

  /* distance feature */
  if(nrfeatures > 3) {
    if(ps->stack) {
      dist = ps->dot - top_no;
    } else {
      dist = 0;
    }
  }

  switch(nrfeatures) {
  case 4:
    snprintf(result, length, "%s,%s/%d/%s|%s:",
             topcat, attach, dist, dotcat, lookaheadcat);
    break;
  case 3:
    snprintf(result, length, "%s,%s/%s|%s:",
             topcat, attach, dotcat, lookaheadcat);
    break;
  case 2:
    snprintf(result, length, "%s,%s/%s:", topcat, attach, dotcat);
    break;
  case 1:
    snprintf(result, length, "%s,%s:", topcat, attach);
    break;
  case 0:
    snprintf(result, length, "%s:", topcat);
    break;
  }
}

/* ----------------------------------------------------------------------
   Find most appropriate parse state in the table.

   The current parse state is encoded and looked up in table. If no entry
   exists there, the context is reduced until there is one. If there is
   no matching entry in the table at all, NULL is returned.
*/
CorpusCount read_count(ParseState ps) {
  CorpusCount cc = NULL;
  int i = 4;

  while(!cc) {
    char buffer[1024];
    CorpusCount cc;
    encode_state(ps, i, buffer, 1024);
    cc = hashGet(table,(Pointer)buffer);

#ifdef DEBUGENCODE_STATE
    cdgPrintf(CDG_DEBUG, "state = %s\n", buffer);
#endif
    if(cc) {
      return cc;
    }
    i--;
    if(i<0) break;
  }
  return NULL;
}

/* ----------------------------------------------------------------------
   Is A possible in PS?

   SHIFT is possible whenever there is input left.

   REDUCE is only possible there is a top stack word. Note that in the
   original implementation, REDUCE was only possible if that word was
   already subordinated. We relax that condition

   LEFT and RIGHT are possible whenever there are two visible words,
   even when the grammar forbids all subordinations between them.
*/
Boolean possible(ParseState ps, Action a) {

  int index;
  LevelValue lv;

  switch(a) {

  case left:
    if(!ps->stack || ps->dot == ps->P->maxTimepoint) {
      return FALSE;
    }
    index = top(ps) * ps->P->nrLevels + main_level_no;
    lv = vectorElement(ps->a->slots, index);
    return lv->regents == NULL;

  case right:
    if(!ps->stack || ps->dot == ps->P->maxTimepoint) {
      return FALSE;
    }
    return TRUE;

  case shift:
    return ps->dot < ps->P->maxTimepoint;

  case reduce:
    return ps->stack != NULL;
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
   Compute Vector of successors of a particular type.

   If A is not possible, NULL is returned rather than a useless empty
   Vector.
*/
Vector find_typed_successors(ParseState ps, Action a) {

  LevelValue lv;
  Vector result = vectorNew(10);
  List l, m;
  int down, up, oldup = -1, index;
  int horizon = ps->a->subStop;
  
  if(!possible(ps, a)) {
    return NULL;
  }

  switch(a) {

  case shift:
  case reduce:

    /* generate the unique transition */
    vectorAddElement(result, newTransition(ps, a));
    break;

  case left:
  case right:
      /* when judging a transition that introduces a new LV,
         we should extend the range of time points to be evaluated by one,
         so that the score of the new LV itself is also considered. */
    aRestrict(ps->a, P->minTimepoint, ps->dot+1);

    /* generate list of possible transitions */
    if(a == left) {
      down = top(ps), up = ps->dot;
    } else {
      down = ps->dot, up = top(ps);
    }

    /* search for Patches that establish this edge */
    index = down * ps->P->nrLevels + main_level_no;
    lv = vectorElement(ps->a->slots, index);
    if(lv && lv->regents) {
      oldup = ((LexemNode)lv->regents->item)->arc->from;
    }
    aRedirect(ps->a, index, up);
    for(l = m = dSearch(ps->a, (Pointer)index,
              OD_INT | OD_LEXICAL | OD_LABEL);
        l != NULL; l = l->next) {
      Transition t = newTransition(ps, a);
      t->p = l->item;

      pLexicalize(t->p, 0, FALSE);

      aRestrict(ps->a, P->minTimepoint, ps->dot+1);
      pJudge(t->p);
      
      vectorAddElement(result, t);
    }
    listDelete(m);
    aRedirect(ps->a, index, oldup);

    /* reset evaluation limit */
    if(right == a) {
      aRestrict(ps->a, P->minTimepoint, horizon);
    }
 }
  return result;
}


/* ----------------------------------------------------------------------
   Compute the list of successor parse states to PS.

   The result is a list of possible states reachable from PS. Depending on
   which policy has been specified, various methods are used to select
   possible choices. The order of elements is unpredictable.

   The following policies currently exist:

   policy=exact:RLslrS: The exact action at the corresponding point in the
   program is used (where r=REDUCE, R=RIGHT, s=shift, L=LEFT).
   Implies beam_width == 1.

   policy=best: prefer the transitions that result in the best Badness.

   policy=nivre: prefer transitions in the order LEFT--RIGHT--REDUCE--SHIFT.

   policy=table: the most closely corresponding parse state is determined
   via read_count(), and transitions of its most frequent type are preferred.

   policy=hybrid: the judgements of `best' and `table' are combined in
   non-obvious ways.

   It is not guaranteed that any successors are found.

   With some decision policies, only one Transition is returned even
   if beam_width is larger than 1.

   Even policies other than `best' will usually try to use the more
   grammatical of two Transitions if both are of the same type.
   
*/
List find_successors(ParseState ps) {

  List result = NULL;
  Vector v = NULL;

  /* predetermined policy */
  if(!strncmp(reductionPolicy, "exact", 5)) {
    if(program) {
      Action a;
      switch(program[0]) {
      case 'L':
        a = left; break;
      case 'R':
        a = right; break;
      case 's':
        a = shift; break;
      case 'r':
        a = reduce; break;
      default:
        cdgPrintf(CDG_DEBUG, "ERROR: invalid action spec `%c'\n", program[0]);
        exit(1);
      }
      program++;
      v = find_typed_successors(ps, a);
      if(!v) {
        cdgPrintf(CDG_ERROR,
                  "ERROR: %s transition not possible, exiting.\n",
                  names[a]);
        return NULL;
      }
    }
  }

  else if(!strcmp(reductionPolicy,"best")) {
    int i, j;
    v = vectorNew(10);
    for(i = 0; i < nr_actions; i++) {
      Vector vv = find_typed_successors(ps, (Action)i);
      if(vv) {
        for(j = 0; j < vectorSize(vv); j++) {
          Transition t = vectorElement(vv, j);
          vectorAddElement(v, t);
        }
        vectorDelete(vv);
      }
    }
  }

  /* Nivre's base line policy */
  else if(!strcmp(reductionPolicy, "nivre")) {
    v = vectorNew(1);
    /* Just return one Transition in any case, disregarding
       alternatives allowed by the grammar. */
    if(possible(ps, left)) {
      vectorAddElement(v, newTransition(ps, left));
    }
    if(possible(ps, right)) {
      vectorAddElement(v, newTransition(ps, right));
    }
    if(possible(ps, reduce)) {
      vectorAddElement(v, newTransition(ps, reduce));
    }
    if(possible(ps, shift)) {
      vectorAddElement(v, newTransition(ps, shift));
    }

    cdgPrintf(CDG_ERROR,
              "ERROR: No transition possible, this should not happen.\n");
    exit(1);
  }

  /* corpus-based decisions */
  if(!strcmp(reductionPolicy, "table")) {
    CorpusCount cc = read_count(ps);
    int i, j;
    Vector vv;
    v = vectorNew(1);
    for(i = 0; i < nr_actions; i++) {
      /* if this state has corpus counts, and A's count is 0, ignore A */
      if(cc && cc->count[i] == 0) {
        continue;
      }
      if(!cc && !possible(ps, i)) {
        continue;
      }
      vv = find_typed_successors(ps, (Action)i);
      if(vv) {
        for(j = 0; j < vectorSize(vv); j++) {
          Transition t = vectorElement(vv, j);
          if(cc) {
            t->score = cc->count[i] / (float)cc->total;
          }
          vectorAddElement(v, t);
        }
        vectorDelete(vv);
      }
    }
  }

#if 0
  /* grammar & corpus conjoined */
  if(!strncmp(reductionPolicy, "hybrid", 6)) {
    CorpusCount cc = read_count(ps);

    v = vectorNew(10);
    if(cc) {
      float record = 0.0;
      Action choice = shift;
      /* try all four choices,
         and combine the grammar's and the corpus's judgement */
      for(i = 0; i < nr_actions; i++) {
        Action a = i;
        if(try(ps, a)) {
          Badness new = ps->predict[a];
          double f, grammar, corpus;
          double heur = corpusstrength;

          grammar = exp(-new->hard) * new->soft;
          corpus = cc->count[a]/(double)cc->total;
          f = heur * corpus + (1-heur) * grammar;

#ifdef DEBUGDECIDE
          cdgPrintf(CDG_ERROR,
                    "%s: hard %d soft %4.3e, corpus %d/%d\n",
                    names[a],
                    new->hard,
                    new->soft,
                    cc->count[a], cc->total);
          cdgPrintf(CDG_ERROR,
                    "grammar: %4.3e, corpus: %4.3e, f: %4.3e\n",
                    grammar, corpus, f);

          if(a == left) {
            pDebug(ps->a, leftPatch);
          }
          if(a == right) {
            pDebug(ps->a, rightPatch);
          }
#endif

          if(f > record) {
            record = f;
            choice = a;
          }
        }
      }
      return choice;

    } else {

      /* fall back */
      try(ps, shift);
      return shift;
    }
  }
#endif

  /* we have promised not to return more then beam_width items, but there
     are no limited list insertion functions. Moreover, there is no way of
     efficiently truncating a List at position n (because ListStruct is
     hidden). Therefore we can only emulate the functionality by hand. */
  {
    int i, j;
    vectorSort(v, tCompare);
    for(i = beam_width; i < vectorSize(v); i++) {
      Transition t = vectorElement(v, i);
      freeTransition(t);
    }
    j = vectorSize(v);
    for(i = beam_width; i < j; i++) {
      vectorRemoveElementAt(v, beam_width);
    }
  }

  result = vectorToList(v);
  vectorDelete(v);
  return result;

}

/* Actually perform a transition.

   If LEFT or RIGHT is specified and t->p is non-NULL, that Patch is
   applied, otherwise the Analysis is transformed via a raw aRedirect()
   call.

   If A is not currently possible, the behaviour is undefined.

   The Transition t is deleted.
*/
ParseState transit(ParseState ps, Transition t) {

  ParseState ps2 = psClone(ps);
  int up, down;
  
  cdgPrintf(CDG_DEBUG, "applying ");
  tPrint(CDG_INFO, t);

  switch(t->a) {

  case left:
  case right:
    if(t->a == left) {
      down = top(ps2), up = ps2->dot;
    } else {
      down = ps2->dot, up = top(ps2);
    }

    /* If the transition has already computed a Patch, use that. */
    if(t->p) {
      lvPrint(CDG_INFO, vectorElement(t->p->selection,0), NULL, 0);
      cdgPrintf(CDG_INFO, "\n");
      apply(ps2->a, t->p);
      t->p = NULL;
    }

    /* Otherwise just issue a raw redirect. */
    else {
      int index = down * ps2->P->nrLevels + main_level_no;
      aRedirect(ps2->a, index, up);
    }

    ps2->structure[down] = up;
    aComputeAux(ps2->a, down, FALSE, FALSE);
    if(t->a == left) {
      pop(ps2);
    } else {
      push(ps2, ps2->dot++);
    }
    break;

  case reduce:
    pop(ps2);
    break;

  case shift:
    push(ps2, ps2->dot++);
    break;
  }

  /* Only table-driven parsing assigns scores to the transitions
     themselves. With other methods, t->score is always 1, so it doesn't
     hurt to multiply unconditionally.  */
  if(cdgMultiply == cdgAggregation) {
    ps2->score *= t->score;
  } else {
    ps2->score = min(ps2->score, t->score);
  }

  aRestrict(ps2->a, ps2->a->P->minTimepoint, ps2->dot+1);
  
  dump_state(ps2);
  freeTransition(t);
  
  return ps2;
}

/* ----------------------------------------------------------------------
   Shift-reduce-inspired parsing.

   The constraint net named (or implied) is subjected to the shift-reduce
   algorithm as described by Nivre(2003).

   Various policies are implemented for deciding between possible
   choices at each stage, selectable by the `policy=X' option.

   The entire parsing run is deterministic only if beamwidth == 1.
   Otherwise, backtracking or similar searches are performed.

   Returns FALSE if an error occurs before parsing.

   The result is registered as a normal Parse.
*/
Boolean shift_reduce(int no, char **args) {

  Agenda ag;
  int i;
  ParseState ps;
  Analysis result = NULL;
  List l;
  
  Level mainlevel = inputGetMainlevel(inputCurrentGrammar);
  if(mainlevel) {
    main_level_no = mainlevel->no;
  }

  /* interpret options */
  if(!parse_options(no, args)) {
    return FALSE;
  }

  if(!P) {
    if(!(P = net2problem(net, 0.0, NULL))) {
      return FALSE;
    }
  }
  
  timerStart(P->profileTime);
  statecount = 0;
  ps = newParseState(P->current);
  /* trivialize the analysis */
  for(i = 0; i < P->nrSlots; i++) {
    turn_to_nil(ps, i);
  }

  l = lgMostProbablePath(P->net->lexemgraph);
  selectPath(ps->a, l);
  listDelete(l);

  /* Instruct the evaluation functions to confine themselves to the
     current horizon. Beyond that are essentially random LVs that should
     not influence the local decisions; e.g., we don't want a ROOT
     subordination deprecated because there still are other ROOT bindings
     farther to the right. */
  aRestrict(ps->a, P->minTimepoint, ps->dot+1);

  ag = agNew(100, freeParseState);
  agInsert(ag, ps->score, ps);

  while(!agIsEmpty(ag)) {
    List l, m;
    ps = agRemoveBest(ag);
#ifdef DEBUGSHIFTREDUCE
    cdgPrintf(CDG_DEBUG, "remove [%d]\n", ps->no);
#endif

    if(result && bCompare(result->b, ps->a->b)) {
      freeParseState(ps);
      continue;
    }

    m = find_successors(ps);
    for(l = m; l != NULL; l = l->next) {
      Transition t = l->item;
      ParseState new = transit(ps, t);

      /* find solution */
      if(new->dot == P->maxTimepoint) {
        if(!result || bCompare(new->a->b, result->b)) {
#ifdef DEBUGSHIFTREDUCE
          cdgPrintf(CDG_DEBUG, "retain [%d]\n", new->no);
          aPrint(CDG_DEBUG, new->a);
#endif
          if(result) {
            aDelete(result);
          }
          result = new->a;
          new->a = NULL;
        }
        freeParseState(new);
      }
      else {
#ifdef DEBUGSHIFTREDUCE
        cdgPrintf(CDG_DEBUG, "insert [%d]\n", new->no);
#endif
        agInsert(ag, new->a->b->soft - new->a->b->hard, new);
      }
    }
    freeParseState(ps);
    listDelete(m);
  }

  agDelete(ag);

  if(result) {
    aPrint(CDG_INFO, result);
    printConflicts(CDG_INFO, result);
    aVerify(result);
    addParse(result);
    if(!P->best || bCompare(result->b, P->best->b)) {
      newMaximum(result);
      P->current = aClone(result);
    }
  }

  {
    double totalDuration = timerElapsed(P->profileTime) / 1000.0;
    cdgPrintf(CDG_PROFILE, "PROFILE: %.3fs spent shifting \n", totalDuration);
  }

  P = NULL;
  aDelete(result);
  return TRUE;

}

/* ----------------------------------------------------------------------
 * Register CDG variables
 */
void shiftInitialize(void) {
  min = worstBadness();
  setRegister("shift-reduce-table",
              SET_STRING, &table_file, NULL, NULL, NULL, NULL);
}

/* ----------------------------------------------------------------------
   Return all memory used by the module
*/
void shiftFinalize(void) {
  freeTable(table);
  bDelete(min);
}

/* ------------------------------------------------------------------------- */
/* -- ENDOFFILE ------------------------------------------------------------ */
/** @} */


