/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/* ---------------------------------------------------------------------------
 * @addtogroup Write Write - functions that write to different document types
 * @author Horia F. Pop (see also AUTHORS and THANKS for more)
 * @date 19/11/99
 *
 * $Id: write.c,v 1.75 2006-02-22 13:17:39 foth Exp $
 *
 * \todo Most of the functions in this module are badly named, formated
 *       and not documented.
 *
 * @{
 */

/* ---------------------------------------------------------------------- */
#include <config.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <libgen.h>
#include "cdg.h"
#include "hook.h"
#include "increment.h"
#include "write.h"
#include "set.h"

/* -- DEFINES ----------------------------------------------------------- */

/*
 * for XFig printing functions
 */

#define write_ycoord  20000   /* y-coordinate of the nodes              */
#define write_radius     50   /* radius of circles for the nodes        */
#define write_distance 2000   /* distance between two consecutive nodes */
#define write_height    700   /* height between multiple arcs           */

/* -- VARIABLES --------------------------------------------------------- */
static int xmlDepth = 0;
static FILE *xmlFile = NULL;
static String xmlFileName = NULL;
static List xmlTagStack = NULL;
static String xmlBuffer = NULL; /* used in convAttrStr */
static int xmlBufferSize = 100;


/* -- FUNCTIONS --------------------------------------------------------- */

static void write_words_as_prolog(FILE*,Vector);
static void write_annos_as_prolog(FILE*,List);
static void write_specs_as_prolog(FILE*,List);
static void write_int_vector_as_prolog(FILE*,Vector);
static void write_string_vector_as_prolog(FILE*,Vector);
static void write_string_list_as_prolog(FILE*,List);

/* ----------------------------------------------------------------------
 * writeInitialize
 */
void writeInitialize(void)
{
  writeXmlFileSet("stdout");
  setRegister("xml", SET_STRING, &xmlFileName, NULL, NULL, writeCallback, NULL);
  xmlBuffer = memMalloc(xmlBufferSize);
}

/* ----------------------------------------------------------------------
 * writeFinalize
 */
void writeFinalize(void)
{
  writeXmlEndAll();

  if (xmlFile != stdout && xmlFile != stderr) {
    fclose(xmlFile);
    xmlFile = NULL;
    xmlFileName = NULL;
  }
  
  cdgFreeString(xmlFileName);
  memFree(xmlBuffer);
}

/* ----------------------------------------------------------------------
 * writes a LaTeX table of the net
 * -- each node is written on a column
 */

void writeStringForColumn(FILE * file, String s)
{
  fprintf(file, "\\shortstack[c]{");
  if (*s != '\0')
    {
      if (*s == '_')
	{
	  fprintf(file, "\\_");
	  s++;
	}
      else
	fprintf(file, "%1.1s", s++);
    }
  while (*s != '\0')
    {
      if (*s == '_')
	{
	  fprintf(file, "\\\\\\_");
	  s++;
	}
      else
	fprintf(file, "\\\\%1.1s", s++);
    }
  fprintf(file, "}");
}

/* ---------------------------------------------------------------------------
 * FIXME: write comments
 */
void writeConstraintNetColumns(ConstraintNet net, String filename)
{

  ConstraintNode n;
  LevelValue lv;
  Level level;
  int noOfValues, noOfValuesPerLexem, noOfValuesPerLevel;
  int i, j;
  List l;
  FILE *file;

  if ((file = fopen(filename, "w")) == NULL) {
    cdgPrintf(CDG_ERROR, "writeConstraintNet: ERROR: Cannot write file '%s'\n",
	      filename);
  } else {
    cdgPrintf(CDG_INFO, "INFO: writing constraint net '%s' to file '%s'\n",
	      net->id, filename);

    noOfValues = 0;
    for (i = 0; i < vectorSize(net->nodes); i++) {
      n = (ConstraintNode) vectorElement(net->nodes, i);
      for (j = 0; j < vectorSize(n->values); j++) {
	lv = (LevelValue) vectorElement(n->values, j);
	if (!lv->isDeleted)
	  noOfValues++;
      }
    }

    fprintf(file, "\\begin{tabular}{|*{%d}{c|}}\n", noOfValues + 3);
    fprintf(file, "\\hline\n");

    /* starting line 1 */
    /* skip over level, modifier and value */
    fprintf(file, "\\multicolumn{3}{|c||}{}\n");

    for (l = inputCurrentGrammar->levels; l != NULL; l = l->next) {
      level = (Level) l->item;
      noOfValuesPerLevel = 0;
      for (i = 0; i < vectorSize(net->nodes); i++) {
	n = (ConstraintNode) vectorElement(net->nodes, i);
	if (n->level != level)
	  continue;
	for (j = 0; j < vectorSize(n->values); j++) {
	  lv = (LevelValue) vectorElement(n->values, j);
	  if (!lv->isDeleted)
	    noOfValuesPerLevel++;
	}
      }
      fprintf(file, "& \\multicolumn{%d}{c|}{%s}\n", noOfValuesPerLevel,
	      level->id);
    }
    fprintf(file, "\\\\\n\\hline\n");
    /* ending line 1 */

    /* starting line 2 */
    /* skip over level, modifier and value */
    fprintf(file, "\\multicolumn{3}{|c||}{}\n");
    for (i = 0; i < vectorSize(net->nodes); i++) {
      n = (ConstraintNode) vectorElement(net->nodes, i);
      noOfValuesPerLexem = 0;
      for (j = 0; j < vectorSize(n->values); j++) {
	lv = (LevelValue) vectorElement(n->values, j);
	if (!lv->isDeleted)
	  noOfValuesPerLexem++;
      }
      fprintf(file, "& \\multicolumn{%d}{c|}{", noOfValuesPerLexem);
      writeStringForColumn(file, n->gn->arc->word);
      fprintf(file, "}\n");
    }
    fprintf(file, "\\\\\n\\hline\n");
    /* ending line 2 */

    /* starting line 3 */
    /* skip over level, modifier and value */
    fprintf(file, "\\multicolumn{3}{|c||}{}\n");
    for (i = 0; i < vectorSize(net->nodes); i++) {
      n = (ConstraintNode) vectorElement(net->nodes, i);
      for (j = 0; j < vectorSize(n->values); j++) {
	lv = (LevelValue) vectorElement(n->values, j);
	if (!lv->isDeleted) {
	  fprintf(file, "& ");
	  writeStringForColumn(file, lv->label);
	  fprintf(file, "\n");
	}
      }
    }
    fprintf(file, "\\\\\n\\hline\n");
    /* ending line 3 */

    /* starting line 4 */
    /* skip over level, modifier and value */
    fprintf(file, "\\multicolumn{3}{|c||}{}\n");
    for (i = 0; i < vectorSize(net->nodes); i++) {
      n = (ConstraintNode) vectorElement(net->nodes, i);
      for (j = 0; j < vectorSize(n->values); j++) {
	lv = (LevelValue) vectorElement(n->values, j);
	if (!lv->isDeleted) {
	  if (spec(lv->regents)) {
	    fprintf(file, "& ");
	    writeStringForColumn(file, ((LexemNode)lv->regents->item)->lexem->description);
	    fprintf(file, "\n");
	  } else {
	    fprintf(file, "& ");
	    writeStringForColumn(file, "nil");
	    fprintf(file, "\n");
	  }
	}
      }
    }
    fprintf(file, "\\\\\n\\hline\\hline\n");
    /* ending line 4 */

    fprintf(file, "\\end{tabular}\n");
    fclose(file);
  }
}

/* ----------------------------------------------------------------------
 * writes a LaTeX table of the net
 * -- each node is written on a line
 */
void writeStringForLine(FILE * file, String s)
{
  while (*s != '\0')
    {
      if (*s == '_')
	{
	  fprintf(file, "\\_");
	  s++;
	}
      else
	fprintf(file, "%1.1s", s++);
    }
}

/* ---------------------------------------------------------------------------
 * FIXME: write comments
 */
void writeConstraintNetLine(ConstraintNet net, String filename)
{
  ConstraintNode n;
  LevelValue lv;
  int i, j;
  FILE *file;

  if ((file = fopen(filename, "w")) == NULL) {
    cdgPrintf(CDG_ERROR, "writeConstraintNet: ERROR: Cannot write file '%s'\n",
	      filename);
  } else {
    /* 
     * uses longtable, provided by the tools package;
     * longtable has the advantage of allowing us to split it
     * so that it will be written on more than one page;
     * the regular table and tabular environments lack
     * this facility
     */
    cdgPrintf(CDG_INFO, "INFO: writing constraint net '%s' to file '%s'\n",
	      net->id, filename);

    fprintf(file, "\\begin{longtable}{||*{8}{c|}|}\n");
    fprintf(file, "\\hline \\hline\n");

    for (i = 0; i < vectorSize(net->nodes); i++) {
      n = (ConstraintNode) vectorElement(net->nodes, i);
      for (j = 0; j < vectorSize(n->values); j++) {
	lv = (LevelValue) vectorElement(n->values, j);

	if (j == 0) {
	  /* write node nr. */
	  fprintf(file, "%5d & ", i);

	  /* write n->level->id */
	  writeStringForLine(file, n->level->id);
	  fprintf(file, " & ");
	} else {
	  fprintf(file, "& & ");
	}

	/* write lv->score and lv->modifier info */
	fprintf(file, " %0.2f & (%02d-%02d) & ",
		lv->score,
                ((LexemNode)lv->dependents->item)->arc->from,
                ((LexemNode)lv->dependents->item)->arc->to);

	/* write lv->modifiee info */
	if (spec(lv->regents))
	  fprintf(file, " (%02d-%02d) & ",
		  ((LexemNode)lv->regents->item)->arc->from,
                  ((LexemNode)lv->regents->item)->arc->to);
	else
	  fprintf(file, " & ");

	/* write n->lexemnode->lexem->description */
	writeStringForLine(file, n->gn->arc->word);

	/* write lv->label */
	if (!lv->isDeleted) {
	  fprintf(file, " & ");
	  writeStringForLine(file, lv->label);
	}

	/* write lv->modifiee->lexem->description */
	if (!lv->isDeleted) {
	  if (spec(lv->regents)) {
	    fprintf(file, " & ");
	    writeStringForLine(file, ((LexemNode)lv->regents->item)->lexem->description);
	  } else
	    fprintf(file, " & nil");
	}

	/* end of table line */
	if (j == vectorSize(n->values) - 1)
	  fprintf(file, " \\\\ \\hline \\hline\n");
	else
	  fprintf(file, " \\\\ \\cline{3-8}\n");
      }
    }
    fprintf(file, "\\hline \\end{longtable}\n");
    fclose(file);
  }
}

/* ----------------------------------------------------------------------
 * writes a LaTeX table of the net
 */
void writeConstraintNet(ConstraintNet net, String filename)
{
  /*
   * uses only the lines version
   * the columns version produces an extremely wide table
   * and it will need to be improved
   */
  writeConstraintNetLine(net, filename);
}

/* 
 * simple functions to write objects to XFig file
 */

/*
 * writes the XFIG header
 */
void writeXFigHeader(FILE *file)
{
  fprintf(file, "#FIG 3.2\n");
  fprintf(file, "Landscape\n");
  fprintf(file, "Center\n");                                                 
  fprintf(file, "Inches\n");
  fprintf(file, "A4\n");
  fprintf(file, "100.00\n");
  fprintf(file, "Single\n");
  fprintf(file, "-2\n");
  fprintf(file, "1200 2\n");
}

/*
 * writes a circle with the center in (x0, y0) and of radius r
 */
void writeXFigCircle(FILE *file, int x0, int y0, int r)
{
  fprintf(file, "1 3 0 1 0 7 0 0 -1 0.0000 1 0.0000 %d %d %d %d %d %d %d %d\n",
          x0, y0, r, r, x0, y0, x0 + r, y0);
}

/*
 * writes a solid circle with the center in (x0, y0) and of radius r
 */
void writeXFigSolidCircle(FILE *file, int x0, int y0, int r)
{
  fprintf(file, "1 3 0 1 0 7 0 0 0 0.0000 1 0.0000 %d %d %d %d %d %d %d %d\n",
          x0, y0, r, r, x0, y0, x0 + r, y0);
}

/*
 * writes a line from (x1, y1) to (x2, y2)
 */
void writeXFigLine(FILE *file, int x1, int y1, int x2, int y2)
{
  fprintf(file, "2 1 0 1 0 7 0 0 -1 0.0000 0 0 -1 0 0 2\n");
  fprintf(file, "         %d %d %d %d\n",
	  x1, y1, x2, y2);
}

/*
 * writes the text s in a box of length l, and with the string origin at (x, y)
 * align = 0 (left), 1 (center), 2(right)
 */
void writeXFigText(FILE *file, int l, int x, int y, String s, int align)
{
  fprintf(file, "4 %d 0 0 0 0 12 0.0000 2 180 %d %d %d %s\\001\n",
	  align, l, x, y, s);
}

/*
 * writes a spline curve from (x1, y1) to (x2, y2), with three fixed points situated 
 * at X coordinates of x1+d/8, x1+d/2, x2-d/8 and
 * at heights of ho + h * 6/7, ho + h and ho + h * 6/7 over the line segment
 */
void writeXFigSpline(FILE *file, int x1, int y1, int x2, int y2, int ho, int h)
{
  fprintf(file, "3 2 0 1 0 7 0 0 -1 0.0000 0 0 0 5\n");
  fprintf(file, "         %d %d %d %d %d %d %d %d %d %d\n",
	  x1, y1, 
	  x1 + (x2 - x1) / 8, y1 + (y2 - y1) / 8 + ho + h * 6 / 7,
	  (x1 + x2) / 2, (y1 + y2) / 2 + ho + h,
	  x2 - (x2 - x1) / 8, y2 - (y2 - y1) / 8 + ho + h * 6 / 7,
	  x2, y2);
  fprintf(file, "         0.000 -1.000 -1.000 -1.000 0.000\n");
}

/* ---------------------------------------------------------------------- */

Boolean writeCompareInts(int n1, int n2)
{
  return (n1 < n2);
}

/*
 * returns the number of arcs from number1 to number2 in list m0
 */
int writeNumberArcs(List m0, int n1, int n2)
{
  int i;
  i = 0;
  while (m0 != NULL)
    {
      if ((((Arc) m0->item)->from == n1) &&
	  (((Arc) m0->item)->to   == n2))
	i++;
      m0 = m0->next;
    }
  return i;
}

/*
 * swaps two integers
 */
void writeSwapIntegers(int *i1, int *i2)
{
  int s;
  s = *i1;
  *i1 = *i2;
  *i2 = s;
}

/*
 * build the sorted vector of number labels of wordgraph nodes
 */
void writeBuildVector(List m, Vector *vectorOfNumbers)
{
  *vectorOfNumbers = vectorNew(0);
  while (m != NULL)
    {
      int number;
      number = ((Arc) m->item)->from;
      if (!vectorContains(*vectorOfNumbers, (void *)number))
	vectorAddElement (*vectorOfNumbers, (void *)number);
      number = ((Arc) m->item)->to;
      if (!vectorContains(*vectorOfNumbers, (void *)number))
	vectorAddElement (*vectorOfNumbers, (void *)number);
      m = m->next;
    }
  vectorSort(*vectorOfNumbers, writeCompareInts);
}

void writeWordOldOutput(FILE *file, List m)
{
  while (m != NULL)
    {
      fprintf(file, "#  %2d %2d %15s %f\n",
	      ((Arc) m->item)->from,
	      ((Arc) m->item)->to,
	      ((Arc) m->item)->word,
	      ((Arc) m->item)->penalty);
      m = m->next;
    }
}

int writeComputeHeightOver(List m, int index1, int index2, Vector vectorOfNumbers)
{
  int i, h;
  h = 0;
  for (i = index1; i<index2; i++)
    {
      int total;
      total = writeNumberArcs(m,
			      (int)vectorElement(vectorOfNumbers, i),
			      (int)vectorElement(vectorOfNumbers, i+1));
      total /= 2;
      if (total > h)
	h = total;
    }
  return (h);
}

void writeWordLine(FILE *file, Arc arc, int xcoord1, int xcoord2, Boolean withWeights)
{
  char s[50];

  /* writing the wordgraph lines */
  writeXFigLine(file, xcoord1 + write_radius, write_ycoord, xcoord2 - write_radius, write_ycoord);
  /* writing the word on the line */
  if (withWeights)
    sprintf(s, "%s (%4.2f)", arc->word, arc->penalty);
  else
    sprintf(s, "%s", arc->word);
  writeXFigText(file, write_distance, xcoord1 + write_distance / 2, write_ycoord - 50, s, 1);
}

void writeXFigWordArc(FILE *file, Arc arc, int xcoord1, int xcoord2, int heightover, int position, int numberExtraArcs, Boolean withWeights)
{
  char s[50];

  /*
   * writing an arc
   * position == -1 --> above the line
   * position == 1  --> below the line
   */
  writeXFigSpline(file, xcoord1, write_ycoord + position * write_radius, 
		  xcoord2, write_ycoord + position * write_radius, 
		  position * heightover, position * write_height * (numberExtraArcs / 2 + 1));

  /* writing the word on the arc */
  if (withWeights)
    sprintf(s, "%s (%4.2f)", arc->word, arc->penalty);
  else
    sprintf(s, "%s", arc->word);
  writeXFigText(file, write_distance, (xcoord1 + xcoord2) / 2,
		write_ycoord + position * (write_radius + heightover + write_height * (numberExtraArcs / 2 + 1)) + (65 + position * 115),
		s, 1);
}

/*
 * writes lattice to XFig file
 */
void writeLattice(Lattice l, Boolean withWeights, String filename)
{
  int i;
  List m;
  FILE *file;
  Vector vectorOfNumbers;
  int heightover;          /* minimal height of arcs between nonconsecutive nodes */
  int number1, number2;    /* the two numbers of the current arc                  */
  int numberExtraArcs;     /* number of extra arcs                                */
  int numberTotalArcs;     /* total number of arcs                                */
  int index1, index2;      /* indexes of the two nodes                            */
  int xcoord1, xcoord2;    /* x-coordinates of the two nodes                      */

  if ((file = fopen(filename, "w")) == NULL)
    {
      cdgPrintf(CDG_ERROR, "writeLattice: ERROR: Cannot write file '%s'\n", filename);
    }
  else
    {
      cdgPrintf(CDG_INFO, "INFO: writing wordgraph '%s' to file '%s'\n", l->id, filename);

      /* build the vector of node labels */
      writeBuildVector(l->arcs, &vectorOfNumbers);

      /* write the header of XFig file */
      writeXFigHeader(file);

      /* write the nodes of the word graph */
      for (i=0; i<vectorSize(vectorOfNumbers); i++)
	writeXFigCircle(file, (i+1) * write_distance, write_ycoord, write_radius);

      /* for each word arc do ... */
      m = l->arcs;
      while (m != NULL)
	{
	  /* compute the necessary information */
	  number1 = ((Arc) m->item)->from;
	  number2 = ((Arc) m->item)->to;
	  index1 = vectorIndexOf(vectorOfNumbers, (void *)number1, 0);
	  index2 = vectorIndexOf(vectorOfNumbers, (void *)number2, 0);
	  if (index1 > index2) writeSwapIntegers(&index1, &index2);
	  /* want to have index1 < index2 */
	  xcoord1 = write_distance * (index1 + 1);
	  xcoord2 = write_distance * (index2 + 1);
	  numberExtraArcs = writeNumberArcs(m->next, number1, number2);
	  numberTotalArcs = writeNumberArcs(l->arcs, number1, number2);
  
	  if (index2 - index1 == 1)
	    {
	      /* this is an arc between consecutive nodes */

	      if ((numberTotalArcs == 1) || ((numberTotalArcs % 2 == 1) && (numberExtraArcs == 0)))
		{
		  /*
		   * consecutive nodes, linked with a single arc
		   * or linked with more arcs, and this is the last
                   */
		  writeWordLine(file, (Arc) m->item, xcoord1, xcoord2, withWeights);
		}
	      else if (numberExtraArcs % 2 == 1)
		{
		  /*
		   * consecutive nodes linked between more than one arc
		   * there are an odd number of arcs left
		   * writing an arc above the line
		   */
		  writeXFigWordArc(file, (Arc) m->item, xcoord1, xcoord2, 0, -1, numberExtraArcs, withWeights);
		}
	      else
		{
		  /*
		   * consecutive nodes linked between more than one arc
		   * there are an even number of arcs left
		   * writing an arc below the line
		   */
		  writeXFigWordArc(file, (Arc) m->item, xcoord1, xcoord2, 0, 1, numberExtraArcs, withWeights);
		}
	    }
	  else
	    {
	      /* this is an arc between two non-consecutive nodes */

	      /* computing heightover */
	      heightover = writeComputeHeightOver(l->arcs, index1, index2, vectorOfNumbers) * write_height;
	      if (numberExtraArcs % 2 == 1)
		{
		  /*
		   * there are an odd number of arcs left
		   * writing an arc above the line
		   */
		  writeXFigWordArc(file, (Arc) m->item, xcoord1, xcoord2, heightover, -1, numberExtraArcs, withWeights);
		}
	      else
		{
		  /*
		   * there are an even number of arcs left
		   * writing an arc below the line
		   */
		  writeXFigWordArc(file, (Arc) m->item, xcoord1, xcoord2, heightover, 1, numberExtraArcs, withWeights);
		}
	    }
	  m = m->next;
	}
      /* writes the old output as a comment */
      writeWordOldOutput(file, l->arcs);

      fclose(file);
      vectorDelete(vectorOfNumbers);
    }
}                                                                              

/*
 * returns the index of a string in a list of strings
 */
int writeListIndexOf(List list, String string)
{
  int i; 
  int j = -1;
  Boolean cont = TRUE;

  for (i=1; i<=listSize(list) && cont; i++)
    if (strcmp((String) listNthElement(list, i), string) == 0)
      {
	cont = FALSE;
	j = i;
      }
  return j - 1;
}

/*
 * Builds the height of each word
 */
void writeBuildHeight(Parse parse, Vector *heights, int level)
{
  int i, size, h, n, nrlevels;

  size = vectorSize(parse->words);
  nrlevels = listSize(parse->levels);
  *heights = vectorNew(size);

  for (i=0; i<size; i++)
    {
      h = 0;
      n = i;
      while (n != -1)
	{
          n = (int)vectorElement(parse->verticesStructure, n * nrlevels + level);
          h++;
	} 
      vectorAddElement(*heights, (void *)(h-1));
    }
}

void writeBuildLevels(Parse parse, Vector *usedlevels, int main_level)
{
  int i; 
  int j = 0;
  int nrlevels = listSize(parse->levels);

  *usedlevels = vectorNew(nrlevels);
  vectorSetAllElements(*usedlevels, (void *)0);
  for (i=0; i<vectorSize(parse->verticesStructure); i++)
    {
      int n2 = (int)vectorElement(parse->verticesStructure, i);
      int levelnr = i % nrlevels;    
      if (levelnr != main_level && n2 != -1)
        vectorSetElement(*usedlevels, (void *)1, levelnr);
    }
  for (i=0; i<nrlevels; i++)
    if ((int)vectorElement(*usedlevels, i) == 1)
      {
	j++;
	vectorSetElement(*usedlevels, (void *)j, i);
      }
}

/* ----------------------------------------------------------------------------
 * writes a parse to an XFig file and a LaTeX file
 */
void writeParse(Parse parse, String main_level,
                String graph_filename, String info_filename)
{
  int i, hmax, nrnodes, levelvalue, nrlevels;
  Vector heights, usedlevels;
  List n;
  FILE *g_file, *i_file;

  if (!listContains(parse->levels, main_level))
    return;

  levelvalue = writeListIndexOf(parse->levels, main_level);

  if ((g_file = fopen(graph_filename, "w")) == NULL) {
    cdgPrintf(CDG_ERROR,
	      "writeParse: ERROR: Cannot write file '%s'\n", graph_filename);
  } else if ((i_file = fopen(info_filename, "w")) == NULL) {
    cdgPrintf(CDG_ERROR,
	      "writeParse: ERROR: Cannot write file '%s'\n", info_filename);
  } else {
    cdgPrintf(CDG_INFO,
	      "INFO: writing parse to files '%s' and '%s'\n",
	      graph_filename, info_filename);

    /* builds the height of each word */
    writeBuildHeight(parse, &heights, levelvalue);

    /* nrnodes = total number of nodes */
    nrnodes = vectorSize(heights);
    /* nrlevels = total number of levels */
    nrlevels = listSize(parse->levels);

    /* hmax = height of the tree */
    hmax = (int)vectorElement(heights, 0);
    for (i = 1; i < nrnodes; i++)
      if (hmax < (int)vectorElement(heights, i))
	hmax = (int)vectorElement(heights, i);

    /* writes XFig file header */
    writeXFigHeader(g_file);

    for (i = 0; i < vectorSize(parse->words); i++) {
      int x1 = (i + 1) * write_distance;
      int y1 = ((int)vectorElement(heights, i) + 1) * write_height * 2;
      int node =
	(int)vectorElement(parse->verticesStructure,
			     i * nrlevels + levelvalue);

      /* writes tree word */
      writeXFigText(g_file, write_distance, x1,
		    (hmax + 1) * write_height * 2 + write_height,
		    ((Word) vectorElement(parse->words, i))->word, 1);

      /* writes tree node */
      writeXFigSolidCircle(g_file, x1, y1, write_radius);

      if (node > -1) {
	int x2 = (node + 1) * write_distance;
	int y2 = ((int)vectorElement(heights, node) + 1) * write_height * 2;
	String label =
	  (String) vectorElement(parse->verticesLabels,
				   i * nrlevels + levelvalue);

	/* writes tree line with i as modifier */
	writeXFigLine(g_file, x1, y1, x2, y2);

	/* writes label of the previous tree line */
	writeXFigText(g_file, write_distance,
		      x1 + write_radius * ((x1 < x2) ? -1 : 1),
		      y1 + write_radius,
		      (strcmp(label, "") == 0) ? "*" : label,
		      (x1 < x2) ? 2 : 0);
      }
    }

    /* collect information on used levels */
    writeBuildLevels(parse, &usedlevels, levelvalue);

    /* writes arcs from other levels */
    for (i = 0; i < vectorSize(parse->verticesStructure); i++) {
      int node1 = i / nrlevels;
      int node2 = (int)vectorElement(parse->verticesStructure, i);
      int levelnr = i % nrlevels;
      String level = listNthElement(parse->levels, levelnr + 1);
      String label = vectorElement(parse->verticesLabels, i);

      if (levelnr != levelvalue && node2 != -1) {
	String s;

	if (node1 < node2) {
	  if (strcmp(label, "") == 0)
	    s = strAppend(level, "->", NULL);
	  else
	    s = strAppend(level, "(", label, ")->", NULL);
	} else {
	  writeSwapIntegers(&node1, &node2);
	  if (strcmp(label, "") == 0)
	    s = strAppend("<-", level, NULL);
	  else
	    s = strAppend("<-", level, "(", label, ")", NULL);
	}
	writeXFigSpline(g_file,
			(node1 + 1) * write_distance,
			(hmax + 2) * write_height * 2,
			(node2 + 1) * write_distance,
			(hmax + 2) * write_height * 2, 0,
			write_height * (int)vectorElement(usedlevels,
							    levelnr));
	writeXFigText(g_file, write_distance,
		      (node1 + node2 + 2) * write_distance / 2,
		      (hmax + 2) * write_height * 2 +
		      (int)vectorElement(usedlevels,
					   levelnr) * write_height - 50, s, 1);
	cdgFreeString(s);
      }
    }

    /* writes parse information */
    fprintf(i_file, "\\begin{longtable}{ll} \\hline \n");
    fprintf(i_file, "Wordgraph & %s \\\\ \n", parse->latticeId);
    fprintf(i_file, "User & %s \\\\ \n", parse->userName);
    fprintf(i_file, "Date & %s \\\\ \n", ctime(&(parse->dateOfCreation)));
    fprintf(i_file, "Search Strategy & %s \\\\ \n", parse->searchStrategy);
    fprintf(i_file, "Solution No. & %d-of-%d\\\\ \n", parse->solutionNo,
	    parse->noOfSolutions);
    fprintf(i_file, "Score & %f\\\\ \n", parse->score);
    fprintf(i_file, "Optimal & %s \\\\ \n", (parse->optimal ? "Yes" : "No"));
    fprintf(i_file, "\\hline \\end{longtable}\n");
    fprintf(i_file, "\n");

    if (parse->violations) {
      /* writes constraint violations */
      fprintf(i_file, "\\begin{longtable}{lllll} \n");
      fprintf(i_file,
	      "\\multicolumn{5}{c}{Constraint Violations} \\\\ \\hline \n");
      for (n = parse->violations; n; n = n->next) {
	ConstraintViolation cv = (ConstraintViolation) n->item;

	if (cv->lv2 == NULL)
	  fprintf(i_file,
		  "%s & (%f) & : & %d & \\\\ \n",
		  cv->constraint->id, cv->penalty, cv->nodeBindingIndex1);
	else
	  fprintf(i_file,
		  "%s & (%f) & : & %d & %d \\\\ \n",
		  cv->constraint->id, cv->penalty,
		  cv->nodeBindingIndex1, cv->nodeBindingIndex2);
      }
      fprintf(i_file, "\\hline \\end{longtable}\n");
    }
    /* closes the files and frees the memory */
    fclose(g_file);
    fclose(i_file);
    vectorDelete(heights);
    vectorDelete(usedlevels);
  }
}

/* ---------------------------------------------------------------------------
 * writes all parses to different XFig and LaTeX files
 * the filenames will be of the form file-1.fig, file-2.fig, ...; file-1.tex, file-2.tex, ...
 */
void writeParses(List parses, String main_level, String graph_filename,
		 String info_filename)
{
  List n;
  String f = NULL, g = NULL;

  for (n = parses; n; n = n->next) {
    Parse p = (Parse) n->item;

    f = strPrintf("%s-%d.fig", graph_filename, p->solutionNo);
    g = strPrintf("%s-%d.tex", info_filename, p->solutionNo);
    writeParse(p, main_level, f, g);
  }

  cdgFreeString(f);
  cdgFreeString(g);
}

/* -------------------------------------------------------------------------
 * -- XML section ----------------------------------------------------------
 * ------------------------------------------------------------------------- */

/* -------------------------------------------------------------------------
 * trimXmlBuffer
 *
 * returns the trimmed xml buffer
 */
String trimXmlBuffer()
{
  int i;

  /* trim white spaces at the end */
  for (i = strlen(xmlBuffer)-1; i >=0; i--) {
    if(isspace(((unsigned char *)xmlBuffer)[i])) {
      xmlBuffer[i] = 0;
    } else {
      break;
    }
  }

  /* trim white spaces at the start */
  for (i = 0; i < strlen(xmlBuffer); i++) {
    if(!isspace(((unsigned char *)xmlBuffer)[i])) {
      break;
    }
  }

  return xmlBuffer+i;
}

/* -------------------------------------------------------------------------
 * convAttrStr
 *
 * escapes unicode characters in an xml attribute string
 * returns a pointer to the formated string, caution: this is a buffer to
 * static private memory
 *
 * TODO: this is data driven
 */
String convAttrStr(String attribute)
{
  int n, i, j;
  String str;
  int strLen;

  if (attribute == NULL)
    return NULL;

  if (strlen(attribute) > xmlBufferSize) {
    xmlBuffer = memRealloc(xmlBuffer, strlen(attribute));
  }
 
  /* convert the attribute */
  xmlBuffer[0] = 0;
  for (n = i = 0; i < strlen(attribute); i++) {
    switch (attribute[i]) {
      case '&': str = "&#x26;"; break;
      case '"': str = "&#x22;"; break; 
      case '<': str = "&#x3c;"; break;
      case '>': str = "&#x3e;"; break;
      default: str = NULL; break;
    }
    strLen = (str) ? strlen(str) : 1;

    /* grant enuf memory */
    while (n + strLen >= xmlBufferSize) {
      xmlBufferSize = (n + strLen) * 2;
      xmlBuffer = memRealloc(xmlBuffer, xmlBufferSize);
    }

    if (str) {
      for (j = 0; j < strLen; j++, n++) {
	xmlBuffer[n] = str[j];
      }
    } else {
      xmlBuffer[n] = attribute[i];
      n++;
    }
    xmlBuffer[n] = 0;
  }

  return trimXmlBuffer();
}

/* -------------------------------------------------------------------------
 * called from registered string variables of the write module
 */
void writeCallback(String name, String *var) 
{
  if (strRegister("xml") == name) {
    if (strRegister("on") == xmlFileName) {
      xmlFileName = strRegister("stdout");
    }
    if (strRegister("off") == xmlFileName) {
      xmlFileName = strRegister("stdout");
      writeXmlFileSet(xmlFileName);
      hkVerbosity &= ~CDG_XML;
    } else {
      writeXmlFileSet(xmlFileName);
      hkVerbosity |= CDG_XML;
      writeXmlHeader();
    }
  }
}


/* -------------------------------------------------------------------------
 * writeXmlFileSet
 *
 * set the xml output file; the variables xmlFile and xmlFileName are changed
 * accordingly; previous files are closed and guaranteed to be wellformed
 *
 * arguments:
 * - fileName	= the new file which is to be opened
 *		  if it doesn't exist the program aborts
 * 
 * returns:
 * the xmlFile handle
 */
FILE *writeXmlFileSet(String fileName) 
{
  /* finish all open xml tags */
  writeXmlEndAll();

  /* close the previous file */
  if (xmlFile != stdout && xmlFile != stderr && xmlFile != NULL) {
    fclose(xmlFile);
    cdgFreeString(xmlFileName);
  }
  
  /* open a new file*/
  xmlFileName = strRegister(fileName);
  if (xmlFileName == strRegister("stdout")) {
    xmlFile = stdout;
  } else if (xmlFileName == strRegister("stderr")) {
    xmlFile = stderr;
  } else {
    xmlFile = fopen(xmlFileName, "w");
    if (!xmlFile) {
      cdgPrintf(CDG_ERROR, "ERROR: cannot open file `%s' for writing\n",
		xmlFileName);
      abort();
    }
  }

  return xmlFile;
}

/* -------------------------------------------------------------------------
 * writeXmlFileGet
 * 
 * returns the xmlFile handle
 */
FILE *writeXmlFileGet(void)
{
  return xmlFile;
}

/* -------------------------------------------------------------------------
 * writeXmlFlush
 *
 * flushes the xmlFile
 */
void writeXmlFlush(void)
{
  if (xmlFile != NULL) {
    fflush(xmlFile);
  }
}

/* -------------------------------------------------------------------------
 * writeXmlIndent
 *
 * write the correct ammount of indentation to the xml file output stream;
 * correct indentation is calulated in writeXmlStartTag and writeXmlEndTag
 */
void writeXmlIndent()
{
  int i;

  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  for (i = xmlDepth; i > 0; i--) {
    fprintf(xmlFile, "  "); 
  }
}


/* -------------------------------------------------------------------------
 * writeXmlProcessingInstruction
 *
 * write an xml processing instruction, only needed in the header hear
 *
 * example:
 *	writeXmlProcessingInstruction("xml", "version=\"1.0\" encoding=\"iso-8859-1\"")
 *	produces
 *	<?xml version="1.0" encoding="iso-8859-1" ?>
 */
void writeXmlProcessingInstruction(String target, String format, ...)
{
  va_list ap;

  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  writeXmlIndent();
  fprintf(xmlFile, "<?%s ", target);
  if (format != NULL) {
    va_start(ap, format);
    vfprintf(xmlFile, format, ap);
    va_end(ap);
  }
  fprintf(xmlFile, " ?>\n");
}

/* -------------------------------------------------------------------------
 * writeXmlStartTag
 *
 * write an xml start tag; the name of this opend tag is pushed on a stack
 * and poped when writeXmlEndTag is encountered accordingly
 *
 * arguments:
 * - name	= name of the tag
 * - format	= the format of an attribute list, e.g. "type=\"%d\""
 * ...		= the values of the attributes
 *
 * example: 
 *	writeXmlStartTag("person", "age=\"%d\" name=\"%s\"", 12, cosima)
 *	produces 
 *	<person age="12" name="cosima">
 *	and a person tag on the xml tag stack
 */
void writeXmlStartTag(String name, String format, ...)
{
  va_list ap;
    
  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  /* indent */
  writeXmlIndent();
  fprintf(xmlFile, "<%s", name);
  
  /* write an attribute list */
  if (format != NULL) {
    fprintf(xmlFile, " ");
    va_start(ap, format);
    vfprintf(xmlFile, format, ap);
    va_end(ap);
  }

  fprintf(xmlFile, ">\n");

  /* register this node as open */
  xmlTagStack = listPrependElement(xmlTagStack, strRegister(name));

  /* increase indentation level */
  xmlDepth++;
}

/* -------------------------------------------------------------------------
 * writeXmlEmptyTag
 *
 * write an xml empty tag; see writeXmlStartTag
 *
 * example:
 *	writeXmlEmptyTag("person", "age=\"%d\" name=\"%s\"", 12, cosima)
 *	produces 
 *	<person age="12" name="cosima"/>
 */
void writeXmlEmptyTag(String name, String format, ...)
{
  va_list ap;
    
  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  writeXmlIndent();
  fprintf(xmlFile, "<%s", name);
  
  /* write an attribute list */
  if (format != NULL) {
    fprintf(xmlFile, " ");
    va_start(ap, format);
    vfprintf(xmlFile, format, ap);
    va_end(ap);
  }

  fprintf(xmlFile, "/>\n");
}
 
/* -------------------------------------------------------------------------
 * writeXmlEndTag
 *
 * pop the last xml tag from the xml tag stack and write an xml end tag
 * for it
 *
 * example:
 *	</person>
 */
void writeXmlEndTag(void) 
{
  String name;
  List l;

  /* check wellformedness */
  if (xmlTagStack == NULL) {
    return;
  }

  /* pop last tag  from stack */
  name = (String) (xmlTagStack->item);
  l = xmlTagStack;
  xmlTagStack = xmlTagStack->next;
  cdgFreeString(name);
  memFree(l); /* freeListCell */
  
  
  /* decrease indentation level */
  xmlDepth--;

  if (hkVerbosity & CDG_XML) {
    writeXmlIndent();
    fprintf(xmlFile, "</%s>\n", name);
  }
}

/* -------------------------------------------------------------------------
 * writeXmlEndAll
 *
 * pop all xml tags from the xml tag stack and write xml end tags
 */
void writeXmlEndAll(void)
{
  while(xmlTagStack) {
    writeXmlEndTag();
  }
}

/* -------------------------------------------------------------------------
 * writeXmlHeader
 *
 * write a propper header for the cdgp xml logfile format
 */
void writeXmlHeader(void)
{
  time_t timeToday;

  if (!(hkVerbosity & CDG_XML)) {
    return;
  }
  
  /* XML Declaration */
  writeXmlProcessingInstruction("xml", "version=\"1.0\" encoding=\"UTF-8\"");
  
  /* the document type giving an external reference (could be internal) */
  fprintf(xmlFile, "<!DOCTYPE cdgp SYSTEM \"cdgp-logfile.dtd\">\n");
  writeXmlStartTag("cdgp", NULL);
  writeXmlStartTag("header", NULL);
  writeXmlUser(cdgUser, NULL);
  time(&timeToday);
  writeXmlDate(timeToday);
  writeXmlFile(xmlFileName, -1);
  writeXmlEndTag(); /* header */
}

/* -------------------------------------------------------------------------
 * writeXmlUser
 *
 * write an xml user
 *
 * arguments:
 * - name	= the full name of the user
 * - mail	= optionaly his mail adress
 */
void writeXmlUser(String name, String mail)
{
  if (mail != NULL) {
    String thisName = strRegister(convAttrStr(name));
    writeXmlEmptyTag("user", "name=\"%s\" mail=\"%s\"",
	thisName, convAttrStr(mail));
  } else {
    writeXmlEmptyTag("user", "name=\"%s\"", convAttrStr(name));
  }
}

/* -------------------------------------------------------------------------
 * writeXmlDate
 *
 * write the given time in xml format
 */
void writeXmlDate(time_t time)
{
  static String buffer;

  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  buffer = strRegister(ctime(&time));
  buffer[strlen(buffer)-1] = 0;
  writeXmlEmptyTag("date", "value=\"%s\"", buffer);

  cdgFreeString(buffer);
}

/* -------------------------------------------------------------------------
 * writeXmlNumber
 *
 * write a Number representation to the xml output stream
 *
 * arguments:
 * - name	= the name of the Number object
 * - value	= its value
 */
void _writeXmlNumber(String name, long double value)
{
  writeXmlEmptyTag("number", "name=\"%s\" value=\"%.10Lg\"", 
      convAttrStr(name), value);
}

/* -------------------------------------------------------------------------
 * writeXmlBoolean
 *
 * write a Boolean representation to the xml output stream
 *
 * arguments:
 * - name	= the name of the Boolean object
 * - value	= its value
 */
void writeXmlBoolean(String name, Boolean value)
{
  writeXmlEmptyTag("boolean", "name=\"%s\" value=\"%s\"", 
      convAttrStr(name), value?"true":"false");
}

/* -------------------------------------------------------------------------
 * writeXmlReference
 *
 * write the xml representation of an object reference
 *
 * arguments:
 * - name	= the name of the object, e.g. "n001k000-1"
 * - type	= the string representation of the type of the object, e.g. "lattice"
 * - pointer	= optionaly a pointer to that object
 */
void writeXmlReference(String name, String type, Pointer object)
{
  String thisName = strRegister(convAttrStr(name));
  String thisType = strRegister(convAttrStr(type));

  if (object != NULL) {
    writeXmlEmptyTag("ref", "name=\"%s\" type=\"%s\" pointer=\"%p\"",
	thisName, thisType, object);
  } else {
    writeXmlEmptyTag("ref", "name=\"%s\" type=\"%s\"", thisName, thisType);
  }
}

  
/* -------------------------------------------------------------------------
 * writeXmlStatistics
 *
 * write an xml statistics, a set of numbers
 *
 * arguments:
 * - type	= the numerical type of the statistics, e.g. 1,2,3,4
 * - method	= the method which generates this statistics, e.g. netsearch, gls, frobbing
 * - description= a plain text describtion of the statistics
 * - ...	= Numbers, terminated with a NULL
 *
 * example:
 *     writeXmlStatistics(1, "test", "this is a test statistics",
 *	  "score", 1.0, 
 *	  "totalTime", (Number) 500,
 *	  "softTime", (Number) 300,
 *	  "solutionTime", (Number) 400,
 *	  "steps", (Number) 1234,
 *	  NULL);
 *
 * produces
 *
 *	<statistics type="1" desc="this is a test statistics">
 *	   <number name="score" value="1.0"/>
 *	   <number name="totalTime" value="500"/>
 *	   <number name="softTime" value="300"/>
 *	   <number name="solutionTime" value="400"/>
 *	   <number name="steps" value="1234"/>
 *	</statistics>
 *
 */
void writeXmlStatistics(int type, String method, String description, ...) 
{
  va_list ap;
  String name;
  Number value;
  
  writeXmlStartTag("statistics", "type=\"%d\" desc=\"%s\"", 
      type, convAttrStr(description));
  writeXmlMethod(method);

  va_start(ap, description);
  while((name = va_arg(ap, String)) != NULL) {
    value = va_arg(ap, Number);
    writeXmlNumber(name, value);
  }
  va_end(ap);

  writeXmlEndTag();
}

/* -------------------------------------------------------------------------
 * writeXmlNetInfo
 *
 * write ConstraintNet information to the xml output stream
 *
 * arguments:
 * - net	= a constraint net
 */
void writeXmlNetInfo(ConstraintNet net)
{
  int i, vmin, vmax, vtotal;
  ConstraintNode node;
  long double candidates;

  if (!(hkVerbosity & CDG_XML)) 
    return;

  vmin = 99999999;
  vtotal = vmax = 0;
  candidates = 0;
  for (i = 0; i < vectorSize(net->nodes); i++) {
    node = (ConstraintNode) vectorElement(net->nodes, i);
    vmin = min(vmin, node->noValidValues);
    vmax = max(vmax, node->noValidValues);
    vtotal += node->noValidValues;
    if (node->noValidValues) {
      if (!candidates) {
	candidates = node->noValidValues;
      } else {
	candidates *= node->noValidValues;
      }
    }
  }

  writeXmlStartTag("netInfo", "id=\"%s\"", convAttrStr(net->id));
  writeXmlReference(net->lexemgraph->lattice->id, "Lattice", net->lexemgraph->lattice);
  writeXmlNumber("noNodes", vectorSize(net->nodes));
  writeXmlNumber("noEdges",  (net->edges == NULL ? 0 : vectorSize(net->edges)));
  writeXmlNumber("noUnaryEvaluations", net->evalUnary);
  writeXmlNumber("noUnaryStatistics", net->statUnary);
  writeXmlNumber("noBinaryEvaluations", net->evalBinary);
  writeXmlNumber("minValues", vmin);
  writeXmlNumber("maxValues", vmax);
  writeXmlNumber("noValues", vtotal);
  writeXmlNumber("avgValuesPerNode", vtotal / (Number)vectorSize(net->nodes));
  writeXmlNumber("candidates", candidates);
  writeXmlEndTag();
}

/* -------------------------------------------------------------------------
 * writeXmlArc
 *
 * write a Arc representation to the xml output stream
 *
 * arguments:
 * - arc	= an arc
 */
void writeXmlArc(Arc arc)
{
  if (arc->penalty > 0) {
    writeXmlEmptyTag("arc", "from=\"%d\" to=\"%d\" word=\"%s\" penalty=\"%f\"",
	arc->from, arc->to, convAttrStr(arc->word), arc->penalty);
  } else {
    writeXmlEmptyTag("arc", "from=\"%d\" to=\"%d\" word=\"%s\"", 
	arc->from, arc->to, convAttrStr(arc->word));
  }
}

/* -------------------------------------------------------------------------
 * writeXmlLattice
 *
 * write a Lattice representation to the xml output stream
 *
 * arguments:
 * - lattice	= a lattice
 */
void writeXmlLattice(Lattice lattice)
{
  Arc arc;
  List l;

  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  writeXmlStartTag("lattice", "id=\"%s\"", convAttrStr(lattice->id));
  writeXmlFile(lattice->filename, lattice->lineNo);
  writeXmlBoolean("trivial", !latticeBranches(lattice));

  for (l = lattice->arcs; l != NULL; l = l->next) {
    arc = (Arc)(l->item);
    writeXmlArc(arc);
  }

  writeXmlEndTag();
}

/* -------------------------------------------------------------------------
 * writeXmlNotification
 *
 * write an xml notification
 *
 * arguments:
 * - type	= severity of notification, e.g. "warning", "error"
 * - format,... = the description of the notification
 */
void writeXmlNotification(String type, String format, ...)
{
  va_list ap;
  
  va_start(ap, format);
  writeXmlVaNotification(type, format, ap);
  va_end(ap);
}

/* -------------------------------------------------------------------------
 * writeXmlVaNotification
 *
 * write an xml notification using a variable list 
 * see writeXmlNotification
 */
void writeXmlVaNotification(String type, String format, va_list ap)
{
  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  if (format != NULL) {
    String buffer = strCopy(strVPrintf(format, ap));
    String bufptr = buffer;
    String token;

    for (token = strtok(bufptr, "\n"); token; token = strtok(NULL, "\n")) {
      writeXmlEmptyTag("notify", "type=\"%s\" desc=\"%s\"", 
	  type, convAttrStr(token));
    }
    
    memFree(buffer);
  } else {
    writeXmlEmptyTag("notify", "type=\"%s\"", type);
  }
}

/* -------------------------------------------------------------------------
 * writeXmlCommand
 *
 * write an xml command tag
 *
 * arguments:
 * - commandLine	= the command line
 */
void writeXmlCommand(String commandLine)
{
  static int xmlCommandNo = 0;

  if (!(hkVerbosity & CDG_XML))
    return;

  writeXmlStartTag("command",  "id=\"command%d\" desc=\"%s\"", 
      xmlCommandNo++, convAttrStr(commandLine));

}

/* -------------------------------------------------------------------------
 * writeXmlFile
 *
 * write information about some file to the xml output stream
 *
 * arguments:
 * - fileName	= some filename
 * - lineNo	= a line number; lineNos < 0 are omitted
 */
void writeXmlFile(String name, int lineNo)
{
  if (lineNo < 0) {
    writeXmlEmptyTag("file", "name=\"%s\"", convAttrStr(name));
  } else {
    writeXmlEmptyTag("file", "name=\"%s\" lineNo=\"%d\"", 
	convAttrStr(name), lineNo);
  }
}

/* -------------------------------------------------------------------------
 * writeXmlGrammar
 *
 * write xml information about a used grammar
 *
 * arguments:
 * - files	= list of file names
 */
void writeXmlGrammar(List files)
{
  List l;
  String fileName;

  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  writeXmlStartTag("grammar", NULL);
  for (l = files; l != NULL; l = l->next) {
    fileName = (String) (l->item);
    writeXmlFile(fileName, -1);
  }
  writeXmlEndTag(); /* grammar */
}

/* -------------------------------------------------------------------------
 * writeXmlMethod
 *
 * write a method tag 
 *
 * arguments:
 * - method	= the name of the method, e.g. "netsearch", "gls", "frobbing"
 *
 * example output:
 *	<method type="netsearch"/>
 */
void writeXmlMethod (String method) 
{
  writeXmlEmptyTag("method", "type=\"%s\"", convAttrStr(method));
}

/* -------------------------------------------------------------------------
 * writeXmlParse
 *
 * write a Parse representation to xml
 *
 * arguments:
 * - parse	= a parse
 */
void writeXmlParse(Parse parse)
{
  int noLevels;
  List l;
  String constraintId;
  Number constraintPenalty;
  int constraintNode1;
  int constraintNode2;

  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  noLevels = listSize(parse->levels);

  /* write parse info header */
  writeXmlStartTag("parse", "id=\"%s\"", convAttrStr(parse->id));

  writeXmlReference(parse->latticeId, "Lattice", NULL);
  writeXmlUser(parse->userName, NULL);
  writeXmlDate(parse->dateOfCreation);
  writeXmlGrammar(parse->grammarFiles);
  writeXmlMethod(parse->searchStrategy);
  if (parse->comment != NULL) {
    writeXmlStartTag("comment", NULL);
    fprintf(xmlFile, parse->comment);
    writeXmlEndTag();
  }
  writeXmlNumber("solutionNo", parse->solutionNo);
  writeXmlNumber("maxSolutions", parse->noOfSolutions);
  writeXmlNumber("score", parse->score);
  writeXmlBoolean("optimal", parse->optimal);

/* 2002-07-11, Michael Daum: do we realy need these properties in the parse ? */
#if 0
  writeXmlNumber("evalUnary", parse->evalUnary);
  writeXmlNumber("evalBinary", parse->evalBinary);
  writeXmlNumber("noHits", parse->numberOfHits);
  writeXmlNumber("evalTime", parse->evaluationTime);
#endif
  if (parse->statisticComment != NULL) {
    writeXmlStartTag("comment", NULL);
    fprintf(xmlFile, parse->statisticComment);
    writeXmlEndTag();
  }

  /* 2002-07-11, Michael Daum: commented out xml bindings node; most of the
   * time this information is of no interest use the xml annotations for that */

#if 0
  /* write bindings */
  writeXmlStartTag("bindings", NULL);
  for (i = 0; i < vectorSize(parse->verticesStructure); i++) {
    label = (String) vectorElement(parse->verticesLabels, i);
    level = listNthElement(parse->levels, (i % noLevels) + 1);
    modifierWord = vectorElement(parse->words, i / noLevels);

    j = (int)vectorElement(parse->verticesStructure, i);
    if (j > -1) {
      modifieeWord = vectorElement(parse->words, j);
    } else {
      modifieeWord = (Word) j;	       /* -1: NIL -2: NONSPEC */
    }

    /* skip uninteresting bindings */
    if (modifieeWord == (Word) - 1 && strlen(label) == 0)
      continue;

    /* write a binding */
    writeXmlStartTag("binding", "id=\"binding%d\"", i);
    writeXmlReference(level, "Level", level);
    writeXmlWord(modifierWord);
    if (label != NULL && strlen(label) > 0) {
      writeXmlReference(label, "Label", NULL);
    }
    writeXmlWord(modifieeWord);
    writeXmlEndTag();		       /* binding */
  }
  writeXmlEndTag();		       /* bindings */
#endif

  /* write violated constraints */
  if (parse->violations) {
    /* writes constraint violations */
    writeXmlStartTag("violations", NULL);

    for (l = parse->violations; l; l = l->next) {
      ConstraintViolation conv = (ConstraintViolation) l->item;

      constraintId = (String) conv->constraint->id;
      constraintPenalty = (Number) (conv->penalty);
      constraintNode1 = (int)(conv->nodeBindingIndex1);
      if (conv->lv2 == NULL) {
	writeXmlEmptyTag("constraint", "id=\"%s\" penalty=\"%f\" node=\"%d\"",
			 convAttrStr(constraintId), constraintPenalty,
			 constraintNode1);
      } else {
	constraintNode2 = (int)(conv->nodeBindingIndex2);
	writeXmlEmptyTag("constraint",
			 "id=\"%s\" penalty=\"%f\" node=\"%d\" node2=\"%d\"",
			 convAttrStr(constraintId), constraintPenalty,
			 constraintNode1, constraintNode2);
      }
    }
    writeXmlEndTag();		       /* violations */
  }

  writeXmlEndTag();		       /* parse */
}

/* -------------------------------------------------------------------------
 * writeXmlWord
 *
 * write a Word representation to xml; this actualy produces a
 * lexem tag from a word
 *
 * arguments:
 * - word	= a word
 *
 * example output:
 *	<lexem id="dann">
 *	  <arc from="0" to="1" word="dann"/>
 *      </lexem>
 *      <lexem id="NIL"/>
 *      <lexem id="NONSPEC"/>
 */
void writeXmlWord(Word word)
{
  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  if ((int)word > -1) {
    writeXmlStartTag("lexem", "id=\"%s\"", convAttrStr(word->description));
    writeXmlEmptyTag("arc", "from=\"%d\" to=\"%d\" word=\"%s\"",
      word->from, word->to, convAttrStr(word->word));
    writeXmlEndTag(); /*  lexem */
  } else if ((int)word == -1) {
    writeXmlEmptyTag("lexem", "id=\"NIL\"");
  } else {
    writeXmlEmptyTag("lexem", "id=\"NONSPEC\"");
  }

}

/* -------------------------------------------------------------------------
 * writeXmlParses
 *
 * write a set of Parse representations to xml
 * see writeXmlParse
 *
 * arguments:
 * - parses	= list of parse
 */
void writeXmlParses(List parses)
{
  List l;
  Parse parse;

  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  writeXmlStartTag("parses", NULL);
  for (l = parses; l != NULL; l = l->next) {
    parse = (Parse) (l->item);
    writeXmlParse(parse);
  }
  writeXmlEndTag();
}

/* -------------------------------------------------------------------------
 * write an annotaion to xml.
 * @see printAnnoEntry
 */
void writeXmlAnnoEntry(AnnoEntry annoEntry)
{
  List l,m;
  Annotation anno;
  Specification spec;
  int noEdges, noWords;
  String specKind, specName;

  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  /* count number of given words and edges */
  noWords = listSize(annoEntry->annos);
  noEdges = 0; 
  for (l = annoEntry->annos; l != NULL; l = l->next) {
    anno = (Annotation) l->item;
    for (m = anno->specs; m != NULL; m = m->next) {
      spec = (Specification) m->item;
      if (spec->type == STDep) {
	noEdges++;
      }
    }
  }

  writeXmlStartTag("annotation", "id=\"%s\" lattice=\"%s\" nowords=\"%d\" noedges=\"%d\"", 
      annoEntry->id, annoEntry->lattice, noWords, noEdges);
  writeXmlFile(annoEntry->filename, annoEntry->lineNo);

  /* write all arc specs */
  for (l = annoEntry->annos; l != NULL; l = l->next) {
    anno = (Annotation) l->item;
    writeXmlStartTag("arc", "from=\"%d\" to=\"%d\" word=\"%s\"",
	anno->from, anno->to, convAttrStr(anno->word));

    for (m = anno->specs; m != NULL; m = m->next) {
      spec = (Specification) m->item;
      specKind = strRegister(convAttrStr(spec->kind));
      specName = strRegister(convAttrStr(spec->name));
      
      if (spec->type == STTag) {
	writeXmlEmptyTag("tag", "%s=\"%s\"", specKind, specName);
      } else {
	writeXmlEmptyTag("dep", "level=\"%s\" label=\"%s\" modifiee=\"%d\"",
	    specKind, specName, spec->position);
      }
    }
    writeXmlEndTag(); /* arc */
  }

  writeXmlEndTag(); /* annotation */
}

/* -------------------------------------------------------------------------
 * write a parse verification to xml.
 * @see parseVerify
 *
 *  <parseverification>
 *    <ref name="parse20" type="Parse"/>
 *    <ref name="n001k008-1" type="AnnoEntry"/>
 *    <tries>
 *      <number name="OBL3" value="9"/>
 *      ...
 *    </tries>
 *    <structhits>
 *      ...
 *    </structhits>
 *    <labelledhits>
 *      ...
 *    </labelledhits>
 *    <lexicalhits>
 *      ...
 *    </lexicalhits>
 *    <stricthits>
 *      ...
 *    </stricthits>
 *  </parseverification>
 */
void writeXmlParseVerification(ParseVerification pv)
{
  HashIterator hi;
  String key;
  int value;
  List l;
  ParseError pe;

  if (!(hkVerbosity & CDG_XML)) {
    return;
  }
  writeXmlStartTag("parseverification", NULL);
  writeXmlReference(pv->parse->id, "Parse", pv->parse);
  writeXmlReference(pv->annoEntry->id, "AnnoEntry", pv->annoEntry);

  for (l = pv->errors; l; l = l->next) {
    pe = (ParseError)(l->item);
    switch(pe->type) {
      case PEWord:
	writeXmlEmptyTag("error", "type=\"word\" annoWord=\"%s\"", 
	    pe->data.word);
	break;
      case PELevel:
	writeXmlEmptyTag("error", "type=\"level\" annoLevel=\"%s\"", 
	    pe->data.level);
	break;
      case PEStruct:
	writeXmlEmptyTag("error", 
	    "type=\"struct\" level=\"%s\" "
	    "modifier=\"%d\" label=\"%s\" modifiee=\"%d\" "
	    "annoModifiee=\"%d\" annoLabel=\"%s\"",
	    pe->data.structError.level,
	    pe->data.structError.modifier,
	    pe->data.structError.label,
	    pe->data.structError.modifiee,
	    pe->data.structError.annoModifiee,
	    pe->data.structError.annoLabel);
	break;
      case PELabel:
	writeXmlEmptyTag("error", 
	    "type=\"label\" level=\"%s\" "
	    "modifier=\"%d\" label=\"%s\" modifiee=\"%d\" "
	    "annoLabel=\"%s\"",
	    pe->data.labelError.level,
	    pe->data.labelError.modifier,
	    pe->data.labelError.label,
	    pe->data.labelError.modifiee,
	    pe->data.labelError.annoLabel);
	break;
      case PELexem:
	if (pe->data.lexError.value) {
	  writeXmlEmptyTag("error", 
	      "type=\"lexem\" modifier=\"%d\" "
	      "feature=\"%s\" value=\"%s\" "
	      "annoValue=\"%s\"",
	      pe->data.lexError.modifier,
	      pe->data.lexError.feature,
	      pe->data.lexError.value,
	      pe->data.lexError.annoValue);
	} else {
	  writeXmlEmptyTag("error", 
	      "type=\"lexem\" modifier=\"%d\" "
	      "feature=\"%s\" "
	      "annoValue=\"%s\"",
	      pe->data.lexError.modifier,
	      pe->data.lexError.feature,
	      pe->data.lexError.annoValue);
	}
	break;
      default: /* never reach */
	cdgPrintf(CDG_ERROR, "ERROR: unknown parse error type\n");
	abort();
    }
  }

  /* tries */
  writeXmlStartTag("tries", NULL);
  hi = hashIteratorNew(pv->tries);
  while ((key = hashIteratorNextKey(hi)) != NULL) {
    value = (int)hashGet(pv->tries, key) -1;
    writeXmlNumber(key, value);
  }
  hashIteratorDelete(hi);
  writeXmlEndTag();

  /* struct hits */
  writeXmlStartTag("structhits", NULL);
  hi = hashIteratorNew(pv->structHits);
  while ((key = hashIteratorNextKey(hi)) != NULL) {
    value = (int)hashGet(pv->structHits, key) -1;
    writeXmlNumber(key, value);
  }
  hashIteratorDelete(hi);
  writeXmlEndTag();

  /* labelled hits */
  writeXmlStartTag("labelledhits", NULL);
  hi = hashIteratorNew(pv->labelledHits);
  while ((key = hashIteratorNextKey(hi)) != NULL) {
    value = (int)hashGet(pv->labelledHits, key) -1;
    writeXmlNumber(key, value);
  }
  hashIteratorDelete(hi);
  writeXmlEndTag();

  /* lexical */
  writeXmlStartTag("lexicalhits", NULL);
  hi = hashIteratorNew(pv->lexicalHits);
  while ((key = hashIteratorNextKey(hi)) != NULL) {
    value = (int)hashGet(pv->lexicalHits, key) -1;
    writeXmlNumber(key, value);
  }
  hashIteratorDelete(hi);
  writeXmlEndTag();

  /* strict */
  writeXmlStartTag("stricthits", NULL);
  hi = hashIteratorNew(pv->strictHits);
  while ((key = hashIteratorNextKey(hi)) != NULL) {
    value = (int)hashGet(pv->strictHits, key) -1;
    writeXmlNumber(key, value);
  }
  hashIteratorDelete(hi);
  writeXmlEndTag();

  writeXmlEndTag(); /* parseverification */
}

/* -------------------------------------------------------------------------
 * write a parse comparison to xml
 * see parseCompare
 */
void writeXmlParseComparison(Parse parse1, Parse parse2, ParseCmpStat stat)
{
  int errors;

  if (!(hkVerbosity & CDG_XML)) {
    return;
  }

  writeXmlStartTag("parsecomparison", NULL);
  writeXmlReference(parse1->id, "Parse", parse1);
  writeXmlReference(parse2->id, "Parse", parse2);

  errors = stat->errorextrawordref + stat->errorextrawordtest +
    stat->errorwordlex + stat->errorgovword + stat->errorgovlex +
    stat->erroredgelabel;

  writeXmlNumber("errorextrawordref", stat->errorextrawordref);
  writeXmlNumber("errorextrawordtest", stat->errorextrawordtest);
  writeXmlNumber("errorwordlex", stat->errorwordlex);
  writeXmlNumber("errorgovword", stat->errorgovword);
  writeXmlNumber("errorgovlex", stat->errorgovlex);
  writeXmlNumber("erroredgelabel", stat->erroredgelabel);
  writeXmlNumber("totalerrors", errors);
  writeXmlNumber("matchedgebase", stat->matchedgebase);
  writeXmlNumber("matchedgegov", stat->matchedgegov);
  writeXmlNumber("matchedgetot", stat->matchedgetot);
  writeXmlNumber("totalextrawordref", stat->totalextrawordref);
  writeXmlNumber("totalextrawordtest", stat->totalextrawordtest);
  writeXmlNumber("totalwordlex", stat->totalwordlex);
  writeXmlNumber("totalgovword", stat->totalgovword);
  writeXmlNumber("totalgovlex", stat->totalgovlex);
  writeXmlNumber("totaledgelabel", stat->totaledgelabel);
  writeXmlNumber("totaledgebase", stat->totaledgebase);
  writeXmlNumber("totaledgegov", stat->totaledgegov);
  writeXmlNumber("totaledgetot", stat->totaledgetot);
  writeXmlNumber("totaledgetest", stat->totaledgetest);
  writeXmlNumber("totaledgeref", stat->totaledgeref);
  writeXmlNumber("accuracy", stat->accuracy);
  writeXmlNumber("recall", stat->recall);
  writeXmlNumber("fmeasure", stat->fmeasure);

  writeXmlEndTag();
}

/* -------------------------------------------------------------------------
 * Writes an AnnoEntry to the file FOO.cda
 * in a format suitable for input to cdgp. 
 */
void writeAnnotation(AnnoEntry ae, String id) {

  FILE *file;
  Annotation a;
  Specification s;
  List l, m;
  String filename;
  String annoname;
  Boolean firstWord = TRUE;

  filename = strAppend(strRegister(id), strRegister(".cda"), NULL);
  file = fopen(filename, "w");
  annoname = strRegister(basename(id));

  if(!file) {
    cdgPrintf(CDG_WARNING, "WARNING: can't write to %s!\n", filename);
    return;
  }

  /* title */  
  fprintf(file, "\n%s : %s <->",
          formatIdentifier(annoname),
          formatIdentifier(ae->lattice));

  /* iterate over annos */
  for(l = ae->annos; l != NULL; l = l->next) {
    a = l->item;
    
    /* write word */
    fprintf(file, "%s\n  %d %d %s\n",
            firstWord ? "" : ",",
            a->from,
            a->to,
            formatIdentifier(a->word));
    firstWord = FALSE;

    /* write specifications */
    for(m = a->specs; m != NULL; m = m->next) {
      s = m->item;
      switch(s->type) {
      case STDep:
        fprintf(file, "%s -> %s -> %d",
                formatIdentifier(s->kind),
                formatIdentifier(s->name),
                s->position);
        if(s->position > 0) {
          fprintf(file,
                  " // %s",
                  ((Annotation)listNthElement(ae->annos, s->position))->word);
        } else if (s->position == (int)NONSPEC) {
          fprintf(file, " // NONSPEC");
	}
        fprintf(file, "\n");

        break;
      case STTag:
        /* do not record the underspecified value in any case */
        if(strcmp("bot", s->name)) {
          fprintf(file,
                  "%s / %s\n",
                  formatIdentifier(s->kind), formatIdentifier(s->name));
        }
        break;
      }
    }
  }
  fprintf(file, ";\n\n");
  fclose(file);
}

/* -------------------------------------------------------------------------
 * Prolog stuff
 */
void writePrologParse(FILE* s,Parse p)
{
  fprintf(s,"parse(%s,%s,",p->id,p->latticeId);
  write_words_as_prolog(s,p->words);
  fputc(',',s);
  write_string_list_as_prolog(s,p->levels);
  fputc(',',s);
  write_string_list_as_prolog(s,p->labels);
  fprintf(s,",%i,",vectorSize(p->verticesStructure));
  write_int_vector_as_prolog(s,p->verticesStructure);
  fprintf(s,",%i,",vectorSize(p->verticesLabels));
  write_string_vector_as_prolog(s,p->verticesLabels);
  fprintf(s,").\n");
}

void writePrologAE(FILE* s,AnnoEntry ae)
{
  fprintf(s,"annotation(%s,%s,",ae->id,ae->lattice);
  write_annos_as_prolog(s,ae->annos);
  fprintf(s,").\n");
}

void write_words_as_prolog(FILE* s,Vector v)
{
  int i = 0;
  Word w;

  fputc('[',s);

  while (i < vectorSize(v))
  {
    w = vectorElement(v,i++);
    fprintf(s,"word(%i,%i,'%s','%s')",w->from,w->to,w->word,w->description);

    if (i < vectorSize(v))
      fputc(',',s);
    else
      fputc(']',s);
  }
}

void write_annos_as_prolog(FILE* s,List p)
{
  Annotation a;

  fputc('[',s);

  while (p)
  {
    a = p->item;
    fprintf(s,"word(%i,%i,'%s',",a->from,a->to,a->word);
    write_specs_as_prolog(s,a->specs);
    fputc(')',s);
    p = p->next;

    if (p)
      fputc(',',s);
    else
      fputc(']',s);
  }
}

void write_specs_as_prolog(FILE* s,List p)
{
  Specification x;

  fputc('[',s);

  while (p)
  {
    x = p->item;
    if (x->type == STTag)
      fprintf(s,"tag('%s','%s')",x->kind,x->name);
    else if (x->type == STDep)
      fprintf(s,"dep('%s','%s',%i)",x->kind,x->name,x->position);
    else
      cdgPrintf(CDG_ERROR,"ERROR: illegal SpecType: %i\n",x->type);
    p = p->next;

    if (p)
      fputc(',',s);
    else
      fputc(']',s);
  }
}

void write_int_vector_as_prolog(FILE* s,Vector v)
{
  int i = 0;
  int n;

  fputc('[',s);

  while (i < vectorSize(v))
  {
    n = (int)vectorElement(v,i++);
    fprintf(s,"%i",n);

    if (i < vectorSize(v))
      fputc(',',s);
    else
      fputc(']',s);
  }
}

void write_string_vector_as_prolog(FILE* s,Vector v)
{
  int i = 0;
  char* p;

  fputc('[',s);

  while (i < vectorSize(v))
  {
    p = vectorElement(v,i++);
    fprintf(s,"'%s'",p);

    if (i < vectorSize(v))
      fputc(',',s);
    else
      fputc(']',s);
  }
}

void write_string_list_as_prolog(FILE* s,List p)
{
  fputc('[',s);

  while (p)
  {
    fprintf(s,"'%s'",(char*)p->item);
    p = p->next;

    if (p)
      fputc(',',s);
    else
      fputc(']',s);
  }
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
/** @} */
