/*

    The BLAH library, a container library
    Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
    Contact: blah@nats.informatik.uni-hamburg.de

    $Id: array.c,v 1.15 2005-12-01 16:20:13 micha Exp $
    
*/

/* ----------------------------------------------------------------------
 * @defgroup Array Arrays
 *
 * Implementation of an array container.
 *
 * An array is a matrix-like data structure with an arbitrary (but
 * fixed) dimension and arbitrary (but fixed) size. Items are accessed by
 * a tuple of indices in constant time. Some of the functions of this 
 * module use a variable number of arguments like
 *    - arrayNew()
 *    - arraySetElement() and
 *    - arrayElement().
 *
 * The defined dimension constructing and array with arrayNew() 
 * must match the number of indices you provide to arraySetElement()
 * and arrayElement().
 *
 * @{
 */

/* -----INCLUDES------------------------------------------------------ */
#include <stdio.h>
#include "blah.h"

/* ----------------------------------------------------------------------
 * internal structure of an array.
 */
struct ArrayStruct {
  Vector dims;                     /**< vector of dimensions */
  int totalNoOfEntries;            /**< product of dimensions */
  Pointer *entries;                /**< table of entries */
};

/* ----------------------------------------------------------------------
 * creates and returns a new vector.
 *
 * This function constructs a new array with an arbitrary number of
 * dimensions.
 *
 * @param i the first array dimension
 * @param ... optional more dimensions
 * @returns a new Array
 */
Array arrayNew(int i, ...)
{
  Array a;
  va_list ap;
  int counter;

  a = (Array) memMalloc(sizeof (ArrayStruct));
  a->dims = vectorNew(5);

  if (i <= 0) {
    printf("PANIC: illegal array dimen %d!\n", i);
    abort();
  }

  counter = i;
  vectorAddElement(a->dims, (void *)i);

  va_start(ap, i);
  while ((i = va_arg(ap, int)) != 0) {
    vectorAddElement(a->dims, (void *)i);
    counter *= i;
  }
  va_end(ap);

  a->totalNoOfEntries = counter;
  a->entries = (Pointer *) memMalloc(sizeof (Pointer) * counter);
  for (i = 0; i < counter; i++)
    a->entries[i] = NULL;

  return (a);
}

/* ----------------------------------------------------------------------
 * copy an array into a new one.
 *
 * @param a is the source for the cloning
 * @returns a new cloned array
 */
Array arrayClone(Array a)
{
  int i;
  Array new;

  new=(Array)memMalloc(sizeof(ArrayStruct));
  new->dims=vectorClone(a->dims);
  new->totalNoOfEntries=a->totalNoOfEntries;

  new->entries=(Pointer *)memMalloc(sizeof(Pointer)*new->totalNoOfEntries);
  for (i=0; i < new->totalNoOfEntries; i++)
    new->entries[i]=a->entries[i];
  
  return(new);
}

/* ----------------------------------------------------------------------
 * deletes an array. Any future access to the array is illegal.
 * Note that this function does not free the memory from the items
 * contained in the array.
 *
 * @param a the array to be deleted.
 */
void arrayDelete(Array a)
{
  if (a == NULL) {
    fprintf(stderr, "ERROR: arrayDelete: argument must not be NULL\n");
    abort();
  }

  vectorDelete(a->dims);
  memFree(a->entries);
  memFree(a);
}

/* ----------------------------------------------------------------------
 * sets an array element to a new value.
 *
 * @param a the array whose element should be set to a new value.
 * @param new the new value to which the element must be set to.
 * @param ... the index of the new element in the array. 
 * @returns the old value of the element
 */
Pointer arraySetElement(Array a, Pointer new, ...)
{
  va_list ap;
  Pointer old;
  int d, factor, index, i;

  va_start(ap, new);

  index = 0;
  factor = 1;
  for (d = 0; d < vectorSize(a->dims); d++) {
    i = va_arg(ap, int);

    index += i * factor;
    if (i >= (int)vectorElement(a->dims, d)) {
      fprintf(stderr, "ERROR: arraySetElement: index %d out of range\n", i);
      abort();
    }
    factor *= (int)vectorElement(a->dims, d);
  }
  va_end(ap);

  old = a->entries[index];
  a->entries[index] = new;
  return (old);
}

/* ----------------------------------------------------------------------
 * sets all array element to a new value
 * 
 * @param a the array whose values should be set to a new value
 * @param new the new value to which all the elements of the array have to be set to.
 */
void arraySetAllElements(Array a, Pointer new)
{
  int i;

  for (i=0; i < a->totalNoOfEntries; i++)
    a->entries[i]=new;
}

/* ----------------------------------------------------------------------
 * retrieves an array element.
 *
 * @param a the array from which an element has to be retrieved.
 * @param ... the indices that define which element has to be retrieved.
 * @return the element identified by the indices.
 */
Pointer arrayElement(Array a, ...)
{
  va_list ap;
  int d, factor, index, i, noDims=vectorSize(a->dims);

  va_start(ap, a);

  index=0;
  factor=1;
  for (d=0; d < noDims; d++) {
    i=va_arg(ap, int);
    index += i * factor;
    if (i >= (int)vectorElement(a->dims, d)) {
      fprintf(stderr,
	      "arrayElement: index %d of dimension %d out of range\n",
	      i, d);
      abort();
    }
    factor *= (int)vectorElement(a->dims, d);
  }
  va_end(ap);

  return(a->entries[index]);
}

/* ----------------------------------------------------------------------
 * returns value for dimension dim.
 *
 * @param a the array whose dimension has to be retrieved.
 * @param dim the dimension in which the size of the array has to be retrieved.
 * @returns the size of the array in the specified dimension.
 */
int arrayDimension(Array a, int dim)
{
  int noDims=vectorSize(a->dims);

  if (dim>=0 && dim<noDims)
    {
      return (int)vectorElement(a->dims, dim);
    }
  fprintf(stderr, "arrayDimension: illegal dimension %d\n", dim);
  abort();
}

/* ------------------------------------------------------------------- */
/* ------------------------------------------------------------------- */
/** @} */

