;; corpus-error-mode: An emacs mode for viewing listings of
;; constraint violations in a given corpus. Use this mode for
;; viewing the files known_errors/new_errors that ccc-check
;; creates.
;;
;; Usage:
;;
;;  [delete] deletes the current conflict, or the current 
;;           sentence record if point is on the first line.
;;
;;  C-up, C-down move through the list by sentences.
;;
;;  C-c C-c  counts instances of the conflict that the cursor is on.
;;
;;  C-c C-d  visits the constraint that the cursor is on.
;;
;;  C-c C-s  calls ccc-show to display all instances of 
;;           the conflict that the cursor is on.
;;
;;  C-c C-v  visits the .cda file and jumps to the word that
;;           the cursor is on.
;;
;; Functions to check the known corpora are predefined; just use
;; M-x diff-heiseticker-errors to compare the `known' and the
;; `current' errors for that corpus.

(require 'cl)
(require 'derived)


(define-derived-mode corpus-error-mode text-mode "Corpus Error"
  "Major mode for editing logs of known errors in corpora."
  
  (ce-highlight-constraints)
  
  (setq truncate-lines nil
        truncate-partial-width-windows nil))


(defun ce-highlight-constraints ()
  (interactive)

  ;; Highlight names of constraints below threshold
  (save-excursion
    (beginning-of-buffer)
    (let ((num 0)
          (threshold 0.1))
      (while (re-search-forward
              "^.+(\\([0-9]\\.[0-9][0-9][0-9]e[+-][0-9][0-9]\\)):" nil t)
        (when (<= (string-to-number (match-string 1)) threshold)
          (progn
            (incf num)
            (highlight-range
             (match-beginning 0) (match-end 0) 'highlight-regexp-face))))
      (message "%d match%s" num (if (= num 1) "" "es")))))


(defun ce-diff (corpus)
  (interactive "sCorpus: ")
  (cd (format "/data/corpus/ccc/%s" corpus))
  (let ((truncate-lines nil))
    (ediff-files "known_errors" "new_errors")))

(defvar ce-corpora
  (directory-files
   "/data/corpus/ccc" 
   nil "^[a-zA-Z]+$" 'sort 'dirs-only)
  "List of base names of corpora found in the CCC directory.")

(defun define-diff-function (name)
  (fset (intern (format "diff-%s-errors" name)) 
        `(lambda () (interactive) (ce-diff ,name))))

(dolist (c ce-corpora) (define-diff-function c))



(define-key corpus-error-mode-map [delete] 'ce-kill)
(define-key corpus-error-mode-map [(control c) c] 'ce-count)
(define-key corpus-error-mode-map [(control c) d] 'ce-visit-definition)
(define-key corpus-error-mode-map [(control c) s] 'ce-show)
(define-key corpus-error-mode-map [(control c) t] 'ce-this)
(define-key corpus-error-mode-map [(control c) v] 'ce-visit-anno)
(define-key corpus-error-mode-map [(control up)] 'ce-backward-sentence)
(define-key corpus-error-mode-map [(control down)] 'ce-forward-sentence)

(defun ce-backward-sentence () (interactive)
  "Move point to the previous sentence in the log."
  (re-search-backward "^Sentence "))

(defun ce-forward-sentence () (interactive)
  "Move point to the next sentence in the log."
  (let ((regex "^Sentence "))
    (when (looking-at regex) (forward-line 1))
    (while (and (not (eobp))
                (not (looking-at regex)))
      (forward-line 1))))

(defun ce-current-sentence () (interactive)
  "Return identifier of the nearest annotation."
  (let ((regex "^Sentence \\(.+\\)$"))
    (save-excursion
      (unless (looking-at regex)
        (re-search-backward regex)))
    (match-string 1)))

(defun ce-current-conflict () (interactive)
  "Return identifier of the nearest conflict."
  (let ((regex "^\\(.+\\)([0-9]"))
    (unless (looking-at regex)
      (re-search-forward regex))
    (match-string 1)))
        
(defun ce-count () (interactive)
  "Count number of instances of this conflict."
  (save-excursion
    (let ((c (format "^%s" (ce-current-conflict))))
      (goto-char (point-min))
      (let ((i 0))
        (while (re-search-forward c nil 'noerror)
          (incf i))
        (message "`%s' occurs %d times." c i)))))

(defun ce-this () (interactive)
  "Display this sentence in XCDG."
  (save-window-excursion
    (save-excursion
      (shell-command
       (format "ccc-show-sentence '%s' & "  
               (ce-current-sentence))
       (get-buffer-create "*scratch*")))))


(defun ce-show () (interactive)
  "Display instances of this conflict in XCDG."
  (save-window-excursion
    (save-excursion
      (shell-command
       (format "ccc-show-conflicts -f %s '%s' & "  
               (buffer-file-name)
               (ce-current-conflict))
       (get-buffer-create "*scratch*")))))

(defun ce-annotation-file (corpus sentence)
  "Try to find the file that contains the annotation of SENTENCE in CORPUS."
  (interactive)
  (let* ((dir (concat "/data/corpus/ccc/" corpus))
        (file (format "%s/%s.cda" dir sentence)))
    (if (file-exists-p file)
        file
      (let* ((number
             (progn (string-match "s\\([0-9]+\\)" sentence)
                    (string-to-number (match-string 1 sentence))))
             (batch (/ number 10000))
             (subdir (format "%03d" batch)))
        (format "%s/%s/%s.cda" dir subdir sentence)))))

(defun ce-visit-anno () (interactive)
  "Visit the file that defines the annotation described at point.

If the cursor is standing on one word of the sentence, point will be
set to the annotation of that word."
  (let* ((filename (buffer-file-name))
         (corpus
          (progn
            (string-match "/\\([^/]+\\)/[a-z]+_errors" filename)
            (match-string 1 filename)))
         (sentence (ce-current-sentence))
         (word (thing-at-point 'word)))
    (setq word (replace-in-string word "'$" ""))
    (find-file (ce-annotation-file corpus sentence))
    (re-search-forward (concat "^ *[0-9]+ +[0-9]+ '?" word "'? *$"))))

(defun ce-visit-definition () (interactive)
  "Visit the definition of the current constraint."
  (let ((constraint (ce-current-conflict)))
    (find-file (concat (getenv "CDGDIR") "/grammar/deutsch/Grammatik.cdg"))
    (goto-char (point-min))
    (re-search-forward (concat ": *'?" constraint))))


(defun ce-kill () (interactive)
  "Delete the current conflict or sentence."
  (if (looking-at "^Sentence ")
      
      ;; Delete the entire sentence.
      (let ((end (save-excursion
                   (forward-line 1)
                   (re-search-forward "^Sentence " nil 'move)
                   (beginning-of-line)
                   (point))))
        (delete-region (point) end))
    
    ;; Delete only this conflict.
    (while (looking-at "^$") (forward-line -1))
    (let* ((regex "^.+([0-9]")
           (start (save-excursion 
                    (re-search-backward "\n\n" nil 'move)
                    (point)))
           (end (save-excursion
                  (re-search-forward "\n\n" nil 'move)
                  (- (point) 2))))
      (delete-region start end)
      (forward-char 2))))

(defun ce-kill-conflict (name) 
  (interactive "sConflict: ")
  "Delete all records about conflict NAME."
  (save-excursion
    (goto-char (point-min))
    (while (re-search-forward (format "^%s(" name) nil 'noerror)
      (ce-kill)))
  (ce-kill-empty-sentences))

(defun ce-kill-empty-sentences () (interactive)
  "delete all records from which all conflicts have been deleted."
  (save-excursion
    (goto-char (point-min))
    (while (re-search-forward "^Sentence " nil 'noerror)
      (unless
           (save-excursion
             (forward-line 2)
             (looking-at "SYN: "))
        (progn
          (beginning-of-line)
          (ce-kill))))))

               

(setq  auto-mode-alist
       (cons '("^\\(known\\|new\\)_errors$" . corpus-error-mode)
             auto-mode-alist))

(defface highlight-regexp-face
  '((t (:foreground "#ff0000")))
  "face for highlight-regexp.")

(defun highlight-regexp (regexp &optional face property)
  "Highlight all text matching a regexp.

FACE defaults to a combination of red on white.
All extents created by this function have the property 'highlight.
If non-NIL, PROPERTY is also added to each extent.

Returns the number of instances highlighted."

  (interactive "sRegexp: ")
  (save-excursion
    (beginning-of-buffer)
    (let ((num 0))
      (while (re-search-forward regexp nil t)
        (incf num)
        (highlight-range
         (match-beginning 0) (match-end 0) face property))
      (message "%d match%s" num (if (= num 1) "" "es"))
      num)))

(defun highlight-range (from to &optional face property)
  "Highlight specific range of current buffer.

FACE defaults to a combination of red on white.
All extents created by this function have the property 'highlight.
If PROPERTY is a symbol, it is added to each extent with
the value t. If PROPERTY is a plist, it is merged into the extent."
  (let ((ext (make-extent from to))
        (f face))
    (when (null f) (setq f 'highlight-regexp-face))
    (set-extent-property ext 'highlight t)
    (set-extent-face ext f)
    (set-extent-mouse-face ext f)
    (cond
     ((listp property)
      (set-extent-properties ext property))
     ((not (null property))
      (set-extent-property ext property t)))))



(provide 'corpus-error)
