#!/usr/bin/env perl
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.


BEGIN {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0);
  require 'cdgsetup.pl';
}

use warnings;
use strict;
use CDG::DepPlot;

our $VERSION = "v0.2";
our $COPYRIGHT = "Copyright (C) 2005 The CDG Team\n\nThis is free software; see the source for copying conditions.\nThere is NO warranty; not even for MERCHANTABILITY or FITNESS\nFOR A PARTICULAR PURPOSE.";

our $PROGNAME = "depplot";
our $VERSION_INFO_LONG = "Dependency Tree Plotter (".$PROGNAME.") ".$VERSION."\n\n".$COPYRIGHT."\n";
our $VERSION_INFO_SHORT = $PROGNAME." ".$VERSION."\n";

our $DEFAULT_TYPE = "log";
our $DEFAULT_OUTPUT_EXT = ".png";
our $DEFAULT_NO_TYPE = "none";

our $QUIET = 0;
our $ERROR_PREFIX = "ERROR: ";

# flush output linebuffered
$|=1;

sub write {
	if (!$QUIET) { print(shift); }
}

sub error {
	if ($QUIET < 2) { die($ERROR_PREFIX.shift); }
}

sub print_version {
	if (shift == 1) {
		&write($VERSION_INFO_LONG);
	} else {
		&write($VERSION_INFO_SHORT);
	}
}

sub print_help_page {
	&write("USAGE: ".$PROGNAME." [options] [filename] ...\n");
	&write("\n");
	&write("[options] may be any combination of\n");
	&write("--out [filename]\n");
	&write("-o [filename]	sets the Output Filename to [filename]\n");
	&write("--type [type]\n");
	&write("-t [type]	the Parser handles the input as if it\n");
	&write("		were of type [type]\n");
	&write("--quiet\n");
	&write("-q		quiet mode\n");
	&write("--reallyquite\n");
	&write("--shutup\n");
	&write("-qq		really quiet mode\n");
	&write("--help\n");
	&write("-h		help page\n");
	&write("--version	show version info - long\n");
	&write("-v		show version info\n");
	&write("\n");
	&write("[type] may be one of\n");
	&write("log		cdg log file\n");
	&write("cda		cda dependency tree\n");
	&write("xml		xml encoded dependency tree\n");
	&write("none		autodetect\n");
}

sub getTypeFromFileName {
	my $fname = shift;
	$fname =~ /.*\.(.?.?.?)$/;
	my $ext = $1;
	
	if ($ext && $ext =~ /log|cda|xml/) {
		return $ext;
	} else {
		return $DEFAULT_TYPE;
	}
}

sub render {
	my $in = shift;
	my $out = shift;
	my $type = shift;
	
	&write($in);
	
	if (!$type || $type eq $DEFAULT_NO_TYPE) {
		$type = &getTypeFromFileName($in);
	}
	
	open(IN_HANDLE, "<$in") or do {
		&write("\n");
		&error("cannot open $in\nstopped");
	};
	
	# read all input
	my $input = '';
	while (<IN_HANDLE>) { $input .= $_; };
	close(IN_HANDLE);
	
	# parse
	my %parseResult;
	eval { %parseResult = &CDG::DepPlot::parse($input, $type) }; if ($@) {
		&write("\n");
		&error($@);
	}
	
	# render
	eval { &CDG::DepPlot::graph(\%parseResult, $out) }; if ($@) {
		&write("\n");
		&error($@);
	}
	
	&write(" done\n");
	
}

# -------------------------------------------------------------------
# main
# -------------------------------------------------------------------

# process arguments

my $in = '';
my $out = '';
my $type = $DEFAULT_NO_TYPE;


while (my $arg = shift) {
	if ($arg =~ /-o|--out/) {
		$out = shift;
	} elsif ($arg =~ /-t|--type/) {
		$type = shift;
	} elsif ($arg =~ /-q|--quiet/) {
		$QUIET = 1;
	} elsif ($arg =~ /-qq|--reallyquiet|--shutup/) {
		$QUIET = 2;
	} elsif ($arg =~ /-v$/) {
		&print_version(0);
		exit;
	} elsif ($arg =~ /--version/) {
		&print_version(1);
		exit;
	} elsif ($arg =~ /-h|-\?|--help/) {
		&print_help_page;
		exit;
	} else {
		$in = $arg;
		$out = $in.$DEFAULT_OUTPUT_EXT if ($out eq '');
		
		&render($in, $out, $type);
		
		$out = '';
	}
}

