#!/usr/bin/env perl
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

=head1 NAME

Depsy - A Dependency Synthesiser

=head1 SYNOPSIS

=over 10

=item Depsy

=over 10

=item [B<-annos> I<filename>]

=item [B<-annoprefix> I<string>]

=item [B<-debug>]

=item [B<-debug-label-table>]

=item [B<-extralevels> I<string>]

=item [B<-index> I<number>]

=item [B<-input-format|if> I<string>]

=item [B<-levels> I<filename>]

=item [B<-lexicon> I<filename>]

=item [B<-man>]

=item [B<-max|n> I<int>]

=item [B<-output-format|of> I<string>]

=item [B<-plugin> I<filename>]

=item [B<-prefix> I<string>]

=item [B<-range> "I<range-spec>"]

=item [B<-syntax|help|?>]

=item [B<-wordgraphs> I<filename>]

=item [I<inputFile1> I<inputFile2> ...]

=back

=back

=head1 DESCRIPTION

This tool is part of the CDG parsing suite.
It translates annotations given in the negra file format to cdg
annotations, and thereby translate phrase structure representations into
dependency representations of the annotations. Additionally it can generate a
rudimentary lexicon and a level description in order to load the generated data
into the WCDG system.

For debugging purpose the phrase structures extracted from the negra format are
converted to a format resembling the Penn Treebank format which is much easier
to read. Note, that this output is not thought to be correctly with regards to
the Penn Treebank format definition but only to resemble to it. At least there
is additional information written, i.e. the node number linking back to the
lines in the negra format sentence (0-499 for leaves, 500- for inner nodes).
The head phrases analysed are marked as well using C<*> at the lexical level and
C<!!> for the inner nodes (see the function C<markLexHeads> L<"markLexHeads()">).

There are a couple of configurations which are possible in order to customize
the desired translation. Actually reducing the negra phrase structures to
dependency structures is not only a process of information reduction but also
enriching the annotation with information needed with respect to the target
corpus. And last not least, not all phrase structures are annotated with lexical
heads.

=head1 OPTIONS

C<Depsy> reads from STDIN
or from the specified files (I<inputFile1> I<inputFile2> ...)
and stores the results in one or more files in the format specified
by -B<-output-format> (default 'cdg').
To configure that tool you have to write a plugin that you specify
with the B<-plugin> option. Results of are written to
the files that you specify with
B<-annos>, B<-lexicon>, B<-levels> and B<-wordgraphs>. There are two
flavours of output format: 'plain' and 'cdg'. The latter format is
valid CDG input to be used with the CDG parser developed by the same
team.
In general, output-files that aren't specified have no defined default
value and thus no corresponding information is extracted.
In what follows you get a detailed list of all those parameters.

=over 10

=item B<-annos> I<filename>

generate dependency trees and save them in I<filename>.

=item B<-annoprefix> I<string>

(default 'auto-s')

This option lets you specify the prefix string to be added to the
annotation's name.

=item B<-debug>

switch on the debugging mode. The currently processed sentence and other
stuff is displayed on C<STDERR>.

=item B<-debug-label-table>

Display statistics about the label table. For each rule a sample of
the sentences in which the rule was used is given, plus the total
number of applications.

=item  B<-extralevels> I<string>

This option can be set to a comma-separated string. It will cause the
output of Depsy to contain other levels than the SYN level. All
subordinations on these levels will be unlabelled NIL bindings, unless
you change them via dependency conversion rules.

=item B<-index> I<number>

specifies the sentence index which we start to number wordgraphs and annotations.
Note, that the Negra Treebank numbers sentences by itself which
has higher priority than the index that we might want to have giving this
index option. This is actually only usefull if there is no explicite
sentence numbering scheme and we have to deal with that ourselfts as it
is in the Penn Treebank.

=item B<-input-format> I<string>

TODO
(default 'negra')

=item B<-levels> I<filename>

generate a syntax level declaration including all needed labels. A
level declaration is going to be generated only if a I<filename>
is specified.

=item B<-lexicon> I<filename>

extract a rudimentary lexicon from the morphological annotations. This
basically serves the purpose to let you load the generated annotations
into your xcdg to display the annotations as graphical trees. If you don't
specify an I<filename> you will not get a lexicon.

=item B<-man>

print out the complete manual of C<Depsy>.

=item B<-max|n> I<int>

specify the maximum number of sentences that are converted.
If no maximum is specified via this parameter we extract as many as possible.

=item B<-output-format> I<string>

TODO (default 'cdg')

=item B<-plugin> I<filename>

TODO

=item B<-prefix> I<string>

TODO (default 's')

specify the prefix string to be used when naming annotations and wordgraphs.

=item B<-range> "I<range-spec>"

specify a list of intervals of annotation ids to be processed.
You might specify an interval 1-625, which includes all annotations between 1 and 625
or any arbitrary numbers, you might say -400 specifying the first 400 annotations, or
10000- to process the annotations 10000 til the end. You can combine comma separated
a list of intervals, e.g. 1-10,50-70,1000-

=item B<-syntax|help|?>

a short syntax summary message.

=item  B<-wordgraphs> I<filename>

generate wordgraphs as linear cdg sentences in I<filename>. If you
don't use this option you will not get the cdg sentences.

=back

=head1 CONFIGURATION

=head2 Writing plugins

A plugin basically must provide ... tabels that configure the behavior of
C<Depsy>. These are:

=over 10

=item %headTable

The purpos of this hashtable is to establish rules to detect lexical heads of a
phrase structure, an information which is not annotated in most cases. This
information is applied by the function C<markLexHeads> L<"markLexHeads()">.

=item %phraseConvTable

This hashtable maps node tags to a list of perl functions the purpose of which
is to always call these functions in the given order whenever we come past the
node tag. This happens in the function C<convertPhraseTree>
L<"convertPhraseTree()"> after the recursive phrase structure has been build
and before the dependency structure is extracted from that.

=item @labelTable

This is a priority list of in order to configure the label generation of the
dependencies. The C<@labelTable> is used by the function C<makeDependency>
L<"makeDependency()">.

=item %depConvTable

This hashtable maps dependency labels to a list of perl functions the purpose
of which is to always call these functions in the given order whenever we come
past the named label. This happens in the function C<convertDepTree>
L<"convertDepTree()"> after the dependency tree has been constructed.

=item %callbacks

This is the hash of known callbacks.  Below is the list of all known callbacks.
Each callback points to a function which returns 0 or 1 depending on success or
failure. Every known callback is sensitive for this return value.

=over 3

=item C<readSentence(nodes)>

called at the end of L<"readNegraSentence()"> after one sentence
(between BOS and EOS) has been parsed. C<nodes> is a pointer to the
hash to all nodes that have been generated for that sentence. If this
callback returns 0, processing of that sentence is aborted, and the
next one is read in.

=item C<preProcess(edges)>

Called before processSentence operates. The parameter is a pointer to
the root of the phrase tree.

=item C<postProcess(edges)>

called at the end of L<"processSentence()">. C<edges> is a pointer to
the list of all dependency edges that have been constructed. The
return code of the callback becomes the return code of
L<"processSentence()">, that is treebank conversion is aborted on 0
and continued on 1;

=item C<initialize()>

called before starting to read the treebank.

=item C<finalize()>

called at the end of the L<"main()"> routine.

=item C<writeLexiconEntry>

called before writing a lexical entry in L<"writeLexiconEntry()">. This
callback gets a tree node as a parameter.  See L<"getFeatures()"> on how to
generate some default lexicial features for the lexicon entry.

If this callback returns 0, then no lexical entry is printed. Maybe you wanted
to suppress some of these, maybe you generate lexical entries by yourself in
the C<LEX> stream. Otherwise 1 should be returned.

=item C<getFeatures>

called in L<"getFeatures()"> if there's no default morphology. Arguments
are the same as to L<"getFeatures">.

=back

=back

=head1 IMPLEMENTATION

=cut

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

use warnings;
use strict;
use Getopt::Long;
use Pod::Usage;
use Parse::RecDescent;

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

=head2 Global Variables

=head3 Variables set by command line parameters

All of the following variables are initialized by command-line parameters as
far as they are given in a specific call, leaving some variables undefined.

=over 10

=item C<$debug> (default 0)

This variable holds the debugging mode with
values 0 or 1 switched on/off by command-line parameter C<-debug> as described in
section L<"OPTIONS">.  In debug mode all debug messages are generated to STDERR.

=cut

my $debug = 0;

my $debug_label_table = 0;

=item C<$maxSentence>

This variable is an integer value greater zero set by command-line parameter
C<-max> (or C<-n>). No more than C<$maxSentence> are generated by C<$Depsy>.

=cut

my $maxSentence;

=item C<@sentenceIdRange>

range-specification as described above for the command-line parameter C<-range>

=cut

my @sentenceIdRange;

=item C<$wordgraphsFile>

This variable holds the filename where the constructed wordgraphs should
be written in.
See C<-wordgraphsFile> in section L<"OPTIONS">

=cut

my $wordgraphsFile;

=item C<$lexiconFile>

This variable holds the filename of the lexicon to be constructed.
See C<-lexicon> in section L<"OPTIONS">

=cut

my $lexiconFile;

=item C<$annoFile>

This variable holds the filename of the resulting annotations in cdg format.
See C<-annos> in section L<"OPTIONS">

=cut

my $annoFile;

=item C<$levelDclFile>

This variable holds the filename of the level declaration which might
optionally be build.  See C<-levels> in section L<"OPTIONS">
undef

=cut

my $levelDclFile;

=item C<$pluginFile>

This is the filename of the plugin to be loaded.

=cut

my $pluginFile;

=item C<$inputFormat>

This Variable holds the format of the input to be read. Valid values are
"penn" and "negra".

=cut

my $inputFormat = "negra";

=item C<$outputFormat>

This Variable holds the format of the output to be written. Valid values are
"cdg" and "plain".

=cut

my $outputFormat = "cdg";

=item C<%wgPrefix>

This is the prefix string added to the names of the generated wordgraphs.

=cut

my $wgPrefix = "s";

=item C<%annoPrefix>

This is the prefix string added to the names of the generated annotations.
wordgraphs.

=back

=cut

my $annoPrefix = "auto-s";

=head3 Plugin tables

=over 10

=item C<%headTable>

See section L<Writing plugins>.

=cut

my %headTable;

=item C<@labelTable>

See section L<Writing plugins>.

=cut

my @labelTable;

=item C<%phraseConvTable>

See section L<Writing plugins>.

=cut

my %phraseConvTable;

=item C<%depConvTable>

See section L<Writing plugins>.

=cut

my %depConvTable;

=item C<%callbacks>

See section L<Writing plugins>.

=back

=cut

my %callbacks;

=head3 Variables used in the conversion processes

=over 10

=item C<$rootNode>

pointer to the current rootNode of the phrase tree.

=cut

my $rootNode;

=item C<%nrSentences>

number of sentences processed so far. Its value is increased by processSentence
only.

=cut

our $nrSentences = 1;

=item C<%sentenceIndex>

current index of the sentence that is being processed. You can set the initial
value using option C<-index>.

=cut

my $sentenceIndex = 1;

=item C<$extralevels>

undocumentend feature

=cut

our $extralevels = '';

=item C<@extralevels>

undocumentend feature

=cut

our @extralevels = ();

=back

=head3 Storage for the negra sections

The following variables contain the eight negra tables C<ORIGIN>, C<EDITOR>,
C<WORDTAG>, C<MORPHTAG>, C<NODETAG>, C<EDGETAG> and C<SECEDGETAG> as they
are read from the input.

=over 10

=item C<@origins>

This list contains the C<ORIGIN> negra table and contains the name and the
comment field of an origin-id. See the function C<readOrigins> L<"readOrigins()">.

=cut

my @origins;

=item C<%editors>

This hashtable maps editor ids to their login and their name. Actually this
information isn't used anywhere. But as long as the negra sources contain this
info we eat them.  See the function C<readEditors> L<"readEditors()">.

=cut

my %editors;

=item C<%wordTags>

This hashtable maps a tag to its description. The rest of the available information,
i.e.  the id and the flag, aren't analyzed, besides the fact that the actually used
word tags in the sentences aren't checked for existence in this table.
See the function C<readWordTags> L<"readWordTags()">.

=cut

my %wordTags;

=item C<%morphTags>

This hashtable maps the morph tag ids to their realization. the C<%morphTags>
are used for generating the features of a word in an annotation in
C<getFeatures> L<"getFeatures()">.  See the function C<readMorphTags>
L<"readMorphTags()">.

=cut

my %morphTags;

=item C<%nodeTags>

This hashtable maps the node tag to its human readable description,
e.g. C<< "NP" => "noun phrase" >>. It is of no actual use here also.
See the function C<readNodeTags> L<"readNodeTags()">.

=cut

my %nodeTags;

=item C<%edgeTags>

This hashtable maps the edge tag to its human readable description, e.g. C<<
"NK" => "noun kernel modifier" >>. Same usefulness as the above C<%nodeTags>
(none).  See the function C<readEdgeTags> L<"readEdgeTags()">.

=cut

my %edgeTags;

=item C<%secEdgeTags>

TODO

=cut

my %secEdgeTags;

=item C<%lexicon>

TODO

=cut

my %lexicon;

=item C<%labels>

TODO

=cut

my %labels;

=item C<$nolabelWarning>

this flags the 'no label' warning having occured

=back

=cut

my $nolabelWarning = 0;

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

=head2 Data Structures

In this section we describe the most important data structures as they are
used in the implementation. Some data structures have already been described
in the section L<"Global Variables"> as they are global variables used to
store the negra input. But after that we want to discuss at least the
main data types that we are going to stumble across in section L<"Functions">.
This is

=over 10

=item * a I<linear representation>

of the phrase structures as they
are read in from the negra input file, called C<%sentence>
(L<"A linear phrase structure representation">),

=item * which is the converted into a I<recursive data type>

made out of
so called C<%nodes>
(L<"A recursive phrase structure representation">),

=item * and which are then converted into a list of I<dependency edges>

that is some C<@deps>, a list of dependencies
(L<"Dependency structures">).

=back


=head3 A linear phrase structure representation

TODO

See C<readNegraSentence> L<"readNegraSentence()"> and C<makePhraseTree> L<"makePhraseTree()">.


=head3 A recursive phrase structure representation

TODO: elaborate!!!

  tree = {
    parent => tree
    children => <list of trees>,
    nodeTag => <nodeTag>,
    edgeTag => <edgeTag to parent>
    id => <index of word in the sentence>
  }

=head3 Dependency structures

=cut

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

=head2 Functions

=cut

#------------------------------------------------------------------------------

=head3 C<debug()>

print debug messages if the C<$debug> flag is set.

=cut

sub debug {
  return if ! $debug;
  my $m = shift;
  if($m !~ /\n$/) {
    $m .= "\n";
  }
  print STDERR "%% DEBUG: $m";
}

#------------------------------------------------------------------------------

=head3 C<readNegraTreebank()>

This functions reads and processes the input which is suppposed to be
in Negra Export format.

=cut

sub readNegraTreebank {

  while (my $line = <>) {
    chomp($line);

    next if ($line =~ /^%%/);
    next if ($line =~ /^$/);

    # negra file format
    if ($line =~ /^#FORMAT (\d*)/) {
      # do nothing
      next;
    }

    # origin table
    if ($line =~ /^#BOT ORIGIN/) {
      &readOrigins();
      next;
    }

    # editor table
    if ($line =~ /^#BOT EDITOR/) {
      &readEditors();
      next;
    }

    # wordtag table
    if ($line =~ /^#BOT WORDTAG/) {
      &readWordTags();
      next;
    }

    # morphtag table
    if ($line =~ /^#BOT MORPHTAG/) {
      &readMorphTags();
      next;
    }

    # nodetag table
    if ($line =~ /^#BOT NODETAG/) {
      &readNodeTags();
      next;
    }

    # edgetag table
    if ($line =~ /^#BOT EDGETAG/) {
      &readEdgeTags();
      next;
    }

    # secedgetag table
    if ($line =~ /^#BOT SECEDGETAG/) {
      &readSecEdgeTags();
      next;
    }

    # sentence
    if ($line =~ /^#BOS (\d+)\s+(\d+)\s+(\d+)\s+(\d+)/) {
      $sentenceIndex = $1;
      my $editorId = $2;
      my $date = $3;
      my $originId = $4;


      # skip unwanted sentence
      if (@sentenceIdRange) {
	if (!&isIncluded($sentenceIndex, @sentenceIdRange)) {
	  while (my $line = <>) {
	    last if ($line =~ /^#EOS $sentenceIndex/);
	  }
	  next
	}
      }


      # read the sentence
      debug "reading negra sentence no. $sentenceIndex.";
      my $nodes = &readNegraSentence($editorId, $date, $originId);
      next if ! $nodes;

      # generate phrase trees
      &makePhraseTree($nodes);

      # process the phrase tree
      last if ! &processSentence($nodes);
      next;
    }
  }
}

#------------------------------------------------------------------------------

=head3 C<readPennTreebank()>

This functions reads and processes the input which is suppposed to be
in Penn Treebank format. It therefore uses a RecDescent grammar which calls
processSentence for each sentence.

=cut

sub readPennTreebank {

my $pennGrammar = <<'EOG';
{
  my $debug = 0;
  my $maxId = 0;
  my $innerMaxId = 500;
  my $posTagPattern = q{(?:
    \#|\$|(?:'')|,|(?:-LRB-)|(?:-RRB-)|\.|:|(?:``)|(?:\bWP\$?)|(?:\bPRP\$)|
    \b(?:(?:CC)|(?:CD)|(?:DT)|(?:EX)|(?:FW)|(?:IN)|(?:JJ)|(?:JJR)|(?:JJS)|
    (?:LS)|(?:MD)|(?:NN)|(?:NNP)|(?:NNPS)|(?:NNS)|(?:PDT)|(?:POS)|(?:PRP)|
    (?:RB)|(?:RBR)|(?:RBS)|(?:RP)|(?:SYM)|(?:TO)|(?:UH)|(?:VB)|(?:VBD)|
    (?:VBG)|(?:VBN)|(?:VBP)|(?:VBZ)|(?:WDT)|(?:WRB))\b)};

  my $nodeTagPattern = q{[A-Z\|]+};
  my $edgeTagPattern = q{(?:(?:--)|[A-Z]+)};
  my %nodes;	# map node ids to nodes
  my %references; # map trace references to nodes

  #my $secEdgeTagPattern = q{
  #(?:\*)|(?:\*EXP\*)|(?:\*ICH\*)|(?:\*PPA\*)|
  #(?:\*RNR\*)|(?:\*T\*)|(?:0)};
  #
}

#sentence: { print STDERR "DEBUG: trying sentence at\n'$text'\n" if $debug; } <reject>
sentence: '(' sExpr ')'
  {
    print STDERR "DEBUG: found sentence\n" if $debug;
    my $rootNode = {
      "id" => -1,
      "isLexHead" => 1,
      "nodeTag" => "ROOT",
      "edgeTag" => "ROOT",
      "word" => "NIL",
      "children" => [$item[2]],
    };

    # establish traces and coreferences
    foreach my $nodeId (keys %nodes) {
      my $node = $nodes{$nodeId};
      if(defined $node->{traceId}) {
	my $reference = $references{$node->{traceId}};
  	$node->{trace} = $reference;
	$reference->{backTrace} = $node;
      }
      if(defined $node->{corefId}) {
	my $reference = $references{$node->{corefId}};
  	$node->{coref} = $reference;
	$reference->{backRef} = $node;
      }
    }

    # process one sentence
    $return = &::processSentence($rootNode, \%nodes);
    undef %nodes;
    undef %references;
    $maxId = 0;
    $innerMaxId = 500;
  } | {
    print STDERR "ERROR: in $::nrSentences. sentence\n$text\n";
    $return = -1;
  }

#sExpr: { print STDERR "DEBUG: trying sExpr at\n$text\n" if $debug; } <reject>
sExpr: '(' sExprContent ')'
  {
    print STDERR "DEBUG: found sExpr\n" if $debug;
    $return = $item[2];
  }

#sExprContent: { print STDERR "DEBUG: trying sExprContent at\n$text\n" if $debug; } <reject>
sExprContent:
  nodeTag sExpr(s)
  {
    my $nodeTag = $item[1][0];
    my $edgeTag = $item[1][1];
    my $ref1Id = $item[1][2];
    my $ref2Id = $item[1][3];
    my $node = {
      "nodeTag" => $nodeTag,
      "edgeTag" => $edgeTag?$edgeTag:"--",
      "id" => $innerMaxId++,
      "children" => $item[2],
    };
    $nodes{$node->{id}} = $node;

    if (defined $ref1Id) {
      $node->{corefId} = $ref1Id;
    }
    if (defined $ref2Id) {
      $references{$ref2Id} = $node;
    }
    foreach my $child (@{$node->{children}}) {
      $child->{parent} = $node;
    }
    $return = $node;
  }
  | posTag /([^()\s]+)/
  {
    my $node = {
      "nodeTag" => $item[1],
      "edgeTag" => "--",
      "id" => $maxId++,
      "word" => $item[2],
    };
    $nodes{$node->{id}} = $node;
    $return = $node;
  }
  | emptyCat
  {
    my $node = {
      "word" => $item[1][0],
      "edgeTag" => "--",
      "nodeTag" => "-NONE-",
      "id" => $maxId++,
    };
    if ($item[1][1]) {
      $node->{traceId} = $item[1][1];
    }
    $nodes{$node->{id}} = $node;
    $return = $node;
  }

#emptyCat: { print STDERR "DEBUG: trying emptyCat at\n$text\n" if $debug; } <reject>
emptyCat:
  '-NONE-' /0|(\*\?\*)/
  {
    $return = [$item[2]];
  }
  |
  '-NONE-' /(\*(?:[A-Z]+\*)?)(?:\-([0-9]+))?/
  {
    $return = [$1, $2];
  }
  | <error: error parsing emptyCat near\n$text>

#nodeTag: { print STDERR "DEBUG: trying nodeTag at\n$text\n" if $debug; } <reject>
nodeTag: /($nodeTagPattern)((?:-$edgeTagPattern)*)(?:=([0-9]+))?(?:-([0-9]+))?/
  {
    my $nodeTag = $1;
    my $edgeTags = $2;
    my $ref1Id = $3;
    my $ref2Id = $4;
    if ($edgeTags) {
      $edgeTags =~ s/^-//;
    }
    $return = [$nodeTag, $edgeTags, $ref1Id, $ref2Id];
  }
  | <error: error parsing an nodeTag near\n$text>

#posTag: { print STDERR "DEBUG: trying posTag at \n$text\n" if $debug; } <reject>
posTag: /($posTagPattern)/x
  {
    print STDERR "DEBUG: found posTag = $item[1]\n" if $debug;
    $return = $item[1];
  }
  | <error: error parsing a posTag near\n$text>


EOG

  my $parser = Parse::RecDescent->new($pennGrammar)
    or die "ERROR: bad treebank parser\n";

  # call processSentence for each sentence
  undef $/;
  while (my $data = <>) {
    my $retval = $parser->sentence(\$data);
    die "\n" if $retval == -1;
    last if $retval == 0;
  }
  $/ = "\n";
}

#------------------------------------------------------------------------------

=head3 C<processSentence()>

This function is called whenever one sentence has been read from the input
and a recursive phrase data structure has been built in memory.

=cut

sub processSentence {
  my $nodes = shift;

  # keep track of the processed sentences

  # skip unwanted sentence
  return 0 if $maxSentence && $nrSentences > $maxSentence;
  if (@sentenceIdRange) {
    if (!&isIncluded($sentenceIndex, @sentenceIdRange)) {
      return 1;
    }
  }

  print STDERR "${wgPrefix}$sentenceIndex        \r";

  my $result = 1;
  if ($callbacks{preProcess}) {
    $result &&= $callbacks{preProcess}->($rootNode);
  }

  # head marking
  &markLexHeads($rootNode);

  # phrase conversions
  debug "*** phrase conversions\n";
  &phraseSnapShot($rootNode);

  &convertPhraseTree($rootNode);

  # head marking
  &markLexHeads($rootNode);
  debug "*** head marking\n";

  # convert the phrase tree to dependency relations
  debug "*** label conversion\n" if $debug;
  $nolabelWarning = 0;
  my @deps = &makeAllDeps($rootNode);

  # optionally, insert dependencies on other levels than SYN
  my %extradeps;
  if(@extralevels) {
    for (0..$#deps) {
      for my $l (@extralevels) {
        $extradeps{$l}[$_] = { label => '', up => 0 };
      }
    }
  }

  # convert the list of dependencies as configured
  debug "*** dependency conversion\n";
  &depSnapShot(\@deps, \%extradeps);
  &convertDepTree(\@deps, \%extradeps);

  # write cdg wordgraphs and annotations
  &writeWordgraph(\@deps);
  &writeAnnotation(\@deps, \%extradeps);

  # write word to lexicon if it is not already in there
  if ($lexiconFile) {
    foreach my $nodeId (keys %$nodes) {
      my $node = $nodes->{$nodeId};
      next if ! defined $node->{word};
      my $key = "$node->{word} $node->{nodeTag} $node->{morphTag}";
      my $entry = $lexicon{$key};
      next if $entry;
      $lexicon{$key} = 1;
      &writeLexiconEntry($node);
    }
  }

  # callback
  if ($callbacks{postProcess}) {
    $result &&= $callbacks{postProcess}->(\@deps);
  }

  $nrSentences++;
  $sentenceIndex++;
  return $result;
}

#------------------------------------------------------------------------------

=head3 C<readOrigins()>

=cut

sub readOrigins {
  # id, name, (comment)
  my $originPattern = '^(\d+)\s+([a-zA-Z0-9\.\-]+)(\s+%%\s+(.*))?';

  while (my $line = <>) {
    chomp($line);

    next if ($line =~ /^%%/);
    next if ($line =~ /^$/);
    last if ($line =~ /#EOT ORIGIN/);

    if ($line =~ /$originPattern/) {
#      print "// origin id=$1, name=$2, comment=$3\n" if $debug;
      $origins[$1] = { "name" => $2 , "comment" => $3 };
      next;
    }
    print STDERR "WARNING: cannot parse origin '$line'\n";
  }
}

#------------------------------------------------------------------------------

=head3 C<readEditors()>

=cut

sub readEditors {
  # id, login, name
  my $editorsPattern = '^([\-\d_]+)\s+([a-zA-Z0-9\.\-_]+)\s+(.*)';
  while (my $line = <>) {
    chomp($line);

    next if ($line =~ /^%%/);
    next if ($line =~ /^$/);
    last if ($line =~ /#EOT EDITOR/);

    if ($line =~ /$editorsPattern/) {
#      print "// editor id=$1, login=$2, name=$3\n" if $debug;
      $editors{$1} = { "login" => $2 , "name" => $3 };
      next;
    }
    print STDERR "WARNING: cannot parse editor '$line'\n";
  }
}

#------------------------------------------------------------------------------

=head3 C<readWordTags()>

=cut

sub readWordTags {
  # id, tag, flag, description
  my $wordTagsPattern = '^([\-\d]+)\s+([A-Z\-\$,\.\(]+)\s+([YN])+\s+(.*)';

  while (my $line = <>) {
    chomp($line);

    next if ($line =~ /^%%/);
    next if ($line =~ /^$/);
    last if ($line =~ /#EOT WORDTAG/);

    if ($line =~ /$wordTagsPattern/) {
      $wordTags{$2} = $4;
#      print "// wordTags{$2} = $wordTags{$2}\n" if $debug;
      next;
    }
    print STDERR "WARNING: cannot parse wordtag '$line'\n";
  }
}

#------------------------------------------------------------------------------

=head3 C<readMorphTags()>

=cut

sub readMorphTags {
  # id, tag, description
  my $morphTagsPattern = '^([\-\d]+)\s+([a-zA-Z0-9\-\.\*\/]+)\s+(.*)';

  while (my $line = <>) {
    chomp($line);

    next if ($line =~ /^%%/);
    next if ($line =~ /^$/);
    last if ($line =~ /#EOT MORPHTAG/);

    if ($line =~ /$morphTagsPattern/) {
      $morphTags{$2} = $1; # store the id
#      print "// morphTags{$2} = $morphTags{$2}\n" if $debug;
      next;
    }
    print STDERR "WARNING: cannot parse morphtag '$line'\n";
  }
}

#------------------------------------------------------------------------------

=head3 C<readNodeTags()>

=cut

sub readNodeTags {
  # id, tag, description
  my $nodeTagsPattern = '^(\d+)\s+([a-zA-Z\-\.\*]+)\s+(.*)';

  while (my $line = <>) {
    chomp($line);

    next if ($line =~ /^%%/);
    next if ($line =~ /^$/);
    last if ($line =~ /#EOT NODETAG/);

    if ($line =~ /$nodeTagsPattern/) {
      $nodeTags{$2} = $3;
#      print "// $1: nodeTags{$2} = $nodeTags{$2}\n" if $debug;
      next;
    }
    print STDERR "WARNING: cannot parse nodetag '$line'\n";
  }
}
#------------------------------------------------------------------------------

=head3 C<readEdgeTags()>

=cut

sub readEdgeTags {
  # id, tag, description
  my $edgeTagsPattern = '^(\d+)\s+([a-zA-Z0-9\-\.\*]+)\s+(.*)';

  while (my $line = <>) {
    chomp($line);

    next if ($line =~ /^%%/);
    next if ($line =~ /^$/);
    last if ($line =~ /#EOT EDGETAG/);

    if ($line =~ /$edgeTagsPattern/) {
      $edgeTags{$2} = $3;
#      print "// $1: edgeTags{$2} = $3\n" if $debug;
      next;
    }
    print STDERR "WARNING: cannot parse edgetag '$line'\n";
  }
}
#------------------------------------------------------------------------------

=head3 C<readSecEdgeTags()>

=cut

sub readSecEdgeTags {
  # id, tag, description
  my $secEdgeTagsPattern = '^(\d+)\s+([a-zA-Z0-9\-\.\*]+)\s+(.*)';

  while (my $line = <>) {
    chomp($line);

    next if ($line =~ /^%%/);
    next if ($line =~ /^$/);
    last if ($line =~ /#EOT SECEDGETAG/);

    if ($line =~ /$secEdgeTagsPattern/) {
      $secEdgeTags{$2} = $3;
#      print "// $1: secEdgeTags{$2} = $3\n" if $debug;
      next;
    }
    print STDERR "WARNING: cannot parse secedgetag '$line'\n";
  }
}

#------------------------------------------------------------------------------

=head3 C<readNegraSentence()>

=cut

sub readNegraSentence {
  my ($editorId, $date, $originId) = @_;

  # id >= 500, nodeTag, edgeTag, parent
  my $innerNodePattern = '^#(\d+)\s+([A-Z-]+)\s+--\s+([A-Z0-9\-]+)\s+(\d+)';

  # word, nodeTag, morphTag, edgeTag, parent
  my $leafNodePattern = '^(\S+)\s+(\S+)\s+([a-zA-Z0-9\.\*\-]+)\s+([A-Z0-9\-\*\?]+)\s+(\d+)';

  # read from stream BOS to EOS
  my $id = 0;
  my %nodes;
  while (my $line = <>) {
    chomp($line);

    next if ($line =~ /^%%/);
    next if ($line =~ /^$/);
#    next if ($line =~ /^--/);
    last if ($line =~ /\#EOS $sentenceIndex/);

    # inner nodes
    if ($line =~ /$innerNodePattern/) {
      $nodes{$1} = {
        "id" => $1,
	"nodeTag" => $2,
	"edgeTag" => $3,
	"parent" => $4
      };

#      printf "%03d: %s\tcat=%s\tedge=%s\tparent=%03d\n", $1, "", $2, $3, $4 if $debug;
      debug "$1:   \tcat=$2\tedge=$3\tparent=$4";
      next;
    }

    # leaf nodes
    if ($line =~ /$leafNodePattern/) {
      my $word = $1;
      my $nodeTag = $2;
      my $morphTag = $3;
      my $edgeTag = $4;
      my $parent = $5;

      # catch spelling correction made by the annotator
      if($line =~ /%% +([a-zA-Z]+) *$/) {
        $word = $1;
      }

      # hack: get rid of empty categories
      # TODO: implement in readSentence callback
      next if $nodeTag eq "-NONE-";

      if ( $morphTag =~ /^((\*)|(\*\.\*\.\*)|(\*\.\*\.\*\.\*\.\*))$/ ) {
	$morphTag = "--";
      }

      $nodes{$id} = {
	"id" => $id,
	"word" => $word,
	"nodeTag" => $nodeTag,
	"morphTag" => $morphTag,
	"edgeTag" => $edgeTag,
	"parent" => $parent
      };

#      printf "%03d: %s\tcat=%s\tparent=%03d\n", $id, $word, $nodeTag, $parent if $debug;
      debug "$id: $word\tcat=$nodeTag\tedge=$edgeTag\tparent=$parent";
      $id++;
      next;
    }

    print STDERR "WARNING: in sentence $wgPrefix$sentenceIndex, cannot parse '$line'\n";
  }

  # callback
  if ($callbacks{readSentence}) {
    return undef if ! $callbacks{readSentence}->(\%nodes);
  }

  return \%nodes;
}

#------------------------------------------------------------------------------

=head3 C<escapeIdentifier()>

escape and quote a string only if necessary

=cut

sub escapeIdentifier {
  (my $result = shift) =~ s/([^\\])\'/$1\\\'/g;
  $result =~ s/^\'/\\\'/g;
  if($result !~ /^[a-zA-Z\x80=\xff][a-zA-Z0-9\x80=\xff]+$/ ||
     $result =~ /^(true|false|label|level)$/) {
    $result = "'$result'";
  }
  return $result;
}

#------------------------------------------------------------------------------

=head3 C<writeWordgraph()>

=cut

sub writeWordgraph {
  return if ! $wordgraphsFile;

  my $deps = shift;

  print WG escapeIdentifier("$wgPrefix$sentenceIndex");
  print WG ": ";
  my $s = '';
  for(@{$deps}) {
    $s .= escapeIdentifier($$_[0]{word});
    $s .= ", ";
  }
  $s =~ s/, $/;\n/;
  print WG $s;
}

#------------------------------------------------------------------------------

=head3 C<mapCase()>

=cut

sub mapCase {
  my ($negraCase) = @_;

  return "bot" if $negraCase eq "*";
  return "acc" if $negraCase eq "Akk";
  return "dat" if $negraCase eq "Dat";
  return "nom" if $negraCase eq "Nom";
  return "gen" if $negraCase eq "Gen";

  # never reach
  print "WARNING: unknown negra case $negraCase\n";
}

#------------------------------------------------------------------------------

=head3 C<mapGender()>

=cut

sub mapGender {
  my ($item) = @_;
  $item =~ tr/A-Z/a-z/;

  if ($item eq "*") {
    $item = "bot";
  }
  return $item;
}

#------------------------------------------------------------------------------

=head3 C<mapNumber()>

=cut

sub mapNumber {
  my ($item) = @_;
  $item =~ tr/A-Z/a-z/;
  return $item;
}

#------------------------------------------------------------------------------

=head3 C<mapPerson()>

=cut

sub mapPerson {
  my ($item) = @_;

  return "first" if $item == 1;
  return "second" if $item == 2;
  return "third" if $item == 3;
}

#------------------------------------------------------------------------------

=head3 C<mapTense()>

=cut

sub mapTense {
  my ($item) = @_;
  $item = $item eq "Pres"? "present":"past";
  return $item
}
#------------------------------------------------------------------------------

=head3 C<mapDegree()>

=cut

sub mapDegree {
  my ($item) = @_;
  $item = $item eq "Comp"? "comparative": $item eq "Sup"? "superlative": "positive";
  return $item;
}

#------------------------------------------------------------------------------

=head3 C<mapMood()>

=cut

sub mapMood {
  my ($item) = @_;
  $item = $item eq "Ind"? "indicative": "subjunctive";
  # no distinction between subjunctive1 and 2
  return $item
}

#------------------------------------------------------------------------------

=head3 C<mapDefinite()>

=cut

sub mapDefinite {
  my ($item) = @_;
  $item = $item eq "Def"? "yes": "no";
  return $item
}

#------------------------------------------------------------------------------

=head3 C<mapFlexion()>

=cut

sub mapFlexion {
  my ($item) = @_;
  if ($item eq "*") {
    $item = "bot";
  } else {
    $item = $item eq "St"? "strong": $item eq "Sw"? "weak": "mixed";
  }
  return $item;
}

#------------------------------------------------------------------------------

=head3 C<isIncluded()>

check if a number is included in a list of intervals, e.g. 1-2,5-10,12,18-

=cut

sub isIncluded {
  my ($x, @intervals) = @_;

  foreach my $interval (@intervals) {
    if ( $interval =~ /^[0-9]+$/ ) {
      return 1 if $interval == $x;
    } else {
      my ($from, $to) = split("-", $interval);
      $from = 0 if $from eq "";
      $to = $x+1 if $to eq "";
      if ($from <= $x && $x <= $to) {
	return 1;
      }
    }
  }

  return 0;
}

#------------------------------------------------------------------------------

=head3 C<getFeatures()>

=cut

sub getFeatures {
  my ($word, $nodeTag, $morphTag) = @_;
  my $features; # the result
  my @intervals = (); # for the security check
  my $genderPattern = qr'(?:Masc)|(?:Fem)|(?:Neut)|\*';
  my $flexionPattern = qr'(?:St)|(?:Sw)|(?:Mix)|\*';
  my $degreePattern = qr'(?:Comp)|(?:Sup)|(?:Pos)|\*';
  my $casePattern = qr'(?:Nom)|(?:Akk)|(?:Dat)|(?:Gen)';
  my $numberPattern = qr'(?:Sg)|(?:Pl)';
  my $personPattern = qr'1|2|3';
  my $tensePattern = qr'(?:Pres)|(?:Past)';
  my $moodPattern = qr'(?:Ind)|(?:Konj)';
  my $definitePattern = qr'(?:Def)|(?:Indef)';

  # generate lexical features from all possible morphtags
  if ($morphTag eq "--") {
    if ($callbacks{getFeatures}) {
      $features=$callbacks{getFeatures}->($word,$nodeTag,$morphTag);
    }
  }
  elsif ( $morphTag =~ /^($degreePattern)\.($genderPattern)\.($casePattern)\.($numberPattern)\.($flexionPattern)$/) {
    $features = {
      "degree" => &mapDegree($1),
      "gender" => &mapGender($2),
      "case" => &mapCase($3),
      "number" => &mapNumber($4),
      "flexion" => &mapFlexion($5),
    };
    @intervals = ("4-11","23-31","37-52","112-144","245-292","295-325");
  }

  elsif ( $morphTag =~ /^($genderPattern)\.($casePattern)\.($numberPattern)\.($flexionPattern)$/ ) {
    $features = {
      "gender" => &mapGender($1),
      "case"  => &mapCase($2),
      "number" => &mapNumber($3),
    };
    if ( $4 ne "*" ) {
      $features->{flexion} =  &mapFlexion($4);
    }
    @intervals = ("13-15",17,"19-21",33,34,35,"54-57", 165,167,168,170,
		  "172-175","177-183",198,199,"201-203", 205,206,208,209,211,"213-216",
		  "218-222", "224-227", "229-232", "234-237", "239-241");
  }

  elsif ( $morphTag =~ /^($degreePattern)\.($genderPattern)\.\*.\*\.\*$/ ) {
    $features = {
      "degree" => &mapDegree($1),
      "gender" => &mapGender($2),
      "case" => "bot",
      "number" => "bot",
      "flexion" => "bot",
    };
    @intervals = (244);
  }

  elsif ( $morphTag =~ /^($genderPattern)\.($casePattern)\.($numberPattern)$/ ) {
    $features = {
      "gender" => &mapGender($1),
      "case" => &mapCase($2),
      "number" => &mapNumber($3),
    };
    @intervals = (12, 16, 18, 22, 32, 36, 53, 58, 166, 169, 171, 176, 180, 197, 200,
		  204, 207, 210, 212, 217, 223, 228, 233, 238);
  }

  elsif ( $morphTag =~ /^($personPattern)\.($casePattern)\.($numberPattern)$/ ) {
    $features = {
      "person" => &mapPerson($1),
      "case" => &mapCase($2),
      "number" => &mapNumber($3),
    };
    @intervals = ("59-62","76-77","86-89");
  }

  elsif ( $morphTag =~ /^($personPattern)\.($numberPattern)\.($tensePattern)\.($moodPattern)$/ ) {
    $features = {
      "person" => &mapPerson($1),
      "number" => &mapNumber($2),
      "tense" => &mapTense($3),
      "mood" => &mapMood($4),
    };
    @intervals = ("66-68", "72-75", 79, "83-85", "93-96", "105-108");
  }

  elsif ( $morphTag =~ /^($personPattern)\.($numberPattern)\.($genderPattern)\.($casePattern)$/ ) {
    $features = {
      "person" => &mapPerson($1),
      "number" => &mapNumber($2),
      "gender" => &mapGender($3),
      "case" => &mapCase($4),
    };
    @intervals = ("97-104", "63-65", "69-71",78, "80-82", "90-92");
  }

  elsif ( $morphTag =~ /^($casePattern)$/ ) {
    $features = {
      "case" => &mapCase($1),
    };
    @intervals = (109,145,184,330);
  }

  elsif ( $morphTag =~ /^($degreePattern)$/ ) {
    $features = {
      "degree" => &mapDegree($1),
    };
    @intervals = (111, 243, 294);
  }

  elsif ( $morphTag =~ /^($numberPattern)$/ ) {
    $features = {
      "number" => &mapNumber($1),
    };
    @intervals = (242,293);
  }

  elsif ( $morphTag =~ /^($casePattern)\.($genderPattern)$/ ) {
    $features = {
      "case" => &mapCase($1),
      "gender" => &mapGender($2),
    };
    @intervals = (110, "146-148");
  }

  elsif ( $morphTag =~ /^($definitePattern)\.($genderPattern)\.($casePattern)\.($numberPattern)$/ ) {
    $features = {
      "gender" => &mapGender($2),
      "case" => &mapCase($3),
      "number" => &mapNumber($4),
      "definite" => &mapDefinite($1),
    };
    @intervals = ("149-164", "185-196");
  }

  elsif ( $morphTag =~ /^($casePattern)-($casePattern)$/ ) {
    $features = {
      "case" => "(" . &mapCase($1) . "|" . &mapCase($2) . ")",
    };
    @intervals = (326,327,329);
  }

  elsif ( $morphTag =~ /^($casePattern)-($casePattern)-($casePattern)$/ ) {
    $features = {
      "case" => "(" . &mapCase($1) . "|" . &mapCase($2) . "|" . &mapCase($3) . ")",
    };
    @intervals = (328);
  }

  # never reach
  else {
    print STDERR "\nWARNING, THIS SHOULD NEVER HAPPEN: cannot match $nodeTag for $word, $morphTag. \n";
  }

  # common features
  $features->{cat} = $nodeTag;

  # security check of the morphtag being analysed correctly
  if ($#intervals > 0 && !&isIncluded($morphTags{$morphTag}, @intervals)) {
    printf STDERR "ERROR: false classification of %s(%d) in %s\n",
      $morphTag, $morphTags{$morphTag}, join ",", @intervals;
  }

  # 1. escape possible quotes in a feature entry
  # 2. remove `flexion' feature from nouns
  # 3. delete features defined to be "bot"
  foreach my $key (keys %{$features}) {
    if ($features->{$key} eq "bot") {
      delete $features->{$key};
    }
    elsif ($features->{cat} eq 'NN' && $key eq 'flexion') {
#      debug "Removing `flexion' from `$word'\n";
      delete $features->{flexion};
    }
    else {
      #$features->{$key} =~ s/([\'\"])/\\$1/g
    }
  }

  return $features;
}

#------------------------------------------------------------------------------

=head3 C<writeLexiconEntry()>

=cut

sub writeLexiconEntry {
  return if ! $lexiconFile;

  my $node = shift;

  # callback
  if ($callbacks{writeLexiconEntry}) {
    return if ! $callbacks{writeLexiconEntry}->($node);
  }

  my $features =
    &getFeatures($node->{word}, $node->{nodeTag}, $node->{morphTag});

  printf LEX "%20s := [ ", &escapeIdentifier($node->{word});

  my $isFirst = 1;
  foreach my $key (sort keys %{$features}) {
    printf LEX "%s%s:%s",
      $isFirst?" ":", " , $key, &escapeIdentifier($features->{$key});
    $isFirst = 0 if $isFirst;
  }

  print LEX " ];\n";
}

#------------------------------------------------------------------------------

=head3 C<writeLevelDecl()>

This function creates the level declaration if you specified the B<-levels>
I<filename>. By default one C<SYN> level is constructed marked to be the
C<mainlevel>. It declares all labels that have been found during tree conversion.
(see L<%labels>).

=cut

sub writeLevelDecl {
  return if ! $levelDclFile;

  print LEVEL "SYN [mainlevel=yes] # ";
  my @keys = sort keys %labels;
  my $isFirst = 1;
  while (@keys) {
    my $key = shift @keys;
    if ($isFirst) {
      $isFirst = 0;
    } else {
      print LEVEL ", ";
    }
    print LEVEL "$key";
  }
  print LEVEL ";\n";
}



# ----------------------------------------------------------------------

=head3 C<isLeaf()>

Return true if a node in a phrase tree is a leaf node.

=cut

sub isLeaf {
  my $n = shift;
  $$n{id} < 500 && $$n{id} > -1
}

# ----------------------------------------------------------------------

=head3 C<formatPhraseTree()>

Format a phrase tree like this:
[S [NP das][VP ist [ADJP gut]]]
(but with more newlines)

=cut

sub formatPhraseTree;
sub formatPhraseTree {
  my $tree = shift;
  my $level = shift;
  my $result;

  $result .= ' ' x $level;
  $result .= "$$tree{edgeTag}";

  if(isLeaf $tree) {
    $result .= "[$$tree{id} $$tree{nodeTag} ";
    $result .= $$tree{word};
  } else {
    $result .= "[$$tree{id}($$tree{min}-$$tree{max}) $$tree{nodeTag}\n";
  }

  my $nl = 0;
  for(@{$$tree{children}}) {
    $nl = 1;
    $result .= formatPhraseTree $_, $level+2;
  }
  $result .= ' ' x $level if $nl;
  $result .= "]\n";

  $result =~ s/\n\s+\]/\]/g;

  return $result;

}

# ----------------------------------------------------------------------

=head3 C<printPhraseTree()>

Print bracketed representation of phrase tree on STDERR.

=cut

sub printPhraseTree {
  print STDERR formatPhraseTree(shift, 0);
}


# ----------------------------------------------------------------------

=head3 C<computeExtents()>

Compute the linear extent of each constituent
and propagate the info to $$x{min} and $${max} for each node.

Leaf nodes have min == max == id. Inner nodes have the smallest
interval that encloses all leaves. This means that interleaved
constituents have extents that partially overlap.

=cut

sub computeExtents;
sub computeExtents {
  my $tree = shift;

  if(exists $$tree{word} && $$tree{word} ne 'NIL') {
    $$tree{min} = $$tree{id};
    $$tree{max} = $$tree{id};
  } else {
    $$tree{min} = 999;
    $$tree{max} = -1;
    for my $c (@{$$tree{children}}) {
      computeExtents $c;
      if($$c{min} < $$tree{min}) {
        $$tree{min} = $$c{min};
      }
      if($$c{max} > $$tree{max}) {
        $$tree{max} = $$c{max};
      }
    }
  }
}



# ----------------------------------------------------------------------

=head3 C<insertInnerNode()>

Insert a new inner node for TREE under PARENT with the values
specified in ATTS.

=cut

sub insertInnerNode {
  my $tree = shift;
  my $parent = shift;
  my $atts = shift;

  my $maxid = $#{$$tree{inner}};
  my $result =
    {
     id => ++$maxid + 500,
     nodeTag => '??',
     edgeTag => '??',
     children => [],
     parent => $parent,
    };

  # enter specified attributes
  while (my($k,$v) = each %$atts) {
    $$result{$k} = $v;
  }

  computeExtents $result;

  $$tree{inner}[$maxid] = $result;

  # recompute parent's children list
  my @c = @{$$parent{children}};
  push @c, $result;
  $$parent{children} = [sort {$$a{min} <=> $$b{min}} @c];
  computeExtents $tree;

  return $result;
}

# ----------------------------------------------------------------------

=head3 C<spliceWords>

Manipulate phrase tree $tree so that leaves $x through $y
turn into a new word with the specified parameters.

=cut

sub spliceWords {
  my ($n, $x, $y, $att) = @_;

  # build new leaf
  my $new =
    { min => $x,
      max => $x,
      id => $x,
      nodeTag => '??',
      edgeTag => '??',
      parent => $n };

  # enter specified attributes
  while(my($k,$v) = each %{$att}) {
    $$new{$k} = $v;
  }

  # copy features from rightmost subword if unspecified
  if(!exists($$new{features})) {
    $$new{features} = $$rootNode{leaves}[$y]{features};
  }

  debug "Spliced new word `$$new{word}'";

  # recompute children of $n
  my @new_children;
  for my $c (@{$$n{children}}) {
    if($$c{id} < $x || $$c{id} > $y) {
      push @new_children, $c;
    }
  }
  push @new_children, $new;
  $$n{children} = [sort { $$a{min} <=> $$b{min} } @new_children];

  # fix leaf array
  my @leaves;
  my $gap = $y - $x;
  for my $l (@{$$rootNode{leaves}}) {
    next if $$l{id} >= $x && $$l{id} <= $y;
    if($$l{id} > $$new{id}) {
      $$l{id} -= $gap;
    }
    $leaves[$$l{id}] = $l;
  }
  $leaves[$x] = $new;
  $$rootNode{leaves} = [@leaves];

  # fix extents
  computeExtents $rootNode;
}

# ----------------------------------------------------------------------

=head3 C<splitWord()>

Manipulate phrase tree $tree so that the leaf $w
turns into $n new words.

=cut

sub splitWord {
  my ($w, $n) = @_;
  my @result;
  my $p = $$w{parent};

  for(1..$n) {
    my $no = $$w{id}+$_-1;
    # build new leaf
    my $new =
      { min => $no,
        max => $no,
        id => $no,
        nodeTag => '??',
        edgeTag => '??',
        parent => $p,
        word => $$w{word}
      };

    push @result, $new;
  }

  # recompute children of $n
  my @new_children;
  for my $c (@{$$p{children}}) {
    if($$c{id} == $$w{id}) {
      push @new_children, @result;
    } else {
      push @new_children, $c;
    }
  }
  $$p{children} = [@new_children];

  # fix leaf array
  my @leaves;
  my $gap = $n-1;
  for my $l (@{$$rootNode{leaves}}) {
    if($$l{id} > $$w{id}) {
      $$l{id} += $gap;
    }
    $leaves[$$l{id}] = $l;
  }
  for(@result) {
    $leaves[$$_{id}] = $_;
  }
  $$rootNode{leaves} = [@leaves];

  # fix extents
  computeExtents $rootNode;

  return @result;
}


# ----------------------------------------------------------------------

=head3 C<moveNode()>

Manipulate phrase tree $tree so that $node is no longer a child of
its former parent, but a child of $new_parent, all parent/child
relations are consistent again, extents are updated, and everybody
is happy.

=cut

sub moveNode {
  my($node, $new_parent) = @_;
  my $old_parent = $$node{parent};

  debug
    "Move node #$$node{id} " .
    "from #$$old_parent{id} ($$old_parent{nodeTag}) ".
    "to #$$new_parent{id} ($$new_parent{nodeTag})";

  # remove
  my @new_children;
  for my $x(@{$$old_parent{children}}) {
    if($$x{id} != $$node{id}) {
      push @new_children, $x;
    }
  }
  $$old_parent{children} = [@new_children];

  # move
  $$node{parent} = $new_parent;

  # insert
  @new_children = @{$$new_parent{children}};
  push @new_children, $node;
  $$new_parent{children} = [ sort { $$a{min} <=> $$b{min} } @new_children ];

  computeExtents $new_parent;
  computeExtents $old_parent;
}

#----------------------------------------------------------------------

=head3 C<deleteNode()>

Manipulate phrase tree $tree so that $node no longer exists.

=cut

sub deleteNode {
  my $n = shift;
  my $p = $$n{parent};

  my @c;
  for my $x(@{$$p{children}}) {
    if($$x{id} != $$n{id}) {
      push @c, $x;
    }
  }
  $$p{children} = [@c];
  computeExtents $p;
}


# ----------------------------------------------------------------------

=head3 C<moveEdge()>

Manipulate a dependency tree so that edge $e points to $p, and
the ->children arrays are consistent.

=cut

sub moveEdge {
  my ($e,$p) = @_;
  my $o = $$e[2];

  $$e[2] = $p;
  push @{$$p{children}}, $$e[0];

  my @new_children = ();
  for(@{$$o{children}}) {
    push @new_children, $_ if $_ != $$e[0];
  }
  $$o{children} = [@new_children];
}


#------------------------------------------------------------------------------

=head3 C<makePhraseTree()>

This function is used by readNegraTreebank() to construct a nested
phrase data structure from a linear sentence hash.

=cut

sub makePhraseTree {
  our ($sentence) = @_; # used by inner function getChildren
  our @leaves = ();
  our @inner = ();

  #---------------------------------------------------------------------------
  sub getChildren {
    my ($parentId) = @_;
    my @children;
    our $sentence;

    my @ids = keys %{$sentence};
    while (@ids) {
      my $id = shift @ids;
      next if $parentId != $sentence->{$id}{parent};

      my $node = {
	"nodeTag" => $sentence->{$id}{nodeTag},
	"edgeTag" => $sentence->{$id}{edgeTag},
	"morphTag" => $sentence->{$id}{morphTag},
	"id" => $id,
        word => 'NIL'
      };
      if ($id < 500) {
	$node->{word} = $sentence->{$id}{word};
	$node->{features} = &getFeatures($node->{word}, $node->{nodeTag},
	  $sentence->{$id}{morphTag});
        $leaves[$id] = $node;
      } else {
	my @thisChildren = &getChildren($id);
	if (! @thisChildren) {
	  #print STDERR "WARNING: removing empty node $node->{id} $node->{nodeTag}\n";
	  next;
	}

	$node->{children} = [@thisChildren];
	foreach my $child (@{$node->{children}}) {
	  $child->{parent} = $node;
	}
        $inner[$id-500] = $node;
      }
      push @children, $node;
    }

    return sort { &getMinId($a) <=> &getMinId($b) } @children;
  }
  #---------------------------------------------------------------------------

  # set the global root node of the constructed tree
  $rootNode =
    { "id" => -1,
      "isLexHead" => 1,
      "nodeTag" => "ROOT",
      "edgeTag" => "ROOT",
      "word" => "NIL",
      "morphTag" => "NIL",
      "children" => [&getChildren(0)],
      leaves => [@leaves],
      inner => [@inner]
    };
  foreach my $child (@{$rootNode->{children}}) {
    $child->{parent} = $rootNode;
  }

  computeExtents $rootNode;

}

#------------------------------------------------------------------------------

=head3 C<convertDepTree()>

manipulate a dependency tree

=cut

sub convertDepTree {
  return if ! %depConvTable;

  my $deps = shift;
  my $extradeps = shift;

  foreach my $dep (@$deps) {
    my ($modifier, $label, $modifiee) = @$dep;
    my $conversions = $depConvTable{$label};
    if (defined $conversions) {
      foreach my $conversion (@$conversions) {
	debug "at $modifier->{word}($modifier->{id})--$label-->$modifiee->{word}($modifiee->{id})\n" ;
	if($conversion->($dep, $deps, $extradeps)) {
	  &depSnapShot($deps, $extradeps);
	}
      }
    }
  }
}

#------------------------------------------------------------------------------

=head3 C<convertPhraseTree()>

manipulate an annotated phrase tree

=cut

sub convertPhraseTree {
  return if ! %phraseConvTable;

  my ($node, $id) = @_;
  return if $node->{isConverted};# every node is processed only once here

  my $conversions = $phraseConvTable{$node->{nodeTag}};
  if (defined $conversions) {
    foreach my $conversion (@$conversions) {
      if ($conversion->($node)) {
	$node->{isConverted} = 1;
	&phraseSnapShot($rootNode);
      }
    }
  }
  foreach my $child (@{$node->{children}}) {
    &convertPhraseTree($child);
  }
}

#------------------------------------------------------------------------------

=head3 C<markLexHeads()>

=cut

sub markLexHeads {
  my ($node) = @_;

  my $headChild = &findHeadChild($node);
  foreach my $child (@{$node->{children}}) {
    $child->{isLexHead} = (defined $headChild &&
      $child->{id} == $headChild->{id} &&
      $child->{id} < 500) ? 1:0;
    &markLexHeads($child);
  }
}

#------------------------------------------------------------------------------

=head3 C<getMinId()>

=cut

sub getMinId {
  my (@nodes) = @_;
  my $minId = 9999;

  foreach my $node (@nodes) {
    my $nodeId = ($node->{id} < 500) ?
      $node->{id}: &getMinId(@{$node->{children}});
    if ($nodeId < $minId) {
      $minId = $nodeId;
    }
  }

  return $minId;
}

#------------------------------------------------------------------------------

=head3 C<getAllNodes()>

=cut

sub getAllNodes {
  my ($nodes) = @_;
  my @allNodes;

  foreach my $node (@$nodes) {
    push @allNodes, $node;
    if ($node->{id} >= 500 || $node->{id} < 0) {
      push @allNodes, @{&getAllNodes($node->{children})};
    }
  }

  return \@allNodes;
}

#------------------------------------------------------------------------------

=head3 C<writePhraseTree()>

=cut

sub writePhraseTree {
  my ($node, $indent) = @_;
  $indent = 0 if ! defined $indent;

  # write a root node and all its children
  if ($node->{id} == -1) {
    printf STDERR " %s(\n", " " x $indent;
    foreach my $child (@{$node->{children}}) {
      &writePhraseTree($child, $indent+3);
    }
    printf STDERR " %s)\n", " " x $indent;
  }

  # write an inner node  and all its children
  elsif ($node->{id} >= 500) {
    printf STDERR "%03d: %s(%s%s%s%s%s\n",
      $node->{id},
      " " x $indent,
      $node->{nodeTag},
      $node->{edgeTag} eq "--"?"": "-$node->{edgeTag}",
      $node->{coref}?"=$node->{coref}->{id}":"",
      ($node->{backRef} || $node->{backTrace})?"-$node->{id}":"",
      $node->{headChildFlag}?" $node->{headChildFlag}":"";


    foreach my $child (@{$node->{children}}) {
      &writePhraseTree($child,
	$indent+length($node->{nodeTag})+ length($node->{edgeTag}) + 3);
    }
    printf STDERR "     %s)\n", " " x $indent;
  }

  # write a leafnode
  else {
    printf STDERR "%03d: %s(%s%s %s) %s\n",
      $node->{id},
      " " x $indent,
      $node->{nodeTag},
      $node->{edgeTag} eq "--"?"": "-$node->{edgeTag}",
      $node->{trace}?"$node->{word}-$node->{trace}->{id}":$node->{word},
      $node->{isLexHead}?"*":"";
  }
}

#------------------------------------------------------------------------------

=head3 C<makeAllDeps()>

wrapper for makeDeps:
Magerman/Lin differences:
- they dont construct NIL dependencies
- they can handle only a single rooted phrase tree / dependency tree
- we can optionally construct more than one level of description

=cut

sub makeAllDeps {
  my ($rootNode) = @_;
  my @deps;

  foreach my $node (@{$rootNode->{children}}) {
    if ($node->{id} < 500) {
      push @deps, &makeDependency($node, $rootNode);
    } else {
      &makeDeps($node, \@deps); # has a side effect on deps
    }
  }

  # check if all words have edges
  LEAF:
  for my $l (@{$$rootNode{leaves}}) {
    for my $dep (@deps) {
      my $x = $$dep[0]{id};
      my $y = $$l{id};
      next LEAF if $$dep[0]{id} == $$l{id};
    }
    debug "no dependency for word $$l{word} ($$l{id})\n" ;
    push @deps, &makeDependency($l, $rootNode);
  }

  # delete all former children
  foreach my $dep (@deps) {
    my ($modifier, $label, $modifiee) = @$dep;
    $modifier->{children} = undef;
  }

  # fix parent-child relation in the dependency tree
  foreach my $dep (@deps) {
    my ($modifier, $label, $modifiee) = @$dep;
    $modifier->{dep} = $dep;
    $modifier->{parent} = $modifiee;
    push @{$modifiee->{children}}, $modifier;
  }

  return sort { $a->[0]{id} <=> $b->[0]{id} } @deps;
}

#------------------------------------------------------------------------------

=head3 C<makeDeps()>

see (Magermann 1994, p.64-66) and (Lin 1995)
returns the lexical head of the analysed node
as a side effect new dependency edges are asserted into deps

=cut

sub makeDeps {
  my ($node, $deps) = @_;

  # termination of the recursion: a leaf node is its own lexical head
  if ($node->{id} < 500) {
    return $node;
  }

  my $headChild = &findHeadChild($node);
  return $rootNode unless $headChild;
  my $lexHead = &makeDeps($headChild, $deps);

  foreach my $child (@{$node->{children}}) {
    next if $child->{id} == $headChild->{id};
    my $lexHeadOfChild = &makeDeps($child, $deps);
    if ($lexHeadOfChild->{id} != -1) {
      push @{$deps}, &makeDependency($lexHeadOfChild, $lexHead);
    }
  }


  return $lexHead;
}

#------------------------------------------------------------------------------

=head3 C<makeDependency()>

=cut
our @label_rule_application;
our @label_rule_sentences;

sub makeDependency {
  my ($modifier, $modifiee) = @_;
  my $domiNode = &getDominator($modifier, $modifiee);
  my $modifierKey = "$modifier->{nodeTag}-$modifier->{edgeTag}";
  my $modifieeKey = "$modifiee->{nodeTag}-$modifiee->{edgeTag}";
  my $domiNodeKey = $domiNode?"$domiNode->{nodeTag}-$domiNode->{edgeTag}":"";
  my $label;

  debug "`$modifier->{word}':";
  debug "  translate triple ($modifierKey,$modifieeKey,$domiNodeKey)";

  for (my $i = 0; $i <= $#labelTable; $i++) {
    my ($thisLabel, $modifierPattern, $modifieePattern, $parentPattern) =
      @{$labelTable[$i]};
    if ($modifierKey =~ $modifierPattern &&
	$modifieeKey =~ $modifieePattern &&
	$domiNodeKey =~ $parentPattern) {
      $label = $thisLabel;
      debug "  chose label `$label' (rule `$modifierPattern/$modifieePattern/$parentPattern\')\n";
      $label_rule_application[$i]++;
      if($label_rule_sentences[$i] !~ /s$sentenceIndex$/) {
        $label_rule_sentences[$i] .= " s$sentenceIndex";
      }
      last;
    }
  }

  if (! defined $label) {
    $nolabelWarning = 1;
    printf STDERR "WARNING: $nrSentences - no label for $modifier->{word}:<%s> $modifiee->{word}:<%s> %s\n",
      $modifierKey, $modifieeKey, $domiNode?"<$domiNodeKey>":"";# if $debug;

#    if ($modifiee->{id} == -1) {
#      $label = "ROOT";
#    } elsif ($modifier->{isLexHead}) {
#      if ($modifier->{parent}->{id} == -1) {
#	$label = "ROOT";
#      } else {
#	#$label = $modifier->{parent}->{edgeTag};
#	$label = $domiNode->{edgeTag};
#      }
#    } else {
#      $label = $modifier->{edgeTag};
#    }

    $label = "--";
  }

  $labels{$label} = 1 if $label !~ /^_/;
  return [ $modifier, $label, $modifiee ];
}

#------------------------------------------------------------------------------

=head3 C<getDominator()>

given two nodes n1 and n2 of a phrase tree
return the ancestor of n1 which dominates n2

=cut

sub getDominator {
  my ($node1, $node2) = @_;
  my $result;

  for (my $node = $node2; defined $node; $node=$node->{parent}) {
    $node->{flag} = 1;
  }

  for (my $node = $node1->{parent}, my  $prevNode = $node1;
       defined $node; $prevNode = $node, $node=$node->{parent}) {
    if ($node->{flag}) {
      $result = $prevNode;
      last;
    }
  }

  for (my $node = $node2; defined $node; $node=$node->{parent}) {
    undef $node->{flag};
  }

  return $result
}

#------------------------------------------------------------------------------

=head3 C<findHeadChild()>

=cut

sub findHeadChild {
  my ($node) = @_;

  # if there are no children (which should not happen but
  # might, as a result of a rogue phrase conversion rule), return
  # undef to show there is something wrong.

  if($#{$node->{children}} == -1) {
    return undef;
  }

  if ($node->{id} < 500) {
    $node->{headChildFlag} = "!!";
    return $node;
  }

  # security check
  if (! %headTable || ! defined $headTable{$node->{nodeTag}}) {
    print STDERR "WARNING: no head table entry for node $node->{nodeTag}\n";
    return undef;
  }

  # get entry in head table for this node
  my ($direction, @headList) = @{$headTable{$node->{nodeTag}}};

  # try all head patterns
  foreach my $headPattern (@headList) {

    # try all children of this node in the specified order:
    # l = left-to-right, r = right-to-left (reverse)
    my @children =
      ($direction eq "l") ?
        @{$node->{children}} :
          reverse @{$node->{children}};

    foreach my $child (@children) {
      my $combinedTag = "$child->{nodeTag}-$child->{edgeTag}";
      if ($combinedTag =~ /$headPattern/ ) {
	$child->{headChildFlag} = "!!";
	return $child;
      }
    }
  }

  # if no specified child can be found, ....
  # - the first child in left-to-right mode
  # - the last child in right-to-left mode
  my $child = $direction eq "l"?  $node->{children}->[0]: $node->{children}->[-1];
  print STDERR "%% WARNING: fallback to $child->{id} for node $node->{id} "
    . "($node->{nodeTag}-$node->{edgeTag}) in sentence s$nrSentences\n"
    if $debug;
  $child->{headChildFlag} = "??";
  return $child;
}


#------------------------------------------------------------------------------
=head3 C<formatCdgAnnotation()>

returns a dependency annotation formated in CDG output format

=cut

sub formatCdgAnnotation {
  my ($deps, $extradeps) = @_;
  my $result = '';
  my %done;

  # print cdg annotation
  $result .=  "'$annoPrefix$sentenceIndex' : '$wgPrefix$sentenceIndex' <->\n";
  foreach my $dependency (sort {$$a[0]{id} <=> $$b[0]{id}}  @$deps) {

    my ($modifier, $label, $modifiee) = @$dependency;

    if($done{$$modifier{id}}++) {
      print STDERR "WARNING: s$sentenceIndex: we already did word # $$modifier{id}\n";
      next;
    }

    $result .= sprintf "%2d %2d %20s SYN -> %-10s -> %2d %-26s ",
      $modifier->{id}, $modifier->{id} +1,
      &escapeIdentifier($modifier->{word}),
      &escapeIdentifier($label),
      $modifiee->{id}+1,
      $modifiee->{id} == -1 ? "":
	"/* " . $modifiee->{word} . " */";
    foreach my $key (sort keys %{$modifier->{features}}) {
      $result .= sprintf " %s/%s", &escapeIdentifier($key),
          &escapeIdentifier($modifier->{features}{$key});
    }
    if(@extralevels) {
      for my $l (@extralevels) {
        my $no = $$modifier{id};

        # default to the empty string as a label
        my $label = $$extradeps{$l}[$no]{label} || '';

        my $up = $$extradeps{$l}[$no]{up};
        if(!$up) {
          $up = 0;
        }
        $result .= " $l -> "
	. &escapeIdentifier($label)
	. " -> $up ";
      }
    }
    $result .= ",\n";
  }

  $result =~ s/,\n$/;\n/;
  $result .= "\n";
  return $result;

}

#------------------------------------------------------------------------------
=head3 C<formatPlainAnnotation()>

returns a dependency annotation formated in PLAIN output format

=cut

sub formatPlainAnnotation {
  my ($deps, $extradeps) = @_;
  my $result = '';
  my %done;

  # print cdg annotation
  $result .=  "BOA $annoPrefix$sentenceIndex $wgPrefix$sentenceIndex\n";
  foreach my $dependency (sort {$$a[0]{id} <=> $$b[0]{id}}  @$deps) {

    my ($modifier, $label, $modifiee) = @$dependency;

    if($done{$$modifier{id}}++) {
      print STDERR "WARNING: we already did word # $$modifier{id}\n";
      next;
    }

    $result .= sprintf "%03d %-20s ",
      $modifier->{id}+1, &escapeIdentifier($modifier->{word});

    foreach my $key (sort keys %{$modifier->{features}}) {
      $result .= sprintf "%-10s ", &escapeIdentifier($modifier->{features}{$key});
    }

    $result .= sprintf "%-10s %03d ",
      &escapeIdentifier($label), $modifiee->{id}+1;

    $result .= "\n";
  }

  $result .=  "EOA $annoPrefix$sentenceIndex\n\n";
  return $result;

}

#------------------------------------------------------------------------------
=head3 C<formatAnnotation()>

returns a dependency annotation formated in the given output format

=cut

sub formatAnnotation {

  if ($outputFormat eq "cdg") {
    return formatCdgAnnotation shift
  } 
  
  elsif ($outputFormat eq "plain") {
    return formatPlainAnnotation shift
  }

  else {
    die "PANIC: unknown output format '$outputFormat'";
  }
}


#------------------------------------------------------------------------------

=head3 C<writeAnnotation()>

=cut

sub writeAnnotation {
  return if ! $annoFile;
  my ($deps, $extradeps) = @_;

  # print cdg annotation
  print ANNO formatAnnotation($deps, $extradeps);
}

#------------------------------------------------------------------------------

=head3 C<formatNodes()>

=cut

sub formatNodes {
  my $root = shift;
  my $n = getAllNodes([$root]);
  my $t = '';

  for my $node (sort {$$a{id} <=> $$b{id} } @$n) {

    if ($$node{id} == -1 || $$node{id} >=500) {

      # describe inner node
      $t .= "\#$$node{id}\t\t\t";
      $t .= "$$node{nodeTag}\t\t";
      $t .= "--\t\t";
      $t .= "$$node{edgeTag}\t\t";
      my $p = $$node{parent}{id};
      $p = 0 unless defined $p;
      $p = 0 if $p<0;
      $t .= "$p\n";
    } else {

      # describe leaf
      $t .= "$$node{word}\t\t";
      $t .= "$$node{nodeTag}\t\t";
      $t .= $$node{morphTag}?"$$node{morphTag}\t\t":"--\t\t";
      $t .= "$$node{edgeTag}\t\t";
      my $p = $$node{parent}{id};
      $p = 0 if $p<0;
      $t .= "$p\n";
    }
  }
  return $t;
}


# lexically scoped counter for makeSnapShot()
{ my $snapcounter = 0;

#------------------------------------------------------------------------------

=head3 C<makeSnapShot()>

Create a PNG image of the current state of the phrase structure
tree. This is used strategically during phrase tree conversion
to leave a trail of how the transformation went.

=cut

sub phraseSnapShot {
  return unless $debug;

  my $root = shift;
  my $txt = formatNodes($root);
  my $filename = sprintf "s$sentenceIndex-stage%03d", $snapcounter;
  open(SNAP, ">$filename.export") or die "Can't write `>$filename.export': $!\n";
  print SNAP $txt;
  close(SNAP);
  $snapcounter++;
}

#------------------------------------------------------------------------------

=head3 C<depSnapShot()>

=cut

sub depSnapShot {
  return unless $debug;
  
  my $deps = shift;
  my $extradeps = shift;
  my $filename = sprintf "s$sentenceIndex-stage%03d", $snapcounter;
  open(SNAP, ">$filename.cda") or die "Can't write `>$filename.cda': $!\n";
  print SNAP formatAnnotation($deps, $extradeps);
  close(SNAP);
  $snapcounter++;
}
} # end of lexical scope

#------------------------------------------------------------------------------

=head3 C<main()>

This is the first function called comparable to the C<main()> function in C
and elsewhere. It reads all command line parameters, opens the configuration
files, creates the initially empty output files and loops over all input that
is given to C<Depsy>.

=cut

sub main {
  my $help = 0;
  my $man = 0;
  my $argRange;

  $help = 1 unless
    GetOptions(
      "annoprefix:s" => \$annoPrefix,
      "annos:s" => \$annoFile,
      "d|debug+" => \$debug,
      "debug-label-table+" => \$debug_label_table,
      "extralevels:s" => \$extralevels,
      "index:i" => \$sentenceIndex,
      "input-format:s" => \$inputFormat,
      "levels:s" => \$levelDclFile,
      "lexicon:s" => \$lexiconFile,
      "output-format|of:s" => \$outputFormat,
      "man" => \$man,
      "max|n:i" => \$maxSentence,
      "plugin:s" => \$pluginFile,
      "prefix:s" => \$wgPrefix,
      "range:s" => \$argRange,
      "syntax|h|help|?" => \$help,
      "wordgraphs:s" => \$wordgraphsFile,
    ) || pod2usage(2);

  pod2usage(-exitval =>1, -message => "\nThe DEPSY Conversion Tool\n") if $help;
  pod2usage(-verbose => 2) if $man;

  if ($argRange) {
    @sentenceIdRange = split(",", $argRange);
  }

  if($extralevels) {
    @extralevels = split(/,/, $extralevels);
  }


  # validate the inputFormat argument
  $inputFormat = "export" if $inputFormat eq "negra";
  die "ERROR: unknown input-format argument $inputFormat\n"
    if $inputFormat ne "export" && $inputFormat ne "penn";

  # validate the outputFormat argument
  die "ERROR: unknown output-format argument $outputFormat\n"
    if $outputFormat ne "cdg" && $outputFormat ne "plain";

  # read the plugin file
  if ($pluginFile) {
    debug "reading the plugin file $pluginFile";
    undef $/; # eval files at once
    open (PLUGIN, $pluginFile) or die "ERROR: cannot open $pluginFile\n";
    eval <PLUGIN>;
    die "ERROR: while reading $pluginFile $@\n" if $@;
    close PLUGIN;
    $/ = "\n";
    @label_rule_application = map {0} @labelTable;
    @label_rule_sentences = map {''} @labelTable;
    debug "reading plugin file done.";
  }

  # print out some file information
  sub printIntro {
    my ($handle, $commentString) = @_;
    $commentString = "//" if ! defined $commentString;
    print $handle
      $commentString . ' automatically generated by depsy.pl (c) 1997-2004' . "\n" .
      $commentString . ' $Id: depsy.pl,v 1.15 2006-05-10 16:02:04 kdaling Exp $ ' . "\n" .
      $commentString . ' contact: The CDG Team <cdg@nats.informatik.uni-hamburg.de>' . "\n\n";
  }

  debug "open output files.";
  if ($wordgraphsFile) {
    open(WG, ">$wordgraphsFile") or die "ERROR: cannot create $wordgraphsFile\n";
    &printIntro(*WG);
  }
  if ($lexiconFile) {
    open(LEX, ">$lexiconFile") or die "ERROR: cannot create $lexiconFile\n";
    &printIntro(*LEX);
  }
  if ($annoFile) {
    open(ANNO, ">$annoFile") or die "ERROR: cannot create $annoFile\n";
    &printIntro(*ANNO);
  }
  if ($levelDclFile) {
    open(LEVEL, ">$levelDclFile") or die "ERROR: cannot create $levelDclFile\n";
    &printIntro(*LEVEL);
  }

  # callback
  if ($callbacks{initialize}) {
    debug "initialize callbacks.";
    $callbacks{initialize}->();
  }

  # read treebank
  if ($inputFormat eq "export") {
    debug "read treebank in negra format.";
    &readNegraTreebank();
    debug "done reading treebank.";
  } elsif ($inputFormat eq "penn") {
    debug "read treebank in penn format.";
    &readPennTreebank();
    debug "done reading treebank.";
  }

  # write some left information gadgets
  &writeLevelDecl();

  print WG "// EOF\n" if $wordgraphsFile;
  print LEVEL "// EOF\n" if $levelDclFile;
  print LEX "// EOF\n" if $lexiconFile;
  print ANNO "// EOF\n" if $annoFile;

  # callback
  if ($callbacks{finalize}) {
    $callbacks{finalize}->();
  }

  debug "generated " . --$nrSentences . " annotation(s)\n" ;

  if($debug_label_table) {
    for (my $i = 0; $i <= $#labelTable; $i++) {
      printf STDERR "%6d applications: @{$labelTable[$i]}", $label_rule_application[$i];
      if ($label_rule_application[$i] > 0) {
        print STDERR "      ", substr($label_rule_sentences[$i], 0, 40);
      }
      print STDERR "\n";
    }
  }
}
&main();

=head1 AUTHORS

Michael Daum E<lt>micha@nats.informatik.uni-hamburg.deE<gt>

=head1 COPYRIGHT

Copyright (C) 1997-2004 The CDG Team E<lt>cdg@nats.informatik.uni-hamburg.deE<gt>

This file is free software; as a special exception the author gives
unlimited permission to copy and/or distribute it, with or without
modifications, as long as this notice is preserved.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

=cut
