#!/usr/bin/env perl
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# cdg/utils/deutsch-supertagger.pl
#
use warnings;
use strict;

use Getopt::Std;

BEGIN {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0);
  require 'cdgsetup.pl';
}
use CDG;
use CDG::Utils qw(:statistics :functions);

use vars qw($opt_m $opt_n $opt_c $opt_h $opt_p);
getopts('chm:n:p');

die "

Usage: $0

    -c             clean up inconsistent tags as described in mit2005
    -h             read this
    -m <STRING>    specify supertag model (default J)
    -n <NUM>       max. number of different pre-/post-modifiers (default 6)
    -p             read Perfect supertags from .cda file (implies -m J)

" if $opt_h;

my $in = "/tmp/ST.in.$$";
my $out = "/tmp/ST.out.$$";

our $width = $opt_n || 6;

our @output;

# Auto-correct mutually inconsistent tags.
#
# Example: if the prediction is
#
#  es      +EXPL/R+
#  mag     +S/N+AUX,SUBJC
#  sein    PRED+AUX/L+
#
# then it can't *possibly* be right, because it predicts both that
# `sein' has a preceding PRED, *and* that neither of the preceding
# words is a PRED. Rather than require the impossible, in this case we
# simply remove the offending context prediction.
sub clean_tags {
  my $from = shift;
  my $to = shift;

  my @labels;
  for($from..$to) {
    if($output[$_] =~ /OTHER/) {
      $labels[$_] = 'OTHER';
    } else {
      $output[$_] =~ /\+(.+)\/.\+/ or warn "Can't match $output[$_];\n";
      $labels[$_] = $1;
    }
  }
  for my $i ($from..$to) {
    $output[$i] =~ /^(.*)\t(.*)\+(.+)\+(.*)$/;
    my $word = $1;
    my @pre = split(/,/,$2);
    my $mid = $3;
    my @post = split(/,/,$4);
    my $old = join(',',@pre) . "+$mid+" . join(',',@post);

    my @newpre;
  L1:
    for my $label (@pre) {
      for($from..$i-1) {
        if($labels[$_] eq $label) {
          push @newpre, $label;
          next L1;
        }
      }
    }
    @pre = @newpre;

    my @newpost;
  L2:
    for my $label (@post) {
      for($i+1..$to) {
        if($labels[$_] eq $label) {
          push @newpost, $label;
          next L2;
        }
      }
    }
    @post = @newpost;

    my $new = join(',',@newpre) . "+$mid+" . join(',',@newpost);
    if($old ne $new) {
      warn "Correcting `$old' to `$new'\n";
      $output[$i] = "$word\t" . $new;
    }
  }
}

sub print_prediction {
  my $value = shift;
  my $name = shift;

  my $result = " $name ";
  if(!defined $value) {
    $result .= 'none'
  } else {
    $result .= $value
  }
  return $result
}


our @words;
open(IN, ">$in") or die "Can't write `>$in': $!\n";
our $name = '';
while(<>) {
  if(/^#sentence: (.+)/) {
    $name = $1;
    next;
  }
  print IN;
  chomp;
  push @words, $_;
}
close(IN);

if($opt_p) {

  if($opt_m) {
    warn "WARNING: -p implies -m J! Ignoring -m setting.\n";
  }

  # read data from annotation
  my ($corpus,$no) = $name =~ /^(.+)-s(\d+)$/;
  my $dir = sprintf "%03d", ($no / 10000);
  my $file = "/data/corpus/ccc/$corpus/$dir/$corpus-s$no.cda";
  die "Can't find anno `$file': $!\n" unless -r $file;

  commandEval 'set verbosity 0';
  my $ae = cda2ae $file;
  die "Can't read anno from $file\n" unless $ae;
  my $h = AnnoEntry2hash $ae;

  our (@pre, @self, @post);
  for(@{$$h{edges}}) {
    next unless $$_{level} eq 'SYN';

    my $reg = $$_{parent};
    $self[$$_{child}]{label} = $$_{label} || 'ROOT';

    if(0 == $reg) {
      $self[$$_{child}]{dir} = 'N';
    } elsif($$_{child} < $reg) {
      $self[$$_{child}]{dir} = 'R';
      $pre[$reg]{$$_{label}}++;
    } else {
      $self[$$_{child}]{dir} = 'L';
      $post[$reg]{$$_{label}}++;
    }
  }

  for(1..$#self) {
    my $tag = '';
    if(defined keys %{$pre[$_]}) {
      $tag = join(',', sort keys %{$pre[$_]});
    }
    $tag .= "+$self[$_]{label}/$self[$_]{dir}+";
    if(defined keys %{$post[$_]}) {
      $tag .= join(',', sort keys %{$post[$_]});
    }
    my $word = $words[$_-1];
    push @output, "$word\t$tag"
#    print STDERR "deduced tag $tag\n";

  }

} else {

  # Find model files
  my $model = $opt_m || 'J';

  my $dir;
  if (-r "$model.123") {
    $dir = '.';
  } elsif (exists $ENV{CDGDIR}) {
    if (-r "$ENV{CDGDIR}/lib/$model.123") {
      $dir = "$ENV{CDGDIR}/lib"
    } elsif (-r "$ENV{CDGDIR}/utils/$model.123") {
      $dir = "$ENV{CDGDIR}/utils"
    } else {
      my $path = $0;
      $path =~ s![^/]+$!!;
      if ( -r "$path/$model.123") {
        $dir = $path;
      }
    }
  }
  if (!$dir) {
    die "Can't find parameter file `$model.123'. It should be in \$CDGDIR/lib.\n";
  }

  # Call tnt+
  system "cd $dir && tnt+ $model $in $out" and die "tnt+ failed!\n";

  # transform output
  open(OUT, "$out") or die "Can't read `$out': $!\n";
  while (<OUT>) {
    chomp;
    push @output, $_;
    print STDERR "DEBUG: read `$_'\n";
  }

  if ($#output < $#words) {
    die "ERROR: $#words lines of input, $#output lines of output\n";
  }

  if ($opt_c) {
    clean_tags(0,$#words);
  }

}

our $line;
for(@words) {
  $_ = shift @output;
  $line = '';

  if(/^$/) {
    $line .= "\n";
  } elsif(/^(.+)\t+OTHER/) {

    # pose no conditions
    $line .= "$1 pre1 none label none dir none post1 none\n";

  } elsif (/^([^\t]+)\t+(.+)$/) {
    my($word,$st) = ($1,$2);
    my($pre,$label,$dir,$post) = ($st =~ /^(.*)\+(.+)\/([LNR])\+(.*)$/);
    my (@pre,@post);
    my $i=0;
    for(split(/,/, $pre)) {
      $pre[$i] = $_;
      $i++;
    }
    $i = 0;
    for(split(/,/, $post)) {
      $post[$i] = $_;
      $i++;
    }
    $line .= "$word";
    for(1..$width) {
      $line .= print_prediction($pre[$_ - 1],"pre$_");
    }
    $line .= " label $label dir $dir";
    for(1..$width) {
      $line .= print_prediction($post[$_ - 1],"post$_");
    }
    $line .= "\n";
  }
  print $line;
  print STDERR "DEBUG: $line";
}
close(OUT);

unlink $in;
unlink $out;
