#!/usr/bin/env perl
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

# Measure correctness of a tagger on an annotated corpus.
use warnings;
use strict;

use File::Spec;

use Getopt::Std;
use vars qw ($opt_j $opt_n $opt_r $opt_h);
getopts("hj:nr");

our @corpus;
our $jobname = $opt_j || undef;
our %stat;

my $tntdir = $ENV{TNTDIR} || '/opt/pkg/tnt';

# structure of defined taggers;
# you only need to give a command for invocation really.
our %taggers =
  (
   TreeTagger =>
   {
    invocation =>
    '/data/linux/opt/TreeTagger/bin/tree-tagger' .
    ' -quiet -token -threshold 0.1 -prob ' .
    '/data/linux/opt/TreeTagger/lib/papa-german.par'

   },

   tnt =>
   {
    invocation =>
    "$tntdir/tnt -v0 -b$tntdir/models/papa.lex -z200 $tntdir/models/negra.tnt -",
   },

   orig =>
   {
    invocation => "deutsch-tagger.pl -n "
   },

   papa =>
   {
    invocation => "deutsch-tagger.pl"
   },

   'papa-clean' =>
   {
    invocation => "deutsch-tagger.pl -m clean"
   },

   'papa-deutsch' =>
   {
    invocation => "deutsch-tagger.pl -m deutsch"
   }
  );

die "

Usage: $0 <tagger name> <corpus>

   -j: set Job name
   -n: use NEGRA export format rather than flat tagged format
   -r: Reuse previous `words' file

" if $opt_h or $#ARGV != 1;


our $tagger = shift;
die "Defined taggers: " . (join " ", keys %taggers) . "\n"
  unless exists $taggers{$tagger};

our %similar;
for (qw (
       FM:XY NN:XY ADJA:NN ADJA:NN ADJD:ADV ADJD:PTKVZ ADJD:VVPP
       ADV:PTKVZ CARD:NN FM:NE FM:NN NE:ADJA NE:FM NE:NN NE:XY ADJA:FM
       ADJA:NN ADJA:XY NN:CARD NN:FM NN:NE NN:PIS PIAT:PIDAT PIDAT:PIAT
       PRELS:PWS PWS:PRELS VMFIN:VVFIN VVPP:ADJD
        )) {
  $similar{$_} = 1;
}



sub interpret {
  my ($tuple, $line, $cat) = @_;
  while ($line =~ s/(\S+)\s+(\S+)\s+//) {
    my($c,$s) = ($1,$2);
    $tuple->{$c} = $s;
    if (!exists $tuple->{max} ||
        $$tuple{$$tuple{max}} < $s ||
        $$tuple{$$tuple{max}} == $s && $c eq $cat) {
      $tuple->{max} = $c
    }
  }
}

# remove quotes from a string
sub parse_string {
  my $s = shift;
  if ($s =~ /^([\'\"])(.*)\1$/) {
    $s = $2;
    $s =~ s|\\||g;
  }
  return $s;
}


sub read_cda {
  my $file = shift;
  my $last = 'cat';
  my %pair;
  my $id;
  open(IN, "$file") or die "Can't read `$file': $!\n";
  while(<IN>) {
    chomp;
    $id = $1 if /^(.+):.*<->/;
    if(/^ *\d+ +\d+ +(\S+).*\bcat\/(\S+)/) {
      my $word = parse_string $1;
      my $cat = parse_string $2;
      next unless $word;
      push @corpus, { word => $word, cat => $cat };
    } elsif (/^ *\d+ +\d+ +(.+)$/) {
      die "ERROR: In sentence $id: Expected category, got `$_'\n"
        if $last eq 'word';
      my $word = $1;
      $pair{word} = parse_string $word;
      $last = 'word';
    } elsif (/^ *\'?cat\'? *\/ *([^ ]+)/) {
      die "ERROR: Expected word, got `$_'\n"
        if $last eq 'cat';
      $pair{cat} = parse_string $1;
      $last = 'cat';

      # guard against words with embedded whitespace
      $pair{word} =~ s/ /_/g;
      push @corpus, { %pair };
    }
  }
  close(IN);
  push @corpus, { word => '#EOS#', cat => 'XY' };

}

# get annotated categories
print "Creating stream of words...\n";

if ($opt_r) {

  open(TAGS, "tags") or die "Can't read `tags'\n";
  while(<TAGS>) {
    if(/^$/) {
      push @corpus, { word => '#EOS#', cat => 'XY' };
    } else {
      /^(.+)\t(.+)/ or die "Can't match `$_'";
      push @corpus, { word => $1, cat => $2 };
    }
  }
  close(TAGS);

}
  else {

  our $items = shift;
  our @items = split(/,/, $items);
  for my $item (@items) {

    if(-f $item) {

      if (!defined $jobname) {
        # deduce base name of the corpus
        $jobname = File::Spec->rel2abs($item);
        if ($jobname =~ m|/([a-z]+)([-_][^/]+).export|) {
          $jobname = $1;
        } else {
          my ($volume,$directories,$file) = File::Spec->splitpath($jobname);
          $jobname = $file;
        }
      }

      open (TXT, $item) or die "Can't open $item.\n";
      if ($opt_n) {

        # read NEGRA export format
        print "  $item [NEGRA format]\n";
        my $last = 'cat';
        my %pair;
        my $on = 0;
        while (<TXT>) {
          chomp;
          if (/^#BOS/) {
            $on = 1;
            next;
          }
          if (/^#EOS/) {
            push @corpus, { word => '#EOS#', cat => 'XY' };
            $on = 0;
            next;
          }
          if (!$on) {
            next;
          }
          if (/^([^#\t]+)\t+([^\t]+)/) {
            %pair = ( word => $1, cat => $2 );
            # guard against words with embedded whitespace
            $pair{word} =~ s/ /_/g;
            push @corpus, { %pair };
          }
        }
      } else {
        print "  $item [flat format]\n";

        # read flat tagged format
        while (<TXT>) {
          chomp;
          if (/^(\S+)\t(\S+)/) {
            my %pair = ( word => $1, cat => $2 );
            push @corpus, { %pair };
          } else {
            push @corpus, { word => '#EOS#', cat => 'XY' };
          }
        }
      }
      close(TXT);
    } elsif($item =~ /(\S+):(\d+)-(\d+)/) {

      my($corpus,$from,$to) = ($1,$2,$3);
      my $i = $to-$from+1;

      if (!defined $jobname) {
        $jobname = $corpus
      }

      # read CDG annotations
      print "  $corpus [$i annotations]\n";

      my $dir = "/data/corpus/ccc/$corpus";
      for my $i ($from..$to) {
        my $subdir = sprintf "%03d", ($i / 10000);
        my $file= "$dir/$subdir/$corpus-s$i.cda";
        if (! -f $file) {
          $file= "$dir/$subdir/auto-$corpus-s$i.cda";
        }
        next unless -f $file;
        read_cda $file;
      }
    }
  }

  # write list of words only
  open(WORDS, ">words") or die "Can't write to `words'\n";
  for (@corpus) {
    if ($$_{word} eq '#EOS#') {
      print WORDS "\n";
    } else {
      print WORDS "$_->{word}\n";
    }
  }
  close(WORDS);

  open(TAGS, ">tags") or die "Can't write to `tags'\n";
  for (@corpus) {
    if ($$_{word} eq '#EOS#') {
      print TAGS "\n";
    } else {
      print TAGS "$_->{word}\t$_->{cat}\n";
    }
  }
  close(TAGS);

}

# call tagger
print "Calling tagger...\n";
open(TAGGER,
     "cat words |" .
     " $taggers{$tagger}{invocation} 2>${tagger}.${jobname}.err |tee ${tagger}.${jobname}.output |")
  or die "Can't call tagger\n";
my $i = 0;

while (<TAGGER>) {
  next unless s/^([^\t]+)\t+//;

  my $word = $1;

  my $pair = $corpus[$i++];

  if ($$pair{word} eq '#EOS#') {
    $pair = $corpus[$i++];
  }

  die "no word nr $i!" unless defined $$pair{word};

  # TreeTagger sometimes overrules our tokenization,
  # turning gibt+'s into gibt+'+s.
  # There's nothing we can do about this except skip the '.
  if ($word ne $$pair{word}) {
    warn "Mismatch: got `$word', expected `$$pair{word}'\n";
    warn "Have to skip `$pair->{word}'\n";
    <TAGGER>;
    next;
  }

  # actually process one line
  my %tagged = ();
  interpret(\%tagged, $_, $$pair{cat});
  $corpus[$i-1]{tagger} = { %tagged };
}
close(TAGGER);

# print transcript
my $words = 0;
my $errors = 0;
my $baderrors = 0;
my $p = 0;

my $logname = "$tagger.$jobname.log";
open(FILE, ">$logname") or die "Can't write `>$logname': $!\n";
for my $w (@corpus) {

  if ($$w{word} eq '#EOS#') {
    print FILE "\n";
    next;
  }

  $words++;

  if (!exists($w->{tagger})) {
    print FILE "Tokenizer error for $w->{word}\n";
    $errors++;
  } else {
    printf FILE "%-24s", "$w->{cat}:$w->{tagger}{max}";
    printf FILE "%-25s", "$w->{word}";
    for (sort {
      $w->{tagger}{$b} <=> $w->{tagger}{$a} || $a cmp $b }
         grep { $_ ne 'max' }
         keys %{$w->{tagger}}) {
      next if /^max$/;
      if ($w->{cat} eq $_) {
        print FILE "$_ [$w->{tagger}{$_}] ";
      } else {
        print FILE "$_ $w->{tagger}{$_} ";
      }
    }
    print FILE "\n";

  }

  # find errors
  if (exists $w->{tagger}) {
    if ($w->{tagger}{max} ne $w->{cat}) {
      $errors++;
      $stat{"$w->{cat}:$w->{tagger}{max}"}++;
      if (!exists($similar{"$w->{cat}:$w->{tagger}{max}"})) {
        $baderrors++;
      }
    }
    if (exists($w->{tagger}{$w->{cat}})) {
      $p += $w->{tagger}{$w->{cat}};
    }
  }
}

# print statistics about what errors happen
for (sort { $stat{$a} <=> $stat{$b} || $b cmp $a } keys %stat) {
  print FILE exists($similar{$_}) ? ' ' : '*';
  print FILE "$_: $stat{$_}\n";
}

# give recall percentage
my $correct = $words - $errors;
printf FILE "         single:      %4.3f%% ($correct/$words)\n", 100 * $correct/$words;
printf "         single:      %4.3f%% ($correct/$words)\n", 100 * $correct/$words;

# give average p of the correct tag
printf FILE "         multi:       %4.3f%% \n", 100 * $p/$words;
printf "         multi:       %4.3f%% \n", 100 * $p/$words;

$correct = $words - $baderrors;
printf FILE "nonstrict single:      %4.3f%% ($correct/$words)\n", 100 * $correct/$words;
printf "nonstrict single:      %4.3f%% ($correct/$words)\n", 100 * $correct/$words;

printf FILE "nonstrict multi:       %4.3f%% \n", 100 * $p/$words;
printf "nonstrict multi:       %4.3f%% \n", 100 * $p/$words;



close(FILE);
