#!/usr/bin/env perl
# Copyright (C) 1997-2006 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  
# translate wsj sexpression format to negra export format
#
# author: Michael Daum <micha@nats.informatik.uni-hamburg.de>
# initial revision: 2003-09-24


=head1 NAME

Penn2Negra - A Convert Tool

=head1 SYNOPSIS

=over 10

=item Penn2negra

=over 10

=item [B<-debug>]

=item [B<-quiet>]

=item [B<-startnr> I<number>]

=item [I<input-file1> I<input-file2> ...]

=back

=back


=head1 OPTIONS

=over 10

=item B<-startnr> I<number>

start numbering the sentences with I<number>.

=back

=cut



use strict;
use Getopt::Long;
use Pod::Usage;
use Parse::RecDescent;

my $quiet;
our $nrSentences = 1;
our $debug_ = 0;

#------------------------------------------------------------------------------
sub main {
  print "%%  automatically generated by penn2negra.pl (c) 2003 \n" .
        '%% $Id: penn-stts2negra.pl,v 1.3 2006-05-10 12:34:30 kdaling Exp $' . "\n" .
        '%% contact: The CDG Team <cdg@nats.informatik.uni-hamburg.de>' . "\n\n";

my $pennGrammar = <<'EOG';
{ 
  my $maxId = 0;
  my $innerMaxId = 500;

  my $nodeTagPattern = q{\$\*LRB\*|[^\s\(\)=\*-]+|-LRB-|-RRB-|\*E\*|--};
  my $edgeTagPattern = q{(?:(?:--)|[A-Z0-9]+)};
  my %nodes;	# map node ids to nodes
  my %references; # map trace references to nodes
}

sentence: { &::printDebug("trying sentence at\n'$text'"); } <reject>
sentence: '(' sExpr ')' 
  {
    &::printDebug("found sentence");
    my $rootNode = {
      "id" => -1,
      "isLexHead" => 1,
      "nodeTag" => "ROOT",
      "edgeTag" => "ROOT",
      "word" => "NIL",
      "children" => [$item[2]],
    };

    # establish traces and coreferences
    my @todelete = ();
    foreach my $nodeId (keys %nodes) {
      my $node = $nodes{$nodeId};
      if(defined $node->{traceId}) {
	my $reference = $references{$node->{traceId}};
        &::printDebug ("Move node \"$reference->{id}\" from former father \"$reference->{parent}->{id}\" ...");
	&::deleteChild($reference);
        $reference->{parent} = $node->{parent};
	$reference->{edgeTag} = $node->{edgeTag};
	&::replaceChild($node, $reference);
	push @todelete, $node;
#       $node->{trace} = $reference;
#       $reference->{backTrace} = $node;
        &::printDebug ("  ... to new father \"$reference->{parent}->{id}\" with new edge tag \"$reference->{edgeTag}\"");
      }
      if(defined $node->{corefId}) {
        my $reference = $references{$node->{corefId}};
  	$node->{coref} = $reference;
        $reference->{backRef} = $node;
      }
    }
    foreach my $delnode (@todelete) {
      delete($nodes{$delnode->{id}});
    }
    # process one sentence
    $return = &::processSentence($rootNode, \%nodes);
    undef %nodes;
    undef %references;
    $maxId = 0;
    $innerMaxId = 500;
  } | {
    print STDERR "ERROR: in $::nrSentences. sentence\n$text\n";
    $return = -1;
  }

sExpr: { &::printDebug("trying sExpr at\n$text"); } <reject>
sExpr: '(' sExprContent ')'
  {
    &::printDebug("found sExpr");
    $return = $item[2];
  }

sExprContent: { &::printDebug("trying sExprContent at\n$text"); } <reject>
sExprContent: 
  nodeTag sExpr(s)
  { 
    my $nodeTag = $item[1][0];
    my $edgeTag = $item[1][1]?$item[1][1]:"--";
    my $ref1Id = $item[1][2];
    my $ref2Id = $item[1][3];
    my $node = {
      "nodeTag" => $nodeTag,
      "edgeTag" => $edgeTag,
      "id" => $innerMaxId++,
      "children" => $item[2],
    };
    $nodes{$node->{id}} = $node;

    if (defined $ref1Id) {
      $node->{corefId} = $ref1Id;
    } else {
      $ref1Id = "";
    }
    if (defined $ref2Id) {
      $references{$ref2Id} = $node;
      $node->{ref} = 1;
    } else {
      $ref2Id = "";
    }
    &::printDebug("found nodeTag sExpr(s) with node tag \"$nodeTag\", edge tag \"$edgeTag\", coref \"$ref1Id\" and reference \"$ref2Id\"");
    foreach my $child (@{$node->{children}}) {
      $child->{parent} = $node;
    }
    $return = $node;
  }
  | nodeTag /([^()\s]+)/
  {
    my $posTag = $item[1][0];
    my $edgeTag = $item[1][1]?$item[1][1]:"--";
    my $ref1Id = $item[1][2];
    my $ref2Id = $item[1][3];
    my $node = {
      "nodeTag" => $posTag,
      "edgeTag" => $edgeTag,
      "id" => $maxId++,
      "word" => $item[2],
    };
    $nodes{$node->{id}} = $node;
    if (defined $ref1Id) {
      $node->{corefId} = $ref1Id;
    } else {
      $ref1Id = "";
    }
    if (defined $ref2Id) {
      $references{$ref2Id} = $node;
    } else {
      $ref2Id = "";
    }
    &::printDebug("found pos tag \"$posTag\" with edge tag \"$edgeTag\", word \"$item[2]\", coref \"$ref1Id\" and reference \"$ref2Id\"");
    $return = $node;
  }
  | emptyCat 
  {
    my $node = {
      "word" => $item[1][0],
      "edgeTag" => "--",
      "nodeTag" => "-NONE-",
      "id" => $maxId++,
    };
    if ($item[1][1]) {
      $node->{traceId} = $item[1][1];
    }
    if ($item[1][2]) {
      ($node->{edgeTag} = $item[1][2]) =~ s/^-//;
    }
    $nodes{$node->{id}} = $node;
    $return = $node;
  }

emptyCat: { &::printDebug("trying emptyCat at\n$text"); } <reject>
emptyCat:
  '-NONE-' /0|(\*\?\*)/
  {
    $return = [$item[2]];
  }
  |
  '-NONE-' /(\*(?:[A-Z]+\*)?)(?:\-([0-9]+))?/
  {
    $return = [$1, $2];
  }
  |
  /\*T([0-9]+)\*((?:-$edgeTagPattern)*)\s+-/
  {
    &::printDebug("found trace no. \"$1\" with edge tag \"$2\"");
    $return = ["*", $1, $2];
  }
  | <error: error parsing emptyCat near\n$text>

nodeTag: { &::printDebug("trying nodeTag at\n$text"); } <reject>
nodeTag: /($nodeTagPattern)((?:-$edgeTagPattern)*)(?:=([0-9]+))?(?:-\*T([0-9]+)\*)?/
  {
    my $nodeTag = $1;
    my $edgeTags = $2;
    my $ref1Id = $3;
    my $ref2Id = $4;
    if ($edgeTags) {
      $edgeTags =~ s/^-//;
    }
    $return = [$nodeTag, $edgeTags, $ref1Id, $ref2Id];
  } 
  | <error: error parsing an nodeTag near\n$text>



EOG

  my $parser = Parse::RecDescent->new($pennGrammar) 
    or die "ERROR: bad treebank parser\n";

  # call processSentence for each sentence
  $/="";
  while (my $data = <>) {
    my $retval = $parser->sentence(\$data);
    die "\n" if $retval == -1;
    last if $retval == 0;
  }
  $/ = "\n";
}

#------------------------------------------------------------------------------
sub deleteChild {
  my $node = shift;
  my @children = @{$node->{parent}->{children}};
  my @newchildren = ();
  foreach my $child (@children) {
    if ($child != $node) {
      push @newchildren, $child;
    }
  }
  @{$node->{parent}->{children}} = @newchildren;
}

#------------------------------------------------------------------------------
sub replaceChild {
  my $node = shift;
  my $toreplace = shift;
  my @children = @{$node->{parent}->{children}};
  my @newchildren = ();
  foreach my $child (@children) {
    if ($child != $node) {
      push @newchildren, $child;
    } else {
      push @newchildren, $toreplace;
    }
  }
  @{$node->{parent}->{children}} = @newchildren;
}

#------------------------------------------------------------------------------
sub processSentence {
  my $rootNode = shift;
  my $nodes = shift;

  unless ($quiet) {
    print STDERR "s$nrSentences     \r";
  }
  print "#BOS $nrSentences 1 20030916 1\n";

  foreach my $node (sort { $a->{id} <=> $b->{id} } values %{$nodes}) {
    my $word = $node->{word};
    my $nodeTag = $node->{nodeTag};
    if ($nodeTag eq "TOP") { next; }
    my $edgeTag = $node->{edgeTag};
    my $parentNode = $node->{parent};
    my $parentId = $parentNode->{id} || "0";
    if ($node->{parent}->{nodeTag} eq "TOP") { $parentId = "0"; }
    my $traceNode = $node->{trace};
    my $corefNode = $node->{coref};

    if ($traceNode) { next; }
    
    if (defined $word) {
      printf "%-20s ", $word;
    } else {
      printf "#%d ", $node->{id};
    }
    printf "%-5s -- %-5s %-5s", $nodeTag, $edgeTag, $parentId;

    #if ($traceNode) {
    #  print "TRACE $traceNode->{id}";
    #}
    if ($corefNode) {
      print "COREF $corefNode->{id}";
    }

    print "\n";
  }
  print "#EOS $nrSentences\n";
  printDebug("\n\n");

  $nrSentences++;
  return 1;
}

#------------------------------------------------------------------------------

sub printDebug {
  my $string = shift;
  print STDERR "DEBUG: $string\n" if $debug_;
}

#------------------------------------------------------------------------------

sub usage {
        print "usage: penn2negra.pl [-startnr number]\n\n";
        exit 1;
}

my $help = 0;

$help = 1 unless
  GetOptions(
    "d|debug+" => \$debug_,
    "startnr:i" => \$nrSentences,
    "quiet"     => \$quiet,
    "syntax|h|help|?" => \$help,
  ) || pod2usage(2);

pod2usage(-exitval =>1, -message => "\nThe Penn 2 Negra Conversion Tool\n") if $help;

#our $argcnt = -1;
#our $arg = "";
#while ($#ARGV > $argcnt) {
#        $argcnt++;
#        $arg = $ARGV[$argcnt];
#        if ($arg eq "-startnr") {
#                if ($#ARGV > $argcnt) {
#                        $argcnt++;
#                        $nrSentences = $ARGV[$argcnt];
#                } else {
#                        usage();
#                }
#        } else {
#                usage();
#        }
#}

&main();
1;
