#!/usr/bin/env perl
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# cdg/utils/shift-reduce.pl
#
# Construct syntax structure using previously observed state counts.
#
#
# SHIFT-REDUCE PARSING
#
# This is a helper program for CDG that predicts entire labelled syntax
# trees in linear time by shift-reduce parsing. The idea is due to Nivre
# (2003), but the implementation is entirely original.
#
# Shift-reduce-parsing uses a stack and an input pointer. The top word
# of the stack and the current input word are known as `top' and `dot',
# and they are the only words that can ever form a dependency edge.
#
# At any time, one of four actions can be taken:
#
#   SHIFT: The word under the input pointer (`dot') is pushed onto the
#          stack. This is done when that word will acquire more
#          dependents on its right side.
#
#   REDUCE: The top of the stack is removed. This means that this word
#           cannot take part in any more dependencies.
#
#   LEFT: The top word is subordinated under the dot word (and removed
#         from the stack, since it cannot take any more dependents).
#
#   RIGHT: The dot word is subordinated under the top word and pushed
#          onto the stack.
#
# The LEFT, RIGHT and REDUCE moves can be annotated with the label
# that should be assigned to a word. Unless the `-u' switch is given,
# actions will actually be `LEFT(SUBJ)' rather than just `LEFT'.
#
# With the right sequence of moves, this procedure can create any
# projective syntax tree. Obviously, parsing accuracy depends on being
# able to choose the correct decision at every step.
#
#
# PARSE STATE FEATURES
#
# This program takes parsing decisions depending on abstracted features
# of the current state of the parser. These features can include:
#
# 1. The top word of the stack. This may either be its reading or its
# POS tag, or an intermediate representation. For instance, the word
# `lteren' might be represented as itself, or just as `ADJA', or as the
# string `-ren/ADJA'; this is subject to the setting of the lexicalization
# switch `-L'. With the value 0, only the POS tag is used. With the
# value 1, content words are represented by their endings as shown above,
# and function words are represented by their reading and POS tag. With
# the value 2, all words are represetend by their reading and POS tag;
# note that this leads to huge parameter spaces that will probably
# overflow your RAM if you train on more than 10,000 sentences or so.
#
# POS tags can be tied to each other in order to reduce the parameter
# space. The switch `-f' specifies a file for parameter tying; each line
# has the form `NE=NN'. This line would instruct the program not to make
# any differences between NE and NN words.
#
# 2. The set of dependencies involving the top word. This is used when
# the `attachment' switch `-a' is set. This feature represents three
# different subfeatures: whether or not the top word is already attached
# (and possibly with what label), whether or not is already has a left
# dependent, and whether ot not it already has a right dependent.
#
# 3. The distance between top and dot, measured in words. This is used
# when the `-i' switch is set.
#
# 4. The categories of the words following dot. The number of these
# `lookahead' words is specified with the `-l' switch.
#
# Which of these features will be actually used depends on the switches
# passed to the program. The switches that affect feature selection are
# `-a', `-f', `-i', `-l' and `-L'. Note that when reading a parse table,
# the program recreates the settings of these switches that were in
# effect when the table was created. Therefore, these switches will
# effectively be ignored when parsing, since honoring them would likely
# result in mismatched signatures and parse failures.
#
# The recommended settings for creating parse tables are `-i -a -l2 -L1'.
#
#
# SIGNATURES
#
# The description of a parse state according to the features specified
# by the program flags is called a `signature'. A signature is created
# from a parse state with the encode() function. A signature might look
# like this:
#
#    will/VMFIN#110#5#ADJD|$
#
# However, this format is likely to change frequently - tables created
# by old versions of this program will probably not work with newer
# versions. This particular signature might mean that top has the
# category `VMFIN', has a left dependent but no right dependent yet, is
# itself already attached, is five words distant from dot, and dot has
# the category `ADJD' (and is the last word, therefore the second
# lookahead token is `$').
#
#
# PARSING OPTIONS
#
# When parsing new input, the program will compare the current parse
# state to previous observations and take the action that was taken most
# often in this situation. If the exact state was never seen,
# successively more general signatures are tried until eventually only
# the top word category remains. Backing off happens by omitting the
# features in the following fixed order: distance measure, lexical
# information, look ahead tokens (one by one), attachment information.
# If all else fails, SHIFT or REDUCE is taken depending on whhat is
# currently possible.
#
# Backing off normally never happens unless a parse state has never been
# seen even one. You can change the threshold away from 1 with the `-h'
# switch.
#
# Deterministic parsing will be used unless the `-b' switch is given to
# demand backtracking. A backtracking search will try out all previously
# observed events and find the parse with the globally highest
# probability. This takes much longer than a deterministic search, of
# course, and does not even always perform better. Therefore there is
# not much point to using `-b' since it defies the only purpose this
# program has: speed.
#
# If the `-u' switch is given, no dependency labels will be assigned.
# This means that the set of possible parse actions is much smaller and
# can actually increase structural accuracy, since it amounts to a
# larger training set.
#
# If the `-s' switch is given, the set of parsing moves is extended by
# the nonstandard `FARLEFT' and `FARRIGHT' moves. These allow the
# creation of nonprojective dependency edges by attaching either top to
# a word in the lookahead window, or dot to a word within the stack. This
# is supported for training only, so it is only useful to count how many
# nonprojective edges there tend to be. In order to allow these moves
# during parsing, further information about how far the `FAR' is
# supposed to be would be necessary.
#
#
# INVOCATION
#
# shift-reduce.pl eats BDB files with the expected extension `.sr'. If the
# program can find the file `default.sr', the default action is to parse
# input. Use `-p <filename>' to parse with another table, and `-t
# <filename>' to train a new parse table.
#
# Input can be specified in three ways. (a) With the target `-', the
# program will parse word/tag pairs from STDIN and write predictions
# suitable for CDG. You can try this with a command line such as
#
#    echo Hallo, wie geht es dir? | \
#     tokenize.pl | \
#     deutsch-tagger.pl | \
#     shift-reduce.pl -
#
# and watch the output. This is not meaningful with `-t'.
#
# (b) The target may also be a file name ending in `.lat'. All lattices in
# that file will be parsed, and the results will be written to new files
# as CDG annotations under the name `<latticeid>-sr.cda'. (This
# implies the `-r' switch to obtain POS tags.) This method is suitable
# for bulk parsing of huge corpora.
#
# (c) If the target has the form `<corpus name>:<number range>', the
# specified annotations from the named corpus will be processed. The
# base directory for searching such corpora can be specified with the
# `-D' switch.
#
# The number of analyses can be limited with the `-n' switch (this is
# only useful if you give a range of numbers of which not all
# annotations exist). A single number can be used instead of a range. If
# no range is given, all available annotations are used. More than one
# such specification can be given, separated with commas.
#
# In training mode, each of the annotations will be deparsed and the
# events in the parse history will be counted. Note that nonprojective
# dependency trees will be simplified until they can be handled by
# shift-reduce parsing unless the `-s' switch is given.
#
# In parsing mode, each of the sentences will be parsed, and the
# annotations will be used for measuring the accuracy. Note that the POS
# tags from the annotation will be used to guide parsing, as in Nivre
# (2003), therefore the parsing figures obtained will not be quite
# realistic. Use the `-r' switch to use real tagging instead.
#
# The program assumes that the name of the POS tag feature is `cat'; you
# can change this assumption with the `-c' switch. The level used to
# extract syntax structures is assumed to be called `SYN'; you can change
# this assumption with the `-S' switch.
#
# When the `-w' switch is given, the parse result will be written to a
# new CDG annotation even when annotations or STDIN is parsed. If the
# switch `-x' is also given, the new annotations will immediately be
# displayed in XCDG.


use warnings;
use strict;

BEGIN {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0);
  require 'cdgsetup.pl';
}

use CDG;
use DB_File;


use Getopt::Std;
our $opt_a = 0;
our $opt_b = 0;
our $opt_c = 'cat';
our $opt_C = 0;
our $opt_d = 0;
our $opt_D = '/data/corpus/ccc';
our $opt_e = '';
our $opt_f = '';
our $opt_h = 1;
our $opt_i = 0;
our $opt_l = 1;
our $opt_L = 0;
our $opt_n = 0;
our $opt_p = 0;
our $opt_r;
our $opt_s = 0;
our $opt_S = 'SYN';
our $opt_t;
our $opt_u = 0;
our $opt_v = 0;
our $opt_w = 0;
our $opt_x = 0;
our $invocation = "$0 @ARGV";
getopts('abc:Cde:D:f:h:il:L:m:n:p:st:urv:wx');

# default: parse with default.sr
if(!$opt_p && !$opt_t) {
  $opt_p = 'default.sr';
}

# default table name for training is also default.sr
if (!$opt_t) {
  $opt_t = 'default.sr';
}

# debugging imples verbosity
if($opt_d) {
  $opt_v = 2;
}

# displaying implies writing
if($opt_x) {
  $opt_w = 1;
}

# FAR transitions imply training
if($opt_p && $opt_s) {
  die "The special FAR transitions are only supported for training.\n";
}

die "Usage: $0 <spec>

    <spec> may look like this:

      Grundgesetz:1-1000,EU:90   parses range of known corpus
      000.lat                    parses all lattices in file
      -                          parse STDIN

    You probably want to specify a parse table with either
    `-t' (train) or `-p' (parse).

    -a    use top word attachment as a feature (FALSE)
    -b    use backtracking (FALSE)
    -c S  category feature (cat)
    -C    Complement checking: avoid assigning certain labels twice
    -d    debug parsing decisions (implies -p and -v2)
    -D S  base directory for treebanks (default /data/corpus/ccc)
    -e S  extra levels to assume, e.g. `-e Phon,Sem,Prag'
    -f S  use file for category tying (off)
    -h N  threshold for backoff (1)
    -i    encode distance between visible words as a feature (FALSE)
    -L N  lexicalisation level (default 0)
    -l N  lookahead window (1)
    -n N  use only N utterances (all)
    -p S  parse using specified table (default: parse from default.sr)
    -r    use real POS tagger (default: simulate)
    -s    allow special (nonprojective) transitions where necessary (FALSE)
    -S S  syntax level name (SYN)
    -t S  Train and create specified transition table
    -u    use Unlabelled trees
    -v N  verbosity level (0)
    -w    write parse result as new annotation (FALSE)
    -x    immediately launch the tree viewer (implies -p and -w) (FALSE)

" if $#ARGV;

# possible parser moves
our @moves = qw(SHIFT REDUCE LEFT RIGHT);
push @moves, qw(FARRIGHT FARLEFT) if($opt_s);

# possible actions. An action is simply a labelled move.
our @actions;

# edge labels seen in training material
our %labels;

# table of category ties
our %ties = ();
if($opt_f) {
  open(TIES, "$opt_f") or die "Can't read `$opt_f': $!\n";
  while(<TIES>) {
    if(/^(.+)=(.+)$/) {
      $ties{$1} = $2;
    }
  }
}

close(TIES);

# table of observed parse actions per signature
our @observations;

# maximal number of features used in signatures
our $maxdim;

# on-disk repository of information created from @observations in a
# previous run
our %db;

# list of sentence structures pulled from CDG
our @annos;

# list of lattices pulled from CDG
our @lattices;

# number of sentences to load in one go
our $block_size = 1000;
if($opt_r) {
  $block_size = 1000;
}


# number of events before we trust a parsing decision
our $threshold = $opt_h;

# number of permissible agenda items
our $agenda_size = 10;

# Global accuracy statistics
our %accuracy;

# conditionally print output
sub report {
  print STDERR @_ if $opt_v >= shift;
}

# quote and escape string suitably for CDG input
sub print_identifier {
  my $x = shift;
  (my $result = $x) =~ s/([\'\"])/\\$1/g;
  if ($result =~ /[^a-zA-Z\x80-\xff]/) {
    $result = "'$result'";
  }
  return $result;
}

# sort like ls -v does
sub byVName {

  my ($a, $b) = @_;

  # don't change the names themselves
  my ($x,$y) = ($a,$b);

  while ($x ne "" && $y ne "") {

    # skip identical alphabetical parts
    while (substr($x,0,1) eq substr($y,0,1) &&
           $x !~ /^[0-9]/) {
      substr($x,0,1) = "";
      substr($y,0,1) = "";
    }

    # compare alphabetical parts normally
    if ($x !~ /^\d/ || $y !~ /^\d/) {
      return $x cmp $y;
    }

    # compare numeric parts numerically
    $x=~s/^(\d+)//;
    my $nx = $1;
    $y=~s/^(\d+)//;
    my $ny = $1;
    my $result = $nx <=> $ny;
    next unless $result;
    return $result;
  }
}

# Does node N already have a dependent labelled L in parse state S?
sub has_label {
  my($n,$s,$l) = @_;
  my $i = 0;
  for(@{$$s{labels}}) {
    if($_ eq $l && $$s{struc}[$i] == $n) {
      return 1;
    }
    $i++;
  }
  return 0;
}

# Abstract from current parse state
sub encode {
  my $s = shift;
  my $w = shift;
  my $t = shift;

  my $dim = 0;

  my $result;

  # find stack category
  my $topcat;
  if(@{$s->{stack}}) {
    my $top = $s->{stack}[0];
    $topcat = $$t[$top];
  } else {
    $topcat = 'NIL';
  }

  # encode top word and category
  if($opt_L) {

    if(@{$s->{stack}}) {
      my $top = $s->{stack}[0];
      my $word = $$w[$top];
      my $tag = $$t[$top];

      # at level 1, regard only the ending of open-class words
      if(1 == $opt_L) {
        if($tag =~ /ADJA|ADJD|NN|NE|^VV/) {
          $word = '-' . substr($word,-3);
        }
      }
      $result = "$word/$tag";
    } else {
      $result = "NIL/NIL";
    }
    $dim = 2;
  }

  # encode top category only
  else {
    $result = $topcat;
    $dim = 1;
  }

  # encode environment of top word
  if ($opt_a) {

    my $attach;

    if (@{$s->{stack}} &&
        $s->{struc}[$s->{stack}[0]]) {
      $attach = '1';
    } else {
      $attach = '0';
    }

    my $has_left_child = 0;
    my $has_right_child = 0;
    if (@{$s->{stack}}) {
      for (1..$s->{dot}-1) {
        if ($s->{struc}[$_] == $s->{stack}[0]) {
          if ($_ < $s->{stack}[0]) {
            $has_left_child = 1;
          } else {
            $has_right_child = 1;
          }
        }
      }
    }
    if ($has_left_child) {
      $attach .= "1";
    } else {
      $attach .= "0";
    }
    if ($has_right_child) {
      $attach .= "1";
    } else {
      $attach .= "0";
    }

    $result .= "#$attach";
    $dim++;
  }

  # encode distance between top of stack and input pointer
  if ($opt_i) {
    my $d = @{$s->{stack}} ? $s->{dot}-$s->{stack}[0] : 0;
    $result .= "#$d";
    $dim++;
  }

  # encode lookahead categories
  my @lookahead;
  for ($s->{dot}..$s->{dot}+$opt_l-1) {
    $dim++;
    if (defined($$t[$_])) {
      push @lookahead, $$t[$_];
    } else {
      push @lookahead, "\$";
    }
  }
  if(@lookahead) {
    $result .= '#';
    $result .= join('|',@lookahead);
  }

  $maxdim = $dim;

  return $result;

}

# Given a particular state specification, return one that has
# less contextual information.
#
# The possible features are stripped in the following order:
#
# - if there is distance information, it is stripped
#
# - if the top word is lexicalized it is delexicalized
#
# - if there are lookahead tokens, one is removed
#
# - if there is stack attachment information, it is removed
#
# reduce_context() does not care whether the new state has any
# information attached to it or not, it just backs off by exactly one
# feature. A better solution might be to back of to the neighbouring
# state that has the most weight, or to pool the information of
# neighbouring states.
sub reduce_context {
  my $s = shift;

  # remove lexical information
  if($$s =~ s!^[^\#]+/([^\#]+)!$1!) {
    return 1;
  }

  # remove distance information
  if($$s =~ s/#(\d+)#/#/) {
    return 1;
  }

  # remove lookahead tokens
  if($$s =~ s/[|#][^|#]+$//) {
    return 1;
  }

  # remove attachment information
  if($$s =~ s/#([01]{3})#/#/) {
    return 1;
  }

  return 0;
}

# translate AnnotationStruct into perl hash
sub read_anno {
  my $ae = shift;

  my $x =
    {
     id => AnnoEntryStruct_id_get($ae),
     annotated => [ undef ],
     normalized => [ undef ],
     tags => [''],
     words => ['NIL'],
     labels => [''],
     projective => [''],
     history => []
    };

  for (my $m = AnnoEntryStruct_annos_get $ae;$m;$m = ListStruct_next_get $m) {
    my $a = ListStruct_item_get $m;
    my $cat = undef;
    my $struc = undef;

    push @{$x->{words}}, AnnotationStruct_word_get $a;

    for (my $n = AnnotationStruct_specs_get $a;$n;$n = ListStruct_next_get $n) {
      my $s = ListStruct_item_get $n;
      my $t = SpecificationStruct_type_get $s;

      # note tag
      if ($t eq $STTag) {
        my $att = SpecificationStruct_kind_get $s;
        if ($att eq $opt_c) {
          $cat = SpecificationStruct_name_get $s;
        }
      }

      # note edge
      else {
        my $level = SpecificationStruct_kind_get $s;
        if ($level eq $opt_S) {
          $struc = SpecificationStruct_position_get $s;
          my $label = SpecificationStruct_name_get $s;
          push @{$x->{labels}}, $label;
          $labels{$label} = 1;
        }
      }
    }

    # note missing cat
    if (!defined $cat) {
      print STDERR "WARNING: no category for word `$$x{words}[-1]' in $$x{id}\n";
      $cat = '';
    }

    # maybe tie cat
    if (%ties && exists $ties{$cat}) {
      $cat = $ties{$cat};
    }

    # note missing structure
    if (!defined $struc) {
      print STDERR
        "WARNING: no syntactic structure for word `$$x{words}[-1]' in $$x{id}\n";
      $struc = 0;
    }

    push @{$x->{projective}}, 1;
    push @{$x->{annotated}}, $struc;
    push @{$x->{tags}}, $cat;
  }
  report 1, "$$x{id}: @{$x->{words}}\n";

  return $x;
}



# return earlier end of (annotated) edge #$i
sub min {
  my $x = shift;
  my $i = shift;
  return $i < $$x{annotated}[$i] ? $i : $$x{annotated}[$i];
}
# return later end of (annotated) edge #$i
sub max {
  my $x = shift;
  my $i = shift;
  return $i > $$x{annotated}[$i] ? $i : $$x{annotated}[$i];
}
# is (annotated) edge #$i attached at the left?
sub left {
  my $x = shift;
  my $i = shift;
  return $i < $$x{annotated}[$i];
}
# is (annotated) edge #$i attached at the right?
sub right {
  my $x = shift;
  my $i = shift;
  return $$x{annotated}[$i] && $i > $$x{annotated}[$i];
}

# detect nonprojective edges
sub check_anno {

  my $x = shift;

  my @normalized = @{$x->{annotated}};
  my @projective = (1) x ($#{@$x{words}}+1);
  my @labels = @{$$x{labels}};
  my @tags = @{$$x{tags}};
  my @words = @{$$x{words}};

  # check all edge pairs
  for my $X (1..$#{$x->{words}}) {
    next unless $normalized[$X];
    for my $Y ($X+1..$#{$x->{words}}) {
      next unless $normalized[$Y];

      if (min($x,$X) < min($x,$Y) &&
          max($x, $X) < max($x,$Y) &&
          max($x,$X) > min($x,$Y)
          ||
          $X == $normalized[$Y] &&
          right($x,$Y) &&
          left($x,$X) &&
          $normalized[$X] < $Y
          ||
          $Y == $normalized[$X] &&
          right($x,$Y) &&
          left($x,$X) &&
          $normalized[$Y] > $X
         ) {

        report 3, "WARNING: nonprojective edges " .
          "$X:$$x{words}[$X]($$x{labels}[$X]) $Y:$$x{words}[$Y]($$x{labels}[$Y])\n";

        # do not punish an edge twice
        if ($projective[$X] && $projective[$Y]) {

          # try very hard to find the `responsible' edge and mark it
          # !This is heavily grammar-dependent!
          if ($labels[$X] eq 'RE') {
            $projective[$X] = 0;
          } elsif ($labels[$Y] eq 'RE') {
            $projective[$Y] = 0;
          } elsif ($labels[$X] =~ '^(KOM|CJ|KON)$') {
            $projective[$X] = 0;
          } elsif ($labels[$Y] =~ '^(KOM|CJ|KON)$') {
            $projective[$Y] = 0;
          } elsif ($labels[$X] eq 'REL') {
            $projective[$X] = 0;
          } elsif ($labels[$Y] eq 'REL') {
            $projective[$Y] = 0;
          } elsif ($labels[$X] eq 'OBJC' && $tags[$normalized[$X]] eq 'NN') {
            $projective[$X] = 0;
          } elsif ($labels[$Y] eq 'OBJC' && $tags[$normalized[$Y]] eq 'NN') {
            $projective[$Y] = 0;
          } elsif ($labels[$X] =~ /^OBJ|SUBJ|PRED/) {
            $projective[$X] = 0;
          } elsif ($labels[$Y] =~ /^OBJ|SUBJ|PRED/) {
            $projective[$Y] = 0;
          } else {
            # give up
            report 3, "WARNING: weird nonprojective edges " .
              " $words[$X]($labels[$X]) $words[$Y]($labels[$Y])\n";
          }
        }
      }
    }
  }

  $$x{normalized} = [ @normalized ];
  $$x{projective} = [ @projective ];

}

# print current state to screen,
# e.g. "[ soll Produkte ] . und"
sub print_state {
  my $s = shift;
  my $w = shift;

  return '' unless exists $s->{dot};
  my $string = $opt_b ? "($$s{no})" : '';
  $string .= '[';
  for (reverse @{$s->{stack}}) {
    die "no word $_!" unless defined $$w[$_];
    $string .= " $$w[$_]";
  }
  $string .= ' ] . ';
  if ($s->{dot} == $#{$w}+1) {
    $string .= '$';
  } else {
    $string .= "$$w[$s->{dot}]";
  }
  if($opt_b) {
    $string .= sprintf ' (%.2e)', $$s{p};
  }
  return $string;
}


# edge labels that cannot occur more than once for the same regent
our %complement_labels;
for( qw(AUX DET OBJA OBJA2 OBJD OBJG OBJC OBJP PRED SUBJ)) {
  $complement_labels{$_}++;
}

# Would action A, applied to state S, create a tree
# with duplicate complements?
sub check_duplicates {
  my($s,$a) = @_;

  # Does the transition in volve labeling at all?
  if($a !~ /\((.*)\)$/) {
    return 0;
  }

  # Is it a complement label?
  my $label = $1;
  if(!exists $complement_labels{$label}) {
    return 0;
  }

  # What is the intended regent?
  my $regent = 0;
  if($a =~ /^LEFT/) {
    $regent = $$s{dot};
  } elsif($a =~ /^RIGHT/) {
    $regent = $$s{stack}[0];
  }

  # Does this regent already have one of the same?
  if($regent) {
    return has_label($regent, $s, $label);
  } else {
    return 0;
  }
}

# Given the current state of affairs, decide on an action. Also
# returns the state that was actually used to make the decision, since
# the complete current state might not have occured in training.
sub decide {
  my($s,$w,$t) = @_;
  my $sig = encode($s, $w, $t);
  my $action;
  my $count = 0;

  # Find best match among the observed signatures
  until($count >= $threshold) {
    $count = $db{"${sig}total"} || 0;

    # back off if fewer than $threshold observations exist
    if($count < $threshold) {
      report 3, "WARNING: $sig: only $count observations\n";
      if(!reduce_context \$sig) {
        last;
      }
    }
  }

  # decide based on observations
  if($count) {

    # Normally, use the action observed most often.
    $action = $db{"${sig}best"};

    if($opt_C && check_duplicates($s,$action)) {

      # Use the best decision that does not involve duplicating complements.
      my $record = 0;
      for my $a (@actions) {
        $count = $db{"${sig}$a"};
        if($count &&
           $count > $record &&
           !check_duplicates($s,$a)) {
          ($action,$record) = ($a,$count);
        }
      }
    }
  }

  # decide without observations, just by looking at $s
  else {

    # use REDUCE if possible...
    if($$s{stack}[0]) {
      $action = 'REDUCE()'
    }

    # ...otherwise SHIFT
    else {
      $action = 'SHIFT';
    }
  }
  report 3, "decide() = $action\n";

  return ($action, $sig);

}

# determine accuracy
sub measure {

  my $s = shift;
  my $x = shift;
  return if exists $s->{no} && $s->{no} == 0;
  my @parsed = @{$s->{struc}};
  my @labels = @{$s->{labels}};

  my $correct =
    grep {
      $$x{annotated}[$_] eq $parsed[$_] ||
        ($opt_d && report(1,
               sprintf
               "ERROR: governor of `%s' is `%s', should be `%s'\n",
               $$x{words}[$_],
               $$x{words}[$parsed[$_]],
               $$x{words}[$$x{annotated}[$_]]) && 0)
      } (1..$#{@$x{words}});
  my $incorrect = $#{@$x{words}} - $correct;

  my $class = int(($correct + $incorrect) / 5);

  $accuracy{$class}{struc}{good} += $correct;
  $accuracy{total}{struc}{good} += $correct;

  $accuracy{$class}{struc}{bad} += $incorrect;
  $accuracy{total}{struc}{bad} += $incorrect;

  if (defined $correct) {
    printf "$$x{id}: %d/%d (%4.2f%%)",
      $correct, $correct+$incorrect, 100*$correct/($correct+$incorrect);
  }

  # determine labelled accuracy
    $correct =
      grep {
        ($$x{annotated}[$_] eq $parsed[$_] &&
         $$x{labels}[$_] eq $labels[$_])
      } (1..$#{@$x{words}});
  $incorrect = $#{@$x{words}} - $correct;

  $accuracy{$class}{labelled}{good} += $correct;
  $accuracy{total}{labelled}{good} += $correct;

  $accuracy{$class}{labelled}{bad} += $incorrect;
  $accuracy{total}{labelled}{bad} += $incorrect;

  if (defined $correct) {
    printf " / labelled: %d/%d (%4.2f%%)",
      $correct, $correct+$incorrect, 100*$correct/($correct+$incorrect);
  }

  print "\n";

}

# Create .cda from parse result
sub write_anno {

  my ($file,$sol,$id,$w,$t) = @_;
  open(ANNO, ">$file") or die "Can't write `>$file': $!\n";

  my $latid = $id;
  $latid =~ s/^auto-//;

  print ANNO "'$id' : '$latid' <->\n";
  for (1..$#$w) {
    printf ANNO "%d %d %s\n", $_-1, $_, print_identifier $$w[$_];
    print ANNO "$opt_c/'$$t[$_]'\n";
    print ANNO "'$opt_S' -> '$$sol{labels}[$_]' -> $$sol{struc}[$_]\n";

    if($opt_e) {
      for(split(/,/,$opt_e)) {
        print ANNO "'$_' -> '' -> 0\n";
      }
    }

    print ANNO ($_ == $#$w ? ';' : ',');
    print ANNO "\n";
  }
  close(ANNO);
}

# Write parse result as annotation
sub maybe_write {

  my ($s,$id,$w,$t) = @_;

  if ($opt_w) {
    write_anno("$id-sr.cda", $s, "$id-sr", $w, $t);
  }
}

# insert search state sorted by P into array ref
sub insert_sorted {
  my $a = shift;
  my $s = shift;

  my $i = 0;
  $i++ while $i <= $#$a && $$a[$i]{p} > $s->{p};
  if ($i < $agenda_size) {
    splice @$a, $i, 0, $s;
  }

}

# parse sentence, finding most probable tree
sub parse {

  my ($id,$w,$t,$ann,$l,$h) = @_;
  my @words = @$w;
  my @tags = @$t;
  my @annotated = @$ann;
  my @labelled = @$l;
  my @history = @$h;

  my @agenda = ();
  my $no = 0;
  my $solution;
  my $state_error = 0;

  if ($opt_d) {
    $opt_v = 2;
  }

  # initial state
  push @agenda,
    {
     struc => [(0) x ($#words + 1)],
     labels => [('') x ($#words + 1)],
     stack => [],
     p => 1.0,
     dot => 1,
     no => 0,
    };

  while (@agenda) {
    my $s = shift @agenda;

    report 2, "read:  ";
    report 1, print_state($s, $w) . "\n";
    next if $solution && $s->{p} < $solution->{p};

    my($action, $sig);
    if($opt_b) {
      $sig = encode($s,$w,$t);
    } else {
      ($action, $sig) = decide($s,$w,$t);
    }

    my $found = 0;
    for (@actions) {
      if($opt_b) {

        # when backtracking, perform duplicate check manually
        next if check_duplicates($s, $_);

      } else {

        # when parsing deterministically, only use the predicted action
        next if $action ne $_;
      }

      # When backtracking, ignore zero probability transitions
      # unless we have no other option (yet).
      my $count = $db{"$sig$_"} || 0;
      if($opt_b && $found && !$count) {
        next unless $count;
      }

      my $top = $s->{stack}[0];
      my $dot = $s->{dot};

      # Ignore physically impossible transitions no matter what. Such
      # events are never observed, of course, but they may be
      # introduced by backing off.
      next if $dot > $#words && !/^REDUCE/;
      next if !$top && !/^SHIFT/;

      my $total = $db{"${sig}total"};
      my $p = $total ? $count / $total : 0;
      report 2, "$sig: ";
      report 1, $_;
      report 2, sprintf(" (%.3f)", $p);
      my $suc = { struc => [@{$s->{struc}}],
                  labels => [@{$s->{labels}}],
                  stack =>  [@{$s->{stack}}],
                  dot => $s->{dot},
                  p => $s->{p} * $p,
                  no => ++$no
                };

      # transit
      if (/^RIGHT(\((.*)\))?$/) {
        my $label = $2 || '';
        report 1, " ($words[$top]<--$label--$words[$dot])";
        if ($opt_d) {
          if ($annotated[$dot] != $top) {
            print STDERR
              "WARNING: correct governor would be `$words[$annotated[$dot]]'\n";
          }
          if (!$opt_u && $labelled[$dot] ne $label) {
            print STDERR
              "WARNING: correct label would be `$label'\n";
          }
        }
        $suc->{struc}[$dot] = $top;
        $suc->{labels}[$dot] = $label;
        unshift @{$suc->{stack}}, $suc->{dot}++;
      } elsif (/^LEFT(\((.*)\))?$/) {
        my $label = $2 || '';
        report 1, " ($words[$top]--$label-->$words[$dot])";
        if ($opt_d) {
          if ($annotated[$top] != $dot) {
            print STDERR
              "WARNING: correct governor would be `$words[$annotated[$top]]'\n";
          }
          if (!$opt_u && $labelled[$top] ne $label) {
            print STDERR "WARNING: correct label would be `$label'\n";
          }
        }
        $suc->{struc}[$top] = $dot;
        $suc->{labels}[$top] = $label;
        shift @{$suc->{stack}};
      } elsif (/^REDUCE(\((.*)\))?/) {
        my $label = $2 || '';
        if(! $suc->{labels}[$top]) {
          $suc->{labels}[$top] = $label;
        }
        shift @{$suc->{stack}};
      }

      elsif (/^SHIFT$/) {
        unshift @{$suc->{stack}}, $suc->{dot}++;
      }

      report 1, "\n";

      # annotate wrong transitions
      if ($opt_d && !$state_error) {
        my $h = shift @history;
        if ($_ ne $h) {
          print STDERR "WARNING: correct action would have been $h, not $_.\n";

          # after a wrong transition, the transition history gets
          # out of sync, and we can no longer reliably detect wrong
          # transitions just by comparing to @history.
          # Therefore we disable transition debugging altogether.
          $state_error = 1;
        }
      }
      $found++;

      # end state?
      if ($suc->{dot} > $#words &&
          (!$solution || $solution->{p} < $suc->{p})) {

        # promote search state to solution
        $solution = $suc;
        report 1, sprintf("Found solution with p = %.3e\n", $suc->{p});
      }

      # sort new state into agenda
      else {
        insert_sorted \@agenda, $suc;
        if($opt_b) {
          report 2, "wrote: " . print_state ($suc, $w) . "\n";
        }
      }
    }
    unless($found) {
      warn "ARRGH! No transition possible for $sig in `$id'.\n";
      return undef;
    }
  }

  return $solution;

}

# reconstruct shift/reduce history
sub deparse {

  # make statistics about transition frequency
  sub count {
    my $action = shift;
    my $sig = shift;
    my $x = shift;

    if($opt_u) {
      $action =~ s/\(.*\)$//;
    }

    if ($opt_d) {

      # don't write to @observations,
      # just record the individual parse history
      push @{$x->{history}}, $action;

    } else {

      # count the transition for later
      $observations[$maxdim]{$sig}{$action}++;
      report 2, "observed  [$sig]:" . $action. "\n";
    }
  }

  my $x = shift;
  my $s =
    {
     dot => 1,
     stack => [],
     struc => [ (0) x ($#{@$x{words}}+1) ],
     labels => [ ('') x ($#{@$x{words}}+1) ],
     anno => $x
    };

  check_anno $x;

  while ($$s{dot} <= $#{@$x{words}}) {

    report 2, print_state ($s, $$x{words}) . "\n";

    my $sig = encode($s, $$x{words}, $$x{tags});

    my $top = $$s{stack}[0];
    my $dot = $$s{dot};

    # detect left action
    if ($top && $$x{normalized}[$top] == $dot) {
      my $label = $$x{labels}[$top];
      report 1, "$$x{words}[$top]--$label-->$$x{words}[$dot]\n";
      count("LEFT($label)", $sig, $x);
      $labels{$label} = 1;
      $$s{struc}[$top] = $dot;
      shift @{$$s{stack}};
      next;
    }

    # if the top stack word is already subordinated,
    # and has no normal (projective) dependents on the right,
    # reduce it
    if ($top && $$x{normalized}[$top] < $top) {
      my $deps = 0;
      for ($dot..$#{@$x{words}}) {
        if ($$x{projective}[$_] && $$x{normalized}[$_] == $top) {
          $deps = 1;
          last;
        }
      }
      if (!$deps) {
        my $label = $$x{labels}[$top];
        count("REDUCE($label)", $sig, $x);
        shift @{$$s{stack}};
        next;
      }
    }

    # if the next word has a dependent on the left,
    # but that dependent is buried in the stack,
    # then it must be nonprojective.
    # Postulate a special transition.
    my $deps = 0;
    for (@{$$s{stack}}) {
      if ($$x{normalized}[$_] == $dot &&
          !$$s{struc}[$_]) {
        if ($opt_s) {
          count('FARLEFT', $sig, $x);
          report 1, "far left attachment needed for `$$x{words}[$_]'\n";
          $$s{struc}[$_] = $dot;
        } else {
          report 1, "WARNING: forcing `$$x{words}[$_]' to NIL\n";
          $$x{normalized}[$_] = 0;
        }
        next;
      }
    }

    # detect right action
    if ($top && $$x{normalized}[$dot] == $top) {
      my $label = $$x{labels}[$dot];
      report 1, "$$x{words}[$top]<--$label--$$x{words}[$dot]\n";
      count("RIGHT($label)", $sig, $x);
      $labels{$label} = 1;
      $$s{struc}[$dot] = $top;
      unshift @{$$s{stack}}, $$s{dot}++;
      next;
    }

    # if the next word is right attached,
    # but its attachment point is not available
    # (either already gone from the stack or buried in it),
    # then it must be nonprojective.
    # Postulate a special transition.
    if (right($x,$dot)) {

      if ($opt_s) {
        report 1, "far right attachment needed for `$$x{words}[$dot]' ($dot)\n";
        count('FARRIGHT', $sig, $x);
        $$s{struc}[$dot] = $$x{normalized}[$dot];
      } else {
        report 1, "WARNING: forcing `$$x{words}[$dot]' ($dot) to NIL\n";
        $$x{normalized}[$dot] = 0;
        count('SHIFT', $sig, $x);
      }
      unshift @{$$s{stack}}, $$s{dot}++;
      next;
    }

    # if stack is empty, shift
    if (!$top) {
      count('SHIFT', $sig, $x);
      unshift @{$$s{stack}}, $$s{dot}++;
      next;
    }


    # if the attachment is farther to the right, shift
    if ($$x{normalized}[$dot] > $dot || $$x{normalized}[$dot] == 0) {
      count('SHIFT', $sig, $x);
      unshift @{$$s{stack}}, $$s{dot}++;
      next;
    }

    # if we can't recreate this word's edge, pretend it is a root edge
    print STDERR "WARNING: Can't deal with word $$x{words}[$dot] in $$x{id}!\n";
    $$x{normalized}[$dot] = 0;
    unshift @{$$s{stack}}, $$s{dot}++;
  }

  # check to see whether the parsed structure is imperfect
  for (1..$#{@$x{words}}) {
    if ($$x{normalized}[$_] != $$s{struc}[$_]) {
      warn "ERROR: $$x{words}[$_] has governor $$s{struc}[$_]," .
        " should be $$x{normalized}[$_] in $$x{id}\n";
    }
  }
}


# Fill @annos with perl versions of inputCurrentGrammar->annotations
sub import_annos {

  @annos = ();

  # translate
  my $all = InputStruct_annotations_get($inputCurrentGrammar);
  for (my $l = hashListOfKeys $all; $l; $l = ListStruct_next_get $l) {
    my $ae = hashGet($all, ListStruct_item_get($l));
    my $x = read_anno $ae;
    push @annos, $x;
  }
  @annos = sort { byVName($$a{id}, $$b{id}) } @annos;

}

# Load one .cda file
sub load_cda {
  my $file = shift;
  print "$file\r";
  commandEval 'set verbosity 0';
  commandEval "load $file";
}

# Call tagger on all annos in one go and set ->{tags}
sub tag_annos {

  my @allwords;
  my @alltags;

  # collect words
  for my $x (@annos) {
    push @allwords, '';
    push @allwords, @{$x->{words}};
  }

  report 1, "tagging $#allwords words...\n";

  # tag
  my $out = "/tmp/$$.out";
  my $in = "/tmp/$$.in";
  my $error = "/tmp/$$.error";
  open(OUT, ">$out") or die "Can't write `$out': $!\n";
  for (@allwords) {
    next if /^NIL$/;
    print OUT "$_\n";
  }
  close(OUT);
  system("deutsch-tagger.pl < $out > $in 2>$error")
    and die "Can't call deutsch-tagger.pl: ".`cat $error`;
  unlink $error;
  open(IN, "$in")
    or die "Can't read `$in': $!\n";

  while(<IN>) {
    if(/^[^\t]+\t([^\t]+)/) {
      unshift @alltags, $1;
    }
  }
  close(IN);

  # distribute tags
  for my $x (@annos) {
    for (1..$#{$$x{words}}) {
      $$x{tags}[$_] = pop @alltags;
#      print "$$x{words}[$_]/$$x{tags}[$_]\n";
    }
  }
}



# Do our stuff in blocks of $block_size.
#
# Rationale: we don't want to load, tag etc. each annotation
# individually, because starting the tagger process is much more
# expensive than the actual parsing. But on the other hand, we don't
# want to have to store 100,000 annotations simultaneously, because
# that might overflow the RAM, and slows down inputLoad()
# quadratically. A block size of 1000 seems adequate.
sub process_batch {

  my $batch = shift;

  return unless @$batch;

  # read annotations from CDG
  print "Loading...\n";
  for(@$batch) {
    load_cda $_;
  }

  # import to Perl
  print "Importing...\n";
  import_annos();
  commandEval 'reset';

  # maybe re-tag
  if ($opt_r && $opt_p) {
    print "Tagging...\n";
    tag_annos();
  }

  # process
  if ($opt_p) {

    # create structure for sentence
    print "Parsing...\n";
    for my $x (@annos) {
      if ($opt_d) {
        deparse $x;
      }
      my $sol = parse($$x{id}, $$x{words}, $$x{tags}, $$x{annotated}, $$x{labels}, $$x{history});
      if ($sol) {
        measure $sol, $x;
        maybe_write $sol, $$x{id}, $$x{words}, $$x{tags};
      }
    }

  } else {

    # count decisions in parse
    print "Deparsing...\n";
    for my $x (@annos) {
      deparse $x;
    }
  }
  @$batch = ();

  @annos = ();
}

# Parse annotations. Words are taken from .cda files, and the
# structure in them is considered `correct' for the purpose of judging
# accuracy. The tags specified in the annotation are used, which means
# that the experiment is unrealistic unless they have been re-computed
# with the -r flag.
sub process_annos {
  my $specs = shift;
  my $i = 0;
  my @all;
  for my $spec (split(/,/, $specs)) {

    my ($min, $max, $corpus) = (0,1000, $spec);
    if ($spec =~ /^(.+):(\d+)-(\d+)/) {
      ($corpus,$min,$max) = ($1,$2,$3);
    } elsif ($spec =~ /^(.+):(\d+)/) {
      ($corpus,$min,$max) = ($1,$2,$2);
    } else {
      ($corpus,$min,$max) = ($spec, 1, 0);
    }
    my $dir = "$opt_D/$corpus";
    if(!-d $dir) {
      die "Cant find corpus `$corpus'!\n";
    }
    my $limit = $max ? $max : 1e8;
    for my $i ($min..$limit) {
      my $subdir = sprintf "%03d", ($i / 10000);
      my $file= "$dir/$subdir/$corpus-s$i.cda";
      if (! -f $file) {
        $file = "$dir/$subdir/auto-$corpus-s$i.cda";
      }
      if (! -f $file) {
        if (0 == $max) {
          last;
        } else {
          next;
        }
      }

      if ($opt_n) {
        last if ++$i > $opt_n;
      }
      push @all, $file;
      if($#all == $block_size - 1) {
        process_batch \@all;
      }
    }
  }
  process_batch \@all;
}

# Communicate with CDG
sub interactive {

  our @words = (undef);
  our @tags = (undef);
  our $i = 0;

  sub parse_one_sentence {

    $i++;
    my $id = "interactive$i";

    # parse words
    my $sol = parse($id, \@words, \@tags, [], [], []);
    maybe_write $sol, $id, \@words, \@tags;

    # print output to CDG
    for (1..$#words) {
      my $gov = $$sol{struc}[$_];
      my $lab = $$sol{labels}[$_] || '<empty>';
      report 3, "wrote `$words[$_]\tgov\t$gov\tlab\t$lab\n";
      print "$words[$_]\tgov\t$gov\tlab\t$lab\n";
    }
  }

  # read lines printed by CDG
  while (<>) {
    chomp;

    # ignore metadata
    next if /^\#[a-z]+:/;

    report 2, "read `$_'\n";

    if(/^$/) {
      parse_one_sentence();
      @words = @tags = (undef);
      next;
    }

    my($word, $tag) = /^([^\t]+)\t([^\t]+)/;
    if ($word) {
      push @words, $word;
      push @tags, $tag;
    }
  }
  if($#words) {
    parse_one_sentence();
  }

  # display
  if ($opt_x) {
    my $id = 'interactive1';
    system "cdg -x -e 'showparse $id-sr' $id-sr.cda";
  }

}

# Print out lattice hash
sub dump_lattice {
  my $lat = shift;
  print "$$lat{id}\n";
  for(1..$#{$$lat{words}}) {
    print "$$lat{words}[$_]/$$lat{tags}[$_] ";
  }
  print "\n";
}

# translate LatticeStruct to perl hash
sub read_lattice {
  my $lat = shift;
  my $id = LatticeStruct_id_get $lat;

  my @words = (undef);
  for (my $l = LatticeStruct_arcs_get $lat; $l; $l = ListStruct_next_get $l) {
    my $arc = ListStruct_item_get $l;
    push @words, ArcStruct_word_get $arc;
  }

  return { id => $id, words => [ @words ] };
}

# Fill @lattices with perl versions of inputCurrentGrammar->lattices
sub read_lattices {

  @lattices = ();
  my $all = InputStruct_lattices_get($inputCurrentGrammar);
  for (my $l = $all; $l; $l = ListStruct_next_get $l) {
    my $lat = ListStruct_item_get $l;
    my $x = read_lattice $lat;
    push @lattices, $x;
  }
}


# Parse file of CDG lattices
sub do_lattices {
  my $file = shift;
  commandEval "load $file";
  read_lattices();

  # construct bulk tagger call
  my $tmpname = "/tmp/$$.tok";
  open(TMP, ">$tmpname") or die "Can't write `$tmpname': $!\n";
  for my $lat(@lattices) {
    for(@{$lat->{words}}) {
      next unless defined $_;
      print TMP "$_\n";
    }
    print TMP "\n";
  }
  close(TMP);
  open(IN, "deutsch-tagger.pl -q $tmpname |")
    or die "Can't call deutsch-tagger.pl: $!\n";

  my $sentence_no = 0;
  my $word_no = 1;
  while(<IN>) {
    if(/^[^\t]+\t([^\t]+)/) {
      $lattices[$sentence_no]{tags}[$word_no] = $1;
      ++$word_no;
    } else {
      $word_no = 1;
      ++$sentence_no;
    }
  }

  close(IN);
  unlink $tmpname;

  for my $lat(@lattices) {

    # shift-reduce parsing is rather unreliable. If there already is
    # an annotation, even an automatically created one, we are
    # unlikely to do better, so don't overwrite it.
    my $newname = "auto-$$lat{id}.cda";
    next if -f $newname;

    my $sol = parse($$lat{id}, $$lat{words}, $$lat{tags}, [], [], []);
    if($sol) {
      write_anno($newname, $sol, "auto-$$lat{id}", $$lat{words}, $$lat{tags});
    }
  }
}


# create new parse move database
sub write_database {

  # create file
  my %db;
  tie %db, 'DB_File', $opt_t or die "Can't write `$opt_t': $!\n";

  compute_actions(keys %labels);

  # create backoff states and write statistics
  my $i = $maxdim;
  while ($i >= 0) {

    # write n-ary transitions
    for my $sig (sort keys %{$observations[$i]}) {
      report 3, "Writing observation $sig\n";
      my $max = 0;
      my $total = 0;
      my $best;
      for my $act (@actions) {
        my $count = $observations[$i]{$sig}{$act} || 0;
        if ($count) {
          $total += $count;
          if ($count > $max) {
            $best = $act;
            $max = $count;
          }
          report 4, "Writing key $sig$act\n";
          $db{"$sig$act"} = $count;
        }
      }

      if($total) {
        report 4, "Writing key ${sig}best\n";
        $db{"${sig}best"} = $best;
        report 4, "Writing key ${sig}total\n";
        $db{"${sig}total"} = $total;
      }
    }

    # create (n-1)-ary backoff transitions from n-ary transitions
    if ($i > 0) {
      while (my ($sig, $dist) = each %{$observations[$i]}) {

        my $newsig = $sig;
        reduce_context \$newsig;

        for (keys %$dist) {
          $observations[$i-1]{$newsig}{$_} += $$dist{$_};
        }
      }
    }
    $i--;
  }

  # global statistics
  $db{invocation} = $invocation;
  $db{maxdim} = $maxdim;
  $db{labels} = join('+', keys %labels);

  untie %db;

}

# compute @actions from @moves
sub compute_actions {
  if($opt_u) {
    @actions = @moves
  } else {
    my @labels = @_;
    @actions = qw (SHIFT);
    for my $a (qw(LEFT RIGHT REDUCE)) {
      for my $l (@labels) {
        push @actions, "$a($l)";
      }
    }
  }
}

# read previously built transition table
sub open_database {

  # search in the CWD
  my $file = $opt_p;
  if (! -r $file && -r "$file.sr") {
    $file = "$file.sr";
  }

  # search alongside this script
  if (! -r $file) {
    (my $here = $0) =~ s!/[^/]+$!!;
    if (-r "$here/$file") {
      $file = "$here/$file";
    } elsif (-r "$here/$file.sr") {
      $file = "$here/$file.sr";
    }
  }

  # search in $CDGDIR
  if (! -r $file && exists($ENV{CDGDIR})) {
    if (-r "$ENV{CDGDIR}/lib/$file") {
      $file = "$ENV{CDGDIR}/lib/$file";
    } elsif (-r "$ENV{CDGDIR}/lib/$file.sr") {
      $file = "$ENV{CDGDIR}/lib/$file.sr";
    }
  }

  if (! -r $file) {
    die "No such file: $file\n";
  }

  # open file
  tie %db, 'DB_File', $file, O_RDONLY
    or die "Can't open `$file': $!\n";

  report 2, "reading transition table `$file'\n";

  # recreate the settings in effect when this table was created
  my $inv = $db{invocation};
  if ($inv =~ / -a\b/) {
    $opt_a = 1;
  }
  if ($inv =~ / -f *(\S+)/) {
    $opt_f = $1;
  }
  if ($inv =~ / -i\b/) {
    $opt_i = 1;
  }
  if ($inv =~ / -u\b/) {
    $opt_u = 1;
  }
  if ($inv =~ /-l *(\d+)/) {
    $opt_l = $1;
  }
  if ($inv =~ /-L *(\d+)/) {
    $opt_L = $1;
  }

  $maxdim = $db{maxdim};

  # create extended set of actions
  compute_actions(split(/\+/, $db{labels}, -1));

  # Since the empty string can be a label, split() must be given the
  # argument -1, which means `do not omit the last field just because it's
  # empty'.

}

### MAIN ###

if ($opt_p) {
  open_database();
} else {

  # detect attempt to overwrite existing table
  if (-f $opt_t) {
    die "ERROR: won't overwrite transition table `$opt_t'.\n";
  }
}

# iterate over input
my $specs = shift;

if ('-' eq $specs) {
  interactive();
} elsif($specs =~ /\.lat$/) {
  do_lattices($specs)
} else {
  process_annos($specs);
}

# Evaluate parsing accuracy
sub statistics {
  my ($class, $type) = @_;
  my $description = $class;

  if($class =~ /^(\d+)$/) {
    my $from= 1 + $class * 5;
    my $to = $from + 4;
    $description = "$from - $to";
  }

  if (defined $accuracy{$class}{$type}{good}) {
    my $g = $accuracy{$class}{$type}{good};
    my $b = $accuracy{$class}{$type}{bad};

    printf "%20s: %12s (%4.2f%%)",
      "$description $type", "$g/".($g+$b), 100*$g/($g+$b);
  }
}

if ($opt_p) {

  for(sort { length $a <=> length $b or $a <=> $b }
      keys %accuracy) {
    statistics($_, 'struc');
    print "   ";
    statistics($_, 'labelled');
    print "\n";
  }

}

# write data base
if (!$opt_p) {
  print "Creating table...\n";
  write_database();
}
