#!/usr/bin/env perl
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Decide what the word boundaries are in a text and put each word on a
# line by itself.
#
# TODO: some greater change would be helpful:
# general recognition of hints for sentence begin or against it
# (tokenizer doesn't reconize sentence ends after abbreveations, version numbers and (roman) ordinal number
use warnings;
use strict;

use Getopt::Std;
our $opt_G;
our $opt_d;
our $opt_g;
our $opt_h;
our $opt_l;
our $opt_m = -1;
our $opt_p;
our $opt_t;
our $opt_w;
our $opt_x;

getopts("Gdghlm:p:stw:x:");

die "
Usage: $0 [input file]

       -G  use CDG::German to perform German-specific corrections
       -d  Debug
       -g  treat some 's as anglo-saxon Genitive (default: always separate)
       -h  show this Help
       -l  use Line-based output format
       -m  Maximum number of sentences
       -p  specify Prefix for lattice names
       -t  Trust the input about sentence boundaries:
           assume that every paragraph contains exactly one sentence.
       -w  specify file of tokens with Whitespace in them
       -x  warn if sentence length eXceeds this threshold

" if $opt_h;

our $prefix = $opt_p ? "$opt_p-" : '';

our $input_line_no = 0;

if($opt_G) {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0);
  require 'cdgsetup.pl';
  require CDG::German;
  import CDG::German qw(all_cats deutsch_lc);
}


# Some words, usually proper names, have whitespace within them, but
# should be treated as one word anyway. Usually these sequences that
# are clearly recognisable, since otherwise they would be very
# misleading even for the human reader. Therefore we allow the user to
# specify a file of such sequences, and each occurrence of them will
# be reassembled upon sight.
our @whitespace_tokens = ();
if($opt_w && -r $opt_w) {
  open(IN, "$opt_w") or die "Can't read `$opt_w': $!\n";
  while(<IN>) {
    chomp;
    (my $new = $_) =~ s/ /INTERNALWHITESPACE/g;
    push @whitespace_tokens, { from => $_, to => $new };
  }
  close(IN);
}

our $exclaim_names = join '|',
  qw (
      ACT Bat Yahoo Fritz More Live 3Dnow 3DNow PowerNow Powernow FX32 DVDit Jamba
      NewHooOn Now SimNow Snap SCORE click FREE Spacewar Ur-Spacewar BlueFritz Esc
     );

our $abbrevs = join '|',
  qw (
      Jan Feb Mar Apr Jun Jul Aug Oct Nov Dec
      Mo Di Mi Do Fr Sa

      .+/[mM]in Abs AZ Az B/s Bd BGBl Chr Co Corp Dipl Dr II Inc Jr Ltd Kl Max
      Mill Min Mio Mr Mrd Mw[Ss]t Nr Par Pf Proc Prof St (.+-)?Str Tel Ver Vol
      [A-Z] [A-Z]\.-[A-Z] al bzw ca co div entspr etc evtl excl
      exkl \d*ff ggf incl inkl jr lt max min mind s [A-Za-z]+str usf usw
      u v vgl vs www zzgl
     );

# These common abbreviations should be treated as if the full words
# had been written; for instance, `z.B.' should be tokenized into two
# words, so that it can function as a normal PP. In contrast, things
# such as `S.W.I.F.T.' are supposed to be a single word.
our $multi_abbrevs = join '|',
  qw ( d.h. e.V. s.a. u.a. z.B. o..);

# Company names that are entirely lower case. `eBay' is not a Problem,
# but `adidas' looks like a certain indicator against a sentence
# break. This list helps us avoid that mistake.
our $small_names = join '|',
  qw (
      1822direkt 3dfx adidas dpa h2g2 heise i-mode id t@x tesion
      mediantis
      debitel speedlink
     );

our $Monate = join '|',
  qw ( Januar Februar Mrz April Mai Juni Juli
       August September Oktober November Dezember );

our $ordinal_nouns = join '|',
  qw ( Auflage Bericht Domain Entwurf Etage Etappe
      Meisterschaft Generation Gipfel Jahr Jahrhundert
      Kammer Kl Klasse Konferenz Liga ^Mal Minute Mitarbeiter
      Mitglied OG Obergeschoss Order Platz
      Quartal Programm Rang Runde Semester
      Senat Sinfonie Sitzung Stelle
      Stock Stockwerk Strafkammer Studie Symphonie
      Tag Tagung Update Veranstaltung
      Verordnung Version Vertrag Wettbewerb
      Zivilkammer

      Berliner Bremer Hamburger
      Inernationale Internationalen
      Deutschen
      );


# internal marker for end of paragraph
our $EOP = '##EOP##';

our @files = ();

our $lastsentence = '';

our $counter = 0;

our $max_length = 500;
$max_length = $opt_x if defined $opt_x;

# build list of files to tokenize
for my $what (@ARGV) {

  if ( -d $what ) {

    # read each specified file
    opendir(DIR, $what);
    for my $file (readdir(DIR)) {
      push @files, "$what/$file";
    }
    closedir(DIR);
  } elsif ( -f $what ) {
    push @files, $what;
  }
}

# array of tentative tokens as detected by preprocess()
our @buffer;
our $eof = 0;

# the sentence under construction
our @sentence = ();

# process list
our $prev = '';
our $last = '';
our $this;
our $next;

# lists of words used to make tokenization more efficient
my $Monat_re = qr "^($Monate)$";
my $num_re = qr '^[-+]?[0-9.,]+-?$';
my $abbrev_re = qr "^($abbrevs)$";
my $multi_abbrev_re = qr "^($multi_abbrevs)$";
my $small_names_re = qr "^($small_names)$";
my $web_re = qr "([a-z0-9-]+\.)+(com|org|net|gov|mil|edu|de)";
my $ordinal_nouns_re = qr "($ordinal_nouns)(s|es)?$"i;

my $single_quotes = 0;
my $double_quotes = 0;
my $parentheses = 0;
# FIXME: detect ridiculously long parenthesized streches and stop them.

my $sentence_start = '';


# Maybe produce debug output.
sub debug {
  print STDERR "@_\n" if $opt_d;
}


# Produce output for the passed array and empty it.
sub write_sentence ($) {
  my $ref = shift;
  my @x = @$ref;
  return unless @x;

  if($#x >= $max_length) {
    warn "WARNING: Sentence with $#x words: " . join(" ", @x) . "\n";
  }

  # detect sentences that consist only in a quotation mark; this is a
  # sure sign that the computation of quotation levels has been thrown
  # off, probably because of inconsistent quotation marks in the input.
  if(1 == @x && $x[0] =~ /^[\'\"]$/) {
    warn "WARNING: possibly incorrect quotation mark in line $input_line_no\n";
    @$ref = ();

    if($x[0] eq '"') {
      $double_quotes = 1 - $double_quotes;
    } else {
      $single_quotes = 1 - $single_quotes;
    }

    return;

    # FIXME: Instead of suppressing this quote we should reattach it
    # to the previous sentence. This, of course, requires that we do
    # not print sentences at once, but maintain a queue of them.
  }

  # Yet another abbreviation heuristic: Although a sentence may end
  # with an abbrevation, it is very unlikely that it is an ordinal
  # abbrevation. Therefore `3.' is split if it is the last word.
  if ($x[-1] =~ /^(\d+|I+)\.$/ && $#x > 1) {
    pop @x;
    push @x, $1;
    push @x, '.';
  }

  # restore stuff that was neutralized before tokenizing
  for (@x) {

    s/EMOTICON_SMILE/:-)/g;
    s/EMOTICON_FROWN/:-(/g;
    s/INTERNALELLIPSIS/.../g;
    s/INTERNALWHITESPACE/ /g;

  }

  if($opt_l) {
    print join(" ", @x);
    print "\n";
  } else {
    print join("\n", @x);
    print "\n\n";
  }

  $lastsentence = "@x";

  @$ref = ();

  $counter++;

  exit 0 if $counter == $opt_m;

}

# Return number of parenthesis level that this word introduces:
#
#  der             --> 0
#  (               --> 1
#  )               --> -1
#  Mitarbeiter(in) --> 0
sub paren_count {
  my $w = shift;

  if ($w =~ /[:;]-[()]/) {

    # emoticons do not count!
    return 0;
  }

  my $result = ($w =~ tr/(/(/);
  $result += ($w =~ tr/[/[/);
  $result -= ($w =~ tr/)/)/);
  $result -= ($w =~ tr/]/]/);
  return $result;
}

# Read lines of input and apply some initial transformations. Note
# that line breaks can intervene between any two words, so most
# transformations must be done later. At this stage we can only
# apply those transformations that insert space, or those that apply
# to patterns that cannot possibly be broken across lines.
#
# TODO: While inserting space the tokenizer doesn't remember at
# which side the interpunction was cut of
# and thus can't make use of this knowledge
sub preprocess ($) {
  $_ = shift;
  unless(defined $_) {
    $eof = 1;
    return undef;
  }

  chomp;
  $input_line_no++;

  debug "preprocess($_)...";
  s/^\s+//g;
  s/\s+$//g;

  # Bad Nauheim --> 'Bad Nauheim'
  for my $w (@whitespace_tokens) {
    s/$$w{from}/$$w{to}/g;
  }

  # normalize foreign quotation marks
  s/[]/"/g;

  # eliminate some emoticons that would majorly confuse the scanner
  s/[:;]-\)/EMOTICON_SMILE/g;
  s/[:;]-\(/EMOTICON_FROWN/g;

  # notice sentence-internal ...
  s/([A-Za-z])\.\.\.([A-Za-z])/$1 INTERNALELLIPSIS $2/g;

  # reassemble consecutive full stops
  s/\.(\s*\.)+/ ... /g;

  # ja---aber --> ja - aber
  s/--+/ - /g;

  # insert space between words and punctuation:
  # (Wohin steuern die USA?). --> ( Wohin steuern die USA ? ) .

  # Vorwrts! --> Vorwrts !
  s/(.*?)([.,;:!?]|\.\.+|!!+|\?\?+)(\s+|\z|[)\]"`',])/$1 $2 $3/g;

  # Isolate quotation marks from the words that they mark:
  # "put options" --> " put options "
  1 while s/(\s|^)([`'"(\[])(\S)/ $2 $3/g;
  1 while s/(\S)([`'")\]])(\s|$)/$1 $2 /g;

  # But some German words legitimately start with apostrophes,
  # such as 'n (ein) or 's (es), so we reassemble those.
  s/\' +(n|s|ne|nem)\b/\'$1/g;

  # An entirely empty line is probably a paragraph break,
  # so we insert a marker for the sentence delimiter.
  s/^\s*$/$EOP/g;

  my @line = split /\s+/;
  for my $word (@line) {
    push @buffer, $word unless $word =~ /^\s*$/;
  }

  debug "...result: @buffer";

}

# Return the next tentative token
sub get_word() {
  while(!@buffer && !$eof) {
    preprocess(<>);
  }
  if(!@buffer) {
    return undef;
  }
  my $w = shift @buffer;
  return $w;
}

# Stuff a token back onto the input
sub unget_word($) {
  my $w = shift;
  unshift @buffer, $w;
}

# Preview a future word word but do not actually read it.
sub preview_word($) {
  my $n = shift;
  my $w;
  my @tmp;
  for(1..$n) {
    $w = get_word();
    push @tmp, $w;
  }
  unshift @buffer, @tmp;
  return $w || '';
}


# Preview what the next non-punctuation word will be.
sub find_next_proper_word () {
  my @preview;
  my $w = '';
  while ($w !~ /[A-Za-z0-9]/) {
    $w = get_word();
    last unless $w;
    push @preview, $w;
  }
  for(reverse @preview) {
    unget_word($_);
  }
  return $w;
}

# Indicate a paragraph break in the output.
sub write_paragraph () {
  return if $last eq $EOP;
  print "\n";
}

# Combine successive ! or ? into one token
sub combine_exclaim {
  my $this = pop @sentence;
  my $type = $this;
  while(preview_word(1) eq $type) {
    get_word();
    $this .= $type;
  }
  push @sentence, $this;
}

# When more than one bit of punctuation occurs at the end of a
# sentence, add all following bits in front of the EOS marker.
sub push_punctuation {
  debug "push_punctuation";
  while (1) {

    my $next = preview_word(1);
    last unless defined $next;

    # add " to the current sentence if we are within quoted
    # material, otherwise to the next one.
    if ($next eq '"' && $double_quotes == 1 && $this !~ /^:$/) {
      push @sentence, '"';
      debug "push(\")";
      get_word();
      $double_quotes = 0;
    } elsif ($next eq '\'' && $single_quotes == 1) {
      debug "push(\')";
      push @sentence, '\'';
      get_word();
      $single_quotes = 0;
    } elsif ($next eq '.') {
      debug "push(.)";
      push @sentence, '.';
      get_word();
    } elsif ($next eq ')' && $this !~ /^!$/) {
      push @sentence, ')';
      debug "push(\))";
      $parentheses--;
      get_word();
    } elsif ($next eq '!') {
      push @sentence, '!';
      debug "push(\!)";
      get_word();
    } else {
      last;
    }
  }
}

# MAIN Tokenize begins
#
# Now apply transformations that need context to either side. This
# stage transforms the token stream so that each item corresponds to
# exactly one word in the final lattice. We also decide where the
# sentences boundaries will be.

WORD:
while (defined ($this = get_word())) {

  debug "scan($this)";

  $next = preview_word(1);

  # An end-of-paragraph marker might have been inserted by
  # preprocessing. Obey it.
  if ($this eq $EOP) {
    write_sentence(\@sentence);
    write_paragraph();
    $sentence_start = $next;
    next;
  }


# HYPHENS

  # Newspaper text often contains hyphenated words that are broken
  # across lines at the hyphen. If they are rearranged, they are
  # often wrongely broken into two: Ost -Berlin. Detect this.
  if ($this =~ /^-[A-Z]/ &&
      $last =~ /^["'A-Z].*[a-z"']$/) {
    warn "WARNING: broken hyphen: $last $this in line $input_line_no\n";
    my $prev = pop @sentence;
    push @sentence, "$prev$this";
    next WORD;
  }

  if($opt_G) {

    # If one of two adjacent words in unknown, but their concatenation
    # is known, it might be a hyphenation error.
    if($last && $this =~ /^[a-z]/i &&
       ( !all_cats($last) && !all_cats(deutsch_lc($last)) ||
         !all_cats($this) && !all_cats(deutsch_lc($this)) ) &&
       all_cats($last . $this)) {
      warn "WARNING: possible hyphenation error `$last $this' in line $input_line_no\n";
      my $prev = pop @sentence;
      push @sentence, "$prev$this";
      next WORD;
    }

    # Try to undo clitic `s'
    if($this =~ /^([a-z].{1,8}|Weil|Wenn)('?s)$/) {
      my ($root,$clitic) = ($1,$2);
      if(!all_cats($this) &&
         !all_cats(lc $this) &&
         all_cats($root)) {
        warn "WARNING: possible clitic `$this' in line $input_line_no\n";
        push @sentence, $root;
        push @sentence, $clitic;
        next WORD;
      }
    }
  }

  # The opposite error also often happens: Ost- Berlin.
  if($last =~ /[A-Za-z0-9"']-$/ &&
     $this =~ /^[A-Z]/ &&

     # But do not use this heuristic when sections are entirely in
     # uppercase, like this: DER KAPITAL- UND ZAHLUNGSVERKEHR.
     # But detect: CD-ROM.
     (($last.$this) =~ /[a-z]/ ||
     $last =~ /(CD|DVD)/)) {
    warn "WARNING: broken hyphen: $last $this in line $input_line_no\n";
    my $prev = pop @sentence;
    push @sentence, "$last$this";
    next WORD;
  }

# NUMBERS

  # detects lines of numbers which are broken through white spaces
  if ($last =~ /^\d\d+$/ && $this =~ /^\d+$/){
    if ($last !~ /[12]\d{3}/) {
      my $prev = pop @sentence;
      push @sentence, "$prev$this";
      debug "reassemble line of numbers";
      next WORD;
    }
  }


# QUOTES

  # detect "
  if ($this =~ /^[^\"]*\"[^\"]*$/) {
    if ($last =~ /^[A-Z].*-".*$/) {
      my $prev = pop @sentence;
      push @sentence, "$prev$this";
      debug "reassemble quoted words";
      next WORD;
    }

    # detect quoted words like " Tastless"-Angebot
    elsif ($next =~ /^[A-Z][a-z]*"-.*$/) {
      push @sentence, "$this$next";
      debug "reassemble quoted words";
      get_word();
      next WORD;
    }

    # The token `15"-Monitor' does not contain a quotation mark,
    # therefore do not count double_quotes
    elsif ($this =~ /\d\"-/) {
    }

    else {
      $double_quotes = 1 - $double_quotes;
      debug "double_quotes == $double_quotes";
    }
  }

  # detect '
  elsif ($this eq "'") {
    my $prev = $sentence[-2] || '';
    my $two = preview_word(2);

    # detect ' as a genitiv-marker
    if (0 == $single_quotes && defined $last &&
        $last =~ /^[A-Z][a-z]*(ch|[xcysz])$/ &&
	$next !~ /^\d\d$/ &&
	$prev !~ /^\'$/ && $two !~ /^\'$/) {
      my $prev = pop @sentence;
      push @sentence, "$prev$this";
      debug "reassemble genitive $prev$this";
      next WORD;
    }

    # detect tow-digit number of year: '99
    elsif($next =~ /^\d\d$/) {
      push @sentence, "$this$next";
      debug "reassemble number of year";
      get_word();
      next WORD;
    }

    # detect quoted words like Chip-'Foundry '
    elsif ($last =~ /^[A-Z].*-'.*$/) {
      my $prev = pop @sentence;
      push @sentence, "$prev$this";
      debug "reassemble quoted words";
      next WORD;
    }

    # detect quoted words like ' Tastless'-Angebot
    elsif ($next =~ /^[A-Z][a-z]*'-.*$/) {
      push @sentence, "$this$next";
      debug "reassemble quoted words";
      get_word();
      next WORD;
    }

    # do nothing; ' starting a paragraph when a quotation
    # is already running does not count as a delimiter,
    # but merely serves as a reminder,
    elsif($last eq $EOP && $single_quotes == 1) {
            debug "ignore redundant '";
    }

    # count single quote
    else {
      $single_quotes = 1 - $single_quotes;
      debug "single_quotes == $single_quotes";
    }
  }

  # detect `
  elsif ($this eq "`") {
    $single_quotes = 1 - $single_quotes;
    debug "single_quotes == $single_quotes";
  }

  # reassemble compounds with quotes in them
  #
  # Formations like `XML-"Sprache"' are often mistokenized because
  # the literal " is taken to be a single token. We reassemble them
  # here if a word with -" follows a single ".
  if ($this =~ /\"-/) {

    my $w = '';
    my $compound = $this;
    my @t = ();
    while (@sentence && $w !~ /^[,\.;:\"\'\`!?]+$/) {
      $w = pop @sentence;
      debug "w is: $w\n";
      push @t, $w;
      debug "t is: @t\n";
      $compound = "$w $compound";
      debug "compound is: $compound\n";
    }

    # do correction if the previous punctuation mark is a single "
    if ($w eq '"') {
      debug "compound in if at beginning is: $compound\n";
      $compound =~ s/^(.) /$1/;
      debug "compound after s/// is: $compound\n";
      $this = $compound;
      debug "thus this is: $this\n";
    } else {
      debug "else-statement\n";
      push @sentence, $_ for reverse @t;
    }
    debug "and now push this on the sentence\nsentence is @sentence\n";
    push @sentence, $this;
    debug "reassemle quoted compound";
    next WORD;
  }


# PARANTHESES

  $parentheses += (paren_count $this);

  # recover from too many closed parenthesis
  if($parentheses < 0) {

    # detect `a )'
    if ($last =~ /^[a-z]$|^[iv]+$/) {
      my $prev = pop @sentence;
      push @sentence, "$prev$this";
      $parentheses++;
      debug "reassemble $prev$this";
      next WORD;
    } else {
      warn "WARNING: ignoring unopened closing paren in line $input_line_no!\n";
      $parentheses = 0;
    }
  }

  # detect `ver(schlimm)bessern'
  elsif (0 == paren_count $this) {
    # do nothing
  }

  # detect (
  elsif (paren_count($this) > 0) {

    # detect `( un)heimlich'
    if ($next =~ /^[^\(]+\).+$/) {
      push @sentence, "$this$next";
      $parentheses += paren_count $next;
      debug "parentheses == $parentheses";
      get_word();
      next WORD;
    }
  }

  # detect )
  elsif ($this =~ /^[\]\)]/) {

    # detect `Dateiname(n)'
    if ($last =~ /.\(./) {
      my $prev = pop @sentence;
      push @sentence, "$prev$this";
      next WORD;
    }
  }

  # detect `( Pretty Good Privacy)-Programm'
  elsif (paren_count($this) < 0) {
    my @list;
    my $x = '';
    while (@sentence && $x ne '.') {
      $x = pop @sentence;
      if ($x eq '(') {
        push @sentence, "(@list $this";
        next WORD;
      } else {
        unshift @list, $x;
      }
    }
    push @sentence, @list;
    push @sentence, $this;
    next WORD;
  }


# DOT

  # ABBREVIATIONS

  # decide whether a dot serves as a sentence end or not
  if (defined $next && $next eq '.') {

    # recognize common multiple abbreviations, such as `z.B.'
    # (a) if written as one word, break them in half
    if("$this." =~ $multi_abbrev_re) {
      $this .= get_word();
      while ($this =~ s/^(.\.)//) {
        push @sentence, $1;
      }
      next WORD;
    }

    # (b) if written as four words, combine each letter with its stop
    my $two = preview_word(2);
    my $three = preview_word(3);
    if(defined $three &&
       $three eq '.' &&
       "$this.$two." =~ $multi_abbrev_re) {
      get_word();
      get_word();
      get_word();
      push @sentence, "$this.";
      push @sentence, "$two.";
      next WORD;
    }
  }

  # decide whether a dot serves as a sentence end or not
  if ($this eq '.') {

    debug "found full stop!";

    # reassemble mis-tokenized abbrevs with multiple stops in them:
    # `S.W.I.F.T.'
    if ($last =~ /[A-Za-z]\.[A-Za-z]$/) {
      my $prev = pop @sentence;
      push @sentence, "$prev.";
      debug "reassemble $prev.";
      next WORD;
    }

    # combine abbreviation and stop: 'Dr', '.' --> 'Dr.'
    if ($last =~ $abbrev_re) {
      my $prev = pop @sentence;
      push @sentence, "$prev.";
      debug "reassemble abbreviation $prev.";
      next WORD;
    }

  # NUMBERS
    # reassemble mis-tokenized numbers:
    # '1', '.', '000', '.', '000' --> '1.000.000'
    if ($last =~ $num_re && $next =~ /^[0-9]{3}\b/) {
      my $prev = pop @sentence;
      push @sentence, "$prev.$next";
      debug "reassemble number $prev.$next";
      next WORD;
    }


# FIXME

    # FIXME: This can fail if a sentence ends with an abbreviation,
    # since the sentence full stop is suppressed in that case. Also,
    # a sentence can legitimately end in a number: `...auf Platz 9.'
    # Currently we err on the side of caution and do not end the
    # sentence.

    else {
      my $n = find_next_proper_word();
      my $reassemble_stop = 0;
      my $break_sentence = 1;

      # As a further hint, if the next proper word does not start with a
      # capital letter, it is probably part of the same sentence.
      if(defined $n &&
         $n !~ /[A-Z\"\'\`\(]/ &&
         $n !~ $small_names_re &&
         $n !~ $web_re &&
         $n !~ $num_re) {

        debug "small letter follows, no sentence break.";
        $break_sentence = 0;

        # Prohibiting the sentence break still does not tell us
        # whether the . should be joined to the previous word or not.
        # In this example it should:
        #
        #  ... bis zu 12 Pf . pro Minute bezahlen ...
        #
        # but in this it shouldn't:
        #
        #  ... mssen wir unterbinden . " , sagte der Leiter des ...
        #
        # And so we only add the dot to the previous word if it really
        # is followed by a small letter immediately.
        if($next =~ /^[a-z]/) {
          debug "small letter follows immediately, reassembling.";
          $reassemble_stop = 1;
        }
      }


# ORDINAL

      # Ordinal numbers look like cardinal numbers + stop.

      if($last =~ /^\d{1,3}$/) {
        debug "ordinal?";
        my $prev = $sentence[-2] || '';
	debug "prev: $prev, last: is $last, this is: $this, next is: $next";
	if(($prev =~ /^die$/i && $next =~ /^[A-Z][A-Za-z]{0,2}$/) ||
	      ($prev =~ /^(S\.|Seite|Platz)$/) ||
	      ($prev =~ /^seit$/ &&
              $last !~ /^(0?[1-9])|([12][0-9])|(3[01])$/)
	      ){
             debug "suspect cardinal number";
             $break_sentence = 1;
             $reassemble_stop = 0;
        }
        elsif($prev =~ /^(als|die|der|das|den|dem|des|
	   am|frs|ins|beim|vom|zum|zur|im|jede[nmrs]?)$/ix ||
	   # day of date before
           $prev =~ /^(0?[1-9])|([12][0-9])|(3[01])\.$/
           ) {
          debug "suspect ordinal number, word before is: $prev";
          $break_sentence = 0;
          $reassemble_stop = 1;
        }
        elsif($next =~ $Monat_re ||
	   $next =~ $ordinal_nouns_re ||
           $next =~ /^(-|bis)$/ ||
	   # Abbreveations often indicate a previous ordinal number
	   #$next =~ /^[A-Z]+$/ ||
           $next =~ /^(\d+|\))$/) {
          debug "suspect ordinal number, following word is: $next";
          $break_sentence = 0;
          $reassemble_stop = 1;
        }
      }

      # roman ordinal numbers
      if($last =~ /^(((I?[VX])(I{1,3})?)|(I{1,3}?))$/ && $prev !~ /^Pentium$/) {
        my $prev = $sentence[-2] || '';
        debug "suspect roman ordinal number: $prev";
        $break_sentence = 0;
        $reassemble_stop = 1;
      }



# SPEECH

      # If direct speech ends in . and the speech attribution follows,
      # do not end the sentence.
      if ($next eq '"') {
        my $a = preview_word(2);
        if ($a =~ /^[a-z]+$/) {
          debug "`$a' suggests speech attribution";
          push @sentence, $this;
          next WORD;
        }
        my $b = preview_word(3);
        if ($a eq ',' && $b =~ /^[a-z]+$/) {
          debug "`$b' suggests speech attribution";
          push @sentence, $this;
          next WORD;
        }
      }

      my $prev = pop @sentence;
      if(defined $prev) {
        if($reassemble_stop) {
          push @sentence, "$prev.";
        } else {
          push @sentence, $prev;
          push @sentence, '.';
        }
      }

      if(!$break_sentence) {
        next WORD;
      }
    }

    if(!$opt_t) {
      debug "Sentence ends.";
      push_punctuation;
      write_sentence(\@sentence);
      $sentence_start = $next;
      next WORD;
    }
  }


# 's or 'n FIXME: better comment

  # Split off clitic 's:
  #
  # 'gibt's' --> 'gibt' ''s'
  #
  # Of course there are many more words that can carry the clitic s;
  # we need a method here of distinguishing `Barron's' from `soll's'.
  # Since we can't do anything more intelligent, we assume that `HP's'
  # should be one token and `geht's' should be two because of
  # capitalization.
  elsif ($this =~ /^(.+)([\'\][sn])$/) {
    my $base = $1;
    my $ext = $2;
    $ext =~ s//'/g;
    my $split = 1;
    if($opt_g && $base =~ /^[A-Z]/) {
      $split = 0;
    }

    if($split) {
      push @sentence, $base;
      push @sentence, $ext;
    } else {
      push @sentence, "$base$ext";
    }
  }


# PUNCTUATION

  # Detect names with ! in them. There is nothing we can do about
  # them except have a long list of such idiotic names.
  elsif ($this =~ '!' and $last =~ /($exclaim_names)$/o) {
    my $prev = pop @sentence;
    push @sentence, "$prev!";
    next WORD;
  }

  # Detect other sentence-ending punctuation.
  else {
    push @sentence, $this;
    if(!$opt_t) {
      if ($this =~ /^([?!:;]|\.\.\.)$/) {
        my $npw = find_next_proper_word();
        if(defined $npw && $npw =~ /^[A-Z0-9]/) {
          combine_exclaim;
          push_punctuation;
	  if ($parentheses == 0) {
            write_sentence(\@sentence);
            $sentence_start = $next;
	  }
        }
      }
    }
  }


# END

  # If this is the last token, stop the loop.
  if (!defined $next) {
    last;
  }

} continue {
  $prev = $last;
  $last = $this;
}
write_sentence(\@sentence);
