# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## AllWordgraphs - manage all grammar wordgraphs.
##
## \author Michael Daum (see also AUTHORS and THANKS for more)
## $Id: allwordgraphs.tcl,v 1.81 2006-06-21 11:54:13 foth Exp $
## ----------------------------------------------------------------------------
class AllWordgraphs {
  inherit DataBrowser


    # public methods
    constructor {args} {}; ## \type TclList
    public method getNext {latticeId}; ## \type TclString
    public method getPrev {latticeId}; ## \type TclString
    public method getAllLatticeIds {};
    public method getCData {id}; ## \type TclString
    public method refreshrow {row id}; ## \type TclNumber, TclString 
    public method init_data {}

    # private methods
    private method _init_data {}
    private method newbutton_action {}
    private method annobutton_action {}
    private method icbutton_action {}
    private method interactivebutton_action {}
    private method newnet {}


  # private variables

  ## hash mapping wordgraph ids to their C structure
  private variable wordgraph; ## \type TclArray

  ## list of ids sorted as in the table
  private variable allsorted

  ## list of selected ids will be stored in inherited
  ## protected variable DataBrowser::_selection 
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body AllWordgraphs::constructor {args} {
  set _idColumnIndex 0
  set table $itk_component(table)
  $table configure -cols 3 \
    -colstretchmode last \
    -cellanchor w 

  itk_component add newbutton {
    button $itk_component(childsite).new \
	-anchor w \
	-text "New" \
	-command [code $this newbutton_action]
  } { }
  
  itk_component add newnetbutton {
    button $itk_component(childsite).newnet\
	-anchor w \
	-text "New Net" \
	-command [code $this newnet]
  } { }

  itk_component add annobutton {
    button $itk_component(childsite).annotation\
	-anchor w \
	-text "Annotations" \
	-command [code $this annobutton_action]
  } { }

  itk_component add icbutton {
    button $itk_component(childsite).ic\
	-anchor w \
	-text "IC" \
	-command [code $this icbutton_action]
  } { }

  itk_component add interactivebutton {
    button $itk_component(childsite).interactive\
	-anchor w \
	-text "Interactive" \
	-command [code $this interactivebutton_action]
  } { }

  itk_component add newnetDialog {
    NewnetDialog $itk_interior.newnetDialog \
	-master . \
	-title "Xcdg: Newnet" \
	-modality application
  } { }

  # packing
  pack $itk_component(newbutton) \
      $itk_component(newnetbutton) \
      $itk_component(annobutton) \
      -side left -padx 3 

  # register the helpmessages
  .cdgmain help sethelpstr \
      $itk_component(newbutton) "Input a new wordgraph" \
      $itk_component(newnetbutton) "Compute a new constraint net" \
      $itk_component(annobutton) "Find the corresponding parse in the input" \
      $itk_component(icbutton) "Apply incremental search to the utterance" \
      $itk_component(interactivebutton) "Parse user-supplied input." \
      $itk_component(entryfield) "Enter a wordgraph-id" \
      $itk_component(table) "Click to select"

  # get the initial data
  _init_data

  eval itk_initialize $args
}


## ----------------------------------------------------------------------------
## fill a row with the values from a specific wordgraph.
## \param row row index to insert data
## \param ptr Pointer to lattice_id
## ----------------------------------------------------------------------------
body AllWordgraphs::refreshrow {row ptr} {
  set id [LatticeStruct_id_get $ptr]
  set length [listSize [LatticeStruct_arcs_get $ptr]]
  set sentence ""
  for {set m [LatticeStruct_arcs_get $ptr]} {$m != "NULL"} \
      {set m [ListStruct_next_get $m]} {
    set arc [ListStruct_item_get $m]
    append sentence " [ArcStruct_word_get $arc]"
  }

  $itk_component(table) setCell row $row,0 [list $id $length $sentence]
}

## ----------------------------------------------------------------------------
## call _init_data if necessary
## ----------------------------------------------------------------------------
body AllWordgraphs::init_data {} {
  global inputCurrentGrammar

  if {[InputStruct_xcdgwordgraphsToDate_get $inputCurrentGrammar]} {
    return
  }

  .cdgmain busy compute "Init Word Graphs" [code $this _init_data]
}



## ----------------------------------------------------------------------------
## get the data from the cdg-tool
## ----------------------------------------------------------------------------
body AllWordgraphs::_init_data {} {
  global inputCurrentGrammar

  set table $itk_component(table)
  $table configure -state normal
  $table clear all
  $table delete rows 1 [$table cget -rows]
  $table tag config default
  $table tag config colored -bg gray80 
  $table tag config length -anchor e
  $table tag col length 1
  $table tag config id -anchor w
  $table tag col id 0
  $table tag raise id
  $table width 0 20 1 7 2 500
  $table setCell row 0,0 { "Id" "Length" "Words" }

  array unset wordgraph
  set m [InputStruct_lattices_get $inputCurrentGrammar]
  for {set l $m}  { $l != "NULL" } { set l [ListStruct_next_get $l ] } {
    set lattice [ListStruct_item_get $l]
    set lattice_id [LatticeStruct_id_get $lattice]
    set wordgraph($lattice_id) $lattice
  }

  set r 1
  set list [inputGetSortedLatticeNames $inputCurrentGrammar]
  _setCount [listSize $list]
  for {set l $list}  { $l != "NULL" } { set l [ListStruct_next_get $l ] } {

    set name [pointer2string [ListStruct_item_get $l]]
    lappend allsorted $name
    set lattice $wordgraph($name)
    refreshrow $r $lattice

    if {[expr $r % 1000] == 0} {
      .cdgmain help showstr "Init word graphs: $name"
      update idletask
    }

    incr r


  }
  listDelete $list;

  setSelection

  InputStruct_xcdgwordgraphsToDate_set $inputCurrentGrammar 1
  $table configure -state disabled
}

## ----------------------------------------------------------------------------
## actions to take place on pressing the newbutton
## ----------------------------------------------------------------------------
body AllWordgraphs::newbutton_action {} {
  .cdgmain confirm "Under Construction\nPlease use `inputwordgraph' in the shell."
  
  set table $itk_component(table)
  $table xview moveto 0.1
}

## ----------------------------------------------------------------------------
## actions to take place on pressing the annobutton
## ----------------------------------------------------------------------------
body AllWordgraphs::annobutton_action {} {

  if {$_selection == ""} {
    ::cmd::Puts "ERROR: Please select a wordgraph"
    return
  }

  foreach item $_selection {
    set lat [findLattice $item]
    if {$lat == "NULL"} {
      ::cmd::Puts "ERROR: no lattice `$item' found."
      return
    }
    for {set l [findAnnotations $lat 1]} {$l != "NULL" } {set l [ListStruct_next_get $l]} {
     set id [AnnoEntryStruct_id_get [ListStruct_item_get $l]]
      set parse [.cdgmain parses getCData $id]
      .cdgmain parses showparse $id
    }
  }
}

## ----------------------------------------------------------------------------
## actions to take place on pressing the icbutton
## ----------------------------------------------------------------------------
body AllWordgraphs::icbutton_action {} {

  if {$_selection == ""} {
    ::cmd::Puts "ERROR: Please select a wordgraph"
    return
  }

  foreach id $_selection {
    ::cmd::IC $id
  }
}

## ----------------------------------------------------------------------------
## actions to take place on pressing the interactivebutton
## ----------------------------------------------------------------------------
body AllWordgraphs::interactivebutton_action {} {

  # if there already is an interactive window, do nothing
  foreach vis [itcl::find objects -isa VisParses] {
    if {[$vis cget -title] == "IC:interactive"} {
      ::cmd::Puts "ERROR: There already is an open interactive window."
      return
    }
  }

  # build a window to display results
  set interactiveWindow [VisParses ".#auto" \
                             -title "Xcdg - Interactive parsing" \
                             -iconname "Interactive" \
                             -interactive 1 ]
  
  # call IC in interactive mode
  ::cmd::IC interactive


}

## ----------------------------------------------------------------------------
## actions taking place when the newnetbutton is pressed
## ----------------------------------------------------------------------------
body AllWordgraphs::newnet {} {

  if {"$_selection" != ""} {
    $itk_component(newnetDialog) activate $_selection
  }
}

## ----------------------------------------------------------------------------
## get pointer to a lattice_id, cache it in array wordgraph
## ----------------------------------------------------------------------------
body AllWordgraphs::getCData {id} {

  if {![info exists wordgraph($id)]} {
    set found 0
    set lattice [findLattice $id]
    if {$lattice == "NULL"} {
      return "NULL"
    }
    set wordgraph($id) $lattice
  }
  return $wordgraph($id)
}



## ----------------------------------------------------------------------------
## get the lattice next after $id.
## ----------------------------------------------------------------------------
body AllWordgraphs::getNext {latticeId} {

  init_data
  if {![info exists wordgraph($latticeId)]} {
    return ""
  }
  set ids [getAllLatticeIds]
  set index [lsearch -exact $ids $latticeId]
  set index [expr $index + 1]
  return [lindex $ids $index]

}

## ----------------------------------------------------------------------------
## Show the annotation for the lattice above $id.
## ----------------------------------------------------------------------------
body AllWordgraphs::getPrev {latticeId} {

  init_data
  if {![info exists wordgraph($latticeId)]} {
    return ""
  }
  set ids [getAllLatticeIds]
  set index [lsearch -exact $ids $latticeId]
  set index [expr $index - 1]
  return [lindex $ids $index]
}


## ----------------------------------------------------------------------------
## Return list of all lattices' names
## ----------------------------------------------------------------------------
body AllWordgraphs::getAllLatticeIds {} {
  init_data
  return $allsorted
}
