# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
## ----------------------------------------------------------------------------
## Balloon - A tooltip class.
## This is a generic tooltip implementation. Just create a Balloon and
## switch it on() now and then. A Balloon is displayed if the mouse
## pointer does not leave the area where it has been initialized within
## 200 milliseconds by default.
## 
## \author Michael Daum 
## $Id: balloon.tcl,v 1.13 2005-12-01 16:22:13 micha Exp $
## ----------------------------------------------------------------------------
class Balloon {
   inherit ::itk::Toplevel

   itk_option define -delay delay Delay 200

   public method on {x y text args}; ## \type TclNumber, TclNumber, TclString, TclList
   public method off {}
   public method maybe_off {}
   constructor {args} {}; ## \type TclList
   private method _doit {x y text}; ## \type TclNumber, TclNumber, TclString

   ## id of the background job.
   ## When a balloon is going to be displayed we spawn a job with \a after
   ## and store the job id in this variable.
   private variable balloonId 0

   # Should the tooltip vanish upon mouse movement?
   private variable _transient 1
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body Balloon::constructor {args} {
   wm withdraw $itk_interior
   wm overrideredirect $itk_interior 1
   wm transient $itk_interior

   bind $itk_interior <Leave> [code $this off]

   itk_component add text {
      label $itk_interior.text \
		-bd 0 \
		-padx 3 \
		-pady 0 \
                -justify left
   } { 
      rename -foreground -textforeground textForeground Foreground
      rename -background -textbackground textBackground Background
      keep -cursor -relief -anchor -text -font -padx -pady -borderwidth
   }

   pack $itk_component(text) -fill both -expand 1 -side left

   eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## hide a tooltip.
## Call this method to switch off the tooltip balloon.
## ----------------------------------------------------------------------------
body Balloon::off {} {
   set pointerx [winfo pointerx $itk_interior]
   set pointery [winfo pointery $itk_interior]
   if {[winfo containing $pointerx $pointery] == $itk_interior} {
      return
   }

   if {$balloonId > 0} {
      after cancel $balloonId
      set balloonId 0
   } else {
      wm withdraw $itk_interior
   }
}

## ----------------------------------------------------------------------
## Remove the tooltip if it was specified to be transient.
## ----------------------------------------------------------------------
body Balloon::maybe_off {} {
  if {1 == $_transient} off
}

## ----------------------------------------------------------------------------
## do the actual display.
## This method is only called by Balloon::on() when the tcl loop gets idle.
## \param x the x coordinate where the balloon will be displayed 
## \param y the y coordinate where the balloon will be displayed 
## ----------------------------------------------------------------------------
body Balloon::_doit {x y text} {
  set balloon $itk_interior
  set height [winfo reqheight $balloon]
  set width [winfo reqwidth $balloon]
  set screenWidth [winfo screenwidth $balloon]
  set screenHeight [winfo screenheight $balloon]
  
  # guard against window reaching offscreen
  if {[expr $x + $width] > $screenWidth} {
    set x [expr $screenWidth - $width]
  }
  if {$x < 0} {
    set x 0
  }
  
  if {$y > $screenHeight} {
    set y [expr $screenHeight - $height]
  }
  if {$y < 0} {
    set y 0
  }

  configure -text "$text"

  wm geometry $balloon +$x+$y
  wm deiconify $balloon
  raise $balloon
  set balloonId 0
}

## ----------------------------------------------------------------------------
## show a tooltip.
## This method actually displays a tooltip with a certain message and position.
## \param x the x coordinate where the balloon will be displayed 
## \param y the y coordinate where the balloon will be displayed 
## \param text the message text to be shown inside the tooltip
## \param args: If non-empty, draw the balloon immediately (do not wait).
## ----------------------------------------------------------------------------
body Balloon::on {x y text args} {
  if {[string trim $text] == ""} {
    return
  }

  if {$balloonId > 0} {
    if {$text == $itk_option(-text)} {
      return
    }
    after cancel $balloonId
  }

  if [regexp {keep} $args] {
    set _transient 0
  } else {
    set _transient 1
  }  

  if [regexp {now} $args] {
    _doit $x $y $text
  } else {
    set balloonId [after $itk_option(-delay) [code $this _doit $x $y $text]]
  }

}

