# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## ParseTree - visualization of one parse.
##
## ParseTree objects are the central means to give you a nice
## dependency tree drawing. They are only thought to be used as
## components of a TreeEditorParent object being the GUI surrounding a set of
## visualized ParseTree objects. Taking the embedded Parse object into
## account, these tree classes implement the tree editor in XCDG.
##
## \author Michael Daum, Kilian A. Foth, Dietmar Fnning 
## (see also AUTHORS and THANKS for more)
##
## $Id: parsetree.tcl,v 1.208 2006-05-08 14:41:53 foth Exp $
## ----------------------------------------------------------------------------

class ParseTree {
  inherit ::itk::Widget 

  # new options

  ## itk_option: color of a node of the dependency tree
  itk_option define -nodecolor nodeColor Color "black"

  ## itk_option: color of edges up to the parents
  itk_option define -edgecolor edgeColor Color "black"

  ## itk_option: color of vertical lines
  itk_option define -vlinecolor vlineColor Color "gray" 

  ## itk_option: font-family 
  itk_option define -fontfamily fontFamily String "Helvetica" 

  ## itk_option: font-factors 
  itk_option define -fontsizes fontSizes String "7 9 10 12 14 16 18" 

  ## itk_option: color of the words in the sentence
  itk_option define -wordcolor wordColor Color "black"

  ## itk_option: color of the dependency labels
  itk_option define -labelcolor labelColor Color "black"

  ## itk_option: color of highlighted structures
  itk_option define -highlightcolor highlightColor Color "red"

  ## itk_option: color of erroneous structures
  itk_option define -errorcolor errorColor Color "red"

  ## itk_option: decoration of a level component
  itk_option define -levelrelief levelRelief Relief "flat"

  ## itk_option: decoration of a level component
  itk_option define -levelborderwidth levelBorderwidth Borderwidth 1

  ## the embedded Parse object
  private variable _parse "" ;## \type Parse

  private variable _parent ""; ## \type TreeEditorParent
  
  ## if no visparses is specified -> default to this snapmode
  private variable _default_snapmode "nearest"
  
  ## array holding the middle points of the words.
  ## initialized by drawText; indexed by timepoints, e.g. 0
  private variable _textMidPos; ## \type TclArray

  ## holds all items on a canvas belonging to one timepoint.
  ## Example: _itemsOfTimepoint(SYN,3) == 4 27 26 24 25
  private variable _itemsOfTimepoint; ## \type TclArray

  ## array holding the timepoint which a specified id belongs to.
  ## Example: _timepointOfId(SYN,3) = 0
  private variable _timepointOfId; ## \type TclArray

  ## array storing arcs by level and timepoint.
  ## Example: _arcOfTimepoint(3,OBL1) = 47 48
  private variable _arcOfTimepoint; ## \type TclArray

  ## array holding the level that an arc belongs to.
  ## Example: _levelOfItem(89) = DOM
  private variable _levelOfItem; ## \type TclArray

  ## holds ids of canvas items to be highlighted when touching a violation.
  ## Example: _itemsOfViolation(1) == 23 SYN 22 SYN
  private variable _itemsOfViolation; ## \type TclArray

  ## array storing the canvas of a level.
  private variable _levelOfCanvas; ## \type TclArray
 
  ## array indicating whether a level is already drawn.
  private variable _isDrawn; ## \type TclArray

  private variable _redraw_level "";
  private variable _dragItem ""
  private variable _mouseItem ""
  private variable _mouseBinding ""
  private variable _dragStartX 0
  private variable _dragStartY 0
  private variable _zoomstep 0
  private variable _oldColors
  private variable _oldLabel ""
  private variable _fontsizes; ## \type TclArray
  private variable _zoomfactors; ## \type TclArray
  private variable _currentFont ""; ## \type TclString

  ## the word that an edge would snap to if dropped
  private variable _snapWord ""

  private variable _lastxmark 0
  private variable _lastymark 0

  ## Command-History-Object for undo-implementation
  private variable _commandHistory ""

  ## Variables holding state of the keyboard highlighting
  private variable _kb_word 0
  private variable _kb_regent -2
  private variable _kb_labels ""
  private variable _kb_label -1
  private variable _kb_lexemes ""
  private variable _kb_lexeme ""
  private variable _label_prefix ""

  ## Flags for inter-handler communication
  private variable _computing 0
  private variable _released 0
  
  private method getSnapmode {};
  
  public method drawAll {}
  public method evalParse {}
  public method writeToFile {fileName}; ## \type TclString
  public method mirror {}
  public method verify {}
  public method showcycles {}
  public method breakcycles {}
  public method zoom {direction}; ## \type TclNumber
  public method setOrientation {orient}; ## \type TclString
  public method getOrientation {}
  public method parse {args}; ## \type TclList
  public method mainlevelname {};
  public method highlight {which}; ## \type TclString
  public method paintTabs {}
  public method paintLevel {level}; ## \type TclString
  public method backwardLevel {}
  public method forwardLevel {}

  ## Methods used to implement undo-functionality (called from TreeEditorParent)
  public method undo {}
  public method redo {}
  public method addUndoListener {listener}; ## \type TclString
  public method canUndo {}
  public method canRedo {}

  public method retry {}

  private method autoredraw {}
  private method autoeval {}

  constructor {p vp args} {}; ## \type Parse, TreeEditorParent, TclList
  destructor {}
 
  ## Methods used to implement undo-functionality (Internal methods)
  private method undoEdge {modifier modifiee level} ; ## \type TclNumber, TclNumber, TclString
  private method undoLabel {modifier label level}; ## \type TclNumber, TclString, TclString
  private method undoLexeme {from description}; ## \type TclNumber, TclString
  private method _zoom {step}; ## \type TclNumber
  private method allEnter {w}; ## \type Canvas
  private method allLeave {w}; ## \type Canvas
  private method labelOfEdge {w edge}; ## \type Canvas, TclString
  private method wordOfTimepoint {w t}; ## \type Canvas, TclNumber
  private method edgeOfTimepoint {w t}; ## \type Canvas, TclNumber
  private method nodeOfTimepoint {w e t}; ## \type Canvas, TclNumber, TclNumber
  private method edgeEnter {w}; ## \type Canvas
  private method edgeLeave {w}; ## \type Canvas
  private method allSybillings {id1 id2 level}; ## \type TclNumber, TclNumber, TclString
  private method canvasToScreen {canvas x y}; ## \type Canvas, TclNumber, TclNumber
  private method createArc {canvas x1 y1 x2 y2 text bindNo}; ## \type Canvas, TclNumber, TclNumber, TclNumber, TclNumber, TclNumber, TclNumber
  private method drawArcs {}
  private method drawLevel {level}; ## \type TclString
  private method drawText {canvas x y}; ## \type Widget, TclNumber, TclNumber
  private method drawViolas {}
  private method canvasClick {w x y}; ## \type Canvas, TclNumber, TclNumber
  private method edgeClick {w x y}; ## \type Canvas, TclNumber, TclNumber
  private method edgeRightClick {w}; ## \type Canvas
  private method arcRightClick {w}; ## \type Canvas
  private method edgeDrop {w x y}; ## \type Canvas, TclNumber, TclNumber
  private method edgeDrag {w x y}; ## \type Canvas, TclNumber, TclNumber
  private method edgeDragPan {w x y}; ## \type Canvas, TclNumber, TclNumber
  private method edgeDragCancel {w}; ## \type Canvas
  private method fixCoords {w x y}; ## \type Canvas, TclNumber, TclNumber
  private method labelClick {w}; ## \type Canvas
  private method labelRightClick {w}; ## \type Canvas
  private method labelSelect {l i label}; ## \type TclString, TclNumber, TclString
  private method makeUserInterface {w}; ## \type Canvas
  private method moveEdge {canvas x1 y1 x2 y2}; ## \type Canvas, TclNumber, TclNumber, TclNumber, TclNumber
  private method moveLabel {canvas label_id x1 y1 x2 y2}; ## \type Canvas, TclString, TclNumber, TclNumber, TclNumber, TclNumber
  private method multiCanvas {args}; ## \type TclList
  private method multiMark {x y}; ## \type TclNumber, TclNumber
  private method multiDragto {x y}; ## \type TclNumber, TclNumber
  private method noCrossing {id bindings}; ## \type TclString, TclList
  private method reDrawLevel {level}; ## \type TclString
  private method registerItems {level}; ## \type TclString
  private method rgValue {score}; ## \type TclNumber
  private method showPopMenu {w x y items}; ## \type Canvas, TclNumber, TclNumber, TclList
  private method violaBrowse {mode xcoord ycoord}; ## \type TclString, TclNumber, TclNumber
  private method pop_constraint {x y X Y}; ## \type TclNumber,   TclNumber, TclNumber, TclNumber
  private method getAlternativeLexemes {w i} ; ## \type Canvas, TclNumber
  private method wordClick {w}; ## \type Canvas
  private method wordMiddleClick {w}; ## \type Canvas
  private method wordShiftClick {w}; ## \type Canvas
  private method wordRightClick {w}; ## \type Canvas
  private method wordEnter {w}; ## \type Canvas
  private method wordLeave {w}; ## \type Canvas
  private method wordSelect {w from lexeme}; ## \type Canvas, TclNumber, LexemNode
  private method optimizeWord {t}; ## \type TclNumber
  private method compareLexeme {a b}; ## \type LexiconItem, LexiconItem
  private method closestNode {w x y}; ## \type Canvas, TclNumber, TclNumber
  private method itemVisible {w i}; ##\type Canvas, TclNumber
  private method centerOnItem {w i}; ##\type Canvas, TclNumber
  private method yscroll {args}; ## \type TclList
  private method kb_select_word {w k a}; ## \type Canvas, TclWidget, TclKeyBinding
  private method kb_select_edge {w k a}; ## \type Canvas, TclWidget, TclKeyBinding
  private method kb_change_edge {w k a}; ## \type Canvas, TclWidget, TclKeyBinding
  private method kb_select_label {w k a}; ## \type Canvas, TclWidget, TclKeyBinding
  private method kb_init_label_list {w i}; ## \type Canvas, TclWidget
  private method kb_accelerate_label {w k a}; ## \type Canvas, TclWidget, TclKeyBinding
  private method kb_unaccelerate_label {w k a}; ## \type Canvas, TclWidget, TclKeyBinding
  private method kb_change_label {w k a}; ## \type Canvas, TclWidget, TclKeyBinding
  private method kb_select_lexeme {w k a}; ## \type Canvas, TclWidget, TclKeyBinding
  private method kb_change_lexeme {w k a}; ## \type Canvas, TclWidget, TclKeyBinding
  private method mybbox {w item}; ## \type Canvas, TclString

  private method mark {item}; ## \type TclNumber
  private method marked {item}; ## \type TclNumber
  private method unmark {item}; ## \type Canvas, TclNumber
  private method save_color {item}; ## \type Canvas, TclNumber
  private method current_canvas {}
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body ParseTree::constructor {p te args} {
  set _parse $p
  set _parent $te

  ## instantiate command-history-object
  set _commandHistory [CommandHistory $this.commandHistory]

  set levels [mainlevelname]
  if {$_parse != "NULL"} {
    set levels [$_parse getLevels]
  }
  foreach level $levels {
    set _isDrawn($level) 0
  }

   # initialize
   eval itk_initialize $args
  # set the fontsizes
  set i -3
  foreach size $itk_option(-fontsizes) {
    set _fontsizes($i) $size
    incr i
  }
  array set _zoomfactors {-3 0.5 -2 0.76 -1 0.83 0 1 1 1.33 2 1.66 3 2}

  # split pane
  itk_component add pane {
     iwidgets::panedwindow $itk_interior.pane \
       -height 10c -width 10c  -orient horizontal
  } {}
  set pane $itk_component(pane)
  $pane add "Parse"
  set pane1 [$pane childsite "Parse"]
  $pane reset

  # tabular containing the levels
  itk_component add tbn {
     iwidgets::Tabnotebook $pane1.tbn
  } {}

  # conflicts
  if {($_parse != "NULL") && ![$_parse isAbstract]} {
    $pane add "Conflicts"
    set pane2 [$pane childsite "Conflicts"]

    itk_component add conflicts {
      MyTable $pane2.conflicts \
          -exportselection 0 \
          -cols 6 \
          -state disabled \
          -colstretchmode unset \
          -selecttype row \
          -font {Helvetica -12 bold} \
          -background gray90 \
          -titlerows 0  \
          -outerborderwidth 1
    } {}
    bind [$itk_component(conflicts) component table] <ButtonPress-1> \
        [code $this violaBrowse 1 %x %y]
    bind [$itk_component(conflicts) component table] <ButtonPress-2> \
        [code "$this pop_constraint %x %y %X %Y; break"]
    bind [$itk_component(conflicts) component table] <ButtonPress-3> \
        [code $this violaBrowse 2 %x %y]
  }

  # popup menu
  itk_component add popmenu {
     menu $itk_interior.popmenu
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

   # build a scrolled canvas for each level 
   set tbn $itk_component(tbn)
   foreach level $levels {

     # add page
     set page [$tbn add  \
       	-label $level \
       	-command [code $this drawLevel $level]]

     # add canvases
     itk_component add $level {
        iwidgets::scrolledcanvas $page.cs
     } {
        keep -cursor -background
        rename -relief -levelrelief levelRelief Relief
        rename -borderwidth -levelborderwidth levelBorderwidth Borderwidth
     }

     # advise the horizontal scrolling commands
     # so that each scrollbar scrolls all levels.
     set hbar [$itk_component($level) component horizsb]
     $hbar configure -command [code $this multiCanvas xview]

     # advise the vertical scrollbar so that it scrolls one level only,
     # and not if the tree is already completely visible.
     set hbar [$itk_component($level) component vertsb]
     $hbar configure -command [code $this yscroll $itk_component($level)]

     # fast scanning
     bind [$itk_component($level) childsite] <ButtonPress-2> [code $this multiMark %x %y] 
     bind [$itk_component($level) childsite] <B2-Motion> [code $this multiDragto %x %y] 

     pack $itk_component($level) -fill both -side top -expand 1
     set _levelOfCanvas([$itk_component($level) childsite]) $level

   }
  
  # packing
  if {$_parse != "NULL"} {
    #puts $_parse
    if {[$_parse isAbstract]} {
      pack $pane $pane1 $tbn -expand 1 -fill both
    } else {
      pack $pane $pane1 $pane2 $tbn -expand 1 -fill both
      pack $itk_component(conflicts) -expand 1 -fill both
    }
  }

  drawViolas
  paintTabs

  # view main level initially
  set mainlevel [mainlevelname]
  set index [lsearch $levels $mainlevel]
  if {$index == -1} {
    $tbn view 0
  } else {
    $tbn view $index
  }
  if { ($_parse != "NULL") && ![$_parse isAbstract]} {
    $pane fraction 80 20 
  }
}


## ----------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------
body ParseTree::destructor {} {
    delete object $_commandHistory
}



## ----------------------------------------------------------------------------
## part of the initialization process.
## ----------------------------------------------------------------------------
body ParseTree::makeUserInterface {w} {

  # touching a word pops up a tooltip giving its description
  $w bind word <Any-Enter> [code $this wordEnter %W ]
  $w bind word <Any-Leave> [code $this wordLeave %W ]

  # clicking on a word shows its menu of alternatives
  $w bind word <ButtonPress-1> [code $this wordClick %W]
  $w bind word <Shift-ButtonPress-1> [code $this wordShiftClick %W]
  $w bind word <ButtonPress-2> [code $this wordMiddleClick %W]
  $w bind word <ButtonPress-3> [code $this wordRightClick %W]
  
  # touching a label highlights it
  $w bind label <Any-Enter> [code $this allEnter %W ]
  $w bind label <Any-Leave> [code $this allLeave %W ]

  # clicking on a label shows its menu of alternatives
  $w bind label <ButtonPress-1> [code $this labelClick %W]
  $w bind label <ButtonPress-3> [code $this labelRightClick %W]

  # touching an edge highlights it
  $w bind edge <Any-Enter> [code $this edgeEnter %W ]
  $w bind edge <Any-Leave> [code $this edgeLeave %W ]

  # middle clicking on an edge shows its label menu
  $w bind edge <ButtonPress-2> [code $this labelClick %W]

  # dragging an edge moves it around
  $w bind edge <ButtonPress-1> [code $this edgeClick %W %x %y]
  $w bind edge <B1-Motion> [code $this edgeDrag %W %x %y]
  $w bind edge <ButtonRelease-1> [code $this edgeDrop %W %x %y]

  # right clicking an edge redirects it automatically
  $w bind edge <ButtonPress-3> [code $this edgeRightClick %W]
  $w bind arc <ButtonPress-3> [code $this arcRightClick %W]

  # clicking on the canvas pretends the nearest edge was clicked
  bind $w <ButtonPress-1> [code $this canvasClick %W %x %y]
  bind $w <B1-Motion> [code $this edgeDrag %W %x %y]
  bind $w <B1-B2-Motion> [code $this edgeDragPan %W %x %y]
  bind $w <ButtonRelease-1> [code $this edgeDrop %W %x %y]

  # typing ESC cancels an edge move action
  bind $w <Escape> [code $this edgeDragCancel %W]

  # movement keys select a word to manipulate
  bind $w <Left> [code $this kb_select_word %W %A %K]
  bind $w <Right> [code $this kb_select_word %W %A %K]
  bind $w <Up> [code $this kb_select_word %W %A %K]
  bind $w <Down> [code $this kb_select_word %W %A %K]

  # Shift combinations change subordinations
  bind $w <Shift-Left> [code $this kb_select_edge %W %A %K]
  bind $w <Shift-Right> [code $this kb_select_edge %W %A %K]
  bind $w <Shift-Up> [code $this kb_select_edge %W %A %K]
  bind $w <Shift-Down> [code $this kb_select_edge %W %A %K]
  bind $w <Shift-Return> [code $this kb_select_edge %W %A %K]
  bind $w <KeyPress-Shift_L> [code $this kb_select_edge %W %A %K]
  bind $w <KeyPress-Shift_R> [code $this kb_select_edge %W %A %K]
  bind $w <KeyRelease-Shift_L> [code $this kb_change_edge %W %A %K]
  bind $w <KeyRelease-Shift_R> [code $this kb_change_edge %W %A %K]

  # Control combinations change labels
  bind $w <Control-Up> [code $this kb_select_label %W %A %K]
  bind $w <Control-Down> [code $this kb_select_label %W %A %K]
  bind $w <Control-Left> [code $this kb_select_label %W %A %K]
  bind $w <Control-Right> [code $this kb_select_label %W %A %K]
  bind $w <Control-Home> [code $this kb_select_label %W %A %K]
  bind $w <Control-End> [code $this kb_select_label %W %A %K]
  bind $w <Control-Return> [code $this kb_select_label %W %A %K]
  bind $w <KeyPress-Control_L> [code $this kb_select_label %W %A %K]
  bind $w <KeyPress-Control_R> [code $this kb_select_label %W %A %K]
  bind $w <KeyRelease-Control_L> [code $this kb_change_label %W %A %K]
  bind $w <KeyRelease-Control_R> [code $this kb_change_label %W %A %K]

  # Meta combinations accelerate label changes
  bind $w <KeyPress-Meta_L> [code "$this kb_accelerate_label %W %A %K; break"]
  bind $w <KeyPress-Meta_R> [code "$this kb_accelerate_label %W %A %K; break"]
  bind $w <Meta-KeyPress> [code "$this kb_accelerate_label %W %A %K; break"]
  bind $w <KeyRelease-Meta_L> [code "$this kb_unaccelerate_label %W %A %K; break"]
  bind $w <KeyRelease-Meta_R> [code "$this kb_unaccelerate_label %W %A %K; break"]

  # Alt combinations change lexemes
  bind $w <Alt-Up> [code $this kb_select_lexeme %W %A %K]
  bind $w <Alt-Down> [code $this kb_select_lexeme %W %A %K]
  bind $w <Alt-Home> [code $this kb_select_lexeme %W %A %K]
  bind $w <Alt-End> [code $this kb_select_lexeme %W %A %K]
  bind $w <Alt-Return> [code $this kb_select_lexeme %W %A %K]
  bind $w <KeyPress-Alt_L> [code $this kb_select_lexeme %W %A %K]
  bind $w <KeyPress-Alt_R> [code $this kb_select_lexeme %W %A %K]
  bind $w <KeyRelease-Alt_L> [code $this kb_change_lexeme %W %A %K]
  bind $w <KeyRelease-Alt_R> [code $this kb_change_lexeme %W %A %K]

  bind $w <ButtonPress-4> [code $this yscroll $w scroll 1 units]
  bind $w <ButtonPress-5> [code $this yscroll $w scroll -1 units]

}


## ----------------------------------------------------------------------------
## Open tool tip with the full lexicon item of the word.
## ----------------------------------------------------------------------------
body ParseTree::wordMiddleClick {w } {
  set level $_levelOfCanvas($w)
  set _mouseItem [$w find withtag current]
  set timepointId $_timepointOfId($level,$_mouseItem)
  scan $timepointId "%i%c%i" wordposition useless1 useless2
  set item ""

  if {$_parse == "NULL"} {
    return
  }
  
  set item [$_parse getValue $wordposition]
   
  # Format the item for better display in a balloon.
  regsub -all {([^,]+:[^,<]+)(,|$)} $item "\\1\\2\n" item
  
  set level 0
  set lines [split $item "\n"]
  set max 0
  foreach line $lines {
    regsub -all "^ *" $line "" line
    set l [string length $line]
    if {$l > $max} { set max $l }
  }
  set text ""
  set nrlines 0
  foreach line $lines {
    regsub -all "^ *" $line "" line
    if {"" == $line} continue

    if {[regexp {\]} $line]} { incr level -1 }
    set indent [string repeat " " $level]
    set line "$indent$line"
    set l [string length $line]
    if {"" != $text} { append text "\n" }
    append text "$line"
    incr nrlines
    if {[regexp {\[} $line]} { incr level }
  }

  scan [$w bbox current] "%f %f %f %f" x1 y1 x2 y2
  set wordheight [expr $y2 - $y1]

  scan [$w coords current] "%f %f %f %f" x1 y1 x2 y2
  scan [canvasToScreen $w $x1 $y1] "%d %d" x y 
  .cdgmain balloon on $x [expr 5+ int($y + $wordheight)]  $text now

}

## ----------------------------------------------------------------------------
## Open tool tip with all predictions about the word.
## ----------------------------------------------------------------------------
body ParseTree::wordShiftClick {w } {
  set level $_levelOfCanvas($w)
  set _mouseItem [$w find withtag current]
  set timepointId $_timepointOfId($level,$_mouseItem)
  scan $timepointId "%i%c%i" wordposition useless1 useless2
  set item ""

  if {$_parse == "NULL"} {
    return
  }

  set text [$_parse getPredictions $wordposition]

  # remove final newline helpfully inserted by getPredictions()
  regsub ".$" $text "" text

  scan [$w bbox current] "%f %f %f %f" x1 y1 x2 y2
  set wordheight [expr $y2 - $y1]

  scan [$w coords current] "%f %f %f %f" x1 y1 x2 y2
  scan [canvasToScreen $w $x1 $y1] "%d %d" x y 

  .cdgmain balloon on $x 100 $text now keep

}

## ----------------------------------------------------------------------------
## Highlight word and open tool tip with its description.
## ----------------------------------------------------------------------------
body ParseTree::wordEnter {w} {

  allEnter $w
  set level $_levelOfCanvas($w)
  set _mouseItem [$w find withtag current]
  set timepointId $_timepointOfId($level,$_mouseItem)
  scan $timepointId "%i%c%i" wordposition useless1 useless2
  set item ""
  if {$_parse != "NULL" && $wordposition >= 0} {
    set item [$_parse getLexiconItem $wordposition]
  }
  if {"NULL" != $item} {
    if {$item == ""} {;# nonspec mode
      set text "nonspec"
    } else {
      set text [LexiconItemStruct_description_get $item]
    }
    scan [$w bbox current] "%f %f %f %f" x1 y1 x2 y2
    set wordheight [expr $y2 - $y1]
    scan [$w coords current] "%f %f %f %f" x1 y1 x2 y2
    scan [canvasToScreen $w $x1 $y1] "%d %d" x y 
    .cdgmain balloon on $x [expr 5+ int($y + $wordheight)]  $text
  }
}

## ----------------------------------------------------------------------------
## Downlight word and close its tooltip.
## ----------------------------------------------------------------------------
body ParseTree::wordLeave {w} {
  allLeave $w
  .cdgmain balloon maybe_off
}

## ----------------------------------------------------------------------------
## Highlight an element of the tree.
## ----------------------------------------------------------------------------
body ParseTree::allEnter {w} {

  # remember the previous color
  set _mouseItem [$w find withtag current]

  # highlight the item touched by the mouse
  mark $_mouseItem

}

## ----------------------------------------------------------------------------
## Downlight an element of the tree.
## ----------------------------------------------------------------------------
body ParseTree::allLeave {w} {

  unmark $_mouseItem
  set _mouseItem ""

}

## ----------------------------------------------------------------------
## Return the label that goes with $edge.
## ----------------------------------------------------------------------
body ParseTree::labelOfEdge {w edge} {

  set level $_levelOfCanvas($w)
  set t $_timepointOfId($level,$edge)
  foreach i $_itemsOfTimepoint($level,$t) {
    if {[lsearch -exact [$w gettags $i] "label"]>=0} {
      return $i
    }
  }
  return ""
}

## ----------------------------------------------------------------------
## Return the word widget at timepoint $t on canvas $w.
## ----------------------------------------------------------------------
body ParseTree::wordOfTimepoint {w t} {
  set level $_levelOfCanvas($w)
  foreach word $_itemsOfTimepoint($level,$t) {
    set tags [$w gettags $word]
    if {[regexp "word" $tags]} {
      return $word
    }
  }
}

## ----------------------------------------------------------------------
## Return the edge widget at timepoint $t on canvas $w.
## ----------------------------------------------------------------------
body ParseTree::edgeOfTimepoint {w t} {

  set level $_levelOfCanvas($w)
  foreach edge $_itemsOfTimepoint($level,$t) {
    set tags [$w gettags $edge]
    if {[regexp "edge" $tags]} {
      return $edge
    }
  }
}

## ----------------------------------------------------------------------
## Return the node widget at timepoint $t on canvas $w.
##
## Note that for t == -1, there can be several candidates on the canvas,
## and we select the one that is closest to the end of edge $e
## to avoid hectic scrolling.
## ----------------------------------------------------------------------
body ParseTree::nodeOfTimepoint {w e t} {

  set level $_levelOfCanvas($w)
  if {-1 != $t} {
    foreach node $_itemsOfTimepoint($level,$t) {
      set tags [$w gettags $node]
      if {[regexp "node" $tags]} {
        return $node
      }
    }
  } else {
    scan [$w coords $e] "%f %f %f %f" ex ey foo bar 
    set record 99999
    set result ""
    foreach item $_itemsOfTimepoint($level,$t) {
      set tags [$w gettags $item]
      if {[regexp "node" $tags]} {
        scan [$w coords $item] "%f %f" x y
        set distance \
            [expr sqrt(($x - $ex) * ($x - $ex) + ($y - $ey) * ($y - $ey))]
        if {$distance < $record} {
          set record $distance
          set result $item
        }
      }
    }
    return $result
  }
}

## ----------------------------------------------------------------------------
## Executed when an edge is touched.
##
## This is different from allEnter because touching an edge should
## also highlight its label.
## ----------------------------------------------------------------------------
body ParseTree::edgeEnter {w} {

  # paint edge
  allEnter $w

  # paint label
  set edge [$w find withtag current]
  set label [labelOfEdge $w $edge]
  mark $label
}

## ----------------------------------------------------------------------------
## Executed when an edge is left by the mouse pointer.
## ----------------------------------------------------------------------------
body ParseTree::edgeLeave {w} {

  # unpaint corresponding label also
  set level $_levelOfCanvas($w)
  set edge [$w find withtag current]
  set t $_timepointOfId($level,$edge)
  set label [labelOfEdge $w $edge]
  unmark $label

  # unpaint edge
  allLeave $w
}


## ----------------------------------------------------------------------------
## Display tooltip with the constraint definition in it.
## ----------------------------------------------------------------------------
body ParseTree::pop_constraint {x y X Y} {

  set index [$itk_component(conflicts) index @$x,$y]
  scan $index "%d,%d" no col
  if {$no < 0} return
  set violas ""
  if { $_parse != "NULL" } {
  	set violas [$_parse getViolations]
  }
  set conflict [lindex $violas $no]
  set id [lindex $conflict 0]
  set c [findConstraint $id]
  if {$c != "NULL"} {
    set s [formatConstraint $c]
    scan [$_parent getGeometry] \
      "%dx%d+%d+%d" width height xoff yoff
    set mx [expr int($xoff + 0.5 * $height)]
    set my [expr int($yoff + 0.5 * $width)]
    .cdgmain balloon on $mx $my $s now
  }
}

## ----------------------------------------------------------------------------
## Highlight edges of a conflict and scroll to them.
## ----------------------------------------------------------------------------
body ParseTree::violaBrowse {mode xcoord ycoord} {

  # get the violation 
  set index [$itk_component(conflicts) index @$xcoord,$ycoord]

  scan $index "%d,%d" viola_no col
  if {$viola_no < 0} return
  $itk_component(conflicts) selection clear all
  $itk_component(conflicts) selection set $viola_no,0
  
  # paint relevant canvas items
  set violas [$_parse getViolations]
  set viola [lindex $violas $viola_no]
  set score [lindex $viola 3]
  set color [rgValue $score]

  # do not use red/green values if the tree is already read and green,
  # use the highlightcolor instead
  if {1 == [[$_parent getParent] getPaintMode]} {
    set color $itk_option(-highlightcolor)
  }
  
  
  foreach {no level} $_itemsOfViolation($viola_no) {
    if {![info exists seen($level)]} {
      drawLevel $level
      paintLevel $level; # to remove the old highlighting
      set seen($level) 1
    }
    set canvas [$itk_component($level) childsite]
    $canvas itemconfigure $no -fill $color
  }

  # scroll to relevant word
  set bId [lindex $viola $mode]
  set binding [$_parse getBindingById $bId]
  set level [lindex $binding 4]
  set nr [lsearch [$_parse getLevels] $level]
  if {-1 != $nr} {
    $itk_component(tbn) view $nr
    set canvas [$itk_component($level) childsite]
    set time [lindex $binding 0]
    set item [wordOfTimepoint $canvas $time]
    centerOnItem $canvas $item
  }
}

## ----------------------------------------------------------------------
## Maybe paint the edges reddish-green according to their correctness.
## ----------------------------------------------------------------------
body ParseTree::paintLevel {level} {

  if { $_parse == "NULL" } return

  if [$_parse isAbstract] return

  if {0 == [[$_parent getParent] getPaintMode]} {

    # paint everything default colors
    set canvas [$itk_component($level) childsite]
    $canvas itemconfigure arc -fill $itk_option(-edgecolor)
    $canvas itemconfigure edge -fill $itk_option(-edgecolor)
    $canvas itemconfigure word -fill $itk_option(-wordcolor)
    $canvas itemconfigure label -fill $itk_option(-labelcolor)
    return
  }

  # paint everything green
  set color [rgValue 1]
  set canvas [$itk_component($level) childsite]
  $canvas itemconfigure arc -fill $color
  $canvas itemconfigure edge -fill $color
  $canvas itemconfigure word -fill $color
  $canvas itemconfigure label -fill $color

  # evaluate the conflicts in descending order,
  # so the reddest get drawn last
  set violas [$_parse getViolations]
  set i [llength $violas]
  while {$i > 0} {
    incr i -1
    set viola [lindex $violas $i]
    set score [lindex $viola 3]
    set color [rgValue $score]

    foreach mode {1 2} {
      set bId [lindex $viola $mode]
      if {$bId < 0} {
        continue
      }
      foreach {no lev} $_itemsOfViolation($i) {
        if {$level == $lev} {
          $canvas itemconfigure $no -fill $color
        }
      }
    }
  }
  array unset _oldColors "$level,*"
}

## ----------------------------------------------------------------------------
## compare two pointers to LexemeNode by their description.
## ----------------------------------------------------------------------------
body ParseTree::compareLexeme {a b} {
  set x [LexiconItemStruct_description_get [LexemNodeStruct_lexem_get $a]]
  set y [LexiconItemStruct_description_get [LexemNodeStruct_lexem_get $b]]
  return [string compare $x $y]

}

## ----------------------------------------------------------------------
## Find lexemes that can be exchanged for $item.
##
## The return value is a list of lists that each contain a descriptive
## string and the fitting wordSelect callback.
## ----------------------------------------------------------------------
body ParseTree::getAlternativeLexemes {w item} {

  # collect needed data
  set lexemes ""
  if {$_parse != "NULL"} {
    set lexemes [$_parse getLexemes]
  }
  if {$lexemes == ""} {
    return ""
  }

  set level $_levelOfCanvas($w)
  set timepoint $_timepointOfId($level,$item)
  if {$timepoint < 0} {
    return ""
  }
  if {$_parse != "NULL"} {
    set word [$_parse getWord $timepoint]
  }
  set from [WordStruct_from_get $word]
  set to [WordStruct_to_get $word]

  set items ""
  foreach l [lsort -command compareLexeme $lexemes] {
    set le [LexemNodeStruct_lexem_get $l]
    set d [LexiconItemStruct_description_get $le]

    # require exactly the same timespan
    set a [LexemNodeStruct_arc_get $l]
    set f [ArcStruct_from_get $a]
    set t [ArcStruct_to_get $a]
    if {$f != $from} continue
    if {$t != $to} continue
    lappend items [list $d [code $this wordSelect $w $timepoint $l]]
  }

  return $items
}

## ----------------------------------------------------------------------------
## show menu of available lexicon-entries when clicking on a word.
## ----------------------------------------------------------------------------
body ParseTree::wordClick {w} {

  .cdgmain balloon off

  # calculate position of the menu
  scan [$w bbox current] "%d %d %d %d" x1 y1 x2 y2

  set _mouseItem [$w find withtag current]
  set items [getAlternativeLexemes $w $_mouseItem]
  if {$items == ""} {
    return
  }

  showPopMenu $w $x1 $y2 $items
}

## ----------------------------------------------------------------------------
## action taking place when a lexeme is chosen from a wordlist.
## ----------------------------------------------------------------------------
body ParseTree::wordSelect {w from lexeme} {

  set oldDescription ""
  if { $_parse != "NULL" } {
    set oldDescription [$_parse getCurrentReading $from]
  }
   
  set le [LexemNodeStruct_lexem_get $lexeme]
  set description [LexiconItemStruct_description_get $le]
  set reading [LexiconItemStruct_word_get $le]
  if { $_parse != "NULL" } {
    $_parse swapWord $from $description [autoeval]
  }
  ## Saving action to history
  if { $description != $oldDescription} {
      $_commandHistory add [code $this undoLexeme $from $description]\
	  [code $this undoLexeme $from $oldDescription]
  }

  if [autoeval] {
    drawViolas
    paintTabs
  }

  # changing a different lexeme may change the reading, 
  # therefore we must redraw the word on all canvases
  foreach level [$_parse getLevels] {
    if { 1 == $_isDrawn($level)} {
      set canvas [$itk_component($level) childsite]
      set word [wordOfTimepoint $canvas $from]
      $canvas itemconfigure $word -text $reading
    }
  }

  set _kb_word $from

  # See explanation in labelSelect{}.
  set word [wordOfTimepoint $w $from]
  set level $_levelOfCanvas($w)
  set _oldColors($level,$word) [$w itemcget $word -fill]

}


## ----------------------------------------------------------------------------
## show menu of available level-labels when clicking on a label.
## ----------------------------------------------------------------------------
body ParseTree::labelClick {w} {
  scan [$w bbox current] "%d %d %d %d" x1 y1 x2 y2
  set level $_levelOfCanvas($w)
  set _mouseItem [$w find withtag current]
  set timepoint $_timepointOfId($level,$_mouseItem)

  # The current item may not actually be the label in question,
  # but its corresponding edge if it received a middle click.
  # Pretend that the event occurred at the label.
  if {[regexp "edge" [$w gettags $_mouseItem]]} {
    set _mouseItem [labelOfEdge $w $_mouseItem]
  }

  set _mouseBinding ""
  if { $_parse != "NULL" } {
    set _mouseBinding [$_parse getBindingAt $level $timepoint]
  }
  set labels [.cdgmain levels getLabels $level]
  if {$labels == "" && $_parse != "NULL"} {
    set labels [$_parse getLabels]
  }

  set items ""
  foreach label [lsort $labels] {
    lappend items [list $label [code $this labelSelect $level $timepoint $label]]
  }
  showPopMenu $w $x1 $y2 $items

}

## ----------------------------------------------------------------------------
## Auto-correct the current label
## ----------------------------------------------------------------------------
body ParseTree::labelRightClick {w} {

  if { $_parse == "NULL" } return

  # FIXME
  #
  # When a sentence is really long, and the user is really quick with
  # the mouse, a long-running libcdg operation might still be running
  # when the user clicks a second time. Since the entire libCDG is not
  # reentrant, this typically crashes since the second execution
  # thread finds its invariants mysteriously violated. At the moment we
  # protect potentially slow operations by the _computing lock, and
  # simply drop the second event. It should be put into a queue
  # and processed after the previous libcdg call is done.
  if {1 == $_computing} {
    ::cmd::Puts "WARNING: Right click ignored. Click slower!"
    return
  }
  set _computing 1

  set level $_levelOfCanvas($w)
  set canvas [$itk_component($level) childsite]
  set _mouseItem [$w find withtag current]
  set timepoint $_timepointOfId($level,$_mouseItem)
  set label ""
  set _mouseBinding ""
  set label [$_parse optimizeLabel $level $timepoint]
  set _mouseBinding [$_parse getBindingAt $level $timepoint]

  if {"NULL" != "$label"} {
    labelSelect $level $timepoint $label
  }
  set _computing 0
}

## ----------------------------------------------------------------------------
## Auto-correct the current subordination
## ----------------------------------------------------------------------------
body ParseTree::edgeRightClick {w} {
  
  if { $_parse == "NULL" } return

  if {1 == $_computing} {
    ::cmd::Puts "WARNING: Right click ignored. Click slower!"
    return
  }
  set _computing 1

  set level $_levelOfCanvas($w)
  set canvas [$itk_component($level) childsite]
  set _mouseItem [$w find withtag current]
  set modifier $_timepointOfId($level,$_mouseItem)

  set oldregent ""
  set regent ""
  set oldregent [$_parse getModifiee $level $modifier]
  set regent [$_parse optimizeStructure $level $modifier]

  if {$regent != $oldregent} {

    $_parse shiftEdge $level $modifier $regent 1
    reDrawLevel $level
    drawArcs
    drawViolas
    paintTabs
  
    $_commandHistory add \
        [code "$this undoEdge $modifier $regent $level"] \
        [code "$this undoEdge $modifier $oldregent $level"]
  }
  set _computing 0
}


## ----------------------------------------------------------------------------
## Auto-correct the current word
## ----------------------------------------------------------------------------
body ParseTree::wordRightClick {w} {

  if { $_parse == "NULL" } return

  if {1 == $_computing} {
    ::cmd::Puts "WARNING: Right click ignored. Click slower!"
    return
  }
  set _computing 1

  # remove tooltip about to be obsoleted
  .cdgmain balloon off

  # collect params
  scan [$w bbox current] "%d %d %d %d" x1 y1 x2 y2
  set level $_levelOfCanvas($w)
  set canvas [$itk_component($level) childsite]
  set _mouseItem [$w find withtag current]
  set timepoint $_timepointOfId($level,$_mouseItem)

  set l [optimizeWord $timepoint]

  wordSelect $w $timepoint $l
  
  set wordheight [expr $y2 - $y1]
  scan [canvasToScreen $w $x1 $y1] "%d %d" x y 
  set li [LexemNodeStruct_lexem_get $l]
  set d [LexiconItemStruct_description_get $li]
  .cdgmain balloon on $x [expr 5+ int($y + $wordheight)] $d now

  set _computing 0
  
}

## ----------------------------------------------------------------------
## Wrapper for Parse::optimizeWord.
## ----------------------------------------------------------------------
body ParseTree::optimizeWord {t} {

  if { $_parse == "NULL" } { return "" }

  # find description of best word
  set result [$_parse optimizeWord $t]
  if {"NULL" != "$result"} {

    # find corresponding LexemNode
    set word [$_parse getWord $t]
    set from [WordStruct_from_get $word]
    set to [WordStruct_to_get $word]
    set lexemes [$_parse getLexemes]
    foreach ln [lsort -command compareLexeme $lexemes] {
      set li [LexemNodeStruct_lexem_get $ln]
      set d [LexiconItemStruct_description_get $li]
      set a [LexemNodeStruct_arc_get $ln]
      set f [ArcStruct_from_get $a]
      set t [ArcStruct_to_get $a]
      if {$f != $from} continue
      if {$t != $to} continue
      if {![string compare $d $result]} {
        return $ln
      }
    }
  }
}

## ----------------------------------------------------------------------------
## Auto-correct a non-mainlevel subordination
## ----------------------------------------------------------------------------
body ParseTree::arcRightClick {w} {

  if { $_parse == "NULL" } { return }
  if {1 == $_computing} {
    ::cmd::Puts "WARNING: Right click ignored. Click slower!"
    return
  }
  set _computing 1

  set arc [$w find withtag current]
  set level $_levelOfItem($arc)
  set canvas [$itk_component($level) childsite]
  set modifier $_timepointOfId($level,$arc)
  set oldregent [$_parse getModifiee $level $modifier]
  set regent [$_parse optimizeStructure $level $modifier]

  if {$regent != $oldregent} {
    $_parse shiftEdge $level $modifier $regent 1
    reDrawLevel $level
    drawArcs
    drawViolas
    paintTabs
    
    $_commandHistory add \
        [code "$this undoEdge $modifier $regent $level"] \
        [code "$this undoEdge $modifier $oldregent $level"]
  }
  set _computing 0
}


## ----------------------------------------------------------------------------
## Set the label of the edge at $timepoint on $level.
## ----------------------------------------------------------------------------
body ParseTree::labelSelect {level timepoint label} {

  if { $_parse == "NULL" } { return }
  
  set binding [$_parse getBindingAt $level $timepoint]
  set modifier [lindex $binding 0]
  set modifiee [lindex $binding 1]
  set oldlabel [lindex $binding 2]
  set binding_id [lindex $binding 3]
  set level [lindex $binding 4]

  if {$label == $oldlabel} {
    return
  }

  $_commandHistory add [code $this undoLabel $modifier $label $level]\
      [code $this undoLabel $modifier $oldlabel $level]


  set canvas [$itk_component($level) childsite]
  set edge [edgeOfTimepoint $canvas $timepoint]
  set item [labelOfEdge $canvas $edge]
  $canvas itemconfigure $item -text $label
  $_parse swapLabel $level $modifier $label [autoeval]

  drawArcs

  if [autoeval] {
    drawViolas
    paintTabs
  }

  set _kb_word $timepoint

  # This routine is called via a callback via showPopMenu after a
  # label has been clicked. allLeave{} will run after we terminate,
  # and that routine will restore the label's prevous color because
  # normally that's the right thing to do. But if we actually changed
  # the label, then it will probably have received a new color by
  # paintTabs, and the saved color value is outdated here. Therefore
  # we re-set the cache value manually.
  set level $_levelOfCanvas($canvas)
  set _oldColors($level,$item) [$canvas itemcget $item -fill]

}


## ----------------------------------------------------------------------------
## Move the closest dependency edge under the mouse pointer for dragging.
## ----------------------------------------------------------------------------
body ParseTree::canvasClick {w x y} {

  set cur [$w find withtag current]
  if {"" != $cur} return

   scan [fixCoords $w $x $y] "%d %d" logical_x logical_y 

   # fond closest edge to the click
   set record 999999
   foreach e [$w find withtag edge] {
     scan [$w coords $e] "%f %f %f %f" cx1 cy1 cx2 cy2

     # compute distance of edge to click
     set d [point_segment_distance $logical_x $logical_y $cx1 $cy1 $cx2 $cy2]
     if {$d < $record} {
       set record $d
       set theedge $e
     }
   }

  # paint edge
  mark $theedge

  # move edge
  scan [$w coords $theedge] "%f %f %f %f" x1 y1 x2 y2
  set _dragItem $theedge
  set _dragStartX $x2
  set _dragStartY $y2
  moveEdge $w $x1 $y1 $logical_x $logical_y

  # pretend there already was a drag event so that the hypothetical
  # new regent will be highlighted for convenience
  edgeDrag $w $x $y

  # paint label
  set label [labelOfEdge $w $theedge]
  mark $label
}

## ----------------------------------------------------------------------------
## Prepare the touched edge for dragging.
## ----------------------------------------------------------------------------
body ParseTree::edgeClick {w x y} {
  scan [$w coords current] "%f %f %f %f" x1 y1 x2 y2
  set _dragItem [$w find withtag current]
  mark $_dragItem

  set _dragStartX $x2
  set _dragStartY $y2
  
  scan [fixCoords $w $x $y] "%d %d" new_x new_y 
  moveEdge $w $x1 $y1 $new_x $new_y

  set label [labelOfEdge $w $_dragItem]
  mark $label
}


## ----------------------------------------------------------------------------
## Find closest circle on the canvas.
##
## One should think that this should be possible with a normal
## `canvas find' command, but it isn't; you can only search for the
## closest item, OR for all circles.
## ----------------------------------------------------------------------------
body ParseTree::closestNode {w x y} {

  set result ""
  set items [$w find withtag node^word]
  set record 99999

  # find coordinates of the lower end of the moving edge; 
  # selecting this node would create a direct cycle, which look ugly
  # and is very likely wrong anyway.
  foreach item $items {
    scan [$w coords $item] "%f %f" x1 y1
    set distance [expr sqrt(($x1 - $x) * ($x1 - $x) + ($y1 - $y) * ($y1 - $y))]
    if {$distance < $record} {
      set record $distance
      set result $item
    }
  }
  
  return $result

}


## ----------------------------------------------------------------------------
## Move dependency edge around on the canvas.
##
## Moving edges around on the canvas is implemented by setting the
## private variable _dragItem to the edge that is clicked on, moving
## this edge when the mouse is dragged, and resetting the variable
## when the mouse is released.
##
## Ordinarily one would use Tcl's automatically managed canvas tag
## `current' for this; but an edge may be selected not by clicking on
## it, but by clicking on the canvas background in its vicinity, and
## Tcl is not smart enough to assign the `current' tag to an item that
## was moved under the mouse pointer after the click occurred. It also
## doesn't allow setting the `current' tag manually, so we have to
## duplicate its functionality. 
## ----------------------------------------------------------------------------
body ParseTree::edgeDrag {w x y} {
  if {$_dragItem == ""} {
    return
  }

  # actually move the edge
  scan [$w coords $_dragItem] "%f %f %f %f" x1 y1 x2 y2
  scan [fixCoords $w $x $y] "%d %d" new_x new_y
  moveEdge $w $x1 $y1 $new_x $new_y

  # find new snap node
  set level $_levelOfCanvas($w)
  set snapNode [closestNode $w $new_x $new_y]

  # if closestNode returned a word, compute the node for it
  if {[regexp {word} [$w gettags $snapNode]]} {
    set t $_timepointOfId($level,$snapNode)
    set snapNode [nodeOfTimepoint $w 0 $t]
  }

  scan [$w coords $snapNode] "%f %f" x1 y1
  set distance [expr sqrt(($x1 - $new_x) * ($x1 - $new_x) + ($y1 - $new_y) * ($y1 - $new_y))]

  if {[getSnapmode] == "nearest" || $distance < 10} {

    # highlight snap node
    $w itemconfigure node -fill $itk_option(-nodecolor)
    $w itemconfigure node -outline $itk_option(-nodecolor)
    $w itemconfigure $snapNode -fill $itk_option(-highlightcolor)
    $w itemconfigure $snapNode -outline $itk_option(-highlightcolor)
    
    # find new snap word
    set t $_timepointOfId($level,$snapNode)
    set word [wordOfTimepoint $w $t]
    set oldSnapword $_snapWord
    set _snapWord $word
    
    # maybe higlight new snap word
    if {$oldSnapword != $_snapWord} {
      unmark $oldSnapword
      mark $_snapWord
    }
  }
}


## ----------------------------------------------------------------------
## Handler for the unusual event that the user wants to pan while
## holding an edge; this is actually useful if an edge must be moved
## between two attachment points that do not fit on the canvas
## simultaneously.
## ----------------------------------------------------------------------
body ParseTree::edgeDragPan {w x y} {
  edgeDrag $w $x $y
  multiDragto $x $y
}


## ----------------------------------------------------------------------
## Cancel an ongoing move operation, i.e. return the edge to where it
## was before it was picked up. Completed moves are not undone, for
## that you need undo/redo.
## ----------------------------------------------------------------------
body ParseTree::edgeDragCancel {w} {
  if {"" == $_dragItem} return

  # return edge to where it belongs
  scan [$w coords $_dragItem] "%f %f %f %f" x1 y1 x2 y2
  moveEdge $w $x1 $y1 $_dragStartX $_dragStartY

  # downlight tree nodes
  $w itemconfigure node -fill $itk_option(-nodecolor)
  $w itemconfigure node -outline $itk_option(-nodecolor)

  # downlight edge and label
  unmark $_dragItem
  set label [labelOfEdge $w $_dragItem]
  unmark $label
  set _dragItem ""

  # downlight snap target word
  if {"" != $_snapWord} {
    $w itemconfigure node -fill $itk_option(-nodecolor)
    $w itemconfigure node -outline $itk_option(-nodecolor)
    unmark $_snapWord
    set _snapWord ""
  }
}


## ----------------------------------------------------------------------------
## Change the subordination of the dragged edge. Which node it moves to
## is dictated by the current snap mode.
## ----------------------------------------------------------------------------
body ParseTree::edgeDrop {w x y} {

  if { $_parse == "NULL" } { return }

  if {$_dragItem == "" } return

  set label [labelOfEdge $w $_dragItem]

  scan [fixCoords $w $x $y] "%d %d" new_x new_y

  set mode [getSnapmode]
  set level $_levelOfCanvas($w)
  set closest [closestNode $w $new_x $new_y]
  scan [$w coords $closest] "%f %f" x1 y1
  set distance [expr sqrt(($x1 - $new_x) * ($x1 - $new_x) + ($y1 - $new_y) * ($y1 - $new_y))]

  # The edge always snaps to the closest node if it is VERY close,
  # i.e. probably touching. The reasoning is that if the user went to
  # the trouble to drag the edge exactly onto another node, they
  # probably want that node. If we did not allow this, there would be
  # no way at all to make arbitrary structure changes with the mouse under
  # the non-default policies.
  if {$distance < 10} {
    set snapNode $closest
  } else {

    # Otherwise, follow the parent widget's drop policy:
    if {$mode == "nearest"} {
      # use the nearest regent
      set snapNode $closest
    } else {
      set canvas [$itk_component($level) childsite]
      set modifier $_timepointOfId($level,$_dragItem)
      set oldregent [$_parse getModifiee $level $modifier]
      if {$mode == "best"} {
        # use the optimal regent
        set regent [$_parse optimizeStructure $level $modifier]
        set snapNode [nodeOfTimepoint $w $_dragItem $regent]
      } elseif {$mode == "previous"} {
        # use the previous regent
        set snapNode [nodeOfTimepoint $w $_dragItem $oldregent]
      }
    }
  }

  scan [$w coords $_dragItem] "%f %f %f %f" x1 y1 x2 y2
  scan [$w coords $snapNode] "%f %f" x2 y2
  if {$snapNode != "" && !($x2 == $_dragStartX && $y2 == $_dragStartY)} {
  
      # this is a correct manipulation
      moveEdge $w $x1 $y1 [expr $x2 + 2] [expr $y2 + 2]
      set modifier $_timepointOfId($level,$_dragItem)
      set modifiee $_timepointOfId($level,$snapNode)
      if {$modifier == $modifiee} {
	set modifiee -1
      }
      set binding [$_parse getBindingAt $level $modifier]
      set modifier [lindex $binding 0]
      set oldmodifiee [lindex $binding 1]
      set label [lindex $binding 2]
      set where [lindex $binding 3]
      set thislevel [lindex $binding 4]
      set oldBinding [$_parse getBindings $thislevel]
      $_parse shiftEdge $thislevel $modifier $modifiee [autoeval]
      drawArcs
      if [autoeval] {
        drawViolas
        paintTabs
      }
      if [autoredraw] {
          reDrawLevel $thislevel
      } else {
          set _redraw_level $thislevel
      }

      ## if bindings changed, save commands to history
      if { $oldBinding != [$_parse getBindings $thislevel]} {
	  $_commandHistory add\
	      [code "$this undoEdge $modifier $modifiee $thislevel"]\
	      [code "$this undoEdge $modifier $oldmodifiee $thislevel"]
      } 

      set _kb_word $modifier

  } else {

      moveEdge $w $x1 $y1 $_dragStartX $_dragStartY
      unmark $_dragItem
      unmark $label
  }
  set _dragItem ""
}

## ----------------------------------------------------------------------------
## draw violations in the list.
## ---------------------------------------------------------------------------
body ParseTree::drawViolas {} {

  if { $_parse == "NULL" } { return }

  if {[$_parse isAbstract]} return

  set table $itk_component(conflicts)
  $table configure -state normal
  $table clear all
  $table delete rows 0 [$table cget -rows]
  $table tag config sel -fg "" -bg "" -relief raised -bd 1
  $table tag config justRight -anchor e
  $table tag config title -anchor w -background gray90 -foreground black -relief flat
  $table tag col justRight 2 4
  $table width 0 40 1 15 2 10 4 10

  # write the violation rows
  set no 0
  set violas [$_parse getViolations]
  $table configure -rows [expr [llength $violas] +1]
  foreach viola $violas {
    # build viola_text
    set rowString [list [lindex $viola 0]] ;# name
    lappend rowString [format "%4.3e" [lindex $viola 3]] ;# score
    set b1Id [lindex $viola 1]
    set b2Id [lindex $viola 2]

    # find the corresponding binding's id:
    set binding [$_parse getBindingById $b1Id]
    set modifier_id1 [lindex $binding 0]
    set modifiee_id1 [lindex $binding 1]
    set label [lindex $binding 2]
    set bindNo1 [lindex $binding 3]
    set level1 [lindex $binding 4]

    set wordstruct [$_parse getWord $modifier_id1]
    set word1 [WordStruct_word_get $wordstruct]
    if {$modifiee_id1 == -1} {
	set word2 NIL
    } elseif {$modifiee_id1 == -2} {
      set word2 NONSPEC
    } else {
      set wordstruct [$_parse getWord $modifiee_id1]
      set word2 [WordStruct_word_get $wordstruct]
    }
    lappend rowString "$level1:" "$word1-$word2"

    # more work for binary violations
    set bindNo2 ""
    if {$b2Id != -1} {

      # find the corresponding binding's id:
      set binding [$_parse getBindingById $b2Id]
      set modifier_id2 [lindex $binding 0]
      set modifiee_id2 [lindex $binding 1]
      set label [lindex $binding 2]
      set bindNo2 [lindex $binding 3]
      set level2 [lindex $binding 4]
      set wordstruct [$_parse getWord $modifier_id2]
      set word1 [WordStruct_word_get $wordstruct]
      if {$modifiee_id2 == -1} {
	set word2 NIL
      } elseif {$modifiee_id2 == -2} {
	set word2 NONSPEC
      } else {
	set wordstruct [$_parse getWord $modifiee_id2]
        set word2 [WordStruct_word_get $wordstruct]
      }
      lappend rowString "$level2:" "$word1-$word2"
    }

    # draw violation into the list
    $table setCell row $no,0 $rowString

    set _itemsOfViolation($no) ""
    incr no
  }

  set levels [$_parse getLevels]
  foreach level $levels {
    registerItems $level
  }
    
  $table configure -state disabled

  $_parent getBadness
    
}

## ----------------------------------------------------------------------------
## draw the sentence to the canvas.
## ----------------------------------------------------------------------------
body ParseTree::drawText {canvas x y} {

  if { $_parse == "NULL" } { return }

   set xx $x
   set level $_levelOfCanvas($canvas)

   set vector [$_parse getWords]
   for {set i 0} {$i < [vectorSize $vector]} {incr i} {
     set wordstruct [vectorElement $vector $i]
     set word [WordStruct_word_get $wordstruct]
     set word_id [$canvas create text $xx $y \
       -text $word \
       -anchor nw \
       -font $_currentFont \
       -fill $itk_option(-wordcolor) \
                      -tags {word text}]

     scan [$canvas bbox $word_id] "%d %d %d %d" x1 y1 x2 y2
     set width [expr $x2 - $x1]
     set _textMidPos($i) [expr $width / 2.0 + $xx]
     set xx [expr $xx + $width + 20]

     # register all ids into _timepointOfId and _itemsOfTimepoint
     set _timepointOfId($level,$word_id) $i
     lappend _itemsOfTimepoint($level,$i) $word_id
   }

   # add "..." in nonspec mode
   if {[$_parse isNonSpec]} {
     set word_id [$canvas create text $xx $y \
       -text "..." \
       -anchor nw \
       -font $_currentFont \
       -fill $itk_option(-wordcolor) \
                      -tags {word text}]

     scan [$canvas bbox $word_id] "%d %d %d %d" x1 y1 x2 y2
     set width [expr $x2 - $x1]
     set _textMidPos(-2) [expr $width / 2.0 + $xx]
     set _timepointOfId($level,$word_id) -2
     lappend _itemsOfTimepoint($level,-2) $word_id
   }
}

## ----------------------------------------------------------------------------
## Move a label to a new position.
## ----------------------------------------------------------------------------
body ParseTree::moveLabel {w label_id x1 y1 x2 y2} {
   set dx [expr ($x2 - $x1) / 2.0]
   set dy [expr ($y2 - $y1) / 2.0]

   set best_cross  4 
   #how bad is our best solution

   if {$dx == 0.0 && $dy == 0.0} {
      $w coords $label_id $x1 $x2
   } else {
      #the search starting point - a little down the road
      set mx [expr $x1 + ($dx * 0.9)]
      set my [expr $y1 + ($dy * 0.9)]
      
      if {[expr abs($dx)] > [expr abs($dy)]} {
         #seek along the x dimension
         set delta [expr $dy/$dx]
         if {$delta < 0.0} {
            set anchor1 se
            set anchor2 nw
         } else {
            set anchor1 sw
            set anchor2 ne
         }
         #preferences in worst case 
         set best_anchor $anchor1
         set best_ix [expr -($dx * 0.7)]

         $w itemconfigure $label_id -anchor $anchor1   
         for {set ix 0} {$ix < [expr abs($dx)]} {incr ix} {
            #up
            $w coords $label_id [expr $mx + $ix] [expr $my + ($ix * $delta)]
	    set bbox [mybbox $w $label_id]
            set cross [llength [eval $w find overlap $bbox]]
            if { $cross < $best_cross} {
               set best_anchor $anchor1
               set best_ix $ix
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            #another anchor
            $w itemconfigure $label_id -anchor $anchor2
	    set bbox [mybbox $w $label_id]
            set cross [llength [eval $w find overlap $bbox]]
            if { $cross < $best_cross} {
               set best_anchor $anchor2
               set best_ix $ix
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            #down
            $w coords $label_id [expr $mx - $ix] [expr $my - ($ix * $delta)]
	    set bbox [mybbox $w $label_id]
            set cross [llength [eval $w find overlap $bbox]]
            if { $cross < $best_cross} {
               set best_anchor $anchor2
               set best_ix [expr -$ix]
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            #another anchor
            $w itemconfigure $label_id -anchor $anchor1
	    set bbox [mybbox $w $label_id]
            set cross [llength [eval $w find overlap $bbox]]
            if { $cross < $best_cross} {
               set best_anchor $anchor1
               set best_ix [expr -$ix]
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
         }
         $w coords $label_id [expr $mx + $best_ix] [expr $my + ($best_ix * $delta)]
         $w itemconfigure $label_id -anchor $best_anchor
      } else { 
         #seek allong the y dimension

         set delta [expr $dx/$dy]
         if {$delta < 0} {
            set anchor1 se
            set anchor2 nw
         } else {
            set anchor1 sw
            set anchor2 ne
         }

         set best_anchor $anchor1
         set best_iy [expr -($dy * 0.7)]

         $w itemconfigure $label_id -anchor $anchor1
         for {set iy 0} {$iy < [expr abs($dy)]} {incr iy} {
            #up
            $w coords $label_id [expr $mx + ($iy * $delta)] [expr $my + $iy]
	    set bbox [mybbox $w $label_id]
            set cross [llength [eval $w find overlap $bbox]]
            if { $cross < $best_cross} {
               set best_anchor $anchor1
               set best_iy $iy
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            $w itemconfigure $label_id -anchor $anchor2
	    set bbox [mybbox $w $label_id]
            set cross [llength [eval $w find overlap $bbox]]
            if { $cross < $best_cross} {
               set best_anchor $anchor2
               set best_iy $iy
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            #down
            $w coords $label_id [expr $mx - ($iy * $delta)] [expr $my - $iy]
	    set bbox [mybbox $w $label_id]
            set cross [llength [eval $w find overlap $bbox]]
            if { $cross < $best_cross} {
               set best_anchor $anchor2
               set best_iy [expr -$iy]
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
            $w itemconfigure $label_id -anchor $anchor1
	    set bbox [mybbox $w $label_id]
            set cross [llength [eval $w find overlap $bbox]]
            if { $cross < $best_cross} {
               set best_anchor $anchor1
               set best_iy [expr -$iy]
               if {$cross < 3} {
                  break 
               }
               set best_cross $cross
            }
         }
         #set the best solution
         $w coords $label_id [expr $mx + ($best_iy * $delta)] [expr $my + $best_iy]
         $w itemconfigure $label_id -anchor $best_anchor
      }
   }
}


## ----------------------------------------------------------------------------
## move an existing edge and its label to a new position.
## ----------------------------------------------------------------------------
body ParseTree::moveEdge {w x1 y1 x2 y2} { 
  $w coords $_dragItem $x1 $y1 $x2 $y2
  set label [labelOfEdge $w $_dragItem]
  moveLabel $w $label $x1 $y1 $x2 $y2
}

## ----------------------------------------------------------------------------
## checks if all nodes between id1 and id2 are modifiers of id2.
## ----------------------------------------------------------------------------
body ParseTree::allSybillings {id1 id2 level} {

  if { $_parse == "NULL" } { return }

  set children [$_parse getModifiers $level $id2]

  if {$id2 < $id1} {
    set tmp $id1
    set id1 $id2
    set id2 $tmp
  }

  for {set i [expr $id1 + 1]} {$i < $id2} {incr i} {
    if {[lsearch -exact $children $i] < 0} {
      return 0
    }
  }

  return 1
}

## ----------------------------------------------------------------------------
## seeks an edge crossing with this id.
## ----------------------------------------------------------------------------
body ParseTree::noCrossing {id bindings} {
   foreach binding $bindings {
      set modifiee_id [lindex $binding 1]
      if {$modifiee_id == -1} continue
      set modifier_id [lindex $binding 0]
      if {$modifiee_id > $id && $modifier_id < $id} {
         return 0 }
      if {$modifier_id > $id && $modifiee_id < $id} {
         return 0 }
   }
   return 1
}


## ----------------------------------------------------------------------------
## force draw a specified level on its page.
## ----------------------------------------------------------------------------
body ParseTree::reDrawLevel {level} {
   set _isDrawn($level) 0
   drawLevel $level
}

## ----------------------------------------------------------------------------
## draw a specified level on its page.
## ----------------------------------------------------------------------------
body ParseTree::drawLevel {level} {

  # skip already drawn levels
  if { $_isDrawn($level) } {
    return
  } else {
    set _isDrawn($level) 1
  }

  # delete old drawing
  set canvas [$itk_component($level) childsite]
  
  # set the zoom Factor
  set zoomFactor $_zoomfactors($_zoomstep)
  set nodesize [expr 2 + 0.5 * $_zoomstep]

 # set the Font size
  set _currentFont [list $itk_option(-fontfamily) $_fontsizes($_zoomstep) bold]
  set labelfont "$_currentFont italic"
  
  $canvas delete all
  makeUserInterface $canvas
  set lapsTime [clock clicks -milliseconds]
  
  if { $_parse == "NULL" } {
    return;
  }
  
  # before calculating the tree's height, check for cycles and mark them to be ignored
  global EXPERIMENTAL_CYCLE_DRAWING
  
  if {$EXPERIMENTAL_CYCLE_DRAWING} {
    if { ![$_parse markOneEdgePerCycle $level] } {
	  puts "Edge marking failed"
    }
  }
  
  set heightInLinks [$_parse getHeight $level]
    
  set heightInPixels [expr ($heightInLinks * 40.0 + 40) * $zoomFactor]
  set lowerrow 0
  set middlerow -20
  set upperrow -60
  
  # ensure that _itemsOfTimepoint($level,-1) is defined
  # so that nodeOfTimepoint will not choke on cyclical trees.
  # (Isn't there a way of testing whether $a($s) is defined 
  # before accessing it??)
  set _itemsOfTimepoint($level,-1) ""

  # draw the words
  drawText $canvas 10 [expr $heightInPixels + 5.0]

  # draw bindings of level

  set bindings ""
  if { $_parse != "NULL" } {
    set bindings [$_parse getBindings $level]
  }
  
  foreach binding $bindings {
     set modifier_id [lindex $binding 0]
     set modifiee_id [lindex $binding 1]
     set label [lindex $binding 2]
     set label [join $label]
     set marked [$_parse isMarked $level $modifier_id]
     
     # some modifications for marked nodes
     if { $marked } {
       set old_modifiee_id $modifiee_id
       set modifiee_id -1
       set label [string range $label 0 [expr [string length $label] - 10]]
     }
     
     if {$heightInLinks == -1} {
      # draw cyclical structures as bipartite graph
      # TODO: remove this
       set modifier_x $_textMidPos($modifier_id)
       set modifier_y $lowerrow

       # draw lower node
       set node_id [$canvas create oval \
                        [expr $modifier_x-$nodesize] [expr $modifier_y-$nodesize] \
                        [expr $modifier_x+$nodesize] [expr $modifier_y+$nodesize] \
                        -fill $itk_option(-nodecolor) \
                        -outline $itk_option(-nodecolor) \
                        -tags node]
       $canvas raise $node_id
       set _timepointOfId($level,$node_id) $modifier_id
       lappend _itemsOfTimepoint($level,$modifier_id) $node_id

       if {$modifiee_id == -1 } {

         # draw NIL node
         set nil_x $modifier_x
         set nil_y $middlerow
         set node_id [$canvas create oval \
                          [expr $nil_x-$nodesize] [expr $nil_y-$nodesize] \
                          [expr $nil_x+$nodesize] [expr $nil_y+$nodesize] \
                          -fill $itk_option(-nodecolor) \
                          -outline $itk_option(-nodecolor) \
                          -tags node ]
         $canvas raise $node_id
         set _timepointOfId($level,$node_id) -1
         lappend _itemsOfTimepoint($level,-1) $node_id
         set modifiee_x $nil_x
         set modifiee_y $nil_y
       } else {

         # draw upper node
         set modifiee_y $upperrow
         set modifiee_x $_textMidPos($modifiee_id)
         set node_id [$canvas create oval \
                          [expr $modifiee_x-$nodesize] [expr $modifiee_y-$nodesize] \
                          [expr $modifiee_x+$nodesize] [expr $modifiee_y+$nodesize] \
                          -fill $itk_option(-nodecolor) \
                          -outline $itk_option(-nodecolor) \
                          -tags node ]
         $canvas raise $node_id
         set _timepointOfId($level,$node_id) $modifiee_id
         lappend _itemsOfTimepoint($level,$modifiee_id) $node_id
       } 
       # Herein lurks a subtle bug. If none of the bindings of this level
       # are NIL bindings, then no NIL node is ever drawn, and no mouse
       # gesture can create a NIL binding. But this can only happen if
       # *all* of the bindings are part of cycles. Unlikely.
       
       # draw edge
       set edge_id [$canvas create line 0 0 0 0 \
                        -fill $itk_option(-edgecolor) \
                        -tags edge ]
       
       $canvas lower $edge_id
       set label_id [$canvas create text 0 0 \
                         -text "$label" \
                         -anchor sw \
                         -font $labelfont \
                         -fill $itk_option(-labelcolor) \
                         -tags {label text}]
       $canvas lower $label_id
       set _timepointOfId($level,$label_id) $modifier_id
       lappend _itemsOfTimepoint($level,$modifier_id) $label_id
       set edges($label_id) $edge_id
       
       # register items and ids in _itemsOfTimepoint and _timepointOfId
       set _timepointOfId($level,$edge_id) $modifier_id
       lappend _itemsOfTimepoint($level,$modifier_id) $edge_id
       
       set _dragItem $edge_id
       moveEdge $canvas $modifier_x $modifier_y $modifiee_x $modifiee_y 
       $canvas itemconfigure $_dragItem -fill $itk_option(-edgecolor)
       set _dragItem ""

    } else {
      # draw non-cyclical structure

      set modifier_height 0
      if { $_parse != "NULL" } {
        set modifier_height [$_parse getWordHeight $level $modifier_id] 
      }
      set modifier_x $_textMidPos($modifier_id)
      set modifier_y [expr ($modifier_height * 40.0 + 40.0) * $zoomFactor]
      
      if { ($modifiee_id >= 0) || $marked } {
        # non root arc
	
	# TODO: This must be recoded to respect Edge marks
        #if (!$marked) {
	   if {($_parse != "NULL") && [$_parse isLeafNode $level $modifier_id] &&
	       [allSybillings $modifier_id $modifiee_id $level] } {
              #set the node on the bottom
              set modifier_y $heightInPixels
           }
	#}
        
        set modifiee_height 0
        if { $_parse != "NULL" } {
          if {$marked} {
	     set modifiee_height [$_parse getWordHeight $level $old_modifiee_id] 
	  } else {
	     set modifiee_height [$_parse getWordHeight $level $modifiee_id] 
	  }
        }
        if {$marked} {
          set modifiee_x $_textMidPos($old_modifiee_id)
	} else {
          set modifiee_x $_textMidPos($modifiee_id)
	}
        set modifiee_y [expr ($modifiee_height * 40.0 + 40.0) * $zoomFactor]
    
     } else {
	if {$modifiee_id == -1} {
	  # root arc
	  set modifiee_x $modifier_x
	  if {($_parse != "NULL") && [$_parse isLeafNode $level $modifier_id]} {
	     #set the nil node on the bottom
	     set modifier_y $heightInPixels
	  }
	  set modifiee_y [expr $modifier_y - 20.0 * $zoomFactor]

	  # draw nil-node
	  set node_id [$canvas create oval \
	    [expr $modifiee_x-$nodesize] [expr $modifiee_y-$nodesize] \
	    [expr $modifiee_x+$nodesize] [expr $modifiee_y+$nodesize] \
	    -fill $itk_option(-nodecolor) \
	    -outline $itk_option(-nodecolor) \
	    -tags node]
	 $canvas raise $node_id

	  # register all ids into _timepointOfId and _itemsOfTimepoint
	  set _timepointOfId($level,$node_id) $modifiee_id
	  lappend _itemsOfTimepoint($level,$modifiee_id) $node_id
	} else {
	# nonspec arc
        if {$marked} {
          set modifiee_x $_textMidPos($old_modifiee_id)
	} else {
          set modifiee_x $_textMidPos($modifiee_id)
	}
        set modifiee_height 0
        if { $_parse != "NULL" } {
          if {$marked} {
	     set modifiee_height [$_parse getWordHeight $level $old_modifiee_id] 
	  } else {
	     set modifiee_height [$_parse getWordHeight $level $modifiee_id] 
	  }
        }
        set modifiee_y [expr ($modifiee_height * 40.0 + 40.0) * $zoomFactor]
	set node_id [$canvas create oval \
	  [expr $modifiee_x-$nodesize] [expr $modifiee_y-$nodesize] \
	  [expr $modifiee_x+$nodesize] [expr $modifiee_y+$nodesize] \
	  -fill $itk_option(-nodecolor) \
	  -outline $itk_option(-nodecolor) \
	  -tags node]
	$canvas raise $node_id

        # register all ids into _timepointOfId and _itemsOfTimepoint
        set _timepointOfId($level,$node_id) $modifiee_id
        lappend _itemsOfTimepoint($level,$modifiee_id) $node_id
	}
     }

     # draw edge between node and word
     set vline_id [$canvas create line \
		  $modifier_x $modifier_y \
 		  $modifier_x $heightInPixels\
	 	  -fill $itk_option(-vlinecolor) \
		  -tags vline]

     $canvas lower $vline_id

     # draw node
     set node_id [$canvas create oval \
        	 [expr $modifier_x-$nodesize] [expr $modifier_y-$nodesize] \
		 [expr $modifier_x+$nodesize] [expr $modifier_y+$nodesize] \
		 -fill $itk_option(-nodecolor) \
		 -outline $itk_option(-nodecolor) \
		 -tags node]
     $canvas raise $node_id

     # draw edge between modifier and modifiee
      set edge_id [$canvas create line 0 0 0 0 \
                          -fill $itk_option(-edgecolor) \
                          -tags edge ]
         

      $canvas lower $edge_id
      set label_id [$canvas create text 0 0 \
                        -text "$label" \
                        -anchor sw \
                        -font $labelfont \
                        -fill $itk_option(-labelcolor) \
                        -tags {label text}]
		
    
      $canvas lower $label_id
      set _timepointOfId($level,$label_id) $modifier_id
      lappend _itemsOfTimepoint($level,$modifier_id) $label_id
      set edges($label_id) $edge_id

      # register items and ids in _itemsOfTimepoint and _timepointOfId
      set _timepointOfId($level,$edge_id) $modifier_id
      set _timepointOfId($level,$vline_id) $modifier_id
      set _timepointOfId($level,$node_id) $modifier_id
      lappend _itemsOfTimepoint($level,$modifier_id) $edge_id
      lappend _itemsOfTimepoint($level,$modifier_id) $vline_id
      lappend _itemsOfTimepoint($level,$modifier_id) $node_id
      

      set _dragItem $edge_id 
      moveEdge $canvas $modifier_x $modifier_y $modifiee_x $modifiee_y 
      $canvas itemconfigure $_dragItem -fill $itk_option(-edgecolor)
      set _dragItem ""
    }
  }

  # move the labels to the correct position
  foreach label_id [array names edges] {
    eval moveLabel $canvas $label_id [$canvas coords $edges($label_id)]
  }

  if {$level == [mainlevelname]} {
    drawArcs 
  }

  $itk_component($level) configure -autoresize 1
  $itk_component($level) yview moveto 0

  registerItems $level

  paintLevel $level
}

## ----------------------------------------------------------------------------
## Compute the array _itemsOfViolation for this level.
## ----------------------------------------------------------------------------
body ParseTree::registerItems {level} {

  if {!$_isDrawn($level)} {
    return
  }

  set mainlevel [mainlevelname]
  set no 0
  set canvas [$itk_component($level) childsite]

  set violas ""
  if { $_parse != "NULL" } {
    set violas [$_parse getViolations]
  }
  
  foreach viola $violas {
    set b1Id [lindex $viola 1]
    set b2Id [lindex $viola 2]

    # find the corresponding binding's id:
    set binding [$_parse getBindingById $b1Id]
    set modifier_id1 [lindex $binding 0]
    set modifiee_id1 [lindex $binding 1]
    set label [lindex $binding 2]
    set bindNo1 [lindex $binding 3]
    set level1 [lindex $binding 4]

    if {$level == $level1} {
      # register words, edges and labels for the first LV
      if {$modifiee_id1 != -1} {
	foreach id $_itemsOfTimepoint($level,$modifiee_id1) {
	  set id_tags [$canvas gettags $id]
	  if {[regexp "word" $id_tags]} {
	    lappend _itemsOfViolation($no) $id $level
	    break
	  }
	}
      }
      foreach id $_itemsOfTimepoint($level,$modifier_id1) {
        set id_tags [$canvas gettags $id]
        if {[regexp "edge|label|word" $id_tags]} {
          lappend _itemsOfViolation($no) $id $level
        }
      }
    }

    if {$level == [mainlevelname]} {
      # register arcs and labels for the first LV
      catch {
	foreach id $_arcOfTimepoint($level1,$modifier_id1) {
	  lappend _itemsOfViolation($no) $id $level
	}
      }
    }

    # more work for binary violations
    if {$b2Id != -1} {

      # find the corresponding binding's id:
      set binding [$_parse getBindingById $b2Id]
      set modifier_id2 [lindex $binding 0]
      set modifiee_id2 [lindex $binding 1]
      set label [lindex $binding 2]
      set bindNo2 [lindex $binding 3]
      set level2 [lindex $binding 4]

      if {$level == $level2} {
        # register edges and labels for the first LV
	if {$modifiee_id2 != -1} {
	  foreach id $_itemsOfTimepoint($level,$modifiee_id2) {
	    set id_tags [$canvas gettags $id]
	    if {[regexp "word" $id_tags]} {
	      lappend _itemsOfViolation($no) $id $level
	      break
	    }
	  }
	}
	foreach id $_itemsOfTimepoint($level,$modifier_id2) {
          set id_tags [$canvas gettags $id]
          if {[regexp "edge|label|word" $id_tags]} {
            lappend _itemsOfViolation($no) $id $level
          }
        }
      }

      if {$level == [mainlevelname]} {
      # register arcs and labels for the second LV
	catch {
	  foreach id $_arcOfTimepoint($level2,$modifier_id2) {
	    lappend _itemsOfViolation($no) $id $level
	  }
	}
      }
    }
    incr no
  }
}

## ----------------------------------------------------------------------------
## Draw everything into the levels.
## ----------------------------------------------------------------------------
body ParseTree::drawAll {} {

  if { $_parse == "NULL" } { return }

  foreach level [$_parse getLevels] {
    reDrawLevel $level
  }
  if {![$_parse isAbstract]} {
    $itk_component(conflicts) selection clear all
  }
  paintTabs
}

## ----------------------------------------------------------------------------
## Colorize the tabs in the tabnotebook.
## This is done according to the worst violation on that tab.
## ----------------------------------------------------------------------------
body ParseTree::paintTabs {} {

  if {($_parse != "NULL") && [$_parse isAbstract]} return

  set level_index [$itk_component(tbn) index select]
  if {$level_index < 0} {
    set level_index 0
  }

  set tabset [$itk_component(tbn) component tabset]
  $tabset configure -foreground black -selectforeground black

  # find the worst score of a violation for each level
  set levels "Level0"
  if {$_parse != "NULL"} {
  	set levels [$_parse getLevels]
  }
  
  foreach level $levels {
    set levelPenalty($level) 1.0
  }
  

  if {$_parse != "NULL"} {
    foreach viola [$_parse getViolations] {
      set b1Id [lindex $viola 1]
      set b2Id [lindex $viola 2]
      set score [lindex $viola 3]
      set binding [$_parse getBindingById $b1Id]
      set level1 [lindex $binding 4]

      if {$score < $levelPenalty($level1)} {
        set levelPenalty($level1) $score
      }
      if {$b2Id != -1} {

        # find the corresponding binding's id:
        set binding [$_parse getBindingById $b2Id]
        set level2 [lindex $binding 4]
        if {$score < $levelPenalty($level2)} {
          set levelPenalty($level2) $score
        }
      }
    }
  }

  # paint the tabs in the tabnotebook green or red accordingly
  foreach level $levels {

    # get a severity value from parse
    set severity $levelPenalty($level)
    set color [rgValue $severity]

    $tabset tabconfigure [$tabset index $level] \
        -foreground $color \
        -selectforeground $color 

    # maybe paint the tree itself
    paintLevel $level

  }
  
  $itk_component(tbn) view $level_index
}

## ----------------------------------------------------------------------------
## Take a score and return an appropriate red/green color for it.
## ----------------------------------------------------------------------------
body ParseTree::rgValue {score} {

  # non-linear mapping to intensity values because the intensities
  # below 0.4 are pretty much indistinguishable
  # (try to find a closed form for this if you want)
  if {$score > 0.99} {
    set intensity 0.0
  } elseif {$score > 0.95} {
    set intensity 0.1
  } elseif {$score > 0.9} {
    set intensity 0.2 
  } elseif {$score > 0.8} {
    set intensity 0.3 
  } elseif {$score > 0.7} {
    set intensity 0.4 
  } elseif {$score > 0.5} {
    set intensity 0.5 
  } elseif {$score > 0.2} {
    set intensity 0.6 
  } elseif {$score > 0.1} {
    set intensity 0.7
  } elseif {$score > 0.01} {
    set intensity 0.8 
  } elseif {$score > 0.001} {
    set intensity 0.9 
  } else {
    set intensity 1.0
  }

  set redvalue [format "%.0f" [expr 4096 * $intensity]]
  if {$redvalue > 4095} { set redvalue 4095 }
  set greenvalue [format "%.0f" [expr (4095 - $redvalue) / 1.3]]
  set color [format "#%03x%03x%03x" $redvalue $greenvalue 0]
  return $color
}

## ----------------------------------------------------------------------------
## Save the Parse to file as an annotation in cdgp input format.
## ----------------------------------------------------------------------------
body ParseTree::writeToFile {fileName} {

  if { $_parse == "NULL" } { return }
  
  if  {$fileName == ""} {
    set fileName [$this parse getId]
  }

  writeAnnotation [$_parse toAnno] $fileName

}

## ----------------------------------------------------------------------------
## Compute the screen x and y coords of canvas-coords x and y.
## ----------------------------------------------------------------------------
body ParseTree::canvasToScreen {canvas x y} {

   set rootx [winfo rootx $canvas]
   set rooty [winfo rooty $canvas]
   scan [$canvas cget -scrollregion] "%d %d %d %d" min_x min_y max_x max_y
   set left_off [lindex [$canvas xview] 0]
   set up_off [lindex [$canvas yview] 0]
   scan [expr $rootx + $x - $min_x - ($max_x - $min_x) * $left_off] "%d" X
   scan [expr $rooty + $y - $min_y - ($max_y - $min_y) * $up_off] "%d" Y

   return [list $X $Y]
}

## ----------------------------------------------------------------------------
## Translate window-based into canvas-based co-ordinates.
##
## If a scrolledcanvas is scrolled to the right, away from the home
## position, and a mouse click occurs in its upper left corner, then
## the bind event will still receive the coordinates (1,1) through its
## %x and %y parameters, even though the canvas pixel the the mouse is
## touching is something like (320,1). (Actually, because of the way
## the scrolledcanvas widget is written, an offset occurs even in the
## home position.) This routine accounts for that offset. Therefore it
## must be called whenever coordinates received through `bind' are to
## be applied to things on the canvas. 
## ----------------------------------------------------------------------------
body ParseTree::fixCoords {w x y} {

   scan [$w cget -scrollregion] "%d %d %d %d" min_x min_y max_x max_y
   set left_off [lindex [$w xview] 0]
   set up_off [lindex [$w yview] 0]
   scan [expr $x + $min_x + ($max_x - $min_x) * $left_off] "%d" new_x
   scan [expr $y + $min_y + ($max_y - $min_y) * $up_off] "%d" new_y

   return [list $new_x $new_y]
}

## ----------------------------------------------------------------------------
## Build a popup menu and show the provided items.
##
## Items is a list of pairs consisting of a label and the associated command.
## ----------------------------------------------------------------------------
body ParseTree::showPopMenu {w x y items} {

  set popmenu $itk_component(popmenu)
  $popmenu delete 0 end

  foreach item $items {
    $popmenu add command \
	-label " [lindex $item 0] " \
	-command [lindex $item 1]
  }
  update idletask

  set screenHeight [winfo screenheight $w]
  set reqHeight [winfo reqheight $popmenu]
  scan [canvasToScreen $w $x $y] "%d %d" screenX screenY
  set diffHeight [expr $screenHeight - $reqHeight -$screenY]
  set noItems [llength $items]

  if {$diffHeight<0} {

    # distribute items on some cascaded menus
    set heightPerItem [expr $reqHeight / $noItems]
    set maxItems [expr ($screenHeight -$screenY) / $heightPerItem - 2]
    set noCascades [expr ceil($noItems / double($maxItems))]
    $popmenu delete $maxItems end
    set lastCascade $popmenu
    set lastItem $maxItems

    for {set i 1} {$i < $noCascades} {incr i} {
      if {![winfo exists $lastCascade.$i]} {
	set newCascade [menu $lastCascade.$i ]
      } else {
	set newCascade $lastCascade.$i
	$newCascade delete 0 end
      }
      $lastCascade add separator
      $lastCascade add cascade \
	  -label " ... " \
	  -menu $newCascade

      for {set j 0} \
	  {$j < $maxItems && [expr $lastItem + $j] < $noItems} \
	  {incr j} {
	set item [lindex $items [expr $lastItem + $j]]
	$newCascade add command \
	    -label " [lindex $item 0] " \
	    -command [lindex $item 1]
      }
      incr lastItem $j
      set lastCascade $newCascade
    }
    update idletask
  }
  eval tk_popup $popmenu $screenX $screenY
}


## ----------------------------------------------------------------------------
## Draw all bindings not on the main level as arcs under the main tree.
## ----------------------------------------------------------------------------
body ParseTree::drawArcs {} {

  if { $_parse == "NULL" } { return }

  set mainlevel [mainlevelname]
  set canvas [$itk_component($mainlevel) childsite]

  # retract all arcs
  $canvas delete arc
  array unset _arcOfTimepoint

  set heightInLinks [$_parse getHeight $mainlevel]
  set start [expr ($heightInLinks * 40.0 + 80) * $_zoomfactors($_zoomstep)]
  set nrWords [vectorSize [$_parse getWords]]

  # collect bindings in array arcs which are worth drawing
  set arcs ""
  foreach level [$_parse getLevels] {

    # skip main level bindings
    if {$level != $mainlevel} {

      set c_level [findLevel $level]
      if {"NULL" != $c_level} {
        set no_labels [listSize [LevelStruct_labels_get $c_level]]
      } else {
        set no_labels 1
      }

      foreach binding [$_parse getBindings $level] {
	set from [lindex $binding 0]
        set to [lindex $binding 1]
        set label [lindex $binding 2]
        set bindNo [lindex $binding 3]

        # Try very hard to display `interesting' arcs and ignore all others.
        #
        # If an ancillary level has only one label, then it obviously
        # expresses no distinction. In that case, NIL edges are ignored
        # and others are drawn.
        #
        # Otherwise, all edges are drawn except those which have the empty
        # string as a label.
	if {$no_labels == 1 && $to != -1} {
	  lappend arcs [list $from $to $label $bindNo $level]
	}
	if {$no_labels > 1 && $label != ""} {
	  lappend arcs [list $from $to $label $bindNo $level]
	}
      }
    }
  }

  # Now figure out where to draw the arcs so that they don't overlap.
  # We do it like this: the space below the sentence is segmented into
  # rows and columns, starting with row 0 directly under the sentence
  # and column 0 under the first word. Now when an arc is drawn from a
  # to b, it gets drawn into the first row r with unused columns a
  # through b and then marks cells ($r,$a) through ($r,b) as used.
  
  # init cell array
  set noArcs [llength $arcs]
  for {set row 0} {$row < $noArcs} {incr row} {
    for {set column -1} {$column < $nrWords} {incr column} {
      set cell($row,$column) 0
    }
  }

  # draw all arcs
  foreach arc $arcs {

    # get the arc
    set from [lindex $arc 0]
    set to [lindex $arc 1]
    set label [lindex $arc 2]
    set bindno [lindex $arc 3]
    set level [lindex $arc 4]

    set text $label
    if {$label == ""} {
      set text $level
    }

    # a = min(from,to), z = max(from,to)
    if {-1 == $to} {
      set a [expr $from -1]
      set z [expr $from +1]
    } else {
      if {$from > $to} { 
        set a $to
        set z $from 
      } else {
        set a $from
        set z $to
      }
    }
    
    # find first partially free row
    for {set row 0} {$row < $noArcs} {incr row} {
      set rowOK 1
      for {set column $a} {$column < $z} {incr column} {
        if {$cell($row,$column) == 1} {
          set rowOK 0
          break
        }
      }
      if {$rowOK == 1} {
	break
      }
    }

    # mark cells as used
    for {set column $a} {$column < $z} {incr column} {
      set cell($row,$column) 1
    }
   
    set x1 $_textMidPos($from)
    if {-1 == $to} {
      set x2 $x1
    } else {
      set x2 $_textMidPos($to)
    }
    set y1 [expr $start + $row * 40]
    set y2 [expr $start + (1 + $row) * 40]

    # shift arrowheads so that they never overlap
    if {-1 != $to} {
      if {$from < $to} {
        set x2 [expr $x2 - 10]
      } else {
        set x2 [expr $x2 + 10]
      }
    }

    set text "$level:$label"
    if {$label == ""} {
      set text $level
    }

    set items [createArc $canvas $x1 $y1 $x2 $y2 $text $bindno]
 
    # register items
    foreach item $items {
      set _levelOfItem($item) $level
      set _timepointOfId($level,$item) $from
      lappend _arcOfTimepoint($level,$from) $item
    }

  }
}

## ----------------------------------------------------------------------------
## Create a new arc with the specified parameters.
## ----------------------------------------------------------------------------
body ParseTree::createArc {canvas x1 y1 x2 y2 text bindNo} {

  # arc
  set xmid [expr ($x2 + $x1) / 2]
  set arc [$canvas create line $x1 $y1 $xmid $y2 $x2 $y1 \
             -arrow last \
             -smooth 1 \
             -tags [list arc binding$bindNo]]
         
             #-arrowshape "10 10 6"\

  # label
  set textx [expr $x1 + ($x2 - $x1) / 2]
  set texty [expr $y2 - 5]
  set label [$canvas create text $textx $texty\
                 -text $text\
                 -font $_currentFont \
                 -tags [list text arc binding$bindNo]]

  return "$arc $label"

}

## ----------------------------------------------------------------------------
## Scan all levels of the tree horizontally.
##
## \param x the x position of the pointer
## \param y the x position of the pointer
## ----------------------------------------------------------------------------
body ParseTree::multiMark {x y} {

  if { $_parse == "NULL" } { return }

  set _lastxmark $x
  set _lastymark $y
  foreach level [$_parse getLevels] {
    $itk_component($level) scan mark $x $y
  }
}

## ----------------------------------------------------------------------------
## Scan all levels of the tree horizontally.
##
## \param x the x position of the pointer
## \param y the x position of the pointer
## ----------------------------------------------------------------------------
body ParseTree::multiDragto {x y} {

  if { $_parse == "NULL" } { return }

  set mainlevel [mainlevelname]
  set canvas [$itk_component($mainlevel) childsite]
  if {"[$canvas xview]" == "0 1"} {
    set x $_lastxmark
  }
  if {"[$canvas yview]" == "0 1"} {
    set y $_lastymark
  }

  # Restrict the actual Y coordinate used for scrolling.
  #
  # Since all levels of analysis model the same word sequence (the
  # path criterion), all canvases in a ParseTree must be the same
  # width. But the tree structures created by them can differ; usually
  # the main level with its glorious syntax tree is much higher than
  # any auxiliary levels, where only isolated edges appear. Therefore
  # we cannot use the Y coordinate extracted from the click event
  # uncritically for vertical panning on all canvases. Instead, we
  # check whether each canvas is already completely visible, and if it
  # is, we do not pan vertically at all.
  foreach level [$_parse getLevels] {
    set c $itk_component($level)
    set eff_y $y
    if {"[$c yview]" == "0 1"} {
      set eff_y 0
    }
    $c scan dragto $x $eff_y
  }
}

## ----------------------------------------------------------------------------
## Call all level canvases of the tree.
## ----------------------------------------------------------------------------
body ParseTree::multiCanvas {args} {

  if { $_parse == "NULL" } { return }

  foreach level [$_parse getLevels] {
    drawLevel $level
    if {"$args" != ""} {
      set canvas $itk_component($level)

      # suppress scrolling when there is nothing to scroll
      if [regexp {xview +(moveto|scroll)} $args] {
        if {"[$canvas xview]" == "0 1"} continue
      }
      eval $canvas $args
    }
  }
}

## ----------------------------------------------------------------------------
## Apply parseMirror.
## ----------------------------------------------------------------------------
body ParseTree::mirror {} {

  if { $_parse == "NULL" } { return }

  $_parse mirror [autoeval]
  
  if [autoeval] {
    drawViolas
  }

  foreach level [$_parse getLevels] {
    reDrawLevel $level
  }
  paintTabs

}

## ----------------------------------------------------------------------------
## Highlight differences to annotation.
## ----------------------------------------------------------------------------
body ParseTree::verify {} {

  if { $_parse == "NULL" } { return }

  set pv [$_parse verify]
  if {$pv == ""} return

  # If the tree is black and white, highlight differences in the error
  # color (usually bright red). But if it is already green and red,
  # use the highlight color (usually white) because that does not
  # conflict with tree painting.
  set color $itk_option(-errorcolor)
  if {1 == [[$_parent getParent] getPaintMode]} {
    set color $itk_option(-highlightcolor)
  }

  set mainlevel [mainlevelname]

  # ensure that all levels are actually drawn so that we can highlight
  # items in them

  # Walk ParseError list
  global PEStruct PELabel PELexem

  set errors [ParseVerificationStruct_errors_get $pv]

  for {set l $errors} {$l != "NULL"} {set l [ListStruct_next_get $l]} {
    set pe [ListStruct_item_get $l]
    set type [ParseErrorStruct_type_get $pe]
    set data [ParseErrorStruct_data_get $pe]

    if {$PEStruct == $type} {
      set se [ParseErrorStruct_data_structError_get $data]
      set level [ParseErrorStruct_data_structError_level_get $se]
      drawLevel $level
      set where [ParseErrorStruct_data_structError_modifier_get $se]
      set canvas [$itk_component($level) childsite]
      set edge [edgeOfTimepoint $canvas [expr $where - 1]]
      $canvas itemconfigure $edge -fill $color

      set label [ParseErrorStruct_data_structError_label_get $se]
      set annolabel [ParseErrorStruct_data_structError_annoLabel_get $se]
      if {$label != $annolabel} {
        set label [labelOfEdge $canvas $edge]
        $canvas itemconfigure $label -fill $color
      }
    }
    if {$PELabel == $type} {
      set le [ParseErrorStruct_data_labelError_get $data]
      set level [ParseErrorStruct_data_labelError_level_get $le]
      drawLevel $level
      set where [ParseErrorStruct_data_labelError_modifier_get $le]
      set canvas [$itk_component($level) childsite]
      set edge [edgeOfTimepoint $canvas [expr $where - 1]]
      set label [labelOfEdge $canvas $edge]
      $canvas itemconfigure $label -fill $color
    }
    if {$PELexem == $type} {
      set le [ParseErrorStruct_data_lexError_get $data]
      set where [ParseErrorStruct_data_lexError_modifier_get $le]
      set canvas [$itk_component($mainlevel) childsite]
      set word [wordOfTimepoint $canvas [expr $where - 1]]
      $canvas itemconfigure $word -fill $color
    }
  }
  
  parseVerificationDelete $pv
}

## ----------------------------------------------------------------------------
## Highlight all edges participating in cycles on the current level.
## ----------------------------------------------------------------------------
body ParseTree::showcycles {} {

  if { $_parse == "NULL" } { return }

  # process currently shown level
  set level [lindex [$_parse getLevels] [$itk_component(tbn) view]]

  # search bindings for structure
  set bindings [$_parse getBindings $level]
  foreach binding $bindings {
    set down [lindex $binding 0]
    set up [lindex $binding 1]
    set parent($down) $up
  }

  # search structure for cycles
  set max [array size parent]
  for {set i 0} {$i < [array size parent]} {incr i} {
    set cyclic($i) 0
  }

  for {set i 0} {$i < [array size parent]} {incr i} {

    # already got this?
    if { 1 == $cyclic($i) } continue
    
    set counter 0
    set j $i
    while {$j != -1} {
      incr counter
      # found cycle that doesn't involve i?
      # let him live, we'll get to him later
      if {$counter > $max} break

      set j $parent($j)
      # found cycle that involves i?
      if {$j == $i} {
        set cyclic($i) 1
        break
      }
    }
  }

  set canvas [$itk_component($level) childsite]
  set items ""
  foreach id [$canvas find withtag edge] {
    if {1 == $cyclic($_timepointOfId($level,$id))} {
      $canvas itemconfigure $id -fill $itk_option(-highlightcolor)
    }
  }
}


## ----------------------------------------------------------------------------
## Break cycles in the tree on the current level.
## 
## This may be necessary because a cycle can't be broken
## by drag & drop if there isn't already a root binding.
## ----------------------------------------------------------------------------
body ParseTree::breakcycles {} {

  if { $_parse == "NULL" } { return }

  # process currently shown level
  set level [lindex [$_parse getLevels] [$itk_component(tbn) view]]

  # search bindings for structure
  set bindings [$_parse getBindings $level]
  foreach binding $bindings {
    set down [lindex $binding 0]
    set up [lindex $binding 1]
    set parent($down) $up
  }

  # search structure for cycles
  set max [array size parent]
  for {set i 0} {$i < [array size parent]} {incr i} {
    set cyclic($i) 0
  }

  for {set i 0} {$i < [array size parent]} {incr i} {

    set counter 0
    set j $i
    while {$j != -1} {
      set j $parent($j)
      if {$j != -1} {
        set old $parent($j)
      }
      incr counter
      if {$counter > $max || $j == $i} {

        # found cycle?
        $_parse shiftEdge $level $j -1 [autoeval]

        $_commandHistory add\
            [code "$this undoEdge $j -1 $level"]\
            [code "$this undoEdge $j $old $level"]

        reDrawLevel $level
        if [autoeval] {
          drawViolas
          paintTabs
        }
        breakcycles
        return

      }
    }
  }

}


## ----------------------------------------------------------------------------
## Break cycles in the tree on the current level.
## 
## This may be necessary because a cycle can't be broken
## by drag & drop if there isn't already a root binding.
## ----------------------------------------------------------------------------
#body ParseTree::breakcycles {} {
#
#  if { $_parse == "NULL" } { return }
#
#  set level [lindex [$_parse getLevels] [$itk_component(tbn) view]]
#  
#  if {![$_parse markOneEdgePerCycle $level]} {
#    puts "breakcycles: Edge marking failed"
#    return;
#  }
#  
#  $_parse makeMarkedEdgesRootEdges $level 1
#
# TODO: this routine has to be adjusted to the shift of the marking functions to libcdg
# TODO: shift this functionality to libcdg, too
#  return;
#  
#}


## ----------------------------------------------------------------------------
## Increase or decrease zoom level of all canvases.
## ----------------------------------------------------------------------------
body ParseTree::zoom {direction} {

  # Only five zoom levels are allowed: -2/-1/0/1/2
  set newzoom [expr $_zoomstep + $direction]
  if {$newzoom > 3} {
    set newzoom 3
  }
  if {$newzoom < -3} {
    set newzoom -3
  }

  _zoom $newzoom

}


## ----------------------------------------------------------------------------
## Set zoom level for one canvas.
## ----------------------------------------------------------------------------
body ParseTree::_zoom {step} {

  set old_zoomstep $_zoomstep
  set _zoomstep $step

  if {$old_zoomstep == $_zoomstep} {
    return
  }
  
  # If we got here, each level actually needs to be redrawn.
  # It would be nicer just to rescale the canvases and leave
  # them drawn, but drawLevel{} uses the actual size of the
  # text items to group them implicitly, and that calculation
  # becomes invalid when the font size changes, so it looks better
  # to force a redraw here. 
  # This also breaks the time-saving delayed drawing of levels.
  if { $_parse != "NULL" } {
    foreach level [$_parse getLevels] {
      reDrawLevel $level
    }
  }
}

## ----------------------------------------------------------------------------
## Delegate to the _parse.
## ----------------------------------------------------------------------------
body ParseTree::parse {args} {
  if { $_parse == "NULL" } { return }

  eval $_parse $args
}

## ----------------------------------------------------------------------------
## Set the orientation of the panedwindow and reset to default fraction.
## ----------------------------------------------------------------------------
body ParseTree::setOrientation {orient} {
  if { $_parse == "NULL" } { return }
  if {[$_parse isAbstract]} return
  $itk_component(pane) configure -orient $orient 
  $itk_component(pane) fraction 80 20
}

## ----------------------------------------------------------------------------
## Get the orientation of the panedwindow.
## ----------------------------------------------------------------------------
body ParseTree::getOrientation {} {
  return [$itk_component(pane) cget -orient]
}

## ----------------------------------------------------------------------------
## Return the name of the main level.
##
## The main level is the one that gets all other edges drawn under it
## as little arcs. If the grammar does not define one, return the name
## of the first level in the parse.
## ----------------------------------------------------------------------------
body ParseTree::mainlevelname {} {
  if { $_parse == "NULL" } { return "Level0"}

  set result [.cdgmain levels getMainlevelId]
  if {$result != ""} { return $result }

  return [lindex [$_parse getLevels] 0]

}

## ----------------------------------------------------------------------------
## Highlight specified edges in color.
## ----------------------------------------------------------------------------
body ParseTree::highlight {which} {
  
  if { $_parse == "NULL" } { return }

  if { ! [regexp {(^[0-9]+(,[0-9]+)*$)?} $which] } {
    ::cmd::Puts "ERROR: Bad parse edge specification `$which'"
    ::cmd::Puts "ERROR: Highlighting failed."
    return
  }

  set levels [$_parse getLevels]
  set nolevels [llength $levels]
  set nowords [$_parse getWidth]


  regsub -all "," $which " " which
  foreach w [split $which] {

    if {![regexp {^[0-9]+$} $w]} {
      continue
    }

    set l [expr $w % $nolevels]
    set t [expr $w / $nolevels]

    # catch improper index
    if {$t >= $nowords} {
      ::cmd::Puts "WARNING: Can't highlight word $t (only $nowords words)"
      continue
    }
    
    # to be highlighted, the edge must first be drawn
    set level [lindex $levels $l]
    drawLevel $level
    set canvas [$itk_component($level) childsite]
    set edge [edgeOfTimepoint $canvas $t]
    $canvas itemconfigure $edge -fill $itk_option(-highlightcolor)
    centerOnItem $canvas $edge
    set _kb_word $t

    # maybe switch to that level if it isn't the first level?

  }
}


## ----------------------------------------------------------------------
## Change reading for given vertice to description.
## @param from vertice to change
## @param description new reading
## ----------------------------------------------------------------------
body ParseTree::undoLexeme {from description} {
  if { $_parse == "NULL" } { return }
  $_parse swapWord $from $description [autoeval]
  if [autoredraw] {
    drawAll
  }
  if [autoeval] {
    drawViolas
  }
}


## ----------------------------------------------------------------------
## Change modifier's label at given level.
## ---------------------------------------------------------------------- 
body ParseTree::undoLabel {modifier label level} {
  if { $_parse == "NULL" } { return }
  $_parse swapLabel $level $modifier $label [autoeval]
  if [autoredraw] {
    drawAll
  }
  if [autoeval] {
    drawViolas
  }
}



## ----------------------------------------------------------------------
## Shift edge (modifier) to modifiee on given level.
##
## Used to undo edge-moving actions.
## ----------------------------------------------------------------------
body ParseTree::undoEdge {modifier modifiee level} {
  if { $_parse == "NULL" } { return }
  $_parse shiftEdge $level $modifier $modifiee [autoeval]
    # update arcs on main level if necessary

  if [autoredraw] {
    # redraw the entire tree
    drawAll
  } else {

    # only shift the changed edge
    set canvas [$itk_component($level) childsite]
    set edge [edgeOfTimepoint $canvas $modifier]
    set regent [nodeOfTimepoint $canvas $edge $modifiee]
    set dependent [nodeOfTimepoint $canvas $edge $modifier]
    scan [$canvas coords $dependent] "%f %f" x1 y1
    scan [$canvas coords $regent] "%f %f" x2 y2
    set _dragItem $edge
    moveEdge $canvas $x1 $y1 $x2 $y2
    set _dragItem ""

  }
  if [autoeval] {
    drawViolas
  }
}


## ----------------------------------------------------------------------
## Add a listener to be informed of changes to the command-history
## @param listener Object-ID implementing method update {}
## ----------------------------------------------------------------------
body ParseTree::addUndoListener {listener} {
    if { $_commandHistory != "" } {
	$_commandHistory addListener $listener
    }
}


## ----------------------------------------------------------------------
## Call undo on command history.
## ----------------------------------------------------------------------
body ParseTree::undo {} {
    if {[$_commandHistory canUndo] == 1} {
	$_commandHistory undo
    }

  # undo commands can be called via Ctrl-Press events on an enclosing
  # VisParses widget, but ParseTrees believe that releasing Ctrl means
  # "apply the change to the current label previously chosen via keyboard".
  #
  # The effect is that undoing a label change via Ctrl-Z first undoes
  # the change, then triggers the Control-Release event, which
  # promptly sets the label back to what it was before. To stop it
  # from doing this, we reset _kb_label.

  # Not a very nice solution; it will break immediately when VisParses
  # receives any other Ctrl binding. Please change this to a brilliant
  # nice clean implementation when you have the time.
  set _kb_label -1

}

## ----------------------------------------------------------------------
## Call redo on command history.
## ----------------------------------------------------------------------
body ParseTree::redo {} {
    if {[$_commandHistory canRedo] == 1} {
	$_commandHistory redo
    }

  # See comment to `undo'.
  set _kb_label -1
  
}


## ----------------------------------------------------------------------
## Predicate-method rooted to command history
## @return 1 => undo can be executed, 0 => undo can not be executed
## ----------------------------------------------------------------------
body ParseTree::canUndo {} {
    return [$_commandHistory canUndo]
}

## ----------------------------------------------------------------------
## Predicate-method rooted to command history
## @return 1 => undo can be executed, 0 => undo can not be executed
## ----------------------------------------------------------------------
body ParseTree::canRedo {} {
    return [$_commandHistory canRedo]
}


## ----------------------------------------------------------------------
## Check if $i is actually visible on $w.
##
## Returns 0 only if $i is actually scrolled offscreen. $i might also
## be invisible because the entire X window is partially obscured, but
## Tcl cannot detect this.
## ----------------------------------------------------------------------
body ParseTree::itemVisible {w i} {

  scan [$w xview] "%f %f" from to
  scan [$w coords $i] "%f %f" x y
  scan [$w bbox all] "%f %f %f %f" xmin ymin xmax ymax

  set xfrom [expr $xmax * $from]
  set xto [expr $xmax * $to]

  if {$x >= $xfrom & $x <= $xto} {
    return 1
  }

  return 0
}

## ----------------------------------------------------------------------
## Scroll $w horizontally so that $i is visible.
## ----------------------------------------------------------------------
body ParseTree::centerOnItem {w i} {

  scan [$w bbox all] "%f %f %f %f" xmin ymin xmax ymax

  scan [$w xview] "%f %f" from to

  # Which point do we definitely want to see?
  scan [$w coords $i] "%f %f" x y

  # Sometimes Tcl claims that the canvas is showing nothing of the
  # content at all, which is blatantly wrong. This seems to happen
  # when the enclosing VisParses object has not yet been mapped to
  # screen. In this case, just assume the editor will get all screen
  # space; that is probably not quite true, but a lot truer than
  # getting no space at all.
  if {$from == $to} {
    puts "Hey, canvas claims to show nothing! Winging it..."
    scan [wm maxsize .] "%d %d" width dummy
  } else {
    set xfrom [expr $xmax * $from]
    set xto [expr $xmax * $to]
    set width [expr $xto - $xfrom]
  }

  # What does that mean in window fractions? 
  set xfraction [expr ($x - ($width/2)) / $xmax]

  # actually change xview
  multiCanvas xview moveto $xfraction

}

## ----------------------------------------------------------------------
## Handle keyboard editing events: 
## select the word that the next change will apply to.
## ----------------------------------------------------------------------
body ParseTree::kb_select_word {w a k} {

  # Long-running computations must block keyboard navigation. Assume
  # the user chooses word 3, presses Control-Enter (optimize label of
  # word 3) and then [Right] without waiting for the label change.
  # XCDG would compute that the best label for word 3 is `FOO', and
  # then change the label of word 4 to `FOO'! Better just to warn the
  # user to type slower.
  if {1 == $_computing} {
    ::cmd::Puts "WARNING: Key press ignored. Type slower!"
    return
  }

  if { $_parse == "NULL" } { return }
  
  set min 0
  set max [$_parse getWidth]
  set level $_levelOfCanvas($w)

  # Maybe downlight previous word and edge
  set oldword [wordOfTimepoint $w $_kb_word]
  if {[marked $oldword]} {
    unmark $oldword
    $w itemconfigure node -fill $itk_option(-nodecolor)
    $w itemconfigure node -outline $itk_option(-nodecolor)
  }

  if {"Left" == $k} {
    incr _kb_word -1
    if {$_kb_word < $min} {
      set _kb_word [expr $max - 1]
    }
  } elseif {"Right" == $k} {
    incr _kb_word
    if {$_kb_word >= $max} {
      set _kb_word $min
    }
  } elseif {"Up" == $k} {
    set r [$_parse getModifiee $level $_kb_word]
    if {-1 != $r} {
      set _kb_word $r
    }
  } elseif {"Down" == $k} {
    set rs [$_parse getModifiers $level $_kb_word]
    if {1 == [llength $rs]} {
      set _kb_word $rs
    }
  }

  set word [wordOfTimepoint $w $_kb_word]
  mark $word

  # Scroll vertically so that the sentence is visible
  $w yview moveto 1

  # Careful: if the mouse is touching an item, and we move the canvas
  # rather than mouse, the item will not receive a Leave event (Tcl
  # bug). Therefore must notify the canvas manually.
  allLeave $w


  # Maybe scroll horizontally so that the word is actually visible.
  if {! [itemVisible $w $word]} {
    centerOnItem $w $word
  }

}

## ----------------------------------------------------------------------
## Handle Shift-ed keyboard editing events:
## select hypothetical new regent
## ----------------------------------------------------------------------
body ParseTree::kb_select_edge {w a k} {

  if { $_parse == "NULL" } { return }
  
  set _computing 1

  set min -1
  set max [$_parse getWidth]
  set level $_levelOfCanvas($w)
  set oldregent [$_parse getModifiee $level $_kb_word]
  set oldword [wordOfTimepoint $w $oldregent]
  set edge [edgeOfTimepoint $w $_kb_word]

  set binding [$_parse getBindingAt $level $_kb_word]
  set a [lindex $binding 0]
  set b [lindex $binding 1]

  # if no keyboard editing is in progress, start from the current regent
  if {$_kb_regent == -2} {
    set _kb_regent $b
  }

  # shift hypothetical regent

  # Left and right are pretty unambiguous; the new regent should be
  # increased or decreased. NIL is considered to be regent -1 (that
  # is, to the left of the first word). The only question is whether
  # the first `right' command should change the regent to the word
  # after the old regent or the word after the old dependent. We use
  # the latter since that results in the smaller change to the tree.
  if {"Left" == $k} {

    if {-1 == $_kb_regent} {
      set _kb_regent $a
    }

    incr _kb_regent -1
    if {$_kb_regent == $_kb_word} {
      incr _kb_regent -1
    }
    if {$_kb_regent < $min} {
      set _kb_regent [expr $max - 1]
      if {$_kb_regent == $_kb_word} {
        set _kb_regent [expr $max - 2]
      }
    }

  } elseif {"Right" == $k} {
    if {-1 == $_kb_regent} {
      set _kb_regent $a
    }

    incr _kb_regent
    if {$_kb_regent == $_kb_word} {
      incr _kb_regent
    }
    if {$_kb_regent >= $max} {
      set _kb_regent -1
    }

  } elseif {"Up" == $k} {
    
    # Shifting the edge up means attaching it to the regent of its
    # regent. The regent of NIL is supposed to be NIL.
    if {$_kb_regent == -1} {
      set r -1
    } else {
      set r [$_parse getModifiee $level $_kb_regent]
    }
    if {$r == $_kb_word} {
      set r $_kb_regent
    }
    set _kb_regent $r

  } elseif {"Down" == $k} {

    # Shifting an edge down is ambiguous. Since there are already Left
    # and Right commands, we take it to mean that the word should be
    # attached to the nearest sibling that does not cause projectivity
    # errors.
    set xs [$_parse getModifiers $level $_kb_regent]
    set new_regent -1

    # find nearest dependent
    set record $max
    foreach x $xs {
      if {$x == $_kb_word} continue
      set dist [expr abs($x - $_kb_word)]
      if {$dist < $record} {
        set record $dist
        set new_regent $x
      }
    }

    if {-1 == $new_regent} {
      set _computing 0
      return
    }
    set _kb_regent $new_regent

  } elseif {"Return" == $k} {

    # Return emulates the right mouse button: optimize this subordination
    set _kb_regent [$_parse optimizeStructure $level $_kb_word]
    
  }

  # Downlight previously highlighted stuff
  $w itemconfigure node -fill $itk_option(-nodecolor)
  $w itemconfigure node -outline $itk_option(-nodecolor)

  # Highlight both the edge and the regent's node
  set word [wordOfTimepoint $w $_kb_word]
  set node1 [nodeOfTimepoint $w $edge $_kb_word]
  set node2 [nodeOfTimepoint $w $edge $_kb_regent]

  # It is possible that nodeOfTimepoint fails, if NIL was requested
  # and the canvas contains no NIL binding. The editing move is still
  # possible, but we cannot preview it, so we don't try.
  if {"" != $node2} {

    # If the edge is not already highlighted, remember its previous color
    if {! [marked $edge]} {
      mark $edge
    }

    # Highlight edge and regent node
    $w itemconfigure $node2 -fill $itk_option(-highlightcolor)
    $w itemconfigure $node2 -outline $itk_option(-highlightcolor)

    # Move edge
    set _dragItem $edge
    scan [$w coords $node1] "%f %f" x1 y1
    scan [$w coords $node2] "%f %f" x2 y2
    moveEdge $w $x1 $y1 $x2 $y2
    set _dragItem ""
    
    # maybe scroll to keep the edge visible
    if {! [itemVisible $w $node2]} {
      centerOnItem $w $node2
    }
  }

  set _computing 0

  # maybe call back release handler
  if {1 == $_released} {
    kb_change_edge $w $a $k
  }
}

## ----------------------------------------------------------------------
## Handle Shift-ed keyboard editing events:
## actually change the subordination 
## ----------------------------------------------------------------------
body ParseTree::kb_change_edge {w a k} {

  if { $_parse == "NULL" } { return }
  
  # maybe defer action until later
  if {1 == $_computing} {
    set _released 1
    return
  }
  
  set level $_levelOfCanvas($w)
  set oldregent [$_parse getModifiee $level $_kb_word]

  # actually edit
  if {$oldregent != $_kb_regent} {
    
    $_parse shiftEdge $level $_kb_word $_kb_regent [autoeval]
    
    # supply history
    $_commandHistory add\
        [code "$this undoEdge $_kb_word $_kb_regent $level"]\
        [code "$this undoEdge $_kb_word $oldregent $level"]
    
    # redraw
    drawArcs
    if [autoeval] {
      drawViolas
      paintTabs
    }
    if [autoredraw] {
      reDrawLevel $level
    } else {
      set _redraw_level $level
    }
  } else {
    set edge [edgeOfTimepoint $w $_kb_word]
    unmark $edge
    $w itemconfigure node -fill $itk_option(-nodecolor)
    $w itemconfigure node -outline $itk_option(-nodecolor)
  }
  
  # close keyboard editing action
  set _kb_regent -2
  set _released 0

}



## ----------------------------------------------------------------------
## Build the list _kb_labels, used by Control and Meta events
## ----------------------------------------------------------------------
body ParseTree::kb_init_label_list {w item} {

  # Make list of alternatives
  set level $_levelOfCanvas($w)
  set edge [edgeOfTimepoint $w $_kb_word]
  set item [labelOfEdge $w $edge]
  set current [$w itemcget $item -text]

  set _kb_labels [lsort [.cdgmain levels getLabels $level]]
  if {$_kb_labels == ""} {
    set _kb_labels [lsort [$_parse getLabels]]
  }

  # Find current index
  set _kb_label [lsearch $_kb_labels $current]
  
  if { -1 == $_kb_label} {
    #puts "WARNING: _kb_label is -1!"
  }
  
  # If the label is not already highlighted, remember its previous color
  if {! [marked $item]} {
    mark $item
  }
  set _oldLabel [$w itemcget $item -text]

}

## ----------------------------------------------------------------------
## Handle Control-ed keyboard events:
## select the hypothetical new label
## ----------------------------------------------------------------------
body ParseTree::kb_select_label {w a k} {
  
  if { $_parse == "NULL" } { return }
  
  set old_value $_computing
  set _computing 1

  set level $_levelOfCanvas($w)
  set edge [edgeOfTimepoint $w $_kb_word]
  set item [labelOfEdge $w $edge]
  set binding [$_parse getBindingAt $level $_kb_word]
  set a [lindex $binding 0]
  set c [lindex $binding 2]
  set current $c

  # When Control is pressed, nothing happens, but the list of
  # possible alternative labels is built because we might be needing
  # it soon. 
  if {[regexp {^Control_[LR]$} $k]} {
    
    kb_init_label_list $w $item

  } elseif {$k == "Up"} {

    # When Up and Down are pressed, the hypothetical new label is
    # changed, and the change is reflected in the string widget of that
    # label, but neither the Tcl parse nor the C Parse are actually
    # changed yet, because that takes appreciable time.
    if {$_kb_label > 0} {
      incr _kb_label -1
    }

  } elseif {$k == "Down"} {
    if {[expr 1 + $_kb_label] < [llength $_kb_labels]} {
      incr _kb_label
    }

  } elseif {$k == "Right" || $k == "Left"} {

    # Left and right have no meaning for label selection, but if we
    # didn't handle them they would be handled by kb_select_word,
    # which would leave the wrong label highlighted.
    set _computing $old_value
    return

  } elseif {$k == "Home"} {
    set _kb_label 0
    
  } elseif {$k == "End"} {
    set _kb_label [expr [llength $_kb_labels] - 1]
    
  } elseif {$k == "Return"} {

    # Return emulates the right mouse button: optimize the current label.
    set label [$_parse optimizeLabel $level $_kb_word]
    if {"NULL" != $label} {
      set _kb_label [lsearch $_kb_labels $label]
    }
  }
  set text [lindex $_kb_labels $_kb_label]
  $w itemconfigure $item -text $text

  set _computing 0

  # Maybe call the key-release handler back
  # if it was called too early to do its work
  if {1 == $_released} {
    kb_change_label $w $a $k
  }
}

## ----------------------------------------------------------------------
## Handle Meta-key events: quick jump to a specific label
## ----------------------------------------------------------------------
body ParseTree::kb_accelerate_label {w a k} {

  set level $_levelOfCanvas($w)
  set edge [edgeOfTimepoint $w $_kb_word]
  set item [labelOfEdge $w $edge]


  # When Meta is pressed, nothing happens, but the list of
  # possible alternative labels is built because we might be needing
  # it soon. 
  if {[regexp {^Meta_[LR]$} $k]} {
    kb_init_label_list $w $item
    return
  }

  # ignore unprintables such as Shift_left
  if {"" == $a} return

  set level $_levelOfCanvas($w)
  set labels [.cdgmain levels getLabels $level]

  # add character to search string
  append _label_prefix $a
  
  # find first matching label
  set found ""
  set i 0
  set len [string length $_label_prefix]
  foreach l $_kb_labels {
    if {0 == [string compare -nocase -length $len $_label_prefix $l]} {
      set found $l
      break
    }
    incr i
  }

  if {"" == $found} {
    return
  }

  set _kb_label $i
  set text [lindex $_kb_labels $_kb_label]
  $w itemconfigure $item -text $text

}


## ----------------------------------------------------------------------
## Handle Meta key release: Actually select the specified label
## ----------------------------------------------------------------------
body ParseTree::kb_unaccelerate_label {w a k} {

  set _label_prefix ""
  kb_change_label $w $a $k

}


## ----------------------------------------------------------------------
## Handle Control-ed keyboard editing events:
## actually change the label 
## ----------------------------------------------------------------------
body ParseTree::kb_change_label {w a k} {

  # detect spurious Control-Release events, i.e. those that
  # result from other Control-ed actions.
  if {-1 == $_kb_label} return

  # If this handler triggers while another handler is running, 
  # we do not yet know which label to use. This can happen when
  # parseOptimizeLabel() has been called.
  #
  # We would like to just `vwait _computing' and then use the result,
  # but Tcl doesn't work like that; the computation cannot continue
  # until we ourselves exit. So we exit, but set a flag that will let
  # kb_select_label know to call us back when it is ready.
  if {1 == $_computing} {
    set _released 1
    return
  }

  set level $_levelOfCanvas($w)
  set label [lindex $_kb_labels $_kb_label]

  # If the label was not changed, no paint action will happen,
  # so we must reset the old color ourselves.
  set edge [edgeOfTimepoint $w $_kb_word]
  set item [labelOfEdge $w $edge]
  if {$label == $_oldLabel} {
    unmark $item
  } else {
    labelSelect $level $_kb_word $label
  }

  set _kb_label -1

  set _released 0
}


## ----------------------------------------------------------------------
## Handle Alt keyboard events:
## select the hypothetical new lexeme
## ----------------------------------------------------------------------
body ParseTree::kb_select_lexeme {w a k} {

  if { $_parse == "NULL" } { return }
  
  set _computing 1

  set level $_levelOfCanvas($w)
  set item [wordOfTimepoint $w $_kb_word]

  # When Alt is pressed, nothing happens, but the list of
  # possible alternative lexemes is built because we might be needing
  # it soon. 
  if {[regexp {^Alt_[LR]$} $k]} {
    set li [$_parse getLexiconItem $_kb_word]
    if {"NULL" != $li} {
      set desc [LexiconItemStruct_description_get $li]
    }

    set _kb_lexemes [getAlternativeLexemes $w $item]
    set i 0
    foreach pair $_kb_lexemes {
      if {[lindex $pair 0] == $desc} {
        break
      }
      incr i
    }
    set _kb_lexeme $i

  } elseif {$k == "Up"} {

    # When Up and Down are pressed, the hypothetical new lexeme is
    # changed, and the change is reflected in the string widget of that
    # lexeme, but neither the Tcl parse nor the C Parse are actually
    # changed yet, because that takes appreciable time.
    if {$_kb_lexeme > 0} {
      incr _kb_lexeme -1
    }

  } elseif {$k == "Down"} {

    if {$_kb_lexeme < [llength $_kb_lexemes]} {
      incr _kb_lexeme
    }

  } elseif {$k == "Home"} {
    set _kb_lexeme 0
    
  } elseif {$k == "End"} {
    set _kb_lexeme [expr [llength $_kb_lexemes] - 1]

  } elseif {$k == "Return"} {

    # Enter emulates the right mouse click: optimize current word.
    set ln [optimizeWord $_kb_word]
    set li [LexemNodeStruct_lexem_get $ln]
    set d [LexiconItemStruct_description_get $li]
    set i 0

    foreach pair $_kb_lexemes {
      if {[lindex $pair 0] == $d} {
        break
      }
      incr i
    }
    set _kb_lexeme $i
  }

  # show tooltip
  set pair [lindex $_kb_lexemes $_kb_lexeme]
  set text [lindex $pair 0]
  
  scan [$w bbox $item] "%f %f %f %f" x1 y1 x2 y2
  set wordheight [expr $y2 - $y1]
  scan [$w coords $item] "%f %f %f %f" x1 y1 x2 y2
  scan [canvasToScreen $w $x1 $y1] "%d %d" x y 
  .cdgmain balloon on $x [expr 5+ int($y + $wordheight)]  $text now
  set _computing 0
  
  # maybe call back release handler
  if {1 == $_released} {
    kb_change_lexeme $w $a $k
  }


}

## ----------------------------------------------------------------------
## Handle Alt keyboard editing events:
## actually change the lexeme 
## ----------------------------------------------------------------------
body ParseTree::kb_change_lexeme {w a k} {

  # maybe defer action until later
  if {1 == $_computing} {
    set _released 1
    return
  }

  # actually change the parse
  set command [lindex [lindex $_kb_lexemes $_kb_lexeme] 1]
  eval $command

  # remove tooltip
  .cdgmain balloon off

  set _released 0
}


## ----------------------------------------------------------------------
## Switch to the previous level of description.
## ----------------------------------------------------------------------
body ParseTree::backwardLevel {} {
  set tbn $itk_component(tbn)
  $tbn prev
}

## ----------------------------------------------------------------------
## Switch to the next level of description.
## ----------------------------------------------------------------------
body ParseTree::forwardLevel {} {
  set tbn $itk_component(tbn)
  $tbn next
}

## ----------------------------------------------------------------------
## Scroll canvas vertically if it isn't already fully visible.
## ----------------------------------------------------------------------
body ParseTree::yscroll {args} {
  set canvas [lindex $args 0]
  set rest [lrange $args 1 end]
  if {"[$canvas yview]" != "0 1"} {
    eval $canvas yview $rest
  }
}


## ----------------------------------------------------------------------
## Return the currently visible scrolledcanvas.
## ----------------------------------------------------------------------
body ParseTree::current_canvas {} {
  set tbn $itk_component(tbn)
  set label [$tbn pageconfigure [$tbn view] -label]
  set level [lindex $label 4]
  return [$itk_component($level) childsite]
}


## ----------------------------------------------------------------------
## Remember the current color of a canvas item.
## ----------------------------------------------------------------------
body ParseTree::save_color {item} {
  set canvas [current_canvas]
  set level $_levelOfCanvas($canvas)
  if { ! [info exists _oldColors($level,$item)] } {
    set _oldColors($level,$item) [$canvas itemcget $item -fill]
  }
}

## ----------------------------------------------------------------------
## Is the item marked as `active'?
## ----------------------------------------------------------------------
body ParseTree::marked {item} {
  set canvas [current_canvas]
  set level $_levelOfCanvas($canvas)
  return [info exists _oldColors($level,$item)]
}

## ----------------------------------------------------------------------
## Mark a canvas item as `active' (and remember its old color).
## ----------------------------------------------------------------------
body ParseTree::mark {item} {
  set canvas [current_canvas]
  save_color $item
  $canvas itemconfigure $item -fill $itk_option(-highlightcolor)
}

## ----------------------------------------------------------------------
## Restore the previous color of a canvas item.
## ----------------------------------------------------------------------
body ParseTree::unmark {item} {

  if { "" == $item} return

  # An item can be highlighted both because it has been touched by the
  # mouse and because it is selected through keyboard editing, but the
  # color cache is only a pad, not a stack. Therefore we must not
  # unmark the same item twice.
  if {![marked $item]} return

  set canvas [current_canvas]
  set level $_levelOfCanvas($canvas)

  # This should not happen, all highlighting events should set _oldColors!
  if {! [info exists _oldColors($level,$item)]} {
    set _oldColors($level,$item) black
    puts "WARNING: cannot restore unsaved color of item $item!"
  } 

  # set item to previous color
  $canvas itemconfigure $item -fill $_oldColors($level,$item)
  array unset _oldColors "$level,$item"
}


## ----------------------------------------------------------------------
## How should edges behave when dropped by the user?
## ----------------------------------------------------------------------
body ParseTree::getSnapmode {} {
  return [[$_parent getParent] getSnapmode]
}

## ----------------------------------------------------------------------
## Should changes to the tree automatically trigger a redraw?
## ----------------------------------------------------------------------
body ParseTree::autoredraw {} {
  return [[$_parent getParent] getRedraw]
}

## ----------------------------------------------------------------------
## Should changes to the tree automatically trigger an evaluation?
## ----------------------------------------------------------------------
body ParseTree::autoeval {} {
  return [[$_parent getParent] getEval]
}


## ----------------------------------------------------------------------
## Return the bounding box of an item on a canvas.
##
## This wrapper is necessary because [$canvas bbox] sometimes returns
## the empty string instead of four numbers. This function always
## returns four coordinates; if the built-in bbox returns the empty
## string, 0 0 0 0 is returned.
## ----------------------------------------------------------------------
body ParseTree::mybbox {w item} {
  set bbox [$w bbox $item]
  if {$bbox == ""} {
    return "0 0 0 0"
  } else {
    return $bbox
  }
}

## ----------------------------------------------------------------------------
## Try to parse the sentence again, but respecting the user's previous
## edits.
## ----------------------------------------------------------------------------
body ParseTree::retry {} {
  $_parse retry [$_commandHistory getHistory]
}

## ----------------------------------------------------------------------
## Evaluate grammar on the parse.
## ----------------------------------------------------------------------
body ParseTree::evalParse {} {
  $_parse evalParse
  drawViolas
  paintTabs
}
