# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## CdgShell - imitate the CDG shell in tcl.
##
## \author Michael Daum (see also AUTHORS and THANKS for more)
## $Id: shell.tcl,v 1.72 2006-01-11 16:14:07 micha Exp $
##-----------------------------------------------------------------------------
class CdgShell {
   inherit ::iwidgets::Scrolledtext InfoSystemListener

   ## itk_option: configure the shell background color
   itk_option define -background background Background "gray90"

   ## itk_option: configure the shell prompt. the actual value will be
   ## set to one of -prompt1 or -prompt2
   itk_option define -prompt prompt Prompt "cdg> "

   ## itk_option: configure the shell first prompt
   itk_option define -prompt1 prompt1 Prompt "cdg> "

   ## itk_option: configure the shell second prompt
   itk_option define -prompt2 prompt2 Prompt "tcltk# "

   ## shell command history
   private variable history; ## \type TclArray

   ## index in the shell command history
   private variable historyIndex

   ## the maximum nr of commands stored in the history
   public variable maxIndex 100

   ## boolean flag with 0: command incomplete, 1: complete
   private variable completeFlag 1

   ## boolean flag with 0: application, 1: safe shell. the ctrl-t key lets
   ## you switch between the two interpreters. the application interpreter is
   ## the one executing the XCDG
   private variable interpFlag 1

   ## interpreter where all commands are executed
   private variable safeInterp ""

   ## boolean flag 0: print command result, 1: don't
   private variable silentFlag 0

   ## array of additional commands
   private variable commands; ## \type TclArray

   ## list of autotag information
   private variable tagging {}

   ## start of the commandline
   private variable firstPosition 0.0

   ## end of the commandline
   private variable lastPosition 0.0

   ## temporarily store the old shell background color here
   private variable oldBackground ""

   ## flag indicating wether next insert deletes line
   private variable needsLineFeed 0	

   private method fgets_action {} 
   private method return_action {} 
   private method tab_action {} 
   private method tabtab_action {} 
   private method up_action {} 
   private method down_action {} 
   private method left_action {} 
   private method home_action {} 
   private method end_action {} 
   private method backspace_action {} 
   private method meta_bs_action {} 
   private method delete_action {} 
   private method control_d_action {} 
   private method control_q_action {} 
   private method control_c_action {} 
   private method control_l_action {} 
   private method b1_action {x y} ; ## \type TclNumber, TclNumber
   private method double_1_action {x y} ; ## \type TclNumber, TclNumber
   private method triple_1_action {x y} ; ## \type TclNumber, TclNumber
   private method prio_action {} 
   private method next_action {} 
   private method shift_up_action {} 
   private method shift_down_action {} 
   private method switch_interp {} 
   private method getCompletions {name {parameter ""}}; ## \type TclString, TclString

   public method updateCmd {args}; ## \type TclList
   public method getCmd {}
   public method deleteCmd {} 
   public method resetCmd {args}; ## \type TclList
   public method safeEval {cmd}; ## \type TclCommand
   public method safeSource {file}; ## \type TclString
   public method insert {string}; ## \type TclString
   public method clear {}
   public method setCursor {pos}; ## \type TclString
   public method autotag {pattern args}; ## \type TclString, TclList
   public method fgets {}

   # satisfy interface InfoSystemListener
   public method insertMessage { message }; ## \type TclString
   
   constructor {args} {}; ## \type TclList

};

##-----------------------------------------------------------------------------
## constructor
##-----------------------------------------------------------------------------
body CdgShell::constructor {args} {
  array set commands {

    # original CDG commands
    activate               {::cmd::Activate section}
    annos2prolog           {::cmd::Annos2Prolog ""}
    annotation             {::cmd::Annotation anno}
    anno2parse             {::cmd::Anno2Parse anno}
    chunk                  {::cmd::Chunk lattice}
    compareparses          {::cmd::Compareparses parse}
    compile                {::cmd::Compile ""}
    constraint             {::cmd::Constraint constraint}
    deactivate             {::cmd::Deactivate section}
    distance               {::cmd::Distance net}
    edges                  {::cmd::Edges net}
    frobbing               {::cmd::Frobbing net}
    gls                    {::cmd::Gls net}
    ngls                   {::cmd::NewGls net}
    hierarchy              {::cmd::Hierarchy hierarchy}
    hook                   {::cmd::Hook hkHooks}
    incrementalcompletion  {::cmd::IncrementalCompletion lattice}
    inputwordgraph         {::cmd::Inputwordgraph ""}
    isearch                {::cmd::Isearch lattice}
    level                  {::cmd::Level level}
    levelsort              {::cmd::Levelsort level}
    lexicon                {::cmd::Lexicon word}
    license                {::cmd::License ""}
    load                   {::cmd::Load file}
    ls                     {::cmd::Ls file}
    net                    {::cmd::Net net}
    netdelete              {::cmd::Netdelete net}
    netsearch              {::cmd::Netsearch net}
    newnet                 {::cmd::Newnet lattice}
    nonspeccompatible      {::cmd::Nonspeccompatible constraint}
    parses2prolog          {::cmd::Parses2Prolog ""}
    predict                {::cmd::Predict ""}
    printparse             {::cmd::PrintParse parse}
    printparses            {::cmd::PrintParses net}
    quit                   {::cmd::Quit ""}
    renewnet               {::cmd::Renewnet net}
    reset                  {::cmd::Reset ""}
    search                 {::cmd::Search ""}
    section                {::cmd::Section ""}
    set                    {::cmd::Set variable}
    shift-reduce           {::cmd::Shift net}
    showlevel              {::cmd::Showlevel level}
    source                 {::cmd::Source file}
    status                 {::cmd::Status ""}
    tagger                 {::cmd::Tagger ""}
    testing                {::cmd::Testing ""}
    time                   {::time ""}
    useconstraint          {::cmd::Useconstraint constraint}
    uselevel               {::cmd::Uselevel level}
    verify                 {::cmd::Verify parse}
    version                {::cmd::Version}
    weight                 {::cmd::Weight constraint}
    wordgraph              {::cmd::Wordgraph lattice}
    writeannotation        {::cmd::Writeannotation net}
    writenet               {::cmd::Writenet net}
    writeparse             {::cmd::WriteParses net}
    writewordgraph         {::cmd::WriteWordgraph lattice}

    # additional XCDG commands
    clear                  {::cmd::Clear "" }
    cd                     {::cmd::Cd file}
    deleteparse            {::cmd::Deleteparse parse}
    options                {::cmd::Options file}
    printf                 {::cmd::Printf ""}
    puts                   {::cmd::Puts ""}
    pwd                    {::pwd ""}
    regexpr                {::regexpr ""}
    regsub                 {::regsub ""}
    run                    {::cmd::Source file}
    showparse              {::cmd::Showparse parse/anno}


  }

  set text [component text]

#  bind $text <KeyPress>        [code "$this updateCmd"]
  bind $text <Return>          [code "$this return_action; break" ] 
  bind $text <Tab>             [code "$this tab_action; break"]
  bind $text <Tab><Tab>        [code "$this tabtab_action; break" ]
  bind $text <Control-a>       [code "$this updateCmd; $this home_action ; break" ]
  bind $text <Control-c>       [code "$this updateCmd; $this control_c_action ; break" ]
  bind $text <Control-l>       [code "$this control_l_action ; break" ]
  bind $text <Control-d>       [code "$this control_d_action" ]
  bind $text <Control-q>       [code "$this control_q_action" ]
  bind $text <Up>              [code "$this updateCmd; $this up_action ; break" ]
  bind $text <Down>            [code "$this updateCmd; $this down_action ; break" ]
  bind $text <Home>            [code "$this updateCmd; $this home_action ; break" ]
  bind $text <End>             [code "$this updateCmd; $this end_action ; break" ]
  bind $text <BackSpace>       [code "$this updateCmd; $this backspace_action ; break" ]
  bind $text <Meta-BackSpace>  [code "$this updateCmd; $this meta_bs_action ; break" ]
  bind $text <Delete>          [code "$this updateCmd; $this delete_action ; break" ]
  bind $text <Left>            [code "$this updateCmd; $this left_action ; break" ]
  
  bind $text <1>               [code "$this b1_action  %x %y; break" ]
  bind $text <Double-1>        [code "$this double_1_action  %x %y; break" ]
  bind $text <Triple-1>        [code "$this triple_1_action  %x %y; break" ]

  bind $text <Prior>           [code "$this updateCmd; $this prio_action ; break" ]
  bind $text <Next>            [code "$this updateCmd; $this next_action ; break" ]

#  Sense ?
#  bind $text <Shift-Prior>     [code "$this prio_action ; break" ]
#  bind $text <Shift-Next>      [code "$this next_action ; break" ]

  bind $text <Shift-Up>        [code "$this updateCmd; $this shift_up_action ; break" ]
  bind $text <Shift-Down>      [code "$this updateCmd; $this shift_down_action ; break" ]

  # disable a couple of things

  #   bind $text <Select> break

  bind $text <Shift-Left> break
  bind $text <Shift-Right> break
  bind $text <Shift-Home> break
  bind $text <Shift-End> break
  bind $text <Shift-Tab> break
  #   bind $text <Shift-Select> break

  bind $text <Control-Left> break
  bind $text <Control-Right> break
  bind $text <Control-Up> break
  bind $text <Control-Down> break
  bind $text <Control-Prior> break
  bind $text <Control-Next> break
  bind $text <Control-Home> break
  bind $text <Control-Shift-Home> break
  bind $text <Control-End> break
  bind $text <Control-Tab> break
  bind $text <Control-i> break
  bind $text <Control-space> break
  bind $text <Control-slash> break
  bind $text <Control-backslash> break

  bind $text <Shift-Control-End> break
  bind $text <Shift-Control-Left> break
  bind $text <Shift-Control-Right> break
  bind $text <Shift-Control-Up> break
  bind $text <Shift-Control-Down> break
  bind $text <Shift-Control-Tab> break
  bind $text <Control-Shift-space> break

  bind $text <Control-i> [code "$this switch_interp; break"]

  focus $text

  set safeInterp [interp create -safe]

  set knownCmds [$safeInterp eval {info commands}]
  foreach srcCmd [array names commands] {
    if {[lsearch $knownCmds $srcCmd] >=0} {
      $safeInterp eval [list rename $srcCmd ""]
    }
    $safeInterp alias $srcCmd [lindex $commands($srcCmd) 0]
  }

  foreach srcCmd {fcopy fblocked fileevent flush read seek tell gets} {
    $safeInterp eval [list rename $srcCmd ""]
  }

  set history(0) ""
  set history(1) ""
  set historyIndex(0) -1
  set historyIndex(1) -1

  eval itk_initialize $args

  resetCmd
  deleteCmd
}


##----------------------------------------------------------------------------- 
## option -background.
##----------------------------------------------------------------------------- 
configbody CdgShell::background {
   configure -textbackground $itk_option(-background)
}

##----------------------------------------------------------------------------- 
## add a syntax highlightening rule.
##----------------------------------------------------------------------------- 
body CdgShell::autotag {pattern args} {
   lappend tagging $pattern
   eval tag configure $pattern $args
}

##----------------------------------------------------------------------------- 
## evaluate a command in the safe interpreter.
##----------------------------------------------------------------------------- 
body CdgShell::safeEval {cmd} {
  global errorInfo

  if {0 == $silentFlag} {
    insert \n
  }

   set oldCursor [cget -cursor]
   . configure -cursor watch
   update idletask
   
   set result ""

   set errorInfo ""
   if {$interpFlag} {

      # special treatment of inputwordgraph
      if {[regexp -lineanchor {^ *inputwordgraph (.*)$} $cmd dummy args]} {
       set cmd "inputwordgraph [string map {\" \\\" \{ \\\{ \} \\\}} $args]"
      }

      if {[catch { set result [$safeInterp eval $cmd] } errMsg] } {
	 insert "ERROR: $errMsg\n"
	 if {[.cdgmain cget -showTrace] > 0} {
	    insert "$errorInfo\n"
	 }
      }
   } else {
      if {[catch { set result [uplevel "#0" $cmd] } errMsg] } {
	 insert "ERROR: $errMsg\n"
	 if {[.cdgmain cget -showTrace] > 0} {
	    insert "$errorInfo\n"
	 }
      }
   }

   . configure -cursor $oldCursor

   if {0 == $silentFlag && $result != ""} {
     insert "$result\n"
   }

   resetCmd
   deleteCmd

   return $result
}

##----------------------------------------------------------------------------- 
## source the commands in the safe interpreter.
## \param file filename of the file to be read
##----------------------------------------------------------------------------- 
body CdgShell::safeSource {file} {
  set oldflag $silentFlag
  set silentFlag 1
  set fid [open [glob $file]]
  set script [read $fid]
  close $fid
  set lines [split $script \n]
  foreach line $lines {
    safeEval $line
  }
  set silentFlag $oldflag
  return
}

##----------------------------------------------------------------------------- 
## set the insert cursor.
## \param pos a valid tcl text widget index specification.
##----------------------------------------------------------------------------- 
body CdgShell::setCursor {pos} {
   mark set insert $pos
#   tag remove sel 1.0 end
   see insert
}

##-----------------------------------------------------------------------------
## print a string to the shell output.
## \param string the message to be printed
##-----------------------------------------------------------------------------
body CdgShell::insert {string} {
   global errorInfo

   set tempErrorInfo $errorInfo
   
   # insertions overwrite selections:
   # a habit of mine is to type something then press shift+cursormovement
   # and then replace the marked stuff with something else
   catch {
     if {[compare sel.first <= insert]
	     && [compare sel.last >= insert]} {
	 delete sel.first sel.last
     }
   }
   set errorInfo $tempErrorInfo

   ::scan [index end] %d logFirst
   incr logFirst -1

   # split string by formfeeds if available
   set splitStringList [split $string "\r"]  
   set isfirst 1
   foreach splitString $splitStringList {
      if {!$needsLineFeed && $isfirst} {
	Scrolledtext::insert end "$splitString"
	set isfirst 0
      } else {
	deleteCmd
	Scrolledtext::insert end "$splitString"
      }
   }

   # if the last list is empty then the last character is a linefeed
   # ergo the next output must first delete the line
   # we register this in `needsLineFeed'
   if {"[lindex $splitStringList end]" == ""} {
     set needsLineFeed 1
   } else {
     set needsLineFeed 0
   }
   
   ::scan [index end] %d logLast

   foreach pattern $tagging {
      for {set i $logFirst} {$i < $logLast} {incr i} {
	 mark set last $i.0
	 while {[regexp -indices $pattern [get last "last lineend"] indices]} {
	    
	    mark set first "last + [lindex $indices 0] chars"
	    mark set last  "last + 1 chars + [lindex $indices 1] chars"
	    tag add $pattern first last
	 }
      }
   }

   set firstPosition [index insert]
   set lastPosition $firstPosition
}


##----------------------------------------------------------------------------- 
## delete the current command-line
##----------------------------------------------------------------------------- 
body CdgShell::deleteCmd {} {
   ::scan $firstPosition "%d.%d" row col
   delete $row.0 $lastPosition
   
   setCursor end
   set firstPosition [index insert]
   set lastPosition $firstPosition
}


##-----------------------------------------------------------------------------
## get a single line of input from the user without interpreting it.
## This proc is only used during subprompts, e.g. as used by frobbing 
##-----------------------------------------------------------------------------
body CdgShell::fgets {} {
  
  global fgets_result 
  set fgets_result ""

  # temporarily redirect the <Return> event
  set oldBinding [bind $itk_component(text) <Return>] 
  bind $itk_component(text) <Return> [code "$this fgets_action ; break" ] 

  # fgets_action sets this variable
  # (has to be global because vwait can't handle class variables)
  vwait fgets_result
  
  # restore old event binding
  bind $itk_component(text) <Return> $oldBinding

  # return string typed by the user
  return $fgets_result
 
}

##------------------------------------------------------------------------------
## alternative &lt;Return&gt; handler,
## This version gets a line, but does not call cdg commands.
##------------------------------------------------------------------------------
body CdgShell::fgets_action {} {

  # communicate with fgets{}
  global fgets_result 
  set fgets_result [getCmd]

  # display the return key
  insert "\n"

}


##----------------------------------------------------------------------------- 
## the return key binding.
##----------------------------------------------------------------------------- 
body CdgShell::return_action {} {
   set cmd [getCmd]
   if {$cmd == ""} {
      insert "\n"
      resetCmd 
      return
   }


   set completeFlag [info complete $cmd]
   if {$completeFlag} {
      safeEval "$cmd"
      lappend history($interpFlag) $cmd
      if {$historyIndex($interpFlag) > $maxIndex} {
	 set history($interpFlag) [lreplace history($interpFlag) 0 0]
      }
      set historyIndex($interpFlag) [llength $history($interpFlag)] 

      if {$interpFlag} {
	 configure -prompt $itk_option(-prompt1)
      } else {
	 configure -prompt $itk_option(-prompt2)
      }
      resetCmd
   } else {
      updateCmd "\n"
   }
   setCursor end
}

##----------------------------------------------------------------------------- 
## get the set of possible commandline completions.
##
## \param name the string to be completed. If it is "file",
## then \c parameter is used to fork an \c "ls \c -adF"
##
## \param parameter used when name = "file"
##
## \returns the set of possible completions
##----------------------------------------------------------------------------- 
body CdgShell::getCompletions {name {parameter ""}} {

  set result ""
 
  catch {
    switch $name {
      file {
        set glob [exec /bin/sh -c "ls -adF $parameter*"]
        foreach file $glob {
          regsub {[*@=|]$} $file "" file
          lappend result $file
        }
      }
      hierarchy {
        set result [.cdgmain hierarchies getAllHierarchyIds]
      }
      net {
        set result [.cdgmain networks getAllNetIds]
      }
      parse {
        set result [.cdgmain parses getAllparseIds]
      }
      parse/anno {
      # allow names of parses or annotations
        set result [.cdgmain parses getAllparseIds]
        set result "$result [.cdgmain parses getAllAnnoIds]"
      }
      anno {
        set result [.cdgmain parses getAllAnnoIds]
      }
      constraint {
        set result [.cdgmain constraints getAllConstraintIds]
      }
      lattice {
        set result [.cdgmain wordgraphs getAllLatticeIds]
      }
      level {
        set result [.cdgmain levels getAllLevelIds]
      }
      section {
        set result [.cdgmain levels getAllSectionIds]
      }
      variable {
        set l [setReturnListOfNames]
        set result ""
        while {$l != "NULL"} {
          set item [pointer2string [ListStruct_item_get $l]]
          lappend result $item
          set l [ListStruct_next_get $l]
        }
        return $result
      }
      word {
        set result [.cdgmain lexemes getAllWords]
      }
      default {
        set result [::cmd::Set $name]
      }
    }
  }
  return $result
}

##----------------------------------------------------------------------------- 
## double tab action.
## This method is part of the commandline completion.
##----------------------------------------------------------------------------- 
body CdgShell::tabtab_action {} {
   set cmdline [getCmd]
   regsub -all {\{} $cmdline {\\\{} cmdline
   regsub -all {\}} $cmdline {\\\}} cmdline
   set no [llength $cmdline]
   set len [string length $cmdline]
   set lastChar [string index $cmdline [expr $len -1]]
   set result ""
   set pivot [lindex $cmdline 0]
   regsub -all {\.} $pivot {\.} pivot
   regsub -all {\[} $pivot {\[} pivot

   if { $lastChar == " " || $no > 1} {
      # argument expansion ########################################
      set flag 0
      if {[info exists commands($pivot)]} {
	 set alter [lindex $commands($pivot) 1]
	 if {"$alter" != ""} {
	    if {$lastChar == " "} {
	       set result [lsort [getCompletions $alter]]
	    } else {
	       set newArg [lindex $cmdline end]
	       set result [lsort [getCompletions $alter $newArg]]
	       set result [lfilter $result "^$newArg.*"]
	    }
	 }
      }
   } else {
      # command expansion ########################################
      set flag 1
      if {$interpFlag} {
	 set result [lsort [$safeInterp eval info commands "$pivot*"]]
      } else {
	 set result [lsort [uplevel "#0" info commands "$pivot*"]]
	 if {[llength $result] > 50} {
	    insert "\nERROR: too many commands\n"
	    resetCmd $cmdline
	    return
	 }
      }
   }
   # print found stuff
   switch [llength $result] {
      0  { }
      1  {
	    deleteCmd
	    if {[string range $result end end] != "/" } {
	       append result " "
	    }
	    if {$flag} {
	       resetCmd "$result"
	    } else {
	       resetCmd "$pivot $result"
	    }

	 }
      default {
	    insert "\n$result\n"
	    resetCmd $cmdline
      }
   }
}

##----------------------------------------------------------------------------- 
## tab action.
## This is part of the commandline completion suite.
##----------------------------------------------------------------------------- 
body CdgShell::tab_action {} {
   set cmdline [getCmd]
   if {$cmdline == ""} {
      return
   }
   regsub -all {\{} $cmdline {\\\{} cmdline
   regsub -all {\}} $cmdline {\\\}} cmdline
   set no [llength $cmdline]
   set len [string length $cmdline]
   set lastChar [string index $cmdline [expr $len -1]]
   set result ""
   set pivot [lindex $cmdline 0]
   regsub -all {\.} $pivot {\.} pivot
   regsub -all {\[} $pivot {\[} pivot

   if { $lastChar == " " || $no > 1} {
      # argument expansion ########################################
      set flag 0
      if {[info exists commands($pivot)]} {
	 set alter [lindex $commands($pivot) 1]
	 if {"$alter" != ""} {
	    if {$lastChar == " "} {
	       set result [lsort [getCompletions $alter]]
	    } else {
	       set newArg [lindex $cmdline end]
	       set result [lsort [getCompletions $alter $newArg]]
	       set result [lfilter $result "^$newArg.*"]
	    }
	 }
      }
   } else {
      # command expansion ########################################
      # show all possible commands (no widgets)
      set flag 1
      if {$interpFlag} {
	 set result [lsort [$safeInterp eval info commands "$pivot*"]]
      } else {
	 set result [lsort [uplevel "#0" info commands "$pivot*"]]
      }
   }

   # test if we have to compare
   set noitems [llength $result]
   if {$noitems == 0} {
      return;
   }
   if {$noitems == 1} {
      deleteCmd
      if {[string range $result end end] != "/"} {
	 append result " "
      }
      if {$flag} {
	 resetCmd "$result"
      } else {
	 if {$no > 1} {
	    resetCmd "[lrange $cmdline 0 [expr [llength $cmdline] -2]] $result"
	 } else {
	    resetCmd "[lindex $cmdline 0] $result"
	 }
      }
      return
   }
   
   #
   # extract common prefix
   #

   # copy list to array for faster access, cut away common part
   set pivot [lindex $cmdline end]
   set i 0
   foreach item $result {
      regexp "^${pivot}(.*)" $item dummy matches($i)
      incr i
   }

   # compare all first characters and collect common ones in expansion
   set expansion ""
   set found 1
   while {$found} {
      set items [lsort [array names matches]]

      # if there's nothing left to do, we quit from loop
      set noitems [llength $items] 
      if {$noitems == 0} {
	 break;
      }
   
      # get the first char from the first item
      set firstItem [lindex $items 0]
      set char [string index $matches($firstItem) 0]
      if {"$char" == ""} {
	 catch {unset matches($firstItem)}
        break;
      }

      # if there exists a non-matching target, then quit loop
      foreach i $items {
	 set found \
	    [regexp "^${char}(.*)" $matches($i) dummy matches($i)]
	 if {!$found} {
	    break;
	 }
      }

      # add common char to expansion
      if {$found} {
	 append expansion $char
      }
   }

   # print found stuff
   deleteCmd
   resetCmd "$cmdline$expansion"
}

##----------------------------------------------------------------------------- 
## step backwards thru the commandline history.
##----------------------------------------------------------------------------- 
body CdgShell::up_action {} {
   set currentPosition [index insert]
   ::scan $currentPosition "%d.%d" currentRow currentCol
   ::scan $firstPosition "%d.%d" firstRow firstCol
   if {$currentRow == $firstRow} {
      set index [expr $historyIndex($interpFlag) -1]
      set line [string trim [lindex $history($interpFlag) $index]]

      deleteCmd
      resetCmd $line
      if { "$line" != "" } {
	 incr historyIndex($interpFlag) -1
      } else {
	 set historyIndex($interpFlag) -1 
      }
   } else {
      if {[expr $currentRow - $firstRow] == 1} {
	 setCursor $firstPosition
      } else {
	 setCursor insert-1l
      }
   }
}

##----------------------------------------------------------------------------- 
## scroll forward thru the commandline history.
##----------------------------------------------------------------------------- 
body CdgShell::down_action {} {
   set currentPosition [index insert]
   ::scan $currentPosition "%d.%d" currentRow currentCol
   ::scan $lastPosition "%d.%d" lastRow lastCol
   if {$currentRow == $lastRow} {
      set index [expr $historyIndex($interpFlag) +1 ]
      set line [string trim [lindex $history($interpFlag) $index]]

      deleteCmd
      resetCmd $line
      if { "$line" != "" } {
	 incr historyIndex($interpFlag) 
      } else {
	 set historyIndex($interpFlag) [llength $history($interpFlag)]
      }
   } else {
      setCursor insert+1l
   }
}

##----------------------------------------------------------------------------- 
## restrict cursor movement to the commandline end.
##----------------------------------------------------------------------------- 
body CdgShell::left_action {} {
   ::scan [index insert] "%d.%d" currentRow currentCol
   ::scan $firstPosition "%d.%d" firstRow firstCol
   
   if {$currentRow > $firstRow || $currentCol > $firstCol} {
      setCursor insert-1c
   }
}

##----------------------------------------------------------------------------- 
## set insert cursor to the end of commandline.
##----------------------------------------------------------------------------- 
body CdgShell::end_action {} {
   setCursor $lastPosition
}

##----------------------------------------------------------------------------- 
## set the insert cursor to the start of commandline.
##----------------------------------------------------------------------------- 
body CdgShell::home_action {} {
   setCursor $firstPosition
}

##----------------------------------------------------------------------------- 
## Erase character in front of cursor
##----------------------------------------------------------------------------- 
body CdgShell::backspace_action {} {
   ::scan [index insert] "%d.%d" currentRow currentCol
   ::scan $firstPosition "%d.%d" firstRow firstCol
   
   if {$currentRow > $firstRow || $currentCol > $firstCol} {
     delete insert-1c insert
     see insert
   }
}

##----------------------------------------------------------------------------- 
## Erase word in front of cursor
##----------------------------------------------------------------------------- 
body CdgShell::meta_bs_action {} {

  # erase white space if there is any
  while {[regexp { $} [getCmd]]} {
    delete insert-1c insert
    see insert
  }

  # then erase the previous word
  while {[regexp {[^ ]$} [getCmd]]} {
    delete insert-1c insert
    see insert
  }
}

##----------------------------------------------------------------------------- 
## delete doesnt delete the selection.
##----------------------------------------------------------------------------- 
body CdgShell::delete_action {} {
   delete insert insert+1c
   see insert
}

##----------------------------------------------------------------------------- 
## kind of control-c
##----------------------------------------------------------------------------- 
body CdgShell::control_c_action {} {
  global cdgCtrlCTrapped cdgCtrlCAllowed

  if {$cdgCtrlCAllowed} {
    set cdgCtrlCTrapped 1
  }
  setCursor end
  deleteCmd
  resetCmd
}

##----------------------------------------------------------------------------- 
## terminate the application.
##----------------------------------------------------------------------------- 
body CdgShell::control_q_action {} {
  control_d_action
}

##----------------------------------------------------------------------------- 
## terminate the application.
##----------------------------------------------------------------------------- 
body CdgShell::control_d_action {} {

  # quit if pressed in an empty line...
  if {[getCmd] == ""} { ::cmd::Quit }

  # ...otherwise chain to default binding 
}

##----------------------------------------------------------------------------- 
## clear the shell screen.
##----------------------------------------------------------------------------- 
body CdgShell::clear {} {
   delete 1.0 end
}

##----------------------------------------------------------------------------- 
## clear the shell screen preventing the commandline content.
##----------------------------------------------------------------------------- 
body CdgShell::control_l_action {} {
   set cmd [getCmd]
   clear
   resetCmd $cmd
}

##----------------------------------------------------------------------------- 
## button1 only sets the selection anchor.
##----------------------------------------------------------------------------- 
body CdgShell::b1_action {x y} {
   global tkPriv
  
   set tkPriv(selectMode) char
   set tkPriv(mouseMoved) 0
   set tkPriv(pressX) $x

   mark set anchor @$x,$y
   tag remove sel 0.0 end
}

##----------------------------------------------------------------------------- 
## double-1 without changing the insert-position.
##----------------------------------------------------------------------------- 
body CdgShell::double_1_action {x y} {
   set w [component text]
   global tkPriv
   
   set tkPriv(selectMode) word

   tkTextSelectTo $w $x $y
}

##----------------------------------------------------------------------------- 
## Triple-1 without changing the insert-position.
##----------------------------------------------------------------------------- 
body CdgShell::triple_1_action {x y} {
   set w [component text]
   global tkPriv
   
   set tkPriv(selectMode) line

   tkTextSelectTo $w $x $y
}

##----------------------------------------------------------------------------- 
## simply scroll up one page, no cursor positioning.
##----------------------------------------------------------------------------- 
body CdgShell::prio_action {} {
   yview scroll -1 pages
}

##----------------------------------------------------------------------------- 
## simply scroll down one page, no cursor positioning.
##----------------------------------------------------------------------------- 
body CdgShell::next_action {} {
   yview scroll 1 pages
}

##----------------------------------------------------------------------------- 
## simply scroll up one line, no cursor positioning.
##----------------------------------------------------------------------------- 
body CdgShell::shift_up_action {} {
   yview scroll -1 units
}

##----------------------------------------------------------------------------- 
## simply scroll down one line, no cursor positioning.
##----------------------------------------------------------------------------- 
body CdgShell::shift_down_action {} {
   yview scroll 1 units
}

##----------------------------------------------------------------------------- 
## expand commandline if needed.
##----------------------------------------------------------------------------- 
body CdgShell::updateCmd {args} {
   if {$args != {}} {
      Scrolledtext::insert end [join $args]
   }
   
   set testPosition [index insert]
   ::scan $testPosition "%d.%d" testRow testCol
   ::scan $lastPosition "%d.%d" lastRow lastCol
   if {($testRow > $lastRow) || \
       ($testRow == $lastRow && $testCol > $lastCol)} {
      set lastPosition $testPosition
   }

   set testPosition [index end-1c]
   ::scan $testPosition "%d.%d" testRow testCol
   ::scan $lastPosition "%d.%d" lastRow lastCol
   if {($testRow > $lastRow) || \
       ($testRow == $lastRow && $testCol > $lastCol)} {
      set lastPosition $testPosition
   }
}

##----------------------------------------------------------------------------- 
## get actual command.
##----------------------------------------------------------------------------- 
body CdgShell::getCmd {} {
  updateCmd
   set cmd [get $firstPosition $lastPosition]
   return $cmd
}

##----------------------------------------------------------------------------- 
## reset commandline.
##----------------------------------------------------------------------------- 
body CdgShell::resetCmd {{cmd ""}} {
   setCursor end
   insert "$itk_option(-prompt)"
   set firstPosition [index insert]
   if {"$cmd" != ""} {
      updateCmd "$cmd"
      set lastPosition [index insert]
   } else {
      set lastPosition $firstPosition
   }
   return
}

##----------------------------------------------------------------------------- 
## switch interpreter. 
## This method allows you to access the interpreter that executes the XCDG application.
##----------------------------------------------------------------------------- 
body CdgShell::switch_interp {} {
   if {$completeFlag} {
      if {$interpFlag} {
	 set interpFlag 0
	 set oldBackground $itk_option(-textbackground)
	 configure \
	    -prompt $itk_option(-prompt2) \
	    -textbackground "gray80" 
      } else {
	 set interpFlag 1
	 configure \
	    -prompt $itk_option(-prompt1) \
	    -textbackground $oldBackground
      }
      deleteCmd
      resetCmd
   }
}

#----------------------------------------------------------------------------- 
# redefine the paste not to change the insert-cursor
proc tkTextPaste {w x y} {; ## \type TclWidget, TclNumber, TclNumber
   catch {
      .cdgmain shell updateCmd [selection get -displayof $w]
      .cdgmain shell setCursor end
      if {[$w cget -state] == "normal"} {focus $w}
   }
}

#-----------------------------------------------------------------------------
#-----------------------------------------------------------------------------

body CdgShell::insertMessage { message } {
	insert $message
	setCursor end
}
