# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## \file textutils.tcl
## This file contains some arbitrary usefull textutilities, mainly used in
## the CdgShell.
##
## \author Michael Daum
##
## $Id: textutils.tcl,v 1.9 2005-12-01 16:22:14 micha Exp $
## ----------------------------------------------------------------------------

##-----------------------------------------------------------------------------
## execute a script on every match in a text widget.
## \param w a text widget
## \param from the starting position from which to scan the text widget
## \param to the ending position from which to scan the text widget
## \param pattern the regular expression pattern to be matched
## \param script a pice of code executed in the callers context
##-----------------------------------------------------------------------------
proc forAllMatches {w from to pattern script} {; ## \type TextWidget, Index, Index, TclString, TclCommand
   for {set i $from} {$i < $to} {incr i} {
      $w mark set last $i.0
      while {[regexp -indices $pattern [$w get last "last lineend"] indices]} {
	 
	 $w mark set first "last + [lindex $indices 0] chars"
	 $w mark set last  "last + 1 chars + [lindex $indices 1] chars"
	 uplevel $script
      }
   }
}

##-----------------------------------------------------------------------------
## filter out items from a TclList.
## \param list a TclList
## \param pattern a regular expression pattern to be matched on each list item
## \returns a new TclList only consisting of the items matching the \c pattern
##-----------------------------------------------------------------------------
proc lfilter {list pattern} {; ## \type TclList, TclString
   set result ""
   foreach element $list {
      if {[regexp $pattern $element]} {
	 lappend result $element
      }
   }
   return $result
}


## ----------------------------------------------------------------------------
## compare strings intelligently like ls -v does
## ----------------------------------------------------------------------------
proc smartCompare {a b} {; ## \type TclString, TclString

  while {$a != "" && $b != ""} {

    # remove common prefixes (except numbers)
    while {[string index $a 0] == [string index $b 0]} {
      if {[regexp {^[0-9]} $a]} {
        break
      }
      regsub {^.} $a "" a
      regsub {^.} $b "" b
    }

    # if numbers follow, compare the numbers
    if {[regexp {^[0-9]+} $a numa] && [regexp {^[0-9]+} $b numb]} {
      # unequal numbers found?

      # have to prefix 1 here so expr won't choke on numbers like `028'
      if {[expr "1$numa" < "1$numb"]} {
        return -1
      }
      if {[expr "1$numa" > "1$numb"]} {
        return 1
      }
      # remove equal numbers
      regsub {^[0-9]+} $a "" a
      regsub {^[0-9]+} $b "" b
      continue
    } else {
      # compare normally
      return [string compare $a $b]
    }
  }
  return 0;
}

## ----------------------------------------------------------------------------
## \{
## \name Distances between line segments
##
## The following code to compute distances between line segments and
## points stolen shamelessly from Rory Daulton (rorydaulton@email.com),
## who posted it to delphi-talk@elists.org.
## ----------------------------------------------------------------------------

## ----------------------------------------------------------------------------
## Convert two directed line segments, starting from the same
## point, to vectors
## ----------------------------------------------------------------------------
proc segments_to_vectors {ex0 ey0 ex1 ey1 x y} {; ## \type TclNumber, TclNumber, TclNumber, TclNumber, TclNumber, TclNumber

  set x1 [expr $ex1 - $ex0]
  set y1 [expr $ey1 - $ey0]
  set x2 [expr $x - $ex0]
  set y2 [expr $y - $ey0]

  return [list $x1 $y1 $x2 $y2]

}

## ----------------------------------------------------------------------------
## Find the dot product of two vectors
## ----------------------------------------------------------------------------
proc dot {x0 y0 x1 y1} {; ## \type TclNumber, TclNumber, TclNumber, TclNumber
  return [expr $x0 * $x1 + $y0 * $y1]
}

## ----------------------------------------------------------------------------
## Find the length of a line segment
## ----------------------------------------------------------------------------
proc segment_length {x0 y0 x1 y1} {; ## \type TclNumber, TclNumber, TclNumber, TclNumber
  return [expr sqrt (pow($x0-$x1, 2) + pow($y0-$y1, 2))]
}

## ----------------------------------------------------------------------------
## Calculate the signed area of the parallelogram that is formed by
## two vectors (by the origin and two given points).
## ----------------------------------------------------------------------------
proc par_area {x1 y1 x2 y2} {; ## \type TclNumber, TclNumber, TclNumber, TclNumber
  return [expr $x1 * $y2 - $y1 * $x2]
}

## ----------------------------------------------------------------------------
## Give distance of a point from a line segment.
## ----------------------------------------------------------------------------
proc point_segment_distance {x y ex0 ey0 ex1 ey1} {; ## \type TclNumber, TclNumber, TclNumber, TclNumber, TclNumber, TclNumber

  # convert the points to vectors
  scan [segments_to_vectors $ex0 $ey0 $ex1 $ey1 $x $y]\
  "%f %f %f %f"  svx svy pvx pvy

  # use the dot product to check the angles, to decide whether to use the
  # end points or a point in the segment
  set dp [dot $svx $svy $pvx $pvy]

  if {$dp <= 0} {
    
    # angle at segment start is not acute: use distance from point to
    #   segment's start
    set result [expr sqrt ( $pvx*$pvx + $pvy*$pvy) ]

  } else {

    set sl [expr $svx*$svx + $svy*$svy]
    if {$dp >= $sl} {

      # angle at segment end is not acute: use distance from point to
      #   segment's end
      set result [segment_length $x $y $ex1 $ey1]
    } else {

      # both angles are acute: use distance from point to the entire line
      set result [expr abs( [par_area $svx $svy $pvx $pvy] / sqrt( $sl ))]
    } 
  }

  return $result

}
## ----------------------------------------------------------------------------
## \} end if chapter "Distances of line segments"
## ----------------------------------------------------------------------------
