# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## VisParses - the grafical tree editor.
## 
## A VisParses object instanciate one toplevel window that embedds a set
## of ParseTree objects to be displayed. There are two common situations that
## determine the set of parses: (1) displaying annotations, (2) displaying
## parser solutions. In the first case one VisParses object can fetch all
## annotations loaded into the system. In the second case one VisParses object
## displays all solutions a specific parsing flavour has found for one wordgraph.
## So in the latter case another parser cannot reuse the same VisParses object
## to visualize its solution record also, but needs a new VisParses object of
## its own, giving you two separate toplevel windows. I think this is the best
## compromise here between lots of toplevel windows and embedded objects in
## separate tabs of one window.
##
## As you see below we don't provide a separate destructor to destruct
## embedded ParseTree objects but add ParseTree objects directly as itk-components
## to the VisParses widget. So they get destructed automatically.
##
## \author Michael Daum, Kilian A. Foth, Dietmar Fnning 
## (see also AUTHORS and THANKS for more)
##
## $Id visparses.tcl $
## ----------------------------------------------------------------------------
class VisParses {
   inherit ::itk::Toplevel TreeEditorParent

   itk_option define -iconname iconName IconName "xcdg"
   itk_option define -interactive interactive Interactive 0

   itk_option define -autoredraw autoredraw Autoredraw 1
   itk_option define -autoeval autoeval Autoeval 1
   itk_option define -painttree painttree Painttree 0

   public method addParse {parse}; ## \type ParseTree
   public method removeParse {parse}; ## \type ParseTree
   public method view {no}; ## \type TclNumber
   public method getSize {}
   public method center {}
   public method readWord {}
   public method getBadness {}
   public method getGeometry {}
   public method setGeometry {geometry} ; ## \type TclString
   public method update {}
   
   # TreeEditorParent implementation
   public method setTitle { title }; ## \type TclString
   public method getTitle {};
   public method setUndo {state}; ## \type TclNumber
   public method setRedo {state}; ## \type TclNumber
   
   # Snapmode
   private variable _snapmode "nearest"
   public method getSnapmode {}

   private variable _redrawFlag 1
   private method _doRedraw {}
   public method getRedraw {}
   public method setRedraw {x}; ## \type TclNumber

   private variable _evalFlag 1
   private method _doEval {}
   public method getEval {}
   public method setEval {x}; ## \type TclNumber

   private variable _paintFlag 1
   private method _doPaint {}
   public method getPaintMode {}
   public method setPaintMode {x}; ## \type TclNumber
   
   constructor {args} {}; ## \type TclList

   private method _setOrientation {orien}; ## \type TclString
   
   private variable _currentOrient
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body VisParses::constructor {args} {
  
  global queue env
  set queue ""

   wm withdraw $itk_interior
   
   # Tree Editor
   itk_component add treeeditor {
	TreeEditor $itk_interior.treeeditor $this
   } {
   }

   # menu
   itk_component add menuframe {
      frame $itk_interior.menuframe
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   # file menu
   itk_component add filebutton {
      menubutton $itk_component(menuframe).filebutton \
		-text "File" \
		-menu $itk_component(menuframe).filebutton.menu
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   itk_component add filemenu {
      menu $itk_component(filebutton).menu 
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }


#    $itk_component(filemenu) add command \
# 		-label "Load ..." \
# 		-underline 0 \
# 		-command [code $itk_component(treeeditor) loadAction]
   $itk_component(filemenu) add command \
		-label "Save" \
		-underline 0 \
                -command [code $itk_component(treeeditor) saveAction]
   $itk_component(filemenu) add command \
		-label "Save as..." \
		-underline 0 \
                -command [code $itk_component(treeeditor) saveAsAction]
   $itk_component(filemenu) add command \
		-label "Print ..." \
		-underline 0 \
		-command [code $itk_component(treeeditor) print]
   $itk_component(filemenu) add separator
   $itk_component(filemenu) add command \
		-label "Close" \
		-underline 0 \
		-command [list destroy $itk_interior]
   $itk_component(filemenu) add command \
		-label "Quit" \
		-underline 0 \
		-command ::cmd::Quit

   # editmenu
   itk_component add editbutton {
      menubutton $itk_component(menuframe).editbutton \
		-text "Edit" \
		-menu $itk_component(menuframe).editbutton.menu
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   itk_component add editmenu {
      menu $itk_component(editbutton).menu 
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

  ## Configure Edit-Menu with Undo-/Redo-Entries
  $itk_component(editmenu) add command \
                -label "Undo" \
      -command [code $itk_component(treeeditor) undo]\
      -accelerator <Ctrl-z>


  $itk_component(editmenu) add command \
                -label "Redo" \
      -command [code $itk_component(treeeditor) redo]\
      -accelerator <Ctrl-r>

  $itk_component(editmenu) entryconfigure 0 -state disabled
  $itk_component(editmenu) entryconfigure 1 -state disabled

  $itk_component(editmenu) add separator

  $itk_component(editmenu) add command \
      -label "Mirror" \
      -command [code $itk_component(treeeditor) mirrorAction]

  $itk_component(editmenu) add command \
      -label "Break cycles" \
      -command [code $itk_component(treeeditor) breakcycleAction]
  $itk_component(editmenu) add command \
      -label "Parse again" \
      -command [code $itk_component(treeeditor) retry]

   # viewmenu
   itk_component add viewbutton {
      menubutton $itk_component(menuframe).viewbutton \
		-text "View" \
		-menu $itk_component(menuframe).viewbutton.menu
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   itk_component add viewmenu {
      menu $itk_component(viewbutton).menu 
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   itk_component add zoommenu {
      menu $itk_component(viewmenu).zoommenu 
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

   $itk_component(viewmenu) add command \
      -label "Redraw" \
      -command [code $itk_component(treeeditor) refresh]
   $itk_component(viewmenu) add command \
      -label "Eval" \
      -command [code $itk_component(treeeditor) reeval]
   $itk_component(viewmenu) add command \
      -label "Zoom in"\
      -command [code $itk_component(treeeditor) zoomIn] \
      -accelerator <KP_Add> 
   $itk_component(viewmenu) add command \
      -label "Zoom out"\
      -command [code $itk_component(treeeditor) zoomOut] \
      -accelerator <KP_Subtract>

   $itk_component(viewmenu) add separator
  
   $itk_component(viewmenu) add command \
      -label "Show cycles" \
      -command [code $itk_component(treeeditor) showcycleAction] \
      -accelerator <c>
   $itk_component(viewmenu) add command \
      -label "Verify" \
      -command [code $itk_component(treeeditor) verifyAction] \
      -accelerator <v>

   $itk_component(viewmenu) add separator
   $itk_component(viewmenu) add command \
      -label "Next" \
      -command [code $itk_component(treeeditor) nextAction] \
      -accelerator <n>
   $itk_component(viewmenu) add command \
      -label "Previous" \
      -command [code $itk_component(treeeditor) previousAction] \
      -accelerator <p>

   # settings menu
   itk_component add settingsbutton {
      menubutton $itk_component(menuframe).settingsbutton \
		-text "Settings" \
		-menu $itk_component(menuframe).settingsbutton.menu
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }
   itk_component add settingsmenu {
      menu $itk_component(settingsbutton).menu 
   } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
   }

  $itk_component(settingsmenu) add checkbutton \
      -label "Auto-redraw" \
      -variable [scope _redrawFlag] \
      -command [code $this _doRedraw]

  $itk_component(settingsmenu) add checkbutton \
      -label "Auto-eval" \
      -variable [scope _evalFlag] \
      -command [code $this _doEval]
  
  $itk_component(settingsmenu) add checkbutton \
      -label "Paint tree" \
      -variable [scope _paintFlag] \
      -command [code $this _doPaint]

  $itk_component(settingsmenu) add separator

  set _currentOrient [$itk_component(treeeditor) cget -currentOrient]
  $itk_component(settingsmenu) add radiobutton \
    -label "Horizontal" \
    -variable [scope _currentOrient] \
    -value "horizontal" \
    -command [code $this _setOrientation "horizontal"]
  $itk_component(settingsmenu) add radiobutton \
    -label "Vertical" \
    -variable [scope _currentOrient] \
    -value "vertical" \
    -command [code $this _setOrientation "vertical"]

  $itk_component(settingsmenu) add separator

  $itk_component(settingsmenu) add radiobutton -label "Snap to nearest node" \
   -variable [scope _snapmode] -value "nearest"

  $itk_component(settingsmenu) add radiobutton -label "Snap to previous node" \
    -variable [scope _snapmode] -value "previous"

  $itk_component(settingsmenu) add radiobutton -label "Snap to best node" \
    -variable [scope _snapmode] -value "best"



  # packing
  pack $itk_component(menuframe) -fill x -side top
  pack $itk_component(treeeditor) -fill both -side top -expand 1

  pack $itk_component(filebutton) \
        $itk_component(editbutton) \
       	$itk_component(viewbutton) \
       	$itk_component(settingsbutton) \
       	-side left -padx 5 -pady 5
  
  eval itk_initialize $args

  set _paintFlag $itk_option(-painttree)
  set _redrawFlag $itk_option(-autoredraw)
  set _evalFlag $itk_option(-autoeval)

  setGeometry [wm geometry .]
  wm deiconify $itk_interior

}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body VisParses::_setOrientation {orien} {
  $itk_component(treeeditor) setOrientation $orien
  set _currentOrient [$itk_component(treeeditor) getOrientation]
}

## ----------------------------------------------------------------------------
## delegates to TreeEditor::addParse
## and also adjusts the title option
## ----------------------------------------------------------------------------
body VisParses::addParse {parse} {
  set method [$parse getSearchStrategy]
  set lattice [$parse getLatticeId]
  set newid "$method:$lattice"
  configure -title $newid
  return [$itk_component(treeeditor) addParse $parse]
}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body VisParses::removeParse {parse} {
	$itk_component(treeeditor) removeParse $parse
}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body VisParses::view {no} {
	$itk_component(treeeditor) view $no
}

## ----------------------------------------------------------------------------
## Return how many pages are present
## ----------------------------------------------------------------------------
body VisParses::getSize {} {
  return [$itk_component(treeeditor) getSize]
}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body VisParses::center {} {
	$itk_component(treeeditor) center
}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body VisParses::readWord {} {
	return [$itk_component(treeeditor) readWord]
}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body VisParses::getBadness {} {
	$itk_component(treeeditor) getBadness
}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body VisParses::update {} {
	$itk_component(treeeditor) update
}

## ----------------------------------------------------------------------------
## return my geometry
## ----------------------------------------------------------------------------
body VisParses::getGeometry {} {
  return [wm geometry $itk_interior]
}

## ----------------------------------------------------------------------------
## set my geometry
## ----------------------------------------------------------------------------
body VisParses::setGeometry {geometry} {
  wm geometry $itk_interior $geometry
}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body VisParses::setTitle { title } {
	configure -title $title
}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body VisParses::getTitle {} {
	return $itk_option(-title)
}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body VisParses::setUndo {state} {
  if { $state == 1 } {
    $itk_component(editmenu) entryconfigure 0 -state normal
  } else {
    $itk_component(editmenu) entryconfigure 0 -state disabled
  }
}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body VisParses::setRedo { state } {
  if { $state == 1 } {
    $itk_component(editmenu) entryconfigure 1 -state normal
  } else {
    $itk_component(editmenu) entryconfigure 1 -state disabled
  }
}

## ----------------------------------------------------------------------
## returns the current snap mode
## ----------------------------------------------------------------------
body VisParses::getSnapmode {} {
  return $_snapmode
}


## ----------------------------------------------------------------------
## Notify current parsetree of redraw mode change.
## ----------------------------------------------------------------------
body VisParses::_doRedraw {} {
    ## Check if the graphical button is correct
    $itk_component(treeeditor) chkRedrawGfx
  if {1 == $_redrawFlag} {
    $itk_component(treeeditor) redraw
  }
}

## ----------------------------------------------------------------------
## Notify current parsetree of eval mode change.
## ----------------------------------------------------------------------
body VisParses::_doEval {} {
  if {1 == $_evalFlag} {
    $itk_component(treeeditor) reeval
  }
}

## ----------------------------------------------------------------------
## Notify all parsetrees of paint mode change.
## ----------------------------------------------------------------------
body VisParses::_doPaint {} {
  $itk_component(treeeditor) paint
}
 
## ----------------------------------------------------------------------
## Return $_paintFlag
## ----------------------------------------------------------------------
body VisParses::getPaintMode {} {
  return $_paintFlag
}
 
## ----------------------------------------------------------------------
## Return $_redrawFlag
## ----------------------------------------------------------------------
body VisParses::getRedraw {} {
  return $_redrawFlag
}

## ----------------------------------------------------------------------
## Return $_redrawFlag
## ----------------------------------------------------------------------
body VisParses::getEval {} {
  return $_evalFlag
}
 
## ----------------------------------------------------------------------
## Set $_paintFlag
## ----------------------------------------------------------------------
body VisParses::setPaintMode {x} {
  set _paintFlag $x
  _doPaint
}
 
## ----------------------------------------------------------------------
## Set $_redrawFlag
## ----------------------------------------------------------------------
body VisParses::setRedraw {x} {
  set _redrawFlag $x
  _doRedraw
}
 
## ----------------------------------------------------------------------
## Set $_evalFlag
## ----------------------------------------------------------------------
body VisParses::setEval {x} {
  set _evalFlag $x
  _doEval
}
 
