# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaConfig - the configuration document
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## $Id: YadaConfig.tcl,v 1.18 2005-12-01 16:23:26 micha Exp $
## ----------------------------------------------------------------------------
class YadaConfig {
  inherit YadaDocument


  # variables ----------------------------------------------------------------
  private variable _currentDocumentName ""
  
  # methods ------------------------------------------------------------------
  public method init {}
  public method load {}
  public method save {}
  public method activationHandle {}
  public method deActivationHandle {}
  public method document {docName args}; ## \type TclString, TclList

  constructor {args} {}; ## \type TclList

  private method _activateDocument {docName}; ## \type TclString
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaConfig::constructor {args} {

  #
  # the big tabno
  #
  itk_component add tabno {
    iwidgets::tabnotebook $itk_component(childsite).tabno \
      -borderwidth 2 \
      -tabpos e \
      -backdrop gray90 \
      -angle 10 \
      -bevelamount 4 \
      -raiseselect 1 \
  } {}

  #
  # experiments
  #
  set experimentPage [$itk_component(tabno) add \
      -label "Experiments"]

  itk_component add experiments {
    YadaExperiments $experimentPage.experiments -name "experiments"
  } {}

  $itk_component(tabno) pageconfigure "Experiments" \
    -command [code $this _activateDocument "experiments"]

  #
  # the grammars 
  #
  set grammarPage [$itk_component(tabno) add \
      -label "Grammars"]

  itk_component add grammars {
    YadaGrammars $grammarPage.grammars -name "grammars"
  } {}

  $itk_component(tabno) pageconfigure "Grammars" \
    -command [code $this _activateDocument "grammars"]


  #
  # the machines
  #
  set machinesPage [$itk_component(tabno) add \
      -label "Machines"]

  itk_component add machines {
    YadaMachines $machinesPage.machines -name "machines"
  } {}

  $itk_component(tabno) pageconfigure "Machines" \
    -command [code $this _activateDocument "machines"]


  #
  # misc preferences
  #
  set prefsPage [$itk_component(tabno) add \
	-label "Preferences"]

  itk_component add prefs {
    YadaPreferences $prefsPage.prefs
  } {}

  $itk_component(tabno) pageconfigure "Preferences" \
    -command [code $this _activateDocument "prefs"]

  # bindings

  # packing + gridding
  pack $itk_component(tabno) -side top -expand 1 -fill both -padx 2 -pady 5
  pack $itk_component(experiments) -expand 1 -fill both
  pack $itk_component(grammars) -expand 1 -fill both
  pack $itk_component(machines) -expand 1 -fill both
  pack $itk_component(prefs) -expand 1 -fill both
  

  eval itk_initialize $args
  .main registerDocument $this
}


## ----------------------------------------------------------------------------
## load
## ----------------------------------------------------------------------------
body YadaConfig::load {} {
  global env

  ### TODO UUUUUUUUUUUUAAAAAAAAAAAAAAAAA

  if { [$itk_component(tabno) view] == 0} {
    if {[catch {$itk_component(experiments) load} errorMessage] } {
      printMessage "Cannot load: $errorMessage"
    }
  } elseif { [$itk_component(tabno) view] == 1} {
    if {[catch {$itk_component(grammars) load} errorMessage] } {
      printMessage "Cannot load: $errorMessage"
    } else {
      $itk_component(grammars) show
    }
  } elseif { [$itk_component(tabno) view] == 2} {
    if {[catch {$itk_component(machines) load} errorMessage] } {
      printMessage "Cannot load: $errorMessage"
    } 
  } elseif { [$itk_component(tabno) view] == 3} {
    if { [catch {$itk_component(prefs) load} errorMessage] } {
      printMessage "Cannot load: $errorMessage"
    } 
  } else {
    printMessage "WARNING: YadaConfig::load under construction"
  }
}

## ----------------------------------------------------------------------------
## save
## ----------------------------------------------------------------------------
body YadaConfig::save {} {
  global env    

  ### TODO UUUUUUUUUUUUAAAAAAAAAAAAAAAAA

  if { [$itk_component(tabno) view] == 0} {
    $itk_component(experiments) save
  } elseif { [$itk_component(tabno) view] == 1} {
    $itk_component(grammars) save
    $itk_component(grammars) show
  } elseif { [$itk_component(tabno) view ] == 2 } {
    $itk_component(machines) save
  } elseif { [$itk_component(tabno) view ] == 3 } {
    $itk_component(prefs) save
  } else {
    printMessage "WARNING: YadaConfig::save under construction"
  }
}

## ----------------------------------------------------------------------------
## initialize the YADA configuration.
## This method initializes the yada config document and all its subdocuments,
## that is all init scripts are sourced. Doing so we generate YadaConfigItem
## objects and collet them all separately in the appropriate subdocuments.
## So the following init scripts in YADA_ETC are sourced in the given order:
## - \c configure.tcl: this is an application level init script 
## - \c grammars.tcl: generate all YadaGrammar objects and assert them into YadaGrammars
## - \c machines.tcl: generate all YadaMachine objects and assert them into YadaMachines
## - \c experiments.tcl: generate all YadaExperiment objects and assert them into YadaExperiments
## Hint: the init scripts might source further init scripts. This lets you
## organize different init scripts for different purposes.
## ----------------------------------------------------------------------------
body YadaConfig::init {} {
  if {$_isInitialized} {
    return
  }

  chain

  # load data from rcFiles
  global env
  set rcFileName [file join $env(YADA_ETC) "configure.tcl"]

  if {[file exists $rcFileName]} {
    # load resource file
    if {[catch {source $rcFileName}]} {
      bgerror "There's an error in your init file!"
    }
  } else {
    # set some hardcoded default experiments
    # puts "INFO: configure file `$rcFileName' not found, using defaults"
  }

  # init subddocs, oder matters
  $itk_component(prefs) init
  $itk_component(grammars) init
  $itk_component(machines) init
  $itk_component(experiments) init

  # default selection
  _activateDocument "experiments"
  $itk_component(tabno) view "Experiments"
}

## ----------------------------------------------------------------------------
## activationHandle
## ----------------------------------------------------------------------------
body YadaConfig::activationHandle {} {
  chain
  $itk_component(fileMenu) entryconfigure "Close" -state disabled
  _activateDocument $_currentDocumentName
}

## ----------------------------------------------------------------------------
## _activateDocument
## ----------------------------------------------------------------------------
body YadaConfig::_activateDocument {docName} {

  # call the deactivation handle
  if {$_currentDocumentName != "" &&  $_currentDocumentName != $docName} {
    $itk_component($_currentDocumentName) deActivationHandle
  }

  # keep track of the active document
  set _currentDocumentName $docName

  # remove the edit menu
  set editButton [.main component editButton]
  $editButton configure -menu ""

  # call the activation handle
  $itk_component($docName) activationHandle

}

## ----------------------------------------------------------------------------
## deActivationHandle
## ----------------------------------------------------------------------------
body YadaConfig::deActivationHandle {} {
  chain
  $itk_component(fileMenu) entryconfigure "Close" -state active
  if {$_currentDocumentName != ""} {
    $itk_component($_currentDocumentName) deActivationHandle
  }
}

## ----------------------------------------------------------------------------
## document
## ----------------------------------------------------------------------------
body YadaConfig::document {docName args} {
  eval $itk_component($docName) $args
}
