# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaConfigDocument - abstract base class for all configure documents.
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## $Id: YadaConfigDocument.tcl,v 1.14 2005-12-01 16:23:26 micha Exp $
## ----------------------------------------------------------------------------
class YadaConfigDocument {
  inherit itk::Widget YadaPlainDocument

  # variables ----------------------------------------------------------------

  ## the known aspects of a YadaConfigItem
  protected variable _allAspects ""; ## \type TclList

  ## the currently selected item
  private variable _currentItem ""; ## \type YadaConfigItem

  ## position within the _allItems array
  private variable _currentItemNo 0; ## \type TclNumber

  ## array to store all items managed by this document
  private variable _allItems; ## \type YadaConfigItem[]

  ## info string displaying the current position within all items
  private variable _counter "0/0"; ## \type TclString

  # methods ------------------------------------------------------------------
  public method init {}
  public method activationHandle {}
  public method getItem {name}; ## \type TclString
  public method getItemNo {name}; ## \type TclString
  public method getAllItems {args}; ## \type TclList
  public method getAllItemNames {args}; ## \type TclList
  public method addItem {item}; ## \type YadaConfigItem
  public method removeItem {item}; ## \type YadaConfigItem
  public method setCurrentItem {item}; ## \type YadaConfigItem
  public method getCurrentItem {}
  public method getWindowTitle {}
  public method updateMenus {} ; ## \virtual
  public method updateCombos {}; ## \virtual
  public method load {}
  public method save {}
  public method saveAll {}

  constructor {args} {}; ## \type TclList
  destructor {}
  

  private method _undo {}
  private method _selectCommand {}
  private method _next {}
  private method _prev {}
  private method _setCounter {}
  

  protected method _askChange {}
  protected method _mark {aspect {newData "undef"} {oldData "undef"}}; ## \type TclString, TclString, TclString
  protected method _unmark {{aspect ""}}; ## \type TclString
  protected method _getMarked {}
  protected method _commit {} 

  protected method _select {{itemName ""}} ;## \purevirtual \type TclString
  protected method _selectFirst {}
  protected method _new {} ; ## \purevirtual
  protected method _delete {} ;## \purevirtual
  protected method _defaults {} ;## \purevirtual
  protected method _validateCommand {aspect newValue {oldValue ""}}; ## \type TclString, TclString, TclString
  protected method _isValidItem {item} ; ## \virtual \type YadaConfigItem

  protected method isValidPersistenceFile {fileName}; ## \purevirtual \type TclString

};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaConfigDocument::constructor {args} {
  #
  # create custom file menu
  #
  set fileButton [.main component fileButton]
  for {
    set i 1
    set fileMenu $fileButton.menu$i
  } {[winfo exists $fileMenu]} {incr i} {
    set fileMenu $fileButton.menu$i
  }
  itk_component add fileMenu {
      menu $fileMenu
  } {
      keep -background -cursor
      rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  ## create submenu entry "New"
  itk_component add newMenu {
     menu $itk_component(fileMenu).newMenu 
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  ## add file menu entries
  $itk_component(fileMenu) add command \
       	-label "Load" \
       	-underline 0 \
       	-command [code $this load] 
  $itk_component(fileMenu) add command \
       	-label "Save" \
       	-underline 0 \
       	-command [code $this save] 
  $itk_component(fileMenu) add command \
       	-label "Save All" \
       	-underline 0 \
       	-command [code $this saveAll] 
  $itk_component(fileMenu) add cascade \
	-label "New" \
	-underline 0 \
	-menu $itk_component(newMenu)
  $itk_component(fileMenu) add separator
  $itk_component(fileMenu) add command \
	-label "Quit" \
	-underline 0 \
	-command {.main quit}


  $itk_component(newMenu) add command \
	-label "Runner" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaRunner Runner}
  $itk_component(newMenu) add command \
	-label "Difference" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaDifference Difference}
  $itk_component(newMenu) add command \
	-label "Gls Statistics" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaGlsStats GlsStatistics}
  $itk_component(newMenu) add command \
	-label "Ranking" \
	-underline 0 \
	-command {YadaDocument::newDocument YadaRanking Ranking}
 

  # edit menu
  set editButton [.main component editButton]
  for {
    set i 1
    set editMenu [subst $editButton.menu$i]
  } {[winfo exists $editMenu]} {incr i} {
    set editMenu [subst $editButton.menu$i]
  }
  
  itk_component add editMenu {
     menu $editMenu
  } {
     keep -background -cursor
     rename -borderwidth -menuborderwidth menuBorderwidth Borderwidth
  }

  $itk_component(editMenu) add command \
       	-label "Undo" \
       	-underline 0 \
       	-command [code $this _undo]
#  $itk_component(editMenu) add command \
       	-label "Commit" \
       	-underline 0 \
       	-command [code $this _commit]
  $itk_component(editMenu) add command \
       	-label "New" \
       	-underline 0 \
       	-command [code $this _new]
  $itk_component(editMenu) add command \
       	-label "Delete" \
       	-underline 0 \
       	-command [code $this _delete]
  $itk_component(editMenu) add separator
  $itk_component(editMenu) add command \
       	-label "Find" \
       	-underline 0 \
       	-command [code $this _find] \
	-state disabled

  # buttons
  itk_component add buttonFrame {
    frame $itk_interior.buttonFrame
  }
  itk_component add topButtonFrame {
    frame $itk_interior.buttonFrame.top
  }
  itk_component add botButtonFrame {
    frame $itk_interior.buttonFrame.bot
  }
  itk_component add commitButton {
    button $itk_component(topButtonFrame).commitButton \
	-text "Ok" \
	-command [code $this _commit]
  }
  itk_component add newButton {
    button $itk_component(topButtonFrame).newButton \
	-text "New" \
	-command [code $this _new]
  }
  itk_component add deleteButton {
    button $itk_component(topButtonFrame).deleteButton \
	-text "Delete" \
	-command [code $this _delete]
  }
  itk_component add undoButton {
    button $itk_component(topButtonFrame).undoButton \
      -text "Cancel" \
      -command [code $this _undo]
  }
  itk_component add naviFrame {
    frame $itk_component(botButtonFrame).naviFrame
  }
  itk_component add nextButton {
    button $itk_component(naviFrame).nextButton \
      -text ">" \
      -command [code $this _next]
  }
  itk_component add prevButton {
    button $itk_component(naviFrame).prevButton \
      -text "<" \
      -command [code $this _prev]
  }

  # counter
  itk_component add counter {
    label $itk_interior.counter \
      -textvariable [scope _counter] \
      -anchor se  \
      -font {-adobe-helvetica-medium-r-normal--10-*-*-*-*-*-*-*}
  } { }

  # childsite
  itk_component add childsite {
    frame $itk_interior.childsite 
  } 

  # name
  itk_component add nameLabel {
    label $itk_component(childsite).nameLabel \
      -text "Name:" \
      -anchor nw
  } {}

  itk_component add nameMark {
    frame $itk_component(childsite).nameMark \
      -borderwidth 2
  } {}

  itk_component add name {
    iwidgets::combobox $itk_component(nameMark).name \
      -completion false \
      -borderwidth 2 \
      -unique true \
      -selectioncommand [code $this _selectCommand]
  }  {}
  set entry [$itk_component(name) component entry]
  $entry configure \
    -validate all \
    -validatecommand [code $this _validateCommand "name" %P]


  #
  # packing + gridding
  #
  grid columnconfigure $itk_interior 0 -weight 1
  grid rowconfigure $itk_interior 0 -weight 1
  grid columnconfigure $itk_component(childsite) 1 -weight 1
  grid $itk_component(childsite) -sticky  news -padx 10 -pady 10 -row 0 -column 0 -rowspan 2
  grid $itk_component(buttonFrame) -sticky ns -padx 10 -row 0 -column 2 -pady 10
  grid $itk_component(counter) -sticky se -padx 0 -row 1 -column 2 -pady 5 -padx 5

  grid $itk_component(nameLabel) -sticky ew -row 0 -column 0 
  grid $itk_component(nameMark) -sticky ew -row 0 -column 1 -pady 10 -columnspan 2
  pack $itk_component(name) -fill both -expand 1

  pack $itk_component(topButtonFrame) -side top -fill both -expand 1
  pack $itk_component(botButtonFrame) -side top -fill both -expand 1
  pack $itk_component(naviFrame) -side bottom -fill x
  pack $itk_component(commitButton) -side top -fill x -pady 5
  pack $itk_component(undoButton) -side top -fill x -pady 5
  pack $itk_component(newButton) \
       $itk_component(deleteButton) -side top -fill x -pady 5
  pack $itk_component(nextButton) \
       $itk_component(prevButton) -side right -fill x -pady 0


  # bindings
  set entry [$itk_component(name) component entry]
  bind $entry <Up> +[code $this _selectCommand]
  bind $entry <Down> +[code $this _selectCommand]

  eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body YadaConfigDocument::destructor {} {
  foreach item [array names _allItems] {
    itcl::delete $_allItems($item)
  }
}

## ----------------------------------------------------------------------------
## init
## ----------------------------------------------------------------------------
body YadaConfigDocument::init {} {
  if {$_isInitialized} {
    return
  }

  chain
  updateCombos
}

## ----------------------------------------------------------------------------
## _selectCommand
## ----------------------------------------------------------------------------
body YadaConfigDocument::_selectCommand {} {
  _unmark
  after idle [code $this _select]
}

## ----------------------------------------------------------------------------
## select the first item that is not named "<none>". If there are no items
## at all the "<none>" item is selected
## ----------------------------------------------------------------------------
body YadaConfigDocument::_selectFirst {} {
  set noItems [$itk_component(name) size]
  for {set i 0} {$i < $noItems} {incr i} {
    set item [$itk_component(name) get $i]
    if {$item != "<none>"} {
      _select $item
      return $item
    }
  }

  _select "<none>"
  return "<none>"
}

## ----------------------------------------------------------------------------
## _askChange
## ----------------------------------------------------------------------------
body YadaConfigDocument::_askChange {} {
  set result 1

  set markedAspects [_getMarked]
  if {$markedAspects != ""} {
    set msg [iwidgets::messagedialog .msg \
      -text "Are you sure that you want to discard your changes?" \
      -title "Yada - Caution"  \
      -image [YadaImages::get question] \
      -modality application \
      -master $itk_interior]
    $msg buttonconfigure OK -text "Yes"
    $msg buttonconfigure Cancel -text "No"
    $msg center .
    set result [$msg activate]
    destroy $msg
  }

  return $result
}

## ----------------------------------------------------------------------------
## activationHandle
## ----------------------------------------------------------------------------
body YadaConfigDocument::activationHandle {} {
  chain

  ## activate custom file menu
  set fileButton [.main component fileButton]
  $fileButton configure -menu $itk_component(fileMenu)

  # set the menus
  set editButton [.main component editButton]
  $editButton configure -menu $itk_component(editMenu)

  # update title and menus
  updateWindowTitle
  updateMenus
}

## ----------------------------------------------------------------------------
## _unmark
## ----------------------------------------------------------------------------
body YadaConfigDocument::_unmark {{aspect ""}} {
  if {$aspect == ""} {
    foreach aspect $_allAspects {
      $itk_component(${aspect}Mark) configure -background gray90
    }
  } else {
    $itk_component(${aspect}Mark) configure -background gray90
  }
}

## ----------------------------------------------------------------------------
## _mark
## ----------------------------------------------------------------------------
body YadaConfigDocument::_mark {aspect {newData "undef"} {oldData "undef"}} {

  set itemName [$itk_component(name) get]

  if {$itemName == "<none>" || $itemName == ""} {
    return
  }

  if {$newData == "undef"} {
    set newData [$itk_component($aspect) get]
  }
  if {$oldData == "undef"} {
    set oldData [string trim [$_currentItem cget -$aspect]]
  }

  set mark $itk_component(${aspect}Mark)

  if {$newData != $oldData} {
    $mark configure -background red4
  } else {
    $mark configure -background gray90
  }
}

## ----------------------------------------------------------------------------
## _getMarked
## ----------------------------------------------------------------------------
body YadaConfigDocument::_getMarked {} {
  set result ""
  foreach aspect $_allAspects {
    if {[$itk_component(${aspect}Mark) cget -background] == "red4"} {
      lappend result $aspect
    }
  }

  return $result
}

## ----------------------------------------------------------------------------
## _undo
## ----------------------------------------------------------------------------
body YadaConfigDocument::_undo {} {
  if {$_currentItem != ""} {
    _select [$_currentItem cget -name]
  }
}

## ----------------------------------------------------------------------------
## _validateCommand
## ----------------------------------------------------------------------------
body YadaConfigDocument::_validateCommand {aspect newValue {oldValue ""}} {

  if {$_currentItem == ""} {
    return 0
  }

  if {$oldValue == ""} {
    if {$newValue == ""} {
      return 1
    }
    set oldValue [string trim [$_currentItem cget -$aspect]]
  }

  set mark $itk_component(${aspect}Mark)
  if {$newValue != $oldValue} {
    $mark configure -background red4
  } else {
    $mark configure -background gray90
  }

  return 1
}

## ----------------------------------------------------------------------------
## _next
## ----------------------------------------------------------------------------
body YadaConfigDocument::_next {} {
  set itemNames [getAllItemNames]
  set index $_currentItemNo
  incr index
  set itemName [lindex $itemNames $index]
  if {$itemName != ""} {
    _select $itemName
  }
}

## ----------------------------------------------------------------------------
## _prev
## ----------------------------------------------------------------------------
body YadaConfigDocument::_prev {} {
  set itemNames [getAllItemNames]
  set index $_currentItemNo
  incr index -1
  set itemName [lindex $itemNames $index]
  if {$itemName != ""} {
    _select $itemName
  }
}

## ----------------------------------------------------------------------------
## getCurrentItem
## \returns the currently displayed YadaConfigItem
## ----------------------------------------------------------------------------
body YadaConfigDocument::getCurrentItem {} {
  return $_currentItem
}

## ----------------------------------------------------------------------------
## update the counter label.
## This method sets the _counter to reflect the number of items in this document.
## ----------------------------------------------------------------------------
body YadaConfigDocument::_setCounter {} {
  set noItems [llength [array names _allItems]]
  if {$noItems == 0} {
    set _counter "0/0"
  } else {
    set _counter "$_currentItemNo/[expr $noItems -1]"
  }
}

## ----------------------------------------------------------------------------
## setCurrentItem
## ----------------------------------------------------------------------------
body YadaConfigDocument::setCurrentItem {item} {
  set itemName ""
  if {$item == ""} {
    set itemName [lindex [getAllItemNames] 0]
    set _currentItem [getItem $itemName]
  } else {
    set itemName [$item cget -name]
    if {![info exists _allItems($itemName)]} {
      if {![addItem $item]} {
	set $item ""
      }
    }
  }

  set _currentItem $item

  if {$_currentItem == ""} {
    set _currentItemNo "0"
  } else {
    set _currentItemNo [getItemNo $itemName]
  }
  _setCounter
}

## ----------------------------------------------------------------------------
## unregister a YadaConfigItem from this config document. the item object 
## itself is not deconstructed..
## @param item the YadaConfigItem object to be unregistered
## ----------------------------------------------------------------------------
body YadaConfigDocument::removeItem {item} {
  set itemName [$item cget -name]

  if {![info exists _allItems($itemName)]} {
    return
  }

  if {$_currentItem == $item} {
    setCurrentItem ""
  }

  unset _allItems($itemName)
  _setCounter
}

## ----------------------------------------------------------------------------
## register a YadaConfigItem to this config document. use removeItem() to 
## unregister it again. The item will only be added if it qualifies to be
## valid.
## @param item the YadaConfigItem
## @returns 1 if the item has been registered successfully, and 0 if the item
## is not valid inside this YadaConfigDocument
## ----------------------------------------------------------------------------
body YadaConfigDocument::addItem {item} {
  if {$item == ""} {
    return 0
  }

  set itemName [$item cget -name]
  if {![_isValidItem $item]} {
    return 0
  }

  set _allItems($itemName) $item
  updateCombos
  _setCounter

  return 1
}

## ----------------------------------------------------------------------------
## getItemNo
## ----------------------------------------------------------------------------
body YadaConfigDocument::getItemNo {itemName} {
  set allNames [getAllItemNames]
  return [lsearch $allNames $itemName]
}

## ----------------------------------------------------------------------------
## getItem
## ----------------------------------------------------------------------------
body YadaConfigDocument::getItem {itemName} {
  if {[info exists _allItems($itemName)]} {
    return $_allItems($itemName)
  } else {
    return ""
  }
}

## ----------------------------------------------------------------------------
## getAllItems
## \returns a list of all known YadaConfigItem objects in this document
## ----------------------------------------------------------------------------
body YadaConfigDocument::getAllItems {args} {
  set result ""

  foreach itemName [eval getAllItemNames $args] {
    lappend result $_allItems($itemName)
  }

  return $result
}

## ----------------------------------------------------------------------------
## getAllItemNames
## @returns a list of all known names of YadaConfigItem objects in this document
## ----------------------------------------------------------------------------
body YadaConfigDocument::getAllItemNames {args} {
  set result ""

  if {[llength $args] == 0} {
    return [lsort -dictionary [array names _allItems]]
  } 

  foreach itemName [lsort -dictionary [array names _allItems]] {
    foreach pattern $args {
      if {[string match $pattern $itemName]} {
	lappend result $itemName
	break
      }
    }
  }

  return $result
}

## ----------------------------------------------------------------------------
## _commit
## ----------------------------------------------------------------------------
body YadaConfigDocument::_commit {} {

  # check new item name
  set newName [$itk_component(name) get]
  set oldName [$_currentItem cget -name]
  
  if {$newName != $oldName} {

    if {[info exists _allItems($newName)]} {
      printMessage "ERROR: '$newName' exists already." error
      return 0
    }

    # relocate item in array
    unset _allItems($oldName)
    set _allItems($newName) $_currentItem
    $_currentItem configure -name $newName
    _setCounter
  }

  updateCombos

  return 1
}

## ----------------------------------------------------------------------------
## @returns a title string representing the state of this document and the
## current config item
## ----------------------------------------------------------------------------
body YadaConfigDocument::getWindowTitle {} {
  set currentItem [getCurrentItem]
  if {$currentItem == ""} {
    return "Configure.[chain]"
  }
  set currentItemName [$currentItem cget -name]
  if {$currentItemName == "" || $currentItemName == "<none>"} {
    return "Configure.[chain]"
  }

  return "Configure.[chain]: [$currentItem getWindowTitle]"
}

## ----------------------------------------------------------------------------
## update the state of menu and button actions and toggle them to
## @c normal or @c disabled if the action is valid or not respectively
## ----------------------------------------------------------------------------
body YadaConfigDocument::updateMenus {} {

  set currentItem [getCurrentItem]
  if {$currentItem == "" } {
    return
  }
 
  # set the state of the "Load" action
  set loadButton [[[.main getDocument Configure] component toolbar] component loadButton]
  if {[$currentItem hasSavedState] } {
    $itk_component(fileMenu) entryconfigure 0 -state normal
    $loadButton configure -state normal
  } else {
    $itk_component(fileMenu) entryconfigure 0 -state disabled
    $loadButton configure -state disabled
  }

  # set the state of the "Save" action
  set saveButton [[[.main getDocument Configure] component toolbar] component saveButton]
  if {[$currentItem cget -isModified] && [string compare [$currentItem cget -name] "<none>"] != 0 } {
    $itk_component(fileMenu) entryconfigure 1 -state normal   
    $saveButton configure -state normal
  } else {
    $itk_component(fileMenu) entryconfigure 1 -state disabled
    $saveButton configure -state disabled
  }

  # set the state of the "Save All" action
  set state "disabled"
  foreach item [getAllItems] {
    if {[$item cget -name] == "<none>"} {
      continue
    }
    if {[$item cget -isModified]} {
      set state "normal"
      break
    }
  }
  $itk_component(fileMenu) entryconfigure 2 -state $state   
  
}

## ----------------------------------------------------------------------------
## update lists of all known comboboxes, i.e. the list of item names 
## ----------------------------------------------------------------------------
body YadaConfigDocument::updateCombos {} {
  set listBox [$itk_component(name) component list]
  set knownNames [array names _allItems]
  $listBox delete 0 end
  eval $listBox insert 0 [lsort -dictionary $knownNames]
}

## ----------------------------------------------------------------------------
## loads the currently selected item
## ----------------------------------------------------------------------------
body YadaConfigDocument::load {} {
  set currentItem [getCurrentItem]

  if {$currentItem == "" || [$currentItem cget -name] == "<none>"} {
    return
  }

  set oldCursors [setBusy]
  if { [catch {$currentItem load} errorMessage] } {
    resetBusy $oldCursors
    error $errorMessage
  }
  resetBusy $oldCursors
  updateWindowTitle
  updateMenus
}

## ----------------------------------------------------------------------------
## saves the currently selected item
## ----------------------------------------------------------------------------
body YadaConfigDocument::save {} {
  set currentItem [getCurrentItem]
  if {$currentItem == "" || $currentItem == "<none>"} {
    return
  }
  
  set currentItemName [$currentItem cget -name]
  if {$currentItemName == "<none>"} {
    return
  }
  
  set oldCursors [setBusy]
  $currentItem save
  resetBusy $oldCursors
  updateWindowTitle
  updateMenus
}

## ----------------------------------------------------------------------------
## saves all itemes
## ----------------------------------------------------------------------------
body YadaConfigDocument::saveAll {} {
  set oldCursors [setBusy]
  foreach item [getAllItems] {
    if {[$item cget -name] != "<none>"} {
      $item save
    }
  }
  resetBusy $oldCursors
  updateWindowTitle
  updateMenus
}

## ----------------------------------------------------------------------------
## test this item to qualify valid inside this YadaConfigDocument. We check for
## uniqueness here only, that is all items must have a unique name
## @returns 1 on success, 0 otherwise
## ----------------------------------------------------------------------------
body YadaConfigDocument::_isValidItem {item} {

  set itemName [$item cget -name]
  set otherItem [getItem $itemName]
  if {$otherItem != ""} {
    if {$item != $otherItem} {
      printMessage "`$itemName' already exists!" error
      return 0
    }
  } 

  return 1
}
