# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaGrammar - bundled information about a CDG grammar
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## $Id: YadaGrammar.tcl,v 1.31 2005-12-01 16:23:27 micha Exp $
## ----------------------------------------------------------------------------
class YadaGrammar {
  inherit YadaConfigItem

  # variables ----------------------------------------------------------------
  
  ## list of all files making up the grammar
  public variable files {}; ## \type TclList

  ## array mapping wordgraph ids to YadaWordgraph objects
  private variable _wordgraphs; ## \type YadaWordgraph[]

  ## array mapping annotation ids to YadaAnnotation objects
  private variable _annotations; ## \type YadaAnnotation[]

  ## flag indicating whether the grammar has been scanned for wordgraphs
  private variable _isScanned 0; ## \type TclBoolean

  # methods ------------------------------------------------------------------

  constructor {args} {}; ## \type TclList
  destructor {}

  public method deleteWordgraphs {}
  public method deleteAnnotations {}
  public method getWordgraphs {}
  public method getAnnotations {}
  public method getWordgraph {wgName}; ## \type TclString
  public method getAnnotation {annoName}; ## \type TclString
  public method addWordgraph {wordgraph}; ## \type YadaWordgraph
  public method addAnnotation {annotation}; ## \type YadaAnnotation
  public method setSelection {args}; ## \type TclList
  public method getSelection {}
  public method isSelected {wgName}; ## \type TclString
  public method scan {}
  public method rescan {}
  public method clone {args}; ## \type TclList
  public method toDom {}
  public method fromDom {document}; ## \type domDoc
  public method fromDomNode {grammarNode}; ## \type domNode
  public method isScanned {}
  public method getFileName {}; ## overriding virtual method
 
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaGrammar::constructor {args} {

  eval configure $args
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body YadaGrammar::destructor {} {
  deleteWordgraphs
  deleteAnnotations
}

## ----------------------------------------------------------------------------
## serializes this object into a dom document
## @returns a new dom document
## ----------------------------------------------------------------------------
body YadaGrammar::toDom {} {
  set document [dom createDocument "yada"]
  set root [$document documentElement]

  $root appendFromList [list grammar "name $name isScanned $_isScanned" {}]
  set grammarNode [$root firstChild]

  foreach file $files {
    $grammarNode appendFromList [list file "name $file" {}]
  }

  # store the wordgraphs
  foreach wordgraph [getWordgraphs] {
    set wordgraphDocument [$wordgraph toDom]
    $grammarNode appendChild [$wordgraphDocument documentElement]
    $wordgraphDocument delete
  }

  # store the annotations
  foreach annotation [getAnnotations] {
    set annoDocument [$annotation toDom]
    $grammarNode appendChild [$annoDocument documentElement]
    $annoDocument delete
  }
  
  return $document
}

## ----------------------------------------------------------------------------
## initialize a grammar from a dom document
## @param document a dom document containing a /yada/grammar
## ----------------------------------------------------------------------------
body YadaGrammar::fromDom {document} {
  set rootNode [$document documentElement]
  set grammarNode [$rootNode selectNodes "/yada/grammar"]
  fromDomNode $grammarNode
}

## ----------------------------------------------------------------------------
## initialize a grammar from a dom node
## @param grammarNode containing the grammar attributes
## ----------------------------------------------------------------------------
body YadaGrammar::fromDomNode {grammarNode} {
  set files {}
  foreach nodeFile [$grammarNode selectNodes "./file"] {
    lappend files [$nodeFile getAttribute "name"]
  }

  set name [$grammarNode getAttribute "name"] 
  set _isScanned [$grammarNode getAttribute "isScanned"]

  # clean up
  deleteWordgraphs
  deleteAnnotations

  # read all wordgraph specs
  foreach wordgraphNode [$grammarNode selectNodes "./wordgraph"] {
    set newWordgraph [YadaWordgraph ::#auto $this]
    $newWordgraph fromDomNode $wordgraphNode
    addWordgraph $newWordgraph
  }

  # read all annotation specs
  foreach annoNode [$grammarNode selectNodes "./annotation"] {
    set newAnno [YadaAnnotation ::#auto $this]
    $newAnno fromDomNode $annoNode
    addAnnotation $newAnno
  }
}

## ----------------------------------------------------------------------------
## deleteWordgraphs
## ----------------------------------------------------------------------------
body YadaGrammar::deleteWordgraphs {} {
  foreach wgName [array names _wordgraphs] {
    itcl::delete object $_wordgraphs($wgName)
    unset _wordgraphs($wgName)
  }
  set isModified 1
}

## ----------------------------------------------------------------------------
## deleteAnnotations
## ----------------------------------------------------------------------------
body YadaGrammar::deleteAnnotations {} {
  foreach annoName [array names _annotations] {
    itcl::delete object $_annotations($annoName)
    unset _annotations($annoName)
  }
  set isModified 1
}


## ----------------------------------------------------------------------------
## getWordgraphs
## ----------------------------------------------------------------------------
body YadaGrammar::getWordgraphs {} {
  set result {}
  foreach wgName [lsort -dictionary [array names _wordgraphs]] {
    lappend result $_wordgraphs($wgName)
  }
  return $result
}

## ----------------------------------------------------------------------------
## getAnnotations
## ----------------------------------------------------------------------------
body YadaGrammar::getAnnotations {} {
  set result {}
  foreach annoName [lsort -dictionary [array names _annotations]] {
    lappend result $_annotations($annoName)
  }
  return $result
}


## ----------------------------------------------------------------------------
## getWordgraph
## ----------------------------------------------------------------------------
body YadaGrammar::getWordgraph {wgName} {
  if {[info exists _wordgraphs($wgName)]} {
    return $_wordgraphs($wgName)
  } else {
    return ""
  }
}

## ----------------------------------------------------------------------------
## getAnnotation
## ----------------------------------------------------------------------------
body YadaGrammar::getAnnotation {annoName} {
  if {[info exists _annotations($annoName)]} {
    return $_annotations($wgName)
  } else {
    return ""
  }
}


## ----------------------------------------------------------------------------
## addWordgraph
## registers a YadaWordgraph object
## ----------------------------------------------------------------------------
body YadaGrammar::addWordgraph {wordgraph} {
  set wgName [$wordgraph cget -name]
  if {[info exists _wordgraphs($wgName)]} {
    itcl::delete object $_wordgraphs($wgName)
  }
  set _wordgraphs($wgName) $wordgraph
}

## ----------------------------------------------------------------------------
## addAnnotation
## registers a YadaAnnotation object
## ----------------------------------------------------------------------------
body YadaGrammar::addAnnotation {annotation} {
  set annoName [$annotation cget -name]
  if {[info exists _annotations($annoName)]} {
    itcl::delete object $_annotations($annoName)
  }
  set _annotations($annoName) $annotation
  
  # also: register this annotation to its wordgraph
  set wgName [$annotation cget -wordgraphName]
  set yadaWordgraph [getWordgraph $wgName]
  if {$yadaWordgraph != ""} {
    $yadaWordgraph addAnnotation $annotation
  }
}


## ----------------------------------------------------------------------------
## scan
## ----------------------------------------------------------------------------
body YadaGrammar::scan {} {
  if {!$_isScanned} {
    rescan
  }
}

## ----------------------------------------------------------------------------
## rescan
## ----------------------------------------------------------------------------
body YadaGrammar::rescan {} {
  global env

  .main printStatus "... rescanning grammar '$name'"

  set isModified 1

  # remove all contained objects from the grammar
  set oldSelection [getSelection]
  deleteWordgraphs
  deleteAnnotations

  # build a temporary xml log file
  set xmlFile [getSecureFileName [file join $env(YADA_TMP) _yada[pid].xml.gz]]


  # build a cdg process which is to be asked about wordgraphs and annotations
  set machine [.main getMachine "localhost"]

  set process [CdgProcess ::#auto \
    -printCommand "" \
    -xmlFile $xmlFile \
    -initFile "none" \
  ]

  # start a cdgp process
  set thisPid [$process start]
  if {$thisPid <= 0} {
    .main printStatus ""
    printMessage "ERROR: cannot start cdgp process $env(YADA_CDGP)"
    return
  }

  # load the grammar files
  foreach file $files {
    if {[regexp {.*\.gz} $file]} {
      $process tell "set preprocessor /usr/bin/zcat"
      $process tell "set prepsuffix .gz"
    } elseif {[regexp {.*\.m4} $file]} {
      $process tell "set preprocessor /usr/bin/m4 -s -P"
      $process tell "set prepsuffix .m4"
    }
    $process tell "load $file"
  }

  # give us the wordgraphs
  $process tell "wordgraph\nannotation\nquit\n"
  .main printStatus "... started cdgp process"
  $process wait

  #itcl::delete object $process

  # get the document data
  .main printStatus "... opening source file '$xmlFile'"
  set fp [open "|zcat $xmlFile" r]
  .main printStatus "... reading source file '$xmlFile'"
  set data [read $fp]
  close $fp

  ### for debugging
  file delete $xmlFile

  # parse it
  .main printStatus "... parsing source file '$xmlFile'"
  set document [dom parse $data]

  set node [$document documentElement]
  if {$node == ""} {
    printMessage "ERROR: no cdgp logfile while recaning grammar." error
  }

  # parse all lattice nodes
  foreach latticeNode [$node getElementsByTagName "lattice"] {
    set yadaWordgraph [YadaWordgraph ::#auto $this]
    $yadaWordgraph fromLatticeNode $latticeNode
    .main printStatus "... adding wordgraph '[$yadaWordgraph cget -name]'"
    addWordgraph $yadaWordgraph
  }
  
  # parse all annotations
  foreach annoNode [$node getElementsByTagName "annotation"] {
    set yadaAnnotation [YadaAnnotation ::#auto $this]
    $yadaAnnotation fromDomNode $annoNode
    .main printStatus "... adding annotation '[$yadaAnnotation cget -name]'"
    addAnnotation $yadaAnnotation
  }

  # delete it
  $document delete
  set document ""
  set _isScanned 1
  .main printStatus "done rescanning"

  # reestablish the selection
  eval setSelection $oldSelection
}

## ----------------------------------------------------------------------------
## select the list of wordgraphs named in args
## ----------------------------------------------------------------------------
body YadaGrammar::setSelection {args} {
  foreach wgName [array names _wordgraphs] {
    set wordgraph $_wordgraphs($wgName)
    $wordgraph configure -isSelected 0
  }
  foreach wgName $args {
    set wordgraph $_wordgraphs($wgName)
    $wordgraph configure -isSelected 1
  }
  set isModified 1
}

## ----------------------------------------------------------------------------
## @returns a list of selected wordgraph names
## ----------------------------------------------------------------------------
body YadaGrammar::getSelection {} {
  set selection {}
  foreach wordgraph [getWordgraphs] {
    if {[$wordgraph cget -isSelected]} {
      lappend selection [$wordgraph cget -name]
    }
  }
  return $selection
}

## ----------------------------------------------------------------------------
## @returns 1 if the given wordgraph is selected
## ----------------------------------------------------------------------------
body YadaGrammar::isSelected {wgName} {
  return [$_wordgraphs($wgName) cget -isSelected]
}

## ----------------------------------------------------------------------------
## clone
## ----------------------------------------------------------------------------
body YadaGrammar::clone {args} {

  # build a new grammar
  set clone [newGrammar -name $name -files $files]

  # clone all wordgraphs
  foreach wgName [array names _wordgraphs] {
    set newWordgraph [$_wordgraphs($wgName) clone]
    $clone addWordgraph $newWordgraph
  }

  # clone all annotations
  foreach annoName [array names _annotations] {
    set newAnno [$_annotations($annoName) clone]
    $clone addAnnotation $newAnno
  }

  eval $clone configure $args
 
  return $clone
}

## ----------------------------------------------------------------------------
## files callback
## ----------------------------------------------------------------------------
configbody YadaGrammar::files {
 
  if {$files == "<none>"} {
    return
  }
  
  set thisFiles $files
  set files ""
  foreach file $thisFiles {
    catch {lappend files [glob $file]}
  }
  set isModified 1
}

## ----------------------------------------------------------------------------
## Returns the value of private member variable _isScanned
## ----------------------------------------------------------------------------
body YadaGrammar::isScanned {} {
  return $_isScanned
}

## ----------------------------------------------------------------------------
## get the filename where the persistent state is saved in
## ----------------------------------------------------------------------------
body YadaGrammar::getFileName {} {
  global env
  set fileName [file join $env(YADA_ETC) grammars ${name}.xml]
  return [getSecureFileName $fileName]
}
