# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaMachine - bundled information about a yada machine
## \note Actually this class should be split into a YadaScheduler and a
## YadaMachine the latter of which only contains the configurational part of
## the current implementation. As things are we add this class to the YadaScheduler
## module \a and the YadaConfiguration module, a construction that might be a little
## wired.
## \ingroup YadaConfiguration
## \ingroup YadaScheduler
##
## \author Michael Daum
##
## $Id: YadaMachine.tcl,v 1.20 2005-12-01 16:23:27 micha Exp $
## ----------------------------------------------------------------------------
class YadaMachine {
  inherit YadaConfigItem

  # variables ----------------------------------------------------------------

  ## list of YadaNode objects for this machine
  private variable _nodes {}; ## \type TclList

  ## list of all YadaJob s queued up to be scheduled on all nodes
  private variable _jobQueue ""

  ## statistics about the average duration of one YadaJob.
  ## The machine updates this figure the more jobs have been seen.
  private variable _timePerJob 0

  ## number of jobs already run on this machine.
  ## This number is related to _timePerJob computing the average duration
  ## seen in _noComputedJobs.
  private variable _noComputedJobs 0

  ## used for profiling jobs on this node.
  ## This hash mapps YadaJob objects to a clock tick.
  private variable _timerOfJob; ## \type TclArray

  ## state of the machin
  private variable _state "idle"; ## \type TclString
  
  # methods ------------------------------------------------------------------
  public method newNode {args}; ## \type TclList
  public method removeNode {node}; ## \type YadaNode
  public method getNodes {}
  public method getActiveNodes {}
  public method getCpus {}
  public method getLoad {}
  public method getTimePerJob {}
  public method resetStats {}

  public method start {}
  public method stopJob {job}; ## \type YadaJob
  public method queueJob {job}; ## \type YadaJob

  public method toDom {}
  public method fromDom {document}; ## \type domDoc
  public method fromDomNode {machineNode}; ## \type domNode
  public method getFileName {}; ## overriding virtual method
  public method clone {args}; ## \type TclList

  private method _getNextNode {}
  private method _addTime {time}; ## \type TclNumber

  constructor {args} {}; ## \type TclList
  destructor {}

};


## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaMachine::constructor {args} {
  set name "localhost"
  resetStats
  eval configure $args
}

## ----------------------------------------------------------------------------
## serializes this object into a dom document
## @returns a new dom document
## ----------------------------------------------------------------------------
body YadaMachine::toDom {} {
  set document [dom createDocument "yada"]
  set root [$document documentElement]

  set machineNode [$document createElement "machine"]
  $root appendChild $machineNode
  $machineNode setAttribute "name" $name
  
  foreach node $_nodes {
    set nodeNode [$document createElement "node"]
    $machineNode appendChild $nodeNode
    $nodeNode setAttribute "host" [$node cget -host] 
    $nodeNode setAttribute "cpus" [$node cget -cpus]
    $nodeNode setAttribute "access" [$node cget -access]
    $nodeNode setAttribute "isActive" [$node cget -isActive]
  }

  return $document
}

## ----------------------------------------------------------------------------
## initialize this machine from a dom document
## @param document a dom document containing a /yada/machine
## ----------------------------------------------------------------------------
body YadaMachine::fromDom {document} {
  set rootNode [$document documentElement]
  set machineNode [$rootNode selectNodes "/yada/machine"]
  fromDomNode $machineNode
}

## ----------------------------------------------------------------------------
## initialize this machine from a dom node
## @param machineNode containing a machine object description
## ----------------------------------------------------------------------------
body YadaMachine::fromDomNode {machineNode} {

  set name [$machineNode getAttribute "name"]

  # kick occasional leftovers
  foreach node $_nodes {
    itcl::delete object $node
  }
  set _nodes {}

  # create new nodes for this machine
  foreach nodeNode [$machineNode selectNodes "./node"] {
    set host [$nodeNode getAttribute "host"]
    set cpus [$nodeNode getAttribute "cpus"]
    set access [$nodeNode getAttribute "access"]
    set isActive [$nodeNode getAttribute "isActive"]

    newNode -host $host -access $access -cpus $cpus -isActive $isActive
  }
}

## ----------------------------------------------------------------------------
## queueJob
## ----------------------------------------------------------------------------
body YadaMachine::queueJob {job} {
  lappend _jobQueue $job
  $job configure -machine $this
}

## ----------------------------------------------------------------------------
## _getNextNode
## This returns the next best node on the cluster, that is the one with
## the least jobs running on it.
## ----------------------------------------------------------------------------
body YadaMachine::_getNextNode {} {

  set minNode ""
  set minLoad 9999999

  # calculate the minimal node in the cluster
  foreach node [getActiveNodes] {
    set load [$node getLoad]
    if {$load < $minLoad} {
      set minNode $node
      set minLoad $load
    }
  }

  return $minNode
}

## ----------------------------------------------------------------------------
## remove a node from this machine.
## This does only release the association with this machine, that is it will
## not destroy the YadaNode object. Actually the YadaNode::destructor calls
## this method itself to make sure that no dangling reference to a destroyed
## object is left in here.
## ----------------------------------------------------------------------------
body YadaMachine::removeNode {node} {
  set index [lsearch -exact $_nodes $node]
  if {$index < 0} {
    return
  }

  set _nodes [lreplace $_nodes $index $index]
}
 

## ----------------------------------------------------------------------------
## create a YadaNode for this machine
## ----------------------------------------------------------------------------
body YadaMachine::newNode {args} {
  set node [eval YadaNode ::#auto $this $args]
  lappend _nodes $node
  return $node
}

## ----------------------------------------------------------------------------
## returns all YadaNode objects of this machine.
## ----------------------------------------------------------------------------
body YadaMachine::getNodes {} {
  return $_nodes
}

## ----------------------------------------------------------------------------
## returns all active YadaNode objects of this machine.
## ----------------------------------------------------------------------------
body YadaMachine::getActiveNodes {} {
  set nodes {}
  foreach node [getNodes] {
    if {[$node cget -isActive]} {
      lappend nodes $node
    }
  }
  return $nodes
}

## ----------------------------------------------------------------------------
## getCpus
## ----------------------------------------------------------------------------
body YadaMachine::getCpus {} {
  set cpus 0
  foreach node [getNodes] {
    if {[$node cget -isActive]} {
      incr cpus [$node cget -cpus]
    }
  }
  return $cpus
}

## ----------------------------------------------------------------------------
## the collective load on all nodes.
## ----------------------------------------------------------------------------
body YadaMachine::getLoad {} {
  set noJobs 0
  foreach node [getNodes] {
    incr noJobs [$node getLoad]
  }
  return $noJobs
}

## ----------------------------------------------------------------------------
## start
## remove the next YadaJob objects from the queue and run them
## ----------------------------------------------------------------------------
body YadaMachine::start {} {
  if {$_jobQueue == ""} {
    return
  }
  if {$_state == "starting"} {
    return 
  }

  # check current capacity
  set cpus [getCpus]
  set load [getLoad]
  set capacity [expr $cpus - $load]
  if {$capacity <= 0} {
    return
  }
  
  # get current cluster load
  foreach node [getActiveNodes] {
    set nodeLoad($node) [$node getLoad]
  }
  set _state "starting"

  # schedule the next jobs 
  set i $capacity;
  while {$i > 0} {
    set job [lindex $_jobQueue 0]
    if {$job == ""}  {
      break
    }
    set _jobQueue [lreplace $_jobQueue 0 0]
    $job setState "dequeuing"

    # must we do this job?
    set xmlFile [$job getXmlFileName]
    if {[file exists $xmlFile]} {
      $job setState "reloading"
      stopJob $job
    } else {
      # get minimal loaded node
      set minNode ""
      set minLoad 9999999
      foreach node [array names nodeLoad] {
	if {$nodeLoad($node) < $minLoad}  {
	  set minNode $node
	  set minLoad $nodeLoad($node)
	}
      }
      set _timerOfJob($job) [clock seconds]
      $minNode startJob $job
      incr nodeLoad($minNode)
      incr i -1
    }
  }
  set _state "running"
}

## ----------------------------------------------------------------------------
## stopJob
## ----------------------------------------------------------------------------
body YadaMachine::stopJob {job} {
  if {[info command $job] == ""} {
    return
  }
  if {[$job hasPassedState "stopping"]} {
    return
  }

  if {[info exists _timerOfJob($job)]} {
    set duration [expr ([clock seconds] - $_timerOfJob($job)) * 1000]
    _addTime $duration
    unset _timerOfJob($job)
  }

  # remove from job queue
  set index [lsearch -exact $_jobQueue $job]
  if {$index >= 0} {
    set _jobQueue [lreplace $_jobQueue $index $index]
  }

  # notify the node
  set node [$job cget -node]
  if {$node != ""} {
    $node stopJob $job
  } else {
    catch {itcl::delete object $job}
  }

  if {$_jobQueue == ""}  {
    set _state "idle"
  }
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body YadaMachine::destructor {} {
  foreach node $_nodes {
    if {[info command $node] != ""} {
      itcl::delete object $node
    }
  }
}

## ----------------------------------------------------------------------------
## update machine statistics.
## This machine adds the given time to the internal timer statistics.
## Incremental averaging is computed like this:
## - \f$x1 \f$
## - \f$ x_1 * 1/2 + x_2/2 = (x_1 + x_2)/2 \f$
## - \f$ (x_1 * 1/2 + x_2/2) * 2/3 + x_3/3 = (x_1 + x_2 + x_3)/3 \f$
## - \f$ ((x_1 * 1/2 + x_2/2) * 2/3 + x_3/3) * 3/4 + x4/4 = (x_1 + x_2 + x_3 + x_4)/4 \f$
## - ...
## \param time a temporal span
## \returns the computed average time per job as calculated for this machine.
## ----------------------------------------------------------------------------
body YadaMachine::_addTime {time} {
  if {$time <= 0} {
    return $_timePerJob
  }

  incr _noComputedJobs

  set _timePerJob \
    [expr $_timePerJob * (($_noComputedJobs -1.0) / ($_noComputedJobs + 0.0)) \
	  + $time / ($_noComputedJobs + 0.0)]

  return $_timePerJob
}

## ----------------------------------------------------------------------------
## getTimePerJob
## ----------------------------------------------------------------------------
body YadaMachine::getTimePerJob {} {
  return $_timePerJob
}

## ----------------------------------------------------------------------------
## get the filename where the persistent state is saved in
## ----------------------------------------------------------------------------
body YadaMachine::getFileName {} {
  global env
  set fileName [file join $env(YADA_ETC) machines ${name}.xml]
  return [getSecureFileName $fileName]
}

## ----------------------------------------------------------------------------
## clone
## ----------------------------------------------------------------------------
body YadaMachine::clone {args} {
  set clone [newMachine -name $name]

  foreach node $_nodes {
    set host [$node cget -host]
    set cpus [$node cget -cpus]
    set access [$node cget -access]
    set isActive [$node cget -isActive]
    $clone newNode \
      -host $host -cpus $cpus -access $access -isActive $isActive
  }

  eval $clone configure $args
  return $clone
}

## ----------------------------------------------------------------------------
## reset statistics about jobs and their exec time.
## ----------------------------------------------------------------------------
body YadaMachine::resetStats {} {
  set _timePerJob 60000
  set _noComputedJobs 1
}
