# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaPlainDocument - abstract yada document.
## \ingroup YadaDocuments
##
## \author Michael Daum 
##
## $Id: YadaPlainDocument.tcl,v 1.16 2005-12-01 16:23:27 micha Exp $
## ----------------------------------------------------------------------------
class YadaPlainDocument {

  # variables ----------------------------------------------------------------
  
  ## name of the document
  public variable name ""
  
  ## flag indicating whether this object functions as combined model-view object (value 1),
  ## or as a pure model object (value 0: has to be set explicitly when instantiating object)
  public variable hasView 1; ## \type TclBoolean
  
  ## flag indicating initialization.
  ## Possible falues are 0 and 1. init() sets this to 1 when it is called
  ## in the init chain calls of inheriting classes, 
  ## i.e. YadaConfigDocument::init() or YadaRunnableDocument::init().
  protected variable _isInitialized 0; ## \type TclBoolean

  protected variable _domDocument ""; ## \type domDocument

  ## Denotes time of most recent modification of document's XML file
  ## CAUSED by this object. This allows recognizing cases where a second
  ## object modifies the same file concurrently, in which case the need
  ## may arise to synchronize the object with the new file contents.
  protected variable _lastModificationTime 0


  ## flag indicating activation / deactivation state
  private variable _isVisible 0; ## \type TclBoolean

  ## flag indicating a modified model
  private variable _isModified 0; ## \type TclBoolean

  # methods ------------------------------------------------------------------
  public method load {}; ## \virtual
  public method save {}
  public method toDom {}; ## \virtual
  public method fromDom {document}; ## \virtual \type domDoc
  public method fromDomNode {node}; ## \virtual \type domNode
  public method activationHandle {}; ## \virtual
  public method deActivationHandle {}; ## \virtual
  public method init {} ;## \virtual
  public method setToggle {name {value ""}}; ## \type TclString, TclRef, \virtual
  public method evalXpath {xpath}; ## \type TclString

  public method hasSavedState {}
  public method isVisible {}
  public method isModified {}
  public method setModified {}
  public method resetModified {}
  public method updateWindowTitle {}
  public method updateMenus {}; ## \virtual
  public method getWindowTitle {}
  public method getFileName {} ; ## \purevirtual
  public method getZippedFileName {}
  public method lastChanged {} 
  
  protected method updateModificationTime {}

  constructor {} {}
  destructor {}

};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaPlainDocument::constructor {} {
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body YadaPlainDocument::destructor {} {
  if {$_domDocument != ""} {
    $_domDocument delete
  }
}


## ----------------------------------------------------------------------------
## activationHandle
## ----------------------------------------------------------------------------
body YadaPlainDocument::activationHandle {} {
  set _isVisible 1
  updateWindowTitle
  updateMenus
}

## ----------------------------------------------------------------------------
## deActivationHandle
## ----------------------------------------------------------------------------
body YadaPlainDocument::deActivationHandle {} {
  set _isVisible 0
}

## ----------------------------------------------------------------------------
## init
## ----------------------------------------------------------------------------
body YadaPlainDocument::init {} {
  if {!$_isInitialized} {
    set _isInitialized 1
  }
}

## ----------------------------------------------------------------------------
## setToggle
## ----------------------------------------------------------------------------
body YadaPlainDocument::setToggle {name {value ""}} {
}


## ----------------------------------------------------------------------------
## loads current document from file (file name depends on document name)
## @returns 0 if the document can't be loaded
## ----------------------------------------------------------------------------
body YadaPlainDocument::load {} {
  set oldCursors [setBusy]

  set zippedFileName [getZippedFileName]
  if {![file exists $zippedFileName]} {  
    resetBusy $oldCursors
    return 0
  }

  set fileID [open "|zcat $zippedFileName" {RDONLY}]
  set xmlData [read  $fileID]
  close $fileID
  if {$_domDocument != ""} {
    $_domDocument delete
  }
  set _domDocument [dom parse $xmlData]
  resetModified
  fromDom $_domDocument

  resetBusy $oldCursors
  updateWindowTitle
  updateModificationTime
  updateMenus

  return 1
}

## ----------------------------------------------------------------------------
## Saves current document to file (file name depends on document name)
## ----------------------------------------------------------------------------
body YadaPlainDocument::save {} {
  set oldCursors [setBusy]
  if {$_domDocument != ""} {
    $_domDocument delete
  }
  set _domDocument [toDom]	
  set fileNameToSave [getFileName]

  if { [file exists ${fileNameToSave}.gz] } {
      file delete ${fileNameToSave}.gz
  }

  set fileID [open $fileNameToSave {CREAT WRONLY}]
  set rootNode [$_domDocument documentElement]
  puts $fileID [$rootNode asXML]
  close $fileID

  if {[catch {exec gzip -f $fileNameToSave} errMsg]} {
      _resetBusy $oldCursors
      error "Saving failed: $errMsg"
  }
  resetModified
  resetBusy $oldCursors
  updateWindowTitle
  updateModificationTime
  updateMenus
}

## ----------------------------------------------------------------------------
## evaluate an xpath expression in the related xml document
## @param xpath a vaild xpath expression for a yada document
## @returns the result value on success and the empty string on failure
## ----------------------------------------------------------------------------
body YadaPlainDocument::evalXpath {xpath} {

  if {$_domDocument == ""} {
    set zippedFileName [getZippedFileName]
    if {![file exists $zippedFileName]} {  
      return ""
    }

    set fileID [open "|zcat $zippedFileName" {RDONLY}]
    set xmlData [read  $fileID]
    close $fileID
    set _domDocument [dom parse $xmlData]
  }

  set rootNode [$_domDocument documentElement]
  return [$rootNode selectNodes $xpath]
}


## ----------------------------------------------------------------------------
## @returns 1 if this document is currently active
## ----------------------------------------------------------------------------
body YadaPlainDocument::isVisible {} {
  return $_isVisible
}

## ----------------------------------------------------------------------------
## @returns 1, if configuration data has changed, 0 otherwise
## ----------------------------------------------------------------------------
body YadaPlainDocument::isModified {} {
  return $_isModified
}

## ----------------------------------------------------------------------------
## Marks item as modified
## ----------------------------------------------------------------------------
body YadaPlainDocument::setModified {} {
  set _isModified 1
  updateWindowTitle
  updateMenus
}

## ----------------------------------------------------------------------------
## Removes modification mark
## ----------------------------------------------------------------------------
body YadaPlainDocument::resetModified {} {
  set _isModified 0
  updateWindowTitle
  updateMenus
}

## ----------------------------------------------------------------------------
## Predicate returning 1, if file containing persistent state of YadaGrammar
## exists, otherwise 0
## ----------------------------------------------------------------------------
body YadaPlainDocument::hasSavedState {} {
  set zippedFileName [getZippedFileName]
  if {[file exists $zippedFileName]} {
    return 1
  } else {
    return 0
  }
}

## ----------------------------------------------------------------------------
## get the title for this document
## this function will be overloaded and called in a chain to concatenate
## the resulting title string
## @returns the title prefix for all YadaPlainDocument classes
## ----------------------------------------------------------------------------
body YadaPlainDocument::getWindowTitle {} {
  return "[string totitle $name 0 1]"
}

## ----------------------------------------------------------------------------
## displays title in main window (reflecting modification state of document)
## ----------------------------------------------------------------------------
body YadaPlainDocument::updateWindowTitle {} {

  if {!$hasView || !$_isVisible} {
    return
  }
 set title [getWindowTitle]
 if {[isModified]} {
   wm title .  "Yada.$title *** changed *** " 
 } else {
   wm title . "Yada.$title"
 }
}

## ----------------------------------------------------------------------------
## Returns time of last update to document XML file or an empty string,
## if there is no such file 
## ----------------------------------------------------------------------------
body YadaPlainDocument::lastChanged {} {
  if {[hasSavedState]} {
    set zippedFileName [getZippedFileName]
    file stat $zippedFileName fileAttribs
    return $fileAttribs(ctime)
  } else {
      return ""
  }
}

## ----------------------------------------------------------------------------
## Synchronizes internal modification attribute with modification time of
## document's XML file
## ----------------------------------------------------------------------------
body YadaPlainDocument::updateModificationTime {} {
  if {[hasSavedState]} {
    file stat [getZippedFileName] fileAttribs
    set _lastModificationTime $fileAttribs(ctime)
  }
}

## ----------------------------------------------------------------------------
## get filename of compressed data
## ----------------------------------------------------------------------------
body YadaPlainDocument::getZippedFileName {} {
  return "[getFileName].gz"
}

## ----------------------------------------------------------------------------
## base updateMenus
## ----------------------------------------------------------------------------
body YadaPlainDocument::updateMenus {} {
}

## ----------------------------------------------------------------------------
## base getFileName
## ----------------------------------------------------------------------------
body YadaPlainDocument::getFileName {} {
  return ""
}
