# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## \file YadaProc.tcl
## Bunch of small helper functions.
## This module collects all helpers that don't fit in a proper class for themselves.
## Actually for those cases the YadaMain might be taken, but ...
##
## \author Michael Daum
##
## $Id YadaProc.tcl$
## ----------------------------------------------------------------------------

## ----------------------------------------------------------------------------
## convert a time in numerical format to a human readeable format.
## \param time time in milliseconds
## \returns a string representation
## ----------------------------------------------------------------------------
proc timeToString {time} {; ## \type TclNumber
  
  if {$time <= 0.0} {
    return "0 ms"
  }
  
  # calculate the units
  set time [expr double($time)]
  set w    [expr floor($time / 604800000.0)]
  set time [expr $time - $w * 604800000.0]
  set d    [expr floor($time / 86400000.0)]
  set time [expr $time - $d * 86400000.0]
  set h    [expr floor($time / 3600000.0)]
  set time [expr $time - $h * 3600000.0]
  set m    [expr floor($time / 60000.0)]
  set time [expr $time - $m * 60000.0]
  set s    [expr floor($time / 1000.0)]
  set ms   [expr $time - $s * 1000]

  # build the string 
  set isFirst 1
  set result ""
  foreach unit {w d h m s ms} {
    set value [subst \$$unit]
    if {$value > 0.0} {
      if {$isFirst} {
	set isFirst 0
      } else {
	append result " "
      }
      append result [format "%.0f %s" $value $unit]
    }
  }

  return $result
}

## ----------------------------------------------------------------------------
## get the index of an entry in a listbox.
## \param listBox the tcl command of the list box 
## \param entry an entry in the \a listBox
## \returns the index > 0 if the \a entry exists, -1 if not so
## ----------------------------------------------------------------------------
proc getIndexOfName {listBox entry} {; ## \type TclCommand, TclString
  set listOfEntries [$listBox get 0 end]
  set index [lsearch $listOfEntries $entry]
  return $index
}

## ----------------------------------------------------------------------------
## open a message dialog with a given text.
## \param message the notification text
## \param image an image name, see YadaImages for possible values
## ----------------------------------------------------------------------------
proc printMessage {message {image "info"}} {; ## \type TclString, TclString
  if {![winfo exists .msg]} {
    iwidgets::messagedialog .msg \
  } 
  .msg configure \
    -text $message \
    -title "Yada - Message"  \
    -image [YadaImages::get $image] \
    -modality application \
    -textpadx 10 \
    -master .

  .msg hide Cancel
  .msg center .
  .msg activate
  destroy .msg
}

## ----------------------------------------------------------------------------
## extract notifications.
## This helper extracts all notifications in a cdgp DomDocument. That is, it
## searches for \\<notify ...\\> nodes and gets the \a desc attribute from it.
## \param document the DomDocument that we are interested in
## \param severity optionally specify the \c type attribute of the notification
## that we want to extract
## \returns the TclList of all messages in all notifications
## ----------------------------------------------------------------------------
proc getAllNotifications {document {severity ""}} {; ## \type DomDocument, TclString

  set messages ""

  # get the cdgp logfile root element
  set cdgpNode [$document getElement "cdgp"]
  if {$cdgpNode != ""} {

    if {$severity != ""} {
      set nodes [$cdgpNode descendant all "notify" "type" $severity]
    } else {
      set nodes [$cdgpNode descendant all "notify"]
    }

    foreach node $nodes {
      lappend messages [$node getAttribute "desc"]
    }
  } 

  return $messages
}

## ----------------------------------------------------------------------------
## Sets mouse cursor to indicate busy processing
## @return list of changed widgets and previous cursor settings
##         ({<widget> <cursor> ....})
## ----------------------------------------------------------------------------
proc setBusy {} {
    set busy {}
    set list {.}
    # Traverse the widget hierarchy to locate widgets with 
    # a nondefault -cursor setting.
    #
    while {$list != ""} {
        set next {}
        foreach w $list {
            catch {set cursor [$w cget -cursor]}
            if {[winfo toplevel $w] == $w || $cursor != "" } {
                lappend busy $w $cursor
                set cursor {}
            }
            set next [concat $next [winfo children $w]]
        }
        set list $next
    }

    # Change the cursor:
    #
    foreach {w _} $busy {
        catch {$w configure -cursor watch}
    }
    update idletasks

    return $busy
}


## ----------------------------------------------------------------------------
## Restores mouse cursor settings for widget hierarchy corresponding
## to saved settings in first argument (should be identical to return value
## of proc setBusy)
## ----------------------------------------------------------------------------
proc resetBusy { listPrevSettings } {; ## \type TclList
    # Restore the original cursor settings.
    #
    foreach {w cursor} $listPrevSettings {
        catch {$w configure -cursor $cursor}
    }
}

## ----------------------------------------------------------------------------
## View a file given a name and popup a dialog.
## ----------------------------------------------------------------------------
proc viewFile {fileName {title ""}} {; ## \type TclString, TclString

  if {![file exists $fileName]} {
    return ""
  }
  set fd [open "|zcat $fileName" r]
  set data [read $fd]
  close $fd

  if {$title == ""} {
    set title "Yada.Log - $fileName"
  }

  set yadaLog [YadaLogDialog .#auto \
      -modality none \
      -title $title \
      -hscrollmode dynamic \
      -wrap none \
      -textbackground gray90 \
      -textfont {-misc-fixed-medium-normal-semicondensed-*-13-*-*-*-*-*-*-*} \
      -width 800 \
      -height 600]

  $yadaLog autotag INFO -foreground black -background green3 
  $yadaLog autotag WARNING -foreground black  -background yellow2 
  $yadaLog autotag ERROR -foreground white -background red3 
  $yadaLog autotag DEBUG -foreground white -background blue3 
  $yadaLog autotag PROFILE -foreground black -background lightblue 
  $yadaLog print "INFO: loading $fileName\n$data\n"
  $yadaLog see 0.0
  $yadaLog center

  return $yadaLog
}

## ----------------------------------------------------------------------------
## interface to create a new machine
## ----------------------------------------------------------------------------
proc newMachine {args} {; ## \type TclList
  return [eval YadaMachine ::#auto $args]
}

## ----------------------------------------------------------------------------
## interface to create a new experiment
## ----------------------------------------------------------------------------
proc newExperiment {args} {; ## \type TclList
  return [eval YadaExperiment ::#auto $args]
}

## ----------------------------------------------------------------------------
## interface to create a new grammar
## ----------------------------------------------------------------------------
proc newGrammar {args} {; ## \type TclList
  return [eval YadaGrammar ::#auto $args]
}

## ----------------------------------------------------------------------------
## returns a secured version of the given filename
## ----------------------------------------------------------------------------
proc getSecureFileName {fileName} {; ## \type TclString
  set oldFileName $fileName
  set fileName [string tolower $fileName]
  regsub -all {[^a-z0-9/\.]} $fileName {_} fileName
  regsub -all {__*} $fileName {_} fileName
  regsub {_*$} $fileName {} fileName

  return $fileName
}

## ----------------------------------------------------------------------------
## debug procedure: print information on the console
## ----------------------------------------------------------------------------
set watch_flag 0
proc watch {{text ""}} {
  global watch_flag
  if {$watch_flag == 1} {
    puts $text
  }
}
