# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaWordgraph - bundled information about a wordgraph.
## This class represents a wordgraph in a YadaGrammar. Scanning a YadaGrammar
## (see YadaGrammar::scan()) results in bunch of YadaWordgraph s being constructed.
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## $Id: YadaWordgraph.tcl,v 1.14 2005-12-01 16:23:27 micha Exp $
## ----------------------------------------------------------------------------
class YadaWordgraph {

  # variables ----------------------------------------------------------------

  ## name of the wordgraph
  public variable name ""; ## \type TclString

  ## filename which this wordgraph has been extracted from
  public variable file ""; ## \type TclString

  ## linenumber in the file where this wordgraph has been defined
  public variable lineNo 0; ## \type TclNumber

  ## flag indicating whether the wordgraph is trivial (linear) or not
  public variable isTrivial 0; ## \type TclBoolean

  ## flag indicating whether wordgraph has been selected
  public variable isSelected 0; ## \type TclBoolean

  ## list of word arcs making up the wordgraph
  ## these are tripples of the form "from to word"
  public variable arcs {}; ## \type TclList

  ## name of related annotation
  private variable _annotations {}; ## \type TclList

  ## the grammar object we belong to
  private variable _grammar ""; ## \type YadaGrammar

  # methods ------------------------------------------------------------------
  public method addAnnotation {annotation}; ## \type TclString
  public method getAnnotations {}
  public method getBestAnnotation {}
  public method getWords {}
  public method clone {args}; ## \type TclList
  public method toDom {}
  public method fromDom {document}; ## \type domDoc
  public method fromDomNode {wordgraphNode}; ## \type domNode
  public method fromLatticeNode {latticeNode}; ## \type DomNode

  constructor {grammar args} {}; ## \type YadaGrammar, TclList
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaWordgraph::constructor {grammar args} {
  set _grammar $grammar
  eval configure $args
}

## ----------------------------------------------------------------------------
## serializes this wordgraph into a dom document
## @returns a new dom document
## ----------------------------------------------------------------------------
body YadaWordgraph::toDom {} {
  set document [dom createDocument "wordgraph"]
  set rootNode [$document documentElement]

  $rootNode setAttribute "name" $name
  $rootNode setAttribute "isTrivial" $isTrivial
  $rootNode setAttribute "isSelected" $isSelected 

  set fileNode [$document createElement "file"]
  $rootNode appendChild $fileNode
  $fileNode setAttribute "name" $file
  $fileNode setAttribute "lineNo" $lineNo

  foreach {from to word} $arcs {
    set arcNode [$document createElement "arc"]
    $rootNode appendChild $arcNode
    $arcNode setAttribute "from" $from
    $arcNode setAttribute "to" $to
    $arcNode setAttribute "word" $word
  }


  return $document
}

## ----------------------------------------------------------------------------
## initialize wordgraph with dom document
## note, that only the first wordgraph found is used, i.e. this function is 
## pointless ;)
## @param document containing a /yada/grammar/wordgraph
## ----------------------------------------------------------------------------
body YadaWordgraph::fromDom {document} {
  set rootNode [$document documentElement]
  set wordgraphNode [$rootNode selectNodes {/yada/grammar/wordgraph[1]}]
  fromDomNode $wordgraphNode
}

## ----------------------------------------------------------------------------
## initialize wordgraph with dom node 
## @param rootNode root DOM-Node (Tag <wordgraph>)
## ----------------------------------------------------------------------------
body YadaWordgraph::fromDomNode {wordgraphNode} {

  set fileNode [$wordgraphNode selectNodes "file"]
  set arcs {}
  foreach arcNode [$wordgraphNode selectNodes "arc"] {
    lappend arcs \
      [$arcNode getAttribute "from"] \
      [$arcNode getAttribute "to"] \
      [$arcNode getAttribute "word"]
  }

  set name [$wordgraphNode getAttribute "name"]
  set file [$fileNode getAttribute "name"] 
  set lineNo [$fileNode getAttribute "lineNo"] 
  set isTrivial [$wordgraphNode getAttribute "isTrivial"] 
  set isSelected [$wordgraphNode getAttribute "isSelected"] 
}

## ----------------------------------------------------------------------------
## registers an annotation object to refere to this wordgraph
## ----------------------------------------------------------------------------
body YadaWordgraph::addAnnotation {annotation} {
  if {[lsearch $_annotations $annotation] < 0} {
    lappend _annotations $annotation
  }
}

## ----------------------------------------------------------------------------
## @returns the list of YadaAnnotation objects
## ----------------------------------------------------------------------------
body YadaWordgraph::getAnnotations {} {
  return $_annotations
}

## ----------------------------------------------------------------------------
## @returns the 'best' annotation, that is one that is called 'gold...'. If
## there's no gold-annotation then we return the first one known ... quite
## arbitrary ;)
## ----------------------------------------------------------------------------
body YadaWordgraph::getBestAnnotation {} {
  
  foreach anno $_annotations {
    if {[regexp -lineanchor {^gold} [$anno cget -name]]} {
      return $anno
    }
  }

  return [lindex $_annotations 0]
}

## ----------------------------------------------------------------------------
## initialize this wordgraph from a lattice node of a cdgp log file
## ----------------------------------------------------------------------------
body YadaWordgraph::fromLatticeNode {latticeNode} {
  if {[$latticeNode nodeName] != "lattice"} {
    error "ERROR: '$latticeNode' is not a lattice node"
  }

  # get the lattice id
  configure -name [$latticeNode getAttribute "id"]
  
  # get the file info
  set node [$latticeNode descendant 1 "file"]
  if {$node == ""} {
    error "ERROR: illegal xml lattice format"
  }
  set file [$node getAttribute "name"]
  set lineNo [$node getAttribute "lineNo"]

  # is it a trivial wordgraph
  set node [$latticeNode descendant 1 "boolean" "name" "trivial"]
  if {$node == ""} {
    error "ERROR: illegal xml lattice format"
  }
  set isTrivial [$node getAttribute "value"]
  
  # scan the arcs
  set arcNodes [$latticeNode getElementsByTagName "arc"]
  if {$arcNodes == ""} {
    error "ERROR: illegal xml lattice format"
  }
  set arcs ""
  foreach node $arcNodes {
    set from [$node getAttribute "from"]
    set to [$node getAttribute "to"]
    set word "[$node getAttribute word]"
    lappend arcs $from $to "$word"
  }
}

## ----------------------------------------------------------------------------
## return the list of words in this wordgraph
## ----------------------------------------------------------------------------
body YadaWordgraph::getWords {} {
  set words ""
  foreach {from to word} $arcs {
    lappend words "$word"
  }

  return $words
}

## ----------------------------------------------------------------------------
## clone
## ----------------------------------------------------------------------------
body YadaWordgraph::clone {args} {

  # build the clone
  set newWordgraph [YadaWordgraph ::#auto $_grammar \
    -name $name \
    -file $file \
    -lineNo $lineNo \
    -isTrivial $isTrivial \
    -isSelected $isSelected \
    -arcs $arcs]

  eval $newWordgraph configure $args

  return $newWordgraph
}
