/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Birth: 10.5.1
 */

/* -------------------------------------------------------------------------
 * @addtogroup Compiler Compiler - compile a constraint grammar
 * \author Michael Daum (see also AUTHORS and THANKS for more)
 *
 * $Id: compile.c,v 1.94 2006-01-17 15:59:49 foth Exp $
 * 
 * This module implements a compiler in order to translate constraints into C code.
 * @{
 */

/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>
#include <unistd.h>
#include <time.h>
#include <libgen.h>
#include <stdio.h>
#include <sys/stat.h>
#include <pwd.h> 
#include <string.h> 
#include <assert.h> 
#include <dlfcn.h> 
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include "cdg.h"
#include "hook.h"
#include "command.h"
#include "input.h"
#include "set.h"
#include "timer.h"
#include "constraintnet.h"
#include "eval.h"


/* -- MACROS -------------------------------------------------------------- */
#define INIT_GRAMMAR "initGrammar" /**< abbrev. */
#define FINIT_GRAMMAR "finitGrammar" /**< abbrev. */

/* -- TYPE DEFINITIONS ---------------------------------------------------- */

/* ---------------------------------------------------------------------------
 * definition of a \c CDG compiler object.
 */
typedef struct CompilerStruct {
  String ccFileName;	     /**< filename for C code */
  FILE * ccFile;	     /**< filehandle of the file in ccFileName */
  String objFileName;	     /**< filename for object code */
  String soFileName;	     /**< filename for C dll grammar */
  int indent;		     /**< current indentation level */
  String indentString;	     /**< current indentation */
  Vector indentStrings;	     /**< possible indentation */
  Boolean needsIndent;	     /**< TRUE by comPrintln, FALSE by comPrint */
  Vector directions;	     /**< collection of all defined directions */
  Vector connexions;	     /**< collection of all defined connexions */
  Constraint currentConstraint;	/**< the constraint being currently translated */
  Formula currentFormula;	/**< the Formula being currently translated */
  int maxValues;	     /**< max number of variables to eval all constraints */
  int maxLookupStrings;	     /**< max number of Strings used in 'lookup' constraints */
  Vector strings;	     /**< list of strings to allocated statically */
  Boolean translateOnly;     /**< if TRUE we dont compile+load a shared object */
} CompilerStruct;
typedef struct CompilerStruct *Compiler; /**< type of a Compiler */

/* ---------------------------------------------------------------------------
 * definition of a approver object.
 */
typedef struct {
  int noRounds;		     /**< no of times to eval all lv pairs */
  Timer timer;		     /**< profiling */
  unsigned long totalInterpretedUnaryTime; /**< profiling information */
  unsigned long totalCompiledUnaryTime; /**< profiling information */
  unsigned long totalInterpretedBinaryTime; /**< profiling information */
  unsigned long totalCompiledBinaryTime; /**< profiling information */
} ApproverStruct;
typedef ApproverStruct *Approver; /**< type of an Approver */

/* ---------------------------------------------------------------------------
 * default make settings.
 * This structure bundles all the information that is needed to compile and
 * link a binary \c CDG grammar. This is called \c MakeInfo in according to
 * the normal place where such information is kept.
 */
typedef struct MakeInfoStruct {
  String cc;		     /**< CC */
  String ld;		     /**< LD */
  String cFlags;	     /**< CFLAGS */
  String ldFlags;	     /**< LDFLAGS */
  String ldLibs;	     /**< LDLIBS */
  String includes;	     /**< INCLUDES */
} MakeInfoStruct;
/** A Pointer to a MakeInfoStruct */
typedef struct MakeInfoStruct *MakeInfo;

/* ---------------------------------------------------------------------------
 * possible return types of formulas, predicates, functions, terms.
 */
typedef enum {
  RTBoolean = (1L << 0),		/**< Boolean return type */
  RTNumber = (1L << 1),			/**< Number return type */
  RTString = (1L << 2),			/**< String return type */
  RTAVNode = (1L << 3),			/**< AVNode return type */
  RTList = (1L << 4),			/**< List return type */
  RTConjunction = (1L << 5),		/**< Conjunction return type */
  RTDisjunction = (1L << 6),		/**< Disjunction return type */
  RTError = (1L << 7),			/**< Syntax error "return type" */
  RTLexemNode = (1L << 8),		/**< LexemNode return type */
  RTGraphemNode = (1L << 9),		/**< GraphemNode return type */
  RTNoError = (1L << 10),		/**< indicate that no error is possible 
					 * in the returning function */
  RTPeek = (1L << 11),			/**< indicate a lexicon access 
					 * in the returning function */
  RTLexemPosition = (1L << 12),		/**< indicate the position where the
					 * returning function is peeking
					 * 0: dependent, 1: regent */
} ReturnType;

/* -- VARIABLES --------------------------------------------------------- */
static Compiler com;	/**< this is allocated on every call to comCompile */
static MakeInfo makeInfo; /**< this is allocated once on comInitialize */

/* -- FUNCTIONS --------------------------------------------------------- */

/* main functions */
Boolean comTranslate(void);
Boolean comMake(void);

/* constructor, destructor */
Compiler comNew(void);
void comFree(Compiler);
MakeInfo comMakeInfoNew(void);
void comMakeInfoFree(MakeInfo);
Approver comNewApprover(void);
void comFreeApprover(Approver);

/* output helper */
void comIndent(void);
void comOutdent(void);
void comPrint(String, ...);
void comPrintln(String, ...);
void comWriteWarning(String, ...);
void comWriteError(String, ...);
void comWriteHeader(void);
void comWriteDeclarations(void);
void comWriteFunctions(void);
void comWriteInitFunction(void);
void comWriteFinitFunction(void);
String comEscapeQuotes(String);

/* string registrator */
int comRegisterString(String);

/* stringifier */
String comDirectionToString(Direction);
String comConnextionToString(Connexion);
String comValueTypeToString(Value);
String comFormulaTypeToString(Formula);
String comTermTypeToString(Term);
String comReturnTypeToString(ReturnType);

/* grammar analyzer */
void comAnalyzeGrammar(void);
int comTermDepth(Term);
int comFormulaDepth(Formula);
int comConstraintDepth(Constraint);
int comPredicateDepth(String, List);
int comFunctionDepth(String, List);
int comMaxLookupStrings(Constraint);
int comMaxLookupStringsInFormula(Formula);
int comMaxLookupStringsInTerm(Term);

/* type inference */
ReturnType comReturnTypeOfFormula(Formula);
ReturnType comReturnTypeOfTerm(Term);
ReturnType comReturnTypeOfFunction(String, List);
ReturnType comReturnTypeOfPredicate(String, List);

/* translators */
void comTranslateUnaryConstraints(void);
void comTranslateBinaryConstraints(void);
void comTranslateConstraint(Constraint);
void comTranslateFormula(Formula, int);
void comTranslateTerm(Term, Boolean, Boolean, int);
void comTranslatePredicate(String, List, int);
void comTranslateFunction(String, List, int);
void comTranslateLexicalAccess(Term);
void comTranslateLexemNodeAccess(Term, int);
void comTranslateMatch(List, int);
void comTranslateMinMax(String, List, int);
void comTranslateAbs(List, int);
void comTranslateDistance(List, int);
void comTranslateStartStop(String, List, int);
void comTranslateGuard(String, List, int);
void comTranslateExists(List);
void comTranslateSubsumes(String, List, int);
void comTranslateChunkHead(String, List, int);
void comTranslateUnder(String, List, int);
void comTranslateBetween(String, List, int);
void comTranslateBottomPeek(Term, int);
void comTranslateTopPeek(Term, int);
void comTranslateArithmetics(Term, int);
void comTranslateEquation(Formula, int);
void comTranslateUnEquation(Formula, int);
void comTranslatePrint(List, int);
void comTranslateDirection(Formula);
void comTranslateConnexion(Formula);
void comTranslateString(Term, int);
void comTranslateNumber(Term, int);
void comTranslatePhrasequotes(List, int);
void comTranslateLookup(List, int);
void comTranslateQuotes(List, int);
void comTranslateParens(List, int);
void comTranslateParent(List, int);
void comTranslatePts(List, int);
void comTranslateHeight(List, int);
void comTranslateIs(String, List, int);
void comTranslateConnected(String, List, int);
void comTranslateHas(String, List, int);

/* index finder */
int comIndexOfHierarchy(String);
int comIndexOfConstraint(String);
int comIndexOfVarInfo(VarInfo);

/* comparators */
Boolean comCompareNets(Approver, ConstraintNet, ConstraintNet);
Boolean comCompareAllLvs(Approver, ConstraintNet, ConstraintNet);
Boolean comCompareAllLvPairs(Approver, ConstraintNet net);
LevelValue comFindComparableLv(ConstraintNet, LevelValue);
Boolean comCompareLvs(LevelValue, LevelValue);
Boolean comCompareWithContext(Approver, ConstraintNet net, List contextList);

/* -------------------------------------------------------------------------
 * constructor for a Compiler
 */
Compiler comNew(void) 
{
  int i = 0;
  String indentString = 
    strRegister("                                                                                ");

  Compiler thisCom = (Compiler)memMalloc(sizeof(CompilerStruct));

  thisCom->ccFileName = NULL;
  thisCom->objFileName = NULL;
  thisCom->soFileName = NULL;
  thisCom->directions = vectorNew(5);
  thisCom->connexions = vectorNew(8);
  thisCom->indent = 0;
  thisCom->indentString = strRegister("");
  thisCom->indentStrings = vectorNew(40);
  thisCom->needsIndent = TRUE;
  thisCom->currentConstraint = NULL;
  thisCom->strings = vectorNew(1000);
  thisCom->translateOnly = FALSE;
  
  vectorAddElement(thisCom->directions, (Pointer)Left);
  vectorAddElement(thisCom->directions, (Pointer)Right);
  vectorAddElement(thisCom->directions, (Pointer)Nil);
  vectorAddElement(thisCom->directions, (Pointer)NonNil);
  vectorAddElement(thisCom->directions, (Pointer)AnyDir);

  vectorAddElement(thisCom->connexions, (Pointer)Top);
  vectorAddElement(thisCom->connexions, (Pointer)Bottom);
  vectorAddElement(thisCom->connexions, (Pointer)Over);   
  vectorAddElement(thisCom->connexions,  (Pointer)Under);   
  vectorAddElement(thisCom->connexions, (Pointer)None);   
  vectorAddElement(thisCom->connexions, (Pointer)Inverse);   
  vectorAddElement(thisCom->connexions, (Pointer)Identical);
  vectorAddElement(thisCom->connexions, (Pointer)AnyCon);   

  for (i = 0; i < vectorCapacity(thisCom->indentStrings) * 2 - 2; i+=2) {
    indentString[i] = 0;
    vectorAddElement(thisCom->indentStrings, strRegister(indentString));
    indentString[i] = ' ';
  }

  return thisCom;
}

/* ---------------------------------------------------------------------------
 * destructor for a Compiler
 */
void comFree(Compiler thisCom) 
{
  int i;
  
  if (thisCom == NULL)
    return;
  
  for (i = 0; i < vectorSize(thisCom->indentStrings); i++) {
    cdgFreeString(vectorElement(thisCom->indentStrings, i));
  }
  for (i = 0; i < vectorSize(thisCom->strings); i++) {
    cdgFreeString(vectorElement(thisCom->strings, i));
  }
  vectorDelete(thisCom->indentStrings);
  vectorDelete(thisCom->strings);
  vectorDelete(thisCom->directions);
  vectorDelete(thisCom->connexions);

  memFree(thisCom);
}

/* ---------------------------------------------------------------------------
 * constructor for a MakeInfo
 */
MakeInfo comMakeInfoNew(void)
{
  MakeInfo thisMakeInfo = (MakeInfo)memMalloc(sizeof(MakeInfoStruct));

  thisMakeInfo->cc = NULL;
  thisMakeInfo->ld = NULL;
  thisMakeInfo->cFlags = NULL;
  thisMakeInfo->ldFlags = NULL;
  thisMakeInfo->ldLibs = NULL;
  thisMakeInfo->includes = NULL;

  return thisMakeInfo;
}

/* ---------------------------------------------------------------------------
 * destructor of a MakeInfo
 */
void comMakeInfoFree(MakeInfo thisMakeInfo)
{
  if (!thisMakeInfo)
    return;
  
  cdgFreeString(thisMakeInfo->cc);
  cdgFreeString(thisMakeInfo->ld);
  cdgFreeString(thisMakeInfo->cFlags);
  cdgFreeString(thisMakeInfo->ldFlags);
  cdgFreeString(thisMakeInfo->ldLibs);
  cdgFreeString(thisMakeInfo->includes);
  memFree(thisMakeInfo);
}

/* ---------------------------------------------------------------------------
 * increase indentation
 */
void comIndent(void)
{
  com->indent++;
  com->indentString = vectorElement(com->indentStrings, 
      min(vectorSize(com->indentStrings), com->indent));
}

/* ---------------------------------------------------------------------------
 * decrease indentation
 */
void comOutdent(void)
{
  com->indent--;
  assert(com->indent >= 0);
  com->indentString = vectorElement(com->indentStrings, 
      min(vectorSize(com->indentStrings), com->indent));
}

/* ----------------------------------------------------------------------
 * grant a newline at the end
 */
void comPrintln(String format, ...)
{
  va_list ap;

  if (com->needsIndent) {
    fprintf(com->ccFile, "%s", com->indentString);
  } 
  va_start(ap, format);
  vfprintf(com->ccFile, format, ap);
  va_end(ap);

  fprintf(com->ccFile, "\n");
  com->needsIndent = TRUE;
}

/* ----------------------------------------------------------------------
 * escape quotes in strings
 */
String comEscapeQuotes(String str) {
  String s = NULL, cp = NULL, rt = "";
  
  if (! strchr(str, '\"')) {
    return str;
  }

  cp = strCopy(str);

  /* escape all quotes */
  while ((s = strchr(cp, '\"'))) {
     *s = '\0'; 
     rt = strAppend(rt, strPrintf("%s\\\"", cp), NULL);
     cp = s + 1;  
  }
  
  rt = strAppend(rt, cp, NULL);
  
  return rt;
}

/* ----------------------------------------------------------------------
 * write a preprocessor warning 
 */
void comWriteWarning(String format, ...)
{
  va_list ap;

  fprintf(com->ccFile, "\n#warning \"");
  
  va_start(ap, format);
  vfprintf(com->ccFile, format, ap);
  va_end(ap);

  if (com->currentConstraint) {
    fprintf(com->ccFile, " in constraint `%s'",
	comEscapeQuotes(com->currentConstraint->id));
  }

  fprintf(com->ccFile, "\"\n");
  com->needsIndent = TRUE;
}

/* ----------------------------------------------------------------------
 * write a preprocessor error 
 */
void comWriteError(String format, ...)
{
  va_list ap;

  fprintf(com->ccFile, "\n#error \"");
  
  va_start(ap, format);
  vfprintf(com->ccFile, format, ap);
  va_end(ap);

  if (com->currentConstraint) {
    fprintf(com->ccFile, " in constraint `%s'",
	comEscapeQuotes(com->currentConstraint->id));
  }

  fprintf(com->ccFile, "\"\n");
  com->needsIndent = TRUE;
}

/* ----------------------------------------------------------------------
 * no good c program without its own printf
 */
void comPrint(String format, ...)
{
  va_list ap;

  if (com->needsIndent) {
    fprintf(com->ccFile, "%s", com->indentString);
  } 
  va_start(ap, format);
  vfprintf(com->ccFile, format, ap);
  va_end(ap);
  
  com->needsIndent = FALSE;
}

/* ----------------------------------------------------------------------
 * string representation of a returntype
 */
String comReturnTypeToString(ReturnType returnType)
{
  switch ((unsigned long) returnType) {
    case RTBoolean: return "RTBoolean";
    case RTNumber: return "RTNumber";
    case RTString: return "RTString";
    case RTAVNode: return "RTAVNode";
    case RTList: return "RTList";
    case RTConjunction: return "RTConjunction";
    case RTDisjunction: return "RTDisjunction";
    case RTError: return "RTError";
    case RTLexemNode: return "RTModifieeLexemNode";
    case RTLexemNode | RTLexemPosition: return "RTModifierLexemNode";
    case RTGraphemNode: return "RTGraphemNode";
    case RTPeek: return "RTPeek";

    case RTBoolean | RTNoError: return "RTBoolean!";
    case RTNumber | RTNoError: return "RTNumber!";
    case RTString | RTNoError: return "RTString!";
    case RTAVNode | RTNoError: return "RTAVNode!";
    case RTList | RTNoError: return "RTList!";
    case RTConjunction | RTNoError: return "RTConjunction!";
    case RTDisjunction | RTNoError: return "RTDisjunction!";
    case RTError | RTNoError: return "RTError!";
    case RTLexemNode | RTNoError: return "RTModifieeLexemNode!";
    case RTLexemNode | RTNoError | RTLexemPosition: return "RTModifierLexemNode!";
    case RTGraphemNode | RTNoError: return "RTGraphemNode!";
    case RTPeek | RTNoError: return "RTPeek!";
  }
  return "horst";
}

/* ----------------------------------------------------------------------
 * string representation of a term type
 */
String comTermTypeToString(Term term)
{
  switch (term->type) {
    case TTTopPeek: return "TTTopPeek";
    case TTBottomPeek: return "TTBottomPeek";
    case TTLabel: return "TTLabel";
    case TTLevel: return "TTLevel";
    case TTAdd: return "TTAdd";
    case TTSubtract: return "TTSubtract";
    case TTMultiply: return "TTMultiply";
    case TTDivide: return "TTDivide";
    case TTFunction: return "TTFunction";
    case TTString: return "TTString";
    case TTNumber: return "TTNumber";
  }
  return "horst";
}

/* ----------------------------------------------------------------------
 * string representation of a formula type
 */
String comFormulaTypeToString(Formula formula) 
{
  switch (formula->type) {
    case FTEqual: return "FTEqual";
    case FTGreater: return "FTGreater";
    case FTLess: return "FTLess";
    case FTGreaterEqual: return "FTGreaterEqual";
    case FTLessEqual: return "FTLessEqual";
    case FTNotEqual: return "FTNotEqual";
    case FTConjunction: return "FTConjunction";
    case FTDisjunction: return "FTDisjunction";
    case FTImplication: return "FTImplication";
    case FTBiimplication: return "FTBiimplication";
    case FTPredicate: return "FTPredicate";
    case FTNegation: return "FTNegation";
    case FTConstant: return "FTConstant";
    case FTConnexion: return "FTConnexion";
    case FTDirection: return "FTDirection";
  }
  return "horst";
}

/* ----------------------------------------------------------------------
 * string representation of a value type
 */
String comValueTypeToString(Value value)
{
  switch(value->type) {
    case VTNumber: return "VTNumber";		
    case VTString: return "VTString";		
    case VTAVNode: return "VTAVNode";		
    case VTList: return "VTList";              
    case VTConjunction: return "VTConjunction";	
    case VTDisjunction: return "VTDisjunction";	
    case VTError: return "VTError";             
    case VTLexemNode: return "VTLexemNode";         
    case VTGraphemNode: return "VTGraphemNode";       
  }
  return "horst";
}

/* ----------------------------------------------------------------------
 * string representation of a connexion
 */
String comConnextionToString(Connexion connexion)
{
  switch (connexion) {
    case Top: return "Top";
    case Bottom: return "Bottom";
    case Over: return "Over";
    case Under: return "Under";
    case None: return "None";
    case Inverse: return "Inverse";
    case Identical: return "Identical";
    case AnyCon: return "AnyCon";
  }
  return "horst";
}

/* ----------------------------------------------------------------------
 * string representation of a direction
 */
String comDirectionToString(Direction direction)
{
  switch (direction) {
    case Left: return "Left";
    case Right: return "Right";
    case Nil: return "Nil"; 
    case NonNil: return "NonNil";
    case AnyDir: return "AnyDir";
  }
  return "horst";
}

/* ----------------------------------------------------------------------
 * write header 
 */
void comWriteHeader(void)
{
  time_t now = time((time_t *)NULL);
  struct passwd *pwp = getpwuid(getuid());

  comPrintln("/*");
  comPrintln(" * THIS IS A MACHINE GENERATED FILE. DO NOT CHANGE ANYTHING HERE SINCE");
  comPrintln(" * IT WILL BE OVERWRITTEN NEXT TIME.");
  comPrintln(" *");
  comPrintln(" * Compiled by %s (%s) on %s",
      pwp ? pwp->pw_name : "UNKNOWN",
      pwp ?  pwp->pw_gecos : "UNKNOWN", 
      ctime(&now));

  comPrintln(" */\n");
  comPrintln("/* -- INCLUDES ------------------------------------------------ */");
  comPrintln("#include <stdio.h>");
  comPrintln("#include <string.h>");
  comPrintln("#include <math.h>");
  comPrintln("#include <assert.h>");
  comPrintln("#include <blah.h>");
  comPrintln("#include \"cdg.h\"");
  comPrintln("#include \"command.h\"");
  comPrintln("#include \"compile.h\"");
  comPrintln("#include \"lexemgraph.h\"");
  comPrintln("#include \"eval.h\"");
  comPrintln("#include \"input.h\"");
  comPrintln("#include \"levelvalue.h\"");
  comPrintln("#include \"hook.h\"");
  comPrintln("#include \"predicates.h\"");
  comPrintln("#include \"functions.h\"");
  comPrintln("#include \"parsing.h\"");
  comPrintln("#include \"chunker.h\"");
}

/* ----------------------------------------------------------------------
 * write declaration section
 */
void comWriteDeclarations(void)
{
  int i;
  List l, m;
  Constraint constraint;


  comPrintln("");
  comPrintln("/* -- VARIABLES ----------------------------------------------- */");
  
  if(com->maxValues) {
    comPrint("static ValueStruct ");
  }

  for (i = 0; i < com->maxValues; i++) {
    comPrint("vs%d", i);
    if (i == com->maxValues - 1) {
      comPrintln(";");
    } else if (i%12 == 11) {
      comPrintln(",");
      comPrint("  ");
    } else {
      comPrint(", ");
    }
  }

  /* max lookup strings */
  comPrintln("static String *lookupStrings;");

  /* hierarchies array */
  comPrintln("typedef struct {");
  comIndent();
  comPrintln("String id;");
  comPrintln("Hierarchy hierarchy;");
  comOutdent();
  comPrintln("} HierarchyStubStruct;");
  comPrintln("typedef HierarchyStubStruct *HierarchyStub;");
  comPrintln("static HierarchyStubStruct hierarchies[%d];",
      hashSize(inputCurrentGrammar->hierarchies));
  comPrintln("static int maxHierachies = %d;", 
      hashSize(inputCurrentGrammar->hierarchies));

  /* constraints array */
  comPrintln("typedef struct {");
  comIndent();
  comPrintln("String id;");
  comPrintln("Constraint constraint;");
  comPrintln("BooleanFunction *cfunction;");
  comOutdent();
  comPrintln("} ConstraintStubStruct;");
  comPrintln("static ConstraintStubStruct constraints[%d];",
      hashSize(inputCurrentGrammar->constraints));
  comPrintln("typedef ConstraintStubStruct *ConstraintStub;");
  comPrintln("static int maxConstraints = %d;", 
      hashSize(inputCurrentGrammar->constraints));

  /* static globals */
  comPrintln("static Hierarchy hierarchy;");
  comPrintln("static String *strings;");
  comPrintln("static int a __attribute__ ((unused)), b __attribute__ ((unused)), c __attribute__ ((unused)), d __attribute__ ((unused));");
  comPrintln("static LevelValue lv __attribute__ ((unused));");
  comPrintln("static Constraint constraint __attribute__ ((unused));");
  comPrintln("static void *hasImplArgs[11] __attribute__ ((unused));");
    comPrintln("static Boolean bool __attribute__ ((unused));");
  comPrintln("");

  /* common forward declarations */
  comPrintln("Boolean %s(Input);", INIT_GRAMMAR);
  comPrintln("void %s(Input);", FINIT_GRAMMAR);
  comPrintln("");
  comPrintln("static HierarchyStub _findHierarchyStub(String id);");
  comPrintln("");
  comPrintln("static ValueType _matchList(String constraint, List candidates, String sort, Value result) __attribute__ ((unused));");
  comPrintln("static ValueType _matchString(String constraint, String sortA, String sortB, Value result) __attribute__ ((unused));");
  comPrintln("static Boolean _subsumes(String constraint, String sortA, String sortB) __attribute__ ((unused));");
  comPrintln("static Boolean _compatible(String constraint, String sortA, String sortB) __attribute__ ((unused));");
  comPrintln("static Hierarchy _findHierarchy(String id) __attribute__ ((unused));");
  comPrintln("static Connexion _getConnextion(LevelValue, LevelValue) __attribute__ ((unused));");
  comPrintln("static String  _lookup(String hashtableName) __attribute__ ((unused));");
  comPrintln("static Boolean  _connected(int a, int b, int nrlevels, int levelno, LexemGraph lg, Vector context) __attribute__ ((unused));");
  
  /* forward declarations of all constraint c-functions */
  comPrintln("/* forward declaration of unary constraints */");
  m = hashListOfKeys(inputCurrentGrammar->constraints);
  for (l = m; l != NULL; l = l->next) {
    constraint = hashGet(inputCurrentGrammar->constraints, l->item);
    if (constraint->sig->arity == 1) {
      comPrintln("Boolean _evalConstraint_%d(LexemGraph, Vector context, LevelValue);", constraint->no);
    }
  }
  comPrintln("/* forward declaration of binary constraints */");
  for (l = m; l != NULL; l = l->next) {
    constraint = hashGet(inputCurrentGrammar->constraints, l->item);
    if (constraint->sig->arity == 2) {
      comPrintln("Boolean _evalConstraint_%d(LexemGraph, Vector context, LevelValue, LevelValue);", constraint->no);
    }
  }
  listDelete(m);

  comPrintln("");
}

/* ----------------------------------------------------------------------
 * write helper functions
 */
void comWriteFunctions(void)
{
  /* 
   * _matchList 
   */
  comPrintln("");
  comPrintln("/* ------------------------------------------------------------");
  comPrintln(" * implement match(string, list, string), see funcMatch");
  comPrintln(" * returns VTNumber or VTError");
  comPrintln(" */");
  comPrintln("static ValueType _matchList(String constraint, List candidates, String sort, Value result) ");
  comPrintln("{");
  comIndent();
  comPrintln("int a, b;");
  comPrintln("List l;");
  comPrintln("Value value;");
  comPrintln("");
  comPrintln("b = (int)hashGet(hierarchy->types, sort) - 1;");
  comPrintln("if (b < 0) {");
  comIndent();
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: 2nd type `%%s' not found in hierarchy `%%s' in `matchList' in constraint `%%s'\\n\","
      "sort, hierarchy->id, constraint);");
  comPrintln("return VTError;");
  comOutdent();
  comPrintln("}");
  comPrintln("");
  comPrintln("result->data.number = 0.0; result->type = VTNumber;");
  comPrintln("for (l = candidates; l != NULL; l = l->next->next) {");
  comIndent();
  comPrintln("value = (Value)l->item;");
  comPrintln("a = (int)hashGet(hierarchy->types, value->data.string) - 1;");
  comPrintln("if (a < 0) {");
  comIndent();
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: type `%%s' not found in hierarchy `%%s' in `matchList' in constraint `%%s'\\n\","
      "value->data.string, hierarchy->id, constraint);");
  comPrintln("return VTError;");
  comOutdent();
  comPrintln("}");
  comPrintln("if ((int)arrayElement(hierarchy->distance, a, b) >= 0) {");
  comIndent();
  comPrintln("result->data.number = ((Value)l->next->item)->data.number;");
  comPrintln("break;");
  comOutdent();
  comPrintln("}");
  comOutdent();
  comPrintln("}");
  comPrintln("return VTNumber;");
  comOutdent();
  comPrintln("}");
  comPrintln("");
      
  /* 
   * _matchString 
   */
  comPrintln("");
  comPrintln("/* ------------------------------------------------------------");
  comPrintln(" * implement match(string, string, string), see funcMatch");
  comPrintln(" * returns VTNumber or VTError");
  comPrintln(" */");
  comPrintln("static ValueType _matchString(String constraint, String sortA, String sortB, Value result)");
  comPrintln("{");
  comIndent();
  comPrintln("int a, b;");
  comPrintln("");
  comPrintln("a = (int)hashGet(hierarchy->types, sortA) - 1;");
  comPrintln("if (a < 0) {");
  comIndent();
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: 1rst type `%%s' not found in hierarchy `%%s' in `matchString' in constraint `%%s'\\n\","
      "sortA, hierarchy->id, constraint);");
  comPrintln("return VTError;");
  comOutdent();
  comPrintln("}");
  comPrintln("");
  comPrintln("b = (int)hashGet(hierarchy->types, sortB) - 1;");
  comPrintln("if (b < 0) {");
  comIndent();
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: 2nd type `%%s' not found in hierarchy `%%s' in `matchString' in constraint `%%s'\\n\","
      "sortB, hierarchy->id, constraint);");
  comPrintln("return VTError;");
  comOutdent();
  comPrintln("}");

  comPrintln("");
  comPrintln("result->data.number = ((int)arrayElement(hierarchy->distance, a, b) >= 0) ? 1.0 : 0.0;");
  comPrintln("return VTNumber;");
  comOutdent();
  comPrintln("}");
  comPrintln("");
      
  /* 
   * _subsumes 
   */
  comPrintln("");
  comPrintln("/* ------------------------------------------------------------");
  comPrintln(" * implement subsumes(string, string, string), see predSubsumes");
  comPrintln(" * returns VTNumber or VTError");
  comPrintln(" */");
  comPrintln("static Boolean _subsumes(String constraint, String sortA, String sortB)");
  comPrintln("{");
  comIndent();
  comPrintln("int a, b;");
  comPrintln("");
  comPrintln("a = (int)hashGet(hierarchy->types, sortA) - 1;");
  comPrintln("if (a < 0) {");
  comIndent();
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: 1rst type `%%s' not found in hierarchy `%%s' in `subsumes' in constraint `%%s'\\n\","
      "sortA, hierarchy->id, constraint);");
  comPrintln("return FALSE;");
  comOutdent();
  comPrintln("}");
  comPrintln("");
  comPrintln("b = (int)hashGet(hierarchy->types, sortB) - 1;");
  comPrintln("if (b < 0) {");
  comIndent();
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: 2nd type `%%s' not found in hierarchy `%%s' in `subsumes' in constraint `%%s'\\n\","
      "sortB, hierarchy->id, constraint);");
  comPrintln("return FALSE;");
  comOutdent();
  comPrintln("}");
  comPrintln("return ((int)arrayElement(hierarchy->distance, a, b) >= 0);");
  comOutdent();
  comPrintln("}");
  comPrintln("");
      
  /* 
   * _compatible 
   */
  comPrintln("");
  comPrintln("/* ------------------------------------------------------------");
  comPrintln(" * implement compatible(string, string, string), see predCompatible");
  comPrintln(" * returns VTNumber or VTError");
  comPrintln(" */");
  comPrintln("static Boolean _compatible(String constraint, String sortA, String sortB)");
  comPrintln("{");
  comIndent();
  comPrintln("int a, b;");
  comPrintln("");
  comPrintln("a = (int)hashGet(hierarchy->types, sortA) - 1;");
  comPrintln("if (a < 0) {");
  comIndent();
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: 1rst type `%%s' not found in hierarchy `%%s' in `compatible' in constraint `%%s'\\n\","
      "sortA, hierarchy->id, constraint);");
  comPrintln("return FALSE;");
  comOutdent();
  comPrintln("}");
  comPrintln("");
  comPrintln("b = (int)hashGet(hierarchy->types, sortB) - 1;");
  comPrintln("if (b < 0) {");
  comIndent();
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: 2nd type `%%s' not found in hierarchy `%%s' in `compatible' in constraint `%%s'\\n\","
      "sortB, hierarchy->id, constraint);");
  comPrintln("return FALSE;");
  comOutdent();
  comPrintln("}");
  comPrintln("return (((int)arrayElement(hierarchy->distance, a, b) >= 0) ||");
  comPrintln("        ((int)arrayElement(hierarchy->distance, b, a) >= 0));");
  comOutdent();
  comPrintln("}");
  comPrintln("");
      
  /* 
   * _findHierarchyStub
   */
  comPrintln("");
  comPrintln("/* ------------------------------------------------------------");
  comPrintln(" * returns the named hierarchy stub, NULL if not found");
  comPrintln(" */");
  comPrintln("static HierarchyStub _findHierarchyStub(String id)");
  comPrintln("{");
  comIndent();
  comPrintln("int i;");
  comPrintln("");
  comPrintln("for (i = 0; i < maxHierachies; i++) {");
  comIndent();
  comPrintln("if (strcmp(id, hierarchies[i].id) == 0) {");
  comIndent();
  comPrintln("return &hierarchies[i];");
  comOutdent();
  comPrintln("}");
  comOutdent();
  comPrintln("}");
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: hierarchy `%%s' not found in `findHierarchy'\\n\", id);");
  comPrintln("return NULL;");
  comOutdent();
  comPrintln("}");
  comPrintln("");

  /* 
   * _findHierarchy
   */
  comPrintln("");
  comPrintln("/* ------------------------------------------------------------");
  comPrintln(" * returns the named hierarchy, NULL if not found");
  comPrintln(" */");
  comPrintln("static Hierarchy _findHierarchy(String id)");
  comPrintln("{");
  comIndent();
  comPrintln("HierarchyStub hierarchyStub = _findHierarchyStub(id);");
  comPrintln("return hierarchyStub?hierarchyStub->hierarchy:NULL;");
  comOutdent();
  comPrintln("}");
  comPrintln("");

  /* 
   * _getConnextion
   */
  comPrintln("");
  comPrintln("/* ------------------------------------------------------------");
  comPrintln(" * returns the connexion of two lvs and caches the result");
  comPrintln(" */");
  comPrintln("static Connexion _getConnextion(LevelValue lv1, LevelValue lv2)");
  comPrintln("{");
  comIndent();
  comPrintln("static Connexion connexion;");
  comPrintln("static LevelValue currentLv1 = NULL, currentLv2 = NULL;");
  comPrintln("if (lv1 != currentLv1 || lv2 != currentLv2) {");
  comIndent();
  comPrintln("currentLv1 = lv1;");
  comPrintln("currentLv2 = lv2;");
  comPrintln("connexion = getCon(lv1, lv2);");
  comOutdent();
  comPrintln("}");
  comPrintln("return connexion;");
  comOutdent();
  comPrintln("}");
  comPrintln("");

  /* 
   * _lookup
   */
  comPrintln("");
  comPrintln("/* ------------------------------------------------------------");
  comPrintln(" * returns the string result of the hashtable lookup");
  comPrintln(" */");
  comPrintln("static String  _lookup(String hashtableName)");
  comPrintln("{");
  comIndent();
  comPrintln("Hashtable hashtable;");
  comPrintln("Value result;");
  comPrintln("String key =  NULL;");
  comPrintln("int i = 0;"); 
  comPrintln("");
  comPrintln("if (lookupStrings[0]) {");
  comPrintln("	hashtable = hashGet(inputCurrentGrammar->maps, hashtableName);");
  comPrintln("	if (hashtable) {");
  comPrintln("	  for (i = 0; lookupStrings[i]; i++) {");
#ifdef DEBUG_LOOKUP
  comPrintln("	    cdgPrintf(CDG_DEBUG, \"LookupStrings[%%d] = %%s\\n\", i, lookupStrings[i]);");
#endif
  comPrintln("	    key = key ? strAppend(key, strRegister(\":\"), lookupStrings[i], NULL) : lookupStrings[i] ;");
  comPrintln("	  }");
#ifdef DEBUG_LOOKUP
  comPrintln("	  cdgPrintf(CDG_DEBUG, \"key = %%s\\n\", key);");
#endif
  comPrintln("	  if ((result = hashGet(hashtable, key))) {");
#ifdef DEBUG_LOOKUP
  comPrintln("	    cdgPrintf(CDG_DEBUG, \"Lookup result = %%s\\n\", result->data.string);");
#endif
  comPrintln("	    return result->data.string;");
  comPrintln("	  }");
  comPrintln("	}else {");
  comPrintln("	  cdgPrintf(CDG_WARNING, \"WARNING: Map %%s not found\\n\", hashtableName);");
  comPrintln("	}");
  comPrintln("}");
  comPrintln("return strRegister(\"\");");
  comOutdent();
  comPrintln("}");

  /*
   * _connected
   */
  comPrintln("");
  comPrintln("/* ------------------------------------------------------------");
  comPrintln(" * returns TRUE if a and b are part of the same tree");
  comPrintln(" */");
  comPrintln("static Boolean _connected(int a, int b, int nrlevels, int levelno, LexemGraph lg, Vector context)");
  comPrintln("{");
  comPrintln("  int count = 0, roota = -1, rootb = -1;");
  comPrintln("  LevelValue lv;");
  comPrintln("  int max = lg->max - lg->min;");
  comPrintln("  ");
  comPrintln("  lv = vectorElement(context, nrlevels * a + levelno);");
  comPrintln("  while(lv) {");
  comPrintln("    if(!spec(lv->regents)) {");
  comPrintln("      roota = ((LexemNode)(lv->dependents->item))->arc->from;");
  comPrintln("      break;");
  comPrintln("    }");
  comPrintln("    else if(NONSPEC == lv->regents) {");
  comPrintln("      return FALSE;");
  comPrintln("    }");
  comPrintln("    else if(count++ > max) {");
  comPrintln("      roota = a;");
  comPrintln("      break;");
  comPrintln("    }");
  comPrintln("    else {");
  comPrintln("      lv = vectorElement(context, nrlevels * ((LexemNode)(lv->regents->item))->arc->from + levelno);");
  comPrintln("    }");
  comPrintln("  }");
  comPrintln("  ");
  comPrintln("  count = 0;");
  comPrintln("  lv = vectorElement(context, nrlevels * b + levelno);");
  comPrintln("  while(lv) {");
  comPrintln("    if(!spec(lv->regents)) {");
  comPrintln("      rootb = ((LexemNode)(lv->dependents->item))->arc->from;");
  comPrintln("      break;");
  comPrintln("    }");
  comPrintln("    else if(NONSPEC == lv->regents) {");
  comPrintln("      return FALSE;");
  comPrintln("    }");
  comPrintln("    else if(count++ > max) {");
  comPrintln("      rootb = b;");
  comPrintln("      break;");
  comPrintln("    }");
  comPrintln("    else {");
  comPrintln("      lv = vectorElement(context, nrlevels * ((LexemNode)(lv->regents->item))->arc->from + levelno);");
  comPrintln("    }");
  comPrintln("  }");
  comPrintln("  ");
  comPrintln("  return roota == rootb;");
  comPrintln("}");

  /* initGrammar and finitGrammar */
  comWriteInitFunction();
  comWriteFinitFunction();
  assert(com->indent == 0);

}

/* ----------------------------------------------------------------------
 * write the grammar finalization function
 */
void comWriteFinitFunction(void)
{
  comPrintln("/* ------------------------------------------------------------");
  comPrintln(" * finalize the dll grammar");
  comPrintln(" */");
  comPrintln("void %s(Input thisInput)", FINIT_GRAMMAR);
  comPrintln("{");
  comIndent();
  comPrintln("int i;");
  comPrintln("Constraint constraint;");
  comPrintln("HashIterator hi;");

  /* free the strings */
  comPrintln("for (i = 0; i < %d; i++) {", vectorSize(com->strings));
  comIndent();
  comPrintln("cdgFreeString(strings[i]);");
  comOutdent();
  comPrintln("}");
  comPrintln("memFree(strings);");

  /* reset unlink the c constraints */
  comPrintln("hi = hashIteratorNew(thisInput->constraints);");
  comPrintln("while (NULL != (constraint = (Constraint)hashIteratorNextValue(hi))) {");
  comIndent();
  comPrintln("constraint->cfunction = NULL;");
  comOutdent();
  comPrintln("}");
  comPrintln("hashIteratorDelete(hi);");

  comOutdent();
  comPrintln("}");
  comPrintln("");
}
/* ----------------------------------------------------------------------
 * write the grammar initialization function
 */
void comWriteInitFunction(void)
{
  Hierarchy hierarchy;
  Constraint constraint;
  int i;
  HashIterator hi;
  
  comPrintln("/* ------------------------------------------------------------");
  comPrintln(" * initialize the dll grammar");
  comPrintln(" */");
  comPrintln("Boolean %s(Input thisInput)", INIT_GRAMMAR);
  comPrintln("{");
  comIndent();
  comPrintln("int i;");
  comPrintln("Hierarchy hierarchy = NULL;");
  comPrintln("Constraint constraint = NULL;");
      
  /* remember the correct ids of the hierarchies */
  for (i = 0, hi = hashIteratorNew(inputCurrentGrammar->hierarchies);
       NULL != (hierarchy = hashIteratorNextValue(hi));
       i++) {
    comPrintln("hierarchies[%d].id = strRegister(\"%s\");", i, 
	comEscapeQuotes(hierarchy->id));
    comPrintln("hierarchies[%d].hierarchy = NULL;", i);
  }
  hashIteratorDelete(hi);
  comPrintln("");

  /* remember the correct ids of the constraints */
  for (i = 0, hi = hashIteratorNew(inputCurrentGrammar->constraints);
       NULL != (constraint = hashIteratorNextValue(hi));
       i++) {
    comPrintln("constraints[%d].id = strRegister(\"%s\");", i, 
	comEscapeQuotes(constraint->id));
    comPrintln("constraints[%d].constraint = NULL;", i);
    comPrintln("constraints[%d].cfunction = &_evalConstraint_%d;", i, constraint->no);
  }
  hashIteratorDelete(hi);
  comPrintln("");

  /* registration of static strings */
  comPrintln("strings = (String *)memMalloc(sizeof(String) * %d);",
      vectorSize(com->strings));
  for (i = 0; i < vectorSize(com->strings); i++) {
    String string = (String)vectorElement(com->strings, i);
    comPrintln("strings[%d] = strRegister(\"%s\");", i, comEscapeQuotes(string));
  }

  /* allocate memory for the lookup strings */
  comPrintln("lookupStrings = (String *)memMalloc(sizeof(String) * %d);",
      (com->maxLookupStrings + 1));

  /* manual runtime linkage of the hierarchies */
  comPrintln("for (i = 0; i < maxHierachies; i++) {");
  comIndent();
  comPrintln("if (! (hierarchy = (Hierarchy)findHierarchy(hierarchies[i].id))) {");
  comIndent();
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: cannot find hierarchy `%%s'\\n\", hierarchies[i].id);");
  comPrintln("return FALSE;");
  comOutdent();
  comPrintln("}");

  comPrintln("hierarchies[i].hierarchy = hierarchy;");
  comOutdent();
  comPrintln("}");
  
  /* manual runtime linkage of the constraints */
  comPrintln("for(i = 0; i < maxConstraints; i++) {");
  comIndent();
  comPrintln("if (! (constraint = (Constraint)hashGet(thisInput->constraints, constraints[i].id))) {");
  comIndent();
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: cannot find constraint `%%s'\\n\", constraints[i].id);");
  comPrintln("return FALSE;");
  comOutdent();
  comPrintln("}");

  comPrintln("constraints[i].constraint = constraint;");
  comPrintln("if (constraint->cfunction != NULL)");
  comIndent();
  comPrintln("cdgPrintf(CDG_WARNING, \"WARNING: constraint `%%s' has already been compiled\\n\", constraint->id);");
  comOutdent();
  comPrintln("constraint->cfunction = constraints[i].cfunction;");
  comOutdent();
  comPrintln("}");

  comPrintln("return TRUE;");
  
  comOutdent();
  comPrintln("}");
  comPrintln("");
}

/* ----------------------------------------------------------------------
 * get the index of a hierarchy
 */
int comIndexOfHierarchy(String id)
{
  int i;
  Hierarchy hierarchy;
  HashIterator hi;
  
  for (i = 0, hi = hashIteratorNew(inputCurrentGrammar->hierarchies);
       NULL != (hierarchy = hashIteratorNextValue(hi));
       i++) {
    if (strcmp(hierarchy->id, id) == 0) {
      return i;
    }
  }

  comWriteError("unknown hierarchy `%s'", id);
  return -1;
}

/* ----------------------------------------------------------------------
 * get the index of a constraint
 */
int comIndexOfConstraint(String id)
{
  int i;
  Constraint constraint;
  HashIterator hi;
  
  for (i = 0, hi = hashIteratorNew(inputCurrentGrammar->constraints);
       NULL != (constraint = hashIteratorNextValue(hi));
       i++) {
    if (strcmp(constraint->id, id) == 0) {
      return i;
    }
  }

  comWriteError("unknown constraint `%s'", id);
  return -1;
}

/* ----------------------------------------------------------------------
 * get index of assigned variable 
 */
int comIndexOfVarInfo(VarInfo varInfo)
{
  int index = listIndex(com->currentConstraint->vars, varInfo);

  assert(index > 0);
  
  return index;
}

/* ----------------------------------------------------------------------
 * register a string to be allocated statically later
 * returns the index in strings
 */
int comRegisterString(String string)
{
  int i;
  String oldString;
  String newString = strRegister(string);

  
  for (i = 0; i < vectorSize(com->strings); i++) {
    oldString = (String)vectorElement(com->strings, i);
    if (oldString == newString) {
      return i;
    }
  }

  return vectorAddElement(com->strings, newString);
}

/* ----------------------------------------------------------------------
 * translate a ^id and \@id
 */
void comTranslateLexemNodeAccess(Term term, int tmpIndex)
{
  String position = "horst";
  String pathItem;

  /* used in parent() */
  if (term->type == TTFunction) {
    comPrint("((LexemNode)vs%d.data.lexemnode)", tmpIndex);
    return;
  }

  /* security checks */
  if (term->type == TTBottomPeek) {
    position = "dependent";
  } else if (term->type == TTTopPeek) {
    position = "regent";
  }
  else { /* never reach */
    comWriteError("comTranslateLexemNodeAccess called with invalid term argument");
    return;
  } 

  pathItem = (String) term->data.peek.path->item; 
  if (strcmp("id", pathItem) != 0) {
    comWriteError("comTranslateLemNodeAccess called with invalid path item `%s'",
	pathItem);
    return;
  }

  comPrint("((LexemNode)(lv%d->%ss->item))", 
      comIndexOfVarInfo(term->data.peek.varinfo), position);
}

/* ----------------------------------------------------------------------
 * translate a lexical access
 */
void comTranslateLexicalAccess(Term term)
{
  String position = "horst";
  String pathItem;

  /* security checks */
  if (term->type == TTBottomPeek) {
    position = "dependent";
  } else if (term->type == TTTopPeek) {
    position = "regent";
  } else { /* never reach */
    comWriteError("comTranslateLexicalAccess called with invalid term argument");
    return;
  }
    
  pathItem = (String) term->data.peek.path->item; 
  if (strcmp("id", pathItem) == 0 ||
      strcmp("from", pathItem) == 0 ||
      strcmp("to", pathItem) == 0 ||
      strcmp("word", pathItem) == 0) {
    comWriteError("comTranslateLexicalAccess called with invalid path item `%s'",
	pathItem);
    return;
  }
    
  comPrint("((LexemNode)(lv%d->%ss->item))->lexem->values[%d/*%s*/]", 
       comIndexOfVarInfo(term->data.peek.varinfo),
       position, term->data.peek.pathindex, sprintPath(term->data.peek.path));
}

/* ----------------------------------------------------------------------
 * translate a function to c code
 */
void comTranslateFunction(String functor, List args, int tmpIndex)
{
  comPrintln("( /* %s */ ", functor);
  comIndent();
  
  /* distance */
  if (strcmp("distance", functor) == 0) {
    comTranslateDistance(args, tmpIndex);
  } 
  
  /* abs */
  else if (strcmp("abs", functor) == 0) {
    comTranslateAbs(args, tmpIndex);
  } 
  
  /* min, max */
  else if (strcmp("min", functor) == 0 ||
           strcmp("max", functor) == 0) {
    comTranslateMinMax(functor, args, tmpIndex);
  } 
  
  /* match */
  else if (strcmp("match", functor) == 0) {
    comTranslateMatch(args, tmpIndex);
  } 

  /* phrasequotes */
  else if (strcmp("phrasequotes", functor) == 0) {
    comTranslatePhrasequotes(args, tmpIndex);
  } 

  /* lookup  */
  else if (strcmp("lookup", functor) == 0) {
    comTranslateLookup(args, tmpIndex);
  } 

  /* quotes  */
  else if (strcmp("quotes", functor) == 0) {
    comTranslateQuotes(args, tmpIndex);
  } 

  /* parens  */
  else if (strcmp("parens", functor) == 0) {
    comTranslateParens(args, tmpIndex);
  } 

  /* parent  */
  else if (strcmp("parent", functor) == 0) {
    comTranslateParent(args, tmpIndex);
  } 
  
  /* pts  */
  else if (strcmp("pts", functor) == 0) {
    comTranslatePts(args, tmpIndex);
  } 

  /* height  */
  else if (strcmp("height", functor) == 0) {
    comTranslateHeight(args, tmpIndex);
  } 

  /* never reach */
  else {
    comWriteError("compile.c:%d: unknown function `%s'", __LINE__, functor);
  }

  
  comOutdent();
  comPrint(")");
}

/* ----------------------------------------------------------------------
 * translate the distance function
 */
void comTranslateDistance(List args, int tmpIndex)
{
  Term term1 = (Term)listNthElement(args, 1);
  Term term2 = (Term)listNthElement(args, 2);
  ReturnType returnType1 = comReturnTypeOfTerm(term1) & ~RTNoError;
  ReturnType returnType2 = comReturnTypeOfTerm(term2) & ~RTNoError;

  comTranslateTerm(term1, FALSE, TRUE, tmpIndex);
  comTranslateTerm(term2, FALSE, TRUE, tmpIndex+1);

  switch(returnType1 & ~RTLexemPosition) {
    case RTLexemNode:
      comPrint("(");
      if (returnType1 & RTLexemPosition) {
	comPrint("(");
	comTranslateLexemNodeAccess(term1, tmpIndex);
	comPrintln(" == NULL) ? (vs%d.data.number = 0.0):", tmpIndex);
      }
      switch(returnType2 & ~RTLexemPosition) {
	case RTLexemNode:
	  comPrint("(");
	  if (returnType2 & RTLexemPosition) {
	    comPrint("(");
	    comTranslateLexemNodeAccess(term2, tmpIndex);
	    comPrintln(" == NULL) ? (vs%d.data.number = 0.0):", tmpIndex);
	  }
	  comPrint("(vs%d.data.number = lgDistanceOfNodes(lg, ", tmpIndex);
	  comTranslateLexemNodeAccess(term1, tmpIndex);
	  comPrint(", ");
	  comTranslateLexemNodeAccess(term2, tmpIndex);
	  comPrintln(")))),");
	  comPrintln("(vs%d.type = VTNumber)", tmpIndex);
	  break;
	default: /* never reach */
	  comWriteError("illegal return type `%s' in `distance'",
	      comReturnTypeToString(returnType2));
      }
      break;
    default: /* never reach */
      comWriteError("illegal return type `%s' in `distance'",
	  comReturnTypeToString(returnType1));
  }
}

/* ----------------------------------------------------------------------
 * translate the abs function
 */
void comTranslateAbs(List args, int tmpIndex)
{
  Term term = (Term)args->item;
  ReturnType returnType = comReturnTypeOfTerm(term);

  comTranslateTerm(term, FALSE, TRUE, tmpIndex);
  returnType &= ~RTNoError;
  switch(returnType) {
    case RTPeek:
      comPrintln(" != VTNumber?", tmpIndex);
      comPrintln("(cdgPrintf(CDG_ERROR, \"ERROR: type mismatch in `abs'\"), vs%d.type = VTError) :",
	  tmpIndex);
      comPrint("((vs%d.data.number = fabs(", tmpIndex);
      comTranslateLexicalAccess(term);
      comPrintln("->data.number)), vs%d.type = VTNumber))", tmpIndex);
      break;
    case RTNumber:
      comPrint("((vs%d.data.number = fabs(", tmpIndex); 
      comTranslateNumber(term, tmpIndex);
      comPrintln(")), vs%d.type = VTNumber)", tmpIndex);
      break;
    default: /* never reach */
      comWriteError("illegal return type `%s' in `abs'",
	  comReturnTypeToString(returnType));
  }
}

/* ----------------------------------------------------------------------
 * translate the phrasequotes function
 */
void comTranslatePhrasequotes(List args, int tmpIndex)
{
  Term term = (Term)args->item;
  ReturnType returnType = comReturnTypeOfTerm(term);
  
  /* top peek */
  comPrint("(! ");
  comTranslateLexemNodeAccess(term, tmpIndex);
  if (returnType & RTLexemPosition) {
    comPrint(" || ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrint(" == NONSPEC");
  }
  comPrintln(") ?");
  comPrintln("cdgPrintf(CDG_ERROR,\"ERROR: argument of 'phrasequotes' isn't a lexem node\"), (vs%d.type = VTError) :", tmpIndex);
  comPrint("(vs%d.data.number = ", tmpIndex);
  comTranslateLexemNodeAccess(term, tmpIndex);
  comPrint("->arc->phraseQuotes), (");
  comPrint("vs%d.type = VTNumber)", tmpIndex);
}

/* ----------------------------------------------------------------------
 * translate the lookup function
 */
void comTranslateLookup(List args, int tmpIndex)
{
  List l;
  ReturnType returnType;
  Term term;
  int i = 0;
  Hashtable hashtable;
  String hashtableName;
  
  term = (Term)args->item;
  returnType = comReturnTypeOfTerm(term) & ~RTNoError;
  hashtableName = term->data.string;

  /* find map */
  if ((hashtable = (hashGet(inputCurrentGrammar->maps, hashtableName)))) {
    for (l = args->next; l; l = l->next) {
      term = (Term)l->item;
      returnType = comReturnTypeOfTerm(term) & ~RTNoError;

      comPrintln("(lookupStrings[%d] = (", i++);
      comTranslateTerm(term, FALSE, FALSE, tmpIndex);
      if (returnType & RTString) {
	if (term->type != TTString) {
	  comPrintln("!= VTString ?");
	  comPrintln("(cdgPrintf(CDG_WARNING, \"WARNING: Type Mismatch (arg %d) in lookup in constraint %s\\n\"), NULL) :",
	      i + 1, comEscapeQuotes(com->currentConstraint->id));
	}
	comTranslateString(term, tmpIndex);
      } else if (returnType & RTPeek) {
	comPrintln("!= VTString ?");
	comPrintln("(cdgPrintf(CDG_WARNING, \"WARNING: Type Mismatch (arg %d) in lookup in constraint %s\\n\"), NULL) :",
	    i + 1, comEscapeQuotes(com->currentConstraint->id));
	comTranslateLexicalAccess(term);
	comPrint("->data.string");
      } else {
	abort(); // never reach
      }
      comPrintln(")),");
    }
	
    comPrintln("(lookupStrings[%d] = NULL),", i);
    comPrintln("(vs%d.data.string = _lookup(\"%s\"))", tmpIndex, hashtableName);
  } else {
    comPrintln("vs%d.data.string = strRegister(\"\"), vs%d.type = VTString",
	tmpIndex, tmpIndex);
  }
}

/* ----------------------------------------------------------------------
 * translate the quotes function
 */
void comTranslateQuotes(List args, int tmpIndex)
{
  Term term = (Term)args->item;
  ReturnType returnType = comReturnTypeOfTerm(term);
  
  comPrint("! ");
  comTranslateLexemNodeAccess(term, tmpIndex);
  if (returnType & RTLexemPosition) {
      comPrint(" || ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrint(" == NONSPEC");
  }
    comPrintln(" ?");
    comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: argument of function 'quotes' isn't a lexem node \"),");
    comPrintln("(vs%d.type = VTError) :", tmpIndex);
    comPrint("(vs%d.data.number = ", tmpIndex);
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrintln("->arc->quotes),");
    comPrintln("(vs%d.type = VTNumber)");
}

/* ----------------------------------------------------------------------
 * translate the parens function
 */
void comTranslateParens(List args, int tmpIndex)
{
  Term term = (Term)args->item;
  ReturnType returnType = comReturnTypeOfTerm(term);
  
  comPrint("! ");
  comTranslateLexemNodeAccess(term, tmpIndex);
  if (returnType & RTLexemPosition) {
      comPrint(" || ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrint(" == NONSPEC");
  }
    comPrintln(" ?");
    comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: argument of function 'parens' isn't a lexem node \"),");
    comPrintln("(vs%d.type = VTError) :", tmpIndex);
    comPrint("(vs%d.data.number = ", tmpIndex);
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrintln("->arc->parens),");
    comPrintln("(vs%d.type = VTNumber)");
}
    
/* ----------------------------------------------------------------------
 * translate the parent function
 */
void comTranslateParent(List args, int tmpIndex)
{
  Term term = (Term) args->item;
  ReturnType returnType = comReturnTypeOfTerm(term);
  int nrlevels = inputCurrentGrammar->noOfLevels;
  VarInfo vi = com->currentConstraint->vars->item;

  comPrint("! ");
  comTranslateLexemNodeAccess(term, tmpIndex);
  if (returnType & RTLexemPosition) {
    comPrint(" || ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrint(" == NONSPEC");
  }
  comPrintln(" ?");
  comPrintln("(cdgPrintf(CDG_ERROR, \"ERROR: argument of function 'parent' isn't a lexem node \"),");
  comPrintln("(vs%d.type = VTError)) :", tmpIndex);

  comPrint("((vs%d.data.number = funcParentImpl(lg, ", tmpIndex);
  comPrint("vectorElement(context, %d * ", nrlevels);
  comTranslateLexemNodeAccess(term, tmpIndex);
  comPrint("->arc->from + %d))),", vi->level->no);
  comPrintln("(vs%d.type = VTLexemNode))");
}

/* ----------------------------------------------------------------------
 * translate the pts function
 */
void comTranslatePts(List args, int tmpIndex)
{
  Term term = (Term)args->item;

  comPrint("! ");
  comTranslateLexemNodeAccess(term, tmpIndex);
  comPrintln(" ?");
  comPrintln("cdgPrintf(CDG_ERROR, \"ERROR: argument of function 'quotes' isn't a lexem node \"),");
  comPrintln("(vs%d.type = VTError) :", tmpIndex);
  comPrint("(vs%d.data.number = lgTagscore(", tmpIndex);
  comTranslateLexemNodeAccess(term, tmpIndex);
  comPrintln(")),");
  comPrintln("(vs%d.type = VTNumber)");

}

/* ----------------------------------------------------------------------
 * translate the `height' function
 */
void comTranslateHeight(List args, int tmpIndex)
{
  Term term = (Term) args->item;
  Level level;

  if (NULL == (level = findLevel(term->data.peek.varinfo->levelname))) {
    comWriteError("ERROR: level `%s' not found!", term->data.peek.varinfo->levelname);
    return;
  }

  comPrint("! ");
  comTranslateLexemNodeAccess(term, tmpIndex);
  comPrintln(" ?");
  comPrintln("(cdgPrintf(CDG_ERROR, \"ERROR: argument of function 'height' isn't a lexem node \"),");
  comPrintln("(vs%d.type = VTError)) :", tmpIndex);
  comPrint("((vs%d.data.number = ", tmpIndex);

  comPrint("funcHeightImpl(");
  comTranslateLexemNodeAccess(term, tmpIndex);
  comPrint(", %d, lg, context)),", level->no);
  comPrintln("(vs%d.type = VTNumber))");

}

/* ----------------------------------------------------------------------
 * translate the min and max function
 */
void comTranslateMinMax(String functor, List args, int tmpIndex)
{
  List l;
  Boolean isFirst = TRUE;
  Term term = (Term)args->item;
  Term lastTerm;
  ReturnType returnType = comReturnTypeOfTerm(term) & ~RTNoError;
  int lastTmpIndex;

  comTranslateTerm(term, FALSE, TRUE, tmpIndex);

  /* peek return type */
  if (returnType == RTPeek) {
    comPrintln(" != VTNumber ?");
    comPrintln("(cdgPrintf(CDG_ERROR,\"ERROR: 1rst argument of `%s' isn't numerical\"),", functor);
    comPrintln("(vs%d.type == VTError)) :", tmpIndex);
  } 

  comPrintln("(vs%d.type = VTNumber, ", tmpIndex);

  /* loop thru all argument terms */
  lastTmpIndex = tmpIndex;
  for (l = args->next; l != NULL; l = l->next) {
    lastTerm = term;
    term = (Term)l->item;
    returnType = comReturnTypeOfTerm(term) & ~RTNoError;
    if (isFirst) {
      isFirst = FALSE;
    } else {
      comPrintln(",");
    }
    comTranslateTerm(term, FALSE, TRUE, tmpIndex+1);

    /* peek return type */
    if (returnType == RTPeek) {
      comPrintln(" != VTNumber ?");
      comPrintln("(cdgPrintf(CDG_ERROR,\"ERROR: 1rst argument of `%s' isn't numerical\"),", functor);
      comPrintln("(vs%d.type == VTError)) :", tmpIndex);
    }
    /*  granted to be RTNumber */
    comPrint("(vs%d.data.number = %s(", tmpIndex, functor); 
    comTranslateNumber(lastTerm, lastTmpIndex);
    comPrint(", ");
    comTranslateNumber(term, tmpIndex+1);
    comPrintln("))");

    lastTmpIndex = tmpIndex+1;
  }

  comPrintln(")");
}

/* ----------------------------------------------------------------------
 * translate the match function
 */
void comTranslateMatch(List args, int tmpIndex)
{
  int i;
  Term term = (Term)listNthElement(args, 1);
  Term termA = (Term)listNthElement(args, 2);
  Term termB = (Term)listNthElement(args, 3);
  ReturnType returnType1 = comReturnTypeOfTerm(term) & ~RTNoError;
  ReturnType returnType2 = comReturnTypeOfTerm(termA) & ~RTNoError;
  ReturnType returnType3 = comReturnTypeOfTerm(termB) & ~RTNoError;

  if (term->type != TTString) {
    comTranslateTerm(term, FALSE, TRUE, tmpIndex);
  }

  /* check the hierarchy */
  if (returnType1 == RTString) {

    /* extra rule for static strings */
    if (term->type == TTString) {
      Hierarchy hierarchy =
        hashGet(inputCurrentGrammar->hierarchies, term->data.string);
      if (!hierarchy) {
	comWriteError("hierarchy `%s' not found");
      } else {
        i = comIndexOfHierarchy(hierarchy->id);
	comPrintln("(hierarchy = hierarchies[%d].hierarchy),", i);
      }
    } 
    
    /* its a string but not provided staticaly */
    else { 
      comPrint("(hierarchy = _findHierarchy("); 
      comTranslateString(term, tmpIndex);
      comPrintln(")),");
      comPrintln("(hierarchy == NULL)?");
      comPrintln("(vs%d.type = VTError) :", tmpIndex);
    }
  } 

  /* hierarchy lexically encoded */
  else if (returnType1 == RTPeek) { 
    comPrintln(" != VTString ?", tmpIndex);
    comPrint("(cdgPrintf(CDG_ERROR, \"ERROR: type mismatch of 1rst argument in `match'\\n\"), ");
    comPrintln("(vs%d.type = VTError)) :", tmpIndex);
    comPrint("(hierarchy = _findHierarchy(");
    comTranslateLexicalAccess(term);
    comPrintln("->data.string)),", tmpIndex);
    comPrintln("(hierarchy == NULL)?");
    comPrintln("(vs%d.type = VTError) :", tmpIndex);
  }

  /* hierarchy unknown at compile time */
  else { 
    comPrintln("((vs%d.type != VTString)?", tmpIndex);
    comPrint("(cdgPrintf(CDG_ERROR, \"ERROR: type mismatch of 1rst argument in `match'\\n\"), ");
    comPrintln("(vs%d.type = VTError)) :", tmpIndex);
    comPrint("(hierarchy = _findHierarchy(");
    comTranslateString(term, tmpIndex);
    comPrintln("))),");
    comPrintln("(hierarchy == NULL)?");
    comPrintln("(vs%d.type = VTError) :", tmpIndex);
  }

  
  /* check termB */
  if (returnType3 & RTPeek) {

    comTranslateTerm(termB, FALSE, TRUE, tmpIndex+1);
    comPrintln(" != VTString ? (vs%d.type = VTError) :", tmpIndex);

    comTranslateTerm(termA, FALSE, TRUE, tmpIndex);
    /* check termA */
    if (returnType2 == RTPeek) {
      comPrintln(" != VTString ?");
      comPrint("((");
      comTranslateLexicalAccess(termA);
      comPrintln("->type != VTList)? (vs%d.type = VTError):", tmpIndex);
      comPrint("(vs%d.type = _matchList(\"%s\", ", tmpIndex, comEscapeQuotes(com->currentConstraint->id));
      comTranslateLexicalAccess(termA);
      comPrint("->data.list, ");
      comTranslateLexicalAccess(termB);
      comPrintln("->data.string, &vs%d))) :", tmpIndex);
      comPrint("(vs%d.type = _matchString(\"%s\", ", tmpIndex, comEscapeQuotes(com->currentConstraint->id));
      comTranslateLexicalAccess(termA);
      comPrint("->data.string, ");
      comTranslateLexicalAccess(termB);
      comPrintln("->data.string, &vs%d))", tmpIndex);
    } else if (returnType2 == RTString) {
      comPrint("(vs%d.type = _matchString(\"%s\", ",
	  tmpIndex, comEscapeQuotes(com->currentConstraint->id));
      comTranslateString(termA, tmpIndex);
      comPrint(", ");
      comTranslateLexicalAccess(termB);
      comPrintln("->data.string, &vs%d))", tmpIndex);
    } else if (returnType2 == RTList) { /* not detected, but hey */
      comPrint("(vs%d.type = _matchList(\"%s\", vs%d.data.list, ",
	  tmpIndex, comEscapeQuotes(com->currentConstraint->id), tmpIndex);
      comTranslateLexicalAccess(termB);
      comPrintln("->data.string, &vs%d))", tmpIndex);
    } 
  } 

  if (returnType3 & RTString) {
    comTranslateTerm(termA, FALSE, TRUE, tmpIndex);
    /* check termA */
    if (returnType2 == RTPeek) {
      comPrintln(" != VTString ?");
      comPrint("((");
      comTranslateLexicalAccess(termA);
      comPrintln("->type != VTList)? (vs%d.type = VTError):", tmpIndex);
      comPrint("(vs%d.type = _matchList(\"%s\", ", tmpIndex, comEscapeQuotes(com->currentConstraint->id));
      comTranslateLexicalAccess(termA);
      comPrint("->data.list, ");
      comTranslateString(termB, tmpIndex+1);
      comPrintln(", &vs%d))) :", tmpIndex);
      comPrint("(vs%d.type = _matchString(\"%s\", ", tmpIndex, comEscapeQuotes(com->currentConstraint->id));
      comTranslateLexicalAccess(termA);
      comPrint("->data.string, ");
      comTranslateString(termB, tmpIndex+1);
      comPrintln(", &vs%d))", tmpIndex);
    } else if (returnType2 == RTString) {
      comPrint("(vs%d.type = _matchString(\"%s\", ",
	  tmpIndex, comEscapeQuotes(com->currentConstraint->id));
      comTranslateString(termA, tmpIndex);
      comPrint(", ");
      comTranslateString(termB, tmpIndex+1);
      comPrintln(", &vs%d))", tmpIndex);
    } else if (returnType2 == RTList) { /* not detected, but hey */
      comPrint("(vs%d.type = _matchList(\"%s\", vs%d.data.list, ",
	  tmpIndex, comEscapeQuotes(com->currentConstraint->id), tmpIndex);
      comTranslateString(termB, tmpIndex+1);
      comPrintln(", &vs%d))", tmpIndex);
    } 
  } 
}

/* ----------------------------------------------------------------------
 * translate a number access
 */
void comTranslateNumber(Term term, int tmpIndex)
{
  switch(term->type) {
    case TTNumber:
      comPrint("%g", term->data.number);
      break;
    default:
      comPrint("vs%d.data.number", tmpIndex);
  }
}

/* ----------------------------------------------------------------------
 * translate a string access
 */
void comTranslateString(Term term, int tmpIndex)
{
  switch (term->type) {
    case TTString:
      comPrint("strings[%d] /* %s */", 
	  comRegisterString(term->data.string), term->data.string);
      break;
    case TTLabel:
      comPrint("lv%d->label",
	  comIndexOfVarInfo(term->data.peek.varinfo));
      break;
    case TTLevel:
      comPrint("lv%d->level->id",
	  comIndexOfVarInfo(term->data.label));
      break;
    default: 
      comPrint("vs%d.data.string", tmpIndex);
  }
}

/* ----------------------------------------------------------------------
 * translate a term to c code
 */
void comTranslateTerm(Term term, 
    Boolean insideFormula, Boolean needsSeparator, int tmpIndex)
{
  ReturnType returnType = comReturnTypeOfTerm(term);

  /* switch off specific terms */
  if (term->type == TTBottomPeek || term->type == TTTopPeek) {
    String pathItem = (String) term->data.peek.path->item; 
    if (strcmp("from", pathItem) != 0 &&
        strcmp("to", pathItem) != 0 &&
	/*strcmp("id", pathItem) != 0 &&*/
	strcmp("word", pathItem) != 0) {
      needsSeparator = FALSE;
    }
  } else if (term->type == TTString ||
      term->type == TTLabel ||
      term->type == TTLevel ||
      term->type == TTNumber) {
    needsSeparator = FALSE;
  }
      
  if (needsSeparator && !(returnType & RTNoError)) {
    comPrint("(");
  }

  switch (term->type) {
    /* ---------------------------------------------------------------------- */
    case TTTopPeek:
      comTranslateTopPeek(term, tmpIndex);
      break;
    /* ---------------------------------------------------------------------- */
    case TTBottomPeek: 
      comTranslateBottomPeek(term, tmpIndex);
      break;
    /* ---------------------------------------------------------------------- */
    case TTLabel:
    case TTLevel:
    case TTString:
    case TTNumber:
      break;
    /* ---------------------------------------------------------------------- */
    case TTAdd: 
    case TTSubtract: 
    case TTMultiply:    
    case TTDivide:   
      comTranslateArithmetics(term, tmpIndex);
      break;
    /* ---------------------------------------------------------------------- */
    case TTFunction:
      comTranslateFunction(term->data.function.functor, 
	  term->data.function.args, tmpIndex);
      break;
  }

  if (needsSeparator) {
    if (!(returnType & RTNoError)) {
      if (insideFormula) {
	comPrintln(" != VTError) &&");
      } else {
	comPrintln(" == VTError)? VTError : ", 
	    tmpIndex, tmpIndex);
      }
    } else {
      comPrintln(",");
    }
  }
}

/* ----------------------------------------------------------------------
 * translate the arithmetic term operations
 */
void comTranslateArithmetics(Term term, int tmpIndex)
{ /* CONT */
  ReturnType returnType1 = comReturnTypeOfTerm(term->data.operation.op1) & ~RTNoError;
  ReturnType returnType2 = comReturnTypeOfTerm(term->data.operation.op2) & ~RTNoError;
  String operator = "horst";

  switch (term->type) {
    case TTAdd: 
      operator = "+";
      comPrintln("( /* add */");
      break;
    case TTSubtract: 
      operator = "-";
      comPrintln("( /* substract */");
      break;
    case TTMultiply:    
      operator = "*";
      comPrintln("( /* multiplty */");
      break;
    case TTDivide:   
      operator = "/";
      comPrintln("( /* divide */");
      break;
    default:
      break;
  }
  comIndent();
  comTranslateTerm(term->data.operation.op1, FALSE, TRUE, tmpIndex);

  switch (returnType1) {
    /* first term is a number */
    case RTNumber:
      comTranslateTerm(term->data.operation.op2, FALSE, TRUE, tmpIndex+1);
      switch (returnType2) {
	case RTPeek:
	  comPrintln(" != VTNumber ?");
	  comPrintln(
	    "(cdgPrintf(CDG_ERROR, "
	    "\"ERROR: type mismatch in term `%s'\\n\"), vs%d.type = VTError) :",
	    operator, tmpIndex);
	  comPrint("((vs%d.data.number = ", tmpIndex);
	  comTranslateNumber(term->data.operation.op1, tmpIndex);
	  comPrint(" %s ", operator);
	  comTranslateLexicalAccess(term->data.operation.op2);
	  comPrintln("->data.number), vs%d.type = VTNumber)", tmpIndex);
	  break;
	case RTNumber:
	  comPrint("((vs%d.data.number = ", tmpIndex); 
	  comTranslateNumber(term->data.operation.op1, tmpIndex);
	  comPrint(" %s ", operator);
	  comTranslateNumber(term->data.operation.op2, tmpIndex+1);
	  comPrintln("), vs%d.type = VTNumber)", tmpIndex);
	  break;
	default:
	  comWriteError("2nd argument of type `%s' not a number",
	      comReturnTypeToString(returnType2));
      }
      break;
    /* first term is a lexical access */
    case RTPeek:
      comPrint(" != VTNumber ?");
      comPrintln(
	  "(cdgPrintf(CDG_ERROR, "
	  "\"ERROR: type mismatch of 1rst argument in term `%s'\\n\"), vs%d.type = VTError) :",
	  operator, tmpIndex);
      comTranslateTerm(term->data.operation.op2, FALSE, TRUE, tmpIndex+1);
      switch (returnType2) {
	case RTPeek:
	  comPrintln(" != VTNumber ?");
	  comPrintln(
	    "(cdgPrintf(CDG_ERROR, "
	    "\"ERROR: type mismatch of 2nd argument in term `%s'\\n\"), vs%d.type = VTError) :",
	    operator, tmpIndex);
	  comPrint("((vs%d.data.number = ", tmpIndex);
	  comTranslateLexicalAccess(term->data.operation.op1);
	  comPrint("->data.number %s ", operator);
	  comTranslateLexicalAccess(term->data.operation.op2);
	  comPrintln("->data.number), vs%d.type = VTNumber)", tmpIndex);
	  break;
	case RTNumber:
	  comPrint("((");
	  comTranslateLexicalAccess(term->data.operation.op1);
	  comPrintln("->type != VTNumber)?");
	  comPrintln(
	    "(cdgPrintf(CDG_ERROR, "
	    "\"ERROR: type mismatch in term `%s'\\n\"), vs%d.type = VTError) :",
	    operator, tmpIndex);
	  comPrint("((vs%d.data.number = ", tmpIndex);
	  comTranslateLexicalAccess(term->data.operation.op1);
	  comPrint("->data.number %s ", operator);
	  comTranslateNumber(term->data.operation.op2, tmpIndex+1);
	  comPrintln("), vs%d.type = VTNumber)", tmpIndex);
	  break;
	default:
	  comWriteError("2nd argument of type `%s' not a number",
	      comReturnTypeToString(returnType2));
      }
      break;
    default:
	  comWriteError("1rst argument of type `%s' not a number",
	      comReturnTypeToString(returnType1));
  }
  comOutdent();
  comPrint(")");
}

/* ----------------------------------------------------------------------
 * translate a bottom peek
 */
void comTranslateBottomPeek(Term term, int tmpIndex)
{
  String pathItem = (String) term->data.peek.path->item; 
  int varInfoIndex = comIndexOfVarInfo(term->data.peek.varinfo);
  List l;
  String att;

  /* id */
  if (strcmp("id", pathItem) == 0) {
    comPrintln("/* @id */");
    return;
  }

  /* from */
  if (strcmp("from", pathItem) == 0) {
    comPrint("((vs%d.data.number = ((LexemNode)(lv%d->dependents->item))->arc->from), VTNumber)",
      tmpIndex, varInfoIndex);
  } 
  
  /* to */
  else if (strcmp("to", pathItem) == 0) {
    comPrint("((vs%d.data.number = ((LexemNode)(lv%d->dependents->item))->arc->to), VTNumber)",
      tmpIndex, varInfoIndex, tmpIndex);
  } 
  
  /* word */
  else if (strcmp("word", pathItem) == 0) {
    comPrint("((vs%d.data.string = ((LexemNode)(lv%d->dependents->item))->lexem->word), VTString)",
      tmpIndex, varInfoIndex);
  } 

  /* chunk_start */
  else if (strcmp("chunk_start", pathItem) == 0) {
    comPrintln("(((LexemNode)(lv%d->dependents->item))->grapheme->chunk ?", varInfoIndex);
    comPrintln("(vs%d.data.number = ((LexemNode)(lv%d->dependents->item))->grapheme->chunk->from->arc->from), vs%d.type = VTNumber :", 
	tmpIndex, varInfoIndex, tmpIndex);
    comPrintln("VTError)");
  }

  /* chunk_end */
  else if (strcmp("chunk_end", pathItem) == 0) {
    comPrintln("(((LexemNode)(lv%d->dependents->item))->grapheme->chunk ?", varInfoIndex);
    comPrintln("(vs%d.data.number = ((LexemNode)(lv%d->dependents->item))->grapheme->chunk->to->arc->to), vs%d.type = VTNumber :", 
	tmpIndex, varInfoIndex, tmpIndex);
    comPrintln("VTError)");
  }

  /* chunk_type */
  else if (strcmp("chunk_type", pathItem) == 0) {
    comPrintln("(((LexemNode)(lv%d->dependents->item))->grapheme->chunk ?", varInfoIndex);
    comPrintln("(vs%d.data.string = chunkerStringOfChunkType(((LexemNode)(lv%d->dependents->item))->grapheme->chunk)), vs%d.type = VTString :", 
	tmpIndex, varInfoIndex, tmpIndex);
    comPrintln("VTError)");
  }
  
  /* peek */
  else { 
    comPrint("/* peek regent ");
    for (l = term->data.peek.path; l != NULL; l = l->next) {
      att = (String)l->item;
      comPrint("%s", att);
      if (l->next) 
	comPrint(".");
    }
    comPrintln(" */");
    comTranslateLexicalAccess(term);
    comPrint("->type");
  }
}

/* ----------------------------------------------------------------------
 * translate a top peek
 */
void comTranslateTopPeek(Term term, int tmpIndex)
{
  String pathItem = (String) term->data.peek.path->item; 
  int varInfoIndex = comIndexOfVarInfo(term->data.peek.varinfo);
  List l;
  String att;

  /* id */
  if (strcmp("id", pathItem) == 0) {
    comPrintln("/* ^id */");
    return;
  }

  /* from */
  if (strcmp("from", pathItem) == 0) {
    comPrintln("((lv%d->regents == NULL)?", varInfoIndex);
    comPrintln("(vs%d.type = VTError) : ", tmpIndex);
    comPrint("((vs%d.data.number = (lv%d->regents == NONSPEC)?",
	tmpIndex, varInfoIndex);
    comPrint("lg->max+2 : ((LexemNode)(lv%d->regents->item))->arc->from),", varInfoIndex);
    comPrintln("(vs%d.type = VTNumber)))", tmpIndex);
  } 
  
  /* to */
  else if (strcmp("to", pathItem) == 0) {
    comPrintln("((lv%d->regents == NULL)?", varInfoIndex);
    comPrintln("(vs%d.type = VTError) : ", tmpIndex);
    comPrint("((vs%d.data.number = (lv%d->regents == NONSPEC)?",
	tmpIndex, varInfoIndex);
    comPrint("lg->max+2 : ((LexemNode)(lv%d->regents->item))->arc->to),", varInfoIndex);
    comPrintln("(vs%d.type = VTNumber)))", tmpIndex);
  } 
  
  /* word */
  else if (strcmp("word", pathItem) == 0) {
    comPrintln("((lv%d->regents == NONSPEC || lv%d->regents == NULL)?",
	varInfoIndex, varInfoIndex);
    comIndent();
    comPrintln("(vs%d.type = VTError) :", tmpIndex);
    comPrintln("((vs%d.data.string = ((LexemNode)(lv%d->regents->item))->lexem->word), (vs%d.type = VTString))",
      tmpIndex, varInfoIndex, tmpIndex);
    comOutdent();
    comPrint(")");
  } 

  /* chunk_start */
  else if (strcmp("chunk_start", pathItem) == 0) {
    comPrintln("((lv%d->regents == NONSPEC || lv%d->regents == NULL)?",
	varInfoIndex, varInfoIndex);
    comIndent();
    comPrintln("(vs%d.type = VTError) :", tmpIndex);
    comPrintln("(((LexemNode)(lv%d->regents->item))->grapheme->chunk ?", varInfoIndex);
    comPrintln("(vs%d.data.number = ((LexemNode)(lv%d->regents->item))->grapheme->chunk->from->arc->from), vs%d.type = VTNumber :", 
	tmpIndex, varInfoIndex, tmpIndex);
    comPrintln("VTError))");
  }

  /* chunk_end */
  else if (strcmp("chunk_end", pathItem) == 0) {
    comPrintln("((lv%d->regents == NONSPEC || lv%d->regents == NULL)?",
	varInfoIndex, varInfoIndex);
    comIndent();
    comPrintln("(vs%d.type = VTError) :", tmpIndex);
    comPrintln("(((LexemNode)(lv%d->regents->item))->grapheme->chunk ?", varInfoIndex);
    comPrintln("(vs%d.data.number = ((LexemNode)(lv%d->regents->item))->grapheme->chunk->to->arc->to), vs%d.type = VTNumber :", 
	tmpIndex, varInfoIndex, tmpIndex);
    comPrintln("VTError))");
  }
  
  /* chunk_type */
  else if (strcmp("chunk_type", pathItem) == 0) {
    comPrintln("((lv%d->regents == NONSPEC || lv%d->regents == NULL)?",
	varInfoIndex, varInfoIndex);
    comIndent();
    comPrintln("(vs%d.type = VTError) :", tmpIndex);
    comPrintln("(((LexemNode)(lv%d->regents->item))->grapheme->chunk ?", varInfoIndex);
    comPrintln("(vs%d.data.string = chunkerStringOfChunkType(((LexemNode)(lv%d->regents->item))->grapheme->chunk)), vs%d.type = VTString :", 
	tmpIndex, varInfoIndex, tmpIndex);
    comPrintln("VTError))");
  }

  /* peek */
  else { 
    comPrint("/* peek regent ");
    for (l = term->data.peek.path; l != NULL; l = l->next) {
      att = (String)l->item;
      comPrint("%s", att);
      if (l->next) 
	comPrint(".");
    }
    comPrintln(" */");
    comPrintln("((lv%d->regents == NONSPEC || lv%d->regents == NULL)?", 
	varInfoIndex, varInfoIndex);
    comPrint("(vs%d.type = VTError) : ", tmpIndex);
    comTranslateLexicalAccess(term);
    comPrint("->type)");
  }
}

/* ----------------------------------------------------------------------
 * translate the print predicate
 */
void comTranslatePrint(List args, int tmpIndex)
{
  Term term;
  List l;
  ReturnType returnType;

  for (l = args; l != NULL; l = l->next) {
    term = (Term)l->item;
    returnType = comReturnTypeOfTerm(term) & ~RTNoError;
    comTranslateTerm(term, FALSE, TRUE, tmpIndex);
    if (returnType == RTPeek) {
      comPrint("printValue(CDG_DEFAULT, ");
      comTranslateLexicalAccess(term);
      comPrintln(", 0),");
    } else {
      comPrintln("printValue(CDG_DEFAULT, &vs%d, 0),",
	  tmpIndex);
    }
  }
  comPrintln("TRUE");
}

/* ----------------------------------------------------------------------
 * translate a predicate to c code
 */
void comTranslatePredicate(String functor, List args, int tmpIndex)
{
  comPrintln("( /* %s */", functor);
  comIndent();

  /* start, stop */
  if (strcmp("start", functor) == 0 ||
      strcmp("stop", functor) == 0) {
    comTranslateStartStop(functor, args, tmpIndex);
  } 

  /* root, spec, nonspec */
  else if (strcmp("root", functor) == 0 ||
      strcmp("spec", functor) == 0 ||
      strcmp("nonspec", functor) == 0) {
    comTranslateGuard(functor, args, tmpIndex);
  } 

  /* print */
  else if (strcmp("print", functor) == 0) {
    comTranslatePrint(args, tmpIndex);
  } 

  /* exists */
  else if (strcmp("exists", functor) == 0) {
    comTranslateExists(args);
  } 

  /* subsumes */
  else if (strcmp("subsumes", functor) == 0 ||
           strcmp("compatible", functor) == 0) {
    comTranslateSubsumes(functor, args, tmpIndex);
  } 

  /* chunk head */
  else if (strcmp("chunk_head", functor) == 0) {
    comTranslateChunkHead(functor, args, tmpIndex);
  } 

  /* under */
  else if (strcmp("under", functor) == 0) {
    comTranslateUnder(functor, args, tmpIndex);
  }
  
  /* between */
  else if (strcmp("between", functor) == 0) {
    comTranslateBetween(functor, args, tmpIndex);
  }

  /* is */
  else if (strcmp("is", functor) == 0) {
    comTranslateIs(functor, args, tmpIndex);
  }

  /* connected */
  else if (strcmp("connected", functor) == 0) {
    comTranslateConnected(functor, args, tmpIndex);
  }

  /* has */
  else if (strcmp("has", functor) == 0) {
    comTranslateHas(functor, args, tmpIndex);
  }

  comOutdent();
  comPrint(")");
}

/* ----------------------------------------------------------------------
 * translate the subsumes and compatible predicates
 */
void comTranslateSubsumes(String functor, List args, int tmpIndex)
{
  int i;
  Term term = (Term)listNthElement(args, 1);
  Term termA = (Term)listNthElement(args, 2);
  Term termB = (Term)listNthElement(args, 3);
  ReturnType returnType1 = comReturnTypeOfTerm(term) & ~RTNoError;
  ReturnType returnType2 = comReturnTypeOfTerm(termA) & ~RTNoError;
  ReturnType returnType3 = comReturnTypeOfTerm(termB) & ~RTNoError;

  /* translate hierarchy term */
  if (term->type != TTString) {
    comTranslateTerm(term, TRUE, TRUE, tmpIndex);
  }
  
  /* check the hierarchy */
  if (returnType1 == RTString) {

    /* extra rule for static strings */
    if (term->type == TTString) {
      i = comIndexOfHierarchy(term->data.string);
      if (i == -1) {
	comWriteError("hierarchy `%s' not found", term->data.string);
      } else {
	comPrintln("(hierarchy = hierarchies[%d].hierarchy),", i);
      }
    } 
    
    /* its a string but not provided staticaly */
    else { 
      comPrint("(hierarchy = _findHierarchy(");
      comTranslateString(term, tmpIndex);
      comPrintln(")),");
      comPrintln("(hierarchy != NULL) &&");
    }
  } 
  /* hierarchy provided lexicaly */
  else if (returnType1 == RTPeek) { 
    comPrintln(" != VTString?");
    comPrintln("(cdgPrintf(CDG_ERROR, \"ERROR: type mismatch of 1rst argument in `%s' of constraint `%s' in line %%d\\n\", __LINE__), FALSE) :",
	functor, comEscapeQuotes(com->currentConstraint->id));
    comPrint("((hierarchy = _findHierarchy(");
    comTranslateLexicalAccess(term);
    comPrint("->data.string)), ");
    comPrintln("(hierarchy != NULL)) &&");
  }
  /* hierarchy unknown at compile time */
  else { 
    comPrintln("((vs%d.type != VTString)?", tmpIndex);
    comPrintln("(cdgPrintf(CDG_ERROR, \"ERROR: type mismatch of 1rst argument in `%s' of constraint `%s' in line %%d\\n\", __LINE__), FALSE) :",
	functor, comEscapeQuotes(com->currentConstraint->id));
    comPrint("((hierarchy = _findHierarchy(");
    comTranslateString(term, tmpIndex);
    comPrint(")), ");
    comPrintln("(hierarchy != NULL))) &&");
  }

  /* translate termA */
  comPrint("(");
  comTranslateTerm(termA, TRUE, TRUE, tmpIndex);
  
  /* check sort a */
  if (returnType2 == RTPeek) {
    comPrintln(" != VTString ?");
    comPrintln("(cdgPrintf(CDG_ERROR, \"ERROR: type mismatch of 2nd argument in `%s' of constraint `%s' in line %%d\\n\", __LINE__), FALSE) :",
	functor, comEscapeQuotes(com->currentConstraint->id));
  }

  /* translate termB */
  comPrint("(");
  comTranslateTerm(termB, TRUE, TRUE, tmpIndex+1);

  /* check sort b */
  if (returnType3 == RTPeek) {
    comPrintln(" != VTString ?");
    comPrintln("(cdgPrintf(CDG_ERROR, \"ERROR: type mismatch of 3nd argument of in `%s' of constraint `%s' in line %%d\\n\", __LINE__), FALSE):",
	functor, comEscapeQuotes(com->currentConstraint->id));
  }

  comPrint("(");
  if (returnType2 == RTPeek) {
    if (returnType3 == RTPeek) {
      comPrint("_%s(\"%s\", ", functor, comEscapeQuotes(com->currentConstraint->id));
      comTranslateLexicalAccess(termA);
      comPrint("->data.string, ");
      comTranslateLexicalAccess(termB);
      comPrint("->data.string)");
    } else {
      comPrint("_%s(\"%s\", ", functor, comEscapeQuotes(com->currentConstraint->id));
      comTranslateLexicalAccess(termA);
      comPrint("->data.string, ");
      comTranslateString(termB, tmpIndex+1);
      comPrint(")");
    }
  } else if (returnType3 == RTPeek) {
    comPrint("_%s(\"%s\", ", functor, comEscapeQuotes(com->currentConstraint->id));
    comTranslateString(termA, tmpIndex);
    comPrint(", ");
    comTranslateLexicalAccess(termB);
    comPrint("->data.string)");
  } else {
    comPrint("_%s(\"%s\", ", 
	functor, comEscapeQuotes(com->currentConstraint->id), tmpIndex, tmpIndex+1);
    comTranslateString(termA, tmpIndex);
    comPrint(", ");
    comTranslateString(termB, tmpIndex+1);
    comPrint(")");
  }
  comPrintln(")))");
}

/* ----------------------------------------------------------------------
 * translate the exists predicate
 */
void comTranslateExists(List args)
{
  String att;
  Term term = (Term)args->item;
  int varInfoIndex = comIndexOfVarInfo(term->data.peek.varinfo);
  String pathItem = (String) term->data.peek.path->item; 
  List l;
  
  /* verbosity */
  comPrint("/* %s ", (term->type == TTTopPeek)?"regent":"dependent");
  for (l = term->data.peek.path; l != NULL; l = l->next) {
    att = (String)l->item;
    comPrint("%s", att);
    if (l->next) 
      comPrint(".");
  }
  comPrintln(" */");

  /* id */
  if (strcmp("id", pathItem) == 0) {
    comPrintln("TRUE");
  }

  /* from, to */
  if (strcmp("from", pathItem) == 0 ||
      strcmp("to", pathItem) == 0) {
    if (term->type == TTTopPeek) {
      comPrintln("(lv%d->regents != NULL)", varInfoIndex);
    } else {
      comPrintln("TRUE");
    }
  }
  /* word */
  if (strcmp("word", pathItem) == 0) {
    if (term->type == TTTopPeek) {
      comPrintln("(lv%d->regents != NULL && lv%d->regents != NONSPEC)",
	  varInfoIndex, varInfoIndex);
    } else {
      comPrintln("TRUE");
    }
  }

  /* peek */
  else if (term->type == TTTopPeek) {
    comPrintln("((lv%d->regents != NONSPEC && lv%d->regents != NULL) &&", 
	varInfoIndex, varInfoIndex);
    comPrintln("(((LexemNode)(lv%d->regents->item))->lexem->values[%d]->type != VTError))",
	varInfoIndex, term->data.peek.pathindex);
  } else {
    comPrintln("(((LexemNode)(lv%d->dependents->item))->lexem->values[%d]->type != VTError)",
	varInfoIndex, term->data.peek.pathindex);
  }
}
/* ----------------------------------------------------------------------
 * translate the chunk_head predicate
 */
void comTranslateChunkHead(String functor, List args, int tmpIndex) 
{
  Term term = (Term)args->item;
  ReturnType returnType = comReturnTypeOfTerm(term) & ~RTNoError;

  if (returnType & RTLexemNode){
    if (returnType & RTLexemPosition) {
      comPrint("(");
      comTranslateLexemNodeAccess(term, tmpIndex);
      comPrint(" != NULL) && (");
      comTranslateLexemNodeAccess(term, tmpIndex);
      comPrint(" != NONSPEC) &&");
    }
      comPrint("((");
      comTranslateLexemNodeAccess(term, tmpIndex);
      comPrintln("->grapheme->chunk) ?");
      comPrint("(");
      comTranslateLexemNodeAccess(term, tmpIndex);
      comPrint("->grapheme == ");
      comTranslateLexemNodeAccess(term, tmpIndex);
      comPrint("->grapheme->chunk->head) : FALSE)");
  } else {
    /* never reach */
    comWriteError("illegal argument type `%s' in functor `%s'",
    comReturnTypeToString(returnType), functor);
  }
  
}
    
/* ----------------------------------------------------------------------
 * translate the 'under' predicate
 */
void comTranslateUnder(String functor, List args, int tmpIndex)
{
  Term term1, term2;
  ReturnType returnType1, returnType2;

  term1 = (Term)args->item;
  returnType1 = comReturnTypeOfTerm(term1) & ~RTNoError;
  term2= (Term)args->next->item;
  returnType2 = comReturnTypeOfTerm(term2) & ~RTNoError;
  
  
  if (returnType1 & RTLexemNode) {
    if (returnType2 & RTLexemNode) {
      comPrint("(");
	comPrint("(! ");
	comTranslateLexemNodeAccess(term1, tmpIndex);
        comPrint(" || ");
	comTranslateLexemNodeAccess(term1, tmpIndex);
	comPrintln(" == NONSPEC) ? FALSE : ");
      
	comPrint("(! ");
	comTranslateLexemNodeAccess(term2, tmpIndex);
        comPrint(" || ");
	comTranslateLexemNodeAccess(term2, tmpIndex);
	comPrintln(" == NONSPEC) ? FALSE : ");
	comPrint("predUnderImpl(");
	comTranslateLexemNodeAccess(term1, tmpIndex);
	comPrint(", ");
	comTranslateLexemNodeAccess(term2, tmpIndex);
	comPrint(", lg, context)");
      comPrint(")");
      return;
    } 
  } 

  comWriteError("illegal argument type in functor '%s'", functor);

}

/* ----------------------------------------------------------------------
 * translate the 'between' predicate
 */
void comTranslateBetween(String functor, List args, int tmpIndex)
{
  Term term;
  ReturnType returnType;

  /* assign a */
  term = (Term)args->item;
  returnType = comReturnTypeOfTerm(term);
  
  comTranslateTerm(term, TRUE, TRUE, tmpIndex);
  if (returnType & RTLexemNode) {
    comPrint("(! ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    if (returnType & RTLexemPosition) {
      comPrint(" || ");
      comTranslateLexemNodeAccess(term, tmpIndex);
      comPrint(" == NONSPEC");
    }
    comPrintln(") ?");
    comPrintln("FALSE :" );
    comPrint("(a = ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrint("->arc->from)");
  } 
  if (returnType & RTNumber) {
    comPrint("(a = ");
    comTranslateNumber(term, tmpIndex);
    comPrintln("), (a < lg->min) ? (a = lg->min) :");
    comPrintln("(a > lg->max) ? (a = lg->max) :");
    comPrintln("TRUE");
  }

  comPrintln(", ");

  /* assign b */
  term = (Term)args->next->item;
  returnType = comReturnTypeOfTerm(term);

  comTranslateTerm(term, TRUE, TRUE, tmpIndex);
  if (returnType & RTLexemNode) {
    comPrint("(! ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    if (returnType & RTLexemPosition) {
      comPrint(" || ");
      comTranslateLexemNodeAccess(term, tmpIndex);
      comPrint(" == NONSPEC");
    }
    comPrintln(") ?");
    comPrintln("FALSE :" );
    comPrint("(b = ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrint("->arc->from)");
  } 
  if (returnType & RTNumber) {
    comPrint("(b = ");
    comTranslateNumber(term, tmpIndex);
    comPrintln("), (b < lg->min) ? (b = lg->min) :");
    comPrintln("(b > lg->max) ? (b = lg->max) :");
    comPrintln("TRUE");
  }

  comPrintln(", ");

  /* call implementation */
  term = (Term)listNthElement(args, 3);
  returnType = comReturnTypeOfTerm(term);
  
  if (returnType & RTString) {
    comPrint("(a > b) ? predBetweenImpl(&b, &a, ");
    comTranslateString(term, tmpIndex);
    comPrintln(", lg, context) :");
    comPrint("predBetweenImpl(&a, &b, ");
    comTranslateString(term, tmpIndex);
    comPrint(", lg, context)");
  }
  
}

/* ----------------------------------------------------------------------
 * translate the 'is' predicate
 */
void comTranslateIs(String functor, List args, int tmpIndex)
{
  Term term1, term2;
  ReturnType returnType;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  Level level = inputGetMainlevel(inputCurrentGrammar);
  int levelno = level ? level->no : 0;
  
  /* recursive `is' not yet done */
  if (listSize(args) >= 3) {
    comWriteError("ternary `is' not yet supported.");
    return;
  }

  /* first and second argument */
  
  term1 = (Term)args->item;
  returnType = comReturnTypeOfTerm(term1);
    
  if (returnType & RTLexemNode) {
    comPrint("(! ");
    comTranslateLexemNodeAccess(term1, tmpIndex);
    if (returnType & RTLexemPosition) {
      comPrint(" || ");
      comTranslateLexemNodeAccess(term1, tmpIndex);
      comPrint(" == NONSPEC");
    }
    comPrintln(") ?");
    comPrintln("FALSE :" );
  } 

  term2 = (Term)args->next->item;
  
  comPrint("(lv = ", tmpIndex);
  comPrint("vectorElement(context, %d * ", nrlevels);
  comTranslateLexemNodeAccess(term1, tmpIndex);
  comPrintln("->arc->from + %d))", levelno);
  comPrint("  && (lv->label == ");
  comTranslateString(term2, tmpIndex);
  comPrintln(") ? TRUE : FALSE");
}

/* ----------------------------------------------------------------------
 * translate the 'connected' predicate
 */
void comTranslateConnected(String functor, List args, int tmpIndex)
{
  Term term;
  ReturnType returnType;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  VarInfo vi = com->currentConstraint->vars->item;
  int levelno = vi->level->no;

  term = (Term)args->item;
  returnType = comReturnTypeOfTerm(term);
  if (returnType & RTLexemNode) {
    comPrint("(! ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    if (returnType & RTLexemPosition) {
      comPrint(" || ");
      comTranslateLexemNodeAccess(term, tmpIndex);
      comPrint(" == NONSPEC");
    }
    comPrintln(") ?");
    comPrintln("FALSE :" );
    comPrint("(a = ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrint("->arc->from)");
  } 
  comPrintln(",");

  term = (Term)args->next->item;
  returnType = comReturnTypeOfTerm(term);
  if (returnType & RTLexemNode) {
    comPrint("(! ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    if (returnType & RTLexemPosition) {
      comPrint(" || ");
      comTranslateLexemNodeAccess(term, tmpIndex);
      comPrint(" == NONSPEC");
    }
    comPrintln(") ?");
    comPrintln("FALSE :" );
    comPrint("(b = ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrint("->arc->from)");
  } 
  comPrintln(",");

  comPrintln("_connected(a, b, %d, %d, lg, context)", nrlevels, levelno);
}

/* ----------------------------------------------------------------------
 * translate the 'has' predicate
 */
void comTranslateHas(String functor, List args, int tmpIndex)
{
  Level level = inputGetMainlevel(inputCurrentGrammar);
  int mainno = level ? level->no : 0;
  Term term;
  ReturnType returnType;
  int nrArgs = listSize(args);
  Boolean useHierarchy = FALSE;

  comPrintln("! context ? TRUE :");

  /* first argument */
  term = (Term)args->item;
  returnType = comReturnTypeOfTerm(term);

  comPrint("(! ");
  comTranslateLexemNodeAccess(term, tmpIndex);
  if (returnType & RTLexemPosition) {
    comPrint(" || ");
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrint(" == NONSPEC");
  }
  comPrintln(") ?");
  comPrintln("FALSE :" );
  
  /* use cache if possible */
  if (5 > listSize(args)) {
    comPrintln("lock_counter > 0 ?");
    comPrint("(a = has_cache[%d * lock_width + ", com->currentFormula->index);
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrintln("->arc->from]) :");
    comPrintln("(a = -1),");
    comPrintln("(lock_counter > 0 && 2 == a) ? TRUE : ");
    comPrintln("(lock_counter > 0 && 1 == a) ? FALSE : ");
  }
  
  /* second argument -> assign Constraint constraint*/
  term = (Term)listNthElement(args, 2);
  returnType = comReturnTypeOfTerm(term);

  comPrint("((constraint = hashGet(inputCurrentGrammar->constraints, ");
  if (returnType & RTString) {
    comTranslateString(term, tmpIndex);
  }
  if (returnType & RTPeek) {
    comTranslateLexicalAccess(term);
    comPrint("->data.string");
  }
  comPrintln(")),");
  comPrintln("(constraint = (constraint && 1 != listSize(constraint->vars)) ? NULL : constraint),");

  /* third argument */ 
  if (nrArgs >= 3) {
    useHierarchy = TRUE;
    term = (Term)listNthElement(args, 3);

    comPrint("! (hierarchy = findHierarchy(");
    comTranslateString(term, tmpIndex);
    comPrintln(")) ? FALSE :");
  }

  /* build arguments array and call implementation */
  term = (Term)args->item;
  comPrintln("(hasImplArgs[0] =");
  comIndent();
    /* lexemNode */
    comTranslateLexemNodeAccess(term, tmpIndex);
    comPrintln("),");
  comOutdent();
  comPrintln("(hasImplArgs[1] =");
  comIndent();
    /* label */
    term = (Term)listNthElement(args, 2);
    returnType = comReturnTypeOfTerm(term);
    if (returnType & RTString) {
      comTranslateString(term, tmpIndex);
    }
    if (returnType & RTPeek) {
      comTranslateLexicalAccess(term);
      comPrint("->data.string");
    }
    comPrintln("),");
  comOutdent();
  comPrintln("(hasImplArgs[2] =");
  comIndent();
    /* constraint */
    comPrintln("constraint),");
  comOutdent();
  comPrintln("(hasImplArgs[3] =");
  comIndent();
    /* hierarchy */
    if (useHierarchy) {
      comPrintln("hierarchy),");
    } else {
      comPrintln("NULL),");
    }
  comOutdent();
  comPrintln("(hasImplArgs[4] =");
  comIndent();
    /* labels */
    if (nrArgs >= 4) {
      comTranslateString((Term)listNthElement(args, 4), tmpIndex);
    } else {
      comPrint("NULL");
    }
    comPrintln("),");
  comOutdent();
  comPrintln("(hasImplArgs[5] =");
  comIndent();
    /* recursive */
    if (nrArgs >= 4) {
      comPrint("((bool = TRUE), &bool)");
    } else {
      comPrint("((bool = FALSE), &bool)");
    }
    comPrintln("),");
  comOutdent();
  comPrintln("(hasImplArgs[6] = ");
  comIndent();
    /* min */
    comPrint("((a = (");
    if (nrArgs >= 5) {
      term = (Term)listNthElement(args, 5);
      comTranslateTerm(term, FALSE, TRUE, tmpIndex);
      comTranslateNumber(term, tmpIndex);
      comPrintln(" > 0 ?");
      comTranslateTerm(term, FALSE, TRUE, tmpIndex);
      comTranslateNumber(term, tmpIndex);
      comPrintln(" :");
    } 
    comPrintln("lg->min)), &a)),");
  comOutdent();
  comPrintln("(hasImplArgs[7] = ");
  comIndent();
    /* max */
    comPrint("((b = (");
    // FIXME: a `max' argument of -1 should be interpreted as lg->max!
    if (nrArgs >= 6) {
      term = (Term)listNthElement(args, 6);
      comTranslateTerm(term, FALSE, TRUE, tmpIndex + 1);
      comTranslateNumber(term, tmpIndex + 1);
      comPrintln(" < lg->max ?");
      comTranslateTerm(term, FALSE, TRUE, tmpIndex + 1);
      comTranslateNumber(term, tmpIndex + 1);
      comPrintln(" :");
    }
    comPrintln("lg->max)), b = ((b == -1) ? lg->max : b), &b)),");
  comOutdent();
  comPrintln("(hasImplArgs[8] = ");
  comIndent();
    /* context */
    comPrintln("context),");
  comOutdent();
  comPrintln("(hasImplArgs[9] = ");
  comIndent();
    /* nrlevels */
    comPrintln("((c = %d), &c)),", inputCurrentGrammar->noOfLevels);
  comOutdent();
  comPrintln("(hasImplArgs[10] = ");
  comIndent();
    /* mainno */
    comPrintln("((d = %d), &d)),", mainno);
  comOutdent();

  /* call preHasImpl */
  comPrintln("predHasImpl(hasImplArgs))");
}

/* ----------------------------------------------------------------------
 * translate the root, spec and nonspec guard-predicates
 */
void comTranslateGuard(String functor, List args, int tmpIndex)
{
  Term term = (Term)args->item;
  ReturnType returnType = comReturnTypeOfTerm(term) & ~RTNoError;
  String operator = "horst";

  /* all dependent guards aren't needed */
  if (!(returnType & RTLexemPosition)) {
    comPrintln("TRUE");
  }
  
  if (strcmp(functor, "root") == 0) {
    operator = "== NULL";
  } else if (strcmp(functor, "nonspec") == 0) {
    operator = "== NONSPEC";
  } else {
    operator = "!= NONSPEC";
  }

  comTranslateTerm(term, TRUE, TRUE, tmpIndex);
  switch (returnType & ~RTLexemPosition) {
    case RTLexemNode:
      comPrint("(");
      comTranslateLexemNodeAccess(term, tmpIndex);
      comPrintln(" %s)", operator);
      break;
    default: /* never reach */
      comWriteError("illegal argument type `%s' in term `%s'",
	  comReturnTypeToString(returnType), functor);
  }
}

/* ----------------------------------------------------------------------
 * translate the start and stop predicate
 */
void comTranslateStartStop(String functor, List args, int tmpIndex)
{
  Term term = (Term)args->item;
  ReturnType returnType = comReturnTypeOfTerm(term) & ~RTNoError;
  String postfix = "horst";

  if (strcmp("start", functor) == 0) {
    postfix = "from == lg->min";
  } else {
    postfix = "to == lg->max";
  }

  comTranslateTerm(term, TRUE, TRUE, tmpIndex);

  switch(returnType & ~RTLexemPosition) {
    case RTLexemNode:
      if (returnType & RTLexemPosition) {
	comPrint("(");
	comTranslateLexemNodeAccess(term, tmpIndex);
	comPrintln(" != NULL) && ");
	comPrint("(");
	comTranslateLexemNodeAccess(term, tmpIndex);
	comPrintln(" != NONSPEC) &&");
      }
      comPrint("(");
      comTranslateLexemNodeAccess(term, tmpIndex);
      comPrint("->arc->%s)", postfix);
      break;
    default: /* never reach */
      comWriteError("illegal argument type `%s' in term `%s'",
	  comReturnTypeToString(returnType), functor);
  }
}

/* ----------------------------------------------------------------------
 * translate the equal and not equal formulas
 */
void comTranslateEquation(Formula formula, int tmpIndex)
{
  ReturnType returnType1 = comReturnTypeOfTerm(formula->data.relation.op1) & ~RTNoError;
  ReturnType returnType2 = comReturnTypeOfTerm(formula->data.relation.op2) & ~RTNoError;
  String operator = "horst";

  if (formula->type == FTNotEqual) {
    operator = "!=";
    comPrintln("( /* not equal */");
  } else {
    operator = "==";
    comPrintln("( /* equal */");
  }

  comIndent();
  comTranslateTerm(formula->data.relation.op1, TRUE, TRUE, tmpIndex);

  /* check known return types */
  switch (returnType1 & ~RTLexemPosition) {
    /* first term returns a string --------------------------------------------- */
    case RTString:
      comTranslateTerm(formula->data.relation.op2, TRUE, TRUE, tmpIndex+1); 
      switch(returnType2) {
	case RTString:
	  comPrint("("); 
	  comTranslateString(formula->data.relation.op1, tmpIndex);
	  comPrint(" %s ", operator);
	  comTranslateString(formula->data.relation.op2, tmpIndex+1);
	  comPrintln(")"); 
	  break;
	case RTPeek:
	  comPrintln(" == VTError ? FALSE :"); /* should be a grammar error */
	  comTranslateLexicalAccess(formula->data.relation.op2);
	  comPrintln("->type == VTString ?");
	  comIndent();
	  comPrint("(");
	  comTranslateString(formula->data.relation.op1, tmpIndex);
	  comPrint(" %s ", operator);
	  comTranslateLexicalAccess(formula->data.relation.op2);
	  comPrintln("->data.string) : ");
	  comPrint("(cdgPrintf(CDG_ERROR, "
	      "\"ERROR: 2nd argument not a string in formula `%s' in constraint `%s', line %%d\\n\", __LINE__), ",
	      operator, comEscapeQuotes(com->currentConstraint->id));
	  comPrintln("FALSE)");
	  comOutdent();
	  break;
	default:
	  if (formula->type == FTEqual) {
	    comPrintln("FALSE");
	    comWriteWarning("comparing type `%s' to `%s'",
		comReturnTypeToString(returnType1),
		comReturnTypeToString(returnType2));
	  } else {
	    comPrintln("TRUE");
	    comWriteWarning("comparing type `%s' to `%s'",
		comReturnTypeToString(returnType1),
		comReturnTypeToString(returnType2));
	  }
      }
      break;
    /* first term returns a number --------------------------------------------- */
    case RTNumber:
      comTranslateTerm(formula->data.relation.op2, TRUE, TRUE, tmpIndex+1); 
      switch (returnType2) {
	case RTNumber:
	  comPrint("("); 
	  comTranslateNumber(formula->data.relation.op1, tmpIndex);
	  comPrint(" %s ", operator);
	  comTranslateNumber(formula->data.relation.op2, tmpIndex+1);
	  comPrintln(")");
	  break;
	case RTPeek:
	  comPrintln(" == VTError ? FALSE :"); /* should be a grammar error */
	  comTranslateLexicalAccess(formula->data.relation.op2);
	  comPrintln("->type == VTNumber ?");
	  comIndent();
	  comPrint("(");
	  comTranslateNumber(formula->data.relation.op1, tmpIndex);
	  comPrint(" %s ", operator);
	  comTranslateLexicalAccess(formula->data.relation.op2);
	  comPrintln("->data.number) : ");
	  comPrint("(cdgPrintf(CDG_ERROR, "
	      "\"ERROR: 2nd argument not a number in formula `%s' in constraint `%s', line %%d\\n\", __LINE__),",
	      operator, comEscapeQuotes(com->currentConstraint->id));
	  comPrintln("FALSE)");
	  comOutdent();
	  break;
	default:
	  if (formula->type == FTEqual) {
	    comPrintln("FALSE");
	    comWriteWarning("comparing type `%s' to `%s'",
		comReturnTypeToString(returnType1),
		comReturnTypeToString(returnType2));
	  } else {
	    comPrintln("TRUE");
	    comWriteWarning("comparing type `%s' to `%s'",
		comReturnTypeToString(returnType1),
		comReturnTypeToString(returnType2));
	  }
      }
      break;
    /* first term returns a lexemnode ------------------------------------------ */
    case RTLexemNode:
      comTranslateTerm(formula->data.relation.op2, TRUE, TRUE, tmpIndex+1); 
      switch (returnType2 & ~RTLexemPosition) {
	case RTLexemNode: 
	  /* TODO: this can be speeded up by dividing this disjunction, alas ... */
	  comPrint("(");
	  if ((returnType1 & RTLexemPosition) || (returnType2 & RTLexemPosition)) {
	    comTranslateLexemNodeAccess(formula->data.relation.op1, tmpIndex);
	    comPrintln(" == NULL)?");
	    comPrint("(");
	    comTranslateLexemNodeAccess(formula->data.relation.op2, tmpIndex);
	    comPrintln(" %s NULL) :", operator);
	    comPrint("((");
	    comTranslateLexemNodeAccess(formula->data.relation.op2, tmpIndex);
	    comPrintln(" != NULL) &&");
	    comPrint("(");
	    comTranslateLexemNodeAccess(formula->data.relation.op1, tmpIndex);
	    comPrintln(" != NONSPEC) &&");
	    comPrint("(");
	    comTranslateLexemNodeAccess(formula->data.relation.op2, tmpIndex);
	    comPrintln(" != NONSPEC) &&");
	  }
	  comPrint("(");
	  comTranslateLexemNodeAccess(formula->data.relation.op1, tmpIndex);
	  comPrint("->grapheme %s ", operator);
	  comTranslateLexemNodeAccess(formula->data.relation.op2, tmpIndex);
	  comPrintln("->grapheme))");
	  break;
	default:
	  if (formula->type == FTEqual) {
	    comPrintln("FALSE");
	    comWriteWarning("comparing type `%s' to `%s'",
		comReturnTypeToString(returnType1),
		comReturnTypeToString(returnType2));
	  } else {
	    comPrintln("TRUE");
	    comWriteWarning("comparing type `%s' to `%s'",
		comReturnTypeToString(returnType1),
		comReturnTypeToString(returnType2));
	  }
      }
      break;
    /* first term is a lexical access ------------------------------------------ */
    case RTPeek:
      switch (returnType2) {
	case RTNumber:
	  comPrintln(" == VTError ? FALSE :"); /* should be a grammar error */
	  comTranslateLexicalAccess(formula->data.relation.op1);
	  comPrintln("->type != VTNumber ?");
	  comPrintln("(cdgPrintf(CDG_ERROR, "
	      "\"ERROR: 1rst argument not a number in formula `%s' in constraint `%s', line %%d\\n\", __LINE__), FALSE):",
	      operator, comEscapeQuotes(com->currentConstraint->id));
	  comPrint("(");
	  comTranslateTerm(formula->data.relation.op2, TRUE, TRUE, tmpIndex+1); 
	  comPrint("(");
	  comTranslateLexicalAccess(formula->data.relation.op1);
	  comPrint("->data.number %s ", operator); 
	  comTranslateNumber(formula->data.relation.op2, tmpIndex+1);
	  comPrintln("))");
	  break;
	case RTString:
	  comPrintln(" == VTError ? FALSE :"); /* should be a grammar error */
	  comTranslateLexicalAccess(formula->data.relation.op1);
	  comPrintln("->type != VTString ?");
	  comPrintln("(cdgPrintf(CDG_ERROR, "
	      "\"ERROR: 1rst argument not a string in formula `%s' in constraint `%s', line %%d\\n\", __LINE__), FALSE):", 
	      operator, comEscapeQuotes(com->currentConstraint->id));
	  comPrint("(");
	  comTranslateTerm(formula->data.relation.op2, TRUE, TRUE, tmpIndex+1); 
	  comPrint("(");
	  comTranslateLexicalAccess(formula->data.relation.op1);
	  comPrint("->data.string %s ", operator); 
	  comTranslateString(formula->data.relation.op2, tmpIndex+1);
	  comPrintln("))");
	  break;
	case RTPeek: /* first and second are lexical accesses */
	  comPrintln(" == VTError ? FALSE :"); /* should be a grammar error */
	  comTranslateTerm(formula->data.relation.op2, FALSE, TRUE, tmpIndex+1); 
	  comPrintln(" == VTError ? FALSE :"); /* should be a grammar error */
	  comTranslateLexicalAccess(formula->data.relation.op1);
	  comPrint("->type != ");
	  comTranslateLexicalAccess(formula->data.relation.op2);
	  comPrintln("->type ?");
	  comPrintln("(cdgPrintf(CDG_ERROR, "
	      "\"ERROR: type mismatch in formula `%s' in constraint `%s', line %%d\\n\", __LINE__), FALSE):",
	      operator, comEscapeQuotes(com->currentConstraint->id));
	  comPrint("((");
	  comTranslateLexicalAccess(formula->data.relation.op1);
	  comPrintln("->type == VTString)? ");
	  comPrint("(");
	  comTranslateLexicalAccess(formula->data.relation.op1);
	  comPrint("->data.string %s ", operator);
	  comTranslateLexicalAccess(formula->data.relation.op2);
	  comPrintln("->data.string) : ");
	  comPrint("(");
	  comTranslateLexicalAccess(formula->data.relation.op1);
	  comPrint("->data.number %s ", operator);
	  comTranslateLexicalAccess(formula->data.relation.op2);
	  comPrintln("->data.number))");
	  break;
	default:
	  if (formula->type == FTEqual) {
	    comPrintln("FALSE");
	    comWriteWarning("comparing type `%s' to `%s'",
		comReturnTypeToString(returnType1),
		comReturnTypeToString(returnType2));
	  } else {
	    comPrintln("TRUE");
	    comWriteWarning("comparing type `%s' to `%s'",
		comReturnTypeToString(returnType1),
		comReturnTypeToString(returnType2));
	  }
      }
      break;
    default:
      comPrintln("FALSE");
      comWriteError("unknown return type `%s'",
	  comReturnTypeToString(returnType1)); /* never reach */
  }
  comOutdent();
  comPrint(")");
}
  
/* ----------------------------------------------------------------------
 * translate the <, <=, >, >= formulas
 */
void comTranslateUnEquation(Formula formula, int tmpIndex)
{
  ReturnType returnType1 = comReturnTypeOfTerm(formula->data.relation.op1) & ~RTNoError;
  ReturnType returnType2 = comReturnTypeOfTerm(formula->data.relation.op2) & ~RTNoError;
  String operator = "horst";

  switch (formula->type) {
    case FTGreater:
      comPrintln("( /* greater */ ");
      operator = ">";
      break;
    case FTGreaterEqual:
      comPrintln("( /* greater equal */");
      operator = ">=";
      break;
    case FTLess:
      comPrintln("( /* less */");
      operator = "<";
      break;
    case FTLessEqual:
      comPrintln("( /* less equal */");
      operator = "<=";
      break;
    default: /* never reach */
      break;
  }
  comIndent();
  comTranslateTerm(formula->data.relation.op1, TRUE, TRUE, tmpIndex);

  /* check known return types */
  switch(returnType1) {
    case RTPeek:
      switch(returnType2) {
	case RTPeek:
	  comPrint(" != VTNumber || (");
	  comTranslateTerm(formula->data.relation.op2, TRUE, TRUE, tmpIndex+1);
	  comPrint(" != VTNumber ?");
#if 0
	  comPrintln("(cdgPrintf(CDG_ERROR, "
	      "\"type mismatch in formula `%s' in constraint `%s'\\n\"), FALSE) :", 
	      operator, comEscapeQuotes(com->currentConstraint->id));
#else
	  comPrintln("FALSE :");
#endif
	  comPrint("(");
	  comTranslateLexicalAccess(formula->data.relation.op1);
	  comPrint("->data.number %s ", operator);
	  comTranslateLexicalAccess(formula->data.relation.op2);
	  comPrintln("->data.number))");
	  break;
	case RTNumber:
	  comPrintln(" != VTNumber ?", tmpIndex);
#if 0
	  comPrintln("(cdgPrintf(CDG_ERROR, "
	      "\"type mismatch in formula `%s' in constraint `%s'\\n\"), FALSE) :", 
	      operator, comEscapeQuotes(com->currentConstraint->id));
#else
	  comPrintln("FALSE :");
#endif
	  comPrint("(");
	  comTranslateTerm(formula->data.relation.op2, TRUE, TRUE, tmpIndex+1);
	  comPrint("(");
	  comTranslateLexicalAccess(formula->data.relation.op1);
	  comPrint("->data.number %s ", operator); 
	  comTranslateNumber(formula->data.relation.op2, tmpIndex+1);
	  comPrintln("))");
	  break;
	default: /* never reach */
	  comWriteError("2nd argument has an illegal return type `%s' in formula `%s'",
	      comReturnTypeToString(returnType2), operator);
      }
      break;
    case RTNumber:
      switch(returnType2) {
	case RTPeek: 
	  comPrint("(");
	  comTranslateTerm(formula->data.relation.op2, TRUE, TRUE, tmpIndex+1);
	  comPrintln(" != VTNumber ?", tmpIndex);
#if 0
	  comPrintln("(cdgPrintf(CDG_ERROR, "
	      "\"type mismatch in formula `%s' in constraint `%s'\\n\"), FALSE) :", 
	      operator, comEscapeQuotes(com->currentConstraint->id));
#else
	  comPrintln("FALSE :");
#endif
	  comPrint("("); 
	  comTranslateNumber(formula->data.relation.op1, tmpIndex);
	  comPrint(" %s ", operator);
	  comTranslateLexicalAccess(formula->data.relation.op2);
	  comPrintln("->data.number))");
	  break;
	case RTNumber:
	  comPrint("("); 
	  comTranslateTerm(formula->data.relation.op2, TRUE, TRUE, tmpIndex+1);
	  comPrint("("); 
	  comTranslateNumber(formula->data.relation.op1, tmpIndex);
	  comPrint(" %s ", operator);
	  comTranslateNumber(formula->data.relation.op2, tmpIndex+1);
	  comPrintln("))");
	  break;
	default: /* never reach */
	  comWriteError("2nd argument has an illegal return type `%s' in formula `%s'",
	      comReturnTypeToString(returnType2), operator);
      }
      break;
    default: /* never reach */
      comWriteError("1rst argument has an illegal return type `%s' in formula `%s'",
	  comReturnTypeToString(returnType1), operator);
  }

  comOutdent();
  comPrint(")");
}

/* ----------------------------------------------------------------------
 * translate the direction formulas
 */
void comTranslateDirection(Formula formula)
{
    comPrintln("( /* direction `%s' */", 
	comDirectionToString(formula->data.direction.d));
    comIndent();

    /* anydir */
    if (formula->data.direction.d == AnyDir) {
      comPrintln("TRUE");
    } 
    
    /* nonnil */
    else if (formula->data.direction.d == NonNil) {
	comPrintln("lv%d->regents != NULL",
	    comIndexOfVarInfo(formula->data.direction.var));
    } 
    
    /* nil */
    else if (formula->data.direction.d == Nil) {
      comPrintln("lv%d->regents == NULL",
	  comIndexOfVarInfo(formula->data.direction.var));
    } 
    
    /* left */
    else if (formula->data.direction.d == Left) {
      comPrintln("lv%d->regents != NULL && lv%d->regents != NONSPEC &&",
	  comIndexOfVarInfo(formula->data.direction.var),
	  comIndexOfVarInfo(formula->data.direction.var));
      comPrintln("((LexemNode)(lv%d->regents->item))->arc->from < ((LexemNode)(lv%d->dependents->item))->arc->from",
	  comIndexOfVarInfo(formula->data.direction.var),
	  comIndexOfVarInfo(formula->data.direction.var));
    }

    /* right */
    else if (formula->data.direction.d == Right) {
      comPrintln("lv%d->regents != NULL &&",
	  comIndexOfVarInfo(formula->data.direction.var));
      comPrint("(lv%d->regents == NONSPEC || ",
	  comIndexOfVarInfo(formula->data.direction.var));
      comPrintln("((LexemNode)(lv%d->regents->item))->arc->from >= ((LexemNode)(lv%d->dependents->item))->arc->from)",
	  comIndexOfVarInfo(formula->data.direction.var),
	  comIndexOfVarInfo(formula->data.direction.var));
    } else { 
      /* never reach */
      comWriteError("illegal direction");
    }
    comOutdent();
    comPrint(")");
}

/* ----------------------------------------------------------------------
 * translate a formula to c code
 */
void comTranslateFormula(Formula formula, int tmpIndex)
{
  ReturnType returnType1;
  String operator = "horst";
  int nextTmpIndex;

  com->currentFormula = formula;
  returnType1 = comReturnTypeOfFormula(formula);
  if (returnType1 == RTError)
    return;
  
  if (returnType1 & ~RTNoError & ~RTBoolean) {
    comPrint("FALSE");
    comWriteWarning("illegal formula of type `%s'",
	comFormulaTypeToString(formula));
    return;
  }
  
  switch(formula->type) {
    /* ---------------------------------------------------------------------- */
    case FTConstant:
      comPrint("%s", (formula->data.constant)?"TRUE":"FALSE");
      break;
    /* ---------------------------------------------------------------------- */
    case FTNegation:
      comPrintln("!( /* negation */");
      comIndent();
      comTranslateFormula(formula->data.formula, tmpIndex);
      comOutdent();
      comPrintln("");
      comPrint(")");
      break;
    /* ---------------------------------------------------------------------- */
    case FTEqual:
    case FTNotEqual:
      comTranslateEquation(formula, tmpIndex);
      break;
    /* ---------------------------------------------------------------------- */
    case FTGreater:
    case FTGreaterEqual:
    case FTLess:
    case FTLessEqual:
      comTranslateUnEquation(formula, tmpIndex);
      break;
    /* ---------------------------------------------------------------------- */
    case FTConjunction:
    case FTDisjunction:
    case FTImplication:
    case FTBiimplication:
      nextTmpIndex = tmpIndex;
      switch(formula->type) {
	case FTConjunction:
	  comPrintln("( /* and */");
	  operator = "&&";
	  break;
	case FTDisjunction:
	  comPrintln("( /* or */");
	  operator = "||";
	  break;
	case FTImplication:
	  comPrintln("(!( /* impl */");
	  operator = "||";
	  break;
	case FTBiimplication:
	  comPrintln("( /* biimpl */");
	  operator = "==";
	  nextTmpIndex = tmpIndex + comFormulaDepth(formula->data.junction.arg1);
	  break;
	default: /* never reach */
	  break;
      }
      comIndent(); /* TODO: must we check VTError here? */
      comTranslateFormula(formula->data.junction.arg1, tmpIndex);
      if (formula->type == FTImplication) {
	comPrint(")");
      } 
      comPrintln(" %s ", operator);
      comTranslateFormula(formula->data.junction.arg2, nextTmpIndex);
      comOutdent();
      comPrintln("");
      comPrint(")");
      break;
    /* ---------------------------------------------------------------------- */
    case FTPredicate:
      comTranslatePredicate(formula->data.predicate.functor, 
	  formula->data.predicate.args, tmpIndex);
      break;
    /* ---------------------------------------------------------------------- */
    case FTConnexion:
      comTranslateConnexion(formula);
      break;
    /* ---------------------------------------------------------------------- */
    case FTDirection:
      comTranslateDirection(formula);
      break;
  }
}

/* ----------------------------------------------------------------------
 * translate a connexion
 * in case of the unsymmetric connexions 'Under' & 'Over' a distinction between
 * the order (lv1, lv2) and (lv2, lv1) is made. in the other cases this doesn't
 * matter.
 */
void comTranslateConnexion(Formula formula) {
      int lv1Index = comIndexOfVarInfo(formula->data.connexion.var1);
      int lv2Index = comIndexOfVarInfo(formula->data.connexion.var2);

      comPrintln("( /* connexion `%s' */", 
	  comConnextionToString(formula->data.connexion.c));
      comIndent();

      /* AnyCon */
      if (formula->data.connexion.c == AnyCon) {
	comPrintln("TRUE");
      }

      /* Top, Bottom */
      else if (formula->data.connexion.c == Top || formula->data.connexion.c == Bottom) {
	comPrintln("_getConnextion(lv1, lv2) == %s || _getConnextion(lv1, lv2) == Identical", 
	    comConnextionToString(formula->data.connexion.c));
      }
      
      /* Over, Under */
      else if (formula->data.connexion.c == Over || formula->data.connexion.c == Under) {
	comPrintln("_getConnextion(lv%d, lv%d) == %s || _getConnextion(lv%d, lv%d) == Inverse", 
	    lv1Index, lv2Index, comConnextionToString(formula->data.connexion.c), lv1Index, lv2Index);
      }

      /* None, Inverse, Identical */
      else {
	comPrintln("_getConnextion(lv1, lv2) == %s", 
	    comConnextionToString(formula->data.connexion.c));
      }
      comOutdent();
      comPrintln(")");
}

/* ----------------------------------------------------------------------
 * translate each unary constraint into a separate c-function
 */
void comTranslateUnaryConstraints(void)
{
  Constraint constraint;
  HashIterator hi = hashIteratorNew(inputCurrentGrammar->constraints);

  while((constraint = hashIteratorNextValue(hi))) {
    if (constraint->sig->arity != 1) 
      continue;
    comTranslateConstraint(constraint);
  }
  hashIteratorDelete(hi);
  assert(com->indent == 0);
}

/* ----------------------------------------------------------------------
 * translate a constraint into a c-function
 */
void comTranslateConstraint(Constraint constraint)
{
  ReturnType returnType;
  com->currentConstraint = constraint;

  /* write header */
  comPrintln("");
  comPrintln("/* ----------------------------------------------------------------------");
  comPrintln(" * evaluate constraint %s", comEscapeQuotes(constraint->id));
  comPrintln("*/");
  if (constraint->sig->arity == 1) {
    comPrintln("Boolean _evalConstraint_%d(LexemGraph lg, Vector context, LevelValue lv1)", constraint->no);
  } else {
    comPrintln("Boolean _evalConstraint_%d(LexemGraph lg, Vector context,  LevelValue lv1, LevelValue lv2)", constraint->no);
  }
  comPrintln("{");
  comIndent();

  /* write some comments */
  comPrintln("");
  comPrintln("/*");
  comPrintln(" * constraint: %s", comEscapeQuotes(constraint->id));
  comPrintln(" * no        : %d", constraint->no);
  comPrintln(" * filename  : %s", constraint->filename);
  comPrintln(" * lineno    : %d", constraint->lineNo);
  if (constraint->sig->arity == 1) {
    comPrintln(" * Signature : %s:%s",
	constraint->sig->l1->id, comDirectionToString(constraint->sig->d1));
  } else {
    comPrintln(" * Signature : %s:%s %s %s:%s",
	constraint->sig->l1->id, comDirectionToString(constraint->sig->d1),
	comConnextionToString(constraint->sig->conn),
	constraint->sig->l2->id, comDirectionToString(constraint->sig->d2));
  }

  comPrintln(" */");
  comPrintln("");

  /* the formula */
  comPrintln("if(!");
  comIndent();
  comTranslateFormula(constraint->formula, 0);
  comOutdent();
  comPrintln("");
  comPrintln(") {");
  comIndent();

  /* the penalty */
  if (constraint->penaltyTerm != NULL) {
    comPrintln("constraints[%d].constraint->penalty = (",
	comIndexOfConstraint(constraint->id));
    comIndent();
    returnType = comReturnTypeOfTerm(constraint->penaltyTerm);
    if (!(returnType & RTNoError)) {
      comTranslateTerm(constraint->penaltyTerm, FALSE, FALSE, 0);
      comPrintln(" == VTError ? 0.0:", 0);
    }
    returnType &= ~RTNoError;

    if (returnType == RTPeek) {
      comTranslateTerm(constraint->penaltyTerm, FALSE, FALSE, 0);
      comPrint(" != VTNumber ? 0.0 : ");
      comTranslateLexicalAccess(constraint->penaltyTerm);
      comPrintln("->data.number");
    } else {
      comTranslateTerm(constraint->penaltyTerm, FALSE, TRUE, 0);
      comTranslateNumber(constraint->penaltyTerm, 0);
      comPrintln("");
    }
    comOutdent();
    comPrintln(");");
  }
  
  comPrintln("return FALSE;");
  comOutdent();
  comPrintln("}\n");

  com->currentConstraint = NULL;
  comPrintln("");
  comPrintln("return TRUE;");
  comOutdent();
  comPrintln("}");
}

/* ----------------------------------------------------------------------
 * translate each binary constraint into a separate c-function
 */
void comTranslateBinaryConstraints(void)
{
  HashIterator hi = hashIteratorNew(inputCurrentGrammar->constraints);
  Constraint constraint;
    
  while((constraint = hashIteratorNextValue(hi))) {
    if (constraint->sig->arity != 2) 
      continue;
    comTranslateConstraint(constraint);
  }
  hashIteratorDelete(hi);
  assert(com->indent == 0);
}

/* ----------------------------------------------------------------------
 * translate the current cdg to C
 */
Boolean comTranslate(void)
{
  Boolean success = TRUE;

  cdgPrintf(CDG_INFO, "INFO: translating current grammar to `%s'\n", 
      com->ccFileName);  
  cdgFlush();

  /* write openers */
  comWriteHeader();

  /* analyze the grammar */
  comAnalyzeGrammar();

  /* write some declarations */
  comWriteDeclarations();

  /* translate every constraint */
  comTranslateUnaryConstraints();
  comTranslateBinaryConstraints();

  /* write helper functions */
  comWriteFunctions();

  return success;
}

/* ----------------------------------------------------------------------
 * compile a shared object from the recently translated code
 */
Boolean comMake(void)
{
  String compileCmd = strAppend(makeInfo->cc, " ",
				makeInfo->cFlags, " ",
				makeInfo->includes, " -c ",
				com->ccFileName, NULL); // this is a registered string now
  String linkCmd = strAppend(makeInfo->ld, " ",
			     makeInfo->ldFlags, " ",
			     makeInfo->ldLibs, " ",
			     com->objFileName, " -o ", 
			     com->soFileName, NULL); // this is a registered string now
  Boolean isOk = TRUE;

  /* call the compiler */
  cdgPrintf(CDG_INFO, "INFO: compiling `%s' to `%s'\n",
	    com->ccFileName, com->objFileName);
  cdgPrintf(CDG_DEBUG, "DEBUG: %s\n", compileCmd);
  cdgFlush();
  if (system(compileCmd) != 0) {
    cdgPrintf(CDG_ERROR, "ERROR: compiler error, message should have been provided\n");
    isOk = FALSE;
  } else {

    /* call the linker */
    cdgPrintf(CDG_INFO, "INFO: linking `%s' to `%s'\n",
	      com->objFileName, com->soFileName);
    cdgPrintf(CDG_DEBUG, "DEBUG: %s\n", linkCmd);
    cdgFlush();
    if (system(linkCmd) != 0) {
      cdgPrintf(CDG_ERROR, "ERROR: linker error, message should have been provided\n");
      isOk = FALSE;
    }
  }
  return isOk;
}

/* ----------------------------------------------------------------------------
 * load a dll grammar
 */
Input comLoad(String filename)
{
  void *dllHandle;
  String sofilename;
  Input input = inputNewInput();

  if (filename == NULL) {
    sofilename = com->soFileName;
  } else {
    sofilename = filename;
  }

  cdgPrintf(CDG_INFO, "INFO: loading `%s'\n", sofilename);

  /* get the dll handle */
  dllHandle = dlopen(sofilename, RTLD_NOW);
  if (!dllHandle) {
    cdgPrintf(CDG_ERROR, "ERROR: can't load `%s': %s\n", sofilename, dlerror());
    return NULL;
  }

  inputRegisterFilename(input, sofilename);
  input->dllHandle = dllHandle;

  return input;
}

/* ----------------------------------------------------------------------------
 * check the return type of a predicate
 */
ReturnType comReturnTypeOfPredicate(String functor, List args)
{
  Term term;
  ReturnType returnType1, returnType2, returnType3;

  /* start, stop, root, spec, nonspec */
  if (strcmp("start", functor) == 0 ||
      strcmp("stop", functor) == 0 ||
      strcmp("root", functor) == 0 ||
      strcmp("nonspec", functor) == 0 ||
      strcmp("spec", functor) == 0) {
    if (listSize(args) != 1 || args == NULL || args->item == NULL) {
      comWriteWarning("`%s' needs one argument", functor);
      return RTError;
    }
    returnType1 = comReturnTypeOfTerm(args->item) & ~RTLexemPosition;
    if (returnType1 == RTError)
      return RTError;
    if (returnType1  == (RTLexemNode | RTNoError)) {
      return RTBoolean | RTNoError;
    } else if (returnType1 == RTLexemNode) {
      return RTBoolean;
    }
    comWriteWarning("argument of type `%s' isn't a lexemnode in `%s'",
	comReturnTypeToString(returnType1), functor);
    return RTError;
  } 

  /* print */
  if (strcmp("print", functor) == 0) {
    return RTBoolean | RTNoError;
  } 

  /* subsumes, compatible */
  if (strcmp("subsumes", functor) == 0 ||
      strcmp("compatible", functor) == 0) {
    if (listSize(args) != 3) {
      comWriteWarning("predicate `%s' needs three arguments", functor);
      return RTError;
    }
    
    returnType1 = comReturnTypeOfTerm(listNthElement(args, 1)) & ~RTNoError;
    if (returnType1 == RTError)
      return RTError;
    if (!(returnType1 & (RTString | RTPeek))) {
      comWriteWarning("1rst argument of type `%s' isn't a string in `%s'",
	  comReturnTypeToString(returnType1), functor); 
      return RTError;
    }
    
    returnType2 = comReturnTypeOfTerm(listNthElement(args, 2)) & ~RTNoError;
    if (returnType2 == RTError)
      return RTError;
    if (!(returnType2 & (RTString | RTPeek))) {
      comWriteWarning("2nd argument of type `%s' isn't a string in `%s'",
	  comReturnTypeToString(returnType2), functor); 
      return RTError;
    }

    returnType3 = comReturnTypeOfTerm(listNthElement(args, 3)) & ~RTNoError;
    if (returnType3 == RTError)
      return RTError;
    if (!(returnType2 & (RTString | RTPeek))) {
      comWriteWarning("3nd argument of type `%s' isn't a string in `%s'",
	  comReturnTypeToString(returnType3), functor); 
      return RTError;
    }
    return RTBoolean;
  } 

  /* exists */
  if (strcmp("exists", functor) == 0) {
    if (listSize(args) != 1 || args == NULL || args->item == NULL) {
      comWriteWarning("predicate `exists' needs one argument");
      return RTError;
    } 
    term = (Term)args->item;
    if (term->type != TTTopPeek && term->type != TTBottomPeek) {
      comWriteWarning("predicate `exists' needs a lexical access");
      return RTError;
    } 
    return RTBoolean | RTNoError;
  } 
  
  /* chunk head */
  if (strcmp("chunk_head", functor) == 0) {
    if (listSize(args) != 1) {
      comWriteWarning("'%s' needs one argument", functor);
      return RTError;
    }
    returnType1 = (comReturnTypeOfTerm(args->item) & ~RTLexemPosition);
    if (returnType1 == RTError) {
      return RTError;
    }
    if (returnType1 == (RTLexemNode | RTNoError)) {
      return RTBoolean;
    }
    comWriteWarning("argument of type `%s' isn't a lexemnode in `%s'",
	comReturnTypeToString(returnType1), functor);
    return RTError;
  } 

  /* under */
  if (strcmp("under", functor) == 0) {
    if (listSize(args) != 2) {
      comWriteWarning("'%s' needs two arguments", functor);
      return RTError;
    }
    term = (Term)args->item;
    returnType1 = (comReturnTypeOfTerm(term) & ~RTLexemPosition);
    if (returnType1 == RTError) {
      return RTError;
    }
    term = (Term)args->item;
    returnType2 = (comReturnTypeOfTerm(term) & ~RTLexemPosition);
    if (returnType2 == RTError) {
      return RTError;
    }
    if (returnType1 == (RTLexemNode | RTNoError)) {
      if (returnType2 == (RTLexemNode | RTNoError)) {
	return RTBoolean;
      }
    }
    comWriteWarning("argument of type `%s' isn't a lexemnode in `%s'",
	comReturnTypeToString(returnType1), functor);
    return RTError;
  }

  /* between */
  if (strcmp("between", functor) == 0) {
    if (listSize(args) != 3) {
      comWriteWarning("'%s' needs 3 arguments", functor);
      return RTError;
    }
    term = (Term)args->item;
    returnType1 = comReturnTypeOfTerm(term);
    term = (Term)args->next->item;
    returnType2 = comReturnTypeOfTerm(term);
    term = (Term)listNthElement(args, 3);
    returnType3 = comReturnTypeOfTerm(term);

    if ((returnType1 | returnType2 | returnType3) & RTError) {
      return RTError;
    }
    if (returnType1 & (RTLexemNode | RTNumber)) {
      if (returnType2 & (RTLexemNode | RTNumber)) {
	if (returnType3 & RTString) {
	  return RTBoolean;
	}
	comWriteWarning("third argument of type `%s' isn't a string in `%s'",
	    comReturnTypeToString(returnType3), functor);
	return RTError;
      }
      comWriteWarning("second argument of type `%s' isn't a lexemnode or number in `%s'",
	  comReturnTypeToString(returnType2), functor);
      return RTError;
    }
    comWriteWarning("first argument of type `%s' isn't a lexemnode or number in `%s'",
	comReturnTypeToString(returnType1), functor);
    return RTError; 
  }

  /* is */
  if (strcmp("is", functor) == 0) {
    if ((listSize(args) < 2) || (listSize(args) > 3)) {
      comWriteWarning("'%s' needs 2 or 3 arguments", functor);
      return RTError;
    }
    returnType1 = comReturnTypeOfTerm(args->item) & ~RTNoError;
    if (returnType1 & RTError) {
      return RTError;
    }
    if (! (returnType1 & RTLexemNode)) {
      comWriteWarning("first argument of type `%s' isn't a lexemnode in `%s'",
	  comReturnTypeToString(returnType1), functor);
      return RTError; 
    }
    returnType2 = comReturnTypeOfTerm(args->next->item) & ~RTNoError;
    if (returnType2 & RTError) {
      return RTError;
    }
    if (! (returnType2 & RTString)) {
      comWriteWarning("second argument of type `%s' isn't a string in `%s'",
	  comReturnTypeToString(returnType1), functor);
      return RTError; 
    }
    /* optional third argument */
    if (listSize(args) == 3) {
      returnType3 = comReturnTypeOfTerm(listNthElement(args, 3)) & ~RTNoError;
      if (! (returnType3 & RTString)) {
	comWriteWarning("third argument of type `%s' isn't a string in `%s'",
	    comReturnTypeToString(returnType1), functor);
	return RTError; 
      }
    }
    return RTBoolean;
  }
  
  /* connected */
  if (strcmp("connected", functor) == 0) {
    if (listSize(args) != 2) {
      comWriteWarning("'%s' needs 2 arguments", functor);
      return RTError;
    }
    term = (Term)args->item;
    returnType1 = (comReturnTypeOfTerm(term) & ~RTLexemPosition);
    if (returnType1 == RTError) {
      return RTError;
    }
    if (! (returnType1 & RTLexemNode)) {
      comWriteWarning("first argument of type `%s' isn't a lexem node in `%s'",
	  comReturnTypeToString(returnType1), functor);
      return RTError; 
    }
    term = (Term)args->next->item;
    returnType2 = (comReturnTypeOfTerm(term) & ~RTLexemPosition);
    if (returnType2 == RTError) {
      return RTError;
    }
    if (! (returnType2 & RTLexemNode)) {
      comWriteWarning("second argument of type `%s' isn't a lexem node in `%s'",
	  comReturnTypeToString(returnType2), functor);
      return RTError; 
    }
    return RTBoolean;
  }

  /* has */
  if (strcmp("has", functor) == 0) {
    if ((listSize(args) < 2) || (listSize(args) > 6)) {
      comWriteWarning("'%s' needs 2 to 6 arguments", functor);
      return RTError;
    }
    /* first argument */
    term = (Term)listNthElement(args, 1);
    returnType1 = comReturnTypeOfTerm(term);
    if (returnType1 & RTError) {
      return RTError;
    }
    if (! (returnType1 & RTLexemNode)) {
      comWriteWarning("first argument of type `%s' isn't a lexemnode in `%s'",
	  comReturnTypeToString(returnType1), functor);
      return RTError;
    }
    /* second argument */
    term = (Term)listNthElement(args, 2);
    returnType1 = comReturnTypeOfTerm(term);
    if (returnType1 & RTError) {
      return RTError;
    }
    if (! (returnType1 & (RTString | RTPeek))) {
      comWriteWarning("second argument of type `%s' isn't a string in `%s'",
	  comReturnTypeToString(returnType1), functor);
      return RTError;
    }
    /* third argument */
    if (listSize(args) > 2) {
      term = (Term)listNthElement(args, 3);
      returnType1 = comReturnTypeOfTerm(term);
      if (returnType1 & RTError) {
	return RTError;
      }
      if (! (returnType1 & RTString)) {
	comWriteWarning("third argument of type `%s' isn't a string in `%s'",
	    comReturnTypeToString(returnType1), functor);
	return RTError;
      }
    }
    /* fourth argument */
    if (listSize(args) > 3) {
      term = (Term)listNthElement(args, 4);
      returnType1 = comReturnTypeOfTerm(term);
      if (returnType1 & RTError) {
	return RTError;
      }
      if (! (returnType1 & RTString)) {
	comWriteWarning("fourth argument of type `%s' isn't a string in `%s'",
	    comReturnTypeToString(returnType1), functor);
	return RTError;
      }
    }
    /* fifth argument */
    if (listSize(args) > 4) {
      term = (Term)listNthElement(args, 5);
      returnType1 = comReturnTypeOfTerm(term);
      if (returnType1 & RTError) {
	return RTError;
      }
      if (! (returnType1 & RTNumber)) {
	comWriteWarning("fifth argument of type `%s' isn't a number in `%s'",
	    comReturnTypeToString(returnType1), functor);
	return RTError;
      }
    }
    /* sixth argument */
    if (listSize(args) > 5) {
      term = (Term)listNthElement(args, 6);
      returnType1 = comReturnTypeOfTerm(term);
      if (returnType1 & RTError) {
	return RTError;
      }
      if (! (returnType1 & RTNumber)) {
	comWriteWarning("sixth argument of type `%s' isn't a number in `%s'",
	    comReturnTypeToString(returnType1), functor);
	return RTError;
      }
    }
    return RTBoolean;
  }
  
  comWriteError("unknown predicate `%s'", functor);
  return RTError;
}

/* ----------------------------------------------------------------------------
 * check the return type of a function
 */
ReturnType comReturnTypeOfFunction(String functor, List args)
{
  List l;
  Term term;
  ReturnType returnType1, returnType2, returnType3;
  int i;
  
  /* distance */
  if (strcmp("distance", functor) == 0) {
    if (listSize(args) != 2 ||
     	args->item == NULL ||
       	args->next->item == NULL) {
      comWriteWarning("function `distance' needs two lexemnodes as arguments");
      return RTError;
    } 
    returnType1 = comReturnTypeOfTerm(listNthElement(args, 1)) & ~RTLexemPosition;
    if (returnType1 == RTError)
      return RTError;
    if (returnType1 & ~RTNoError & ~RTLexemNode) {
      comWriteWarning(
	  "1rst argument of type `%s' isn't a lexemnodes in `distance'",
	  comReturnTypeToString(returnType1));
      return RTError;
    } 
    returnType2 = comReturnTypeOfTerm(listNthElement(args, 2)) & ~RTLexemPosition;
    if (returnType2 == RTError)
      return RTError;
    if (returnType2 & ~RTNoError & ~RTLexemNode) {
      comWriteWarning(
      	  "2nd argument of type `%s' isn't a lexemnodes in `distance'",
	  comReturnTypeToString(returnType2));
      return RTError;
    } 

    if (returnType1 & RTNoError & returnType2) {
      return RTNumber | RTNoError;
    } else {
      return RTNumber;
    }
  } 
  
  /* abs */
  else if (strcmp("abs", functor) == 0) {
    if (listSize(args) != 1 || args->item == NULL) {
      comWriteWarning("function `abs' needs one argument");
      return RTError;
    } else {
      returnType1 = comReturnTypeOfTerm(args->item);

      if (returnType1 == RTError)
	return RTError;
      if (returnType1 & ~RTNoError & ~RTNumber) {
	comWriteWarning("argument type `%s' isn't numerical in `abs'",
	    comReturnTypeToString(returnType1));
	return RTError;
      }

      if (returnType1 & RTNoError) {
	return RTNumber | RTNoError;
      } else {
	return RTNumber;
      }
    }
  } 
  
  /* min, max */
  else if (strcmp("min", functor) == 0 ||
      strcmp("max", functor) == 0) {
    if (listSize(args) <= 1) {
      comWriteWarning("function `%s' needs at least two arguments",
	  functor);
      return RTError;
    } else {
      Boolean errorPossible = FALSE;

      /* check the arguments */
      for (l = args, i = 0; l != NULL; i++, l = l->next) {
	term = (Term)l->item;
	returnType1 = comReturnTypeOfTerm(term);
	if (returnType1 == RTError)
	  return RTError;
	if (returnType1 & ~RTNoError & ~RTNumber) {
	  comWriteWarning("%d. argument type `%s' isn't numerical in `%s'",
	      i, comReturnTypeToString(returnType1), functor);
	  return RTError;
	}
	errorPossible &= !(returnType1 & RTNoError);
      }

      if (errorPossible) {
	return RTNumber;
      }  else {
	return RTNumber | RTNoError;
      }
    }
  } 
  
  /* match */
  else if (strcmp("match", functor) == 0) {
    if (listSize(args) != 3) {
      comWriteWarning("function `match' needs three arguments");
      return RTError;
    }
    returnType1 = comReturnTypeOfTerm(listNthElement(args, 1)) & ~RTNoError;
    if (returnType1 == RTError)
      return RTError;
    if (returnType1 & ~RTString) {
      comWriteWarning("1rst argument of type `%s' isn't a string in `match'",
	  comReturnTypeToString(returnType1));
      return RTError;
    }
    returnType2 = comReturnTypeOfTerm(listNthElement(args, 2)) & ~RTNoError;
    if (returnType2 == RTError)
      return RTError;
    if (returnType2 & ~RTList & ~RTString & ~RTPeek) {
      comWriteWarning("2nd argument of type `%s' isn't a string in `match'",
	  comReturnTypeToString(returnType2));
      return RTError;
    }
    returnType3 = comReturnTypeOfTerm(listNthElement(args, 3)) & ~RTNoError;
    if (returnType3 == RTError)
      return RTError;
    if (returnType3 & ~RTString & ~RTPeek) {
      comWriteWarning("3rd argument of type `%s' isn't a string in `match'",
	  comReturnTypeToString(returnType3));
      return RTError;
    }
    return RTNumber;
  } 
  
  /* phraseqotes */
  else if (strcmp("phrasequotes", functor) == 0) {
    if (listSize(args) != 1) {
      comWriteWarning("function `match' needs one argument");
      return RTError;
    }
    returnType1 = comReturnTypeOfTerm(args->item) & ~RTNoError;
    if (returnType1 & RTLexemNode) {
      return RTNumber;
    }
    if (returnType1 & ~RTLexemNode & ~RTLexemPosition) {
      comWriteWarning("argument of function 'phrasequotes' not a lexeme node");
      return RTError;
    }
    return RTError; // never reach
  }

  /* lookup */
  else if (strcmp("lookup", functor) == 0) {
    if (listSize(args) < 2) {
      comWriteWarning("function `lookup' needs 2 or more arguments");
      return RTError;
    } else {
      Boolean errorPossible = FALSE;

      /* check first argument */
      returnType1 = comReturnTypeOfTerm((Term)args->item);
      if (returnType1 & RTError)
	return RTError;
      if (returnType1 & ~RTNoError & ~RTString) {
	comWriteWarning("First argument type `%s' isn't a string in 'lookup'",
	    comReturnTypeToString(returnType1));
	return RTError;
      }
      
      /* check the other arguments */
      for (l = args->next, i = 2; l != NULL; i++, l = l->next) {
	term = (Term)l->item;
	returnType1 = comReturnTypeOfTerm(term);
	if (returnType1 == RTError)
	  return RTError;
	if (returnType1 & ~RTNoError & ~RTString & ~RTPeek) {
	  comWriteWarning("%d. argument type `%s' isn't a string or peek in 'lookup'",
	      i, comReturnTypeToString(returnType1));
	  return RTError;
	}
	errorPossible &= !(returnType1 & RTNoError);
      }
      
      if (errorPossible) {
	return RTString;
      }  else {
	return RTString | RTNoError;
      }
    }
  }

  /* quotes */
  else if (strcmp("quotes", functor) == 0) {
    if (listSize(args) != 1) {
      comWriteWarning("function `quotes' needs one argument");
      return RTError;
    }
    term = (Term)args->item;
    returnType1 = comReturnTypeOfTerm(term);
    if (returnType1 & RTError) {
      return RTError;
    }
    if (returnType1 & ~RTLexemNode & ~RTLexemPosition & ~RTNoError) {
      comWriteWarning("First argument type `%s' isn't a LexemNode in 'quotes'",
	  comReturnTypeToString(returnType1));
      return RTError;
    }
    return RTNumber;
  }


  /* parent */
  else if (strcmp("parent", functor) == 0) {
    if (listSize(args) != 1) {
      comWriteWarning("function `parent' needs one argument");
      return RTError;
    }
    term = (Term)args->item;
    returnType1 = comReturnTypeOfTerm(term);
    if (returnType1 & RTError) {
      return RTError;
    }
    if (returnType1 & ~RTLexemNode & ~RTLexemPosition & ~RTNoError) {
      comWriteWarning("First argument type `%s' isn't a LexemNode in 'parent'",
	  comReturnTypeToString(returnType1));
      return RTError;
    }
    return RTNumber;
  }

  /* pts */
  else if (strcmp("parens", functor) == 0 ||
           strcmp("pts", functor) == 0 ||
           strcmp("height", functor) == 0) {
    if (listSize(args) != 1) {
      comWriteWarning("function `pts' needs one argument");
      return RTError;
    }
    term = (Term)args->item;
    returnType1 = comReturnTypeOfTerm(term);
    if (returnType1 & RTError) {
      return RTError;
    }
    if (returnType1 & ~RTLexemNode & ~RTLexemPosition & ~RTNoError) {
      comWriteWarning("First argument type `%s' isn't a LexemNode in 'pts'",
	  comReturnTypeToString(returnType1));
      return RTError;
    }
    return RTNumber;
  }
   
  /* default */
  else {
    comWriteError("compile.c:%d: unknown function `%s'", __LINE__, functor);
    return RTError;
  }
}

/* ----------------------------------------------------------------------------
 * check the return type of a term 
 */
ReturnType comReturnTypeOfTerm(Term term)
{
  ReturnType returnType1, returnType2;

  switch(term->type) {
    case TTString:
    case TTLabel:
    case TTLevel:
      return RTString | RTNoError;
    case TTNumber:
      return RTNumber | RTNoError;
    case TTAdd:
    case TTSubtract:
    case TTMultiply:
    case TTDivide:
      returnType1 = comReturnTypeOfTerm(term->data.operation.op1);
      if (returnType1 == RTError) 
	return RTError;
      returnType2 = comReturnTypeOfTerm(term->data.operation.op2);
      if (returnType1 == RTError) 
	return RTError;

      if (returnType1 & ~RTNoError & (RTNumber | RTPeek)) {
	if (returnType2 & ~RTNoError & (RTNumber | RTPeek)) { 
	  if (returnType1 & returnType2 & RTNoError) {
	    return RTNumber | RTNoError;
	  } else {
	    return RTNumber;
	  }
	}
      }
      comWriteWarning("argument types `%s' and `%s' aren't numerical in `%s'",
	  comReturnTypeToString(returnType1),
	  comReturnTypeToString(returnType2),
	  comTermTypeToString(term));
      break;
    case TTFunction:
      return comReturnTypeOfFunction(
	  term->data.function.functor, term->data.function.args);
    case TTTopPeek:
      if (term->data.peek.path == NULL) {
	return RTAVNode;
      } else {
	String pathItem = term->data.peek.path->item;
	if (strcmp("id", pathItem) == 0) {
	  return RTLexemNode | RTNoError | RTLexemPosition;
	} else if (strcmp("from", pathItem) == 0) {
	  return RTNumber;
	} else if (strcmp("to", pathItem) == 0) {
	  return RTNumber;
	} else if (strcmp("word", pathItem) == 0) {
	  return RTString;
	} else {
	  return RTPeek | RTNoError; /* peek checks errors itself */
	}
      }
      break;
    case TTBottomPeek:
      if (term->data.peek.path == NULL) {
	return RTAVNode;
      } else {
	String pathItem = term->data.peek.path->item;
	if (strcmp("from", pathItem) == 0) {
	  return RTNumber | RTNoError;
	} else if (strcmp("to", pathItem) == 0) {
	  return RTNumber | RTNoError;
	} else if (strcmp("word", pathItem) == 0) {
	  return RTString | RTNoError;
	} else if (strcmp("id", pathItem) == 0) {
	  return RTLexemNode | RTNoError;
	} else {
	  return RTPeek | RTNoError; /* peek checks errors itself */
	}
      }
      break;
  }
  
  return RTError;
}

/* ----------------------------------------------------------------------------
 * check the return type of a formula 
 */
ReturnType comReturnTypeOfFormula(Formula formula)
{
  ReturnType returnType1, returnType2;
  
  switch (formula->type) {
    case FTConnexion:
    case FTDirection:
    case FTConstant:
      return RTBoolean | RTNoError;
    case FTNegation:
      return comReturnTypeOfFormula(formula->data.formula);
    case FTConjunction:
    case FTDisjunction:
    case FTImplication:
    case FTBiimplication:
      returnType1 = comReturnTypeOfFormula(formula->data.junction.arg1);
      if (returnType1 == RTError)
	return RTError;
      returnType2 = comReturnTypeOfFormula(formula->data.junction.arg2);
      if (returnType2 == RTError)
	return RTError;

      if (returnType1 & RTBoolean & ~RTNoError) {
	if (returnType2 & RTBoolean & ~RTNoError) { 
	  if (returnType1 & RTNoError & returnType2) {
	    return RTBoolean | RTNoError;
	  } else {
	    return RTBoolean;
	  }
	} 
      }
      comWriteWarning("argument types `%s' and `%s' aren't boolean in `%s'",
	  comReturnTypeToString(returnType1),
	  comReturnTypeToString(returnType2),
	  comFormulaTypeToString(formula));
      break;
    case FTEqual:
    case FTNotEqual:
      returnType1 = comReturnTypeOfTerm(formula->data.relation.op1) & ~RTLexemPosition;
      if (returnType1 == RTError)
	return RTError;
      returnType2 = comReturnTypeOfTerm(formula->data.relation.op2) & ~RTLexemPosition;
      if (returnType2 == RTError)
	return RTError;

      if (returnType1 & ~RTNoError & returnType2) {
	if (returnType1 & RTNoError & returnType2) {
	  return RTBoolean | RTNoError;
	} else {
	  return RTBoolean;
	}
      } 
      if (returnType1 & ~RTNoError & RTPeek) {
	if (returnType2 & ~RTLexemNode) {
	  return RTBoolean;
	}
      } else if (returnType1 & ~RTNoError & RTLexemNode) {
	if (returnType2 & ~RTNoError & ~RTPeek) {
	  return RTBoolean;
	}
      } else if (returnType2 & ~RTNoError & RTPeek) {
	return RTBoolean;
      }

      comWriteWarning("argument types `%s' and `%s' don't match in `%s'",
	  comReturnTypeToString(returnType1),
	  comReturnTypeToString(returnType2),
	  comFormulaTypeToString(formula));
      break;
    case FTGreater:
    case FTLess:
    case FTGreaterEqual:
    case FTLessEqual:
      returnType1 = comReturnTypeOfTerm(formula->data.relation.op1);
      if (returnType1 == RTError)
	return RTError;
      returnType2 = comReturnTypeOfTerm(formula->data.relation.op2);
      if (returnType2 == RTError)
	return RTError;

      if (returnType1 & ~RTNoError & (RTNumber | RTPeek)) {
	if (returnType2 & ~RTNoError  & (RTNumber | RTPeek)) { 
	  if (returnType1 & RTNoError & returnType2 & ~RTPeek) {
	    return RTBoolean | RTNoError;
	  } else {
	    return RTBoolean;
	  }
	} 
      }
      comWriteWarning("argument types `%s' and `%s' aren't numerical in `%s'",
	  comReturnTypeToString(returnType1),
	  comReturnTypeToString(returnType2),
	  comFormulaTypeToString(formula));
      break;
    case FTPredicate:
      return comReturnTypeOfPredicate(
	  formula->data.predicate.functor, formula->data.predicate.args);
 }

  return RTError;
}

/* ----------------------------------------------------------------------------
 * analyze the currently loaded grammar
 */
void comAnalyzeGrammar(void)
{
  HashIterator hi = hashIteratorNew(inputCurrentGrammar->constraints);
  Constraint constraint;
  int depth;
  int maxStrings;
  

  /* compute the number of variables needed to evaluate all constraints */
  com->maxValues = 0;
  while((constraint = hashIteratorNextValue(hi))) {
    depth = comConstraintDepth(constraint);
    if (depth > com->maxValues) {
      com->maxValues = depth;
    }
  }
  hashIteratorDelete(hi);

  /* compute the max. number of strings used in 'lookup' functions */
  hi = hashIteratorNew(inputCurrentGrammar->constraints);
  com->maxLookupStrings = 0;
  while((constraint = hashIteratorNextValue(hi))) {
    maxStrings = comMaxLookupStrings(constraint);
    if (maxStrings > com->maxLookupStrings) {
      com->maxLookupStrings = maxStrings;
    }
  }
  hashIteratorDelete(hi);
}

/* ----------------------------------------------------------------------------
 * compute the number of strings needed for evaluationg 'lookup' functions in
 * a constraint
 */
int comMaxLookupStrings(Constraint constraint) {
  int max1 = comMaxLookupStringsInTerm(constraint->penaltyTerm);
  int max2 = comMaxLookupStringsInFormula(constraint->formula);

  return max(max1, max2);
}

/* ----------------------------------------------------------------------------
 * compute the number of strings needed for evaluationg 'lookup' functions in
 * a formla
 */
int comMaxLookupStringsInFormula(Formula formula) {
  int tmp1 = 0, tmp2 = 0;

  switch (formula->type) {
    case FTConstant:
    case FTConnexion:
    case FTDirection:
      break;
    case FTNegation:
      tmp1 = comMaxLookupStringsInFormula(formula->data.formula);
      break;
    case FTConjunction:
    case FTDisjunction:
    case FTImplication:
    case FTBiimplication:
      tmp1 = comMaxLookupStringsInFormula(formula->data.junction.arg1);
      tmp2 = comMaxLookupStringsInFormula(formula->data.junction.arg2);
      break;
    case FTEqual:
    case FTNotEqual:
    case FTGreater:
    case FTLess:
    case FTGreaterEqual:
    case FTLessEqual:
      tmp1 = comMaxLookupStringsInTerm(formula->data.relation.op1);
      tmp2 = comMaxLookupStringsInTerm(formula->data.relation.op1);
      break;
    case FTPredicate:
      break;
  }

  return max(tmp1, tmp2);
}

/* ----------------------------------------------------------------------------
 * compute the number of strings needed for evaluationg 'lookup' functions in
 * a term 
 */
int comMaxLookupStringsInTerm(Term term)
{
  int tmp1 = 0;
  List l;

  if (term == NULL)
    return 0;

  switch (term->type) {
    case TTTopPeek:
    case TTBottomPeek: 
    case TTLabel:
    case TTLevel:
    case TTString:
    case TTNumber:
    case TTAdd: 
    case TTSubtract:   
    case TTMultiply:    
    case TTDivide:   
    case TTFunction:
      if (strcmp(term->data.function.functor, "lookup") == 0) {
	for (l = term->data.function.args->next; l != NULL; l = l->next) {
	  tmp1++;
	}
      }
  }

  return tmp1;
}

/* ----------------------------------------------------------------------------
 * compute the number of variables needed to evaluate a constraint
 */
int comConstraintDepth(Constraint constraint)
{
  int depth1 = comTermDepth(constraint->penaltyTerm);
  int depth2 = comFormulaDepth(constraint->formula);

  return max(depth1, depth2);
}

/* ----------------------------------------------------------------------------
 * compute the number of variables needed to evaluate a formula
 */
int comFormulaDepth(Formula formula)
{
  int maxDepth = 0, depth;

  switch (formula->type) {
    case FTConstant:
    case FTConnexion:
    case FTDirection:
      break;
    case FTNegation:
      maxDepth = comFormulaDepth(formula->data.formula);
      break;
    case FTConjunction:
    case FTDisjunction:
    case FTImplication:
      maxDepth = comFormulaDepth(formula->data.junction.arg1);
      depth = comFormulaDepth(formula->data.junction.arg2);
      maxDepth = max(maxDepth, depth);
      break;
    case FTBiimplication:
      maxDepth = 
	comFormulaDepth(formula->data.junction.arg1) +
	comFormulaDepth(formula->data.junction.arg2);
      break;
    case FTEqual:
    case FTNotEqual:
    case FTGreater:
    case FTLess:
    case FTGreaterEqual:
    case FTLessEqual:
      maxDepth = comTermDepth(formula->data.relation.op1);
      depth = comTermDepth(formula->data.relation.op2)+1;
      maxDepth = max(maxDepth, depth);
      break;
    case FTPredicate:
      maxDepth = comPredicateDepth(formula->data.predicate.functor, 
	  formula->data.predicate.args);
      break;
  }

  return maxDepth;
}

/* ----------------------------------------------------------------------------
 * compute the number of variables needed to evaluate a function
 */
int comFunctionDepth(String functor, List args)
{
  int maxDepth = 0;
  int depth;
  Term term;
  List l;

  /* distance */
  if (strcmp("distance", functor) == 0) {
    term = (Term)listNthElement(args, 1);
    maxDepth = comTermDepth(term);
    term = (Term)listNthElement(args, 2);
    depth = comTermDepth(term)+1;
    maxDepth = max(maxDepth, depth);
  } 
  
  /* abs */
  else if (strcmp("abs", functor) == 0) {
    term = (Term)listNthElement(args, 1);
    maxDepth = comTermDepth(term);
  } 
  
  /* min, max */
  else if (strcmp("min", functor) == 0 ||
           strcmp("max", functor) == 0) {

    term = (Term)args->item;
    maxDepth = comTermDepth(term);
    for (l = args->next; l != NULL; l = l->next) {
      term = (Term)l->item;
      depth = comTermDepth(term)+1;
      maxDepth = max(maxDepth, depth);
    }
  } 
  
  /* match */
  else if (strcmp("match", functor) == 0) {
    term = (Term)listNthElement(args, 1);
    maxDepth = comTermDepth(term);
    
    term = (Term)listNthElement(args, 2);
    depth = comTermDepth(term);
    maxDepth = max(maxDepth, depth);
    
    term = (Term)listNthElement(args, 3);
    depth = comTermDepth(term)+1;
    maxDepth = max(maxDepth, depth);
  } 

  /* parent */
  else if (strcmp("parent", functor) == 0) {
    term = (Term)listNthElement(args, 1);
    maxDepth = comTermDepth(term) + 1;
  }

  return maxDepth;
}

/* ----------------------------------------------------------------------------
 * compute the number of variables needed to evaluate a predicate
 */
int comPredicateDepth(String functor, List args)
{
  int maxDepth = 0;
  int depth;
  Term term;
  List l;

  /* start, stop, root, spec, nonspec */
  if (strcmp("start", functor) == 0 ||
      strcmp("stop", functor) == 0 ||
      strcmp("spec", functor) == 0 ||
      strcmp("nonspec", functor) == 0 ||
      strcmp("nonspec", functor) == 0 ||
      strcmp("root", functor) == 0) {
    term = (Term)args->item;
    maxDepth = comTermDepth(term);
  } 

  /* print */
  else if (strcmp("print", functor) == 0) {
    for (l = args; l != NULL; l = l->next) {
      term = (Term)l->item;
      depth = comTermDepth(term);
      maxDepth = max(maxDepth, depth);
    }
  } 

  /* exists */
  else if (strcmp("exists", functor) == 0) {
    /* no variables needed */
  } 

  /* subsumes */
  else if (strcmp("subsumes", functor) == 0 ||
           strcmp("compatible", functor) == 0) {

    term = (Term)listNthElement(args, 1);
    depth = comTermDepth(term);
    maxDepth = max(maxDepth, depth);

    term = (Term)listNthElement(args, 2);
    depth = comTermDepth(term);
    maxDepth = max(maxDepth, depth);

    term = (Term)listNthElement(args, 3);
    depth = comTermDepth(term)+1;
    maxDepth = max(maxDepth, depth);
  } 

  return maxDepth;
}

/* ----------------------------------------------------------------------------
 * compute the number of variables needed to evaluate a term
 */
int comTermDepth(Term term)
{
  int maxDepth = 0;
  int depth;

  if (term == NULL)
    return 0;

  switch (term->type) {
    case TTTopPeek:
    case TTBottomPeek: 
      {
	String pathItem = (String) term->data.peek.path->item; 
	if (strcmp("id", pathItem) == 0) {
	  maxDepth = 0;
	} else {
	  maxDepth = 1;
	}
      }
      break;
    case TTLabel:
    case TTLevel:
    case TTString:
    case TTNumber:
      maxDepth = 0;
      break;
    case TTAdd: 
    case TTSubtract:   
    case TTMultiply:    
    case TTDivide:   
      maxDepth = comTermDepth(term->data.operation.op1);
      depth = comTermDepth(term->data.operation.op2)+1;
      maxDepth = max(maxDepth, depth);
      break;
    case TTFunction:
      maxDepth = comFunctionDepth(term->data.function.functor,
	  term->data.function.args);
      break;
  }

  return maxDepth;
}

/* ----------------------------------------------------------------------------
 * comCompile: entry function to this package.
 *	- translate the currently loaded cdg into C
 *	- make a shared object out of it
 *	- load it
 *
 *	returns the name of the so file, NULL on failure
 */
String comCompile(int no, char **args)
{
  Boolean isOk = TRUE;
  int i;
  String outfileName = NULL, sofilename = NULL;

  /* construct a compiler instance */
  com = comNew();

  /* parse the commandline */
  for (i = 0; i < no && args[i] != NULL; i++) {
    if (args[i][0] == '-' && strlen(args[i]) > 1) {
      switch (args[i][1]) {
	case 'o':		       /* outfile */
	  outfileName = args[++i];
	  break;
	case 'c':		       /* compile only flag */
	  com->translateOnly = TRUE;
	  break;
	case 'h':		       /* help */
	case '?':
	  cdgPrintf(CDG_DEFAULT,
		    "\n"
		    "syntax: compile [-o <filename>] [-c] [-h]\n"
		    "        -o <filename> : define an outputfile\n"
		    "        -c            : translate only, dont compile + dont load \n"
		    "        -h            : this message\n\n"
		    "settings:\n"
		    "      CC : %s\n"
		    "      LD : %s\n"
		    "  CFLAGS : %s\n"
		    " LDFLAGS : %s\n"
		    "  LDLIBS : %s\n"
		    "INCLUDES : %s\n\n",
		    makeInfo->cc,
		    makeInfo->ld,
		    makeInfo->cFlags,
		    makeInfo->ldFlags,
		    makeInfo->ldLibs,
		    makeInfo->includes);
	  isOk = FALSE;
	  break;
      }
    }
    if (!isOk) {
      break;
    }
  }

  /* check parameters */
  if (isOk && listSize(inputCurrentGrammar->files) == 0) {
    cdgPrintf(CDG_ERROR, "ERROR: no files loaded\n");
    isOk = FALSE;
  }
  /* if no filename was specified then generate a default one */
  if (isOk && outfileName == NULL) {
    outfileName = basename((String) inputCurrentGrammar->files->item);
  }

  /* construct the needed filenames */
  if (isOk) {
    String ext = strrchr(outfileName, '.');

    if (ext != NULL) {
      ext[0] = 0;
    }
    com->ccFileName = strAppend(outfileName, ".c", NULL);
    com->objFileName = strAppend(outfileName, ".o", NULL);
    /* the linker requires an absolut path */
    com->soFileName = strAppend(dirname((String) inputCurrentGrammar->files->item), 
	"/", basename((String) inputCurrentGrammar->files->item), ".so", NULL);
    com->ccFile = fopen(com->ccFileName, "w");
    if (com->ccFile == NULL) {
      cdgPrintf(CDG_ERROR, "ERROR: cannot open file `%s'\n", com->ccFileName);
      isOk = FALSE;
    }
  }

  /* translate, compile and load */
  isOk = isOk && comTranslate();
  if (isOk && com->ccFile)
    fclose(com->ccFile);
  if (isOk && (!com->translateOnly && comMake())) {
    sofilename = strRegister(com->soFileName);
  } else {
    sofilename = NULL;
  }

  /* finalize the compiler */
  comFree(com);
  return sofilename;
}

/* ----------------------------------------------------------------------------
 * initialize the compiler module 
 */
void comInitialize(void)
{
  /* build a default makeInfo */
  makeInfo = comMakeInfoNew();

  /* these are some default values that should work for most of us */
  makeInfo->cc = strRegister("gcc");
  makeInfo->ld = strRegister("gcc");
  makeInfo->cFlags = strRegister(CFLAGS);
  makeInfo->ldFlags = strPrintf("-shared " LDFLAGS);
  makeInfo->ldLibs = strRegister(LDLIBS);
  makeInfo->includes = strPrintf(INCLUDES);

  /* register make variables */
  setRegister("CC", SET_STRING, &makeInfo->cc, NULL, NULL, NULL, NULL);
  setRegister("LD", SET_STRING, &makeInfo->ld, NULL, NULL, NULL, NULL);
  setRegister("CFLAGS", SET_STRING, &makeInfo->cFlags, NULL, NULL, NULL, NULL);
  setRegister("LDFLAGS", SET_STRING, &makeInfo->ldFlags, NULL, NULL, NULL, NULL);
  setRegister("LDLIBS", SET_STRING, &makeInfo->ldLibs, NULL, NULL, NULL, NULL);
  setRegister("INCLUDES", SET_STRING, &makeInfo->includes, NULL, NULL, NULL, NULL);
}

/* ----------------------------------------------------------------------------
 * initialize a dll grammar
 */
Boolean comInitGrammar(Input thisInput)
{
  Boolean (*thisInitGrammar)();

  if (thisInput->dllHandle == NULL)
    return TRUE;

  /* get the init function of the dll grammar */
  thisInitGrammar = (Boolean (*)())dlsym(thisInput->dllHandle, INIT_GRAMMAR);
  if (!thisInitGrammar) {
    cdgPrintf(CDG_ERROR, "ERROR: can't find function `%s': %s\n",
	      INIT_GRAMMAR, dlerror());
    return FALSE;
  }

  /* initialize the dll grammar */
  return (*thisInitGrammar) (thisInput);
}

/* ----------------------------------------------------------------------------
 * finalize a dll grammar
 */
void comFinitGrammar(Input thisInput)
{
  void (*thisFinitGrammar)(Input);

  if (thisInput->dllHandle == NULL)
    return;

  thisFinitGrammar = (void (*)(Input))dlsym(thisInput->dllHandle, FINIT_GRAMMAR);

  if (!thisFinitGrammar) {
    cdgPrintf(CDG_ERROR, "ERROR: can't find function `%s': `%s'\n",
	FINIT_GRAMMAR, dlerror());
  } else {
    (*thisFinitGrammar)(thisInput);
  }
  dlclose(thisInput->dllHandle);
  thisInput->dllHandle = NULL;
}

/* ----------------------------------------------------------------------------
 * finalize the compiler module 
 */
void comFinalize(void)
{
  comMakeInfoFree(makeInfo);
}
/* ----------------------------------------------------------------------------
 * allocate a new approver
 */
Approver comNewApprover(void)
{
  Approver ap = (Approver)memMalloc(sizeof(ApproverStruct));

  ap->noRounds = 1;
  ap->timer = timerNew();
  ap->totalInterpretedUnaryTime = 0;
  ap->totalCompiledUnaryTime = 0;
  ap->totalInterpretedBinaryTime = 0;
  ap->totalCompiledBinaryTime = 0;

  return ap;
}

/* ----------------------------------------------------------------------------
 * free an approver
 */
void comFreeApprover(Approver ap)
{
  if (ap == NULL)
    return;
  
  timerFree(ap->timer);
  memFree(ap);
}

/* ----------------------------------------------------------------------------
 * find a lv using the same dependent lexemes, label and regent lexemes
 */
LevelValue comFindComparableLv(ConstraintNet net, LevelValue lv)
{
  LevelValue testLv;
  int i;

  for (i = 0; i < vectorSize(net->values);  i++) {
    testLv = (LevelValue)vectorElement(net->values, i);
    if ((
	  (lv->regents == NULL && testLv->regents == NULL) ||
	  (lv->regents == NONSPEC && testLv->regents == NONSPEC) ||
	  (
	    lv->regents != NULL && testLv->regents != NULL &&
	    lv->regents != NONSPEC && testLv->regents != NONSPEC &&
	    ((LexemNode)(lv->regents->item))->arc->from == ((LexemNode)(testLv->regents->item))->arc->from &&
	    ((LexemNode)(lv->regents->item))->arc->to == ((LexemNode)(testLv->regents->item))->arc->to &&
	    ((LexemNode)(lv->regents->item))->lexem->description == ((LexemNode)(testLv->regents->item))->lexem->description
	  )
	) &&
	((LexemNode)(lv->dependents->item))->arc->from == ((LexemNode)(testLv->dependents->item))->arc->from &&
	((LexemNode)(lv->dependents->item))->arc->to == ((LexemNode)(testLv->dependents->item))->arc->to &&
	  ((LexemNode)(lv->dependents->item))->lexem->description == ((LexemNode)(testLv->dependents->item))->lexem->description &&
	strcmp(lv->label, testLv->label) == 0 &&
	lv->level == testLv->level) {
      return testLv;
    }
  }

  return NULL;
}

/* ----------------------------------------------------------------------------
 * compare two levelvalues 
 * returns TRUE if equal, else FALSE
 */
Boolean comCompareLvs(LevelValue lv1, LevelValue lv2)
{
  Boolean isEqual = TRUE;
  List l;
  Constraint c1;

  if (lv2 == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: value %d not found\n", lv1->indexWRTNet);
    return FALSE;
  }
  
  /* compare position in net */
  if (lv1->indexWRTNet != lv2->indexWRTNet) {
    cdgPrintf(CDG_ERROR, "WARNING: value is at position %d, should be at %d\n",
	lv2->indexWRTNet, lv1->indexWRTNet);
  }
  
  /* compare dependent */
  if (((LexemNode)(lv1->dependents->item))->lexem->description != ((LexemNode)(lv2->dependents->item))->lexem->description) {
    cdgPrintf(CDG_ERROR, "ERROR: value %d uses a different lexem `%s' in dependent position, should be `%s'\n",
	lv2->indexWRTNet, ((LexemNode)(lv2->dependents->item))->lexem->description,
	((LexemNode)(lv1->dependents->item))->lexem->description);
    isEqual = FALSE;
  }

  /* compare regent */
  if (lv1->regents == NULL) {
    if (lv2->regents != NULL) {
      cdgPrintf(CDG_ERROR, "ERROR: value %d should be NIL\n", lv2->indexWRTNet);
      isEqual = FALSE;
    }
  } else if (lv1->regents == NONSPEC) {
    if (lv2->regents != NONSPEC) {
      cdgPrintf(CDG_ERROR, "ERROR: value %d should be NONSPEC\n", lv2->indexWRTNet);
      isEqual = FALSE;
    }
  }

  if (lv1->regents != NULL && lv1->regents != NONSPEC &&
      lv2->regents != NULL && lv2->regents != NONSPEC &&
      ((LexemNode)(lv1->regents->item))->lexem->description != ((LexemNode)(lv2->regents->item))->lexem->description) {
    cdgPrintf(CDG_ERROR, "ERROR: value %d uses a different lexem `%s' in regent position, should be `%s'\n",
	lv2->indexWRTNet, ((LexemNode)(lv2->regents->item))->lexem->description,
	((LexemNode)(lv1->regents->item))->lexem->description);
    isEqual = FALSE;
  }

  /* compare levels */
  if (lv1->level != lv2->level) {
    cdgPrintf(CDG_ERROR, "ERROR: value %d uses a different level `%s', should be `%s'\n",
	lv2->indexWRTNet, lv2->level->id, lv1->level->id);
    isEqual = FALSE;
  }

  /* compare labels */
  if (strcmp(lv1->label, lv2->label) != 0) {
    cdgPrintf(CDG_ERROR, "ERROR: value %d uses a different label `%s', should be `%s'\n",
	lv2->indexWRTNet, lv2->label, lv1->label);
    isEqual = FALSE;
  }

  /* compare score */
  if (fabs(lv1->score - lv2->score) > 1e-10) {
    cdgPrintf(CDG_ERROR, "ERROR: value %d has a different score %4.3e, should be %4.3e\n",
	lv2->indexWRTNet, lv2->score, lv1->score);
    isEqual = FALSE;

    cdgPrintf(CDG_ERROR, "       conflicts:\n", lv2->indexWRTNet);
    for (l = lv2->constraints; l != NULL; l = l->next) {
      c1 = (Constraint)l->item;
      cdgPrintf(CDG_ERROR, "       %s\n", c1->id);
    }
    cdgPrintf(CDG_ERROR, "\n");
    if (lv1->constraints == NULL) { 
      cdgPrintf(CDG_ERROR, "       should be none.\n");
    } else {
      cdgPrintf(CDG_ERROR, "       should be:\n");
      for (l = lv1->constraints; l != NULL; l = l->next) {
	c1 = (Constraint)l->item;
	cdgPrintf(CDG_ERROR, "       %s\n", c1->id);
      }
    }
  }

  if (!isEqual) {
    cdgPrintf(CDG_ERROR, "\nOrig:");
    lvPrint(CDG_ERROR, lv1, NULL, 0);
    cdgPrintf(CDG_ERROR, "\nTest:");
    lvPrint(CDG_ERROR, lv2, NULL, 0);
    cdgPrintf(CDG_ERROR, "\n");
#if 1
    cdgPrintf(CDG_ERROR, "Dependent %d:\n", ((LexemNode)(lv2->dependents->item))->no);
    printLexiconItem(CDG_ERROR, ((LexemNode)(lv2->dependents->item))->lexem);
    cdgPrintf(CDG_ERROR, "\n");
    if (lv2->regents != NULL && lv2->regents != NONSPEC) {
      cdgPrintf(CDG_ERROR, "Regent %d:\n", ((LexemNode)(lv2->regents->item))->no);
      printLexiconItem(CDG_ERROR, ((LexemNode)(lv2->regents->item))->lexem);
      cdgPrintf(CDG_ERROR, "\n");
    }
#endif
  }
  
  return isEqual;
}

/* ----------------------------------------------------------------------------
 * compare each levelvalue in both networks
 * returns TRUE if all are equal, else FALSE
 */
Boolean comCompareAllLvs(Approver ap, ConstraintNet net1, ConstraintNet net2)
{
  LevelValue lv1, lv2;
  int i;
  Boolean hasEqualValues = TRUE;

  /* compare all values, must be in same order */
  if (vectorSize(net1->values) > vectorSize(net2->values)) {
    for (i = 0; i < vectorSize(net1->values) && hasEqualValues && !cdgCtrlCTrapped; i++) {
      lv1 = (LevelValue)vectorElement(net1->values, i);
      lv2 = comFindComparableLv(net2, lv1);

      if (lv2 == NULL) {
	cdgPrintf(CDG_ERROR, "ERROR: lv %d not found in net\n", lv1->indexWRTNet);
	lvPrint(CDG_ERROR, lv1, NULL, 0);
	cdgPrintf(CDG_ERROR, "\n");
	hasEqualValues = FALSE;
      } else {
	hasEqualValues &= comCompareLvs(lv1, lv2);
      }
    }
  } else {
    for (i = 0; i < vectorSize(net2->values) && hasEqualValues && !cdgCtrlCTrapped; i++) {
      lv2 = (LevelValue)vectorElement(net2->values, i);
      lv1 = comFindComparableLv(net1, lv2);
      if (lv1 == NULL) {
	cdgPrintf(CDG_ERROR, "ERROR: lv %d not found in net\n", lv2->indexWRTNet);
	lvPrint(CDG_ERROR, lv2, NULL, 0);
	cdgPrintf(CDG_ERROR, "\n");
	hasEqualValues = FALSE;
      } else {
	hasEqualValues &= comCompareLvs(lv1, lv2);
      }
    }
  }

  return hasEqualValues;
}

/* ----------------------------------------------------------------------------
 * compare each pair of levelvalues in both networks
 */
Boolean comCompareAllLvPairs(Approver ap, ConstraintNet net)
{
  Boolean hasEqualValuePairs = TRUE;
  int i,j,n;
  List l;
  ConstraintViolation cv;
  LevelValue lv1, lv2;
  BadnessStruct bad1, bad2;
  List conflicts1 = NULL, conflicts2 = NULL;
  Number score1 = 1.0, score2 = 1.0;

  /* check all value pairs */
  for (i = 0; i < vectorSize(net->values) && hasEqualValuePairs && !cdgCtrlCTrapped; i++) {
    lv1 = (LevelValue)vectorRemoveElementAt(net->values, i);
    for (j = i+1; j < vectorSize(net->values) && hasEqualValuePairs && !cdgCtrlCTrapped; j++) {
      lv2 = (LevelValue)vectorElement(net->values, j);

      /* refresh cache */
      scDelete(net->cache);
      if (scUseCache) {
	net->cache = scNew(vectorSize(net->values));
      } else {
	net->cache = NULL;
      }

      /* interpreted grammar */
      evalEvaluationMethod = EMTInterpreted;
      for (n = ap->noRounds; n > 0; n--) {
	listForEachDelete(conflicts1, cvDelete);
	conflicts1 = NULL;
	bad1.no = 0;
	bad1.hard = 0;
	bad1.soft = 1.0;
	timerStart(ap->timer);
	score1 = evalBinary(lv1, lv2, net, NULL, FALSE, &bad1, &conflicts1);
	ap->totalInterpretedBinaryTime += timerElapsed(ap->timer);
      }

      /* refresh cache */
      scDelete(net->cache);
      if (scUseCache) {
	net->cache = scNew(vectorSize(net->values));
      } else {
	net->cache = NULL;
      }

      /* compiled grammar */
      evalEvaluationMethod = EMTCompiled;
      for (n = ap->noRounds; n > 0; n--) {
	listForEachDelete(conflicts2, cvDelete);
	conflicts2 = NULL;
	bad2.no = 0;
	bad2.hard = 0;
	bad2.soft = 1.0;
	timerStart(ap->timer);
	score2 = evalBinary(lv1, lv2, net, NULL, FALSE, &bad2, &conflicts2);
	ap->totalCompiledBinaryTime += timerElapsed(ap->timer);
      }

      /* compare score */
      if (fabs(score1 - score2) > 1e-10) {
	hasEqualValuePairs = FALSE;
	cdgPrintf(CDG_ERROR, "ERROR: score %4.3e differs, should be %4.3e\n",
	    score2, score1);
      }

      /* compare badness */
      if (bad1.no == bad2.no &&
	  bad1.hard == bad2.hard &&
	  fabs(bad1.soft - bad2.soft) < 1e-10) {
	if (fabs(score1 - score2) > 1e-10) {
	  cdgPrintf(CDG_ERROR, "ERROR: same badness with different score %4.3e, should be %4.3e\n",
	      score2, score1);
	  hasEqualValuePairs = FALSE;
	}
      } else {
	hasEqualValuePairs = FALSE;
	cdgPrintf(CDG_ERROR, "ERROR: badness (%d,%d,%4.3e) differs, ",
	    bad2.no, bad2.hard, bad2.soft);
	cdgPrintf(CDG_ERROR, "should be (%d,%d,%4.3e)\n",
	    bad1.no, bad1.hard, bad1.soft);
      }

      /* show comparrison */
      if (!hasEqualValuePairs) {
	List conflicts = listSort(conflicts1, cvCompare);
	listDelete(conflicts1);
	conflicts1 = conflicts;
	conflicts = listSort(conflicts2, cvCompare);
	listDelete(conflicts2);
	conflicts2 = conflicts;

	cdgPrintf(CDG_ERROR, "violations:\n");
	for (l = conflicts2; l != NULL; l = l->next) {
	  cv = (ConstraintViolation)l->item;
	  if (cv->constraint->sig->arity == 1) { 
	    cdgPrintf(CDG_ERROR, "   %s:%4.3e:%d\n",
		cv->constraint->id, cv->penalty,
		cv->lv1->indexWRTNet);
	  } else {
	    cdgPrintf(CDG_ERROR, "%s:%4.3e:%d:%d\n",
		cv->constraint->id, cv->penalty,
		cv->lv1->indexWRTNet, cv->lv2->indexWRTNet);
	  }
	}
	cdgPrintf(CDG_ERROR, "should be:\n");
	for (l = conflicts1; l != NULL; l = l->next) {
	  cv = (ConstraintViolation)l->item;
	  if (cv->constraint->sig->arity == 1) { 
	    cdgPrintf(CDG_ERROR, "   %s:%4.3e:%d\n",
		cv->constraint->id, cv->penalty,
		cv->lv1->indexWRTNet);
	  } else {
	    cdgPrintf(CDG_ERROR, "%s:%4.3e:%d:%d\n",
		cv->constraint->id, cv->penalty,
		cv->lv1->indexWRTNet, cv->lv2->indexWRTNet);
	  }
	}
      }
    
      /* more info */
      if (!hasEqualValuePairs) {
	cdgPrintf(CDG_ERROR, "\nlattice: %s\n", ((LexemNode)(lv1->dependents->item))->lexemgraph->lattice->id);
	cdgPrintf(CDG_ERROR, "%03d: ", lv1->indexWRTNet);
	lvPrint(CDG_ERROR, lv1, NULL, 0);
	cdgPrintf(CDG_ERROR, "\n%03d: ", lv2->indexWRTNet);
	lvPrint(CDG_ERROR, lv2, NULL, 0);
	cdgPrintf(CDG_ERROR, "\n");

#if 1
	cdgPrintf(CDG_ERROR, "\nModifier1:\n");
	printLexiconItem(CDG_ERROR, ((LexemNode)(lv1->dependents->item))->lexem);
	cdgPrintf(CDG_ERROR, "\n");
	cdgPrintf(CDG_ERROR, "Modifiee1:\n");
	if (lv1->regents == NULL) {
	  cdgPrintf(CDG_ERROR, "NIL\n");
	} else if (lv1->regents == NONSPEC) {
	  cdgPrintf(CDG_ERROR, "NONSPEC\n");
	} else {
	  printLexiconItem(CDG_ERROR, ((LexemNode)(lv1->regents->item))->lexem);
	  cdgPrintf(CDG_ERROR, "\n");
	}
	cdgPrintf(CDG_ERROR, "\nModifier2:\n");
	printLexiconItem(CDG_ERROR, ((LexemNode)(lv2->dependents->item))->lexem);
	cdgPrintf(CDG_ERROR, "\n");
	cdgPrintf(CDG_ERROR, "Modifiee2:\n");
	if (lv2->regents == NULL) {
	  cdgPrintf(CDG_ERROR, "NIL\n");
	} else if (lv2->regents == NONSPEC) {
	  cdgPrintf(CDG_ERROR, "NONSPEC\n");
	} else {
	  printLexiconItem(CDG_ERROR, ((LexemNode)(lv2->regents->item))->lexem);
	  cdgPrintf(CDG_ERROR, "\n");
	}
#endif
      }
    }
  }

  return hasEqualValuePairs;
}

/* ----------------------------------------------------------------------------
 * compare the nets after applying context_sensitive constraints
 */
Boolean comCompareWithContext(Approver ap, ConstraintNet net, List contextList)
{
  Boolean hasEqualValuePairs = TRUE;
  int i,j,n;
  List l;
  ConstraintViolation cv;
  LevelValue lv1, lv2;
  BadnessStruct bad1, bad2;
  List conflicts1 = NULL, conflicts2 = NULL;
  Number score1 = 1.0, score2 = 1.0;
  Vector context;

  while (contextList) {
    context = contextList->item;
    /* check all value pairs */
    for (i = 0; i < vectorSize(net->values) && hasEqualValuePairs && !cdgCtrlCTrapped; i++) {
      lv1 = (LevelValue)vectorRemoveElementAt(net->values, i);
      for (j = i+1; j < vectorSize(net->values) && hasEqualValuePairs && !cdgCtrlCTrapped; j++) {
	lv2 = (LevelValue)vectorElement(net->values, j);

	/* refresh cache */
	scDelete(net->cache);
	if (scUseCache) {
	  net->cache = scNew(vectorSize(net->values));
	} else {
	  net->cache = NULL;
	}

	/* interpreted grammar */
	evalEvaluationMethod = EMTInterpreted;
	for (n = ap->noRounds; n > 0; n--) {
	  listForEachDelete(conflicts1, cvDelete);
	  conflicts1 = NULL;
	  bad1.no = 0;
	  bad1.hard = 0;
	  bad1.soft = 1.0;
	  timerStart(ap->timer);
	  score1 = evalBinary(lv1, lv2, net, context, TRUE, &bad1, &conflicts1);
	  ap->totalInterpretedBinaryTime += timerElapsed(ap->timer);
	}

	/* refresh cache */
	scDelete(net->cache);
	if (scUseCache) {
	  net->cache = scNew(vectorSize(net->values));
	} else {
	  net->cache = NULL;
	}

	/* compiled grammar */
	evalEvaluationMethod = EMTCompiled;
	for (n = ap->noRounds; n > 0; n--) {
	  listForEachDelete(conflicts2, cvDelete);
	  conflicts2 = NULL;
	  bad2.no = 0;
	  bad2.hard = 0;
	  bad2.soft = 1.0;
	  timerStart(ap->timer);
	  score2 = evalBinary(lv1, lv2, net, context, TRUE, &bad2, &conflicts2);
	  ap->totalCompiledBinaryTime += timerElapsed(ap->timer);
	}

	/* compare score */
	if (fabs(score1 - score2) > 1e-10) {
	  hasEqualValuePairs = FALSE;
	  cdgPrintf(CDG_ERROR, "ERROR: score %4.3e differs, should be %4.3e\n",
	      score2, score1);
	}

	/* compare badness */
	if (bad1.no == bad2.no &&
	    bad1.hard == bad2.hard &&
	    fabs(bad1.soft - bad2.soft) < 1e-10) {
	  if (fabs(score1 - score2) > 1e-10) {
	    cdgPrintf(CDG_ERROR, "ERROR: same badness with different score %4.3e, should be %4.3e\n",
		score2, score1);
	    hasEqualValuePairs = FALSE;
	  }
	} else {
	  hasEqualValuePairs = FALSE;
	  cdgPrintf(CDG_ERROR, "ERROR: badness (%d,%d,%4.3e) differs, ",
	      bad2.no, bad2.hard, bad2.soft);
	  cdgPrintf(CDG_ERROR, "should be (%d,%d,%4.3e)\n",
	      bad1.no, bad1.hard, bad1.soft);
	}

	/* show comparrison */
	if (!hasEqualValuePairs) {
	  List conflicts = listSort(conflicts1, cvCompare);
	  listDelete(conflicts1);
	  conflicts1 = conflicts;
	  conflicts = listSort(conflicts2, cvCompare);
	  listDelete(conflicts2);
	  conflicts2 = conflicts;

	  cdgPrintf(CDG_ERROR, "violations:\n");
	  for (l = conflicts2; l != NULL; l = l->next) {
	    cv = (ConstraintViolation)l->item;
	    if (cv->constraint->sig->arity == 1) { 
	      cdgPrintf(CDG_ERROR, "   %s:%4.3e:%d\n",
		  cv->constraint->id, cv->penalty,
		  cv->lv1->indexWRTNet);
	    } else {
	      cdgPrintf(CDG_ERROR, "%s:%4.3e:%d:%d\n",
		  cv->constraint->id, cv->penalty,
		  cv->lv1->indexWRTNet, cv->lv2->indexWRTNet);
	    }
	  }
	  cdgPrintf(CDG_ERROR, "should be:\n");
	  for (l = conflicts1; l != NULL; l = l->next) {
	    cv = (ConstraintViolation)l->item;
	    if (cv->constraint->sig->arity == 1) { 
	      cdgPrintf(CDG_ERROR, "   %s:%4.3e:%d\n",
		  cv->constraint->id, cv->penalty,
		  cv->lv1->indexWRTNet);
	    } else {
	      cdgPrintf(CDG_ERROR, "%s:%4.3e:%d:%d\n",
		  cv->constraint->id, cv->penalty,
		  cv->lv1->indexWRTNet, cv->lv2->indexWRTNet);
	    }
	  }
	}
      
	/* more info */
	if (!hasEqualValuePairs) {
	  cdgPrintf(CDG_ERROR, "\nlattice: %s\n", ((LexemNode)(lv1->dependents->item))->lexemgraph->lattice->id);
	  cdgPrintf(CDG_ERROR, "%03d: ", lv1->indexWRTNet);
	  lvPrint(CDG_ERROR, lv1, NULL, 0);
	  cdgPrintf(CDG_ERROR, "\n%03d: ", lv2->indexWRTNet);
	  lvPrint(CDG_ERROR, lv2, NULL, 0);
	  cdgPrintf(CDG_ERROR, "\n");

#if 1
	  cdgPrintf(CDG_ERROR, "\nModifier1:\n");
	  printLexiconItem(CDG_ERROR, ((LexemNode)(lv1->dependents->item))->lexem);
	  cdgPrintf(CDG_ERROR, "\n");
	  cdgPrintf(CDG_ERROR, "Modifiee1:\n");
	  if (lv1->regents == NULL) {
	    cdgPrintf(CDG_ERROR, "NIL\n");
	  } else if (lv1->regents == NONSPEC) {
	    cdgPrintf(CDG_ERROR, "NONSPEC\n");
	  } else {
	    printLexiconItem(CDG_ERROR, ((LexemNode)(lv1->regents->item))->lexem);
	    cdgPrintf(CDG_ERROR, "\n");
	  }
	  cdgPrintf(CDG_ERROR, "\nModifier2:\n");
	  printLexiconItem(CDG_ERROR, ((LexemNode)(lv2->dependents->item))->lexem);
	  cdgPrintf(CDG_ERROR, "\n");
	  cdgPrintf(CDG_ERROR, "Modifiee2:\n");
	  if (lv2->regents == NULL) {
	    cdgPrintf(CDG_ERROR, "NIL\n");
	  } else if (lv2->regents == NONSPEC) {
	    cdgPrintf(CDG_ERROR, "NONSPEC\n");
	  } else {
	    printLexiconItem(CDG_ERROR, ((LexemNode)(lv2->regents->item))->lexem);
	    cdgPrintf(CDG_ERROR, "\n");
	  }
#endif
	}
      }
    }
    contextList = contextList->next;
  }

  return hasEqualValuePairs;
}

/* ----------------------------------------------------------------------------
 * compare two constraint networks
 */
Boolean comCompareNets(Approver ap, ConstraintNet net1, ConstraintNet net2)
{
  Boolean hasEqualNets = TRUE;
  int noConflicts1, noConflicts2;
  int i, j, maxNumberOfContexts = 20;
  LevelValue lv, lValue;
  List net1ContextList = NULL;
  Vector context;
  ConstraintNode node;

  /* compare number of values */
  if (vectorSize(net1->values) != vectorSize(net2->values)) {
    cdgPrintf(CDG_ERROR, "ERROR: net has %d lvs, should be %d lvs\n",
	vectorSize(net2->values), vectorSize(net1->values));
    hasEqualNets = FALSE;
  }
  
  /* compare no conflicts */
  noConflicts1 = 0;
  for (i = 0; i < vectorSize(net1->values); i++) {
    lv = (LevelValue)vectorElement(net1->values, i);
    noConflicts1 += listSize(lv->constraints);
  }
  noConflicts2 = 0;
  for (i = 0; i < vectorSize(net2->values); i++) {
    lv = (LevelValue)vectorElement(net2->values, i);
    noConflicts2 += listSize(lv->constraints);
  }
  if (noConflicts1 != noConflicts2) {
    cdgPrintf(CDG_ERROR, "ERROR: net has %d %s conflicts\n",
	abs(noConflicts1 - noConflicts2), 
	(noConflicts1 < noConflicts2)?"more":"fewer");
  }

  /* build a list of contexts for net1 */
  for (i = 0; i < maxNumberOfContexts; i++) {
    context = vectorNew(vectorSize(net1->nodes));
    vectorSetAllElements(context, NULL);
    for (j = 0; j < vectorSize(net1->nodes); j++) {
      node = vectorElement(net1->nodes, j);
      lValue = vectorElement(node->values, i % vectorSize(node->values));
      vectorSetElement(context, lValue, lvIndex(lValue));
    }
    net1ContextList = listAppendElement(net1ContextList, context);
  }
      
  hasEqualNets = comCompareAllLvs(ap, net1, net2) && 
    comCompareAllLvPairs(ap, net1) &&
    comCompareWithContext(ap, net1, net1ContextList) && 
    hasEqualNets;

  return hasEqualNets;
}

/* ----------------------------------------------------------------------------
 * approve a compiled grammar
 */
void comApprove(int no, char **args) 
{
  Approver ap;
  Lattice lattice;
  ConstraintNet net1, net2;
  int i = 0, noLattices;
  Boolean hasEqualNets = TRUE;
  Boolean isAtty = isatty(fileno(stdout));
  unsigned long oldVerbosity = hkVerbosity;
  EvalMethodType oldEvaluationMethod = evalEvaluationMethod;
  List l;
  
  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;
//  hkVerbosity &= ~(CDG_INFO | CDG_DEBUG | CDG_PROFILE);
  
  if (inputCurrentGrammar == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: no grammar loaded\n");
    return;
  }
  if (inputCurrentGrammar->dllHandle == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: no dll grammar loaded\n");
    return;
  }

  ap = comNewApprover();
  noLattices = listSize(inputCurrentGrammar->lattices);

  /* 
   * build constraint nets for each available lattice
   */
  cdgPrintf(CDG_DEFAULT, "\n");
  for(l = inputCurrentGrammar->lattices;
      l != NULL && hasEqualNets && !cdgCtrlCTrapped;
      l = l->next) {
    lattice = l->item;
    

    /* 
     * build networks and eval binary
     */

    /* interpreted grammar */
    evalEvaluationMethod = EMTInterpreted;
    timerStart(ap->timer);
    net1 = cnBuild(lattice, TRUE);
    ap->totalInterpretedUnaryTime += timerElapsed(ap->timer);
    if (!net1) {
      cdgPrintf(CDG_WARNING, "WARNING: cannot build net for lattice `%s'\n", lattice->id);
      continue;
    }

    /* compiled grammar */
    evalEvaluationMethod = EMTCompiled;
    timerStart(ap->timer);
    net2 = cnBuild(lattice, TRUE);
    ap->totalCompiledUnaryTime += timerElapsed(ap->timer);
    if (!net2) {
      cdgPrintf(CDG_WARNING, "WARNING: cannot build net for lattice `%s'\n", lattice->id);
      cnDelete(net1);
      continue;
    }

    /* compare the two nets */
    hasEqualNets = comCompareNets(ap, net1, net2);

    cdgPrintf(CDG_DEFAULT, "%s(%03d/%03d) %s: %lums:%lums:%lums vs. %lums:%lums:%lums, speedup %4.2f:%4.2f:%4.2f%s", 
	isAtty?"\r":"",
	i+1, noLattices, lattice->id,
	ap->totalInterpretedUnaryTime, 
	ap->totalInterpretedBinaryTime, 
	ap->totalInterpretedUnaryTime+ap->totalInterpretedBinaryTime, 
	ap->totalCompiledUnaryTime, 
	ap->totalCompiledBinaryTime, 
	ap->totalCompiledUnaryTime+ap->totalCompiledBinaryTime, 
	ap->totalInterpretedUnaryTime / (0.0 + ap->totalCompiledUnaryTime),
	ap->totalInterpretedBinaryTime / (0.0 + ap->totalCompiledBinaryTime),
	(ap->totalInterpretedUnaryTime + ap->totalInterpretedBinaryTime) / 
	( 0.0 + ap->totalCompiledUnaryTime + ap->totalCompiledBinaryTime),
	isAtty?"":"\n");
    cdgFlush();

    /* clean up in each loop */
    /*
    cnDelete(net1);
    cnDelete(net2);
    */

  }
  
  cdgPrintf(CDG_DEFAULT, "\n");
  comFreeApprover(ap);

  cdgCtrlCTrapped = cdgCtrlCAllowed = FALSE;
  hkVerbosity = oldVerbosity;
  evalEvaluationMethod = oldEvaluationMethod;
}
#if 0
/* -------------------------------------------------------------------------
 * build a constraint net and evaluate a constraint on two levelvalues.
 * compare the results of this operation with a compiled vs. an evaluated
 * constraint
 * arguments are: 1. net, 2. lv1Index, 3. lv2Index, 4. constraintId
 */
void comAnalyzeConstraint(int no, char **args) {
  int constrain
  int lv1Index = 
  


  if (inputCurrentGrammar == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: no grammar loaded\n");
    return;
  }
  if (inputCurrentGrammar->dllHandle == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: no dll grammar loaded\n");
    return;
  }

  
  
  
  evalEvaluationMethod = EMTInterpreted;





  evalEvaluationMethod = EMTCompiled;

  
}
#endif


/* ------------------------------------------------------------------------- */
/* -- ENDOFFILE ------------------------------------------------------------ */
/** @} */
